/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ident "$Header: bp.c 1.5 90/03/13 $"

/*
 * bp.c
 *	Binary patch a COFF file or memory
 *
 *	Original author: Bob Beck
 *	Mods by Len Wyatt and Phil Hochstetler and Dave Stewart
 */

#include <stdio.h>
#include <a.out.h>
#include <sys/types.h>
#include <sys/fcntl.h>
#include <ldfcn.h>

#define	USAGE \
"usage: %s [-qdh] [-s symfile] filename symbol[+offset] [new_value]\n"
#define streq(s1, s2)   (strcmp(s1, s2) == 0)


int	quiet;
int	readonly;
int	numbase;
int	binpatch;
static	short sect_text;         /* text section number */
static	short sect_data;         /* data section number */
static	short sect_bss;          /* bss  section number */

char	*strchr();
char	*malloc();
char	*pvalue();
void	perror(), exit();
long	lseek();

/*
 * Print usage message.
 */
usage(prog)
	char	*prog;
{
	fprintf(stderr, USAGE, prog);
	exit(1);
}

main(argc, argv)
	int	argc;
	char	**argv;
{
	register char *argp;
	char	*prog = argv[0];
	char	*file;
	char	*symfile = NULL;
	char	*symbol;
	char	*plus;
	int	value = 0;
	int	offset = 0;

	while(--argc) {
		argp = *++argv;
		if (*argp == '-') while(*++argp) switch(*argp) {
		case 'q':
			++quiet;
			break;
		case 's':
			if (--argc <= 0)
				usage(prog);
			symfile = *++argv;
			break;
		case 'd':
			numbase = 10;
			break;
		case 'h':
			numbase = 16;
			break;
		default:
			usage(prog);
		} else
			break;
	}
	if (argc < 2 || argc > 3)
		usage(prog);

	file = argv[0];
	symbol = argv[1];
	if (plus = strchr(symbol, '+')) {
		offset = getnum(plus+1);
		*plus = '\0';
		if (plus == symbol)
			symbol = NULL;
	}
	if (argc == 3)
		value = getnum(argv[2]);
	else
		++readonly;
	exit(modfile(file, symfile, symbol, offset, value));
}

/*
 * modfile()
 *	Modify a file at given symbol offset.
 *
 * if symbol is NULL, use only offset.
 * If symfile non-null, use this for symbols.
 *
 * Returns appropriate exit status (0 ==> hunkey dorey).
 */

modfile(file, symfile, symbol, offset, value)
	char	*file;
	char	*symfile;
	char	*symbol;
{
	LDFILE	*ldptr = NULL;
	int	fd;
	int	symval = 0;
	int	start_off = 0;
	ushort	scnindex;
	long	pos;
	int	oldval;
	SCNHDR	secthead;

	if (symfile == NULL) {
		symfile = file;
	}
	/*
	 * Open symfile.
	 */
	if ((ldptr = ldopen(symfile, ldptr)) == NULL) {
		fprintf(stderr, "%s: cannot open\n", symfile);
		return(1);
	}

	fd = open(file, readonly ? O_RDONLY : O_RDWR);
	if (fd < 0) {
		perror(file);
		return(1);
	}


	/*
	 * If need a symbol, look it up.
	 * Use symfile if explicitly given, else the file itself.
	 */
	if (symbol != NULL) {
		if (strcmp(symfile, file) == 0)
			binpatch++;
		if (!getsymval(symfile, symbol, &symval, &scnindex)) {
			(void) ldclose(ldptr);
			return(1);
		}
		/*
		 * If patching a binary, must have a seek offset which
		 * accounts for the offset of the target section and
		 * for the address of the symbol.
		 */
		if (binpatch) {
			if (ldshread(ldptr, scnindex, &secthead) == FAILURE) {
				fprintf(stderr, "%s: couldn't read sec %d\n",
					symfile, scnindex);
				return(1);
			}
			start_off = secthead.s_scnptr - secthead.s_vaddr ;
		}
	}

	/*
	 * Seek in file to where we want to 
	 * patch, and write the new value.
	 */
	pos = start_off + symval + offset;
	if (pos < 0) {
		fprintf(stderr, "Offset out of range: 0x%x\n", pos);
		(void) close(fd);
		return(1);
	}
	(void) lseek(fd, (off_t)pos, 0);
	if (!quiet || readonly) {
		if (read(fd, (char *)&oldval, sizeof(oldval)) != sizeof(oldval)) {
			perror(file);
			(void) close(fd);
			return(1);
		}
		if (readonly)
			value = oldval;
		(void) lseek(fd, (off_t)pos, 0);
	}
	if (!readonly &&
	    write(fd, (char *)&value, sizeof(value)) != sizeof(value)) {
			perror(file);
			(void) close(fd);
			return(1);

	}
	if (!quiet) {
		printf("%s: ", file);
		if (symbol == NULL)
			printf("%s", pvalue(symval+offset));
		else if (offset == 0)
			printf("%s", symbol);
		else
			printf("%s+%s", symbol, pvalue(offset));
		if (!readonly)
			printf(" was %s", pvalue(oldval));
		printf(" is %s\n", pvalue(value));
	}
	(void) close(fd);
	return(0);
}

/*
 * pvalue()
 *	printf a number in numbase format
 *
 * Returns pointer to static area containing
 * the printed number
 */

char *
pvalue(n)
	int n;
{
	static char lbuf[256];
	
	switch (numbase) {
	case 10:
		(void) sprintf(lbuf, "%d", n);
		break;
	case 16:
	default:
		(void) sprintf(lbuf, "0x%x", n);
		break;
	}
	return (lbuf);
}

/*
 * getsymval()
 *	Given name of file containing symbol table and name of
 *	symbol, return it's value.
 *
 * Returns non-zero for sucess, else zero for failure.
 */

getsymval(symfile, symbol, symvalp, scnindexp)
	char	*symfile;
	char	*symbol;
	int	*symvalp;
	ushort	*scnindexp;
{
	LDFILE	*ldptr = NULL;
	long num_syms;		/* number of symbols in file */
	long symindex;		/* symbol index as we process them */
	SYMENT sym;		/* structure to hold current symbol */
	char *sname;		/* symbol name */
	char *ldgetname();

	if ((ldptr = ldopen(symfile, ldptr)) == NULL) {
		fprintf(stderr, "%s: could not open symbol file\n", symfile);
		return(0);
	}

	check_file_type(ldptr);
	scan_sections(ldptr);

	num_syms = HEADER(ldptr).f_nsyms;
	if (num_syms < 0) {
		fprintf(stderr, "%s: negative symbol count\n", symfile);
	}
	if (num_syms == 0) {
		printf("Warning: no symbols\n");
		return(0);
	}

	for (symindex = 0; symindex < num_syms; symindex++) {
		if (ldtbread(ldptr, symindex, &sym) != SUCCESS) {
			printf("Error reading symbol %d\n", symindex);
			return(0);
		}
		if ((sname = ldgetname(ldptr, &sym)) == NULL) {
			symindex += sym.n_numaux;
			continue;
		}

		/*
		 * skip the symbol if it isn't in a useful section, and
		 * if it isn't a storage class we are interested in.
		 */
		if ((sym.n_scnum == sect_text || sym.n_scnum == sect_data ||
		     sym.n_scnum == sect_bss) &&
		     (sym.n_sclass == C_EXT || sym.n_sclass == C_EXT) &&
		     strcmp(sname, symbol) == 0) {

			/*
			 * don't allow patching a binary's bss, since
			 * there really isn't any data in the file
			 * which represents it.
			 */
			if (binpatch && sym.n_scnum == sect_bss) {
				symindex += sym.n_numaux;
				continue;
			}

			*symvalp = sym.n_value;
			*scnindexp = sym.n_scnum;
			return(1);
		}
		symindex += sym.n_numaux;
	}
	fprintf(stderr, "No %s in %s.\n", symbol, symfile);
	return(0);
}


/*
 * getnum()
 *	Return an integer value converted from a string.
 *
 * Leading "0x" returns hex value, else decimal.
 */

getnum(s)
	register char *s;
{
	int	value;

	if (s[0] == '0' && s[1] == 'x') {
		if (numbase == 0)
			numbase = 16;
		(void) sscanf(&s[2], "%x", &value);
		return(value);
	}
	if (numbase == 0)
		numbase = 10;
	return(atoi(s));
}


/*
 * Quick check that file is in COFF format and that magic number
 * is correct.
 */

check_file_type(ldptr)
	LDFILE	*ldptr;
{
	if (ISCOFF(HEADER(ldptr).f_magic)) {
	    if (HEADER(ldptr).f_magic != I386MAGIC
	     && HEADER(ldptr).f_magic != SVI386MAGIC)
		printf("Warning: bad magic number.  Will try to process\n");
	} else {
	    printf("Error: File is not in COFF format.\n");
	    exit(1);
	}
}

/*
 * Scan the sections and get section numbers for text, data, and bss.
 */

scan_sections(ldptr)
	LDFILE	*ldptr;
{
	unsigned short sec,		/* current section */
		       nsec;		/* last section */
	SCNHDR	secthead;		/* section header structure */

	/* File last section number, give up if none. */

	nsec = HEADER(ldptr).f_nscns;
	if (nsec == 0) {
		printf("Error: section count is 0.\n");
		exit(1);
	}

	for (sec = 1; sec <= nsec; sec++) {
	    if (ldshread(ldptr, sec, &secthead) != SUCCESS) {
		printf("Error: ldshread fails for section %d\n", sec);
		exit(1);
	    }
	    if (streq(".text", secthead.s_name)){
		if (sect_text != 0) {
		    printf("Error: dupilcate text sections: %d, %d\n",
		      sect_text, sec);
		    exit(1);
		}
		sect_text = sec;
	    } else if (streq(".data", secthead.s_name)) {
		if (sect_data != 0) {
		    printf("Error: dupilcate data sections: %d, %d\n",
		      sect_data, sec);
		    exit(1);
		}
		sect_data = sec;
	    } else if (streq(".bss", secthead.s_name)) {
		if (sect_bss != 0) {
		    printf("Error: dupilcate bss sections: %d, %d\n",
		      sect_bss, sec);
		    exit(1);
		}
		sect_bss = sec;
	    }
	}

	if (sect_text == 0 || sect_data == 0 || sect_bss == 0) {
		fprintf(stderr, "Error: missed .text, .data, or .bss section\n");
		fprintf(stderr, "       .text = %d\n", sect_text);
		fprintf(stderr, "       .data = %d\n", sect_data);
		fprintf(stderr, "       .bss = %d\n", sect_bss);
		exit(0);
	}
}
