/*
 * Support routines for csh.
 *
 * These assist in port to System V / POSIX system.
 *
 *
 * $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 *
 */

#ident	"$Header: support.c 1.8 90/06/27 $"

#include <sys/types.h>
#include <sys/procstats.h>
#include <fcntl.h>
#include <string.h>

#ifdef DEBUG
static char *_blanks = "                    ";
static char *_zeroes = "00000000000000000000";
static char *uc_digs = "0123456789ABCDEF";
static char *lc_digs = "0123456789abcdef";
static char  *lc_nan = "nan0x";
static char  *uc_nan = "NAN0X";
static char  *lc_inf = "inf";
static char  *uc_inf = "INF";
#endif

/*
 * C version of bcopy that handles overlapped copies.
 *
 * bcopy(source, destination, count)
 *	char *source, *destination;
 *	unsigned int count;
 *
 * There are four cases to consider:
 *
 * 1. S < D no overlap		3. S < D with overlap**
 * 2. D < S no overlap		4. D < S with overlap
 *
 * ** - Case 3 must be done in the reverse direction.
 *
 */

void
bcopy(s, d, count)
	char *s, *d;
	int count;
{
	if (s == d)				/* no work to do */
		return;

	if (d < s) {				/* case 2 or 3 */
		while ((count--) > 0)
			*d++ = *s++;
		return;
	}

	if ((s + count) > d) {			/* case 3 */
		s += count - 1;
		d += count - 1;
		while ((count--) > 0)
			*d-- = *s--;
		return;
	}

	while ((count--) > 0)			/* case 4 */
		*d++ = *s++;
}

/*
 * killpg() is easy to simulate.  Use kill, with negative process number.
 */

killpg(pgrp, sig)
int pgrp, sig;
{
	return(kill(-(pgrp), sig));
}

/*
 * Compute the difference in process stats.
 */

subtract_process_stats(p2, p1, pr)
struct process_stats *p2, *p1, *pr;
{
	pr->ps_utime.tv_sec = p2->ps_utime.tv_sec - p1->ps_utime.tv_sec;
	pr->ps_utime.tv_usec = p2->ps_utime.tv_usec - p1->ps_utime.tv_usec;
	if (pr->ps_utime.tv_usec < 0) {
		pr->ps_utime.tv_sec -= 1;
		pr->ps_utime.tv_usec += 1000000;
	}
	pr->ps_stime.tv_sec = p2->ps_stime.tv_sec - p1->ps_stime.tv_sec;
	pr->ps_stime.tv_usec = p2->ps_stime.tv_usec - p1->ps_stime.tv_usec;
	if (pr->ps_stime.tv_usec < 0) {
		pr->ps_stime.tv_sec -= 1;
		pr->ps_stime.tv_usec += 1000000;
	}

	pr->ps_maxrss = p2->ps_maxrss - p1->ps_maxrss;
	pr->ps_pagein = p2->ps_pagein - p1->ps_pagein;
	pr->ps_reclaim = p2->ps_reclaim - p1->ps_reclaim;
	pr->ps_zerofill = p2->ps_zerofill - p1->ps_zerofill;
	pr->ps_pffincr = p2->ps_pffincr - p1->ps_pffincr;
	pr->ps_pffdecr = p2->ps_pffdecr - p1->ps_pffdecr;
	pr->ps_swap = p2->ps_swap - p1->ps_swap;
	pr->ps_syscall = p2->ps_syscall - p1->ps_syscall;
	pr->ps_volcsw = p2->ps_volcsw - p1->ps_volcsw;
	pr->ps_involcsw = p2->ps_involcsw - p1->ps_involcsw;
	pr->ps_signal = p2->ps_signal - p1->ps_signal;
	pr->ps_lread = p2->ps_lread - p1->ps_lread;
	pr->ps_lwrite = p2->ps_lwrite - p1->ps_lwrite;
	pr->ps_bread = p2->ps_bread - p1->ps_bread;
	pr->ps_bwrite = p2->ps_bwrite - p1->ps_bwrite;
	pr->ps_phread = p2->ps_phread - p1->ps_phread;
	pr->ps_phwrite = p2->ps_phwrite - p1->ps_phwrite;
}


/*
 * Support for signals.
 */


#include <stdio.h>
#include <signal.h>

typedef int (*sigtype)();

extern int errno;

/* Set and test a bit.  Bits numbered 1 to 32 */

#define SETBIT(x, y)	x |= (1 << ((y) -1) )
#define ISSET(x, y)	((x & (1 << ((y) - 1))) != 0)

#ifdef DEBUG

/*
 * dprintf()
 *
 * This will write on csh log file.
 */

#define CSH_LOG_FILE	"csh.log"

/*VARARGS*/
dprintf(format,arg1,arg2,arg3,arg4,arg5,arg6)
char *format;
int arg1,arg2,arg3,arg4,arg5,arg6;
{
	int fd;
	char dbuf1[1024];
	char dbuf2[1024];
	pid_t d_pid;
	int len;

	d_pid = getpid();
	d_sprintf(dbuf1, "%d: ", d_pid);
	d_sprintf(dbuf2, format, arg1,arg2,arg3,arg4,arg5,arg6);
	strcat(dbuf1, dbuf2);
	strcat(dbuf1, "\n");
	len = strlen(dbuf1);


	if ((fd = open(CSH_LOG_FILE, O_WRONLY|O_CREAT|O_APPEND, 0666)) < 0)
		return;
	write(fd, dbuf1, len);
	close(fd);
}

#endif

/* #define SHOW_SIGNALS	1	/* to assist in debugging signals */

#ifdef SHOW_SIGNALS
char *show_sig_mask();
#endif

int debug_signals = 0;


/*
 * igsetmask(mask)
 *
 * Set a new signal mask.  Return old mask.
 */

sigsetmask(mask)
	int mask;
{
	sigset_t set, oset;
	int m;
	register int i;

	sigemptyset(&set);
	sigemptyset(&oset);
	for (i=1; i <= MAXSIG; i++)
		if (ISSET(mask, i))
			sigaddset(&set, i);
	
#ifdef SHOW_SIGS
	if (debug_signals) {
		dprintf("+ sigsetmask(%s)", show_sig_mask(mask));
	}
#endif

	if (sigprocmask(SIG_SETMASK, &set, &oset))
		printf("sigsetmask(0x%x) - sigprocmask failed, errno %d",
			mask, errno);

	m = 0;
	for (i=1; i < MAXSIG; i++)
		if (sigismember(&oset, i))
			SETBIT(m, i);
	
	return(m);
}

/*
 * sigblock(mask)
 *
 * Add "mask" set of signals to the present signal mask.
 * Return old mask.
 */

sigblock(mask)
	int mask;
{
	sigset_t set, oset;
	int m;
	register int i;

	set = 0;				/* XXX */
	oset = 0;				/* XXX */

	/* Get present set of signals. */

	if (sigprocmask(SIG_SETMASK, NULL, &set))
		printf("sigblock(0x%x) - sigprocmask failed, errno %d",
			mask, errno);

	/* Add in signals from mask. */
	for (i=1; i <= MAXSIG; i++)
		if (ISSET(mask, i))
			sigaddset(&set, i);
	
#ifdef SHOW_SIGNALS
	if (debug_signals) {
		dprintf("+ sigblock(%s)\n", show_sig_mask(mask));
	}
#endif

	sigprocmask(SIG_SETMASK, &set, &oset);

	/* Return old mask to user. */

	m = 0;
	for (i=1; i < MAXSIG; i++)
		if (sigismember(&oset, i))
			SETBIT(m, i);
	
	return(m);
}


/*
 * bsd_sigpause(mask)
 *
 * Set new signal mask and wait for signal;
 * Old mask is restored on signal.
 */

bsd_sigpause(mask)
	int mask;
{
	sigset_t set;
	register int i;

	sigemptyset(&set);

	for (i=1; i <= MAXSIG; i++)
		if (ISSET(mask, i))
			sigaddset(&set, i);
	
#ifdef SHOW_SIGNALS
	if (debug_signals) {
		dprintf("+ bsd_sigpause(%s)\n", show_sig_mask(mask));
	}
#endif

	sigsuspend(&set);
}

/*
 * bsd_signal()
 *
 * Simulate the BSD style of signal() in which signals are reset
 * after they are caught.  They are also blocked during the signal call.
 */

sigtype
bsd_signal(sig, func)
	int sig;
	sigtype func;
{
	struct sigaction act, oact;
	sigset_t set;
	sigtype r_func;

	if (sig < 0 || sig > MAXSIG) {
		printf("error: bsd_signal(%d) signal out of range\n", sig);
		return;
	}

	sigemptyset(&set);

	act.sa_handler = (void(*)()) func;	/* user function */
	act.sa_mask = set;			/* signal mask */
	act.sa_flags = 0;			/* no special actions */

	if (sigaction(sig, &act, &oact)) {
		printf("error: bsd_signal(%d) - sigaction failed, errno %d\n",
		    sig, errno);
		return((sigtype) SIG_IGN);
	}

	r_func = (sigtype) oact.sa_handler;
	return(r_func);
	
}

/*
 * show_call(s)
 *	char *s
 *
 * Show signal call, for debugging.
 */

show_call(str)
char *str;
{
#ifdef SHOW_SIGNALS
	if (debug_signals) {
		dprintf("+ %s\n", str);
	}
#endif
}


#ifdef DEBUG

/*
 * char *
 * show_sig(sig)
 *
 * Display signal name.
 */

static char *sig_list[] = {
"none",
"SIGHUP",  "SIGINT",   "SIGQUIT",  "SIGILL",   "SIGTRAP",  /* 1, ...   5 */
"SIGIOT",  "SIGEMT",   "SIGFPE",   "SIGKILL",  "SIGBUS",   /* 6, ...  10 */
"SIGSEGV", "SIGSYS",   "SIGPIPE",  "SIGALRM",  "SIGTERM",  /* 11, ... 15 */
"SIGUSR1", "SIGUSR2",  "SIGCLD",   "SIGTPWR",  "SIGWIND",  /* 16, ... 20 */
"SIG21",   "SIGPOLL",  "SIG23",    "SIGCONT",  "SIGSTOP",  /* 21, ... 25 */
"SIGTSTP", "SIGTTIN",  "SIGTTOU",  "SIG29",    "SIG30",    /* 26, ... 30 */
"SIG31",   "SIG32"					   /* 31, ... 32 */
};

char *show_sig(sig)
	int sig;
{
	char buf[100];

	if (sig < 1 || sig > NSIG) {
		d_sprintf(buf, "BAD SIG #%d", sig);
		return(buf);
	}
	return(sig_list[sig]);
}

/*
 * char *
 * show_func(f)
 *
 * Display function address or defined const.
 */

char *show_func(f)
	sigtype f;
{
	char buf[100];

	if (f == (sigtype) SIG_DFL)
		return("SIG_DFL");
	else if (f == (sigtype) SIG_IGN)
		return("SIG_IGN");
	else {
		d_sprintf(buf, "0x%x", f);
		return(buf);
	}
}

/*
 * char *show_sig_mask(mask)
 *
 * Display a signal mask -- bit set = display that signal.
 */

char *
show_sig_mask(mask)
	int mask;
{
	int sig, bit;
	int count = 0;
	char buf[800];

	buf[0] = '\0';

	for (sig = 1; sig <= NSIG ; sig++) {
		if (ISSET(mask,sig)) {
			if (count != 0)
				strcat(buf, ", ");
			strcat(buf, sig_list[sig]);
			count++;
		}
	}

	if (count == 0)
		return("0");
	else
		return(buf);
}

/*
 * cur_sig_mask
 *
 * Return string showing current signal mask.
 */

char *
cur_sig_mask()
{
	sigset_t set;
	int m;
	register int i;

	/* Get present set of signals. */

	if (sigprocmask(SIG_SETMASK, NULL, &set)) {
		printf("cur_sig_mask() - sigprocmask failed, errno %d", errno);
		return("");
	}

	m = 0;
	for (i=1; i < MAXSIG; i++)
		if (sigismember(&set, i))
			SETBIT(m, i);
	return(show_sig_mask(m));
}


/*
 * This is support for local sprintf.  Because CSH defines
 * own copy must include this.  This is AT&T code.
 */


#define MAXDIGS 11

/* Maximum total number of digits in E format */
#if u3b || M32 || i386
#define MAXECVT 17
#else
#define MAXECVT 18
#endif

/* Maximum number of digits after decimal point in F format */
#define MAXFCVT 60

/* Maximum significant figures in a floating-point number */
#define MAXFSIG MAXECVT

/* Maximum number of characters in an exponent */
#if u3b || M32 || i386
#define MAXESIZ 5
#else
#define MAXESIZ 4
#endif

/* Maximum (positive) exponent */
#if u3b || M32 || i386
#define MAXEXP 310
#else
#define MAXEXP 40
#endif

/* Data type for flags */
typedef char bool;

/* Convert a digit character to the corresponding number */
#define tonumber(x) ((x)-'0')

/* Convert a number between 0 and 9 to the corresponding digit */
#define todigit(x) ((x)+'0')

/* Max and Min macros */
#define max(a,b) ((a) > (b)? (a): (b))
#define min(a,b) ((a) < (b)? (a): (b))


#define _ctype          (* _libc__ctype)
#define _numeric        (* _libc__numeric)
#define _cur_locale     (* _libc__cur_locale)

#define _getflthw	(* _libc__getflthw)

#define _cleanup (* _libc__cleanup)
#define environ  (* _libc_environ)		
#define end	 (* _libc_end)

#define malloc	(* _libc_malloc)		
#define free	(* _libc_free)	
#define realloc (* _libc_realloc)


#define _sibuf  (* _libc__sibuf)
#define _sobuf  (* _libc__sobuf)
#define _smbuf  (* _libc__smbuf)
#define _iob 	(* _libc__iob)
#define _lastbuf	(* _libc__lastbuf)
#define _bufendtab	(* _libc__bufendtab)


#include <stdio.h>
#include <varargs.h>
#include <values.h>
#include <ctype.h>
#include <nan.h>


static int x_doprnt();

/*VARARGS2*/
int
d_sprintf(string, format, va_alist)
char *string, *format;
va_dcl
{
	register int count;
	FILE siop;
	va_list ap;

	siop._cnt = MAXINT;
	siop._base = siop._ptr = (unsigned char *)string;
	siop._flag = _IOWRT|_IOSTRG;
	siop._file = -1;
	va_start(ap);
	count = x_doprnt(format, ap, &siop);
	va_end(ap);
	*siop._ptr = '\0'; /* plant terminating null character */
	return(count);
}


#define PUT(p, n)     { register unsigned char *newbufptr; \
	              if ((newbufptr = bufptr + n) > bufferend) { \
		              x_dowrite(p, n, iop, &bufptr); \
                      } else { \
		              (void) memcpy((char *) bufptr, p, n); \
			      bufptr = newbufptr; \
                      } \
		    } 
#define PAD(s,n)    { register int nn; \
		    for (nn = n; nn > 20; nn -= 20) \
			       x_dowrite(s, 20, iop, &bufptr); \
                    PUT(s, nn); \
                   }

#define SNLEN     5    	/* Length of string used when printing a NaN */

/* bit positions for flags used in doprnt */

#define LENGTH 	  1	/* l */
#define FPLUS     2	/* + */
#define FMINUS	  4	/* - */
#define FBLANK	  8	/* blank */
#define FSHARP	 16	/* # */
#define PADZERO  32	/* padding zeroes requested via '0' */
#define DOTSEEN  64	/* dot appeared in format specification */
#define SUFFIX	128	/* a suffix is to appear in the output */
#define RZERO	256	/* there will be trailing zeros in output */
#define LZERO	512	/* there will be leading zeroes in output */

/*
 *	C-Library routines for floating conversion
 */

extern char *fcvt(), *ecvt();

extern char _numeric[];

static int
x_lowdigit(valptr)
long *valptr;
{	/* This function computes the decimal low-order digit of the number */
	/* pointed to by valptr, and returns this digit after dividing   */
	/* *valptr by ten.  This function is called ONLY to compute the */
	/* low-order digit of a long whose high-order bit is set. */

	int lowbit = *valptr & 1;
	long value = (*valptr >> 1) & ~HIBITL;

	*valptr = value / 5;
	return(value % 5 * 2 + lowbit + '0');
}

/* The function _dowrite carries out buffer pointer bookkeeping surrounding */
/* a call to fwrite.  It is called only when the end of the file output */
/* buffer is approached or in other unusual situations. */
static
x_dowrite(p, n, iop, ptrptr)
register char *p;
register int	n;
register FILE	*iop;
register unsigned char **ptrptr;
{
	if (iop->_file != -1) {
		iop->_cnt -= (*ptrptr - iop->_ptr);
		iop->_ptr = *ptrptr;
		_bufsync(iop);
		(void) fwrite(p, 1, n, iop);
		*ptrptr = iop->_ptr;
	} else
		*ptrptr = (unsigned char *) memcpy((char *) *ptrptr, p, n) + n;
}

static
x_doprnt(format, args, iop)
register char	*format;
va_list	args;
register FILE	*iop;
{
                
	/* bufptr is used inside of doprnt instead of iop->_ptr; */
	/* bufferend is a copy of _bufend(iop), if it exists.  For */
	/* dummy file descriptors (iop->_file = -1), bufferend */
	/* may be meaningless.				*/

	unsigned char *bufptr, *bufferend;

	/* This variable counts output characters. */
	int	count = 0;

	/* Starting and ending points for value to be printed */
	register char	*bp;
	char *p;

	/* Field width and precision */
	int	width, prec;

	/* Format code */
	register int	fcode;

	/* Number of padding zeroes required on the left and right */
	int	lzero, rzero;

	/* Flags - bit positions defined by LENGTH, FPLUS, FMINUS, FBLANK, */
	/* and FSHARP are set if corresponding character is in format */
	/* Bit position defined by PADZERO means extra space in the field */
	/* should be padded with leading zeroes rather than with blanks */
	register int	flagword;

	/* Values are developed in this buffer */
	char	buf[max(MAXDIGS, 1+max(MAXFCVT+MAXEXP, MAXECVT))];

	/* Pointer to sign, "0x", "0X", or empty */
	char	*prefix;

	/* Exponent or empty */
	char	*suffix;

	/* Buffer to create exponent */
	char	expbuf[MAXESIZ + 1];

	/* Length of prefix and of suffix */
	int	prefixlength, suffixlength;

	/* Combined length of leading zeroes, trailing zeroes, and suffix */
	int 	otherlength;

	/* The value being converted, if integer */
	long	val;

	/* The value being converted, if real */
	double	dval;

	/* Output values from fcvt and ecvt */
	int	decpt, sign;

	/* Pointer to a translate table for digits of whatever radix */
	char	*tab;

	/* Work variables */
	int	k, lradix, mradix;

	/* Variables used to flag an infinities and nans, resp. */
	/* Nan_flg is used with two purposes: to flag a NaN and */
	/* as the length of the string ``NAN0X'' (``nan0x'') */
	 int	 Inf = 0, NaN_flg = 0 ;

	/* Pointer to string "NAN0X" or "nan0x" */
	 char	 *SNAN ;

        /* Flag for negative infinity */
	 int neg_inf = 0;

	/* initialize buffer pointer and buffer end pointer */
	{ register int fno = iop->_file;
	bufptr = iop->_ptr;
	bufferend = (fno == -1) ? 
			(unsigned char *)((long) bufptr | (-1L & ~HIBITL))
				 : _bufend(iop);
	}

	/*
	 *	The main loop -- this loop goes through one iteration
	 *	for each string of ordinary characters or format specification.
	 */
	for ( ; ; ) {
		register int n;

		if ((fcode = *format) != '\0' && fcode != '%') {
			bp = format;
			do {
				format++;
			} while ((fcode = *format) != '\0' && fcode != '%');
		
			count += (n = format - bp); /* n = no. of non-% chars */
			PUT(bp, n);
		}
		if (fcode == '\0') {  /* end of format; return */
			register int nn = bufptr - iop->_ptr;
			iop->_cnt -= nn;
			iop->_ptr = bufptr;
			if (bufptr + iop->_cnt > bufferend && /* in case of */
					iop->_file != -1) /* interrupt */
				_bufsync(iop);	/* during last several lines */
			if (iop->_flag & (_IONBF | _IOLBF) &&
				    (iop->_flag & _IONBF ||
				     memchr((char *)(bufptr-count), '\n', count) != NULL))
				(void) _xflsbuf(iop);
			return(ferror(iop) ? EOF : count);
		}

		/*
		 *	% has been found.
		 *	The following switch is used to parse the format
		 *	specification and to perform the operation specified
		 *	by the format letter.  The program repeatedly goes
		 *	back to this switch until the format letter is
		 *	encountered.
		 */
		width = prefixlength = otherlength = flagword = suffixlength = 0;
		format++;

	charswitch:

		switch (fcode = *format++) {

		case '+':
			flagword |= FPLUS;
			goto charswitch;
		case '-':
			flagword |= FMINUS;
			goto charswitch;
		case ' ':
			flagword |= FBLANK;
			goto charswitch;
		case '#':
			flagword |= FSHARP;
			goto charswitch;

		/* Scan the field width and precision */
		case '.':
			flagword |= DOTSEEN;
			prec = 0;
			goto charswitch;

		case '*':
			if (!(flagword & DOTSEEN)) {
				width = va_arg(args, int);
				if (width < 0) {
					width = -width;
					flagword ^= FMINUS;
				}
			} else {
				prec = va_arg(args, int);
				if (prec < 0)
					prec = 0;
			}
			goto charswitch;

		case '0':	/* obsolescent spec:  leading zero in width */
				/* means pad with leading zeros */
			if (!(flagword & (DOTSEEN | FMINUS)))
				flagword |= PADZERO;
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
		      { register num = fcode - '0';
			while (isdigit(fcode = *format)) {
				num = num * 10 + fcode - '0';
				format++;
			}
			if (flagword & DOTSEEN)
				prec = num;
			else
				width = num;
			goto charswitch;
		      }

		/* Scan the length modifier */
		case 'l':
			flagword |= LENGTH;
			/* No break */
		case 'h':
			goto charswitch; 

		/*
		 *	The character addressed by format must be
		 *	the format letter -- there is nothing
		 *	left for it to be.
		 *
		 *	The status of the +, -, #, and blank
		 *	flags are reflected in the variable
		 *	"flagword".  "width" and "prec" contain
		 *	numbers corresponding to the digit
		 *	strings before and after the decimal
		 *	point, respectively. If there was no
		 *	decimal point, then flagword & DOTSEEN
		 *	is false and the value of prec is meaningless.
		 *
		 *	The following switch cases set things up
		 *	for printing.  What ultimately gets
		 *	printed will be padding blanks, a
		 *	prefix, left padding zeroes, a value,
		 *	right padding zeroes, a suffix, and
		 *	more padding blanks.  Padding blanks
		 *	will not appear simultaneously on both
		 *	the left and the right.  Each case in
		 *	this switch will compute the value, and
		 *	leave in several variables the informa-
		 *	tion necessary to construct what is to
		 *	be printed.  
		 *
		 *	The prefix is a sign, a blank, "0x",
		 *	"0X", or null, and is addressed by
		 *	"prefix".
		 *
		 *	The suffix is either null or an
		 *	exponent, and is addressed by "suffix".
		 *	If there is a suffix, the flagword bit
		 *	SUFFIX will be set.
		 *
		 *	The value to be printed starts at "bp"
		 *	and continues up to and not including
		 *	"p".
		 *
		 *	"lzero" and "rzero" will contain the
		 *	number of padding zeroes required on
		 *	the left and right, respectively.
		 *	The flagword bits LZERO and RZERO tell
		 *	whether padding zeros are required.
		 *
		 *	The number of padding blanks, and
		 *	whether they go on the left or the
		 *	right, will be computed on exit from
		 *	the switch.
		 */



		
		/*
		 *	decimal fixed point representations
		 *
		 *	HIBITL is 100...000
		 *	binary, and is equal to	the maximum
		 *	negative number.
		 *	We assume a 2's complement machine
		 */

		case 'i':
		case 'd':
			/* Fetch the argument to be printed */
			if (flagword & LENGTH)
				val = va_arg(args, long);
			else
				val = va_arg(args, int);

			/* Set buffer pointer to last digit */
	                p = bp = buf + MAXDIGS;

			/* If signed conversion, make sign */
			if (val < 0) {
				prefix = "-";
				prefixlength = 1;
				/*
				 * Negate, checking in
				 * advance for possible
				 * overflow.
				 */
				if (val != HIBITL)
					val = -val;
				else     /* number is -HIBITL; convert last */
					 /* digit now and get positive number */
					*--bp = x_lowdigit(&val);
			} else if (flagword & FPLUS) {
				prefix = "+";
				prefixlength = 1;
			} else if (flagword & FBLANK) {
				prefix = " ";
				prefixlength = 1;
			}

		decimal:
			{ register long qval = val;
				if (qval <= 9) {
					if (qval != 0 || !(flagword & DOTSEEN))
						*--bp = qval + '0';
				} else {
					do {
						n = qval;
						qval /= 10;
						*--bp = n - qval * 10 + '0';
					} while (qval > 9);
					*--bp = qval + '0';
				}
			}

			/* Calculate minimum padding zero requirement */
			if (flagword & DOTSEEN) {
				register leadzeroes = prec - (p - bp);
				if (leadzeroes > 0) {
					otherlength = lzero = leadzeroes;
					flagword |= LZERO;
				}
			}

			break;

		case 'u':
			/* Fetch the argument to be printed */
			if (flagword & LENGTH)
				val = va_arg(args, long);
			else
				val = va_arg(args, unsigned);

			p = bp = buf + MAXDIGS;

			if (val & HIBITL)
				*--bp = x_lowdigit(&val);

			goto decimal;

		/*
		 *	non-decimal fixed point representations
		 *	for radix equal to a power of two
		 *
		 *	"mradix" is one less than the radix for the conversion.
		 *	"lradix" is one less than the base 2 log
		 *	of the radix for the conversion. Conversion is unsigned.
		 *	HIBITL is 100...000
		 *	binary, and is equal to	the maximum
		 *	negative number.
		 *	We assume a 2's complement machine
		 */

		case 'o':
			mradix = 7;
			lradix = 2;
			goto fixed;

		case 'X':
		case 'x':
			mradix = 15;
			lradix = 3;

		fixed:
			/* Fetch the argument to be printed */
			if (flagword & LENGTH)
				val = va_arg(args, long);
			else
				val = va_arg(args, unsigned);

			/* Set translate table for digits */
			  tab = (fcode == 'X') ? uc_digs : lc_digs;

			/* Entry point when printing a double which is a NaN */
		put_pc:
			/* Develop the digits of the value */
			p = bp = buf + MAXDIGS;
			{ register long qval = val;
				if (qval == 0) {
					if (!(flagword & DOTSEEN)) {
						otherlength = lzero = 1;
						flagword |= LZERO;
					}
				} else
					do {
						*--bp = tab[qval & mradix];
						qval = ((qval >> 1) & ~HIBITL)
								 >> lradix;
					} while (qval != 0);
			}

			/* Calculate minimum padding zero requirement */
			if (flagword & DOTSEEN) {
				register leadzeroes = prec - (p - bp);
				if (leadzeroes > 0) {
					otherlength = lzero = leadzeroes;
					flagword |= LZERO;
				}
			}

			/* Handle the # flag */
			if (flagword & FSHARP && val != 0)
				switch (fcode) {
				case 'o':
					if (!(flagword & LZERO)) {
						otherlength = lzero = 1;
						flagword |= LZERO;
					}
					break;
				case 'x':
					prefix = "0x";
					prefixlength = 2;
					break;
				case 'X':
					prefix = "0X";
					prefixlength = 2;
					break;
				}

			break;

		case 'E':
		case 'e':
			/*
			 * E-format.  The general strategy
			 * here is fairly easy: we take
			 * what ecvt gives us and re-format it.
			 */

			/* Establish default precision */
			if (!(flagword & DOTSEEN))
				prec = 6;

			/* Fetch the value */
	                dval = va_arg(args, double);

                        /* Check for NaNs and Infinities */
			if (IsNANorINF(dval))  {
			   if (IsINF(dval)) {
			      if (IsNegNAN(dval)) 
				neg_inf = 1;
			      Inf = 1;
			      bp = (fcode == 'E')? uc_inf: lc_inf;
			      p = bp + 3;
			      break;
                           }
			   else {
				if (IsNegNAN(dval)) {
					prefix = "-";
					prefixlength = 1;
				}
			        val  = GETNaNPC(dval); 
				NaN_flg = SNLEN;
				mradix = 15;
				lradix = 3;
				if (fcode == 'E') {
					SNAN = uc_nan;
					tab =  uc_digs;
				}
				else {
					SNAN =  lc_nan;
					tab =  lc_digs;
				}
				goto put_pc;
                           }
			}
			/* Develop the mantissa */
			bp = ecvt(dval, min(prec + 1, MAXECVT), &decpt, &sign);

			/* Determine the prefix */
		e_merge:
			if (sign) {
				prefix = "-";
				prefixlength = 1;
			} else if (flagword & FPLUS) {
				prefix = "+";
				prefixlength = 1;
			} else if (flagword & FBLANK) {
				prefix = " ";
				prefixlength = 1;
			}

			/* Place the first digit in the buffer*/
			p = &buf[0];
			*p++ = (*bp != '\0') ? *bp++ : '0';

			/* Put in a decimal point if needed */
			if (prec != 0 || (flagword & FSHARP))
				*p++ = _numeric[0];

			/* Create the rest of the mantissa */
			{ register rz = prec;
				for ( ; rz > 0 && *bp != '\0'; --rz)
					*p++ = *bp++;
				if (rz > 0) {
					otherlength = rzero = rz;
					flagword |= RZERO;
				}
			}

			bp = &buf[0];

			/* Create the exponent */
			*(suffix = &expbuf[MAXESIZ]) = '\0';
			if (dval != 0) {
				register int nn = decpt - 1;
				if (nn < 0)
				    nn = -nn;
				for ( ; nn > 9; nn /= 10)
					*--suffix = todigit(nn % 10);
				*--suffix = todigit(nn);
			}

			/* Prepend leading zeroes to the exponent */
			while (suffix > &expbuf[MAXESIZ - 2])
				*--suffix = '0';

			/* Put in the exponent sign */
			*--suffix = (decpt > 0 || dval == 0) ? '+' : '-';

			/* Put in the e */
			*--suffix = isupper(fcode) ? 'E'  : 'e';

			/* compute size of suffix */
			otherlength += (suffixlength = &expbuf[MAXESIZ]
								 - suffix);
			flagword |= SUFFIX;

			break;

		case 'f':
			/*
			 * F-format floating point.  This is a
			 * good deal less simple than E-format.
			 * The overall strategy will be to call
			 * fcvt, reformat its result into buf,
			 * and calculate how many trailing
			 * zeroes will be required.  There will
			 * never be any leading zeroes needed.
			 */

			/* Establish default precision */
			if (!(flagword & DOTSEEN))
				prec = 6;

			/* Fetch the value */
			dval = va_arg(args, double);

                        /* Check for NaNs and Infinities  */
			if (IsNANorINF(dval)) {
			   if (IsINF(dval)) {
			      if (IsNegNAN(dval))
				neg_inf = 1;
                              Inf = 1;
			      bp = lc_inf;
			      p = bp + 3;
			      break;
                           }
                           else {
				if (IsNegNAN(dval)) {
					prefix = "-";
					prefixlength = 1;
				}
			        val  = GETNaNPC(dval);
				NaN_flg = SNLEN;
				mradix = 15;
				lradix = 3;
				tab =  lc_digs;
				SNAN = lc_nan;
				goto put_pc;
                           }
                        } 
			/* Do the conversion */
			bp = fcvt(dval, min(prec, MAXFCVT), &decpt, &sign);

			/* Determine the prefix */
		f_merge:
			if (sign && decpt > -prec && *bp != '0') {
				prefix = "-";
				prefixlength = 1;
			} else if (flagword & FPLUS) {
				prefix = "+";
				prefixlength = 1;
			} else if (flagword & FBLANK) {
				prefix = " ";
				prefixlength = 1;
			}

			/* Initialize buffer pointer */
			p = &buf[0];

			{ register int nn = decpt;

				/* Emit the digits before the decimal point */
				k = 0;
				do {
					*p++ = (nn <= 0 || *bp == '\0' 
						|| k >= MAXFSIG) ?
				    		'0' : (k++, *bp++);
				} while (--nn > 0);

				/* Decide whether we need a decimal point */
				if ((flagword & FSHARP) || prec > 0)
					*p++ = _numeric[0];

				/* Digits (if any) after the decimal point */
				nn = min(prec, MAXFCVT);
				if (prec > nn) {
					flagword |= RZERO;
					otherlength = rzero = prec - nn;
				}
				while (--nn >= 0)
					*p++ = (++decpt <= 0 || *bp == '\0' ||
				   	    k >= MAXFSIG) ? '0' : (k++, *bp++);
			}

			bp = &buf[0];

			break;

		case 'G':
		case 'g':
			/*
			 * g-format.  We play around a bit
			 * and then jump into e or f, as needed.
			 */
		
			/* Establish default precision */
			if (!(flagword & DOTSEEN))
				prec = 6;
			else if (prec == 0)
				prec = 1;

			/* Fetch the value */
			dval = va_arg(args, double);

			/* Check for NaN and Infinities  */
			if (IsNANorINF(dval)) {
			   if (IsINF(dval)) {
			      if (IsNegNAN(dval)) 
				neg_inf = 1;
			   bp = (fcode == 'G') ? uc_inf : lc_inf;
			   p = bp + 3;
                           Inf = 1;
			   break;
                           }
                           else {
				if (IsNegNAN(dval)) {
					prefix = "-";
					prefixlength = 1;
				}
			        val  = GETNaNPC(dval);
				NaN_flg = SNLEN;
				mradix = 15;
				lradix = 3;
				if ( fcode == 'G') {
					SNAN = uc_nan;
					tab = uc_digs;
				}
				else {
					SNAN = lc_nan;
					tab =  lc_digs;
				}
				goto put_pc;
                           }
                        }

			/* Do the conversion */
			bp = ecvt(dval, min(prec, MAXECVT), &decpt, &sign);
			if (dval == 0)
				decpt = 1;

			{ register int kk = prec;
				if (!(flagword & FSHARP)) {
					n = strlen(bp);
					if (n < kk)
						kk = n;
					while (kk >= 1 && bp[kk-1] == '0')
						--kk;
				}
				
				if (decpt < -3 || decpt > prec) {
					prec = kk - 1;
					goto e_merge;
				}
				prec = kk - decpt;
				goto f_merge;
			}

		case '%':
			buf[0] = fcode;
			goto c_merge;

		case 'c':
			buf[0] = va_arg(args, int);
		c_merge:
			p = (bp = &buf[0]) + 1;
			break;

		case 's':
			bp = va_arg(args, char *);
			if (!(flagword & DOTSEEN))
				p = bp + strlen(bp);
			else { /* a strnlen function would  be useful here! */
				register char *qp = bp;
				while (*qp++ != '\0' && --prec >= 0)
					;
				p = qp - 1;
			}
			break;

		default: /* this is technically an error; what we do is to */
			/* back up the format pointer to the offending char */
			/* and continue with the format scan */
			format--;
			continue;

		}
       
		if (Inf) {
		   if (neg_inf) {
			prefix = "-";
			prefixlength = 1;
			neg_inf = 0;
                   }
		   else if (flagword & FPLUS) {
			prefix = "+";
			prefixlength = 1;
			}
		   else if (flagword & FBLANK) {
			prefix = " ";
			prefixlength = 1;
		   }
		   Inf = 0;
		}
		 
		/* Calculate number of padding blanks */
		k = (n = p - bp) + prefixlength + otherlength + NaN_flg;
		if (width <= k)
			count += k;
		else {
			count += width;

			/* Set up for padding zeroes if requested */
			/* Otherwise emit padding blanks unless output is */
			/* to be left-justified.  */

			if (flagword & PADZERO) {
				if (!(flagword & LZERO)) {
					flagword |= LZERO;
					lzero = width - k;
				}
				else
					lzero += width - k;
				k = width; /* cancel padding blanks */
			} else
				/* Blanks on left if required */
				if (!(flagword & FMINUS))
					PAD(_blanks, width - k);
		}

		/* Prefix, if any */
		if (prefixlength != 0)
			PUT(prefix, prefixlength);

		/* If value is NaN, put string NaN*/
		if (NaN_flg) {
			PUT(SNAN,SNLEN);
			NaN_flg = 0;
                }

		/* Zeroes on the left */
		if (flagword & LZERO)
			PAD(_zeroes, lzero);
		
		/* The value itself */
		if (n > 0)
			PUT(bp, n);

		if (flagword & (RZERO | SUFFIX | FMINUS)) {
			/* Zeroes on the right */
			if (flagword & RZERO)
				PAD(_zeroes, rzero);

			/* The suffix */
			if (flagword & SUFFIX)
				PUT(suffix, suffixlength);

			/* Blanks on the right if required */
			if (flagword & FMINUS && width > k)
				PAD(_blanks, width - k);
		}
	}
}


#endif /* DEBUG */
