/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ident	"$Header: devdestroy.c 1.3 90/03/05 $"

/*
 * devdestroy - Deconstruct complex devices into simpler ones.
 */

#include <sys/types.h>
#include <sys/devsw.h>
#include <sys/vtoc.h>
#include <sys/fcntl.h>

#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <constab.h>

#define USAGE "usage: devdestroy [-a]\n\tdevdestroy parent"
#define BUILDDIR	"/etc/devbuild.d"
#define DESTROYOPT	" -d "

#ifdef	lint
struct	streamtab {
	int dummy;
};
#endif	/* lint */

struct constab *ctab;
struct constab *last_entry;
FILE *consfp;
char destroyer[BUFSIZ];

struct constab *find_constab();

extern int   errno;
extern int   optind;
extern char *optarg;
extern int   lockf();
extern void  exit();
extern void  perror();

main(argc,argv)
int  argc;
char **argv;
{
	struct constab *cp;
	char	*dev;
	int	c;
	int	errflag = 0;
	int	all = 0;

	/*
	 *      check for proper arguments
	 */

	while ((c = getopt(argc, argv, "a")) != -1) {
		switch ((char)c) {
			case 'a':
				all++;
				break;
			case '?':
				errflag = 1;
				break;
		}
	}

	if (errflag) {
		(void)fprintf(stderr, "%s\n", USAGE);
		exit(2);
	}

	if ((consfp = fopen(CONSTAB, "a")) == NULL) {
		(void)fprintf(stderr,
			"devdestroy: warning - could not write on <%s>\n",
			CONSTAB);
		return(1);
	}
	rewind(consfp);
	if (lockf(fileno(consfp), F_LOCK, 0L) < 0) {
		(void)fprintf(stderr, "devdestroy: warning: cannot lock <%s>\n",
			CONSTAB);
	}

		
	if (!all) {
		if (argc != 2) {
			(void)fprintf(stderr, "%s\n", USAGE);
			exit(2);
		}

		dev = argv[1];

		if (read_constab(&ctab, &last_entry) != 0) {
			(void)fprintf(stderr, "devdestroy: unable to read %s\n",
								CONSTAB);
			exit(1);
		}

		cp = find_constab(dev, ctab, last_entry);

		if (cp == (struct constab *)0) {
			(void)fprintf(stderr, "devdestroy: unable to find %s in %s\n",
							dev, CONSTAB);
			exit(1);
		}

		if (dodestroy(cp->ct_devtype, dev) != 0) {
			exit(2);
		}

		if (remove_constab(dev, &ctab, &last_entry) == 0) {
			rewind(consfp);
			ftruncate(fileno(consfp), 0);
			(void)write_constab(ctab, last_entry, consfp);
		}
		exit(0);
	}

	if (all) {
		if (read_constab(&ctab, &last_entry) == 0) {
			cp = last_entry;

			/*
			 * destroy devices in reverse order to their
			 * order in devtab/constab.  Thus, any nested
			 * relationships can get broken down in the
			 * correct order.  Note that last_entry will
			 * move up one slot in ctab on remove_constab().
			 */
			while (cp != ctab) {
				dev = cp->ct_parent;
				if (dodestroy(cp->ct_devtype, dev) == 0) {
					(void)remove_constab(dev, &ctab,
								&last_entry);
					cp = last_entry;
				} else {
					cp--;
				}
			}

			/*
			 * Note that we skip the first entry in ctab, since
			 * this is supposed to be the root disk, which
			 * we can't destroy anyway.
			 */

			rewind(consfp);
			ftruncate(fileno(consfp), 0);
			exit(write_constab(ctab, last_entry, consfp));
		} else {
			exit(1);
		}
	}
	exit(0);

	/*NOTREACHED*/
}

/*
 * dodestroy
 *
 *	Destroy the virtual device which is of type devtype and whose
 * parent node is parent.  The program in BUILDDIR whose name matches
 * devtype is invoked with the DESTROYOPT flag and it handles the
 * actual destruction.
 */
dodestroy(devtype, parent)
	char *devtype, *parent;
{
	char *p1;

	(void)strcpy(destroyer, BUILDDIR);
	(void)strcat(destroyer, "/");
	p1 = destroyer + strlen(destroyer);
	(void)strcat(p1, devtype);

	if (access(destroyer, X_OK) < 0) {
		(void)fprintf(stderr,
			"devbuild: illegal virtual device type %s ",
			devtype);
		(void)fprintf(stderr, "(unable to access %s)\n", destroyer);
		return(1);
	}

	(void)strcat(destroyer, DESTROYOPT);
	(void)strcat(destroyer, parent);

	if (system(destroyer) != 0) {
		/*
		 * leave it up to subcommand to report error to stderr
		 */
		return(1);
	}

	return(0);
}

/*
 * remove_constab
 *
 *	Remove the constab entry from ct whose parent matches d.  A side effect
 * of removing an entry from ct is that le (pointer to the last_entry in
 * ct) is now one slot "closer" to ct.
 */
remove_constab(d, ct, le)
	char *d;
	struct constab **ct, **le;
{
	struct constab *cp, *next;
	int foundit = 0;

	for (cp = *ct; cp <= *le; cp++) {
		if (strcmp(d, cp->ct_parent) == 0) {
			foundit++;
			break;
		}
	}

	if (!foundit)
		return(-1);

	for (next = cp + 1; next <= *le; next++, cp++) {
		*cp = *next;
	}

	*le = cp - 1;
	return(0);
}

/*
 * write_constab
 *
 *	Write the constab which is pointed to by ct onto the stream
 * pointer fp.
 */
write_constab(ct, le, fp)
	struct constab *ct, *le;
	FILE *fp;
{
	char **kp;
	struct constab *cp;
	
	for (cp = ct; cp <= le; cp++) {
		(void)fprintf(fp, "%s %s", cp->ct_devtype, cp->ct_parent);
		if (cp->ct_kids) {
			for (kp = cp->ct_kids; kp && *kp; kp++){
				(void)fprintf(fp, " %s", *kp);
			}
		}
		if (cp->ct_hints) {
			for (kp = cp->ct_hints; kp && *kp; kp++){
				(void)fprintf(fp, " %s", *kp);
			}
		}
		(void)fprintf(fp, "\n");
	}
	return(0);
}

/*
 * find_constab
 *
 *	locate the constab entry whose ct_parent entry is known.  Return
 * 0 if parent is not in ctab.
 */
struct constab *
find_constab(parent, ct, le)
	char	*parent;
	struct	constab	*ct, *le;
{
	for ( ; ct <= le; ct++) {
		if (strcmp(ct->ct_parent, parent) == 0) {
			return(ct);
		}
	}
	return((struct constab *)0);
}
