/*
 * $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 *
 */

/*	Copyright (c) 1984 AT&T	*/
/*	  All Rights Reserved  	*/

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/

#ident	"$Header: df.c 1.19 90/06/21 $@(#)df:df.c	1.35"

/* df (with support for 512, 1KB and 2KB block sizes)
 *	Distributed Unix, FS Switching Version
 */

#include <stdio.h>
#include <sys/param.h>
#include <sys/types.h>
#include <mntent.h>
#include <limits.h>
#include <sys/stat.h>
#include <sys/statfs.h>
#include <sys/ufsfilsys.h>
#include <dirent.h>
#include <sys/errno.h>
#include <string.h>
#include <sys/sysmacros.h>

#define EQ(x,y)		(strcmp(x,y)==0)
#define EQN(x,y,z)	(strncmp(x,y,z)==0)
#define MNTTAB	MNTTABNAME
#define MOUNTED fs_name[0]
#define DEVLEN (PATH_MAX)
#define DEVLSZ 200	/* number of unique file systems we can remember */
#define FREMOTE	0x0002	/* file is remote (see <sys/user.h>)  */
#define PHYSBLKSZ 512	/* size of physical disk block */

#define MAX(a,b)	((a)>(b)?(a):(b))
#define N_BLOCKS (fsbtodb(sblock,sblock->fs_size)) /* # of filesystem blocks */
#define FREE_FRAGS (sblock->fs_cstotal.cs_nbfree * sblock->fs_frag + \
	sblock->fs_cstotal.cs_nffree)	/* number of free frags */
#define FREE_BLOCKS (fsbtodb(sblock,FREE_FRAGS))
	/* AVAIL_BLOCKS = MAX(max_avail - used, 0) */
#define AVAIL_BLOCKS ((fsbtodb(sblock, sblock->fs_dsize) * (100 - sblock->fs_minfree) / 100) - \
	(fsbtodb(sblock, sblock->fs_dsize) - FREE_BLOCKS))
#define N_INODES (sblock->fs_ncg * sblock->fs_ipg)	/* number of inodes */
#define FREE_INODES (sblock->fs_cstotal.cs_nifree)	/* total free inodes */

extern char *optarg;
extern int optind, opterr;
extern int errno;
char *devnm();
void nfs_getmntpnt();

struct	mntent	*mntp;
struct stat	S;
struct fs	*sblock;
struct statfs	Fs_info;

int	freflg, totflg, locflg, Superflg;
int	fd;
daddr_t	alloc();
char	*cwd, *getcwd();

char nfsmes[] = "          (%-10.32s): \
is not a file system, directory, or mounted resource\n";

char usage[] = "Usage: \
df [-ltS] [-f] [{<file system> <directory> <mounted resource>}] ...\n";

char *
basename(s)
char *s;
{
	int n = 0;

	while(*s++ != '\0') n++;
	while(n-- > 0)
		if(*(--s) == '/') 
			return(++s);
	return(--s);
}

main(argc, argv)
char **argv;
{
	char	*res_name, *s;
	int	 c, k, errcnt = 0;
	int	 in_mnttab;
	FILE	*mnttab;

	s = basename(argv[0]);
	if(!strncmp(s,"devnm",5)) {

		if((cwd = getcwd((char *)NULL, MAXPATHLEN)) == NULL) {
			fprintf(stderr,gettxt("df:27","devnm: out of memory\n"));
			exit(1);
		}
		while(--argc) {
			if(stat(*++argv, &S) == -1) {
				fprintf(stderr,
				    gettxt("df:28","devnm: can't stat %s\n"),
				    *argv);
				errcnt++;
				continue;
			}
			if (major(S.st_dev) == 255) {
				printf(gettxt("df:34","devnm: %s is on a remote filesystem\n"),
					*argv);
				errcnt++;
				continue;
			}
			res_name = devnm(S.st_dev);
			(void) chdir(cwd);
			if(res_name[0] != '\0') {
				printf("%s %s\n", res_name, *argv);
			} else {
				printf(gettxt("df:3","devnm: %s not found\n"),
				    *argv);
				errcnt++;
			}
		}
		exit(errcnt);
	}
	while((c = getopt(argc,argv,"tflS")) != EOF) {
		switch(c) {

			case 'l':	/* do local resources only */
				locflg = 1;
				break;

			case 'f':	/* freelist count only */
				freflg++;
				break;

			case 't':	/* include total allocations */
				totflg = 1;
				break;

			case 'S':	/* space available to Super-users */
				Superflg++;
				break;

			default:
				fprintf(stderr,gettxt("df:25",usage));
				exit(1);
		}
	}

	/* we are looking for specific file systems
	   in the mount table */
	if(argc > optind) {
		if((cwd = getcwd((char *)NULL, MAXPATHLEN)) == NULL) {
			fprintf(stderr,gettxt("df:33","df: out of memory\n"));
			exit(1);
		}
		for(k = optind; k < argc; k++) {
			if(stat(argv[k], &S) == -1) {
				errcnt++;
				fprintf(stderr,
				    gettxt("df:4","df: cannot open %s\n"),
				    argv[k]);
				continue;
			}
			if((S.st_mode & S_IFMT) == S_IFREG
			|| (S.st_mode & S_IFMT) == S_IFIFO) {
				printf(gettxt("df:26",nfsmes), argv[k]);
				continue;
			}
			/*
			 * If we are on a NFS filesys (major == 255) then
			 *	we have to find the mountpoint.
			 */
			if (major(S.st_dev) == 255) {
				nfs_getmntpnt(&argv[k]);
				res_name = (char *)0;
			} else {
				res_name = devnm(S.st_dev);
			}
			(void) chdir(cwd);
			in_mnttab = 0;

			if ((mnttab = setmntent(MNTTAB, "r")) == NULL) {
				fprintf(stderr,
				    gettxt("df:29","df: cannot open "));
				perror(MNTTAB);
				exit(2);
			}
			while ((mntp = getmntent(mnttab)) != NULL) {
				if (strcmp(mntp->mnt_type, MNTTYPE_IGNORE) == 0) {
					continue;
				}
				if(locflg && !EQ(mntp->mnt_type, MNTTYPE_UFS))
					continue;

				/* does the argument match either
				   the mountpoint or the device? */
				if(EQ(argv[k], mntp->mnt_dir)
				|| EQ(argv[k], mntp->mnt_fsname)) {
					errcnt += printit(mntp->mnt_fsname,
							  mntp->mnt_dir);
					in_mnttab++;
					break;

				/* or is it on one of
					the mounted device */
				} else
				if( (S.st_mode & S_IFMT) == S_IFDIR
				&& EQ(res_name, mntp->mnt_fsname) ) {
					errcnt += printit(res_name, argv[k]);
					in_mnttab++;
					break;
				}
			}
			endmntent(mnttab);

			/* process arguments that were
			   not in /etc/mnttab */
			if (! in_mnttab) {
				errcnt += printit(argv[k], "\0");
			}
		}
	} else {
			/* doing all things in the mount table */

		if ((mnttab = setmntent(MNTTAB, "r")) == NULL) {
			fprintf(stderr, gettxt("df:29","df: cannot open "));
			perror(MNTTAB);
			exit(2);
		}
		while ((mntp = getmntent(mnttab)) != NULL) {
			if (strcmp(mntp->mnt_type, MNTTYPE_IGNORE) == 0) {
				continue;
			}
			if(locflg && !EQ(mntp->mnt_type, MNTTYPE_UFS))
				continue;
			if(stat(mntp->mnt_dir, &S) == -1)
				S.st_dev = -1;
			errcnt += printit(mntp->mnt_fsname, mntp->mnt_dir);
		}
		endmntent(mnttab);
	}

	exit(errcnt);
}

printit(dev, fs_name)
char *dev, *fs_name;
{
	char mark;


	if(MOUNTED && EQ(mntp->mnt_type, MNTTYPE_NFS) && freflg)
	{
		if(statfs(fs_name, &Fs_info, sizeof(struct statfs), 0) < 0) {
			fprintf(stderr, "df: cannot statfs %s\t", fs_name);
			perror(" statfs");
			return(1);
		}
		printf("%-9s (%-16.32s):", fs_name, dev);
		printf(" %8ld blocks\n",
			Superflg ? Fs_info.f_bfree : MAX(Fs_info.f_bavail,0));
		close(fd);
		if(totflg) {
			printf("                   total: ");
			printf("%8ld blocks %7d i-nodes\n",
				Fs_info.f_blocks, Fs_info.f_files);
		}
		return(0);
	}
	if(MOUNTED && !EQ(mntp->mnt_type, MNTTYPE_UFS))
	{
		if(freflg)
			return(0);
	}

	if(!MOUNTED || freflg) {	/* Unmounted file system queries
					   and freelist querries work only
					   on UNIX file systems */
		if((fd = open(dev, 0)) < 0) {
			fprintf(stderr,
			    gettxt("df:4","df: cannot open %s\n"), dev);
			close(fd);
			return(1);
		}
		if(stat(dev, &S) < 0) {
			fprintf(stderr,
			    gettxt("df:5","df: cannot stat %s\n"), dev);
			close(fd);
			return(1);
		}

		if(((S.st_mode & S_IFMT) == S_IFREG) 
		|| ((S.st_mode & S_IFMT) == S_IFIFO)) {
			printf(gettxt("df:26",nfsmes), dev);
			close(fd);
			return(1);
		}

		/* must page align for raw filesystems and
		 * and roundup for zdc's
		 */
		if((sblock = (struct fs *)valloc(SBSIZE)) == NULL) {
			perror(gettxt("df:30","df: cannot valloc superblock"));
			close(fd);
			return(1);
		}

		sync();
		if(lseek(fd, (long)dbtob(SBLOCK), 0) < 0
		|| read(fd, sblock, SBSIZE) != SBSIZE) {
			fprintf(stderr,
			    gettxt("df:31","df: cannot read %s\n"),dev);
			close(fd);
			return(1);
		}	
		if(sblock->fs_magic != FS_MAGIC
		|| sblock->fs_bsize > MAXBSIZE) {
			printf(gettxt("df:26",nfsmes), dev);
			return(1);
		}
		
	} else {	/* mounted file system */

		if(statfs(fs_name, &Fs_info, sizeof(struct statfs), 0) < 0) {
			fprintf(stderr,
			    gettxt("df:7","df: cannot statfs %s\t"), fs_name);
			perror(gettxt("df:32"," statfs"));
			return(1);
		}
	}

	if(!freflg) {
		if(!MOUNTED) 
			printf("  %-6.6s ", sblock->fs_fname);
		else
			printf("%-9s", fs_name);
		printf(" (%-16.32s):", dev);

			/* mark 2nd and subsequent reports on the same
							remote resource.  */
		mark = (firstpik(S.st_dev)) ? ' ' : '*';
		if(!MOUNTED) {
			printf(gettxt("df:8"," %8ld blocks%c %7d i-nodes\n"),
				Superflg ? FREE_BLOCKS : MAX(AVAIL_BLOCKS, 0),
				mark, FREE_INODES);
		} else {
			printf(gettxt("df:8"," %8ld blocks%c %7d i-nodes\n"),
				Superflg ? Fs_info.f_bfree : Fs_info.f_bavail,
					mark, Fs_info.f_ffree);
		}
	} else {
		printf(gettxt("df:9","%-10s(%-12s): %8ld blocks\n"),
			fs_name, dev, Superflg ? FREE_BLOCKS : MAX(AVAIL_BLOCKS,0));
		close(fd);
	}
	if(totflg) {
		printf(gettxt("df:10","                   total: "));
		if(MOUNTED && !freflg)
			printf(gettxt("df:11","%8ld blocks %7d i-nodes\n"),
				Fs_info.f_blocks,
				Fs_info.f_files);
		else
			printf(gettxt("df:11","%8ld blocks %7d i-nodes\n"),
				N_BLOCKS, N_INODES);
	}
	return(0);
}

firstpik(st_dev)
dev_t st_dev;
{
	static dev_t devlst[DEVLSZ];
	static	int i, ndev;

	if(st_dev == -1)
		return(1);	/* stat() failed.. data invalid */
	if(ndev)
		for(i = 0; i < ndev; i++) 
			if(st_dev == devlst[i])
				return(0);
	if(ndev < DEVLSZ)
		devlst[ndev++] = st_dev;
	return(1);
}

struct dirent *dbuf;

char *
devnm(fno)
dev_t fno;
{
	int i;
	static struct devs {
		char *devdir;
		DIR *dfd;
	} devd[] = {		/* in order of desired search */
		"/dev/dsk",0,
		"/dev",0,
		"/dev/rdsk",0
	};
	static char devnam[DEVLEN+1];

	devnam[0] = '\0';
	if(!devd[1].dfd) {	/* if /dev isn't open, nothing happens */
		for(i = 0; i < 3; i++) {
			devd[i].dfd = opendir(devd[i].devdir);
		}
	}

	for(i = 0; i < 3; i++) {
		if((devd[i].dfd != 0) && (devd[i].dfd->dd_fd > 0)
		   && (chdir(devd[i].devdir) == 0)
		   && (dsearch(devd[i].dfd,fno))) {
			strcpy(devnam, devd[i].devdir);
			strcat(devnam,"/");
			strncat(devnam,dbuf->d_name,DEVLEN-strlen(devnam));
			return(devnam);
		}
	}
	return(devnam);
}

dsearch(ddir,fno)
DIR *ddir;
dev_t fno;
{
	struct stat	S;

	rewinddir(ddir);
	while ((dbuf = readdir(ddir)) != NULL) {
		if(!dbuf->d_ino) continue;
		if(stat(dbuf->d_name, &S) == -1) {
			fprintf(stderr,
				gettxt("df:17","devnm: cannot stat %s\n"),
				dbuf->d_name);
			return(0);
		}
		if((fno != S.st_rdev) 
		|| ((S.st_mode & S_IFMT) != S_IFBLK)
		|| (strcmp(dbuf->d_name,"swap") == 0)
		|| (strcmp(dbuf->d_name,"pipe") == 0)
			) continue;
		return(1);
	}
	return(0);
}
#if u3b2 || u3b15
/* heuristic function to determine logical block size of System V file system */

s5bsize(fd)
int fd;
{

	int results[3];
	int count;
	long address;
	long offset;
	char *buf;
	struct dinode *inodes;
	struct direct *dirs;
	char * p1;
	char * p2;
	
	results[1] = 0;
	results[2] = 0;

	buf = (char *)malloc(PHYSBLKSZ);

	for (count = 1; count < 3; count++) {

		address = 2048 * count;
		if (lseek(fd, address, 0) != address)
			continue;
		if (read(fd, buf, PHYSBLKSZ) != PHYSBLKSZ)
			continue;
		inodes = (struct dinode *)buf;
		if ((inodes[1].di_mode & IFMT) != IFDIR)
			continue;
		if (inodes[1].di_nlink < 2)
			continue;
		if ((inodes[1].di_size % sizeof(struct direct)) != 0)
			continue;
	
		p1 = (char *) &address;
		p2 = inodes[1].di_addr;
		*p1++ = 0;
		*p1++ = *p2++;
		*p1++ = *p2++;
		*p1   = *p2;
	
		offset = address << (count + 9);
		if (lseek(fd,offset,0) != offset)
			continue;
		if (read(fd, buf, PHYSBLKSZ) != PHYSBLKSZ)
			continue;
		dirs = (struct direct *)buf;
		if (dirs[0].d_ino != 2 || dirs[1].d_ino != 2 )
			continue;
		if (strcmp(dirs[0].d_name,".") || strcmp(dirs[1].d_name,".."))
			continue;
		results[count] = 1;
		}
	free(buf);
	
	if(results[1])
		return(Fs2b);
	if(results[2])
		return(Fs4b);
	return(-1);
}
#endif

/*
 * If we are on a NFS filesystem we have to find the mountpoint that we
 *	are currently on or below.  Just keep stating and looking to see
 *	when the st_dev changes between us and our parent.  When it does
 *	we just hit the mountpoint.  If it doesn't, just change to our
 *	parent and try again.  Point the argv pointer that we are
 *	passed at the mount point name so that we can look for it up in
 *	our caller.
 */
void
nfs_getmntpnt(name)
	char **name;
{
	struct stat s, parent_s;
	char targetwd[MAXPATHLEN], parent[MAXPATHLEN];

	/*
	 * Get the fullpath name of our argument.
	 */
	if (0 != chdir(*name)) {
		fprintf(stderr, gettxt("df:35","df: chdir failed\n"));
		exit(2);
	}
	*name = targetwd; /* make the argv point to the full path name */
	strcpy(targetwd, ".");
	strcpy(parent, "./..");
	/*
	 * stat the current directory
	 */
	if(stat(targetwd, &s) == -1) {
		fprintf(stderr, gettxt("df:36", "df: couldn't stat file.\n"));
		exit(2);
	}
	for (;;) {
		/*
		 * stat the parent directory
		 */
		if(stat(parent, &parent_s) == -1) {
			fprintf(stderr,
				gettxt("df:36","df: couldn't stat file.\n"));
			exit(2);
		}
		/*
		 * If they have the same major/minor then they are within
		 *	the same filesystem.  So we just move up one level
		 *	and keep looking.  If they are different then
		 *	the current is a mount point and we fall
		 *	through and return.
		 */
		if (parent_s.st_dev == s.st_dev) {
			if (0 != chdir(parent)) {
				fprintf(stderr,
					gettxt("df:35","df: chdir failed\n"));
				exit(2);
			}
			s = parent_s; /* parent stat is now current */
			continue;
		}
		if(getcwd(targetwd, MAXPATHLEN) == NULL) {
			fprintf(stderr, gettxt("df:37", "df: getcwd failed\n"));
			exit(2);
		}
		return;
	}
}
