/*
 * $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 *
 */

/*	Copyright (c) 1984 AT&T	*/
/*	  All Rights Reserved  	*/

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/

#ident	"$Header: grep.c 1.8 90/06/22 $@(#)grep:grep.c	1.13"
extern char *gettxt();
/*
 * grep -- print lines matching (or not matching) a pattern
 *
 *	status returns:
 *		0 - ok, and some matches
 *		1 - ok, but no matches
 *		2 - some error
 */

#include <stdio.h>
#include <ctype.h>
#include <memory.h>
#include <regexpr.h>

char errstr[][64] = {
	"Range endpoint too large." ,
	"Bad number." ,
	"``\\digit'' out of range." ,
	"Illegal or missing delimiter." ,
	"No remembered search string." ,
	"\\( \\) imbalance." ,
	"Too many \\(." ,
	"More than 2 numbers given in \\{ \\}." ,
	"} expected after \\." ,
	"First number exceeds second in \\{ \\}." ,
	"[ ] imbalance." ,
	"Regular expression overflow." ,
	"Illegal byte sequence.",
	"Unknown regexp error code!!" ,
	NULL
};

#define	errmsg(msg, arg)	fprintf(stderr, msg, arg)
#define ESIZE	512
#define	BLKSIZE	512

char	*strrchr();
int	temp;
long	lnum;
char	linebuf[2*8*BUFSIZ];
char	prntbuf[2*8*BUFSIZ];
char	expbuf[ESIZE];
int	nflag;
int	bflag;
int	lflag;
int	cflag;
int	vflag;
int	sflag;
int	iflag;
int	errflg;
int	nfile;
long	tln;
int	nsucc;
int	nlflag;
char	*ptr, *ptrend;

main(argc, argv)
register argc;
char **argv;
{
	register	c;
	register char	*arg;
	extern int	optind;

	while((c=getopt(argc, argv, "blcnsvi")) != -1)
		switch(c) {
		case 'v':
			vflag++;
			break;
		case 'c':
			cflag++;
			break;
		case 'n':
			nflag++;
			break;
		case 'b':
			bflag++;
			break;
		case 's':
			sflag++;
			break;
		case 'l':
			lflag++;
			break;
		case 'i':
			iflag++;
			break;
		case '?':
			errflg++;
		}

	if(errflg || (optind >= argc)) {
		errmsg(gettxt("grep:15", "Usage: grep -blcnsvi pattern file . . .\n"),
			(char *)NULL);
		exit(2);
	}

	argv = &argv[optind];
	argc -= optind;
	nfile = argc - 1;

	if (strrchr(*argv,'\n'))
		regerr(41);

	if (iflag) {
		for(arg = *argv; *arg != NULL; ++arg)
			*arg = (char)tolower((int)((unsigned char)*arg));
	}

	compile(*argv, expbuf, &expbuf[ESIZE]);
	if(regerrno)
		regerr(regerrno);

	if (--argc == 0)
		execute(NULL);
	else
		while (argc-- > 0)
			execute(*++argv);

	exit(nsucc == 2 ? 2 : nsucc == 0);
}

execute(file)
register char *file;
{
	register char *lbuf, *p;
	int count, count1;

	if (file == NULL)
		temp = 0;
	else if ((temp = open(file, 0)) == -1) {
		if (!sflag)
			errmsg(gettxt("grep:16", "grep: can't open %s\n"), file);
		nsucc = 2;
		return;
	}
	/* read in first block of bytes */
	if((count = read(temp, prntbuf, 8*BUFSIZ)) <= 0) {
		close(temp);

		if (cflag) {
			if (nfile>1)
				fprintf(stdout, gettxt("colon:10", "%s:"), file);

			fprintf(stdout, "%ld\n", tln);
		}
		return;
	}
		
	lnum = 0;
	tln = 0;
	ptr = prntbuf;
	for(;;) {
		/* look for next newline */
		if((ptrend = memchr(ptr, '\n', prntbuf + count - ptr)) == NULL) {
			count = prntbuf + count - ptr;
			if(count <= 8*BUFSIZ) {
				/* 
				 * shift end of block to beginning of buffer
				 * if necessary
				 * and fill up buffer until newline 
				 * is found 
				 */
				if(ptr != prntbuf)
				/* assumes memcpy copies correctly with overlap */
					memcpy(prntbuf, ptr, count);
				p = prntbuf + count;
				ptr = prntbuf;
			} else {
				/*
				 * No newline in current block.
				 * Throw it away and get next
				 * block.
				 */
				count = 0;
				ptr = p = prntbuf;
			}
			/*
			 * here is why we have a 2*8*BUFSIZ buffer. In
			 * reading 8*BUFSIZ more data plus whatever was
			 * left over previously, we would overflow a 8*BUFSIZ
			 * buffer. We could read 8*BUFSIZ-count, which would
			 * always fit into an 8*BUFSIZ buffer, but would
			 * end up making reads that were not nice even
			 * multiples of 8*BUFSIZ. In a worst case, count
			 * could be 8*BUFSIZ-1, which means we require
			 * double the amount of buffer to use this scheme.
			 */
			if((count1 = read(temp, p, 8*BUFSIZ)) > 0) {
				count += count1;
				continue;
			}
			/* end of file - last line has no newline */
			ptrend = ptr + count;
			nlflag = 0;
		} else
			nlflag = 1;
		lnum++;
		*ptrend = '\0';
		if (iflag) {
			p = ptr;
			for(lbuf=linebuf; p < ptrend; )
				*lbuf++ = (char)tolower((int)(unsigned char)*p++);
			*lbuf = '\0';
			lbuf = linebuf;
		} else
			lbuf = ptr;

		if((step(lbuf, expbuf) ^ vflag) && succeed(file) == 1)
			break;	/* lflag only once */
		if(!nlflag)
			break;
		ptr = ptrend + 1;
		if(ptr >= prntbuf + count) {
			/* at end of block; read in another block */
			ptr = prntbuf;
			if((count = read(temp, prntbuf, 8*BUFSIZ)) <= 0)
				break;
		}
	}
	close(temp);

	if (cflag) {
		if (nfile>1)
			fprintf(stdout, gettxt("colon:10", "%s:"), file);

		fprintf(stdout, "%ld\n", tln);
	}
	return;
}

succeed(f)
register char *f;
{
	int nchars;
	nsucc = (nsucc == 2) ? 2 : 1;
	if (cflag) {
		tln++;
		return(0);
	}
	if (lflag) {
		fprintf(stdout, "%s\n", f);
		return(1);
	}

	if (nfile > 1)	/* print filename */
		fprintf(stdout, gettxt("colon:10", "%s:"), f);

	if (bflag)	/* print block number */
		fprintf(stdout, gettxt("colon:12", "%ld:"), (lseek(temp, 0L, 1)-1)/BLKSIZE);

	if (nflag)	/* print line number */
		fprintf(stdout, gettxt("colon:12", "%ld:"), lnum);

	if(nlflag) {
		/* newline at end of line */
		*ptrend = '\n';
		nchars = ptrend - ptr + 1;
	} else
		nchars = ptrend - ptr;
	fwrite(ptr, 1, nchars, stdout);
	return(0);
}

regerr(err)
register err;
{
	char	msgid[16];						
	errmsg(gettxt("grep:17", "grep: RE error %d: "), err);
	switch(err) {
		case 11:
			err = 0;
			break;
		case 16:
			err = 1;
			break;
		case 25:
			err = 2;
			break;
		case 36:
			err = 3;
			break;
		case 41:
			err = 4;
			break;
		case 42:
			err = 5;
			break;
		case 43:
			err = 6;
			break;
		case 44:
			err = 7;
			break;
		case 45:
			err = 8;
			break;
		case 46:
			err = 9;
			break;
		case 49:
			err = 10;
			break;
		case 50:
			err = 11;
			break;
		case 67:
			err = 12;
			break;
		default:
			err = 13;
			break;
	}

	sprintf(msgid, "grep:%d",err+1);
	errmsg("%s\n", gettxt(msgid, errstr[err]));
	exit(2);
}
