/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

/*	Copyright (c) 1986 AT&T	*/
/*	  All Rights Reserved  	*/

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/

#ident "$Header: holdline.c 1.1 90/01/19 $@(#)/usr/bin/holdline 1.0 1986/12 AT&T-UP"

/*
 * holdline
 *
 *		This program can be used in place of getty in /etc/inittab
 *		to hold open a device "forever".  Since the last "close"
 *		of a STREAMS device causes the stream to be dismantled,
 *		it is sometimes convenient to have it stay open
 *		indefinitely, to avoid costly re-initialization every
 *		time the device is to be used.  This is particularly
 *		true of line printers, which may require modules and
 *		configurations similar to ttys.
 *
 *		The "getty" in /etc/inittab can be replaced with this
 *		program for those devices which are to be serial
 *		printers (or other devices).
 *
 *	Usage:	holdline DEVICE BAUD_RATE TERM_TYPE CONF_ENTRY
 *
 *	Actions:
 *
 *	0. Close fdescs 0 and 1.  Open the FIRST argument as fd 0, and dup
 *	   it to fd 1.  This sets stdin and stdout to the DEVICE, which
 *	   must be a full path name.
 *	1. Run "stty BAUD sane" with the SECOND argument as the baud rate.
 *	2. Export the THIRD argument into the environment as $TERM.
 *	   (The words "TERM=" don't have to be given in the argument).
 *	   This allows all child processes to have access to the device
 *	   type.
 *	3. Run "setterm -x" with the FOURTH argument as the "-x" parameter.
 *	   (i.e., the entry form the setterm "conf.file" to be run).  This
 *	   entry can be used to set parameters and push modules appropriate
 *	   to the particular device.
 *	4. Close fd 2, and dup fd 1 onto it.  This makes stdin, stdout, and
 *	   stderr all hooked to the DEVICE command line argument.  Up to
 *	   this point, all stderr output will go to the terminal from
 *	   which holdline is invoked (usually the console).
 *	5. Sleep forever.  The SIGTERM signal is waited for, and causes
 *	   a message to be printed before exit.
 */

#include <stdio.h>
#include <fcntl.h>
#include <signal.h>

#define TIMEOUT		30
#define PATH		"PATH=/bin:/usr/bin:/etc:/usr/lib/setterm"

int handle();
char *prog;

main(argc, argv)
	int argc;
	char **argv;
{
	register int fd;
	unsigned char term[128], set[128];

	prog = *argv++;
	if (argc != 5) {
		fprintf(stderr, "Usage: %s device baud term conf_entry\n",
			prog);
		exit(1);
	}
	close(0);
	close(1);
	if ((fd = open(*argv, O_RDWR)) != 0) {
		fprintf(stderr, "%s: failed to re-open stdin.\n", prog);
		goaway(1);
	}
	else {
/* fprintf(stderr, "DEV: [%s]\n", *argv); */
		++argv;
	}
	/*
	 * 1. Run "stty BAUD sane" with the SECOND argument as baud rate.
	 */
/* fprintf(stderr, "BAUD: [%s]\n", *argv); */
	strcpy(set, "stty ");
	strcat(set, *argv++);
	strcat(set, " sane");
	system(set);
	/*
	 * 2. export THIRD arg as $TERM.
	 */
/* fprintf(stderr, "TERM: [%s]\n", *argv); */
	strcpy(term, "TERM=");
	strcat(term, *argv++);
	putenv(term);
	putenv(PATH);	/* set our path for children */
 	/*
	 * 3. run FOURTH arg as arg to "setterm -x".
	 */
/* fprintf(stderr, "CONF: [%s]\n", *argv); */
	strcpy(set, "setterm -x ");
	strcat(set, *argv++);
	system(set);
	/*
	 * 4. close 2. dup 1 to 2.
	 */
	close(2);
	dup(1);
	/*
	 * 5. sleep forever.
	 */
	signal(SIGTERM, handle);
	pause();
	exit(0);
}

goaway(n)
	int n;
{
	sleep(TIMEOUT);
	exit(n);
}

handle()
{
	fprintf(stderr, "%s: SIGTERM caught -- exit.\n", prog);
	exit(0);
}
