/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ident	"$Header: quota.c 1.7 89/08/07 $"

#ident	"@(#)quota.c	1.1 87/07/28 3.2/4.3 NFSSRC"

#ident	"@(#)quota.c 1.1 86/09/25 SMI"	/* form UCB 4.4 06/21/83 */

/*
 * quota.c
 *	Disk quota reporting program.
 */

/* $Log:	quota.c,v $
 */

#include <stdio.h>
#include <mntent.h>
#include <ctype.h>
#include <pwd.h>
#include <errno.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/ufsquota.h>
#include <limits.h>

int	vflag;
#define QFNAME	"quotas"

#define kb(n)   (howmany(dbtob(n), 1024))

extern	struct passwd *getpwuid(), *getpwnam();
extern	void	exit(), perror();
extern	ushort	getuid();
extern	time_t	time();
extern	long	lseek();
extern	char	*strcpy(), *strcat();

main(argc, argv)
	char *argv[];
{
	register char *cp;

	argc--,argv++;
	while (argc > 0) {
		if (argv[0][0] == '-')
			for (cp = &argv[0][1]; *cp; cp++) switch (*cp) {

			case 'v':
				vflag++;
				break;

			default:
				(void) fprintf(stderr,
					"quota: %c: unknown option\n", *cp);
				exit(1);
			}
		else
			break;
		argc--, argv++;
	}
	if (argc == 0) {
		showuid((uid_t) getuid());
		exit(0);
	}
	for (; argc > 0; argc--, argv++) {
		if (alldigits(*argv))
			showuid((uid_t) atoi(*argv));
		else
			showname(*argv);
	}
	exit(0);
	/* NOTREACHED */
}

showuid(uid)
	uid_t uid;
{
	struct passwd *pwd = getpwuid(uid);

	if (uid == 0) {
		if (vflag)
			(void) printf("no disk quota for uid 0\n");
		return;
	}
	if (pwd == NULL)
		showquotas(uid, "(no account)");
	else
		showquotas(uid, pwd->pw_name);
}

showname(name)
	char *name;
{
	struct passwd *pwd = getpwnam(name);

	if (pwd == NULL) {
		(void) fprintf(stderr, "quota: %s: unknown user\n", name);
		return;
	}
	if (pwd->pw_uid == 0) {
		if (vflag)
			(void) printf("no disk quota for %s (uid 0)\n", name);
		return;
	}
	showquotas(pwd->pw_uid, name);
}

showquotas(uid, name)
	uid_t uid;
	char *name;
{
	register struct mntent *mntp;
	FILE *mtab;
	struct dqblk dqblk;
	uid_t myuid;

	myuid = (uid_t) getuid();
	if (uid != myuid && myuid != 0) {
		(void) printf("quota: %s (uid %d): permission denied\n",
			       name, uid);
		return;
	}
	if (vflag)
		heading(uid, name);
	mtab = setmntent(MNTTABNAME, "r");
	while (mntp = getmntent(mtab)) {
		if (strcmp(mntp->mnt_type, MNTTYPE_42) == 0) {
			if (quotactl(Q_GETQUOTA,
			      mntp->mnt_fsname, (int) uid, &dqblk) != 0) {
				if (!vflag || !getdiskquota(mntp, uid, &dqblk))
					continue;
			}
		} else {
			if (vflag || !hasmntopt(mntp, MNTOPT_NOQUOTA))
				getotherquota(mntp, uid);
			continue;
		}
		if (dqblk.dqb_bsoftlimit == 0 && dqblk.dqb_bhardlimit == 0 &&
		    dqblk.dqb_fsoftlimit == 0 && dqblk.dqb_fhardlimit == 0)
			continue;
		if (vflag)
			prquota(mntp, &dqblk);
		else
			warn(mntp, &dqblk);
	}
	(void) endmntent(mtab);
}

warn(mntp, dqp)
	register struct mntent *mntp;
	register struct dqblk *dqp;
{
	time_t t_sec;

	t_sec = time(NULL);
	if (dqp->dqb_bhardlimit &&
	     dqp->dqb_curblocks >= dqp->dqb_bhardlimit) {
		(void) printf("Block limit reached on %s\n", mntp->mnt_dir);
	} else if (dqp->dqb_bsoftlimit &&
	     dqp->dqb_curblocks >= dqp->dqb_bsoftlimit) {
		if (dqp->dqb_btimelimit == 0) {
			(void) printf("Over disk quota on %s, remove %dK\n",
			    mntp->mnt_dir,
			    kb(dqp->dqb_curblocks - dqp->dqb_bsoftlimit + 1));
		} else if (dqp->dqb_btimelimit > t_sec) {
			char btimeleft[80];

			fmttime(btimeleft, dqp->dqb_btimelimit - t_sec);
			(void) printf(
"Over disk quota on %s, remove %dK within %s\n",
			    mntp->mnt_dir,
			    kb(dqp->dqb_curblocks - dqp->dqb_bsoftlimit + 1),
			    btimeleft
			);
		} else {
			(void) printf(
"Over disk quota on %s, time limit has expired, remove %dK\n",
			    mntp->mnt_dir,
			    kb(dqp->dqb_curblocks - dqp->dqb_bsoftlimit + 1)
			);
		}
	}
	if (dqp->dqb_fhardlimit &&
	    dqp->dqb_curfiles >= dqp->dqb_fhardlimit) {
		(void) printf("File count limit reached on %s\n",
			      mntp->mnt_dir);
	} else if (dqp->dqb_fsoftlimit &&
	    dqp->dqb_curfiles >= dqp->dqb_fsoftlimit) {
		if (dqp->dqb_ftimelimit == 0) {
			(void) printf(
"Over file quota on %s, remove %d file%s\n",
			    mntp->mnt_dir,
			    dqp->dqb_curfiles - dqp->dqb_fsoftlimit + 1,
			    ((dqp->dqb_curfiles - dqp->dqb_fsoftlimit + 1) > 1 ?
				"s" :
				"" )
			);
		} else if (dqp->dqb_ftimelimit > t_sec) {
			char ftimeleft[80];

			fmttime(ftimeleft, dqp->dqb_ftimelimit - t_sec);
			(void) printf(
"Over file quota on %s, remove %d file%s within %s\n",
			    mntp->mnt_dir,
			    dqp->dqb_curfiles - dqp->dqb_fsoftlimit + 1,
			    ((dqp->dqb_curfiles - dqp->dqb_fsoftlimit + 1) > 1 ?
				"s" :
				"" ),
			    ftimeleft
			);
		} else {
			(void) printf(
"Over file quota on %s, time limit has expired, remove %d file%s\n",
			    mntp->mnt_dir,
			    dqp->dqb_curfiles - dqp->dqb_fsoftlimit + 1,
			    ((dqp->dqb_curfiles - dqp->dqb_fsoftlimit + 1) > 1 ?
				"s" :
				"" )
			);
		}
	}
}

heading(uid, name)
	uid_t uid;
	char *name;
{
	(void) printf("Disk quotas for %s (uid %d):\n", name, uid);
	(void) printf("%-12s %7s%7s%7s%12s%7s%7s%7s%12s\n"
		, "Filesystem"
		, "usage"
		, "quota"
		, "limit"
		, "timeleft"
		, "files"
		, "quota"
		, "limit"
		, "timeleft"
	);
}

prquota(mntp, dqp)
	register struct mntent *mntp;
	register struct dqblk *dqp;
{
	time_t t_sec;
	char ftimeleft[80], btimeleft[80];
	char *cp;

	t_sec = time(NULL);
	if (dqp->dqb_bsoftlimit && dqp->dqb_curblocks >= dqp->dqb_bsoftlimit) {
		if (dqp->dqb_btimelimit == 0) {
			(void) strcpy(btimeleft, "NOT STARTED");
		} else if (dqp->dqb_btimelimit > t_sec) {
			fmttime(btimeleft, dqp->dqb_btimelimit - t_sec);
		} else {
			(void) strcpy(btimeleft, "EXPIRED");
		}
	} else {
		btimeleft[0] = '\0';
	}
	if (dqp->dqb_fsoftlimit && dqp->dqb_curfiles >= dqp->dqb_fsoftlimit) {
		if (dqp->dqb_ftimelimit == 0) {
			(void) strcpy(ftimeleft, "NOT STARTED");
		} else if (dqp->dqb_ftimelimit > t_sec) {
			fmttime(ftimeleft, dqp->dqb_ftimelimit - t_sec);
		} else {
			(void) strcpy(ftimeleft, "EXPIRED");
		}
	} else {
		ftimeleft[0] = '\0';
	}
	if (strlen(mntp->mnt_dir) > 12) {
		(void) printf("%s\n", mntp->mnt_dir);
		cp = "";
	} else {
		cp = mntp->mnt_dir;
	}
	(void) printf("%-12.12s %7d%7d%7d%12s%7d%7d%7d%12s\n",
	    cp,
	    kb(dqp->dqb_curblocks),
	    kb(dqp->dqb_bsoftlimit),
	    kb(dqp->dqb_bhardlimit),
	    btimeleft,
	    dqp->dqb_curfiles,
	    dqp->dqb_fsoftlimit,
	    dqp->dqb_fhardlimit,
	    ftimeleft
	);
}

fmttime(buf, time_left)
	char *buf;
	register ulong time_left;
{
	int i;
	static struct {
		int c_secs;		/* conversion units in secs */
		char * c_str;		/* unit string */
	} cunits [] = {
		{60*60*24*28, "months"},
		{60*60*24*7, "weeks"},
		{60*60*24, "days"},
		{60*60, "hours"},
		{60, "mins"},
		{1, "secs"}
	};

	if (time_left <= 0) {
		(void) strcpy(buf, "EXPIRED");
		return;
	}
	for (i = 0; i < sizeof(cunits)/sizeof(cunits[0]); i++) {
		if (time_left >= cunits[i].c_secs)
			break;
	}
	(void) sprintf(buf, "%.1f %s", (double)time_left/cunits[i].c_secs,
		 cunits[i].c_str);
}

alldigits(s)
	register char *s;
{
	register c;

	c = *s++;
	do {
		if (!isdigit(c))
			return (0);
	} while (c = *s++);
	return (1);
}

int
getdiskquota(mntp, uid, dqp)
	struct mntent *mntp;
	uid_t uid;
	struct dqblk *dqp;
{
	int fd;
	dev_t fsdev;
	struct stat statb;
	char qfilename[MAXPATHLEN];
	static int nolocalquota = 0;
	extern int errno;

	if (nolocalquota)
		return (0);
	if (stat(mntp->mnt_fsname, &statb) < 0 ||
	    (statb.st_mode & S_IFMT) != S_IFBLK)
		return (0);
	fsdev = statb.st_rdev;
	(void) sprintf(qfilename, "%s/%s", mntp->mnt_dir, QFNAME);
	if (stat(qfilename, &statb) < 0 || statb.st_dev != fsdev)
		return (0);
	if ((fd = open(qfilename, O_RDONLY)) < 0)
		return (0);
	if (quotactl(Q_SYNC,mntp->mnt_fsname,0,NULL) < 0 && errno == EINVAL) {
		nolocalquota++;
		return (0);
	}
	(void) lseek(fd, (long)dqoff(uid), SEEK_SET);
	if (read(fd, dqp, sizeof(struct dqblk)) != sizeof(struct dqblk)) {
		(void) close(fd);
		return (0);
	}
	(void) close(fd);
	return (1);
}

getotherquota(mntp, uid)
	struct mntent *mntp;
	uid_t uid;
{

	char path[PATH_MAX+1];
	char *nargv[5];
	char uid_string[12];
	int count;
	int pid, wpid;
	int wstat;

	path[0] = '\0';
	(void) strcat(path, "/bin/quota_");
	(void) strcat(path, mntp->mnt_type);	/* "/bin/quota_type" */
	nargv[0] = path;
	count = 1;
	if (vflag)
		nargv[count++] = "-v";
	(void) sprintf(uid_string, "%d", uid);
	nargv[count++] = uid_string;
	nargv[count++] = mntp->mnt_fsname;
	nargv[count] = 0;

	switch (pid = fork()) {
	case -1:
		perror("fork");
		exit(8);
		break;

	case 0:
		(void) execv(path, nargv);
		break;
	default:
		while ((wpid = wait(&wstat)) >= 0)
			if (wpid == pid)
				break;

	}
}
