/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ident	"$Header: quotacheck.c 2.10 89/08/07 $"

#ident	"@(#)quotacheck.c	1.2 87/07/24 3.2/4.3NFSSRC"

/*
 * Copyright (c) 1980 Regents of the University of California.
 * All rights reserved.  The Berkeley software License Agreement
 * specifies the terms and conditions for redistribution.
 */

#ifndef lint
char copyright[] =
"@(#) Copyright (c) 1980 Regents of the University of California.\n\
 All rights reserved.\n";
#endif /* not lint */

#ident	"@(#)quotacheck.c	5.6 (Berkeley) 11/3/85"

/*
 * quotacheck.c
 *	Fix up / report on disc quotas & usage
 */

/* $Log:	quotacheck.c,v $
 */

#include <stdio.h>
#include <ctype.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/vnode.h>
#include <sys/ufsinode.h>
#include <sys/ufsfilsys.h>
#include <sys/ufsquota.h>
#include <sys/stat.h>
/* #include <sys/wait.h> */
#include <mntent.h>
#include <pwd.h>

#define MAXMOUNT	50

union un {
	struct	fs	sblk;
	char	dummy[MAXBSIZE];
} *un;
#define	sblock	un->sblk

#define	ITABSZ	256
struct	dinode	*itab;
struct	dinode	*dp;

#define LOGINNAMESIZE	8
struct fileusage {
	struct fileusage *fu_next;
	ulong fu_curfiles;
	ulong fu_curblocks;
	ushort	fu_uid;
	char fu_name[LOGINNAMESIZE + 1];
};
#define FUHASH 997
struct fileusage *fuhead[FUHASH];
struct fileusage *lookup();
struct fileusage *adduid();
int highuid;

int fi;
ino_t ino;
struct	dinode	*ginode();
char		*makerawname();
int		chkquota();

int	vflag;		/* verbose */
int	aflag;		/* all file systems */
int	pflag;		/* fsck like parallel check */

#define QFNAME "quotas"
char *listbuf[MAXMOUNT];
struct fileusage zerofileusage;

extern	struct passwd *getpwent();
extern	void	exit(), sync(), perror(), setpwent(), endpwent();
extern	long	lseek();
extern	char	*strcpy(), *strncpy();
extern	char	*malloc(), *valloc();
extern int	errno;

main(argc, argv)
	int argc;
	char **argv;
{
	register struct mntent *mntp;
	register struct fileusage *fup;
	char **listp;
	int listcnt;
	FILE *mtab, *fstab;
	int errs = 0;
	struct passwd *pw;

again:
	argc--, argv++;
	if (argc > 0 && strcmp(*argv, "-v") == 0) {
		vflag++;
		goto again;
	}
	if (argc > 0 && strcmp(*argv, "-a") == 0) {
		aflag++;
		goto again;
	}
	if (argc > 0 && strcmp(*argv, "-p") == 0) {
		pflag++;
		goto again;
	}
	if (argc <= 0 && !aflag) {
		(void) fprintf(stderr, "Usage:\n\t%s\n\t%s\n",
			"quotacheck [-v] [-p] -a",
			"quotacheck [-v] [-p] filesys ...");
		exit(1);
	}

	(void) setpwent();
	while ((pw = getpwent()) != 0) {
		fup = lookup((ushort)pw->pw_uid);
		if (fup == 0) {
			fup = adduid((ushort)pw->pw_uid);
			(void) strncpy(fup->fu_name, pw->pw_name,
				sizeof(fup->fu_name));
		}
	}
	(void) endpwent();

	if (aflag) {

		/*
		 * Go through fstab and make a list of appropriate
		 * filesystems.
		 */
		listp = listbuf;
		listcnt = 0;
		if ((fstab = setmntent(FSTABNAME, "r")) == NULL) {
			(void) fprintf(stderr, "Can't open ");
			perror(FSTABNAME);
			exit(8);
		}
		while (mntp = getmntent(fstab)) {
			if (strcmp(mntp->mnt_type, MNTTYPE_42) != 0 ||
			    !hasmntopt(mntp, MNTOPT_QUOTA) ||
			    hasmntopt(mntp, MNTOPT_RO))
				continue;
			*listp = malloc((uint) strlen(mntp->mnt_fsname) + 1);
			(void) strcpy(*listp, mntp->mnt_fsname);
			listp++;
			listcnt++;
		}
		(void) endmntent(fstab);
		*listp = (char *)0;
		listp = listbuf;
	} else {
		listp = argv;
		listcnt = argc;
	}
	if (pflag && listcnt >= 2) {

		errs = preen_list(listcnt, listp, chkquota);

	} else {
		if ((mtab = setmntent(MNTTABNAME, "r")) == NULL) {
			(void) fprintf(stderr, "Can't open ");
			perror(MNTTABNAME);
			exit(8);
		}
		while (mntp = getmntent(mtab)) {
			if (strcmp(mntp->mnt_type, MNTTYPE_42) == 0 &&
			    !hasmntopt(mntp, MNTOPT_RO) &&
			    (oneof(mntp->mnt_fsname, listp, listcnt) ||
			     oneof(mntp->mnt_dir, listp, listcnt)) ) {
				errs += chkquota(mntp);
			}
		}
		(void) endmntent(mtab);

		while (listcnt--) {
			if (*listp)
				(void) fprintf(stderr, "Cannot check %s\n",
						*listp);
		}
	}
	if (errs > 255) errs = 255;
	exit(errs);
	/* NOTREACHED */
}

/* physical disk info structure */
struct phys_info {
	struct phys_info *phys_next;	/* linked list */
	struct part_info *phys_part;	/* partition info (head) */
	struct part_info *phys_part_tl;	/* partition info tail pointer */
	char	phys_name[DYNIX_MAXNAMLEN]; /* base name of drive */
	int	phys_pid;		/* PID of fsck checking this device */
};
struct phys_info *phys_head;		/* list head */
struct phys_info *phys_tail;		/* list tail */

/* partition info */
struct part_info {
	struct part_info *part_next;	/* linked list */
	struct mntent	part_mnt;	/* copy of the mount structure */
};


/*
 * preen_list
 *
 * Create a linked list of physical disks that each contain a list of
 * partitions for that disk.  Spawn a parallel process for each physical
 * device.  Avoid seeks by applying 'function' to each partition serially.
 */

preen_list(count, speclist, function)
	int	count;			/* argc */
	char	*speclist[];		/* argv */
	int	(*function)();		/* function to call */
{
	register struct phys_info *phys;
	register struct part_info *part;
	register char	*fs;
	register struct mntent *mntp;
	struct phys_info	*AddDisk();
	char	*DiskBaseName();
	FILE	*mtab;
	int	sumstatus = 0;
	int	w_status;
#ifdef DEBUG
	int	pid;
#endif /* DEBUG */

	if ((mtab = setmntent(MNTTABNAME, "r")) == NULL) {
		(void) fprintf(stderr, "Can't open ");
		perror(MNTTABNAME);
		exit(8);
	}

	phys_head = phys_tail = NULL;	/* just to make sure */

	while (mntp = getmntent(mtab)) {
		fs = mntp->mnt_fsname;
#ifdef DEBUG
		printf("preen_list(): fs = '%s'\n", fs);
		fflush(stdout);
#endif /* DEBUG */

		if (strcmp(mntp->mnt_type, MNTTYPE_42) != 0 ||
		    hasmntopt(mntp, MNTOPT_RO) ||
		    (!oneof(fs, speclist, count) &&
		     !oneof(mntp->mnt_dir, speclist, count)))
			continue;

		phys = AddDisk(DiskBaseName(fs));
#ifdef DEBUG
		printf("preen_list(): fs = '%s', BaseName = '%s'\n",
			fs, phys->phys_name);
		fflush(stdout);
#endif /* DEBUG */
		AddSlice(phys, mntp);
	}
	(void) endmntent(mtab);


	/*
	 * Fork a child process for each physical disk and let
	 * the children traverse the partition lists.
	 */
	for (phys = phys_head; phys; phys = phys->phys_next) {
#ifdef DEBUG
		printf("pid %d: forking()\n", getpid());
		fflush(stdout);
#endif /* DEBUG */
		if ((phys->phys_pid = fork()) < 0) {
			perror("fork");
			exit(8);
		}
		if (phys->phys_pid != 0)
			continue;
		/*
		 * CHILDREN ONLY:
		 *
		 * Execute 'function' on one partition at a time
		 * using the list of partitions specified.  Each
		 * file system in this list is on the same physical
		 * drive.
		 */
		for (part = phys->phys_part; part; part = part->part_next) {
#ifdef DEBUG
			(void) printf("(phys_pid = %d)\t", getpid());
			(void) printf("fsck -p %s\n", 
					part->part_mnt.mnt_fsname);
			(void) fflush(stdout);
#endif /* DEBUG */
			/*
			 * call appropriate function and
			 * collect exit status...
			 */
			sumstatus += (*function)(&part->part_mnt);
		}
		/* CHILDREN EXIT */
		if (sumstatus > 255) sumstatus = 255;
		exit(sumstatus);
	}

	/*
	 * Wait for children.  Exit with the max status collected from
	 * all child exits.
	 */

#ifndef	DEBUG
	while (wait(&w_status) > 0) {
#else
	while ((pid = wait(&w_status)) > 0) {
		printf("waitfor(pid=%d, w_status = 0x%x):\n", pid, w_status);
		fflush(stdout);
#endif	/* DEBUG */
		sumstatus += (w_status >> 8);
	}
	if (sumstatus > 255) sumstatus = 255;
	exit(sumstatus);
}


/*
 * Convert a disk name (either raw or block) to the base name of the disk.
 * This is useful for finding the physical device associated with a
 * partition.
 *
 * Examples of conversion might be:
 *	/dev/rdsk/zd0s0	yields	zd0
 *	/dev/dsk/zd0s6	yields	zd0
 *	/dev/rzd0a	yields	zd0
 *	/dev/zd0g	yields	zd0
 */

static
char *
DiskBaseName(name)
	char	*name;
{
	register char *b, *e;		/* begin and end of BaseName */
	static char newnm[DYNIX_MAXNAMLEN]; /* don't trash 'name' */
	extern char *strrchr();

	b = strrchr(name, '/');
	b = b ? b + 1 : name;		/* if no '/', already basename */
	if (*b == 'r')
		b++;

	e = b;
	while (isalpha(*e))
		e++;
	while (isdigit(*e))
		e++;
	if (*e == '\0')
		return(b);
	(void) strncpy(newnm, b, e - b);
	newnm[e - b] = '\0';
	return(newnm);
}


/*
 * Add a disk to the linked list of physical drives.  If that drive
 * has already been added, return a pointer to it.
 * 'diskname' is the physical device.
 */

static
struct phys_info *
AddDisk(diskname)
	char	*diskname;
{
	register struct phys_info *phys;

#ifdef DEBUG
	printf("AddDisk(): diskname = %s\n", diskname);
	fflush(stdout);
#endif /* DEBUG */
	/* now search the existing list */
	for (phys = phys_head; phys; phys = phys->phys_next) {
		if (strcmp(phys->phys_name, diskname) == 0)
			return(phys);	/* found it */
	}

	/* make a new entry */
	phys = (struct phys_info *) malloc(sizeof(struct phys_info));
#ifdef DEBUG
	printf("AddDisk(): sizeof(phys_info) = %d\n", sizeof(struct phys_info));
	printf("AddDisk(): phys = 0x%x\n", phys);
	fflush(stdout);
#endif /* DEBUG */
	if (phys == NULL) {
		perror("Out of memory (phys_info)");
		exit(8);
	}
	phys->phys_next = NULL;
	phys->phys_part = NULL;
	phys->phys_part_tl = NULL;
	(void) strcpy(phys->phys_name, diskname);
	phys->phys_pid = 0;
	if (phys_tail) {
		phys_tail->phys_next = phys;
		phys_tail = phys;
	} else
		phys_head = phys_tail = phys;
	return(phys);
}


/*
 * Add a partition descripton to list associated with 'phys'.
 * 'specname' is the name of the device special file to be checked.
 */

static
AddSlice(phys, mntp)
	register struct phys_info *phys;
	struct mntent *mntp;
{
	register struct part_info *part;

	part = (struct part_info *) malloc(sizeof(struct part_info));
	if (part == NULL) {
		perror("Out of memory (part_info)");
		exit(8);
	}
	part->part_next = NULL;
	part->part_mnt.mnt_fsname = 
			malloc((uint) strlen(mntp->mnt_fsname) + 1);
	(void) strcpy(part->part_mnt.mnt_fsname, mntp->mnt_fsname);
	part->part_mnt.mnt_dir =
			malloc((uint) strlen(mntp->mnt_dir) + 1);
	(void) strcpy(part->part_mnt.mnt_dir, mntp->mnt_dir);
	part->part_mnt.mnt_type = malloc((uint) strlen(mntp->mnt_type) + 1);
	(void) strcpy(part->part_mnt.mnt_type, mntp->mnt_type);
	part->part_mnt.mnt_opts = malloc((uint) strlen(mntp->mnt_opts) + 1);
	(void) strcpy(part->part_mnt.mnt_opts, mntp->mnt_opts);

#ifdef DEBUG
	printf("AddSlice(): phys = 0x%x, part = 0x%x, name = %s, dir = %s. type = %s, opt = %s\n",
			phys, part, part->part_mnt.mnt_fsname, part->part_mnt.mnt_dir, part->part_mnt.mnt_type, part->part_mnt.mnt_opts);
		fflush(stdout);
#endif /* DEBUG */

	if (phys->phys_part_tl) {
		phys->phys_part_tl->part_next = part;
		phys->phys_part_tl = part;
	} else
		phys->phys_part = phys->phys_part_tl = part;
}

chkquota(mntp)
	register struct mntent *mntp;
{
	register struct fileusage *fup;
	char *fsdev;
	char *fsfile;
	char qffile[MAXPATHLEN + 1];
	dev_t quotadev;
	FILE *qf;
	register ushort uid;
	register ushort highquota;
	int cg, i;
	char *rawdisk;
	struct stat statb;
	struct dqblk dqbuf;
        static int warned = 0;
	extern int errno;

	(void) sprintf(qffile, "%s/%s",	mntp->mnt_dir, QFNAME);
	fsdev = mntp->mnt_fsname;
	fsfile = mntp->mnt_dir;
	rawdisk = makerawname(fsdev);
	if (vflag)
		(void) printf("*** Checking quotas for %s (%s)\n", rawdisk, fsfile);
	fi = open(rawdisk, 0);
	if (fi < 0) {
		perror(rawdisk);
		return (1);
	}
	qf = fopen(qffile, "r+");
	if (qf == NULL) {
		perror(qffile);
		(void) close(fi);
		return (1);
	}
	if (fstat(fileno(qf), &statb) < 0) {
		perror(qffile);
		(void) fclose(qf);
		(void) close(fi);
		return (1);
	}
	quotadev = statb.st_dev;
	if (stat(fsdev, &statb) < 0) {
		perror(fsdev);
		(void) fclose(qf);
		(void) close(fi);
		return (1);
	}
	if (quotadev != statb.st_rdev) {
		(void) fprintf(stderr,
			"%s dev (0x%x) mismatch %s dev (0x%x)\n",
			qffile, quotadev, fsdev, statb.st_rdev);
		(void) fclose(qf);
		(void) close(fi);
		return (1);
	}
	if ((un = (union un *)valloc(sizeof (union un))) == NULL ||
	    (itab = (struct dinode *)valloc(ITABSZ * sizeof(struct dinode)))
	     == NULL) {
		(void) fprintf(stderr, "valloc failed\n");
		(void) fclose(qf);
		(void) close(fi);
		return (1);
	}
        if (quotactl(Q_SYNC, fsdev, 0, NULL) < 0 &&
            errno == EINVAL && !warned && vflag) {
                warned++;
                (void) printf(
"*** Warning: Quotas are not configured into this kernel\n");
        }
        sync();
	bread(SBLOCK, (char *)&sblock, SBSIZE);
	ino = 0;
	for (cg = 0; cg < sblock.fs_ncg; cg++) {
		dp = NULL;
		for (i = 0; i < sblock.fs_ipg; i++)
			(void) acct_space(ginode());
	}
	highquota = 0;
	for (uid = 0; uid <= highuid; uid++) {
		(void) fseek(qf, (long)dqoff(uid), 0);
		(void) fread(&dqbuf, sizeof(struct dqblk), 1, qf);
		if (feof(qf))
			break;
		fup = lookup(uid);
		if (fup == 0)
			fup = &zerofileusage;
		if (dqbuf.dqb_bhardlimit || dqbuf.dqb_bsoftlimit ||
		    dqbuf.dqb_fhardlimit || dqbuf.dqb_fsoftlimit) {
			highquota = uid;
		} else {
			fup->fu_curfiles = 0;
			fup->fu_curblocks = 0;
		}
		if (dqbuf.dqb_curfiles == fup->fu_curfiles &&
		    dqbuf.dqb_curblocks == fup->fu_curblocks) {
			fup->fu_curfiles = 0;
			fup->fu_curblocks = 0;
			continue;
		}
		if (vflag) {
			if (pflag || aflag)
				(void) printf("%s: ", rawdisk);
			if (fup->fu_name[0] != '\0')
				(void) printf("%-10s fixed:", fup->fu_name);
			else
				(void) printf("#%-9d fixed:", uid);
			if (dqbuf.dqb_curfiles != fup->fu_curfiles)
				(void) printf("  files %d -> %d",
				    dqbuf.dqb_curfiles, fup->fu_curfiles);
			if (dqbuf.dqb_curblocks != fup->fu_curblocks)
				(void) printf("  blocks %d -> %d",
				    dqbuf.dqb_curblocks, fup->fu_curblocks);
			(void) printf("\n");
		}
		dqbuf.dqb_curfiles = fup->fu_curfiles;
		dqbuf.dqb_curblocks = fup->fu_curblocks;
		(void) fseek(qf, (long)dqoff(uid), 0);
		(void) fwrite(&dqbuf, sizeof(struct dqblk), 1, qf);
		(void) quotactl(Q_SETQUOTA, fsdev, (int) uid, &dqbuf);
		fup->fu_curfiles = 0;
		fup->fu_curblocks = 0;
	}
	(void) fflush(qf);
	(void) ftruncate(fileno(qf), (highquota + 1) * sizeof(struct dqblk));
	(void) fclose(qf);
	(void) close(fi);
	return (0);
}


acct_space(ip)
	register struct dinode *ip;
{
	register struct fileusage *fup;

	if (ip == NULL)
		return;
	if (ip->di_mode == 0)
		return;
	fup = adduid((ushort)ip->di_uid);
	fup->fu_curfiles++;
	if ((ip->di_mode & IFMT) == IFCHR || (ip->di_mode & IFMT) == IFBLK)
		return;
	fup->fu_curblocks += ip->di_blocks;
}

oneof(target, listp, n)
	char *target;
	register char **listp;
	register int n;
{

	while (n--) {
		if (*listp && strcmp(target, *listp) == 0) {
			*listp = (char *)0;
			return (1);
		}
		listp++;
	}        
	return (0);
}

struct dinode *
ginode()
{
	register unsigned long iblk;

	if (dp == NULL || ++dp >= &itab[ITABSZ]) {
		iblk = itod(&sblock, ino);
		bread((ulong)fsbtodb(&sblock, iblk),
		    (char *)itab, ITABSZ * sizeof(struct dinode));
		dp = &itab[ino % INOPB(&sblock)];
	}
	if (ino++ < ROOTINO)
		return(NULL);
	return(dp);
}

bread(bno, buf, cnt)
	long unsigned bno;
	char *buf;
{
	(void) lseek(fi, (long)dbtob(bno), 0);
	if (read(fi, buf, (uint) cnt) != cnt) {
		perror("read");
		exit(1);
	}
}

struct fileusage *
lookup(uid)
	register ushort uid;
{
	register struct fileusage *fup;

	for (fup = fuhead[uid % FUHASH]; fup != 0; fup = fup->fu_next)
		if (fup->fu_uid == uid)
			return (fup);
	return ((struct fileusage *)0);
}

struct fileusage *
adduid(uid)
	register ushort uid;
{
	struct fileusage *fup, **fhp;
	extern char *calloc();

	fup = lookup(uid);
	if (fup != 0)
		return (fup);
	fup = (struct fileusage *)calloc(1, sizeof(struct fileusage));
	if (fup == 0) {
		(void) fprintf(stderr,
			"out of memory for fileusage structures\n");
		exit(1);
	}
	fhp = &fuhead[uid % FUHASH];
	fup->fu_next = *fhp;
	*fhp = fup;
	fup->fu_uid = uid;
	if (uid > highuid)
		highuid = uid;
	return (fup);
}


/*
 * Convert a block device name to the equivalent raw name, by prepending
 * an "r" after "/dev/".  The new name is stuck into static storage.
 * If "/dev/" is not present - just return the name as is.
 *
 * Note that this routine will (correctly) not convert "/dev/root".
 * Examples of conversion might be:
 *	/dev/zd0a	yields	/dev/rzd0a
 *	/dev/dsk/zd0s0	yields	/dev/rdsk/zd0s0
 */

char *
makerawname(name)
	register char	*name; 
{
	register char	*p;
	static char	newnm[MAXPATHLEN];	/* don't trash 'name' */

	if (strncmp(name, "/dev/", 5)) {
		return(name);
	}
	p = name + 5;
	if (*p == 'r')
		return(name);
	(void) sprintf(newnm, "/dev/r%s", p);
	return(newnm);
}

