/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.

/*	Copyright (c) 1986,1989 AT&T	*/
/*	  All Rights Reserved  	*/

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/

#ident "$Header: cmds.c 1.3 90/09/07 $ @(#)setterm:cmds.c 2.1 1989/02/14 AT&T-UP"

/*
 * cmds.c	Command routines.  Note: routines that can't push a module
 *		and wish to post a fatal error "BAD_PUSH", should copy the
 *		name of the module to "command.action", where it will be
 *		pickedup by the "fatal()" routine.
 */

#include <stdio.h>
#include <curses.h>		/* for terminfo */
#include <term.h>		/* for curses */
#include <sys/param.h>
#include <fcntl.h>		/* file control for open() */
#include <stropts.h>		/* Streams specific ioctls */

#include "header.h"

extern char *getdevt(), *getcom();
extern char *prog, *infoptr;
extern int verify, opt_run, touched;

struct cmd command;		/* command structure for types & args */
struct sel sellist[MAXS];	/* select list */
int nsels;			/* number of items in current select list */

char vars[10][20];		/* for "set" variables */

struct termio tt;		/* for throw/catch */
int th_active = 0;		/* TRUE if a "throw" is active */

/*
 * pop_all	Pop off all modules via calls to "pop()".
 */

pop_all()

{
	if (verify)
		printf("  POPALL\n");
	else {
		while (pop())
			;
	}
}

/*
 * pop		Pop off one module.
 */

pop()

{
	if (verify) {
		printf("  POP\n");
		return FALSE;	/* so "verify" won't run forever! */
	}
	if (ioctl(0, I_POP, 0) >= 0) {
		++touched;
		return TRUE;
	}
	else
		return FALSE;
}

/*
 * popto	Pop all modules up to but not including any module
 *		in the given list.  List is a "|" separated list of
 *		names, similar to the "devt" entry.  If no module
 *		in the stop list is on the Stream, then all modules
 *		will end up being popped.
 */

popto()

{
	unsigned char buf[256];

	if (verify) {
		printf("  POPTO '%s'\n", command.action);
		return;
	}
	while (TRUE) {
		if (ioctl(0, I_LOOK, buf) < 0)	/* no more to pop */
			return;
		if (devequal(buf, command.action))
			return;	/* can't pop this one, so return... */
		else {
			++touched;
			ioctl(0, I_POP, 0);	/* pop this one */
		}
	}
}

/*
 * set		Set a variable to the name of the module on top.
 *		The "pseudo" variable "T" can be set as well, to be
 *		used in other commands.
 */

set()

{
	register int a;
	char *buf[128];
	register char *tty, *tmp;
	char *ttyname();

	a = command.action[0];
	if ((a < 'A' || a > 'F') && (a != 'T'))
		fatal(BAD_SET);
	if (a == 'T')
		a = 6;
	else
		a -= 'A';	/* get in range 0-6 or "T"*/
	if (verify) {
		if (a == 6)
			a = 'T' - 'A';
		printf("  SET %c\n", a+'A');
		return;
	}
	if (a == 6) {	/* set the "T" variable */
		tty = ttyname(0);
		if (tty) {	/* device is a tty... */
			/*
			 * find end of tty name, strip trailing digits,
			 * if any, leaving "generic" name.
			 */
			strcpy(&(vars[6][0]), tty);	/* get ttyname */
			tmp = (char*) &(vars[6][0]) + (strlen(&(vars[6][0])) - 1);
			while (tmp > &(vars[6][0])) {
				if (*tmp >= '0' && *tmp <= '9') {
					*tmp = '\0';
					--tmp;
				}
				else
					break;
			}
		}
		else
			vars[6][0] = '\0';	/* not a tty */
		return;
	}
	else {	/* set var to top module */
		if (ioctl(0, I_LOOK, buf) < 0)
			strcpy(&(vars[a][0]), "\0");
		else
			strcpy(&(vars[a][0]), buf);
	}
}

/*
 * push		Try to push a module; exit on failure.
 */

push()

{
	if (verify) {
		printf("  PUSH '%s'\n", command.action);
		return;
	}
	if (ioctl(0, I_PUSH, command.action) < 0) {
		fatal(BAD_PUSH);
	}
	++touched;
}

/*
 * run		Run the indicated command line by doing a "system()"
 *		on the command line.  If "-t" option specified, then
 *		replace child's environment $TERM with the argument.
 */

run()

{
	register int pid;
	char t[256];
	char *getenv();
	extern int opt_t;

	if (verify)
		printf("  RUN { %s }\n", command.action);
	else {
		if (opt_t) {
			sprintf(t, "TERM=%s", getenv("TERM"));
			putenv(t);
		}
		system(command.action);
	}
}

/*
 * cmd_if	Do the "if" command:	if VAR=NAME?MOD1:MOD2
 */

cmd_if()

{
	register char *hold, *line, *var, *test, *if_side, *else_side;

	line = command.action;
	var = line++;
	if (*line != '=')
		fatal(BAD_IF);
	else
		++line;
	if ((*var < 'A' || *var > 'F') && (*var != 'T'))
		fatal(BAD_IF);
	if (*var == 'T') {		/* requesting tty "type" name */
		*var = 'G' - 'A';
	}
	else
		*var = *var - 'A';	/* normalize to range... */
	test=line;
	while (*line && (*line != '?'))
		++line;
	if (*line != '?')
		fatal(BAD_IF);
	else {
		*line = '\0';	/* null terminate the test name */
		++line;
	}
	if_side = line;
	while (*line && (*line != ':') && (*line != ' ') && (*line != '\t'))
		++line;
	if (*line == ':') {	/* we have an else side */
		*line = '\0';	/* truncate if side */
		++line;
		else_side = line;
		while (*line && (*line != ' ') && (*line != '\t'))
			++line;
		*line = '\0';
	}
	else {
		*line = '\0';
		else_side = (char *) 0;
	}
	if (verify) {
		if (*var == 6)
			*var = 'T' - 'A';
		printf("  IF (%c == '%s')\n", (*var)+'A', test);
		printf("     THEN '%s'\n", if_side);
		if (else_side)
			printf("     ELSE '%s'\n", else_side);
		return;
	}
	/*
	 * now do the real thing...see if "NAME" is the side before
	 * the colon (the "if" side).  Compare the contents of the
	 * variable with "test"...
	 */
	if (strcmp(&(vars[*var][0]), test) == 0) {
		if (strcmp("END", if_side) == 0)
			exit(0);	/* end immediately */
		else if (strcmp("POP", if_side) == 0) {
			if (ioctl(0, I_POP, 0) == 0)
				++touched;
			return;
		}
		else if (strcmp("NUL", if_side) == 0)
			return;
		else {
			if (ioctl(0, I_PUSH, if_side) < 0) {
				strcpy(command.action, if_side);
				fatal(BAD_PUSH);
			}
			else {
				++touched;
				return;	/* successful push of if_side */
			}
		}
	}
	/*
	 * otherwise, check to see if the "NAME" is the side after the
	 * colon (the "else" side.
	 */
	if (else_side && (*else_side != ' ' && *else_side != '\t')) {
		if (strcmp("END", else_side) == 0)
			exit(0);	/* end immediately */
		else if (strcmp("POP", else_side) == 0) {
			if (ioctl(0, I_POP, 0) == 0)
				++touched;
			return;
		}
		else if (strcmp("NUL", else_side) == 0)
			return;
		else {
			if (ioctl(0, I_PUSH, else_side) < 0) {
				strcpy(command.action, else_side);
				fatal(BAD_PUSH);
			}
			else {
				++touched;
				return;	/* successful push of else_side */
			}
		}
	}
}

/*
 * test		Do the "test" command:	test VAR=NAME
 *		If module NAME is in the stream, then set VAR to its name.
 */

test()

{
	register char *hold, *line, *var, *test;

	line = command.action;

	if (verify)
		hold = line;
	var = line++;
	if (*line != '=')
		fatal(BAD_TEST);
	else
		++line;
	if ((*var < 'A' || *var > 'F') && (*var != 'T'))
		fatal(BAD_TEST);
	if (*var == 'T') {		/* requesting tty "type" name */
		*var = 'G' - 'A';
	}
	else
		*var = *var - 'A';	/* normalize to 0-6 */
	test=line;
	while (*line && (*line != ' ') && (*line != '\t'))
		++line;
	*line = '\0';	/* null terminate the test name */
	if (verify) {
		if (*var == 6)
			*var = 'T' - 'A';
		printf("  TEST (%c = '%s')\n", (*var)+'A', test);
		return;
	}
	/*
	 * now do the real thing...
	 */
	if (ioctl(0, I_FIND, test) <= 0) {	/* not found */
		vars[*var][0] = '\0';
	}
	else
		strcpy(&(vars[*var][0]), test);	/* save the name in var */
}

/*
 * select	Do the select command.  Select a module from a list
 *		of possibilities and push it, or do the action
 *		associated with it if it's a special name.
 */

select()

{
	register int i, done;
	register char *var, *line;

	line = command.action;
	var = line++;
	if (*line != '=')
		fatal(BAD_SEL);
	else
		++line;
	if ((*var < 'A' || *var > 'F') && (*var != 'T'))
		fatal(BAD_SEL);
	if (*var == 'T') {		/* requesting tty "type" name */
		*var = 6;
	}
	else
		*var = *var - 'A';	/* normalize to 0-6 */
	done = FALSE;
	nsels = 0;
	while (!done) {
		sellist[nsels].sel_test = line;	/* set test word */
		while (*line && (*line != ':'))
			++line;
		if (! *line)
			fatal(BAD_SEL);
		*line++ = '\0';		/* truncate "test" word */
		sellist[nsels].sel_val = line;	/* set value word */
		while (*line && (*line != ',') &&
			(*line != ' ') && (*line != '\t'))
			++line;
		if (! *line)
			done = TRUE;
		if ((*line == ' ') || (*line == '\t'))
			done = TRUE;
		*line++ = '\0';		/* truncate "value" word */
		++nsels;
	}
	if (verify) {
		if (*var == 6)	/* fake for tty type name */
			*var = 'T' - 'A';
		printf("  SELECT %c:\n", (*var)+'A');
		for (i = 0; i < nsels; i++) {
			printf("\t %s -> %s\n", sellist[i].sel_test,
						sellist[i].sel_val);
		}
		return;
	}
	/*
	 * Search the select list and perform the first matching
	 * action.  It is NOT an error for no action to match.
	 */
	i = 0;
	while (i < nsels) {
		/*
		 * Look for special words on the "test" side:
		 * On the "test" side, only NUL and ELSE allowed.
		 */
		if ((strcmp("NUL", sellist[i].sel_test) == 0) &&
		    (vars[*var][0] == '\0')) {
			/*
			 * Saw "NUL" and the "set" variable was a NULL
			 * string.
			 */
			if (strcmp("NUL", sellist[i].sel_val) != 0) {
				if (ioctl(0, I_PUSH, sellist[i].sel_val) < 0) {
					strcpy(command.action, sellist[i].sel_val);
					fatal(BAD_PUSH);
				}
				else {
					++touched;
					return;	/* successful push */
				}
			}
			else 
				return;		/* On a null val do nothing */
		}
		else if (strcmp("ELSE", sellist[i].sel_test) == 0) {
			if (strcmp("NUL", sellist[i].sel_val) != 0) {
				if (ioctl(0, I_PUSH, sellist[i].sel_val) < 0) {
					strcpy(command.action, sellist[i].sel_val);
					fatal(BAD_PUSH);
				}
				else {
					++touched;
					return;	/* successful push */
				}
			}
			else 
				return;		/* On a null val do nothing */
		}
		else if ((strcmp("POP", sellist[i].sel_test) == 0) ||
			 (strcmp("END", sellist[i].sel_test) == 0)) {
			/*
			 * On pop or end on the "test" side, ERROR.
			 */
			fatal(BAD_SEL);
		}
		/*
		 * Now, test for other module names.
		 */
		if (strcmp(&(vars[*var][0]), sellist[i].sel_test) == 0) {
			/*
			 * found module name.  See what we should do.
			 */
			if (strcmp("END", sellist[i].sel_val) == 0)
				exit(0);	/* end immediately */
			else if (strcmp("POP", sellist[i].sel_val) == 0) {
				if (ioctl(0, I_POP, 0) == 0)
					++touched;
				return;
			}
			else if (strcmp("NUL", sellist[i].sel_val) == 0) {
				return;
			}
			else {
				if (ioctl(0, I_PUSH, sellist[i].sel_val) < 0) {
					strcpy(command.action, sellist[i].sel_val);
					fatal(BAD_PUSH);
				}
				else {
					++touched;
					return;	/* successful push */
				}
			}
		}
		++i; /* skip to next entry */
	}
}

/*
 * throw	Get the current "ioctl(TCGETA)" params and save them.
 */

throw()

{
	if (verify)
		printf("  THROW\n");
	if (th_active)
			fatal(DOUBLET);
	if (!verify) {
		if (ioctl(0, TCGETA, &tt) < 0)
			fatal(TC_FAIL);
	}
	th_active = TRUE;
}

/*
 * catch	Reset the params that were "thrown".  I.e., do a
 *		"ioctl(TCSETA)" from the params saved by a "throw".
 *		If nothing was thrown, do an "stty sane".
 */

catch()

{
	if (verify) {
		printf("  CATCH\n");
	}
	else {
		if (!th_active)		/* is there a throw active? */
			system("stty sane");
		else {
			if (ioctl(0, TCSETA, &tt) < 0)
				fatal(TC_FAIL);
		}
	}
	th_active = FALSE;	/* successful catch */
}

/*
 * parse	Find an entry in CONFFILE that matches our devt entry.
 *		call "runit" on each command found there.  We allow
 *		the "devt" in the configuration file to contain
 *		"|" characters between "real" devt names for entries
 *		that are executable for more than one devt type, e.g.,
 *			Latin1|Latin2	do_stuff...
 */

parse(file, flag)

	char *file;
	int flag;	/* flag true if we're parsing everything */
{
	register FILE *fp;
	register char *compart;
	register int one_done;
	char line[512], devt[128], chk[128];

	if ((fp = fopen(file, "r")) == NULL)
		fatal(NO_CONF);
	if (verify)
		flag = TRUE;
	if (opt_run) {
		flag = FALSE;		/* don't do everything */
		strcpy(devt, infoptr);	/* infoptr is entry to run */
	}
	else {
		if ((device_type == NULL) || (*device_type == '\0'))
			strcpy(devt, "ascii"); /* no devt entry, try "ascii" */
		else
			strcpy(devt, device_type);

#if 0
		/* Obsolete code. */
		if ((devt == NULL || *devt == '\0') && !flag)
			fatal(NO_DEVT);
#endif
	}
	/*
	 * Get a line.  Compare the "devt" part of it (in "chk") with
	 * the devt we're supposed to find.  If found, do "runit()", which
	 * calls the appropriate command routines.  It is an error for
	 * no entry to match.
	 */
	one_done = FALSE;
	while (getline(fp, line)) {
		compart = getdevt(line, chk);
		/*
		 * only use "devequal" if we're not running every entry...
		 */
		if ((flag == TRUE) || devequal(devt, chk)) {
			if (flag == TRUE)
				printf("\nDEVT '%s'\n", chk /* devt */ );
			one_done = TRUE;
			while (compart = getcom(&command, compart)) {
				runit();
			}
			if (!flag)
				return;
		}
	}
	if (!one_done)
		fatal(NO_MATCH);
}

/*
 * runit	Run the command in the "com" structure.
 */

runit()

{
	switch (command.comtype) {
		case CMD_SET:		set(); break;
		case CMD_POP:		pop(); break;
		case CMD_POPALL:	pop_all(); break;
		case CMD_POPTO:		popto(); break;
		case CMD_PUSH:		push(); break;
		case CMD_IF:		cmd_if(); break;
		case CMD_TEST:		test(); break;
		case CMD_RUN:		run(); break;
		case CMD_SEL:		select(); break;
		case CMD_THROW:		throw(); break;
		case CMD_CATCH:		catch(); break;
		default:		fatal(BAD_COM);
	}
}
