/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred. */
/*	Copyright (c) 1986,1989 AT&T	*/
/*	  All Rights Reserved  	*/

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/

#ident "$Header: main.c 1.1 90/01/11 $ @(#)/usr/bin/setterm 2.1 89/02/14 AT&T-UP"

#include <curses.h>			/* for terminfo */
#include <term.h>			/* for curses */
#include <sys/param.h>
#include <fcntl.h>			/* file control for open() */
#include <stropts.h>			/* Streams specific ioctls */

#include "header.h"

#ifdef DEBUG
char *CONFFILE =	"./lib/conf.file";
#else
char *CONFFILE =	"/usr/lib/setterm/conf.file";
#endif

extern struct cmd command;	/* in cmds.c */

char *prog, *infoptr;	/* pointer to our name; pointer to tty type */

int retcode;		/* for setupterm */
int verify;		/* for conf.file verification */
int touched = FALSE;	/* TRUE if we pushed or popped anything */
int opt_run;		/* the "unconditional run" option */
int opt_t = FALSE;	/* TRUE if option "-t" specified */

/*
 * MAIN	------------------------------------------------------------
 */

main(argc, argv)

	int argc;
	char **argv;
{
	char *getenv();
	register char *tmpc;	/* tmp pointer to CONFFILE name */

	prog = *argv;
	tmpc = CONFFILE;

#ifdef DEBUG
	verify = TRUE;
#else
	verify = FALSE;
#endif

	opt_run = FALSE;

	if (argc > 3)
		fatal(USAGE);		/* no return */
	if (argc == 1)
		infoptr = NULL;
	else {					/* argc == 2 or 3 */
		argv++;
		if (**argv == '-') {
			++*argv;
			/* for "-t" option, infoptr will be the name of the
			 * terminal to look up in terminfo.
			 */
			if (**argv == 't') {
				if (((argc == 2) && (*++*argv))
				|| ((argc == 3) && (!*++*argv && *++argv))) {
					infoptr = *argv;
					opt_t = TRUE;
				}
				else
					fatal(USAGE);
			}
			/* for "-x" option, infoptr will be the name of the
			 * devt entry in conf.file to run.
			 */
			else if (**argv == 'x') {
				if (((argc == 2) && (*++*argv))
				|| ((argc == 3) && (!*++*argv && *++argv))) {
					infoptr = *argv;
					opt_run = TRUE;
				}
				else
					fatal(USAGE);
			}
			else if (**argv == 'v')
				verify = TRUE;
			else
				fatal(USAGE);
		}
		else
			fatal(USAGE);
	}
	/*
	 * Do some special-case checking for XT and SXT.
	 */

	if (!opt_run && !verify)
		check_tty();

	/*
	 * Get the tty info and exit if bad.
	 */

	if (!opt_run && !verify) {
		setupterm(infoptr, 1, &retcode);
		if (retcode != 1)
			fatal(NO_INFO);
	}

	/*
	 * open the conf.file and find an entry for the device type.
	 * parse the command line and run each command.  If user set
	 * $SETTLIB in the environment, use the named file to override
	 * the standard CONFFILE.
	 */

	if (tmpc = getenv("SETTLIB"))
		CONFFILE = tmpc;
	parse(CONFFILE, verify);
	exit(0);

}

/*
 * check_tty	Check to make sure we're NOT an SXT or XT device.
 *		The XT because it only supports ASCII, SXT because
 *		the user must run this BEFORE using "shl".  Anything
 *		else is assumed O.K.
 *
 *		We know that the SXT driver works by pushing line
 *		disciplines BELOW it in the stream, and hence they
 *		are inaccessible to us.
 *
 *		This code is obsolete unless you have a specific need
 *		to disallow them.  We may need to run this over any
 *		driver, and it should be O.K.
 */

check_tty()

{
#if 0
	register char *ttyp;
	char *ttyname();

	if (verify)		/* don't bother during file verification */
		return;
	ttyp = ttyname(0);	/* get the name of our stdin device. */
	if (!ttyp)
		fatal(NOT_TTY);
	if (strncmp("/dev/sxt", ttyp, 8) == 0)
		/* naughty naughty... */
		fatal(NO_SXT);
	if (strncmp("/dev/xt", ttyp, 7) == 0)
		fatal(NO_XT);
#else
	return;
#endif
}

/*
 * fatal	Print error message and exit.  On some errors, try
 *		error recovery.  Error recovery consists of calling
 *		ourselves recursively as "setterm -x Error".  We
 *		only try error recovery if !verify.
 */

fatal(no)

	int no;
{
	char *getenv();
	register int rec;	/* error recovery flag */

	rec = 0;
	switch (no) {
	case USAGE:	fprintf(stderr, "Usage: %s [ -t term_name ]\n", prog);
			fprintf(stderr, "   or: %s [ -x conf_entry]\n", prog);
			fprintf(stderr, "   or: %s [ -v ]\n", prog);
			break;
	case NO_INFO:	fprintf(stderr,
				"%s: terminfo entry not found for %s.\n",
				prog,infoptr);
			break;
	case TC_FAIL:	fprintf(stderr, "%s: TCSET/TCGET failed.\n", prog);
			++rec;
			break;
	case NO_SXT:	fprintf(stderr,
				"%s: cannot be run under shell layers.\n",
				prog);
			fprintf(stderr,
				"Please run %s before invoking \"shl\".\n",
				prog);
			break;
	/*
	 * XT may or may not be fatal.  See above.
	 */
	case NO_XT:	fprintf(stderr,
				"%s: cannot be run under layers.\n", prog);
			break;
	case NOT_TTY:	fprintf(stderr,
				"%s: \"ttyname\" failed to return a name.\n",
				prog);
			break;
	case BAD_SET:	fprintf(stderr,
				"%s: illegal \"set\" variable.\n", prog);
			++rec;
			break;
	case BAD_COM:	fprintf(stderr, "%s: illegal command type.\n", prog);
			++rec;
			break;
	case NO_CONF:	fprintf(stderr, "%s: can't open config file \"%s\"\n",
				prog, CONFFILE);
			break;
	case BAD_PUSH:	fprintf(stderr,
				"%s: can't push \"%s\"; not configured?\n",
				prog, command.action);
			++rec;
			break;
	case BAD_IF:	fprintf(stderr, "%s: bad syntax in \"if\" command.\n",
					prog);
			++rec;
			break;
	case BAD_TEST:	fprintf(stderr,
				"%s: bad syntax in \"test\" command.\n",
				prog);
			++rec;
			break;
	case BAD_RUN:	fprintf(stderr, "%s: bad \"run\" command.\n", prog);
			++rec;
			break;
	case BAD_SEL:	fprintf(stderr,
				"%s: bad syntax in \"select\" command.\n",
				prog);
			++rec;
			break;
	case DOUBLET:	fprintf(stderr,
				"%s: two throws, no catch.\n", prog);
			++rec;
			break;
	case NO_MATCH:	fprintf(stderr, "%s: no matching entry found.\n",
				prog);
			break;
	case NO_EXEC:	fprintf(stderr, "%s: RUN command 'exec' failed.\n",
				prog);
			++rec;
			break;
	case NO_FORK:	fprintf(stderr, "%s: fork failed.\n", prog);
			break;
#if 0
/*
 * Obsolete.  Since all ttys are Streamified, it's reasonable to request
 * things even for ASCII terminals.  See what happens in "parse" when
 * there's no "devt" for a terminal.
 */
	case NO_DEVT:	fprintf(stderr, "%s: device is ASCII\n", prog);
			break;
#endif
	default:	fprintf(stderr,
				"%s: internal error (code %d).\n", no);
	}
	if (!rec) exit(1);	/* no recoverable errors */
	if (verify) exit(1);	/* verify option only */
	if (!touched) exit(1);	/* TRUE if the stream was "touched" */
	/*
	 * try recovery... "prog" is our argv[0].
	 */
	if (!strcmp(infoptr, "Error"))
		fprintf(stderr, "%s: Recursion in error recovery.\n", prog);
	else {
		execlp(prog, prog, "-x", "Error", (char *) 0);
		fprintf(stderr, "%s: Error recovery failed.\n", prog);
	}
	exit(1);
}
