/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ifndef lint
static char rcsid[] = "$Header: time.c 1.5 90/08/03 $";
#endif

/*
 * time
 */
#include <errno.h>
#include <sys/types.h>
#include <stdio.h>
#include <signal.h>
#include <sys/param.h>
#include <sys/dir.h>
#include <sys/psl.h>
#include <sys/pte.h>
#include <sys/user.h>
#include <sys/vm.h>
#include <sys/reg.h>
#include <sys/plocal.h>
#include <sys/timer.h>
#include <sys/procstats.h>
#include <sys/proc.h>

#define	TRUE	1
#define	FALSE	0

struct process {
	struct process *p_next;
	struct pt_stop	p_s;
	struct process_stats	p_stats;
	int		p_parent;
	int		p_stopped;
	timeval_t	p_start;
	timeval_t	p_finish;
	char		p_name[MAXCOMLEN];
};
#define	p_pid		p_s.ps_pid
#define p_reason	p_s.ps_reason

struct {
	long user;
	long sys;
	long childuser;
	long childsys;
} buffer;

struct	process	proclist;/* process structure pointer */
timeval_t	before;		/* time command started */
timeval_t	after;		/* time command finished */
int	pid;
int	aflag;		/* time all procs */
int	vflag;		/* give verbose fork/exec/exec notification */
int	rflag;		/* give resource usage */
int	nkids;		/* number of children */
int	ndone;		/* waited for children */

double sex();
struct process *findp();
char *malloc();
char *strcpy();
char *strrchr();
char *memset();
chldhandler();

main(argc, argv)
	int argc;
	char **argv;
{
	extern char	*optarg;
	extern int	optind;
	int	c, opterr = 0;
	struct	process_stats cpst1, cpst2, cpstats;
	timeval_t timeval;
	int status;
	char *myname, *m;

	myname = argv[0];
#ifdef NOWAY
	while (argc > 1 && argv[1][0] == '-') {
		switch(argv[1][1]) {
		case 'a':	/* all procs */
			aflag = TRUE;
			break;
		case 'v':	/* verbose */
			vflag = TRUE;
			break;
		case 'r':	/* give resource usage */
			rflag = TRUE;
			break;
		}
		argc--;
		argv++;
	}
	if (argc<=1)
		exit(0);
#endif /* NOWAY */
	while ((c=getopt(argc, argv, "arv")) != EOF) {
		switch(c) {
		case 'a':	/* all procs */
			aflag = TRUE;
			break;
		case 'r':	/* give resource usage */
			rflag = TRUE;
			break;
		case 'v':	/* verbose */
			vflag = TRUE;
			break;
		}
	}

	if (argc <= optind) {
		exit(0);
	}
	m = strrchr(myname, '/');
	if (m != NULL)
		myname = ++m;
	if (strcmp(myname, "ptime") == 0)
		aflag = TRUE;
	if (opterr) {
		fprintf(stderr,"usage: %s [-arv] command\n", myname);
		exit(2);
	}
	if (aflag == TRUE)
		exit(timeall(&argv[optind]));
	(void) get_process_stats(&before, PS_SELF, 0, &cpst1);
	pid = fork();
	if (pid < 0) {
		perror("time");
		exit(1);
	}
	if (pid == 0) {
		execvp(argv[optind], &argv[optind]);
		perror(argv[optind]);
		exit(1);
	}
	sigset(SIGINT, SIG_IGN);
	sigset(SIGQUIT, SIG_IGN);
	while (wait(&status) != pid)
		;
	(void) get_process_stats(&after, PS_SELF, 0, &cpst2);
	subtract_process_stats(&cpst2, &cpst1, &cpstats);
	if ((status&0377) != 0)
		fprintf(stderr, "Command terminated abnormally.\n");
	after.tv_sec -= before.tv_sec;
	after.tv_usec -= before.tv_usec;
	if (after.tv_usec < 0)
		after.tv_sec--, after.tv_usec += 1000000;
	fprintf(stderr, "\n");
	printt("real", &after);
	printt("user", &cpstats.ps_utime);
	printt("sys ", &cpstats.ps_stime);
	fprintf(stderr, "\n");
	if (rflag)
		printr(&cpstats);
	exit (status>>8);
}

printt(s, tv)
	char *s;
	timeval_t *tv;
{
	fprintf(stderr, "%s%9d.%01d\n", s, tv->tv_sec, tv->tv_usec/100000);
}

timeall(a)
	char **a;
{
	int status;

	(void) sigset(SIGINT, SIG_IGN);
	(void) sigset(SIGTERM, SIG_IGN);
	(void) sigset(SIGCHLD, chldhandler);
	(void) mptrace(XPT_MPDEBUGGER, 0, 0, 0);

	(void) get_process_stats(&before, PS_SELF, 0, 0);
	forkcmd(a);
	waitall();
	(void) get_process_stats(&after, PS_SELF, 0, 0);
	printstats();
	exit(status>>8);
}

forkcmd(a)
	char **a;
{
	int p = fork();
	if (p < 0) {
		perror("time: fork");
		exit(1);
	}
	if (p == 0) {
		(void) sigset(SIGINT, SIG_DFL);
		(void) sigset(SIGTERM, SIG_DFL);
		(void) sigset(SIGCHLD, SIG_DFL);
		execvp(a[0], &a[0]);
		perror(a[0]);
		_exit(127);
	}
}

waitall()
{
	(void) sighold(SIGCHLD);
	while (nkids == 0 || nkids > ndone)
		(void) sigpause(SIGCHLD);
	(void) sigrelse(SIGCHLD);
}

struct proc proc;

chldhandler()
{
	struct pt_stop	status;
	register struct process *p;
	sigset_t	sigset;
	timeval_t now;

	sigfillset(&sigset);

	while(mptrace(XPT_STOPSTAT, 0, (int) &status, 0) != -1) {

		(void) get_process_stats(&now, PS_SELF, 0, 0);
		p = findp(status.ps_pid);
		p->p_reason = status.ps_reason;
		p->p_stopped = TRUE;

		if (vflag) {
			char *why;
			switch(p->p_reason) {
			case PTS_FORK:
				why = "fork";
				break;
			case PTS_EXEC:
				why = "exec";
				break;
			case PTS_EXIT:
				why = "exit";
				break;
			default:
				why = "signal";
				break;
			}
			fprintf(stderr, "%d %s\n", p->p_pid, why);
			fflush(stderr);
		}

		switch(p->p_reason) {

		case PTS_FORK:
			(void) mptrace(XPT_RPROC, p->p_pid, (int)&proc, 0);
			p->p_parent = proc.p_ppid;
			(void) strcpy(p->p_name, findp(p->p_parent)->p_name);
			(void) mptrace(XPT_SETSIGMASK, p->p_pid, &sigset, 0);
			p->p_start = now;
			break;

		case PTS_EXEC:
			getname(p);
			(void) mptrace(XPT_SETSIGMASK, p->p_pid, &sigset, 0);
			if (p->p_parent == 0)
				p->p_start = now;
			break;

		case PTS_EXIT:
			p->p_finish = now;
			getru(p);
			ndone++;
			break;

		default:
			fprintf(stderr, "got signal for proc %d\n", p->p_pid);
			break;
		}

		(void) mptrace(XPT_RPROC, p->p_pid, (int)&proc, 0);
		p->p_parent = proc.p_ppid;
	}

	for (p = proclist.p_next; p != NULL; p = p->p_next) {
		if (p->p_stopped) {
			(void) ptrace(PT_CONTSIG, p->p_pid, 1, 0);
			p->p_stopped = FALSE;
		}
	}
}


struct process *
findp(pid)
	register int pid;
{
	register struct process *p;

	p = &proclist;
	if (p->p_next == NULL) {
nfound:
		p->p_next = (struct process *) malloc(sizeof(struct process));
		p = p->p_next;
		if (p == NULL) {
			fprintf(stderr, "time: out of memory.\n");
			exit(1);
		}
		memset((char *)p, 0, sizeof(struct process));
		p->p_pid = pid;
		++nkids;
		return p;
	}
	for (;;) {
		p = p->p_next;
		if (p->p_pid == pid)
			return(p);
		if (p->p_next == NULL)
			break;
	}
	goto  nfound;
}

#define PSTATSIZE (sizeof(struct process_stats)/sizeof(int) + 1)

getru(p)
	struct process *p;
{
	int buf[PSTATSIZE];
	int offset;
	int i;

	offset = (int) &(((struct user *)0)->u_pst);
	for(i = 0; i < PSTATSIZE; i++) {
		buf[i] = ptrace(PT_RUSER, p->p_pid, offset, 0);
		offset += sizeof(int);
	}
	p->p_stats = *(struct process_stats *)buf;
}

getname(p)
	struct process *p;
{
	register int offset;
	register int *buf;
	register char *cp;
	register int i, j;

	offset = (int) (((struct user *)0)->u_comm);
	buf = (int *) p->p_name;
	cp = p->p_name;
	for (i = 0; i < DYNIX_MAXNAMLEN / sizeof(int) + 1; i++) {
		*buf++ = ptrace(PT_RUSER, p->p_pid, offset, 0);
		offset += sizeof(int);
		for (j = 0; j < sizeof(int); j++) {
			if (*cp++ == '\0')
				return;
		}
	}
}

printr(pstat)
	struct process_stats *pstat;
{
	fprintf(stderr, "%6s%9s%9s%7s%7s%7s%5s%7s%7s\n",
		"MaxRSS", "MajorPF", "MinorPF", "Swaps", "blkI", "blkO", "Nsig", "Vcsw", "Icsw");
#define	R(X) (pstat->ps_/**/X)
	fprintf(stderr, "%6d %8d %8d %6d %6d %6d %4d %6d %6d\n",
		R(maxrss), R(pagein), R(reclaim), R(swap),
		R(bread), R(bwrite), R(signal), R(volcsw), R(involcsw));
}

double
sex(tv)
	timeval_t tv;
{
	return(tv.tv_usec / 1e6 + tv.tv_sec);
}

printstats()
{
	struct process *p;
	struct process_stats *pstat;
	double totu = 0.0, tots = 0.0, tote;
	double u, s, e;

	fprintf(stderr, "%5s %5s %7s %7s %7s %s\n",
		"pid", "ppid", "utime", "stime", "elapsed", "name");
	for (p = proclist.p_next; p != NULL; p = p->p_next) {
		u = sex(p->p_stats.ps_utime);
		s = sex(p->p_stats.ps_stime);
		e = sex(p->p_finish) - sex(p->p_start);
		fprintf(stderr, "%5d %5d %7.1f %7.1f %7.1f %s\n", 
			p->p_pid, p->p_parent, u, s, e, p->p_name);
		totu += u;
		tots += s;
	}
	tote = sex(after) - sex(before);
	fprintf(stderr, "%5s %5s %7.1f %7.1f %7.1f\n", "total", "", totu, tots, tote);
	if (rflag) {
		fprintf(stderr, "\n%5s %5s %6s%9s%9s%7s%7s%7s%5s%7s%7s\n",
			"pid", "ppid", "MaxRSS", "MajorPF", "MinorPF", "Swaps", "blkI", "blkO", "Nsig", "Vcsw", "Icsw");
		for (p = proclist.p_next; p != NULL; p = p->p_next) {
			pstat = &p->p_stats;
			fprintf(stderr, "%5d %5d %6d %8d %8d %6d %6d %6d %4d %6d %6d\n",
				p->p_pid, p->p_parent,
				R(maxrss), R(pagein), R(reclaim), R(swap),
				R(bread), R(bwrite), R(signal), R(volcsw), R(involcsw));
		}
	}
}

/*
 * Compute the difference in process stats.
 */

subtract_process_stats(p2, p1, pr)
struct process_stats *p2, *p1, *pr;
{
	pr->ps_utime.tv_sec = p2->ps_utime.tv_sec - p1->ps_utime.tv_sec;
	pr->ps_utime.tv_usec = p2->ps_utime.tv_usec - p1->ps_utime.tv_usec;
	if (pr->ps_utime.tv_usec < 0) {
		pr->ps_utime.tv_sec -= 1;
		pr->ps_utime.tv_usec += 1000000;
	}
	pr->ps_stime.tv_sec = p2->ps_stime.tv_sec - p1->ps_stime.tv_sec;
	pr->ps_stime.tv_usec = p2->ps_stime.tv_usec - p1->ps_stime.tv_usec;
	if (pr->ps_stime.tv_usec < 0) {
		pr->ps_stime.tv_sec -= 1;
		pr->ps_stime.tv_usec += 1000000;
	}

	pr->ps_maxrss = p2->ps_maxrss - p1->ps_maxrss;
	pr->ps_pagein = p2->ps_pagein - p1->ps_pagein;
	pr->ps_reclaim = p2->ps_reclaim - p1->ps_reclaim;
	pr->ps_zerofill = p2->ps_zerofill - p1->ps_zerofill;
	pr->ps_pffincr = p2->ps_pffincr - p1->ps_pffincr;
	pr->ps_pffdecr = p2->ps_pffdecr - p1->ps_pffdecr;
	pr->ps_swap = p2->ps_swap - p1->ps_swap;
	pr->ps_syscall = p2->ps_syscall - p1->ps_syscall;
	pr->ps_volcsw = p2->ps_volcsw - p1->ps_volcsw;
	pr->ps_involcsw = p2->ps_involcsw - p1->ps_involcsw;
	pr->ps_signal = p2->ps_signal - p1->ps_signal;
	pr->ps_lread = p2->ps_lread - p1->ps_lread;
	pr->ps_lwrite = p2->ps_lwrite - p1->ps_lwrite;
	pr->ps_bread = p2->ps_bread - p1->ps_bread;
	pr->ps_bwrite = p2->ps_bwrite - p1->ps_bwrite;
	pr->ps_phread = p2->ps_phread - p1->ps_phread;
	pr->ps_phwrite = p2->ps_phwrite - p1->ps_phwrite;
}
