
/* 
 * $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

/*	Copyright (c) 1984 AT&T	*/
/*	  All Rights Reserved  	*/

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/
#ident "$Header: args.c 1.1 90/01/19 $@(#)args.c	1.1 89/12/01"
/*
 * Function formal parameters handling.
 * Also modify "main" parameters if necessary.
 */
#include "xs.h"
#include "names.h"
#include "stack.h"
#include "types.h"
#include "anal.h"
#include "args.h"
#include "mfile1.h"
#include "xdefs.h"

extern char *gettxt();
/*
 * Memory allocation saving
 * buffers are never freed, they are put in free lists, according to their
 * types.
 */
static struct args_buf {
  struct args_buf *args_next;
} *args_start;

typedef struct args {
  char *ar_name;
  long ar_off;
  int ar_len;
  TYPEP ar_decl;
  struct args *ar_next;
} ARGS;

static ARGS *ar_func;
static int ar_nr;

/*
 * Free the args name list
 */
void
args_free()
{
  ARGS *ptr;
  struct args_buf *ptr2;
  
  while (ar_func != 0){
    ptr = ar_func;
    ar_func = ar_func->ar_next;
    /*
     * Insert arg in free list
     */
    ptr2 = (struct args_buf *)ptr;
    ptr2->args_next = args_start;
    args_start = ptr2;
  }
  ar_nr = 0;
}

/*
 * Add a name to the arg list
 * Must append it at the end. Also check that an argument is only used
 * once.
 */
void
args_add(name)
  TOKEN name;
{
  ARGS *ptr = ar_func, *prev = ptr;
  while (ptr != 0){
    if (strcmp(name.t_name, ptr->ar_name) == 0){
      uerror(gettxt("xs:17","%s used more than once"));
      return;
    }
    prev = ptr;
    ptr = ptr->ar_next;
  }
  ar_nr++;
  
  if (args_start != 0){
    /*
     * use previously free'd arg
     */
    ptr = (ARGS *)args_start;
    args_start = args_start->args_next;
  } else {
    /*
     * Allocate a new arg
     */
    if ((ptr = (ARGS *)malloc(sizeof(ARGS))) == 0)
      fatal(gettxt("xs:18","Out of memory (args.c (1))"));
  }
  ptr->ar_name = name.t_name;
  ptr->ar_off = name.t_off;
  ptr->ar_len = name.t_len;
  ptr->ar_next = 0;
  ptr->ar_decl = type_null;
  if (ar_func == 0)
    ar_func = ptr;
  else
    prev->ar_next = ptr;
}

/*
 * Determine if a name is in the argument table
 */
static ARGS *
args_isin(name)
  char *name;
{
  ARGS *ptr = ar_func;
  while (ptr != 0){
    if (strcmp(name, ptr->ar_name) == 0)
      return ptr;
    ptr = ptr->ar_next;
  }
  return 0;
}

/*
 * Declare function arguments
 */
void
args_decl(attr, dlist)
  TOKEN attr, dlist;
{
  TYPEP dl, de, aux;
  ARGS *ar;
  dl = dlist.t_type;
  
  if (is_null_type(attr.t_type)){
    uerror(gettxt("xs:19","missing attribute"));
    return;
  }
  while (!is_null_type(dl)){
    de = dl_de(dl);
    if ((ar = args_isin(ty_name(de))) == 0)
      uerror(gettxt("xs:20","%s ignored"), ty_name(de));
    else if (!is_null_type(ar->ar_decl))
      uerror(gettxt("xs:21","%s redeclaration"), ty_name(de));
    else {
      aux = types_new(NODE_IDENT, 0, ty_name(de));
      ar->ar_decl = aux;
      ty_name_new(aux, ty_name(de));
      id_at_new(aux, attr.t_type);
      id_de_new(aux, de);
      id_re_new(aux, type_null);
      id_flag_new(aux, IDENT_DEF);
    }
    dl = dl_next(dl);
  }
}

/*
 * End arguments handling : declare remaining attributes as 'int'
 * and free args table
 */
void
args_def()
{
  TYPEP aux, aux1, aux2;
  ARGS *ptr;
  struct args_buf *ptr2;
  
  while (ar_func != 0){
    if (is_null_type(ar_func->ar_decl)){
      aux = types_new(NODE_IDENT, 0, ar_func->ar_name);
      aux1 = types_new(NODE_ATTR, 0, 0);
      aux2 = types_new(NODE_DECL, 0, 0);
      ty_name_new(aux2, ar_func->ar_name);
      de_flag_new(aux2, DECL_NAME);
      de_next_new(aux2, type_null);
      at_cl_new(aux1, type_null);
      at_ty_new(aux1, types_systype(INT));
      id_at_new(aux, aux1);
      id_de_new(aux, aux2);
      id_re_new(aux, type_null);
      id_flag_new(aux, IDENT_DEF);
      ty_name_new(aux, ar_func->ar_name);
    }
    ptr = ar_func;
    ar_func = ar_func->ar_next;
    ptr2 = (struct args_buf *)ptr;
    ptr2->args_next = args_start;
    args_start = ptr2;
  }
}

static bool is_main = FALSE;
static char *ar_argv;
static bool is_argv = FALSE;

/*
 * do_main : Add parameters to main if necessary
 * Needed to be able to call the strings initialization with the command
 * name to retrieve the messages
 */
void
do_main(decl)
  TOKEN decl;
{
  TYPEP f, aux, aux1, aux2, aux3, aux4;
  ARGS *ptr;
  bool is_argc = FALSE;
  f = decl.t_type;
  
  if (strcmp(ty_name(f), "main") == 0){
    is_main = TRUE;
    
    ptr = ar_func;
    if (ar_nr >= 2)
      /*
       * main has at least 2 parameters. use them.
       * 'argv' is the second one
       */
      ar_argv = ar_func->ar_next->ar_name;
    else {
      is_argv = TRUE;
      if (ar_nr == 0){
        /*
         * Create the argc param.
         * args_def will add the declaration
         */
        if (args_start != 0){
          ptr = (ARGS *)args_start;
          args_start = args_start->args_next;
        } else {
          if ((ptr = (ARGS *)malloc(sizeof(ARGS))) == 0)
            fatal(gettxt("xs:22","Out of memory (args.c (2))"));
        }
        ptr->ar_name = na_unique();
        ptr->ar_off = decl.t_off + decl.t_len - 1;
        ptr->ar_len = strlen(ptr->ar_name);
        ptr->ar_decl = type_null;
        ptr->ar_next = 0;
        ar_nr++;
        ar_func = ptr;
        act_e_append(ptr->ar_off, ptr->ar_name);
        is_argc = TRUE;
      }
      /*
       * Create the argv param
       * This requires a special declaration (char *argv[])
       * attribute is no class(type_null), type char (systype(CHAR)
       * declarator is unsized array (DECL_ARRAY) of pointers (DECL_PTR)
       */
      if (args_start != 0){
        ptr = (ARGS *)args_start;
        args_start = args_start->args_next;
      } else {
        if ((ptr = (ARGS *)malloc(sizeof(ARGS))) == 0)
          fatal(gettxt("xs:23","Out of memory (args.c (3))"));
      }
      ptr->ar_name = na_unique();
      ptr->ar_off = 0;
      ptr->ar_len = 0;
      ptr->ar_decl = type_null;
      ptr->ar_next = 0;
      ar_func->ar_next = ptr;
      ar_nr++;
      act_e_append(is_argc ? ar_func->ar_off : ar_func->ar_off + 
                   ar_func->ar_len, ", %s", ptr->ar_name);
      aux = types_new(NODE_IDENT, 0, ptr->ar_name);
      aux1 = types_new(NODE_ATTR, 0, 0);
      aux2 = types_new(NODE_DECL, 0, 0);
      ty_name_new(aux2, ptr->ar_name);
      de_flag_new(aux2, DECL_NAME);
      aux3 = types_new(NODE_DECL, 0, 0);
      de_flag_new(aux3, DECL_ARRAY);
      aux4 = types_new(NODE_DECL, 0, 0);
      de_flag_new(aux4, DECL_PTR);
      de_next_new(aux4, type_null);
      de_next_new(aux3, aux4);
      de_next_new(aux2, aux3);
      at_cl_new(aux1, type_null);
      at_ty_new(aux1, types_systype(CHAR));
      id_at_new(aux, aux1);
      id_de_new(aux, aux2);
      id_re_new(aux, type_null);
      ty_name_new(aux, ptr->ar_name);
      ar_argv = ptr->ar_name;
      ptr->ar_decl = aux;
    }
  }
}

/*
 * do_main2 : Add the declaration for the argv param.
 */
void
do_main2(lc)
  TOKEN lc;
{
  if (is_main && is_argv){
    act_e_append(lc.t_off - 1, "char *%s[];\n", ar_argv);
    is_argv = FALSE;
  }
}

/*
 * do_main3 : Add the call to the initialisation routine
 * 3 levels on stack : global vars, args, local vars
 */
void
do_main3(pos)
  long pos;
{
  if (is_main && types_nelem() == 3){
    act_e_append(pos, "%s(%s[0]);\n", init_strings, ar_argv);
    is_main = FALSE;
  }
}
