
/* 
 * $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

/*	Copyright (c) 1984 AT&T	*/
/*	  All Rights Reserved  	*/

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/
#ident "$Header: init.c 1.1 90/01/19 $@(#)init.c	1.1 89/12/01"
/*
 * Initialized vars scanning and replacement
 * Scanning algorithm is copied from Unix V.3 C compiler (common/pftn.c)
 * Use an auxiliary stack to keep track of the types tree scanning
 */
#include "xs.h"
#include "names.h"
#include "stack.h"
#include "types.h"
#include "anal.h"
#include "init.h"
#include "body.h"
#include "xdefs.h"

extern char *gettxt();
/*
 * Memory allocation saving
 * buffers are never freed, they are put in free lists, according to their
 * types.
 */
static struct in_buf {
  struct in_buf *in_next;
} *in_start;

#define top		(*(TYPEP *)st_top(st_init))
#define top2		(*(TYPEP *)st_elem(st_init, nelem - 2))
#define nelem		st_nelem(st_init)
#define init_elem(lev)	(*(TYPEP *)st_elem(st_init, lev))

static STACK *st_init;
static bool is_name, is_string;
static TYPEP curident;
static char loc_buf[EXPRLEN];
static TYPEP SCALAR;
static TOKEN loc_string;


reset_init()
{
  SCALAR = type_null;
}

/*
 * Init_push is needed to allocate the TYPEP whose address will be finally
 * pushed.
 */
 
init_push(elem)
  TYPEP elem;
{
  TYPEP *ptr;
  if (in_start != 0){
    /*
     * Use a previously free'd TYPEP
     */
    ptr = (TYPEP *)in_start;
    in_start = in_start->in_next;
  } else {
    /*
     * Allocate a new TYPEP
     */
    ptr = (TYPEP *)malloc(sizeof(TYPEP));
    if (ptr == 0)
      fatal("Out of memory (init.c (1))");
  }
  ptr->ty_level = elem.ty_level;
  ptr->ty_offset = elem.ty_offset;
  ptr->ty_tab = elem.ty_tab;
  st_push(st_init, (char *)ptr);
}

static
pop()
{
  struct in_buf *ptr;
  /*
   * Insert top typep in free list
   */
  ptr = (struct in_buf *)st_pop(st_init);
  ptr->in_next = in_start;
  in_start = ptr;
}

/*
 * Start an initialisation process
 * SCALAR is a fake node that simulates a scalar entry on top of stack
 */
void
init_start(id)
  TYPEP id;
{
  if (is_null_type(id))
    return;
  if (is_null_type(SCALAR))
    SCALAR = g_types_new(NODE_REPLACE, 0, (char *)0);
  curident = id;
  is_init = TRUE;
  do_init = TRUE;
  is_name = is_string = FALSE;
  st_init = st_creat();
  push(de_ty(id_de(id)), de_next(id_de(id)));
}

/*
 * Found a left brace in initializer.
 * Go down one level only and bump flag at new level
 */

void
init_lbrace()
{
  if (!do_init)
    return;
  nextelem(TRUE);
  if (ty_type(top) == NODE_DECL)
    de_fl_new(top, de_fl(top) + 1);
  else
    su_fl_new(top, su_fl(top) + 1);
}

/*
 * Found a right brace
 * Back up stack until a flagged { is found
 */
void
init_rbrace()
{
  TYPEP elem;
  int fl;
  if (!do_init)
    return;
  for (;;){
    if (ty_type(elem = top) == NODE_DECL && (fl = de_fl(elem))){
      de_fl_new(elem, fl - 1);
      endelem();
      return;
    } else if (fl = su_fl(elem)){
      su_fl_new(elem, fl - 1);
      endelem();
      return;
    }
    if (nelem == 0)
      return;
    pop();
  }
}

void      
init_end()
{
  if (!do_init)
    return;
  is_init = FALSE;
  do_init = FALSE;
  st_free(st_init);
  st_init = 0;
}

void
init_expr(expr)
  TOKEN expr;
{
  TYPEP aux, ptr;
  char *make_name();
  bool is_size = FALSE;
  int str_nr, index;
   
  if (!do_init)
    return;

  nextelem(FALSE);

  if (!do_init)
    return;
   
  if (is_name || is_string){
    /*
     * Initialization expression contains a string or a name that was
     * initialized previously here
     */
    if (is_name){
      /*
       * depends on a name : actual initialisation must be performed
       * dynamically
       */
      id_flag_new(curident, id_flag(curident) | IDENT_STRUSE);
      act_e_replace(expr.t_off, expr.t_len, " 0");
    } else {
      /*
       * Init by a string : behavior depends on the identifier type
       */
      TOKEN taux;
      taux.t_type = at_ty(id_at(curident));
      if (is_char(taux)){
        /*
         * Identifier is of character type.
         * Check declarator
         */
        TYPEP de, ty, attr;
        de = id_de(curident);
        if (!is_null_type(de = de_next(de)) && de_flag(de) == DECL_ARRAY &&
             is_null_type(de_next(de))){
          TYPEP ptr;
          /*
           * Found a name followed by an unsized array with no next element
           * and type is character : this is an unsized characters array
           */
          id_flag_new(curident, id_flag(curident) | IDENT_ARRAY|IDENT_STRINIT);
          act_e_replace(de_off(de), de_len(de), " *%s", ty_name(curident));
          ec_replace(curident);
        } else if (is_null_type(de_next(de = id_de(curident))) && 
            ty_flag(ty = at_ty(id_at(curident))) == TYPE_TYPEDEF && 
            !is_null_type(de = de_next(last_de(ty))) &&
            de_flag(de) == DECL_ARRAY && is_null_type(de_next(de))){
          TYPEP ptr;
          /*
           * Same case as above for a typedef
           */
          id_flag_new(curident, id_flag(curident) | IDENT_ARRAY|IDENT_STRINIT);
          de = id_de(curident);
          attr = id_at(curident);
          act_e_replace(at_off(attr), at_len(attr), "\nchar");
          act_e_replace(de_off(de), de_len(de), " *%s", ty_name(curident));
	  ec_replace(curident);
	}
      }
      if (ty_type(top) == NODE_REPLACE && nelem > 1 && 
          ty_type(ptr = top2) == NODE_DECL && de_flag(ptr) == DECL_SIZE){
        /*
         * Identifier is a sized array. Replace by an empty table and
         * get string number
         */
        is_size = TRUE;
        act_e_replace(expr.t_off, expr.t_len, " { 0 }");
        str_nr = string_size(loc_string, index = de_in_max(ptr));
	pop();
      } else {
        /*
         * Identifier is a pointer
         */
        id_flag_new(curident, id_flag(curident) | IDENT_STRINIT);
        act_e_replace(expr.t_off, expr.t_len, " 0");
        if (id_flag(curident) & IDENT_ARRAY)
          /*
           * Generate replacement function call, do not add len address param
           * (will come later (body.c))
           */
          do_string(loc_string, FALSE);
        else
          /*
           * Generate replacement function call, with all parameters
           */
          do_string(loc_string, TRUE);
      }
    }
    is_name = is_string = FALSE;

    /*
     * Build initialization expression according to type
     * Replacement lists are scanned at compound statement end or at
     * file end (see body.c)
     */
    aux = types_new(NODE_REPLACE, 0, 0);
    re_na_new(aux, make_name(is_size));
    re_off_new(aux, expr.t_off);
    re_len_new(aux, expr.t_len);
    re_zero_new(aux, loc_string.t_off);
    re_nr_new(aux, str_nr);
    re_next_new(aux, type_null);
    re_si_new(aux, 0);
    if (is_size){
      /*
       * Replacement for a sized array
       */
      re_flag_new(aux, REPL_SIZE);
      re_alen_new(aux, index);
    } else if (id_flag(curident) & IDENT_ARRAY)
      /*
       * Replacement for an unsized array
       */
      re_flag_new(aux, REPL_ARRAY);
    else
      /*
       * Replacement for anything else (pointers mostly)
       */
      re_flag_new(aux, REPL_SCALAR);

    if (is_null_type(ptr = id_re(curident)))
      id_re_new(curident, aux);
    else {
      while (!is_null_type(re_next(ptr)))
        ptr = re_next(ptr);
      re_next_new(ptr, aux);
    }
    /*
     * Add replace node in proper list
     */
    if (is_global(curident))
      /*
       * Global identifier (static or extern)
       */
      repl_add_global(curident);
    else if (id_flag(curident) & IDENT_STATIC)
      /*
       * Local static identifier
       */
      repl_add_static(-1, curident);
    else
      /*
       * Local dynamic identifier
       */
      repl_add_local(-1, curident);
  }
  endelem();
}

static
nextelem(onelevel)
  bool onelevel;
{
  int flag;
  TYPEP ttop;
  ttop = top;

  do {
    if (nelem == 0)
      cerror("confused nextelem");
    if (ty_type(ttop) == NODE_DECL && ((flag = de_flag(ttop)) == DECL_SIZE ||
    	flag == DECL_ARRAY)){
      /*
       * Scanning an array
       */
      if (flag == DECL_SIZE && de_in_cur(ttop) >= de_in_max(ttop)){
        uerror(gettxt("xs:34","too many array initializers"));
        do_init = FALSE;
        break;
      }
      push(de_ty(ttop), de_next(ttop));
      ttop = top;
      if (!do_init)
        return;
    } else if (ty_type(ttop) == NODE_TYPE) {
      /*
       * Scanning a structure
       */
      if (is_null_type(su_cdl(ttop))){
        if (su_fl(ttop)){
          uerror(gettxt("xs:35","too many struct initializers"));
          do_init = FALSE;
          break;
        } else if (nelem == 0)
          break;
        pop();
	ttop = top;
        continue;
      }
      {
      	TYPEP de;
        de = dl_de(su_cdl(ttop));
        push(tl_ty(su_ctl(ttop)), de_next(de));
        ttop = top;
      }
      if (!do_init)
        return;
    } else
      break;
  } while (!onelevel);
}

static
endelem()
{
  int flag, type;
  TYPEP ttop;
  ttop = top;

  while (((type = ty_type(ttop)) == NODE_DECL && de_fl(ttop) == 0 ||
          type == NODE_TYPE && su_fl(ttop) == 0 ||
          type == NODE_REPLACE) && nelem > 1){
    pop();
    ttop = top;
    if (ty_type(ttop) == NODE_DECL && ((flag = de_flag(ttop)) == DECL_ARRAY ||
    	flag == DECL_SIZE)){
      de_in_cur_new(ttop, de_in_cur(ttop) + 1);
      if (de_flag(ttop) == DECL_ARRAY || de_in_cur(ttop) < de_in_max(ttop))
        break;
    } else if (ty_type(ttop) == NODE_TYPE && ty_flag(ttop) == TYPE_STRUCT){
      su_cdl_new(ttop, dl_next(su_cdl(ttop)));
      if (!is_null_type(su_cdl(ttop))){
        break;
      } else {
        su_ctl_new(ttop, tl_next(su_ctl(ttop)));
        if (!is_null_type(su_ctl(ttop))){
          su_cdl_new(ttop, tl_dl(su_ctl(ttop)));
          break;
        }
      } 
    }
  }
}

static
push(type, elem)
  TYPEP type, elem;
{
  if (is_null_type(elem)){
    push_scalar(type);
    return;
  }
  de_fl_new(elem, 0);

  switch(de_flag(elem)){
  case DECL_PTR:
  case DECL_FUNC:
    init_push(elem);
    return;
  case DECL_ARRAY:
  case DECL_SIZE:
    de_in_cur_new(elem, 0);
    init_push(elem);
    return;
  }
}

static
push_scalar(type)
  TYPEP type;
{
  int flag;

  if (is_null_type(type) || (flag = ty_flag(type)) == TYPE_SYSTEM ||
      flag == TYPE_ENUM){
    init_push(SCALAR);
    return;
  }
  
  if (flag == TYPE_STRUCT || flag == TYPE_UNION){
    TYPEP aux;
    
    if (is_null_type(su_tl(type))){
      if (is_null_type(aux = tdl_locate(flag, ty_name(type))) || 
          is_null_type(su_tl(aux))){
        uerror(gettxt("xs:36","undefined struct/union %s"), ty_name(type));
        do_init = FALSE;
        return;
      }
      su_tl_new(type, su_tl(aux));
    }
    su_ctl_new(type, su_tl(type));
    su_cdl_new(type, tl_dl(su_tl(type)));
    su_fl_new(type, 0);
    init_push(type);
    return;
  }
  
  if (flag == TYPE_TYPEDEF){
    push(td_ty(type), de_next(td_de(type)));
    return;
  }
}

/*
 * Build actual name for replacement expression
 */
static char *
make_name(is_size)
  bool is_size;
{
  register int i;
  TYPEP tptr;
  char *ptr = loc_buf;
  int lnelem = nelem;
  
  if (is_size)
    lnelem--;
  ptr += sprintf(ptr, "%s", ty_name(curident));
  for (i = 0 ; i < lnelem ; i++){
    tptr = init_elem(i);
    switch(ty_type(tptr)){
    case NODE_REPLACE:
      continue;
    case NODE_DECL:
      if (de_flag(tptr) == DECL_PTR)
      	continue;
      if (i == lnelem - 2 && de_flag(tptr) == DECL_ARRAY &&
      		id_flag(curident) & IDENT_ARRAY)
        continue;
      ptr += sprintf(ptr, "[%d]", de_in_cur(tptr));
      continue;
    case NODE_TYPE:
      ptr += sprintf(ptr, ".%s", ty_name(dl_de(su_cdl(tptr))));
      continue;
    }
  }
  return na_store(loc_buf);
}

/*
 * Found a string in an initial expression
 */
void
init_string(string)
  TOKEN string;
{
  loc_string = string;
  is_string = TRUE;
}

/*
 * Found a name in an initial expression that is already used in an
 * initialization expression
 */
void
init_name(name)
  TOKEN name;
{
  TYPEP aux;

  if (!do_init)
    return;
    
  if (is_null_type(aux = a_ident_isin(name.t_name, (int *)0)))
    uerror(gettxt("xs:37","%s undefined"), name.t_name);
  else
    if (id_flag(aux) & (IDENT_STRINIT|IDENT_STRUSE|IDENT_ARRAY))
      is_name = TRUE;
}

ec_replace(id)
  TYPEP id;
{
  TYPEP ptr;
  ptr = id_ec(id);
  while (!is_null_type(ptr)){
    if (ec_typedef(ptr)){
      act_e_replace(ec_off(ptr), ec_len(ptr), "\nchar");
      act_e_replace(ec_a_off(ptr), ec_a_len(ptr), " *%s", ty_name(id));
    } else
      act_e_replace(ec_off(ptr), ec_len(ptr), " *%s", ty_name(id));
      ptr = ec_next(ptr);
  }
}
