
/* 
 * $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

/*	Copyright (c) 1984 AT&T	*/
/*	  All Rights Reserved  	*/

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/
#ident "$Header: xs.c 1.1 90/01/19 $@(#)xs.c	1.1 89/12/01"
#include "xs.h"
#include "names.h"
#include "stack.h"
#include "types.h"
#include "anal.h"
#include "xdefs.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/signal.h>

extern char *gettxt();

static char n_outfile[MAXNAMELEN + 1];		/* output file name */
static char n_msgfile[MAXNAMELEN + 1];		/* messages file name */
static char *outfile = OUTFILE;			/* output name format */
static char *msgfile = MSGFILE;			/* messages name format */
static bool onepass = FALSE;
static bool use_stdout = FALSE;

float_exception ()
{
	signal(SIGFPE,float_exception);
	return;
}

intr_end()
{
  bool doit = FALSE;
  fprintf(stderr, gettxt("xs:78","Interrupt\n"));
  if (unlink (n_msgfile) == 0){
    doit = TRUE;
    fprintf(stderr, "%s", n_msgfile);
  }
  if (!use_stdout){
    if (unlink (n_outfile) == 0){
      if (doit)
        fprintf(stderr, " ");
      else
        doit = TRUE;
      fprintf(stderr, "%s", n_outfile);
    }
  }
  if (doit)
    fprintf(stderr, gettxt("xs:87"," removed\n"));
  exit(3);
}

main(argc, argv)
  int argc;
  char *argv[];
{
  register int i;
  bool do_stdin;

  lxinit_all();

  signal(SIGFPE, float_exception);
  signal(SIGHUP, intr_end);
  signal(SIGINT, intr_end);
  signal(SIGQUIT,intr_end);
  signal(SIGTERM, intr_end);
  for (i = 1 ; i < argc ; i++){
    /*
     * Analyze non input files arguments
     */
    if (argv[i][0] == '-'){
      switch(argv[i][1]){
      case 'O':
        /*
         * Use stdout to output result
         */
        use_stdout = TRUE;
        argv[i][0] = 0;
        continue;
      case 'o':
        /*
         * Output file name format
         */
        argv[i][0] = 1;
        outfile = argv[++i];
        use_stdout = FALSE;
        continue;
      case 'm':
        /*
         * Messages file name format
         */
        argv[i][0] = 1;
        msgfile = argv[++i];
        continue;
      case 'i':
        /*
         * Strings initialization routine name
         */
        argv[i][0] = 1;
        init_strings = argv[++i];
        continue;
      case 'r':
        /*
         * Strings replace routine name
         */
        argv[i][0] = 1;
        replace_strings = argv[++i];
        continue;
      case 'u':
        /*
         * Extra vars "unique" names (prefix only)
         */
        argv[i][0] = 1;
        strings_unique = argv[++i];
        continue;
      case '1':
        /*
         * Stop after pass 1
         */
        onepass = TRUE;
        argv[i][0] = 0;
        continue;
      case 'd':
        argv[i][0] = 1;
        desc_file = argv[++i];
        continue;
      case '\0':
        /*
         * Allow files names beginning with '-'.
         * This happens from time to time
         */
        break;
      default:
        fprintf(stderr, gettxt("xs:79","-%c: Unknown option\n"), argv[i][1]);
        argv[i][0] = 0;
      }
    }
  }
  if (desc_file != 0)
    load_desc_file();

  do_stdin = TRUE;
  for (i = 1 ; i < argc ; i++){
    char *sl, *dot;
    /*
     * Process input files arguments
     */
    if (argv[i][0] < ' '){
      i += argv[i][0];
      continue;
    }
    do_stdin = FALSE;
    if ((fd_in = fopen(argv[i], "r")) == NULL){
      perror(argv[i]);
      continue;
    }
    strcpy(ftitle, argv[i]);

    /*
     * extract input file basename
     */
    sl = strrchr(argv[i], '/');
    if (sl != 0)
      argv[i] = sl + 1;
    dot = strrchr(argv[i], '.');
    if (dot != 0)
      *dot = '\0';

    /*
     * Do it
     */
    do_file(argv[i]);
  }
  
  if (do_stdin){
    /*
     * No input file arguments : use stdin.
     * Take a copy if not a regular file
     * output file on stdout, messges file base name taken from
     * DEFMSGFILE
     */
    struct stat statbuf;

    if (fstat(fileno(stdin), &statbuf) == -1){
      perror("stdin");
      exit(1);
    }

    if ((statbuf.st_mode & S_IFMT) != S_IFREG){
      char buf[BUFSIZ];
      int n;
      
      /*
       * Not a regular file
       */
      if ((fd_in = tmpfile()) == NULL){
        perror("tempfile");
        exit(1);
      }
      while ((n = fread(buf, 1, BUFSIZ, stdin)) != 0){
        if (fwrite(buf, 1, n, fd_in) != n){
          perror("temp file");
          fclose(fd_in);
          exit(1);
        }
      }
      rewind(fd_in);
    } else {
      do_stdin = FALSE;
      fd_in = stdin;
    }
    
    strcpy(ftitle, "stdin");
    /*
     * Do it
     */
    do_file("stdin");
  }
  exit(0);
}

/*
 * Actual file processing
 */
do_file(f_name)
  char *f_name;
{
  char buf[32];
  long time();
  char *ptr;
  
  sprintf(buf, "_%s_%lX", f_name, time((long *)0));
  if ((ptr = strchr(buf, '.')) != 0)
    *ptr = '_';
    
  act_start();
  reset_error();
  reset_var();
  reset_string();
  reset_init();
  lxinit();
  types_creat();
  repl_creat();
  var_set_nr = na_unique();
  sprintf(n_msgfile, msgfile, f_name);
  if ((fd_msg = fopen(n_msgfile, "w")) == NULL){
    perror(n_msgfile);
    fclose(fd_in);
    fclose(fd_out);
    return;
  }
  do_start();
  yyparse();
  func_name = na_store(buf);
  do_global();
  do_symtab();
  if (!onepass){
    if (!use_stdout){
      sprintf(n_outfile, outfile, f_name);
      if ((fd_out = fopen(n_outfile, "w")) == NULL){
        perror(n_outfile);
        fclose(fd_in);
        return;
      }
    } else
      fd_out = stdout;

    act_do();
    fclose(fd_out);
  }
  fclose(fd_in);
  msg_end(fd_msg);
  fclose(fd_msg);
  types_free();
  repl_end();
  act_free();
  na_freeall();
}
