/*
 * $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 *
 */
/*	Copyright (c) 1984 AT&T	*/
/*	  All Rights Reserved  	*/

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/

#ident	"@(#)libc-port:gen/time_comm.c	1.8.1.1"
/*
 * Functions that are common to ctime(3C) and cftime(3C)
 */
#include	<ctype.h>
#include  	<stdio.h>
#include  	<sys/types.h>
#include 	<time.h>
#include 	<string.h>

#define	EPOCH		70
#define SEC_IN_DAY	(24 * 60 * 60)
#define MAXTZNAME	50
#define	year_size(A)	(((A) % 4) ? 365 : 366)
#define dayofweek(A)	(((A) + 7340035L) % 7)

time_t	timezone = 0;
time_t	altzone = 0;
int 	daylight = 0;
extern char 	*getenv();

static time_t 	start_dst ;    /* Start date of alternate time zone */
static time_t  	end_dst;      /* End date of alternate time zone */
static char	tz0[MAXTZNAME+1] = "GMT";
static char	tz1[MAXTZNAME+1] = "   ";
char 		*tzname[] = { tz0, tz1 };
static short	month_size[12] = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

extern int getdst();
extern char *getzname(), *gettime(), *getdigit();
extern void tzsetclock();

struct tm *
localtime(tim)
time_t	*tim;
{
	register struct tm *ct;
	static void getusa();
	long daybegin, dayend;
	time_t	copyt, curr;
	char *pch;

	tzsetclock(*tim);

	copyt = *tim - timezone;
	ct = gmtime(&copyt);
	if (!daylight) {
		tzset();
		return(ct);
	}

	curr = ct->tm_yday*SEC_IN_DAY + ct->tm_hour*60*60 + ct->tm_min*60 + ct->tm_sec;
	if ( start_dst == 0 && end_dst == 0) {
		(void) getusa(&daybegin, &dayend, ct);
		start_dst = daybegin;
		end_dst = dayend;
	} else {
		daybegin = start_dst;
		dayend  = end_dst;
	}
	if (curr >= daybegin && curr < dayend)
	{
		copyt = *tim - altzone;
		ct = gmtime(&copyt);
		ct->tm_isdst = 1;
	}
	tzset();
	return(ct);
}

struct tm *
gmtime(tim)
time_t	*tim;
{
	register int		d0, d1;
	long			hms, day;
	static struct tm	xtime;

	/* Break initial number into days */
	hms = *tim % SEC_IN_DAY;
	day = *tim / SEC_IN_DAY;
	if(hms < 0) {
		hms += SEC_IN_DAY;
		day -= 1;
	}

	/* Generate hours:minutes:seconds */
	xtime.tm_sec = hms % 60;
	d1 = hms / 60;
	xtime.tm_min = d1 % 60;
	d1 /= 60;
	xtime.tm_hour = d1;

	/* Generate day of the week. The addend is 4 mod 7 (1/1/1970 was Thursday) */

	xtime.tm_wday = (day + 7340036L) % 7;

	/* Generate year number */
	if(day >= 0)
		for(d1=EPOCH; day >= year_size(d1); d1++)
			day -= year_size(d1);
	else
		for(d1=EPOCH; day < 0; d1--)
			day += year_size(d1-1);
	xtime.tm_year = d1;
	xtime.tm_yday = d0 = day;

	/* Generate month */
	if(year_size(d1) == 366)
		month_size[1] = 29;
	for(d1 = 0; d0 >= month_size[d1]; d1++)
		d0 -= month_size[d1];
	month_size[1] = 28;
	xtime.tm_mday = d0 + 1;
	xtime.tm_mon = d1;
	xtime.tm_isdst = 0;
	return(&xtime);
}

time_t
mktime(timeptr)
struct tm *timeptr;
{
	register int i, year;
	time_t tyme = 0;
	struct tm *tp2;
	char *pch;
	time_t tzdiff = 0, altzdiff, dststart = 0, dstend;
	time_t sincejan1 = 0;
	
	year = timeptr->tm_year;
	if (timeptr->tm_year > EPOCH) {
		while (--timeptr->tm_year >= EPOCH) {
			tyme += year_size(timeptr->tm_year) * SEC_IN_DAY;
		}
	}
	++timeptr->tm_year;
	while (timeptr->tm_year < EPOCH) {
		tyme -= year_size(++timeptr->tm_year) * SEC_IN_DAY;
	}
	if (timeptr->tm_mon >= 0) {
		while (timeptr->tm_mon >= 12) {
			tyme += year_size(year++) * SEC_IN_DAY;
			timeptr->tm_mon -= 12;
		}
		if (year_size(year) == 366) {
			month_size[1] = 29;
		}
		for (i = 0; i < timeptr->tm_mon; i++) {
			sincejan1 += month_size[i] * SEC_IN_DAY;
		}
		month_size[1] = 28;
	} else {
		--year;
		while (timeptr->tm_mon < -12) {
			sincejan1 -= year_size(year--) * SEC_IN_DAY;
			timeptr->tm_mon += 12;
		}
		if (year_size(year) == 366) {
			month_size[1] = 29;
		}
		for (i = 11; timeptr->tm_mon++ < 0; i--)
			sincejan1 -= month_size[i] * SEC_IN_DAY;
		month_size[1] = 28;
	}

	sincejan1 += (timeptr->tm_mday - 1) * SEC_IN_DAY;
	sincejan1 += timeptr->tm_hour * 60 * 60;
	sincejan1 += timeptr->tm_min * 60;
	sincejan1 += timeptr->tm_sec;

	tyme += sincejan1;

	tzsetclock(tyme);
	if (timeptr->tm_isdst > 0) {
		tyme += altzone;
	} else if (timeptr->tm_isdst == 0) {
		tyme += timezone;
	} else if (sincejan1 >= start_dst && sincejan1 < end_dst) {
		tyme += (altzone - timezone);
	}
	tp2 = localtime(&tyme);

	timeptr->tm_sec = tp2->tm_sec;
	timeptr->tm_min = tp2->tm_min;
	timeptr->tm_hour = tp2->tm_hour;
	timeptr->tm_mday = tp2->tm_mday;
	timeptr->tm_mon = tp2->tm_mon;
	timeptr->tm_year = tp2->tm_year;
	timeptr->tm_wday = tp2->tm_wday;
	timeptr->tm_yday = tp2->tm_yday;
	timeptr->tm_isdst = tp2->tm_isdst;

	return(tyme);
}

void
tzset()
{
	tzsetclock(time((time_t *)0));
}

static void
tzsetclock(clock)
time_t clock;
{
  char *ptr;

  
	if ( (ptr = getenv("TZ")) == NULL)
		return;
	
	/* Get main time zone name and difference from GMT */
	if ( ((ptr = getzname(ptr,tzname[0])) == 0) || ((ptr = gettime(ptr,&timezone,1)) == 0))
		return;
	
	altzone = timezone - 60 * 60;
	start_dst = end_dst = 0;
	daylight = 0;

	/* Get alternate time zone name */
	if ( (ptr = getzname(ptr,tzname[1])) == 0)
		return;
	daylight = 1;

	/* If the difference between alternate time zone and
	 * GMT is not given, use one hour as default.
	 */
	if (*ptr == '\0')
		return;
	if (*ptr != ',')
		if ( (ptr = gettime(ptr,&altzone,1)) == 0 || *ptr != ',')
			return;
	getdst(ptr + 1,&start_dst, &end_dst, clock);
}

static char *
getzname(p,tz)
char *p, *tz;
{
  int n = MAXTZNAME;

  if (!isalpha(*p))
	return(0);
  do
  {
	if (NULL != tz)
		*tz++ = *p ;
  } while (--n > 0 && isalpha(*++p) );
  while(isalpha(*p))
	p++;
  if (NULL != tz)
	  while(--n >= 0)
		*tz++ = '\0';		/* Pad with nulls */
  return(p);	
}

static char *
gettime(p,timez,f)
char *p;
time_t *timez;
int f;
{
	register time_t t = 0;
	int d, sign = 0;

	d = 0;
	if (f)
		if ( (sign = (*p == '-')) || (*p == '+'))
			p++;
	if ( (p = getdigit(p,&d)) != 0)
	{
		t = d * 60 * 60;
		if (*p == ':')
		{
			if( (p = getdigit(p+1,&d)) != 0)
			{
				t += d * 60;
				if (*p == ':')
				{
					if( (p = getdigit(p+1,&d)) != 0)
						t += d;
				}
			}
		}
		if(sign)
			*timez = -t;
		else
			*timez = t;
	} 
	return(p);
}
static char *
getdigit(ptr,d)
char *ptr;
int * d;
{


	if (!isdigit(*ptr))
		return(0);
 	*d = 0;
	do
	{
		*d *= 10;
		*d += *ptr - '0';
	}while( (isdigit(*++ptr)));
	return(ptr);
}
static int
getdst(p,s,e,t)
char *p;
time_t *s, *e, t;
{
	int lsd,led;
	time_t st,et;
	char *dstdate();
	st = et = 2*60*60; 	/* Default for start and end time is 02:00:00 */
	if ( (p = dstdate(p,&lsd, t)) == 0 )
		return(0);
	if ( (*p == '/') &&  ((p = gettime(p+1,&st,0)) == 0) )
			return(0);
	if (*p == ',')
	{
		if ( (p = dstdate(p+1,&led, t)) == 0 )
			return(0);
		if ((*p == '/') &&  ((p = gettime(p+1,&et,0)) == 0) )
				return(0);
	}
	/* Convert the time into seconds */
	*s = (long)(lsd * SEC_IN_DAY + st);
	*e = (long)(led * SEC_IN_DAY + et - (timezone - altzone));
	return(1);
}

static void
getusa(s,e,t)
long *s,*e;
struct tm *t;
{
	static struct {
		int	yrbgn;
		int	daylb;
		int	dayle;
	} daytab[] = {
		87,	96,	303,	/* new legislation - 1st Sun in April	*/
		76,	119,	303,
		75,	58,	303,	/* 1975: Last Sun in Feb - last Sun in Oct */
		74,	5,	333,	/* 1974: Jan 6 - last Sun. in Nov */
		70,	119,	303,	/* start GMT	*/
	};
	int i = 0;

	while (t->tm_year < daytab[i].yrbgn) /* can't be less than 70	*/
		i++;
	*s = daytab[i].daylb; /* fall through the loop when in correct interval	*/
	*e = daytab[i].dayle;

	*s = sunday(t, *s);
	*e = sunday(t, *e);
	*s = (long)(*s * SEC_IN_DAY +2*60*60);
	*e = (long)(*e * SEC_IN_DAY +60*60);
	return;
}
static int
sunday(t, d)
register struct tm *t;
register int d;
{
	if(d >= 58)
		d += year_size(t->tm_year) - 365;
	return(d - (d - t->tm_yday + t->tm_wday + 700) % 7);
}

/*
 * Parses the date portion of the Daylight savings time string within the
 * timezone.  This new function handles the POSIX specified system of giving a
 * date, either Julian 'J', Floating 'M', or just normally ''
 */
char *
dstdate(p, date, now)
char *p;
int *date;
time_t now;
{
	register int d1, m1=0, year;
	int m,n,d;
	long day, dom, first;
/*
 *  Determine the year.  This is needed for the Julian since February 29th
 *  doesn't really 'exist' for the POSIX Julian system.   The year is also 
 *  needed to determine day of the week jan 1st falls on, for the floating 
 *  method of determining start and end dst.
 */

	day = now / SEC_IN_DAY;

	for (year = EPOCH, d1 = 0; d1 + year_size(year) <= day; year++) {
		d1 += year_size(year);
	}

	switch (*p) {
		case 'J':
			if ( (p = getdigit(p+1, date) ) == 0)
				return(0);
#if 0
/* This code looks right to me, but it's ifdef'd out
   because we don't pass the VSX3.201 ANSI.os/time
   tests with it in. -mjb */

			if (year_size(year) == 366  &&  *date >= 60)
				*date -= 1;
#endif
			return(p);
		case 'M':
			if ( (p = getdigit(p+1, &m) ) == 0)
				return 0;
			if ( *p  == '.' ) {
				if ((p = getdigit(p+1, &n)) == 0)
					return 0;
				if ( *p == '.' ) {
					if ( (p = getdigit(p+1, &d) ) == 0)
						return 0;
				}
			}
	/*  We found Dec 31st for the year before, add one to the day. */
			d1++;
			first = d1;
			if(year_size(year) == 366) {
				month_size[1] = 29;
			}
			for (m1 = 0; m > m1 + 1; m1++) {
				d1 += month_size[m1];
			}
			/* Found the first day of the month we want */
			if (dayofweek(d1) > d) {
				dom = 7 - dayofweek(d1) + d;
			} else if (dayofweek(d1) < d) {
				dom = d - dayofweek(d1);
			}
			dom += 7*(n - 1);
			if (dom > month_size[m-1]) {
				dom -= 7;
			}
			d1 += dom;
			month_size[1] = 28;
			*date = d1 - first;
			return(p);
		default:
			return(p = getdigit(p, date));
	}
}
