/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
*/

#ident	"$Header: _locale.c 1.5 90/11/01 $@(#)libc-port:locale/_locale.c "

#include <stdio.h>
#include "_locale.h"
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>

static char lang[] = "LANG";
static char *_loc_envs[LC_MAX + 1][5] = /* env. vars for "" */
{
	{"LC_ALL", "LC_CTYPE",	   lang,   "CHRCLASS",	0}, /* 0       */
	{"LC_ALL", "LC_NUMERIC",   lang,   0,		0}, /* 1       */
	{"LC_ALL", "LC_TIME",	   lang,   "LANGUAGE",	0}, /* 2       */
	{"LC_ALL", "LC_COLLATE",   lang,   0,		0}, /* 3       */
	{"LC_ALL", "LC_MESSAGES",  lang,   0,		0}, /* 4       */
	{"LC_ALL", "LC_MONETARY",  lang,   0,		0}, /* 5 posix */
	{"LC_ALL", lang,	   0,	   0,	     	0}  /* 6 Xopen */
};

/*
 * This is routine is called by _init_locale() during program startup.
 * It makes sure that the "C" locale string is returned if the 
 * cooresponding ENV variable is not defined. 
 * 
 */
char *
_nativeloc(cat)	/* return value for category with "" locale */
int cat;
{
	static char ans[LC_NAMELEN];
	register char *s, **p;

	for (p = _loc_envs[cat]; *p != 0; p++)
		if ((s = (char *)getenv(*p)) != 0 && s[0] != '\0')
			goto found;

	(void)strncpy(ans, "C", LC_NAMELEN - 1);
	return ans;		/* If nothing else defined, ret "C" locale */
found:
	(void)strncpy(ans, s, LC_NAMELEN - 1);
	ans[LC_NAMELEN - 1] = '\0';
	return ans;
}


/*
 * This is routine is similar to _nativeloc, but is Posix and X/open compliant. 
 * 
 * Called by setlocale()
 */
char *
_sl_nativeloc(cat)	/* return value for category with "" locale */
int cat;
{
	static char ans[LC_NAMELEN];
	register char *s, **p;

	for (p = _loc_envs[cat]; *p != 0; p++)
		if ((s = (char *)getenv(*p)) != 0 && s[0] != '\0')
			goto found;

	(void)strncpy(ans, "C", LC_NAMELEN - 1);
	return ans;		/* If nothing else defined, ret "C" locale */

found:
	(void)strncpy(ans, s, LC_NAMELEN - 1);
	ans[LC_NAMELEN - 1] = '\0';
	if (valid_loc(ans))
		return ans;
	
	return NULL;
}


/*
 * Create a complete pathname out of the locales name.
 */
char *
_fullocale(loc, file)	/* "/usr/lib/locale/<loc>/<file>" */
char *loc;
char *file;
{
	static char ans[18 + 2 * LC_NAMELEN] = "/usr/lib/locale/";
	register char *p = ans + 16;

	(void)strcpy(p, loc);
	p += strlen(loc);
	p[0] = '/';
	(void)strcpy(p + 1, file);
	return ans;
}



/*
 * A Simple check for a locales existance. Just stat what should be
 * the directory name.
 */
int
valid_loc(loc)
   char	*loc;
{
	char   fullpath[18 + LC_NAMELEN];
	struct stat	buf;

	if (strcmp(loc, "C")) { 			/* Not C locale */
		strcpy(fullpath, "/usr/lib/locale/");
		strncat(fullpath, loc, LC_NAMELEN); 
		if (stat(fullpath, &buf) < 0)
			return 0;
	}				/* Else assume valid locale */
	return(1);	
}
