
/* 
 * $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

/*	Copyright (c) 1984 AT&T	*/
/*	  All Rights Reserved  	*/

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/


#ident "$Header: catopen.c 1.1 90/01/09 $@(#)catopen.c	1.4 89/12/01"

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <nl_types.h>
#include "nl_type2.h"

char *getenv(), *strchr();
static char path[NL_MAXPATHLEN];
static char def_nlspath[] = DEF_NLSPATH;
static struct stat buff;

static char *
cat_name(name)
  char *name;
{
  char *nlspath;
  char *lang;
  char *ptr, *ptr2, *ptr3, c, d;
  struct stat *buf;
    
  /*
   * A name that contains a slash is a pathname
   */
  if (strchr(name, '/') != 0)
    return name;
    
  /*
   * Get env variables
   */
  if ((nlspath = getenv(NL_PATH)) == 0){
    /*
     * No env var : use default
     */
    nlspath = def_nlspath;
  }
  
  if ((lang = getenv(NL_LANG)) == 0)
    lang = NL_DEF_LANG;
    
  ptr = nlspath;
  ptr2 = path;
  *ptr2 = '\0';
  while (TRUE){
    switch (c = *ptr++){
    case ':':
      if (ptr2 == path){
        /*
         * Leading ':' Check current directory
         */
        if (access(name, 4) == 0)
          return name;
        continue;
      }
    case '\0':
	/*
	 * Found a complete path
	 */
      	*ptr2 = '\0';
        /*
	 * Test to see that path is not a pure directory,
	 * if it is then attempt toopen it and append the
	 * filename "name" to it, in an attempt to succeed
	 * This syntax is not explicitly defined in XPG2, but
	 * is a logical extension to it. (XVS tests it too).
	 */
	
	buf = &buff;
	stat(path, buf);
	if ((buf -> st_mode) & S_IFDIR) {
		strcat(path, "/");
		if (access(strcat(path, name), 4) == 0)
			return path;
		} /* otherwise branch out to end */
	else
		if ( access(path, 4) == 0)
			return path;
	
	if (c == '\0'){
		/*
	 	 * File not found
	 	 */
	  	if (nlspath == def_nlspath)
		    return 0;
		nlspath = def_nlspath;
		ptr = nlspath;
        }
        ptr2 = path;
        *ptr2 = '\0';
        continue;
    case '%':
      /*
       * Expecting var expansion
       */
      switch(c = *ptr++){
      case 'L':
        ptr3 = lang;
        while ((d = *ptr3++) != 0)
          *ptr2++ = d;
        continue;
      case 'l':
        ptr3 = lang;
        while ((d = *ptr3++) != 0 && d != '.' && d != '_')
          *ptr2++ = d;
        continue;
      case 't':
        ptr3 = strchr(lang, '_');
        if (ptr3++ != 0){
          while ((d = *ptr3++) != 0 && d != '.')
            *ptr2++ = d;
        }
        continue;
      case 'c':
        ptr3 = strchr(lang, '.');
        if (ptr3++ != 0){
          while ((d = *ptr3++) != 0)
            *ptr2++ = d;
        }
        continue;
      case 'N':
        ptr3 = name;
        while ((d = *ptr3++) != 0)
          *ptr2++ = d;
        continue;
      case '%':
        *ptr2++ = '%';
        continue;
      default:
        *ptr2++ = '%';
        *ptr2++ = c;
        continue;
      }
    default:
      *ptr2++ = c;
      continue;
    }
  }
}

nl_catd
catopen(name, mode)
  char *name;
  int mode;
{
  nl_catd catd;
  int fd;
  
  /*
   * Allocate space to hold the necessary data
   */
  if ((catd = (nl_catd)malloc(sizeof (nl_catd_t))) == 0)
      return (nl_catd)-1;
  
  /*
   * Get actual file name and open file
   */
  if ((name = cat_name(name)) == 0){
    catd->catd_set = 0;
  } else {
  
    /*
     * Build internal data structures
     */
    if (cat_build(name, catd))
      return catd;
  }

  free(catd);
  return (nl_catd)-1;
}
