	.file "pow.s"
/*
 * $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 *
 */

  / Double Precision Power Function
  /
  / 
  / Functional Description:
  / 	This routine implements the double precision power (BASE ** POWER)
  / 	function, where BASE and POWER are both real numbers. The algorithm
  / 	used was obtained from the original PASCAL library that this replaces.
  / 
  / 	The algorithm for the power function is as follows:
  / 
  / 	IF (base . EQ. 0)
  / 		IF (power .LE. 0)
  / 			return ERROR
  / 		ELSE
  / 			result = 0
  / 		ENDIF
  / 	ELSE
  / 		ipower = trunc(power)
  / 		IF (ipower .NE. power)
  / 			IF (base .GT. 0)
  / 				result = exp(ln(base) * power)
  / 			ELSE
  / 				return ERROR
  / 			ENDIF
  / 		ELSE
  / 			result = __powi (base, ipower)
  / 		ENDIF
  / 	ENDIF
  / 
  / Created:
  / 	11/25/86 	BS - Quantitative Technology Corporation for
  / 				Sequent Computer Systems
  /
  /
	.data
  /
  / DOMERR	11			 error code for illegal value
  / OVFERR	12			 error code for overflow
  /
  /ipower	edx			 integer value of power 
  /base	fp2			 holds base 
  /power	fp4			 holds power 
  /tempf	fp6			 temporary register 
  /
one:	.double	0Dx3ff0000000000000	/fp one
inf:	.double 0Dx7ff0000000000000	/fp infinity
neginf:	.double 0Dxfff0000000000000	/fp -infinity
  /ctxtrnc 0x013a0000			#context register bits for truncate
  /ctxrnd  0x073a0000			#context register bits for rounding
  /
  /
#include "sys.s"

POWERR1:
	.byte	0x70,0x6f,0x77,0x00
POWERR2:
	.byte	0x70,0x6f,0x77,0x3a,0x20,0x44,0x4f,0x4d,0x41,0x49
	.byte	0x4e,0x20,0x65,0x72,0x72,0x6f,0x72,0x0a,0x00
HALF:
	.double	0Dx3fe0000000000000
ZERO:
	.double 0Dx0000000000000000
pow_max_int:
	.double	0Dx41dfffffffc00000
	.data
	.align	2

ENTRY(pow)
	wloadl	4(%esp),%fp2		/get base
	wloadl	12(%esp),%fp4		/get power
	wtstl	%fp2			/test base = 0
	wstctx  %eax			/transfer contents of control reg
	sahf				/receive to flags
	jne	chk_pwr			/if not, go check value of power
	wtstl	%fp4			/test power > 0
	wstctx  %eax			/transfer contents of control reg
	sahf				/receive to flags
	ja	return2			/if so, return a value of 0 (jlt)

pow_domain_err:
	wloadl	4(%esp),%fp2		/get base
	wloadl	12(%esp),%fp4		/get power
	subl	$0x20,%esp		/allocate struct exception exc
	wstorl	%fp2,0x8(%esp)		/exc.arg1 = base
	wstorl	%fp4,0x10(%esp)		/exc.arg2 = power
	movl	$0x1,(%esp)		/exc.type = DOMAIN
	movl	$POWERR1,0x4(%esp)	/exc.name = "pow"
	wloadl	ZERO,%fp2		/exc.retval = 0.0
	wstorl	%fp2,0x18(%esp)
	leal	(%esp),%eax
	pushl	%eax
	call	matherr
	popl	%ecx
	testl	%eax,%eax
	jne	pow_domain_exit

	pushl	$0x12			/length of err string
	pushl	$POWERR2		/"pow: DOMAIN error\n"
	pushl	$0x2			/write to stderr
	call	write
	addl	$0xc,%esp
	movl	$0x21,errno		/errno = EDOM
pow_domain_exit:
	wloadl	0x18(%esp),%fp2		/return exc.retval
	addl	$0x20,%esp		/deallocate struct exception exc
	ret

  /
  / base <> 0
  /

chk_pwr:
	pushl	%ebx			/preserve non-scratch register
	wstctx	%ebx			/save old value of context work
	movl	$0x073a0000,%ecx	/truncate context word
	wldctx	%ecx			/place 1167 in truncate mode
	wfixl	%fp4,%fp6		/get INT(power)
	wstors	%fp6,%edx		/
	wfloatl	%fp6,%fp6		/
	wcmpl	%fp4,%fp6		/is power an integer value?
	wstctx  %eax			/transfer contents of control reg
	wldctx	%ebx			/restore rounding mode
	popl	%ebx			/restore contents of %ebx
	sahf				/receive to flags
	je	up_int			/power is an integer, go compute value
	wtstl	%fp2			/ base > 0?
	wstctx  %eax			/transfer contents of control reg
	sahf				/receive to flags
	jae	ln_exp			/if so, go compute via ln and exp(jge)

  /##############################################################
  / Error return for base < 0 and power is not an integer value #
  /##############################################################

	jmp	pow_domain_err

  /
  / base > 0 and power is not an integer value 
  /
ln_exp:
	subl	$8,%esp
	wstorl	%fp2,(%esp)		/ get value of base
	call	log			/ compute ln(base)
	wmull	20(%esp),%fp2		/ compute ln(base) * power
	wloadl	libm_maxexp,%fp4	/ check ln(base) * power < maxexp
	wcmpl	%fp2,%fp4
	wstctx	%eax
	sahf				/transfer contents of control reg
	ja	overflow0		/ we have an overflow
	wloadl	libm_minexp,%fp4	/ check .n(base) *power > minexp
	wcmpl	%fp2,%fp4
	wstctx	%eax
	sahf				/ transfer contents of control reg
	jb	underflow0		/ we have an underflow
	wstorl	%fp2,(%esp)		/ get ln(base) * power
	call	exp			/ compute exp(ln(base) * power)
	addl	$8,%esp
	ret

  /
  / power is an integer value	
  /

up_int:
	wloadl	%fp2,%fp4		/move base

  / Double Precision Integer Power Function
  /
  / 
  / Functional Description:
  / 	This routine implements the double precision integer power
  / 	(BASE ** IPOWER) function, where BASE is real and IPOWER is integer.
  / 	The algorithm used was obtained from the original PASCAL library that
  / 	this replaces. (NOTE: This routine is called by the flt. pt. power
  / 	function to compute the result when the BASE is less than 0 or when
  / 	the power is an integer value.)
  / 
  / 	The algorithm for the integer power function is as follows:
  / 
  / 	result = 1
  / 	invert = ipower .LT. 0
  / 	IF (ipower .NE. 0)
  / 		ipower = abs(ipower)
  / 		IF (odd(ipower))
  / 			result = base
  / 		ENDIF
  / 		bit = 2
  / 		DO WHILE (ipower .GE. bit)
  / 			base = base * base
  / 			IF ((ipower .AND. bit) .NE. 0)
  / 				result = result * base
  / 			ENDIF
  / 			bit = bit + bit
  / 		ENDDO
  / 		IF (invert)
  / 			result = 1.0 / result
  / 		ENDIF
  / 	ENDIF
  / 
  / Created:
  / 	11/25/86 	BS - Quantitative Technology Corporation for
  / 				Sequent Computer Systems
  /
  / ipower	edx	 integer value of power 
  / invert	al	 flag for inverting(negative power) 
  / bitval	ecx	 integer value of bit mask 
  / result	fp2	 holds result 
  / base		fp4	 holds base 
  /
	wloadl	one,%fp2		/init result to 1.0
	cmpl	$0,%edx			/power < 0?
	setg	%al			/if so, set flag
	jz	return			/if power = 0, return a 1.0
	jg	nabs			/if positive don't perform absolute val
	negl	%edx			/ipower = abs(ipower)
nabs:
	testb	$0x01,%edx		/is ipower even?
	jz	cont			/if so, continue
	wloadl	%fp4,%fp2		/if not, init result to base
cont:
	movl	$2,%ecx			/initialize bit tester value
loopf:
	cmpl	%ecx,%edx		/power < bitval?
	jb	chkinv			/if so, thru. Go check on invert
	wmull	%fp4,%fp4		/base = base * base
	testl	%ecx,%edx		/bit set in power?
	jz	cont1			/if not, don't update result
	wmull	%fp4,%fp2		/if so, update result
cont1:
	addl	%ecx,%ecx		/update bit tester value
	jmp	loopf			/continue with computation
					/
chkinv:
	cmpb	$0,%al			/was power < 0?
	jnz	return			/is not, go return
	wdivl	one,%fp2		/result = 1.0/result

  /
  / finish and clean up stack
  /

return:
	wloadl	inf,%fp4		/check for infinity return (overflow)
	wabsl	%fp2,%fp6
	wcmpl	%fp4,%fp6
	wstctx  %eax			/transfer contents of control reg
	sahf				/receive to flags
	je	overflow
	wloadl	_mindouble,%fp4		/check for number<mindouble (underflow)
	wcmpl	%fp4,%fp6
	wstctx	%eax
	sahf
	ja	underflow
return2:
	ret

underflow0:
	addl	$8,%esp
underflow:
	wloadl	4(%esp),%fp2		/get base
	wloadl	12(%esp),%fp4		/get power
	subl	$0x20,%esp		/allocat struct exception exc
	wstorl	%fp2,0x8(%esp)		/exc.arg1 = base
	wstorl	%fp4,0x10(%esp)		/exc.arg2 = power
	movl	$0x4,(%esp)		/exc.type = UNDERFLOW
	wloads	ZERO,%fp6		/exc.retval = 0.0
	wcvtls	%fp6,%fp4
	wstorl	%fp4,0x18(%esp)
	jmp	pow_range_err

overflow0:
	addl	$8,%esp
overflow:
	wloadl	4(%esp),%fp2		/get base
	wloadl	12(%esp),%fp4		/get power
	subl	$0x20,%esp		/allocat struct exception exc
	wstorl	%fp2,0x8(%esp)		/exc.arg1 = base
	wstorl	%fp4,0x10(%esp)		/exc.arg2 = power
	movl	$0x3,(%esp)		/exc.type = OVERFLOW
	wloadl	_maxdouble,%fp4		/exc.retval = HUGE_VAL
	wstorl	%fp4,0x18(%esp)
	wloadl	%fp2,%fp6		/duplicate base and see if > 0
	wabsl	%fp6,%fp6
	wcmpl	%fp2,%fp6
	wstctx	%eax
	sahf
	je	pow_range_err		/base is positive don't change HUGE_VAL

	subl	$0x8,%esp		/make room for dum
	leal	(%esp),%eax		/load address for dum
	pushl	%eax
	wloadl	0x1c(%esp),%fp2		/put power * 0.5 on stack
	wmull	HALF,%fp2
	subl	$0x8,%esp		/make room on stack for power * 0.5
	wstorl	%fp2,(%esp)
	call	modf			/ modf(power*0.5, &dum)
	addl	$0x14,%esp		/pop args off of stack
	wtstl	%fp2
	wstctx	%eax
	sahf
	je	pow_range_err		/no power is even so exc.retval is ok
	wcmpl	HALF,%fp2		/ is MODF(power*0.5, &dum) != 0.5
	wstctx	%eax
	sahf
	jne	pow_big_err1		/ this is a domain error
	wloadl	0x18(%esp),%fp2		/ power is odd so ...
	wnegl	%fp2,%fp4
	wstorl	%fp4,0x18(%esp)
	jmp	pow_range_err
pow_big_err1:
	addl	$0x20,%esp		/get stack ready for DOMAIN errror
	jmp	pow_domain_err
pow_range_err:
	movl	$POWERR1,0x4(%esp)	/exc.name = "pow"
	leal	(%esp),%eax
	pushl	%eax
	call	matherr
	popl	%ecx
	testl	%eax,%eax
	jne	pow_range_exit
	movl	$0x22,errno		/errno = ERANGE
pow_range_exit:
	wloadl	0x18(%esp),%fp2		/return exc.retval
	addl	$0x20,%esp		/deallocate struct exception exc
	ret

