# $Copyright:	$
# Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
# Sequent Computer Systems, Inc.   All rights reserved.
#  
# This software is furnished under a license and may be used
# only in accordance with the terms of that license and with the
# inclusion of the above copyright notice.   This software may not
# be provided or otherwise made available to, or used by, any
# other person.  No title to or ownership of the software is
# hereby transferred.
#

# $Header: libpps.mk 1.15 89/07/06 $

# Makefile for i386 Parallel Support Library
#
# Defines:
#	ALM		- use multibus atomic lock memory for hardware mutex.
#	OPENT		- kernel support for reading program object file.
#	CONCAT		- mmap() support for concatenating mmap regions.
#	MAP_ZEROFILL	- mmap() support for creating ZFOD space.
#				This is picked up from <sys/mman.h>
#
#

CPP	= /lib/cpp
COPTS	= -DCONCAT
CFLAGS	= -O -Wc,-i $(COPTS) -Wc,-seq

LIB	= libpps.a
GTPP_FILES = Makefile README cdynamic.c cstatic.c shmem.c ifile

OBJS	= ppinit.o \
		c_interface.o f_interface.o team.o microtask.o m_get_deflt.o \
		shbrk.o prbrk.o slock.o sbarrier.o shmalloc.o  \
		cpus_online.o _nargs.o shared_flag.o \
		profbrk.o _p__ffork.o

CFILES	= ppinit.c shbrk.c prbrk.c slock.c sbarrier.c \
		shmalloc.c microtask.c cpus_online.c \
		shared_flag.c _nargs.c \
		profbrk.c team.c c_interface.c \
		f_interface.c m_get_deflt.c _p__ffork.c

SOBJS	= _exit.o _sigaction.o _mmap.o

.s.o:
	$(CPP) -P -DMCOUNT $(INCLUDE) $*.s | sed -f pps.sed \
		| $(AS) -o profiled/$*.o ; \
	$(CPP) -P -UMCOUNT $(INCLUDE) $*.s | sed -f pps.sed \
		| $(AS) -o $*.o

all		: $(LIB)

$(LIB)		: profiled parallel
		  @$(AR) cru libpps.a $(OBJS) $(SOBJS) && \
			$(AR) ma microtask.o libpps.a ppinit.o && \
			$(AR) ma ppinit.o libpps.a shared_flag.o && \
			$(AR) ma ppinit.o libpps.a shbrk.o
		  cd profiled && \
		  $(AR) cru libpps.a $(OBJS) $(SOBJS) && \
			$(AR) ma microtask.o libpps.a ppinit.o && \
			$(AR) ma ppinit.o libpps.a shared_flag.o && \
			$(AR) ma ppinit.o libpps.a shbrk.o

parallel	:& $(OBJS) $(SOBJS)
profiled	:; mkdir profiled
clean		:; rm -rf $(OBJS) $(SOBJS) profiled
clobber		:; rm -rf $(LIB) $(OBJS) $(SOBJS) profiled

install		: all
		install -f $(ROOT)/usr/include/parallel -u bin -g bin -m 444 \
				parallel.h
		install -f $(ROOT)/usr/include/parallel -u bin -g bin -m 444 \
				microtask.h
		install -f $(ROOT)/usr/lib -u bin -g bin -m 444 libpps.a
		cd profiled; \
		install -f $(ROOT)/usr/lib/libp -u bin -g bin -m 444 libpps.a
		cd gtpp.demo; \
		for i in $(GTPP_FILES);\
		do\
			install -f $(ROOT)/usr/sequent/gtpp.demo -u bin -g bin \
				-m 644 $$i;\
		done

cpus_online.o	: cpus_online.c
			$(CC) -p -DMCOUNT $(CFLAGS) -c cpus_online.c ; \
			$(LD)  -r cpus_online.o -o profiled/cpus_online.o ; \
			$(CC) $(CFLAGS) -c cpus_online.c ; \
			mv cpus_online.o cpus_online.O ; \
			$(LD)  -r cpus_online.O -o cpus_online.o; \
			rm -f cpus_online.O

_p__ffork.o	: _p__ffork.c
			$(CC) -p -DMCOUNT $(CFLAGS) -c _p__ffork.c ; \
			$(LD)  -r _p__ffork.o -o profiled/_p__ffork.o ; \
			$(CC) $(CFLAGS) -c _p__ffork.c ; \
			mv _p__ffork.o _p__ffork.O; \
			$(LD)  -r _p__ffork.O -o _p__ffork.o ; \
			rm -f _p__ffork.O

microtask.o	: microtask.c
			$(CC) -p -DMCOUNT $(CFLAGS) -c microtask.c ; \
			$(LD)  -r microtask.o -o profiled/microtask.o ; \
			$(CC) $(CFLAGS) -c microtask.c ; \
			mv microtask.o microtask.O; \
			$(LD)  -r microtask.O -o microtask.o ; \
			rm -f microtask.O

shared_flag.o	: shared_flag.c
			$(CC) -p -DMCOUNT $(CFLAGS) -c shared_flag.c ; \
			$(LD)  -r shared_flag.o -o profiled/shared_flag.o ; \
			$(CC) $(CFLAGS) -c shared_flag.c ; \
			mv shared_flag.o shared_flag.O; \
			$(LD)  -r shared_flag.O -o shared_flag.o ; \
			rm -f shared_flag.O

_nargs.o	: _nargs.c
			$(CC) -p -DMCOUNT $(CFLAGS) -c _nargs.c ; \
			$(LD)  -r _nargs.o -o profiled/_nargs.o ; \
			$(CC) $(CFLAGS) -c _nargs.c ; \
			mv _nargs.o _nargs.O; \
			$(LD)  -r _nargs.O -o _nargs.o ; \
			rm -f _nargs.O

ppinit.o	: ppinit.c
			$(CC) -p -DMCOUNT $(CFLAGS) -c ppinit.c ; \
			$(LD)  -r ppinit.o -o profiled/ppinit.o ; \
			$(CC) $(CFLAGS) -c ppinit.c ; \
			mv ppinit.o ppinit.O; \
			$(LD)  -r ppinit.O -o ppinit.o ; \
			rm -f ppinit.O

prbrk.o		: prbrk.c 
			$(CC) -p -DMCOUNT $(CFLAGS) -c prbrk.c ; \
			$(LD)  -r prbrk.o -o profiled/prbrk.o ; \
			$(CC) $(CFLAGS) -c prbrk.c ; \
			mv prbrk.o prbrk.O; \
			$(LD)  -r prbrk.O -o prbrk.o ; \
			rm -f prbrk.O

team.o	: team.c 
			$(CC) -p -DMCOUNT $(CFLAGS) -c team.c ; \
			$(LD)  -r team.o -o profiled/team.o ; \
			$(CC) $(CFLAGS) -c team.c ; \
			mv team.o team.O; \
			$(LD)  -r team.O -o team.o ; \
			rm -f team.O

m_get_deflt.o	: m_get_deflt.c 
			$(CC) -p -DMCOUNT $(CFLAGS) -c m_get_deflt.c ; \
			$(LD)  -r m_get_deflt.o -o profiled/m_get_deflt.o ; \
			$(CC) $(CFLAGS) -c m_get_deflt.c ; \
			mv m_get_deflt.o m_get_deflt.O; \
			$(LD)  -r m_get_deflt.O -o m_get_deflt.o ; \
			rm -f m_get_deflt.O

f_interface.o	: f_interface.c 
			$(CC) -p -DMCOUNT $(CFLAGS) -c f_interface.c ; \
			$(LD)  -r f_interface.o -o profiled/f_interface.o ; \
			$(CC) $(CFLAGS) -c f_interface.c ; \
			mv f_interface.o f_interface.O; \
			$(LD)  -r f_interface.O -o f_interface.o ; \
			rm -f f_interface.O

c_interface.o	: c_interface.c 
			$(CC) -p -DMCOUNT $(CFLAGS) -c c_interface.c ; \
			$(LD)  -r c_interface.o -o profiled/c_interface.o ; \
			$(CC) $(CFLAGS) -c c_interface.c ; \
			mv c_interface.o c_interface.O; \
			$(LD)  -r c_interface.O -o c_interface.o ; \
			rm -f c_interface.O

sbarrier.o	: sbarrier.c 
			$(CC) -p -DMCOUNT $(CFLAGS) -c sbarrier.c ; \
			$(LD)  -r sbarrier.o -o profiled/sbarrier.o ; \
			$(CC) $(CFLAGS) -c sbarrier.c ; \
			mv sbarrier.o sbarrier.O; \
			$(LD)  -r sbarrier.O -o sbarrier.o ; \
			rm -f sbarrier.O

shbrk.o		: shbrk.c 
			$(CC) -p -DMCOUNT $(CFLAGS) -c shbrk.c ; \
			$(LD)  -r shbrk.o -o profiled/shbrk.o ; \
			$(CC) $(CFLAGS) -c shbrk.c ; \
			mv shbrk.o shbrk.O; \
			$(LD)  -r shbrk.O -o shbrk.o ; \
			rm -f shbrk.O

shmalloc.o	: shmalloc.c 
			$(CC) -p -DMCOUNT $(CFLAGS) -c shmalloc.c ; \
			$(LD)  -r shmalloc.o -o profiled/shmalloc.o ; \
			$(CC) $(CFLAGS) -c shmalloc.c ; \
			mv shmalloc.o shmalloc.O; \
			$(LD)  -r shmalloc.O -o shmalloc.o ; \
			rm -f shmalloc.O

slock.o		: slock.c 
			$(CC) -p -DMCOUNT $(CFLAGS)  -c slock.c ; \
			$(LD)  -r slock.o -o profiled/slock.o ; \
			$(CC) $(CFLAGS)  -c slock.c ; \
			mv slock.o slock.O; \
			$(LD)  -r slock.O -o slock.o ; \
			rm -f slock.O

#
#	Don't want these profiled!!!
#
_mmap.o		: _mmap.s
			$(CPP) -P -UMCOUNT $(INCLUDE) $*.s | sed -f pps.sed \
				| $(AS) -o profiled/$*.o ; \
			$(CPP) -P -UMCOUNT $(INCLUDE) $*.s | sed -f pps.sed \
				| $(AS) -o $*.o

profbrk.o	: profbrk.c 
			$(CC) $(CFLAGS) -c profbrk.c ; \
			mv profbrk.o profbrk.O; \
			$(LD)  -r profbrk.O -o profbrk.o ; \
			cp profbrk.o profiled ; \
			rm -f profbrk.O
FRC:
