/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ident "$Header: ppinit.c 1.4 89/06/17 $"


/*
 * ppinit.c
 *	Parallel program run-time environment initialization.
 *
 * -DOPENT ==> system supports "get RO text fd" primitive.
 *
 * TODO:
 *	Allow user to somehow specify where to create shared-data file,
 *		and not unlink it (for later rendezvous).
 *	Worry about taking signal in sbrk()/shmbrk() (use process local
 *		signal stack; only really issue for multi-tasker).
 */

#include <a.out.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/mman.h>
#include <time.h>
#include <sys/resource.h>

#include <sys/pmap.h>
#include <sys/param.h>
#include <sys/vmparam.h>

#include "parallel.h"
#include "parc.h"

extern	char	*strchr();
extern	char	*getenv();
extern  char	*_shared_end;
extern int _priv_stack_limit;
/*
 * Note require that stack_limit and STACK_HOLE values all be multiples
 * of STACK_INCR
 */
#define	STACK_INCR	(64*1024)		/* stack growth granularity */
int _stack_incr;
#define STACK_HOLE	(0)		/* gap between stacks */
int _stack_hole;
int _stack_limit;

/*
 * Define "_curbrk" to satisfy mmap()'s ref without
 * requiring libc's sbrk().
 */

char	*_curbrk = NULL;

#ifndef	lint
/*
 * Must insure we pull in run-time's version of sbrk().  If program
 * doesn't directly reference, will get libc's sbrk() due to malloc().
 *
 * Could just include the module, but that's less modular.
 */
extern	char	*sbrk();
static	char	*(*sbrkp)() = sbrk;
#endif

/*
 * _ppinit()
 *	Parallel startup for C programs.
 *
 * Called by crt0.o before main().  Arranges shared parts of address space
 * to be shared and initialize shared-memory allocator.
 *
 * Leaves shared-data file-descriptor in _shm_fd.
 */

extern int _do_shared_stack;

int	_shm_fd;				/* fd for shared data */
int	_stk_fd;				/* fd used for parent stack */
int	_pgoff;

static private int *_new_sp;
private int *_priv_sp;
private int _single_stack = 1;

static	char	shm_name[] = " shared data XXXXXX";
static	char	stk_name[] = " stack data XXXXXX";
static int	szinitsh;		/* sizeof init'd shared */
static int	shdoff;			/* file offset of init'd shared */
static int	szshbss;		/* sizeof shared bss */
/*ARGSUSED*/
_ppinit(argc, argv)
	int	argc;
	char	**argv;

{
	register int	szshared;		/* sizeof all shared */
	extern	shared char sshdata, eshdata, eshbss; /* shared bounds */
	extern int *_init_shstk();
	struct rlimit priv_stack_lim;
	register struct scnhdr *scnhdr;
	int	tfd;

	/*
	 * If no shared stuff, nothing to do.
	 */

	if (&sshdata == NULL)
		return;

	_pgoff = getpagesize() - 1;

	/*
	 * Create unnamed "shared" file to map.
	 */

	_shm_fd = _mkuntmp(shm_name);
	if (_shm_fd < 0)
		bad_init("error creating unnamed temp file", errno);

	/*
	 * Map file in at appropriate place.
	 */

	if ((tfd = get_ro_text(argv[0])) < 0) {
#ifdef DEBUG
		printf("get_ro_text\n");
#endif
		bad_init(argv[0], errno);
	}

	szshared = ((int)PGRND(szinitsh + szshbss));
	if (MMAP(_shm_fd, &sshdata, szshared, 0) < 0)  {
		errno = ENOEXEC;
		bad_init("shared region at invalid address", ENOEXEC);
	}
	_shared_end = &sshdata + szshared;

	/*
	 * If there is initialized shared data, find and read it.
	 */

	if (szinitsh != 0) {

		if (lseek(tfd, (off_t)shdoff, 0) == (off_t)-1) {
			bad_init(argv[0], ENOEXEC);
		}
		if (read(tfd, &sshdata, szinitsh) != szinitsh) {
#ifdef DEBUG
			printf("read in shared data %x\n", szinitsh);
#endif
			bad_init(argv[0], ENOEXEC);
		}

		(void) close(tfd);
	}

	/*
	 * get shared and private stack configuration information,
	 * do any rounding that's needed.
	 */

	if (_stack_incr == 0)
		_stack_incr = STACK_INCR;
	if (_stack_hole == 0)
		_stack_hole = STACK_HOLE;
	/* (void) getrlimit(RLIMIT_STACK, &priv_stack_lim); */
	if (_stack_limit == 0)
		_stack_limit = 0x100000;
	_stack_limit = (_stack_limit + _stack_incr) & (~_stack_incr);

	if (_priv_stack_limit == 0)
		_priv_stack_limit = 0x100000;

	if ( _do_shared_stack == 1 ) {
		_stk_fd = _mkuntmp(stk_name);
		if (_stk_fd < 0)
			bad_init("failure creating unnamed temp", errno);
#ifdef DEBUG
		printf("shared stack fd is %d\n", _stk_fd);
#endif
		/*
		 * allocate and initialize the shared stack segment.
		 * new_sp points to the place in the shared stack which
		 * corresponds to our current stack pointer.
		 * only the portion of the stack from &argc
		 * 
		 * Note, that we don't copy the environment and argument
		 * strings and string pointer arrays to the
		 * shared stack.  We rely on the private copies of these
		 * things to be at the same address in every child's
		 * private stack.
		 *
		 * we only copy the stack environment necessary to put
		 * allow us to exit this routine sanely back into crt0
		 * after having switched the stack pointer register to the
		 * shared stack.
		 */

 		_new_sp = _init_shstk((int *)(&argc - 2), (int *)(&argv + 1));
		if ((int)_new_sp < 0) 
			bad_init(argv[0], errno);
#ifdef DEBUG
		printf("new sp %x\n", _new_sp);
#endif
#ifdef i386
		;asm("	movl	%ebp, _priv_sp");
		;asm("	movl	_new_sp,%ebp");
#endif
	}
	/*
	 * Init basic shared-memory allocator.
	 */

	if (_init_shbrk() == -1)
		bad_init("failure initializing shared heap", errno);
}

/*
 * _mkuntmp()
 *	Make an unnamed temporary file, insure we are only reference.
 */

_mkuntmp(as)
	char	*as;
{
	register char *s, *temp;
	register unsigned pid;
	register int i, j;
	register int fd;
	sigset_t newmask, oldmask;
	char name_buffer[1024];

	pid = getpid();
	for (j = 0; j < 5; j++) {

		switch(j) {
		case 0:	
			temp = getenv("TMPPATH");
			if (temp == 0)
				continue;
			(void) strcpy(name_buffer, temp);
			(void) strcat(name_buffer, "/");
			break;
		
		case 1:	
			name_buffer[0] = 0;
			break;
		case 2:
			temp = getenv("HOME");
			if (temp == 0)
				continue;
			(void) strcpy(name_buffer, temp);
			(void) strcat(name_buffer, "/");
			break;
		case 3:
			(void) strcpy(name_buffer, "/usr/tmp/");
			break;
		case 4:
			(void) strcpy(name_buffer, "/tmp/");
			break;
		}
		(void) strcat(name_buffer,as);
		s = name_buffer;
		while (*s++)
			;
		s--;
		while (*--s == 'X') {
			*s = (pid%10) + '0';
			pid /= 10;
		}
		s++;
		i = 'a';
		do {
			if (i=='z')
				break;
			*s = i++;
			(void)sigfillset(&newmask);
			(void)sigprocmask(SIG_BLOCK, &newmask, &oldmask);
			fd = open(name_buffer, O_CREAT|O_EXCL|O_RDWR, 0666);
			if (fd >= 0) 
				(void) unlink(name_buffer);
			(void)sigprocmask(SIG_SETMASK, &oldmask, (sigset_t *)NULL);
		} while((fd < 0) && (errno == EEXIST));

		if (fd >= 0) {
			break;
		}
	}

	(void)fcntl(fd, F_SETFD, 1);
	return(fd);
}

/*
 * get_ro_text()
 *	Get a RO reference to executing a.out.
 *
 * Use syscall if it exists, else determine from program name and environment.
 */

#ifdef	OPENT

static
get_ro_text() { return(approp syscall); }

#else

static
get_ro_text(prog)
	char	*prog;
{
	char	name[1024];
	register char *path;
	register char *colon;
	register int len;
	register int fd;

	if (strchr(prog, '/'))
		return(chk_text(prog));

	if ((path = getenv("PATH")) == NULL)
		path = ":/bin:/usr/bin";

	for (colon = path; colon != NULL; path = colon + 1) {
		if (colon = strchr(path, ':')) {
			len = colon - path;
			(void) strncpy(name, path, len);
			name[len] = '\0';
		} else
			(void) strcpy(name, path);
		if (name[0])
			(void) strcat(name, "/");
		(void) strcat(name, prog);
		if ((fd = chk_text(name)) > 0) {
			return(fd);
		}
	}

	errno = ENOENT;
	return(-1);
}

/*
 * try to check that this is at least an executable a.out format file!!
 *	return an open file descriptor of true.
 *	otherwise, return -1, with errno set to an appropriate error code.
 */

static
chk_text(name)
char *name;
{
	struct exhdr {
		struct filehdr fhdr;
		struct aouthdr ahdr;
		struct scnhdr	shdr[8];
	} exhdr;

	struct scnhdr *scnp;
	int fd, i, txstart, sthdr;
	

	if (access(name, R_OK|X_OK))
		return(-1);

	if ((fd = open(name, O_RDONLY)) < 0) 
		return(-1);

	if (read(fd, (char *)&exhdr, sizeof(exhdr)) < sizeof(exhdr))  {
		errno = ENOEXEC;
		return(-1);
	}

	scnp = exhdr.shdr;
	for (i = 0; i < exhdr.fhdr.f_nscns; i++) {
		switch (scnp[i].s_flags) {
		case STYP_TEXT + STYP_REG:
			txstart = scnp[i].s_vaddr;
			break;
		case STYP_DATA + STYP_SHARED:
			shdoff = scnp[i].s_scnptr;
			szinitsh = scnp[i].s_size;
			break;
		case STYP_BSS + STYP_SHARED:
			szshbss = scnp[i].s_size;
			break;
		}
	}

	/*
	 * compare the file header read from the file with
	 * what's in memory.  If they match, must be the same file.
	 * otherwise, not.
	 * Find the start of the header by striping of a quess at
	 * the size of all the headers (gag, choke, coff).
	 */
	sthdr = txstart & 0xfffff000;
	scnp = exhdr.shdr;
	if (memcmp((char *)sthdr, (char *)&exhdr, sizeof(exhdr)))
	{
		errno = ENOEXEC;
		return(-1);
	}

	return(fd);
}

#endif

/*
 * bad_init()
 *	For some reason, couldn't init -- complain and exit with error status.
 */

static
bad_init(msg, err)
	char	*msg;
	int	err;
{
	perror(msg);
	_exit(err);
}


void
_exit(status)
int status;
{
	_lib_exit(status);
}


int
sigaction(sig, act, oact)
int sig;
struct sigaction *act, *oact;
{
		return(_lib_sigaction(sig, act, oact));
}

void
(*sigset(sig, act))()
int sig;
void (*act)();
{
	struct sigaction newact, oldact;

	newact.sa_handler = act;
	newact.sa_flags = 0;
	sigemptyset(&newact.sa_mask);
	if (sig == (int) SIG_HOLD) {
		sigaddset(&newact.sa_mask, sig);
		sigaction(sig, (struct sigaction *)NULL, &oldact);
		sigprocmask(SIG_BLOCK, &newact.sa_mask, (sigset_t)NULL);
		return(oldact.sa_handler);
	} else {
		sigprocmask(SIG_BLOCK, (sigset_t)NULL, &oldact.sa_mask);
		if (sigismember(&oldact.sa_mask, sig)) {
			sigemptyset(&oldact.sa_mask);
			sigaddset(&oldact.sa_mask, sig);
			sigprocmask(SIG_UNBLOCK, &oldact.sa_mask);
		}
		_lib_sigaction(sig, &newact, &oldact);
		return(oldact.sa_handler);
	}
}
