/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ident "$Header: mntent.c 1.4 89/08/07 $"

/* @(#)mntent.c 1.1 86/02/03 SMI */
/* @(#)mntent.c	2.1 86/04/11 NFSSRC */

/*
 * mntent.c
 *	Read/parse and write /etc/fstab & /etc/mnttab and their ilk.
 */

/* $Log:	mntent.c,v $
 */

#include <stdio.h>
#include <ctype.h>
#include <mntent.h>
#include <string.h>
#include <sys/fcntl.h>

static	struct mntent cur_mnt;
static	char line[BUFSIZ+1];

static char *
mntstr(p)
	register char **p;
{
	char *cp = *p;
	char *retstr;

	while (*cp && isspace(*cp))
		cp++;
	retstr = cp;
	while (*cp && !isspace(*cp))
		cp++;
	if (*cp) {
		*cp = '\0';
		cp++;
	}
	*p = cp;
	return (retstr);
}

static long
mntdigit(p)
	register char **p;
{
	register long value = 0;
	char *cp = *p;

	while (*cp && isspace(*cp))
		cp++;
	for (; *cp && isdigit(*cp); cp++) {
		value *= 10;
		value += *cp - '0';
	}
	while (*cp && !isspace(*cp))
		cp++;
	if (*cp) {
		*cp = '\0';
		cp++;
	}
	*p = cp;
	return (value);
}

static
mnttabscan(mnttabp, mnt)
	FILE *mnttabp;
	struct mntent *mnt;
{
	char *cp;

	mnt->mnt_fsname = "";
	mnt->mnt_dir    = "";
	mnt->mnt_type   = "";
	mnt->mnt_opts   = "";
	mnt->mnt_time   = 0;

loop:
	do {
		cp = fgets(line, 256, mnttabp);
		if (cp == NULL) {
			return (EOF);
		}
	} while (*cp == '#');
	mnt->mnt_fsname = mntstr(&cp);
	if (*mnt->mnt_fsname == '\0')
		goto loop;	/* ignore blank lines */
	if (*cp == '\0')
		return (1);
	mnt->mnt_dir = mntstr(&cp);
	if (*cp == '\0')
		return (2);
	mnt->mnt_type = mntstr(&cp);
	if (*cp == '\0')
		return (3);
	mnt->mnt_opts = mntstr(&cp);
	if (*cp == '\0')
		return (4);
	mnt->mnt_time = (time_t) mntdigit(&cp);
	return (5);
}
	
FILE *setmntent(fname, flag)
	char *fname;
	char *flag;
{
	FILE *mnttabp;
	struct flock lock;

	if ((mnttabp = fopen(fname, flag)) == NULL) {
		return (NULL);
	}
	lock.l_type = F_RDLCK;
	lock.l_whence = 0;
	lock.l_start = 0;
	lock.l_len = 0;

	while (*flag) {
		if (*flag == 'w' || *flag == 'a' || *flag == '+') {
			lock.l_type = F_WRLCK;
		}
		flag++;
	}
	if (fcntl(fileno(mnttabp), F_SETLKW, &lock) < 0) {
		(void) fclose(mnttabp);
		return (NULL);
	}
	return (mnttabp);
}

int
endmntent(mnttabp)
	FILE *mnttabp;
{

	if (mnttabp) {
		(void) fclose(mnttabp);
	}
	return (1);
}

struct mntent *
getmntent(mnttabp)
	FILE *mnttabp;
{
	int nfields;

	if (mnttabp == 0)
		return ((struct mntent *)0);
	nfields = mnttabscan(mnttabp, &cur_mnt);
	if (nfields == EOF || nfields < 4)
		return ((struct mntent *)0);
	return (&cur_mnt);
}

addmntent(mnttabp, mnt)
	FILE *mnttabp;
	register struct mntent *mnt;
{
	if (fseek(mnttabp, 0, 2) < 0) {
		return (1);
	}
	(void) mntprtent(mnttabp, mnt);
	return (0);
}

static char tmpopts[256];

static char *
mntopt(p)
	char **p;
{
	char *cp = *p;
	char *retstr;

	while (*cp && isspace(*cp))
		cp++;
	retstr = cp;
	while (*cp && *cp != ',')
		cp++;
	if (*cp) {
		*cp = '\0';
		cp++;
	}
	*p = cp;
	return (retstr);
}

char *
hasmntopt(mnt, opt)
	register struct mntent *mnt;
	register char *opt;
{
	char *f, *opts;

	(void) strcpy(tmpopts, mnt->mnt_opts);
	opts = tmpopts;
	f = mntopt(&opts);
	for (; *f; f = mntopt(&opts)) {
		if (strncmp(opt, f, strlen(opt)) == 0)
			return (f - tmpopts + mnt->mnt_opts);
	} 
	return (NULL);
}

static
mntprtent(mnttabp, mnt)
	FILE *mnttabp;
	register struct mntent *mnt;
{
	(void) fprintf(mnttabp, "%s %s %s %s %d\n",
	    mnt->mnt_fsname,
	    mnt->mnt_dir,
	    mnt->mnt_type,
	    mnt->mnt_opts,
	    mnt->mnt_time);
	return(0);
}
