/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

/*	Copyright (c) 1984 AT&T	*/
/*	  All Rights Reserved  	*/

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/

#ident "$Header: clone.c 1.21 90/03/06 $"

/* $Log:	clone.c,v $
 */

/*
 * Clone Driver.
 */


#include "sys/types.h"
#include "sys/devsw.h"
#include "sys/mutex.h"
#include "sys/intctl.h"
#include "sys/file.h"
#include "sys/sysmacros.h"
#include "sys/param.h"
#include "sys/errno.h"
#include "sys/signal.h"
#include "sys/user.h"
#include "sys/conf.h"
#include "sys/debug.h"
#include "sys/session.h"
#include "sys/stream.h"
#include "sys/stropts.h"
#include "sys/strstat.h"
#include "sys/cmn_err.h"
#include "sys/buf.h"
#include "sys/vfs.h"
#include "sys/vnode.h"
#include "sys/timer.h"
#include "sys/proc.h"
#include "sys/uio.h"
#include "sys/cmap.h"
#include "sys/ufsinode.h"
#include "sys/ufsfilsys.h"
#include "sys/ufsmount.h"
#include "sys/ufsdir.h"
#include "sys/snode.h"
#include "sys/vm.h"		/* for plocal */
#include "sys/var.h"
#include "sys/poll.h"
#include "sys/strid.h"
#include "sys/systm.h"
#include "sys/pte.h"		/* for plocal */
#include "sys/plocal.h"		/* for l struct */

/*
 * Orginal System V.3 clone on supported one unit per clone device.
 * DYNIX/ptx(R) supports multiple units by encoding the unit number
 * in the high eight bits.
 */

#define   CLONEMAJOR(x)	((int)((x) & 0xff))
#define   CLONEMINOR(x)	((int)(((unsigned)(x)>>8)&0xff))
#define	  CLONEMAKDEV(x,y) ((unsigned) (((y)<<8) | (x)))

int cloneopen();
static struct module_info clone_info = { STRID_CLONE, "CLONE", 0, 0, 0, 0 };
static struct qinit clonerinit = { NULL, NULL, cloneopen, NULL, NULL, &clone_info, NULL };
static struct qinit clonewinit = { NULL, NULL, NULL, NULL, NULL, &clone_info, NULL };
struct streamtab cloneinfo = { &clonerinit, &clonewinit };


extern int clone_c_major;		/* in conf.c */
extern struct devops cloneops;


#ifdef DEBUG
int cdebug = 0;
#define CL_PRINTF(arg)	if (cdebug) printf(arg); else
#else
#define CL_PRINTF(arg)
#endif


/*
 * Clone open.  Dev is the major device of the streams device to open.
 * Look up the device in the cdevsw[].  Attach its qinit structures to
 * the read and write queues and call its open with the q and the dev
 * CLONEOPEN.  Swap in a new vnode with the real device number
 * constructed from the major and the minor returned by the device
 * open.
 *
 * As far as the locking environment is concerned, this routine
 * assumes that sd_plumb sema for this stream is held and thus this is
 * an exclusive access to this stream.  Because of this, clone opens
 * are always "first" opens (since at the end the stream belongs to
 * another vnode and subsequent clone opens will have nothing to do
 * with this stream).  It is also assumed that sd_mutex is not held by
 * the caller.
 */

cloneopen(q, dev, flag, sflag)
	queue_t *q;
{
	int mindev;
	dev_t cdev, rdev;
	register struct streamtab *st;
	struct dev *dp;
	struct file *fp;
	struct vnode *nvp;
	struct snode *sp, *nsp;
	extern struct snode *getspstr();
	extern lock_t tbl_mutex;
	register spl_t spl;

	/*
	 * Get the device to open.
	 */
	if (sflag) {
		CL_PRINTF("cloneopen: FAILING sflag \n");
		return(OPENFAIL);
	}
	/*
	 * Determine device being cloned by extracting bits from
	 * minor device number.
	 */
	cdev = CLONEMAJOR(minor(dev));

	dp = dev_ref(makedev(cdev, 0));
	if (!dp || !(st = DEV_STREAMTAB(dp))) {
		dev_deref(dp);
		u.u_error = ENXIO;
		CL_PRINTF("cloneopen: FAILING i range \n");
		return(OPENFAIL);
	}
	dev_deref(dp);

		
	/*
	 * Substitute the real qinit values for the current ones
	 */
	setq(q, st->st_rdinit, st->st_wrinit);

	/*
	 * Build the device argument for the device open routine
	 */
	rdev = makedev(cdev, CLONEMINOR(minor(dev)));

	/*
	 * Call the device open with the stream flag  CLONEOPEN.  The device
	 * will either fail this or return a minor device number.
	 */
	CL_PRINTF("cloneopen: calling actual open\n");
	if ((mindev = (*q->q_qinfo->qi_qopen)(q, rdev, flag, CLONEOPEN)) == OPENFAIL) {
		CL_PRINTF("cloneopen: actual open FAILED\n");
		/* dont have to undo setq since stream will deallocated in stropen anyway. */
		return(OPENFAIL);
	}
	rdev = makedev(cdev, mindev);
	CL_PRINTF("cloneopen: actual open succeeded\n");

	/*
	 * Allocate a new in-core snode.
	 * cdevvp returns a locked vnode with a reference count of one.
	 */
	nvp = cdevvp(rdev, u.u_uid, u.u_gid);
	/*
	 * Get the fp, sp at top of this stream.
	 */
	fp = u.u_openfp;
	sp = getspstr(fp);
	ASSERT_DEBUG(fp != NULL && sp->s_sptr == q->q_str,
		     "cloneopen: u.u_openfp");

	/* 
	 * Exchange the new vnode for the old one in the setup.
	 */
	CL_PRINTF("cloneopen: switching snodes\n");

	nsp = VTOS(nvp); 	/* already locked */
	/*
	 * Make reference to device returned from driver.
	 * Note: should succeed unless driver forgot to do devsw_add,
	 *   	unwinding for a buggy device driver seems too difficult.
	 */
	nsp->s_devp = nvp->v_rdevp = dev_ref(rdev);
	if (nsp->s_devp == NULL) {
		cmn_err(CE_PANIC, "cloneopen: no dev ref for %x\n", rdev);
		/*
		 *+ The driver called from clone did not create references
		 *+ for the actual devices in the device table.
		 *+ Corrective action:  fix the device driver that is using
		 *+ the clone driver.
		 */
	}

	VN_LOCKNODE(STOV(sp));
	spl = p_lock(&tbl_mutex, SPLSTR);

	nsp->s_sptr = sp->s_sptr;
	sp->s_sptr = NULL;	/* no other clone open will get this stream */
	nsp->s_sptr->sd_rdev = rdev;
	nsp->s_sptr->sd_strtab = st;
	fp->f_vnode = nvp;   /* no need to lock fp */
	rehashstr(nsp->s_sptr); /* rehash since minor has changed */
	v_lock(&tbl_mutex, spl);

	/*
	 * We are in essence closing the clone driver snode and opening
	 * the new snode.  Since spec_close() is not getting called, we
	 * need to fake out releasing the reference.
	 */
	dev_deref(sp->s_devp);

	/*
	 * This is a temporary workaround that permits higher level
	 * routines (e.g., spec_open()) to know if a cloneopen(),
	 * rather than a direct device open, has occurred.  This
	 * will be the pointer to the new dev structure returned
	 * by cloneopen().  u_clonedev only has meaning in the
	 * spec_open() routine.  This workaround should be removed
	 * when mailbug #31990 is addressed.
	 */
	u.u_clonedev = nsp->s_devp;


	VN_PUT(STOV(sp));
	VN_UNLOCKNODE(nvp);

	CL_PRINTF("cloneopen: end\n");
	return(mindev);
}	

/*
 * clonedev_add - add a clone device to the devsw data structure.
 *
 * This routine builds a clone device entry and calls devsw_add;
 * it returns the result of devsw_add. 
 * devsw_add will fail if trying to add a dev which has already been added.
 */
clonedev_add(cdev)
	dev_t cdev;
{
	register struct dev *dp;
	int cminor = CLONEMAKDEV(major(cdev), minor(cdev));

	dp = (struct dev *) calloc(sizeof(struct dev));
	dp->d_dev = makedev(clone_c_major, cminor);
	dp->d_ops = &cloneops;

	return devsw_add(dp);
}

