/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ident	"$Header: cmpt.c 1.39 90/05/03 $"

/*
 * cmpt.c
 *	4.2 non-VTOC disk compatibility driver
 */

/* $Log:	cmpt.c,v $
 */


#include "sys/types.h"
#include "sys/devsw.h"
#include "sys/ioctl.h"
#include "sys/sysmacros.h"
#include "sys/param.h"
#include "sys/intctl.h"
#include "sys/errno.h"
#include "sys/buf.h"
#include "sys/uio.h"
#include "sys/signal.h"
#include "sys/user.h"
#include "sys/file.h"
#include "sys/vnode.h"
#include "sys/mutex.h"
#include "sys/vtoc.h"
#include "sys/autoconf.h"
#include "sys/cmn_err.h"
#include "sys/debug.h"
#include "sys/cmpt.h"
#include "sys/genalloc.h"
#include "sys/systm.h"

#define PCMPT   (PZERO - 1)             /* not signalable */

#define	CMPT_ALLOC_PRIV()	genalloc(&cmpt_privlist, 0)
#define CMPT_FREE_PRIV(P)	genfree(&cmpt_privlist, (P))

#define	DEVTOCMPTPRIV(d)	(struct cmpt_private *)(d)->d_private

extern	int	cmpt_c_major;		/* major no of cmpt device */
extern	int	cmpt_maxtypes;		/* no of configured part tables */
extern	struct	cmptsize *cmptparts[];	/* partition tables */

/*
 * memory management for private data associated with devs
 */

static	struct	genallocator cmpt_privlist;

static	sema_t	cmpt_sema;		/* exclusive access cmpt data strcts*/

static	struct	cmptdt cmptdt;		/* V_BUILD ioctl data */

static	caddr_t	cmptall();

extern	struct dev *fdtodev();

extern	int	cmptopen();
extern	int	cmptclose();
extern	int	cmptstrat();
extern	int	cmptsize();
extern	int	cmptread();
extern	int	cmptwrite();
extern	int	cmptioctl();
extern	int	cmpt_shutdown();
extern	int	cmptdiagstrat();
extern	struct	devops cmptops;

extern	int	cmpt_mem_init;
extern	int	cmpt_mem_click;
extern	int	mount_readonly;

/*
 * cmpt_init - initilize the driver
 * this device is represented by a single dev in devsw.
 */
/*ARGSUSED*/
void
cmpt_init(ndevs)
	int ndevs;
{
	struct dev *newdev;

	newdev = (struct dev *)calloc(sizeof(struct dev));
	newdev->d_dev = makedev(cmpt_c_major, 0);
	newdev->d_ops = &cmptops;
	if (devsw_add(newdev) != 0) {
		cmn_err(CE_WARN, "cmpt: problem adding dev to devsw");
		/*
		 *+ The kernel discovered a problem when attempting to add the cmpt
		 *+ device driver to the system.  This is probably
		 *+ due to an error in driver code.
		 */
	}

	/*
	 * We need to manage the memory used for private data in cmpt
	 * devs.  Since the number of partitions per disk is unknown
	 * at boot time (and thus, the amount of memory needed for private
	 * data), we use genalloc routines to manage the memory.  Could
	 * as easily use the heap when it comes along.
	 */

	if (geninit(&cmpt_privlist, sizeof(struct cmpt_private), cmpt_mem_init,
		calloc, cmpt_mem_click, cmptall) != 0) {

		cmn_err(CE_WARN,"cmpt: problem initializing private data pool");
		/*
		 *+ The kernel discovered a problem when trying to allocate memory
		 *+ for the cmpt driver's private data pool.
		 */
		(void)devsw_remove(newdev);
		return;
	}

	init_sema(&cmpt_sema, 1);
}
/*
 * cmptall - memory allocator used to increment cmpt_privlist
 *
 * As we allocate memory for private data, we may exhaust the freelist.
 * If that's the case, call cmptall to allocate more.
 */
static caddr_t
cmptall(nbytes)
	int	nbytes;
{
	return(wmemall(nbytes, 1));
}


/*
 * cmptopen - open the compatibility partition.  If we are mapped to
 * a device, call its open routine.  Otherwise, we have yet to be
 * bound to a device, so return success.  Note that we return ENXIO if
 * the partition length is zero.  Note also that we honor "soft"
 * write-protect.
 */
cmptopen(dp, mode, type)
	struct dev *dp;
	int	mode;
	int	type;
{
	int	error;
	struct cmpt_private *dsp = DEVTOCMPTPRIV(dp);
	struct dev *dpp;

	if (dsp != NULL && (dpp = dsp->cp_physdev)) {
		if (dsp->cp_size <= 0) {
			return(ENXIO);
		}
		if ((mode & FWRITE) && (dsp->cp_flags & VP_WRPROTECT))
			return(EBUSY);
		error = DEV_OPEN(dpp, mode, type);
		return(error);
	}
	return(0);
}

/*
 * cmptclose - call device's close routine if we are bound to one
 */
cmptclose(dp, flag, type)
	struct dev *dp;
	int	flag;
	int	type;
{
	struct cmpt_private *dsp = DEVTOCMPTPRIV(dp);
	struct dev *dpp;

	if (dsp != NULL && (dpp = dsp->cp_physdev)) {
		DEV_CLOSE(dpp, flag, type);
	}
}

/*ARGSUSED*/
cmptread(dp, uio, flags)
        struct dev *dp;
        struct uio *uio;
{
        int err, diff;
        off_t lim;
	struct cmpt_private *dsp = DEVTOCMPTPRIV(dp);

	lim = dsp->cp_size;
	lim <<= DEV_BSHIFT;
	err = physck(lim, uio, B_READ, &diff);
	if (err != 0) {
		if (err == -1)  /* not an error, but request of 0 bytes */
			err = 0;
		return (err);
	}
	if (flags & IO_ASYNC) {
		err = aphysio(dp, B_READ, uio, diff, flags);
	} else {
		err = physio(dp, (struct buf *)0, B_READ, uio);
		uio->uio_resid += diff;
	}
	return (err);
}

/*ARGSUSED*/
cmptwrite(dp, uio, flags)
        struct dev *dp;
        struct uio *uio;
{
        int err, diff;
        off_t lim;
	struct cmpt_private *dsp = DEVTOCMPTPRIV(dp);

	lim = dsp->cp_size;
	lim <<= DEV_BSHIFT;
	err = physck(lim, uio, B_WRITE, &diff);
	if (err != 0) {
		if (err == -1)  /* not an error, but request of 0 bytes */
			err = 0;
		return (err);
	}
	if (flags & IO_ASYNC) {
		err = aphysio(dp, B_WRITE, uio, diff, flags);
	} else {
		err = physio(dp, (struct buf *)0, B_WRITE, uio);
		uio->uio_resid += diff;
	}
	return (err);
}
	
/*
 * cmptstrat - Do the actual partition checking/offsetting.
 * No action will be taken on I/O completion, so the buf header
 * passed as an argument can be safely reused.  Note that the b_blkno
 * must have its value preserved since the buf cache code expects it
 * to remain unchanged.
 */
cmptstrat(dp, bp, offset)
	struct	dev *dp;
	struct	buf *bp;
	daddr_t	    offset;
{
	struct cmpt_private *dsp = DEVTOCMPTPRIV(dp);
	daddr_t start = dsp->cp_start;

	bp->b_status = B_STATUS_OK;	  /* default status */

	if ((bp->b_blkno + offset + (bp->b_bcount >> DEV_BSHIFT))
							> dsp->cp_size) {
		bp->b_flags |= B_ERROR;
		bp->b_error = EINVAL;
		bp->b_resid = bp->b_bcount;
		biodone(bp);
		return;
	}

	DEV_STRATEGY(dsp->cp_physdev, bp, offset + start);
}

/*
 * cmptioctl - handle I/O control requests. 
 *
 * valid commands:
 *      V_BUILD  - construct a set of compatibility partitions
 *      V_DESTROY - deconstruct the partitions created in V_BUILD
 *      V_WRITEPROTECT - soft write-protect prevents writes on raw device
 *      V_WRITEENABLE - reenable writes on raw device
 */
/*ARGSUSED*/
cmptioctl(dp, cmd, data, flag)
	struct	dev *dp;
	int	cmd;
	caddr_t	data;
	int	flag;
{
	struct	dev *dpp;
	struct	cmptdt *cd = &cmptdt;
	struct	cmpt_private *dsp;
	int	error, type;

	/*
	 * ioctls which may only be done by root
	 */
	if (!suser()) {
		switch(cmd) {
		case V_BUILD:
		case V_DESTROY:
		case V_WRITEPROTECT:
		case V_WRITEENABLE:
			return(EPERM);
		}
	}

	if (dp->d_private == NULL) {

		/*
		 * ioctls which must be done on a cmpt dev which is bound to
		 * a device.
		 */

		switch(cmd) {

		case V_WRITEPROTECT:
		case V_WRITEENABLE:
			return(EINVAL);
		}
	} else {

		/*
		 * ioctls which must be done on the cmpt dev which
		 * is not bound to any device.
		 */

		switch(cmd) {

		case V_BUILD:
		case V_DESTROY:
			return(EINVAL);
		}
	}

	switch(cmd) {

	case V_BUILD:
		if (error = copyin(data, (caddr_t)cd, sizeof(struct cmptdt)))
			return(error);
		type = cd->cd_type;
		if (type > cmpt_maxtypes || type < 0)
			return(EINVAL);

		dpp = fdtodev(cd->cd_fd);
		if (!dpp)
			return(EINVAL);

		error = cmpt_vbuild(dp, dpp, type);
		if (error == 0) {
			fdclose(cd->cd_fd);
			error = u.u_error;
		}

		return(error);


	case V_DESTROY:
		return(cmpt_vdestroy(dp, (int)data));
	
	case V_WRITEPROTECT:
		
		if (mount_readonly == 0)
			return(0);
		/*
		 * disallow write-protecting the diag device
		 */
		if ((dp->d_dev & (DEV_DIAG_MASK|V_RAW_MINOR)) ==
		    (DEV_DIAG_MASK|V_RAW_MINOR))
			return(EINVAL);

		/*
		 * All further writes to the raw device will fail.
		 */
		dsp = DEVTOCMPTPRIV(dp);
		dsp->cp_flags |= CP_WRPROTECT;
		return(0);
		

	case V_WRITEENABLE:

		if (mount_readonly == 0)
			return(0);
		/*
		 * reenable writes to the raw device
		 */
		if ((dp->d_dev & (DEV_DIAG_MASK|V_RAW_MINOR)) ==
		    (DEV_DIAG_MASK|V_RAW_MINOR))
			return(EINVAL);

		dsp = DEVTOCMPTPRIV(dp);
		dsp->cp_flags &= ~CP_WRPROTECT;
		return(0);

#ifdef DEBUG
	case V_IOCTL:
		error = DEV_IOCTL( ((struct cmpt_private *)
				    dp->d_private)->cp_physdev,
				  data, (caddr_t) 0, flag);
		return (error);
#endif /* DEBUG */
	/*
	 * unsupported commands
	 */
	case V_READ:
	case V_WRITE:
	default:
		/*
		 * For raw device, pass ioctl on to real device
		 * driver.
		 */
		if ((dp->d_dev & (DEV_DIAG_MASK|V_RAW_MINOR)) ==
		    (DEV_DIAG_MASK|V_RAW_MINOR)) {
			error = DEV_IOCTL(((struct vtoc_private *)
					  dp->d_private)->vp_physdev,
					  cmd, (caddr_t)data, flag);
			return(error);
		} else
			return(EINVAL);
	}
	/*NOTREACHED*/
}

/*
 * cmpt_vbuild - do the real work of the V_BUILD operation.
 *
 * contruct a cmpt driver onto a disk.  Care must be taken to
 * be sure that we can back out of the operation if there is
 * a consistency problem.  Note that we do not do a dev_deref
 * on the "child" device dev, since we are just inheriting the
 * reference which was established in its open operation.
 */
cmpt_vbuild(dp, dpp, type)
	struct	dev *dp, *dpp;
	int	type;
{
	int	i;
	dev_t	newdevt;
	struct	dev *tmpdev, *td;
	struct	devops *dops, *diagdops;
	struct	cmpt_private *tmppriv, *ddsp;

	if (dp == (struct dev *)0 || dpp == (struct dev *)0) {
		return(EINVAL);
	}

	/*
	 * We're picky here about the minor device number
	 * we are partitioning - insist that the minor number
	 * be one of the unpartitioned devices.  Note that we
	 * also confirm that this is not the diag dev either.
	 */
	if ((dpp->d_dev & (V_RAW_MINOR | DEV_DIAG_MASK)) != V_RAW_MINOR) {
		return(EINVAL);
	}

	/*
	 * try to remove the dev from devsw.  This will only
	 * work if we are the only reference to the dev.
	 * If others have active references, fail the operation.
	 */
	if (devsw_remove(dpp) != 0) {
		return(EBUSY);
	}
		
	/*
	 * minor numbers start with a low byte of 0.
	 */
	newdevt = (dpp->d_dev & ~(V_RAW_MINOR));

	/*
	 * Serialize to prevent tmpdev from referring to a
	 * device which has been destroyed via V_DESTROY.
	 */

	(void)p_sema(&cmpt_sema, PCMPT);

	/*
	 * for each possible new dev, there may be an existing
	 * dev already present.  First see if such devs
	 * exist and fail operation if they do.  This is really
	 * just a sanity-check, since there should not be other
	 * devs present at these major/minors.
	 */
	for (i = 0; i < CMPT_NUMPART; i++) {
		tmpdev = dev_ref(newdevt + i);
		if(tmpdev) {
			v_sema(&cmpt_sema);
			dev_deref(tmpdev);
			(void)devsw_add(dpp);
			return(EBUSY);
		}
	}

	dops = DEVOPS_ALLOC();
	*dops = *dpp->d_ops;
	dops->d_open = cmptopen;
	dops->d_close = cmptclose;
	dops->d_strategy = cmptstrat;
	dops->d_read = cmptread;
	dops->d_write = cmptwrite;
	dops->d_size = cmptsize;
	dops->d_ioctl = cmptioctl;
	dops->d_shutdown = cmpt_shutdown;

	/*
	 * allocate and initialize each partition.  Note that we seed away
	 * the size/start of the "c" partition, to initialize the diag
	 * device's private data.
	 */

	for (i=0; i < CMPT_NUMPART; i++) {
		tmpdev = DEV_ALLOC();
		tmpdev->d_ops = dops;
		tmpdev->d_private = (caddr_t)CMPT_ALLOC_PRIV();
		cmpt_newdev(tmpdev, dpp, newdevt + i, type, i);

		/*
		 * C partition
		 */
		if (i == 2)
			tmppriv = DEVTOCMPTPRIV(tmpdev);

		(void)devsw_add(tmpdev);
		dev_deref(tmpdev);
	}

	/*
	 * allocate and initialize a devops struct for the diag device.
	 */

	diagdops = DEVOPS_ALLOC();

	*diagdops = *dpp->d_ops;
	diagdops->d_open = cmptopen;
	diagdops->d_close = cmptclose;
	diagdops->d_strategy = cmptdiagstrat;
	diagdops->d_read = cmptread;
	diagdops->d_write = cmptwrite;
	diagdops->d_size = 0;
	diagdops->d_ioctl = cmptioctl;
	diagdops->d_shutdown = 0;

	/*
	 * allocate the diag dev and initialize it.  Note that the diag dev
	 * saves a copy of the devops struct used by "normal" partitions,
	 * so that it can be deallocated when the device gets destroyed.
	 * Most of the characteristics of the diag device are inherited from
	 * the "c" partition.
	 */

	td = DEV_ALLOC();
	td->d_private = (caddr_t)CMPT_ALLOC_PRIV();
	td->d_ops = diagdops;
	td->d_flags = dpp->d_flags;
	td->d_dev = newdevt | DEV_DIAG_MASK | V_RAW_MINOR;
	ddsp = DEVTOCMPTPRIV(td);
	*ddsp = *tmppriv;
	ddsp->cp_devops = dops;
	(void)devsw_add(td);
	dev_deref(td);

	/*
	 * phys dev's parent pointer is set to point to one of the
	 * cmpt devs.  This allows physical drivers to
	 * keep a pointer to their unpartitioned dev arround to
	 * use in passing up V_REMDEV interrupts.  This needs to be undone
	 * in V_DESTROY/V_REMDEV operations.
	 */
	dpp->d_parent = tmpdev;
	v_sema(&cmpt_sema);
	return(0);
}

				
/*
 * cmpt_newdev()
 *	initialize a new dev.
 * Partition data is derived from compiled-in table.
 */
static
cmpt_newdev(newdev, dp, newdevt, type, whichpart)
	struct	dev *newdev, *dp;
	dev_t	newdevt;
	int	whichpart;
{
	struct cmpt_private *dsp;

	newdev->d_flags = dp->d_flags;
	newdev->d_dev = newdevt;
	dsp = DEVTOCMPTPRIV(newdev);
	dsp->cp_driver_id = cmpt_c_major;
	dsp->cp_size = cmptparts[type][whichpart].p_size;
	dsp->cp_start = cmptparts[type][whichpart].p_start;
	dsp->cp_physdev = dp;
}

/*
 * cmpt_vdestroy()
 *	remove the cmpt driver, leaving just the raw disk.
 *
 * Note on memory management: The V_BUILD ioctl will allocate
 * memory for two devops structs: one for the diag dev, and
 * one which is shared amongst all of the partition devs.
 * The pointer to this second devops struct is stored in
 * the diagnostic dev's private data field, as an aid to freeing it.
 *
 * This ioctl should only work if applied to /dev/cmpt, with
 * any part of the device to be destroyed as the argument.
 */
static int
cmpt_vdestroy(dp, data)
	struct	dev *dp;
	int	    data;
{
	int	i;
	dev_t	devno;
	struct	dev *dpp, *argdp, *diagdp;
	struct	cmpt_private *dsp;
	int	error, argi = -1;
	struct	dev *dtab[CMPT_NUMPART];	/* devs to be destroyed */

	if (major(dp->d_dev) != cmpt_c_major) {
		return(EINVAL);
	}

	/*
	 * argdp points to dev which is behind the fd passed
	 * in as argument to ioctl.
	 */

	if ((int)data < 0 || !(argdp = fdtodev((int)data))) {
		return(EINVAL);
	}

	dsp = DEVTOCMPTPRIV(argdp);

	/*
	 * Confirm that this is a reasonable dev to destroy.  Must
	 * have "Private" dev data, and must also have the
	 * proper driver id in its private data.
	 */

	if (!dsp || dsp->cp_driver_id != cmpt_c_major) {
		return(EINVAL);
	}

	dpp = dsp->cp_physdev;

	(void)p_sema(&cmpt_sema, PCMPT);
	bzero((caddr_t)dtab, sizeof(dtab));

	/*
	 * clear any dev_t bits related to partition number or whether
	 * this is a diagnostic device
	 */
	devno = argdp->d_dev & ~V_PART_MASK;
	devno &= ~DEV_DIAG_MASK;

	/*
	 * loop through all possible partitions, grabbing
	 * references to them if they exist.
	 */
	for (i=0; i < CMPT_NUMPART; i++) {

		dtab[i] = dev_ref(devno + i);

		/*
		 * If a ref fails, this dev is being used in
		 * another mapping - fail the V_DESTROY.
		 */
		if (dtab[i] == (struct dev *)0) {
			error = EBUSY;
			goto failpop;
		}

		/*
		 * The dev which was handed in as argument needs
		 * to be handled specially.  Decrement the
		 * ref count and attempt to remove it.  If
		 * this fails, zero out the entry in dtab[], so
		 * it does not get the ref count decreased
		 * again (ufs_close() takes care of this.
		 */

		if (dtab[i] == argdp) {
			dev_deref(dtab[i]);
			if (devsw_remove(argdp) != 0) {
				dtab[i] = (struct dev *)0;
				error = EBUSY;
				goto failpop;
			} else {
				argi = i;
			}
		} else if (devsw_remove(dtab[i]) != 0) {

			/*
			 * If we can't remove a dev from devsw, it is
			 * currently in use.  Decrement its ref count,
			 * and zero out the entry in dtab[], so we don't
			 * try to add it back to devsw.
			 */

			dev_deref(dtab[i]);
			dtab[i] = (struct dev *)0;
			error = EBUSY;
			goto failpop;
		}
	}

	/*
	 * clean up the diagnostics device.
	 * Note that since fdtodev() doesn't bump the
	 * ref cnt, we don't need to dev_deref() it (unless
	 * we are trying to V_DESTROY through the diag device).
	 */

	devno |= DEV_DIAG_MASK | V_RAW_MINOR;
	diagdp = dev_ref(devno);

	if (!diagdp) {
		cmn_err(CE_WARN, "cmpt: missing diagnostics device");
		/*
		 *+ A V_DESTROY operation on the cmpt device failed
		 *+ because the diagnostic device was discovered
		 *+ to be missing.  This might indicate that a
		 *+ driver is removing the wrong device
		 *+ objects.
		 */

		error = EINVAL;
		goto failpop;
	} else {
		if (diagdp == argdp) {
			dev_deref(argdp);
		}

		if (devsw_remove(diagdp) != 0) {
			if (diagdp != argdp) {
				dev_deref(diagdp);
			}
			error = EBUSY;
			goto failpop;
		}
	}

	/*
	 * The fd passed in as an argument to the ioctl must have
	 * a close() simulated on it before recovering the memory
	 * associated with its dev.  Note that for the inode to
	 * get closed, the ref count must be > 0
	 */

	fdclose((int)data);

	/*
	 * prevent drive shutdown interrupts from following this
	 * link any more.
	 */

	dpp->d_parent = (struct dev *)0;

	/*
	 * recover memory used by removed devs.  
	 */

	for (i=0; i < CMPT_NUMPART; i++) {
		if (dtab[i]) {
			CMPT_FREE_PRIV(DEVTOCMPTPRIV(dtab[i]));
			DEV_FREE(dtab[i]);
		}
	}

	dsp = DEVTOCMPTPRIV(diagdp);
	DEVOPS_FREE(dsp->cp_devops);
	DEVOPS_FREE(diagdp->d_ops);
	CMPT_FREE_PRIV(DEVTOCMPTPRIV(diagdp));
	DEV_FREE(diagdp);

	/*
	 * clean up the physical disk device
	 */

	(void)devsw_add(dpp);
	dev_deref(dpp);

	v_sema(&cmpt_sema);
	return(0);
failpop:
	/*
	 * Add back devs which were removed from devsw.  Need to
	 * be careful to release references to these devs.  Note
	 * that we don't decrement the ref count on the dev
	 * passed in as an argument, since ufs_close() handles this.
	 */

	for (i = 0; i < CMPT_NUMPART; i++) {
		if (dtab[i]) {
			(void)devsw_add(dtab[i]);
			if (i != argi) {
				dev_deref(dtab[i]);
			}
		}
	}
	v_sema(&cmpt_sema);
	return(error);
}

/*
 *  cmpt_shutdown()
 *	Called by child when the child breaks.  The second argument is a
 * pointer to the struct dev of the child.  The first argument should be
 * a pointer to a cmpt dev
 */
/*ARGSUSED*/
cmpt_shutdown(dp, cdp)
	struct dev *dp, *cdp;
{
	struct	dev *dpp;
	dev_t	devno;
	int	i;
	
#ifdef DEBUG
	printf( "cmpt_shutdown: dp @ 0x%x, dev 0x%x, cdp @ 0x%x, dev 0x%x\n",
	       dp, dp->d_dev, cdp, cdp->d_dev);
#endif /* DEBUG */
	/*
	 * Propagate the shutdown to any parent drivers.
	 */
	devno = dp->d_dev & ~(V_PART_MASK | DEV_DIAG_MASK);
	for (i=0; i < CMPT_NUMPART; i++) {
		dpp = dev_ref(devno + i);
		if (!dpp)
			continue;
		
		if (dpp->d_parent) {
			DEV_SHUTDOWN(dpp->d_parent, dpp);
		}
		
		(void)dev_deref(dpp);
	}
	
	return(0);
}

/*
 * cmptsize()
 *	Used for swap-space partition calculation.
 */
cmptsize(dp)
	register struct dev *dp;
{
	struct cmpt_private *dsp;

	dsp = DEVTOCMPTPRIV(dp);
	if (dsp->cp_size <= 0 || !(dsp->cp_physdev))
		return(-1);
	return(dsp->cp_size);
}

/*
 * cmptdiagstrategy
 *	strategy routine for the diagnostic cmpt device
 *
 * Acts as a filter for I/O requests going through the diagnostic
 * device.  This routine provides read-only access to the entire
 * disk drive.  This access is provided for performance-measuring
 * utilities which need direct access to the whole disk.  Thus,
 * no partition checking is done at this level.
 */
cmptdiagstrat(dp, bp, offset)
	struct	dev *dp;
	struct	buf *bp;
	daddr_t	    offset;
{
	struct cmpt_private *dsp = DEVTOCMPTPRIV(dp);

	bp->b_status = B_STATUS_OK;	  /* default status */

	if ((bp->b_flags & B_READ) != B_READ) {
		bp->b_flags |= B_ERROR;
		bp->b_error = EINVAL;
		bp->b_resid = bp->b_bcount;
		biodone(bp);
		return;
	}

	DEV_STRATEGY(dsp->cp_physdev, bp, offset);
}
