/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ident "$Header: lp.c 1.33 90/05/01 $"

/*
 *	driver for the mlp2000 (Systech line printer)
 *
 */

/* $Log:	lp.c,v $
 */

#ifdef DEBUG
int	lp_debug = 0;
#define LPDEBUG		if (lp_debug) CPRINTF
#endif	/* DEBUG */


#include "sys/types.h"
#include "sys/devsw.h"
#include "sys/param.h"
#include "sys/debug.h"
#include "sys/sysmacros.h"
#include "sys/systm.h"
#include "sys/mutex.h"
#include "sys/signal.h"
#include "sys/user.h"
#include "sys/errno.h"
#include "sys/clock.h"
#include "sys/intctl.h"
#include "sys/gate.h"
#include "sys/uio.h"
#include "sys/buf.h"
#include "sys/session.h"
#include "sys/stream.h"
#include "sys/stropts.h"
#include "sys/strid.h"
#include "sys/file.h"
#include "sys/cmn_err.h"
#include "sys/mbad.h"
#include "sys/mbad_device.h"
#include "sys/lp.h"
#include "sys/lp_device.h"
#include "sys/printer.h"

#include "sys/autoconf.h"
#include "sys/mbad_ioconf.h"

static struct module_info info = {
	STRID_LP, "lp", 0, INFPSZ, 4096, 512,
};

static int lpopen(), lpclose(), lpwput();

static struct qinit rinit = {
	NULL, NULL, lpopen, lpclose, NULL, &info, NULL,
};

static struct qinit winit = {
	lpwput, NULL, NULL, NULL, NULL, &info, NULL,
};

struct streamtab lpinfo = {
	&rinit, &winit, NULL, NULL,
};

int lp_probe(), lp_timer(), lpboot();
void lp_init(), lp_intr();

/*
 * On interrupts, the board number == interrupt level - stintr_base.
 */
extern	struct		lp_printer lpconfig[];
extern	int		lpprinters;		/* # of printer configs */
extern struct devops	lpdevops;		/* devops for lp driver */
extern int		lp_c_major;		/* major device number for lp */
struct	lp_info		**lp_info;		/* software state */
int			lpintr_base;		/* base interrupt vector */
int			lpmaxdev;		/* max number of devices */ 
int			nlp;			/* maximum number of boards */

/*
 * lp_probe
 * 	see if there is a controller present
 */
lp_probe(dev)
	struct gen_device *dev;
{
	register struct lpdevice *base;
	unchar i;

	base = (struct lpdevice *) dev->devreg[0];
	base->lpd_cmdreg = (unchar) 0;		/* disable interrupts */
	i = base->lpd_intr;
#ifdef lint
	CPRINTF("interrupt register:0x%x\n", i);
#endif
	/*
	 * if you get a memory error, there is no board.
	 */
	
	return((base->lpd_intr&LPI_MEMERR)==0);
}

/*
 * lp_init
 *	Use existing lpboot to boot driver.
 */
void 
lp_init( driver )
	struct driver_module *driver;
{
	(void) mbad_boot_mbad_device( driver, lpboot, (int)lp_flags, lp_bin);
}


/*
 * lp_boot
 *	Set up data structures and initialize for each
 *	lp controller in system.
 *	Locks are allocated *one per board*.
 *	One half of the mapping registers are allocated
 *	to each of the possible ports.
 */
lpboot(n, md)
	int	n;				/* # of boards configured */
	register struct mbad_dev *md;		/* pointer to first */
{
	register struct lp_info *lp;
	int	board;
	int	unit;
	int	printer = 0;
	lock_t	*locks;
	int	prindex;
	struct dev *devp;
	
#ifdef DEBUG
	LPDEBUG("lpboot: begin with number of boards = %d\n", n);
#endif /* DEBUG */
	nlp = n;
	lpintr_base = md->md_vector;
	/*
	 * allocate space for softc structure and one lock *per board*,
	 * ie, one lock for every two softc's.
	 */
	if (nlp) {
		lp_info = (struct lp_info **) 
			calloc( (int)(sizeof(struct lp_info *) * nlp * 2));
		locks = (lock_t *) calloc((int)(sizeof(lock_t) * nlp));
	}
	lpmaxdev = (nlp*2) - 1;	

	/*
	 * for each controller ...
	 *	for each channel ...
	 *		allocate and initialize softc struct
	 */
	for (board = 0; board < nlp; md++, board++) {
		for (unit = 0; unit < 2; unit++) {
			lp = lp_info[board * 2 + unit] =
				(struct lp_info *) calloc((int)sizeof(struct lp_info));
			lp->lp_unit = unit;
			lp->lp_base = (struct lpdevice *) md->md_csr;
			if (!(lp->lp_alive = md->md_alive))
				continue;
			lp->lp_mblevel = md->md_level;
			lp->lp_mbdesc = md->md_desc;
			/*
			 * Half the maps to each unit, maximum 8k
			 * per line printer.  If odd number of maps,
			 * unit 0 gets one extra (arbitrary).
			 */
			lp->lp_nmaps = md->md_nmaps >> 1;
			if (unit == 0 && (md->md_nmaps & 01))
				lp->lp_nmaps++;
			lp->lp_basemap = md->md_basemap + (unit * lp->lp_nmaps);
			lp->lp_bufsz = lp->lp_nmaps > 8 ? (8 * MB_MRSIZE) : lp->lp_nmaps * MB_MRSIZE;
			if (lp->lp_nmaps < 1) {
				cmn_err(CE_WARN,
				"lp%d: Line %d no maps, deconfigured, csr=0x%x",
					board, unit, lp->lp_base);
				/*
				 *+ No maps have been configured for the
				 *+ lp device.  Corrective action:  check the system and device 
				 *+ configuration files. 
				 */
				lp->lp_alive = 0;
				continue;
			}
			devp = (struct dev *)calloc(sizeof(struct dev));
			devp->d_dev = makedev(lp_c_major, (board << 1) | unit);
			devp->d_flags = 0;
			devp->d_ops = &lpdevops;
			if (devsw_add(devp) < 0) {
				CPRINTF("lp%d: problem adding device to devsw\n",
					devp->d_dev);
				lp->lp_alive = 0;
				continue;
			}
			/*
			 * If there exist more printers than we configured,
			 * duplicate the last printer's configuration.
			 */
			if (printer >= lpprinters) {
				cmn_err(CE_NOTE, "lp%d: printer %d not configured.",
					board, printer);
				/*
				 *+ There are more printers than have been
				 *+ specified in the lpconfig table.  Excess
				 *+ printers will use the lpconfig data specified for
				 *+ the last entry in the table.
				 */
				cmn_err(CE_CONT, "using printer %d's configuration\n",
					lpprinters - 1);
				prindex = lpprinters - 1;
			} else
				prindex = printer;
			printer++;
			lp->lp_height = lpconfig[prindex].lp_height;
			lp->lp_width = lpconfig[prindex].lp_width;
			lp->lp_ops = lpconfig[prindex].lp_ops;
			lp->lp_lock = &locks[board];
			lp->lp_line = 0;
			lp->lp_dmareq = 0;
			init_lock(lp->lp_lock);
			init_sema(&lp->lp_sema, 0);
			lp->lp_base->lpd_mode = unit|LPMODE;
			LPSTOP(lp->lp_base->lpd_cmdreg);
		}
		if (lp->lp_alive) {
#ifdef DEBUG
			LPDEBUG("lpboot: board %d dev %d alive!\n", board, unit);
#endif /* DEBUG */
			lp->lp_base->lpd_csr |= (unchar)LPINTE;	/* enable interrupts */
			sendsoft(lp->lp_mbdesc->mb_slicaddr, (unchar)(1<<lp->lp_mblevel));
		}
	}
}

/*
 * lpopen
 * 	Open the device. We initialize the data structures
 * 	after the open succeeds. This simplifies the close
 * 	routine in handling the case when the printer is
 * 	offline, the open blocks and we signal out of it.
 */
/*ARGSUSED*/
static int
lpopen(q, dev, flag, sflag)
	dev_t dev;
	queue_t *q;		/* read queue */
{
	struct lp_info *lp;
	int unit = minor(dev);
	spl_t s;

	if (sflag)
		/* non-driver open */
		return (OPENFAIL);

	if (LPBOARD(dev) >= nlp || (lp = lp_info[unit]) == NULL 
				|| !lp->lp_alive)
		/* invalid device */
		return (OPENFAIL);

	s = p_lock(lp->lp_lock, LP_SPL);

	if (q->q_ptr) {
		/*
		 * Device is already open.  This is an exclusive
		 * access device.
		 */
		u.u_error = EBUSY;
		v_lock(lp->lp_lock, s);
		return (OPENFAIL);
	}

	if (!(flag & FNDELAY) && LPOFFLINE(lp)) {
		/*
		 * wait for printer to come online 
		 */
#ifdef DEBUG
		LPDEBUG("lpopen: offline && !NDELAY - waiting\n");
#endif /* DEBUG */
		lp->lp_state |= LP_TIMER | LP_WAIT;
		lp->lp_timeid = timeout(lp_timer, (caddr_t)lp, 2*HZ);
		while (lp->lp_state & LP_TIMER) {
			/*
			 * Must loop until ready in case of SIGSTOP
			 */
			if (p_sema_v_lock(&lp->lp_sema, (PZERO+1)|PCATCH,
					  lp->lp_lock) == PCATCH) {
#ifdef DEBUG
				LPDEBUG("lpopen: woken by interrupt\n");
#endif /* DEBUG */
				u.u_error = EINTR;
				untimeout(lp->lp_timeid);
				return (OPENFAIL);
			}
			s = p_lock(lp->lp_lock, LP_SPL);
		}
	}

	lp->lp_state = 0;
	lp->lp_msg = NULL;
	q->q_ptr = (caddr_t)lp;
	WR(q)->q_ptr = (caddr_t)lp;
	lp->lp_qptr = WR(q);
	v_lock(lp->lp_lock, s);
	return (OPENOK);
}

/*
 * lpclose 
 *	wait for all queued output to drain and then
 *	close device
 */
static int
lpclose(q)
	queue_t *q;		/* read queue */
{
	struct lp_info *lp;
	spl_t s;

	lp = (struct lp_info *)q->q_ptr;

	s = p_lock(lp->lp_lock, LP_SPL); 
	while (lp->lp_state & LP_BUSY) {
#ifdef DEBUG
		LPDEBUG("lpclose: LP_BUSY\n");
#endif /* DEBUG */
		if (p_sema_v_lock(&lp->lp_sema, (PZERO+1)|PCATCH,
				  lp->lp_lock) == PCATCH) {
			/*
			 * woken up by a signal so cleanup
			 */
#ifdef DEBUG 
			LPDEBUG("lpclose: woken by interrupt!\n");
#endif /* DEBUG */
			if (lp->lp_state & LP_TIMER) {
				untimeout(lp->lp_timeid);
			}
			u.u_error = EINTR;
			disable_procs(q);
			s = p_lock(lp->lp_lock, LP_SPL);
			if (lp->lp_msg) {
				freemsg(lp->lp_msg);
				lp->lp_msg = NULL;
			}
			lp->lp_qptr = NULL;
			lp->lp_state = 0; 
			v_lock(lp->lp_lock, s);
			return;
		}
		s = p_lock(lp->lp_lock, LP_SPL);
	}
	ASSERT_DEBUG(!(lp->lp_msg), "lpclose: current message not freed");
	v_lock(lp->lp_lock, s);
	disable_procs(q);
	s = p_lock(lp->lp_lock, LP_SPL);
	lp->lp_qptr = NULL;
	lp->lp_state = 0;
	v_lock(lp->lp_lock, s);
	return;
}

/*
 * lpwput
 *	process streams messages by either handling them
 *	now or queuing them for later
 */	
static int
lpwput(q, mp)
	queue_t *q;		/* write queue */
	register mblk_t *mp;	/* message pointer */
{
	struct lp_info *lp;
	struct pr_config *cfg;
	struct iocblk *iocp;
	spl_t s;

	lp = (struct lp_info *)q->q_ptr;

	switch (mp->b_datap->db_type) {

	case M_FLUSH:
		/* Canonical flush handling */
		if (*mp->b_rptr & FLUSHW) {
			flushq(q, FLUSHDATA);
			s = p_lock(lp->lp_lock, LP_SPL);
			if (lp->lp_msg) {
				freemsg(lp->lp_msg);
				lp->lp_msg = NULL;
			}
			v_lock(lp->lp_lock, s);
		}
		if (*mp->b_rptr & FLUSHR) {
			flushq(RD(q), FLUSHDATA);
			*mp->b_rptr &= ~FLUSHW;
			qreply(q, mp);
		} else
			freemsg(mp);
		break;

	case M_IOCTL:
		iocp = (struct iocblk *)mp->b_rptr;
		switch (iocp->ioc_cmd) {
		case PIOCSETCANON:
			switch (*(int *)mp->b_cont->b_rptr) {
			case PRCAPS:
				s = p_lock(lp->lp_lock, LP_SPL);
				lp->lp_ops = PRCAPS;
				v_lock(lp->lp_lock, s);
				mp->b_datap->db_type = M_IOCACK;
				break;
			case PRRAW:
				s = p_lock(lp->lp_lock, LP_SPL);
				lp->lp_ops = PRRAW;
				v_lock(lp->lp_lock, s);
				mp->b_datap->db_type = M_IOCACK;
				break;
			case PRSPECIAL:
				s = p_lock(lp->lp_lock, LP_SPL);
				lp->lp_ops = PRSPECIAL;
				v_lock(lp->lp_lock, s);
				mp->b_datap->db_type = M_IOCACK;
				break;
			default:
				mp->b_datap->db_type = M_IOCNAK;
				iocp->ioc_error = EINVAL;
				break;
			}
			qreply(q, mp);
			return;
		default:
			break;
		}
		/*FALLTHROUGH*/

	case M_PCIOCTL:
		/*
		 * lp printer doesn't support any priority
		 * ioctl's so NAK it.
		 */
		mp->b_datap->db_type = M_IOCNAK;
		qreply(q, mp);
		break;

	case M_PCPROTO:
		cfg = (struct pr_config *)mp->b_rptr;
		if (cfg->p_type == REQUEST_CFG) {
			cfg->p_type = REQUEST_CFG_RESPONSE;
			cfg->p_ops = lp->lp_ops;
			cfg->p_width = lp->lp_width;
			cfg->p_height = lp->lp_height;
		} else
			freemsg(mp);
		break;

	case M_DATA:
		/*
		 * Queue message and start device if
		 * not busy.
		 */
		putq(q, mp);
		s = p_lock(lp->lp_lock, LP_SPL);
		if (!(lp->lp_state & LP_BUSY))
			lp_out(lp);
		v_lock(lp->lp_lock, s);
		break;

	default:
		cmn_err(CE_WARN, "lpwput: unknown message type %d",
			mp->b_datap->db_type);
		/*
		 *+ The lp driver has received a message it doesn't
		 *+ recognize and is discarding it.
		 */
		freemsg(mp);
		break;

	}	/* end switch */
}

/*
 * lp_out
 *	get next buffer with data and either output via 
 *	lp_start or trigger to be output via lp_timer
 *
 *	must be called with lp_info locked	
 */
lp_out(lp)
	struct lp_info *lp;		/* device specific info */
{
	register mblk_t *bp;
	queue_t *q;

#ifdef DEBUG
	LPDEBUG("lp_out: begin\n");
#endif /* DEBUG */
	q = lp->lp_qptr;
loop:
	if ((bp = lp->lp_msg) ==  NULL) {
		if ((bp = getq(q)) == NULL) {
			vall_sema(&lp->lp_sema);
			return;
		}
		lp->lp_msg = bp;
	}
	if (bp->b_rptr >= bp->b_wptr) {
		bp = lp->lp_msg->b_cont;
		lp->lp_msg->b_cont = NULL;
		freeb(lp->lp_msg);
		lp->lp_msg = bp;
		goto loop;
	}

	/*
	 * Found a message with data in it.  Check if ONLINE
	 */
	lp->lp_state |= LP_BUSY;
	if (LPOFFLINE(lp)) {
		lp->lp_state |= LP_TIMER;
		lp->lp_timeid = timeout(lp_timer, (caddr_t)lp, 2*HZ);
		return;
	}
	lp->lp_count = MIN(lp->lp_bufsz, (ushort)(bp->b_wptr - bp->b_rptr));
	lp->lp_buf = bp->b_rptr;
	bp->b_rptr += lp->lp_count;
#ifdef DEBUG
	LPDEBUG("lp_out: call start\n");
#endif /* DEBUG */
	lp_start(lp);
}


unchar tc[2] = { LPS_TC0, LPS_TC1 };

/*
 * lp_start
 *	start output of current data to device
 *
 * 	must be called with lp_info locked
 */
lp_start(lp)
	struct lp_info *lp;		/* device specific info */
{
	register int addr;
	ushort cnt;

	addr = mbad_physmap(lp->lp_mbdesc, (int)lp->lp_basemap,
			(caddr_t)lp->lp_buf, lp->lp_count, (int)lp->lp_nmaps);
	cnt = lp->lp_count - 1;
	LPSTOP(lp->lp_base->lpd_cmdreg);
	/*
	 * to transfer the LSB first, must instruct the 8237 (clear byte ptr)
	 */
	lp->lp_base->lpd_clbp = 1;
	if (lp->lp_unit == 0) {
		lp->lp_base->lpd_cnt0 = (unchar)cnt;
		lp->lp_base->lpd_cnt0 = (unchar)(cnt >> 8);
		lp->lp_base->lpd_addr0 = (unchar)addr;
		lp->lp_base->lpd_addr0 = (unchar)(addr >> 8);
		lp->lp_base->lpd_hiaddr0 = (unchar)(addr >> 16);
	} else {
		lp->lp_base->lpd_cnt1 = (unchar)cnt;
		lp->lp_base->lpd_cnt1 = (unchar)(cnt >> 8);
		lp->lp_base->lpd_addr1 = (unchar)addr;
		lp->lp_base->lpd_addr1 = (unchar)(addr >> 8);
		lp->lp_base->lpd_hiaddr1 = (unchar)(addr >> 16);
	}

	if (lp->lp_dmareq == 0) {
		lp->lp_base->lpd_req = (unchar)LPR_SET | lp->lp_unit;
		lp->lp_dmareq = 1;
	}
	lp->lp_base->lpd_csr |= (unchar)LPINTE;		/* enable ints */
	lp->lp_base->lpd_bmask = (unchar)lp->lp_unit;	/* clear mask bit */
	LPSTART(lp->lp_base->lpd_cmdreg);
	mbad_reenable(lp->lp_mbdesc, lp->lp_mblevel);
}

/*
 * lp_intr
 *	handle device interrupts and start next output
 *	operation when appropriate
 */
void 
lp_intr(vector)
	int vector;
{
	int unit = (vector - lpintr_base) * 2;
	unchar status;
	struct lp_info *lp;
	spl_t s;
	int i;

	if (unit < 0 || unit > lpmaxdev) {
		cmn_err(CE_WARN, "lp: stray interrupt, unit out of range %d",
			unit);
		/*
		 *+ The lp driver received an interrupt for a nonexistent
		 *+ device.
		 */
		return;
	}
	if ((lp = lp_info[unit]) == NULL || !lp->lp_alive) {
		cmn_err(CE_WARN, "lp: stray interrupt, dead unit %d", unit);
		/*
		 *+ The lp driver received an interrupt for a device that was
		 *+ not configured.
		 */
		return;
	}
	s = p_lock(lp->lp_lock, LP_SPL);
	/*
	 * Any write to interrupt register clears it, which leaves 
	 * a window in which an interrupt can be lost.  Therefore,
	 * check the DMA count for both devices to determine
	 * if a transfer completed.
	 */
	status = (lp->lp_base->lpd_intr) & ~(LPS_TC0 | LPS_TC1);
	lp->lp_base->lpd_intr = 1;		/* clear interrupts */
	status |= (lp->lp_base->lpd_cmdreg & (LPS_TC0|LPS_TC1)); 
	if (status == LPI_REQUEST) {		/* ignore spurious ints */
#ifdef DEBUG
		LPDEBUG("lpintr: spurious int\n");
#endif /* DEBUG */
		v_lock(lp->lp_lock, s);
		mbad_reenable(lp->lp_mbdesc, lp->lp_mblevel);
		return;
	}
	if (status & LPI_MEMERR)
		cmn_err(CE_WARN, "lp%d: multibus memory error", 
			vector - lpintr_base);
		/*
		 *+ The lp driver attempted a transfer to an illegal MULTIBUS 
		 *+ address.
		 */

	for (i = 0; i < 2; i++) {
		lp = lp_info[unit++];
		if (lp->lp_state & LP_BUSY && (status & tc[i])) {
			lp->lp_state &= ~LP_BUSY;
			lp_out(lp);
		}
	}
	v_lock(lp->lp_lock, s);
}

/*
 * lp_timer
 *	poll for device ONLINE, wake up waiters
 */
lp_timer(lp)
	struct lp_info *lp;
{
	spl_t s;

	s = p_lock(lp->lp_lock, LP_SPL);
	if (LPOFFLINE(lp)) {
		lp->lp_timeid = timeout(lp_timer, (caddr_t)lp, 2*HZ);
		v_lock(lp->lp_lock, s);
		return;
	}
	lp->lp_state &= ~LP_TIMER; 
	if (lp->lp_state & LP_WAIT) {
		vall_sema(&lp->lp_sema);
	} else {
		lp_out(lp);
	}
	v_lock(lp->lp_lock, s);
}
