/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ident	"$Header: mbad.c 1.40 90/10/22 $"

/*
 * mbad.c
 *	Various procedures dealing with MBIf's.
 */

/* $Log:	mbad.c,v $
 */

#include "sys/types.h"
#include "sys/signal.h"
#include "sys/param.h"
#include "sys/debug.h"
#include "sys/mutex.h"
#include "sys/systm.h"
#include "sys/map.h"
#include "sys/user.h"
#include "sys/timer.h"
#include "sys/proc.h"
#include "sys/vm.h"
#include "sys/buf.h"
#include "sys/vnode.h"
#include "sys/cmn_err.h"

#include "sys/cfg.h"
#include "sys/slic.h"
#include "sys/slicreg.h"
#include "sys/bic.h"

#include "sys/hwparam.h"
#include "sys/intctl.h"
#include "sys/mftpr.h"
#include "sys/pte.h"
#include "sys/reg.h"
#include "sys/trap.h"
#include "sys/plocal.h"

#include "sys/autoconf.h"
#include "sys/multibus.h"

#include "sys/mbad_ioconf.h"
#include "sys/mbad.h"
#include "sys/mbad_device.h"

#ifdef	DEBUG
int	mbad_debug = 0;		/* get verbose flag */
extern int boot_debug;
static unchar *mbad_unused_vectors;
#endif	/* DEBUG  */

struct mb_desc	*MBAd;		/* base of array */
int	 	NMBAd;		/* # MBAd's to map at boot */
uint	 	MBAdvec;	/* bit-vector of existant MBAd's */
unchar		MBAd_errbase;	/* MBAd error base vector */
struct pte	*MBIOmap;	/* mapping pages for MBAd's */
struct pte	*MBIOpte;	/* current MBIO map pte value */

struct mb_desc	*MBAd_probe();	/* MBAd device prober */
void		MBAd_error();	/* MBAd error catcher */
int		MBAd_nmi();	/* handler for NMI during probe */
label_t		MBAd_nmi_jmp;	/* for NMI recovery during probe */
int		MBAd_idx;	/* index of mbad being probed for devices */

#define	MBAD_CLEAR(idx)		MBAdvec &= ~(1 << (idx))
#define	MBAD_EXISTS(idx)	(MBAdvec & (1 << (idx)))
#define	WILDCARD(md)		((md)->md_idx == -1)

#define	PLURAL(x)		((x)==1?"":"s")

extern struct device_config sb8000_config[];
extern struct mbad_device mbad_devices[];
static void bad_b_iotype(), mbad_setup_range_check();

mbad_probe( device )
	struct gen_device *device;
{
	register struct mbad_device *dev = (struct mbad_device *)device;
	struct ctlr_toc *toc = &CD_LOC->c_toc[SLB_MBABOARD];
	register struct	ctlr_desc *cd = &CD_LOC->c_ctlrs[toc->ct_start];
	int i;

	for (i = 0; i < toc->ct_count; i++, cd++)
		if (dev->config->dev_addr == cd->cd_slic) {
			dev->cd = cd;	/* This is the matching MBAD */
			if (FAILED(cd)) 
				return 0;	/* Short circuit search */
			return 1;	/* Found */
		}
	return 0;			/* Not found */
}

/*
 * mbad_init()
 *	Allocate and fill out MBAd descriptors.
 *
 * Can deal with deconfigured boards (uses MBAdvec).
 *
 * We fill out all the MBAd[*]'s, using MBAdvec to tell sysinit()
 * which ones to map.
 */

#ifdef	ns32000
ulong	virt_mbad;			  /* kernel base virt addr of MBAd's */
extern	ulong	max_ker_vaddr;		/* max kernel virtual address */
#endif	/* ns32000  */

void 
mbad_init( driver )
	struct driver_module *driver;
{
	struct ctlr_toc *toc = &CD_LOC->c_toc[SLB_MBABOARD];
	register struct	ctlr_desc *cd;
	register struct mb_desc *mb;
	register struct mbad_device *dev = 
			(struct mbad_device *)driver->devices;
	int i, j, nmbads = driver->numdev;
	unchar	vec;

	mb = MBAd = (struct mb_desc *)calloc(nmbads * sizeof(struct mb_desc));

#ifdef	ns32000
	if (toc->ct_count) {
		/*
		 * Move MBAd mapping.  Power-up/boot firmware placed the MBAd's
		 * starting at 8Meg.  Must move before do any probes/boots
		 * since some driver save MBAd addresses.
		 *
		 * Have MBAd mapping top out at max_ker_vaddr, and adjust
		 * max_ker_vaddr to reflect this).
		 *
		 * Use ovbcopy() since source may overlap destination.
		 */
		virt_mbad = max_ker_vaddr - toc->ct_count * MBAD_ADDR_SPACE;
		ovbcopy((caddr_t) (CD_LOC->c_pagetable + L1IDX(OLD_VIRT_MBAD)),
			(caddr_t) (CD_LOC->c_pagetable + L1IDX(virt_mbad)),
			(unsigned) (toc->ct_count
					* ((MBAD_ADDR_SPACE/NBPG)/NPTEPG))
					* sizeof (struct pte)
		);
		max_ker_vaddr = virt_mbad;
	} 
#endif	/* ns32000  */

	for (i = 0; i < nmbads; i++, mb++, dev++) {
		if (cd = dev->cd) {
			mb->mb_slicaddr = cd->cd_slic;
			mb->mb_mem = (caddr_t)VA_MBAd(cd->cd_i);
			mb->mb_ios = (struct mb_ios *)
				(VA_MBAd(cd->cd_i) + MBAd_IOwindow);
			mb->mb_dmabase = MBAd_IOwindow;
			if ((cd->cd_diag_flag & (CFG_FAIL|CFG_DECONF)) == 0) {
				++NMBAd;
				MBAdvec |= 1 << i;
			}
		}
	}

	if (NMBAd < toc->ct_count) {
		cd = &CD_LOC->c_ctlrs[toc->ct_start];
		CPRINTF("        # Not using MULTIBUS Adapters: slic");
		for (i = 0; i < toc->ct_count; i++, cd++) {
			for (j = 0, dev = (struct mbad_device *)driver->devices;
			     j < nmbads; j++, dev++)
				if (dev->cd == cd)
					break; /* Matching config entry found */

			if (j == nmbads || !MBAD_EXISTS(j))
				CPRINTF(" %d", cd->cd_slic);
		}
		CPRINTF(".\n");
	}
	/*
	 * Initialize MBAd error reporting and mapping.
	 */

	CPRINTF("        # MBAd DMA window at 0x%x.\n", MBAd_IOwindow);

	bin_alloc[MBAD_ERROR_BIN] =
		roundup(bin_alloc[MBAD_ERROR_BIN], MBAD_ERR_ALIGN);
	MBAd_errbase = ivecpeek(MBAD_ERROR_BIN);

	for (i = 0; i < nmbads; i++) {
		/*
		 * This relies on ivecall() returning consecutive vectors.
		 */
		vec = ivecall(MBAD_ERROR_BIN);
		ivecinit(MBAD_ERROR_BIN, vec, MBAd_error);
		if (MBAD_EXISTS(i)) 
			MBAd_init(&MBAd[i], MBAD_ERROR_BIN, vec);

		bin_alloc[MBAD_ERROR_BIN] =
			roundup(bin_alloc[MBAD_ERROR_BIN], MBAD_ERR_ALIGN);
	}

	/*
	 * If any devices, allocate mapping pages for them.
	 */
	if (NMBAd) {
		callocrnd(NBPG);
#ifdef	i386
		/*
		 * Must allocate for configured # of MBAd's 
		 * since several MBAd's may be mapped by same 
		 * page and their mappings are allocated in
		 * configuration device order, as are their
		 * SB8000 memory windows. Note that these
		 * are assigned to windows in bus order.
		 */
		MBIOmap = (struct pte *)
			calloc(howmany(toc->ct_count * MBAD_ADDR_SPACE,
					NPTEPG*NBPG) * NBPG);
#endif	/* i386  */
#ifdef	ns32000
		/*
		 * Need only allocate ptes for existing MBAd's, since
		 * single page of mapping maps subset of MBAd.
		 */
		MBIOmap = (struct pte *) calloc(NMBAd * NPTEMBAD
						* sizeof(struct pte));
#endif	/* ns32000  */
#ifdef DEBUG
		mbad_unused_vectors =
			(unchar *) calloc( nmbads * 8 * sizeof (unchar));
#endif /* DEBUG */
		MBIOpte = MBIOmap;
	}
}

unchar 
mbad_boot_mbad_device( driver, boot, flags, bin)
	struct driver_module *driver;
	int (*boot)(), flags;
	unchar bin;
{
	struct mbad_old_mbad_device *dev
		= (struct mbad_old_mbad_device *)driver->devices;
	struct mb_desc *mb;
	struct mbad_map_origin *mo;
	struct mbad_dev *md;
	unchar vec;
	int i, j;

	/* Returns basevec */
#ifdef DEBUG
	if (boot_debug)
CPRINTF("mbad_boot_mbad_device: init %s, boot @ 0x%x, flags = 0x%x, bin = %d\n",
		       driver->name, boot, flags, bin); 
#endif /* DEBUG */

	/*
	 * Allocate and fill in all the mbad_dev structures
	 */

	md = (struct mbad_dev *)
		calloc(driver->numdev * sizeof (struct mbad_dev));
	if ( ! md )
		cmn_err( CE_PANIC,
		"mbad_boot_mbad_device: No memory for %s driver mbad_dev array",
			driver->name);
	/*
	 *+ When booting an old mbad device (one that will not work on any
	 *+ MULTIBUS, but only the mbad), the kernel allocates a block of memory for
	 *+ the init routines (for backwards compatibility).  The kernel did not
	 *+ have enough memory left to allocate this structure for the named
	 *+ driver.  Unless a huge number of devices are
	 *+ configured, this error usually points to a serious shortage
	 *+ of memory or to a bug in some driver requesting hugh chunks of
	 *+ kernel memory.
	 */
	
	for ( i = 0; i < driver->numdev; i++, md++ ) {
		/* prepare each device in turn */
		dev[i].mbad_dev = md;
		if ( driver->intr )
			vec = ivecall(bin);

		
		if (dev[i].bus->bustype == NOBUS)
			continue;

		if ( dev[i].bus->owner->driver &&
		    strcmp( dev[i].bus->owner->driver->name, "mbad") != 0 ) {
			cmn_err( CE_WARN,
		     "mbad_boot_mbad_device: %s%d not on an MBAD, deconfigured",
				driver->name, DEVICEN(dev+i));
			/*
			 *+ The specified device is not on a MULTIBUS provided by
			 *+ an mbad; however, the device driver is prepared to
			 *+ deal only with a mbad MULTIBUS.  Corrective action:
			 *+ either insert the
			 *+ board into a MULTIBUS connected to an
			 *+ mbad, or rewrite the device driver to be more
			 *+ general. 
			 */
			continue;
		}

		md->md_flags = flags;
		md->md_bin = bin;
		
		if ( dev[i].map_origin ) {
#ifdef DEBUG
			if (boot_debug)
				CPRINTF( "mbad_boot_mbad_device:  %s%d found\n",
				       driver->name, DEVICEN(dev+i) ); 
#endif /* DEBUG */
			/* Map origin only gets set if alive. */
			mo = (struct mbad_map_origin *)dev[i].map_origin;
			md->md_basemap = mo->basemap;
			md->md_nmaps = mo->nmaps;
			md->md_idx = (struct mbad_device *)dev[i].bus->owner
				- mbad_devices;
			md->md_csr = (ulong)dev[i].devreg;
			md->md_mapwant = dev[i].n_maps;
			md->md_level = dev[i].config->busint;
			md->md_desc = mb = &MBAd[md->md_idx];
			/*
			 * If there is an interrupt, set up MBAd
			 * to respond properly to it.  Also,
			 * check that the same MBAd interrupt
			 * level isn't already in use. 
			 */
			if (driver->intr) {
				if (mb->mb_intrs & (1 << md->md_level)) {
					cmn_err( CE_WARN,
						"%s%d on %s%d, csr 0x%x: MB ",
					       driver->name, DEVICEN(dev),
					       dev->bus->owner->driver->name,
					       mb-MBAd,
					       md->md_csr);
					/*
					 *+ The named device has a MULTIBUS
					 *+ interrupt level that is the same
					 *+ as another already booted device
					 *+ on the same MULTIBUS.  This is not
					 *+ allowed.  Corrective action: 
					 *+ reconfigure the device
					 *+ with a different interrupt
					 *+ (specified with unit number).
					 */
					cmn_err( CE_CONT,
						"level %d%s", md->md_level,
					       " overlaps previous use.\n");
					(void)MBAd_mapall(mb, (int)
							  (-md->md_nmaps));
					continue;
				}
				md->md_vector = vec;
				ivecinit( md->md_bin, vec, driver->intr );
				/*
				 * We know intr_setup exists, because we know
				 * its mbad!  Maps the multibus interrupt of
				 * level md->md_level on the mbad that
				 * controlls the dev[i].bus (multibus) to SLIC
				 * interrupt bin md->md_bin, vector vec,
				 * destination SL_GROUP|GROUP_TMPOS.  The zero
				 * is for multibus interrupt vectors, which
				 * don't exist.
				 */
				j = (*dev[i].bus->intr_setup)
					(dev[i].bus, md->md_level, 0,
					 md->md_bin, vec,
					 SL_GROUP|TMPOS_GROUP);
				if ( j != BUS_INT_OK ) {
					cmn_err( CE_WARN,
				"mbad_boot_mbad_device: %s%d bad mbad intr map",
						driver->name, DEVICEN(dev+i));
					/*
					 *+ While the kernel was attempting to initialize the
					 *+ interrupt mapping for the given
					 *+ device, an unspecified error
					 *+ occurred.  The device is not
					 *+ configured.  If the error is due to a
					 *+ duplicate interrupt, this error
					 *+ message will have been preceded by
					 *+ another error message.  Otherwise,
					 *+ the error occurred because the bus owner isn't an
					 *+ mbad. 
					 */
					continue;
				}
				mb->mb_intrs |= 1 << md->md_level;
			}
			
			md->md_alive = 1;
		} 

	}
#ifdef DEBUG
	if ( boot_debug )
		CPRINTF(
	      "mbad_boot_mbad_device: booting %d %s%s, mbad_dev array @ 0x%x\n",
		       driver->numdev, driver->name,
		       (driver->numdev == 1) ? "" : "s", dev->mbad_dev);
#endif /* DEBUG */
	(*boot)(driver->numdev, dev->mbad_dev);
	return dev->mbad_dev->md_vector;  /* base vector */
}

/*
 * mbad_multibus_probe_setup()
 *	Probe for devices attached to MBAd's.
 *
 * Also, init error handling vectors per MBAd.
 *
 * Allocate enough level-2 mapping pages to map all possible MBAd's
 * (4K page maps 4Meg, so typically one mapping page).
 */

mbad_multibus_probe_setup(dev)
	struct gen_device *dev;
{
	register struct	mb_desc	*mb;
	register int nmaps;
	struct mbad_map_origin *mo;
	int	maps;

	nmaps = dev->nsegs * ((dev->segsize + MB_MRSIZE-1)/MB_MRSIZE
			      + (dev->segalign%MB_MRSIZE? 1 : 0));
	
	if (mb = MBAd_probe(dev)) {

		/*
		 * Allocate whatever map registers it needs.
		 * If can't fit, don't configure the device.
		 */
		maps = MBAd_mapall(mb, nmaps);
		if (maps < 0) {
			cmn_err( CE_WARN,
		      "%s%d on %s%d csr 0x%x: %d map registers is not enough\n",
				dev->driver->name, DEVICEN(dev),
			       dev->bus->owner->driver->name, mb-MBAd,
			       dev->devreg[0], nmaps);
			/*
			 *+ There are not enough map registers left on the 
			 *+ named mbad to fill the needs of the named device.
			 *+ Corrective action:  
			 *+ either configure this device with more segments or
			 *+ a larger segsize, or delete some device(s) from
			 *+ this mbad, as it can provide only a total of
			 *+ 256 DMA maps spread among all attached devices.
			 */
			return 0;
		}
		dev->map_origin = calloc( sizeof (struct mbad_map_origin)
					 +  (nmaps - 1) * sizeof (long) );
		if ( ! dev->map_origin ) {
			cmn_err( CE_PANIC,
		"mbad_multibus_probe_setup: no memory for map struct for %s%d",
				 dev->driver->name, DEVICEN(dev) );
			/*
			 *+ The system ran out of memory and
			 *+ mbad_multibus_probe_setup was unable to allocate
			 *+ the data structure necessary to assign MULTIBUS
			 *+ maps to the named device.  The device is left
			 *+ unconfigured. 
			 */
			return 0;
		}

		mo = (struct mbad_map_origin *) dev->map_origin;
		mo->basemap = maps;
		dev->n_maps = mo->nmaps = nmaps;

		/*
		 * Tell all about it...
		 */
		
		CPRINTF("# %s configuration unit %d on %s is ",
		       dev->driver->name, 
		       (dev->config - dev->on_bus->config_array)
		       /dev->on_bus->nspaces,
		       bus_type[dev->bus->bustype]);
		CPRINTF("csr 0x%x\n        ", dev->devreg[0]);
		return 1;
	}
	return 0;
}

/*
 * MBAd_probe()
 *	Probe for a device on a multibus.
 *
 * Handles wild-card MBAd specification, finding a MBAd with the device
 * if it exists.
 *
 * Returns pointer to MBAd descriptor that device lives on or zero if
 * device not found.
 */

static struct mb_desc *
MBAd_probe(dev)
	struct gen_device *dev;
{
	register struct mb_desc *mb;
	int i, mbad_idx = (struct mbad_device *)dev->bus->owner - mbad_devices;
	extern	int (*probe_nmi)();

	/*
	 * MBAd is either wild-carded or must exist in the configuration.
	 */

	if ( !MBAD_EXISTS(mbad_idx))
		return ((struct mb_desc *)0);

	/*
	 * If no probe procedure, then assume it exists if the
	 * desired MBAd exists.  Note:  MBAD_EXISTS is false for
	 * wildcard MBAd index.
	 */

	if ( ! dev->driver->probe ) {
		if (MBAD_EXISTS(mbad_idx))
			return (&MBAd[mbad_idx]);
		else
			return ((struct mb_desc *)0);
	}

	/*
	 * It has a probe procedure, so we have to try to find it.
	 * We take over NMI vector temporarily here, to handle not
	 * present devices (see machine/locore.s for NMI handling).
	 */

	if (!MBAD_EXISTS(mbad_idx))
		return ((struct mb_desc *)0);

	probe_nmi = MBAd_nmi;
	
	/*
	 * Set up parameters for probe procedure and call
	 * driver.  If driver returns "true", then we found it.
	 *
	 * Use setjmp/longjmp to handle case of device not
	 * present, which results in NMI.
	 */
	
	mb = &MBAd[mbad_idx];
	for ( i = 0; i < dev->nspaces; i++ )
		switch ( dev->config[i].space ) {
		case MULTIBUS_NO_SPACE:
			break;
		case MULTIBUS_IO_SPACE:
			dev->devreg[i] = (caddr_t)
				&mb->mb_ios->mb_io[dev->config[i].dev_addr];
			break;
		case MULTIBUS_MEMORY_SPACE:
			dev->devreg[i] = (caddr_t)
				&mb->mb_mem[dev->config[i].dev_addr];
			break;
		default:
			cmn_err( CE_WARN, 
			   "MBAd_probe: Can't probe %s%d: illegal space %d in ",
				dev->driver->name, DEVICEN(dev), 
				dev->config[i].space);
			/*
			 *+ While the kernel was setting up to probe the named device, it 
			 *+ detected an illegal value in the "space" field of
			 *+ the config record.  Legal values for MULTIBUS are
			 *+ given in the include file sys/multibus.h.  You
			 *+ can use the information that printed to
			 *+ locate the record for the specified device in the _space.c file.
			 *+ The record number (zero based) is printed, followed by
			 *+ the unit and the record within the unit.  This is
			 *+ for devices that can have more than one config
			 *+ record per unit configuration.
			 */

			cmn_err( CE_CONT,
				"config record %d (unit %d, record %d of %d)",
				dev->config-dev->on_bus->config_array+i,
				(dev->config-dev->on_bus->config_array)
				/dev->nspaces, i, dev->nspaces);
			break;
		}
	
	if (setjmp(&MBAd_nmi_jmp)) {
		probe_nmi = NULL;
		return ((struct mb_desc *)0);
	}
	
	if ((*dev->driver->probe)(dev)) {
		probe_nmi = NULL;
		return (mb);
	}
	
	probe_nmi = NULL;
	return ((struct mb_desc *)0);
}

/*
 * mbad_multibus_alloc_maps()
 *	Fill out page-tables for existant MBAd's.  Also arrange that each
 *	MBAd map the c-list.  Called by sysinit() at initialization time.
 *
 * Assumes configure() fills out MBAdvec and allocated mapping pages (MBIOmap).
 *
 * C-list was allocated phys==virt, contiguous, in sysinit().
 */

void
mbad_multibus_alloc_maps( bus )
	struct bus *bus;
{
	extern struct pte *kl1pt_master;
	struct pte *kl1pt = kl1pt_master; /* kernel level 1 page-table */
	register struct pte *pte;
	register unsigned paddr;
	register int i, toc_idx;
	struct	mb_desc *mb;

	/*
	 * If no devices found, then no map to create.
	 * Otherwise map the existing MBAd's.  De-conf'd
	 * MBAd's are left unmapped (0'd pte's).
	 */

	if (MBIOmap == NULL)
		return;

	i = (struct mbad_device *)bus->owner - mbad_devices;
	toc_idx = mbad_devices[i].cd->cd_i;
#ifdef	ns32000
	register int addr;
	register int j;
	int	l1idx;
	
	if (!MBAD_EXISTS(i))
		return;
	pte = MBIOpte;
	l1idx = L1IDX(VA_MBAd(toc_idx));
	addr = 0;
	while(addr < MBAD_ADDR_SPACE) {
		ASSERT_DEBUG(*(int*)(&kl1pt[l1idx])==0,"kl1pt: MBAd");
		ASSERT_DEBUG(((int)pte & (NBPG-1))==0, "MBAd: align");
		*(int*)(&kl1pt[l1idx]) = (int)pte|PG_V|PG_R|PG_M|PG_KW;
		for (j = 0; j < NPTEPG; j++) {
			paddr = PA_MBAd(toc_idx)+addr;
			*(int*)(pte++) = PHYSTOPTE(paddr)
				| PG_V|PG_R|PG_M | PG_KW;
			addr += NBPG;
		}
		++l1idx;
	}
	MBIOpte = pte;
#endif	/* ns32000  */
	
#ifdef	i386
	if (!MBAD_EXISTS(i)) 
		return;
	
	pte = MBIOmap + toc_idx * (MBAD_ADDR_SPACE / NBPG);
	/*
	 * Insure level-1 maps it.  This is redundant except
	 * on NPTEPG*NBPG (4Meg) boundary.
	 */
	
	*(int *) &kl1pt[L1IDX(VA_MBAd(toc_idx)+KSOFF)] =
		PHYSTOPTE((int)pte & PG_PFNUM) | PG_V|PG_R|PG_M|PG_KW;
	
	/*
	 * Map the particular MBAd address space into level-2's.
	 */
	
	for (paddr = PA_MBAd(toc_idx); paddr < PA_MBAd(toc_idx+1); 
	     paddr += NBPG)
		*(int*)(pte++) = PHYSTOPTE(paddr)|PG_V|PG_R|PG_M|PG_KW;
#endif	/* i386  */
	mb = &MBAd[i];
	
	if (mb->mb_dmareg != MB_MAPS)
		CPRINTF("MBAd%d: %d maps unused.\n",
			i, MB_MAPS-mb->mb_dmareg);
#ifdef DEBUG
	if ( mbad_unused_vectors ) {
		void mbad_unused_intr();
		unchar vec;
		int j;

		for ( j = 0; j < 8; j++ )
			if ( !( mb->mb_intrs & 1<<j ) ) {
				vec = ivecall( MBAD_ERROR_BIN );
				ivecinit( MBAD_ERROR_BIN, vec,
					 mbad_unused_intr);
				(void)mbad_multibus_intr_setup( bus, j, 0,
							 MBAD_ERROR_BIN, vec,
							 SL_GROUP|TMPOS_GROUP);
				mbad_unused_vectors[8*i+j] = vec;
			}
	}
#endif /* DEBUG */
}

/*
 * mbad_multibus_dma_setup(dev, length, address, id)
 * For the given multibus dev, using the maps assigned to that dev, create a
 * dma mapping with the given id to be used for doing a DMA transfer of the
 * given length at the given address.  The caller must ensure that only one
 * invocation of this routine is active at any one time. If a DMA request
 * already exists with the given id, the new "mapping" is added to it.
 * If the entire request cannot be mapped, as much of it as can be mapped is
 * mapped.  The actual length mapped (which is less than or equal the requested
 * length) is returned.  The ID must be non-zero, or zero mapped length is
 * returned. 
 */

mbad_multibus_dma_setup( dev, length, address, id )
	struct gen_device *dev;
	long length, id;
	caddr_t address;
{
	int i, j, maps_used, j_max = 0, j_max_at;
	struct bus *bus = dev->bus;
	struct mbad_map_origin *mo = (struct mbad_map_origin *)dev->map_origin;

	maps_used = (length + ((int)address & MB_MRSIZE-1) + MB_MRSIZE-1)
		/ MB_MRSIZE;

	if ( !id )
		return 0;
	
	/* First fit */
	for ( i = 0; i < mo->nmaps - (maps_used-1); i++ ) {
		for ( j = 0; j < maps_used; j++)
			if ( mo->id[i+j] ) {
				if ( j > j_max ) {
					j_max = j;
					j_max_at = i;
				}
				i = i+j;
				break;
			}
		if ( j == maps_used )
			/* found a usable slot */
			break;
	}

	if  ( j != maps_used ) {
		/* Max not available, map the largest chunk. */
		i = j_max_at;
		maps_used = j_max;
		length = MB_MRSIZE * maps_used - ((int)address & MB_MRSIZE-1);
	}
	
	for ( j = 0; j < maps_used; j++ )
		mo->id[i+j] = id;

	if ( maps_used )
		dev->bus_map_addr = (char *) mbad_physmap( 
			&MBAd[(struct mbad_device *)bus->owner - mbad_devices],
			(int)(mo->basemap + i), address,
			(unsigned)length, maps_used);
	return length;
}

/*
 * mbad_multibus_dma_remap(dev, length, address, id)
 * For the given multibus dev, remap the maps assigned to the given dma id to
 * handle a DMA transfer of the given length starting at the given address.
 * Works like mbad_multibus_dma_setup except that no locks are required
 * provided that a given DMA id cannot be used by more than one control thread
 * at a time.  Returns the mapped bus address if successful, otherwise NULL.
 */

char *
mbad_multibus_dma_remap( dev, length, address, id )
	struct gen_device *dev;
	long length, id;
	caddr_t address;
{
	int i, j, maps_used;
	struct bus *bus = dev->bus;
	struct mbad_map_origin *mo = (struct mbad_map_origin *)dev->map_origin;

	if ( !id || length <= 0)
		return ((char *)NULL);
	
	maps_used = (length + ((int)address & MB_MRSIZE-1) + MB_MRSIZE-1)
		/ MB_MRSIZE;

	for ( i = 0; i < mo->nmaps - (maps_used-1) && mo->id[i] != id; i++ ) 
		continue;

	for ( j = i; j < i + maps_used; j++)
		if ( mo->id[j] != id ) 
			return ((char *)NULL);

	return ((char *) mbad_physmap( 
			&MBAd[(struct mbad_device *)bus->owner - mbad_devices],
			(int)(mo->basemap + i), address,
			(unsigned)length, maps_used));
}

void 
mbad_multibus_dma_done( dev, id )
	struct gen_device * dev;
	long id;
{
	struct mbad_map_origin *mo = (struct mbad_map_origin *)dev->map_origin;
	int i;

	for ( i = 0; i < mo->nmaps; i++ )
		if (mo->id[i] == id)
			mo->id[i] = 0;
}

/*ARGSUSED*/
mbad_multibus_intr_setup( bus, bus_level, bus_vector,
			 slic_bin, slic_vector, slic_dest )
	struct bus * bus;
	int bus_level, bus_vector;
	unchar slic_bin, slic_vector, slic_dest;
{
	struct mb_desc *mb;
	/*
	 * For the given bus, which had better be controlled by an alive mbad,
	 * arrange for multibus interrupts to map to a slic mIntr to the given
	 * bin, vector, and destination
	 */

#ifdef DEBUG
	if (boot_debug) {
	    CPRINTF("%s: bus %d lvl %d vec %d slic bin 0x%x vec 0x%x dest 0x%x\n",
		       "mbad_multibus_intr_setup", bus - busses, bus_level,
			bus_vector, slic_bin, slic_vector, slic_dest);
	}
#endif /* DEBUG */
	mb = &MBAd[ (struct mbad_device *)bus->owner - mbad_devices ];
	if (mb->mb_intrs & (1 << bus_level)) {
		cmn_err( CE_WARN,
			"mbad_multibus_intr_setup: duplicate int lvl %d on",
			bus_level);
		/*
		 *+ More than one device on the specified MULTIBUS (which is
		 *+ controlled by the listed mbad) is trying to use the stated
		 *+ MULTIBUS interrupt level.  The first device probed will
		 *+ use that level.  Corrective action:  reconfigure the
		 *+ MULTIBUS devices found on the MULTIBUS so
		 *+ that each device has a unique MULTIBUS interrupt level.
		 */
		cmn_err( CE_CONT, " bus %d (mbad%d)", bus - busses, mb - MBAd);
		return BUS_INTLVL_BAD;
	}
	MBAD_CMD(mb, MCMD_INIT_LINE|(unchar) bus_level);
	MBAD_CMD(mb, slic_dest);
	MBAD_CMD(mb, slic_vector);	  /* message data */
	MBAD_CMD(mb, SL_MINTR|slic_bin);	/* command */
	mb->mb_intrs |= 1 << bus_level;
	
	return BUS_INT_OK;
}

#ifdef DEBUG
void
mbad_unused_intr(vector)
{
	int i, j, numdev;

	for ( i = 0; drivers[i].numdev; i++ )
		if ( strcmp( drivers[i].name, "mbad") == 0 )
			break;

	numdev = drivers[i].numdev;

	for ( j = 0; j < numdev; j++) {
		if (!MBAD_EXISTS(i))
			continue;
		for ( i = 0; i < 8; i++)
			if ( mbad_unused_vectors[8*j+i] == vector ) {
				cmn_err( CE_WARN,
					"mbad%d: unexpected intr, level %d",
					j, i);
				/*
				 *+ A multibus interrupt was detected on an
				 *+ mbad and there is no device on that
				 *+ interrupt level on that mbad
				 */
				return;
			}
	}
	cmn_err( CE_WARN, "mbad_unused_intr: bad vector 0x%x", vector);
}
#endif	/* DEBUG */

void 
mbad_multibus_ack_int(bus, level)
	struct bus *bus;
	unchar level;
{
	spl_t	mb_spl = splhi();

	/* bus->owner->config->dev_addr is slic addr of mbad */
	sendsoft( (unchar)bus->owner->config->dev_addr, (unchar)(1<<(level)));
	splx(mb_spl);
}


/*
 * MBAd_init()
 *	Init MBAd by setting up error vectors and enable mapping registers.
 *
 * Does not set up individual multi-bus interrupt vectors.
 * Called by probe_mbad_devices() very early in init (prior to probing
 * MBAd devices), with only one processor alive and interrupts off.
 */

static
MBAd_init(mb, bin, vecbase)
	register struct mb_desc *mb;
	unchar	bin;			/* bin to take interrupts from MBAd */
	unchar	vecbase;		/* vector from bin for 1st MBAd error */
{
	register int i;
	unchar	MBAd_a_csr;

	/*
	 * Arrange that maskable interrupt to argument bin with argument
	 * vector is the result of nasty MBAd errors.
	 */
	MBAD_CMD(mb, MCMD_INIT_ERRORS);		/* next three cmds are... */
	MBAD_CMD(mb, SL_GROUP|TMPOS_GROUP);	/* destination,		  */
	MBAD_CMD(mb, vecbase);			/* message data,	  */
	MBAD_CMD(mb, SL_MINTR|bin);		/* command		  */

	/*
	 * Set up the IO/DMA space in the MBAd.
	 * We use the values placed in the MBAd's address csr by the
	 * power-up firmware, and modify the IO/DMA space window.
	 */
	MBAd_a_csr = rdslave(mb->mb_slicaddr, SL_A_CSR);
	MBAd_a_csr &= ~(SLB_S0|SLB_S1);
	MBAd_a_csr |= MBAd_IOwindow / (256*1024);
	wrslave(mb->mb_slicaddr, SL_A_CSR, MBAd_a_csr);

	/*
	 * Initialize mapping registers to 1-1 mapping and enable
	 * MBAd to use mapping registers.  1-1 mapping is necessary
	 * for debug and clean initial condition.
	 *
	 * Need to do this early to allow drivers to use mapping in
	 * boot/probe procedures.
	 */
	for(i = 0; i < MB_MAPS; i++) 
		MBADMAP(mb, i) = i;

	mb->mb_ios->mb_ctl[0] = MBC_MAPENAB;		/* ON mapping */
}

/*
 * MBAd_error()
 *	Error interrupts from MBAd's wind up here.
 *
 * Print out logical MBAd number and slic address.
 */

static void
MBAd_error(vector)
	unchar	vector;
{
	register int board;

	board = (vector - MBAd_errbase) / MBAD_ERR_ALIGN;
	cmn_err(CE_NOTE, "MBAd%d hard (access) error, slic %d.\n",
			board, MBAd[board].mb_slicaddr);
	/*
	 *+ The specified mbad generated an error interrupt.
	 */
	access_error((unchar)rdslave(MBAd[board].mb_slicaddr, SL_G_ACCERR));
	cmn_err(CE_PANIC, "MBAd hard error");
	/*
	 *+ There was an unrecoverable error on the MULTIBUS Adapter.
	 */
}

/*
 * MBAd_mapall()
 *	Allocate map-registers on a MBAd.
 *
 * This is static allocation; once allocated, never released.
 */

static
MBAd_mapall(mb, nmaps)
	register struct mb_desc *mb;
	register int	nmaps;
{
	int	val;

	if (nmaps == 0)
		return (0);

	if (mb->mb_dmareg + nmaps > MB_MAPS)
		return (-1);

	val = mb->mb_dmareg;
	mb->mb_dmareg += nmaps;

	return (val);
}


/*
 * MBAd_nmi()
 *	NMI handler while probing MBAd devices.
 *
 * We clear the access error from the processor and longjmp (returning
 * to MBAd_probe, above).
 *
 * This is enterred directly from HW vector table; however, since
 * entry is from drivers 'probe' procedure, MBAd_probe() already
 * considers scratch registers as volitile (thus, no need to save/restore
 * them).
 */

static
MBAd_nmi()
{
#if	defined(ns32000)
	unchar	regval =  rdslave(va_slic->sl_procid, SL_G_ACCERR);
#ifdef	MODELD
	int	timeout;
	/*
	 * Need a "clearnmi" for 32K.
	 */
	timeout = ((~regval)&SLB_ATMSK) == SLB_AETIMOUT;
#endif	/* MODELD */

	/*
	 * Writing anything to the processor's access-error register
	 * clears it.
	 */

	wrslave(va_slic->sl_procid, SL_G_ACCERR, 0xbb);

#else	/* SGS HW  */
#ifdef	MODELD
	/*
	 * SGS and SGS2 HW.
	 */
	int	timeout = clearnmi();
#else	/* !MODELD */
	extern	struct	ctlr_desc *slic_to_config[];
	unchar	my_slic = va_slic->sl_procid;
	unchar	bic_slic = BIC_SLIC(my_slic, slic_to_config[my_slic]->cd_flags);
	unchar	accerr_reg = (my_slic & 0x01) ? BIC_ACCERR1 : BIC_ACCERR0;
	unchar	regval = rdSubslave(bic_slic, PROC_BIC, accerr_reg);

	/*
	 * Writing anything to the BIC's access-error register clears it.
	 */

	wrSubslave(bic_slic, PROC_BIC, accerr_reg, 0xbb);
#endif	/* MODELD */
#endif	/* ns32000  */

	/*
	 * Determine why we got an NMI.  Hopefully it was NOT
	 * due to a Sequent Bus Timeout.
	 */

#ifdef	MODELD
	if (timeout) {
#else	/* !MODELD */
	if (((~regval)&SLB_ATMSK) == SLB_AETIMOUT) {
#endif	/* MODELD */
		/*
		 * We COULD deconfigure the board here, and keep on running.
		 * But, we panic instead, just to play it safe, and protect
		 * naive system administrators.
		 */
		cmn_err(CE_NOTE,
			"mbad %d is NOT FUNCTIONING, SQT BUS TIMEOUT\n",
			MBAd_idx);
		/*
		 *+ The specified mbad experienced a Sequent bus timeout.  
		 */

		cmn_err(CE_PANIC, "mbad TIMEOUT");
		/*
		 *+ The system bus timed out on a request to the MULTIBUS Adapter.
		 */
	}

	/*
	 * "return" to MBAd_probe.
	 */

	longjmp(&MBAd_nmi_jmp);
}

/*
 * mbad_physmap()
 *	Set up DMA transfer to physical memory in MBAd map registers.
 *
 * addr argument is physical-addr of target/source of DMA.
 * Used during boot/probe in drivers to allow multibus devices to
 * see into system memory.
 *
 * Returns DMA address to tell the multi-bus device.
 *
 * No checking on legality of arguments.
 */

unsigned
mbad_physmap(mb, dmabase, addr, len, nmap)
	register struct	mb_desc	*mb;		/* MBAd descriptor */
	register int	dmabase;		/* 1st map register */
	caddr_t		addr;			/* start address to map */
	unsigned	len;			/* # bytes to map */
	int		nmap;			/* # map registers (sanity) */
{
	register int count;
	unsigned offset;
	unsigned val;

	offset = (int)addr & (MB_MRSIZE-1);
	count = (offset + len + (MB_MRSIZE-1)) / MB_MRSIZE;
	ASSERT(count <= nmap, "mbad_physmap: too big");
	/*
	 *+ The number of MULTIBUS Adapter map registers required to
	 *+ perform this DMA request was not sufficient for this transfer
	 *+ size.
	 */

	val = mb->mb_dmabase + dmabase * MB_MRSIZE + offset;

	for (; count--; addr += MB_MRSIZE, ++dmabase) {
		ASSERT(dmabase < MB_MAPS, "mbad_physmap: bad map");
		/*
		 *+ When setting up the MULTIBUS Adapter map registers for
		 *+ a DMA transfer, the kernel found that it was using nonexistent
		 *+ registers.
		 */
#ifdef	DEBUG
		if (mbad_debug)
			CPRINTF("MBAd map[%d] = 0x%x\n", dmabase, PHYSTOMB(addr));
#endif	/* DEBUG  */
		ASSERT((ulong) addr < MAX_MBAD_ADDR_MEM,
					"mbad_physmap: address > 64Meg");
		/*
		 *+ The kernel attempted to map addresses greater than 64 MB.
		 *+ The MULTIBUS Adapter cannot access physical memory
		 *+ at addresses greater than 64 MB.
		 */
		MBADMAP(mb, dmabase) = PHYSTOMB(addr);
	}

	return (val);
}

#if	NBPG >= MB_MRSIZE		/* eg, SGS */

/*
 * mbad_setup()
 *	Takes a "bp" (struct buf *), a mb_desc structure,
 *	a DMA register base number, and sets up the DMA.
 *
 * Returns DMA address to tell the multi-bus device.
 * Does *no* locking.
 * Panics if bad pte found; "can't" happen.
 *
 * B_RAWIO, B_PTEIO, B_PTBIO cases must flush TLB to avoid stale mappings
 * thru Usrptmap[], since this is callable from interrupt procedures.
 *
 * Implementation assumes NBPG >= MB_MRSIZE.
 */

unsigned
mbad_setup(bp, mb, dmabase, nmap)
	struct	buf	*bp;			/* buffer header */
	struct	mb_desc	*mb;			/* MBAd descriptor */
	register int	dmabase;		/* 1st map register */
	int	nmap;				/* # map registers (sanity) */
{
	register struct pte *pte;		/* pte mapping data */
	register unsigned paddr;		/* phys-addr to map */
	int	count;				/* # MB maps to set up */
	unsigned val;				/* MB dma address to use */
	unsigned pgoffset;			/* RAW IO offset in page */
	unsigned mboffset;			/* RAW IO offset in MB map */
	struct pte *vtopte();

	/*
	 * Source/target pte's are found differently based on type
	 * of IO operation.
	 */

	switch(bp->b_iotype) {
	case B_RAWIO:					/* RAW IO */
		/*
		 * In this case, must look into alignment of physical
		 * memory, since we can start on arbitrary boundary.
		 */

		flush_tlb();
		pte = vtopte(bp->b_proc, btop(bp->b_un.b_addr));
		pgoffset = (int)bp->b_un.b_addr & (NBPG-1);
		mboffset = pgoffset & (MB_MRSIZE-1);
		count = (unsigned)(mboffset + bp->b_bcount + (MB_MRSIZE-1))
						>> MB_MRSIZEL2;
		val = mb->mb_dmabase + dmabase * MB_MRSIZE + mboffset;
		paddr = PTETOPHYS(*pte) + (pgoffset & ~(MB_MRSIZE-1));
		break;

	case B_FILIO:					/* file-sys IO */
		/*
		 * Filesys/buffer-cache IO.  These are always cluster aligned
		 * both physically and virtually.
		 */

		pte = &Sysmap[btop(bp->b_un.b_addr)];
		count = (unsigned)(bp->b_bcount + MB_MRSIZE-1) >> MB_MRSIZEL2;
		val = mb->mb_dmabase + dmabase * MB_MRSIZE;
		paddr = PTETOPHYS(*pte);
#ifdef	DEBUG
		if (((int)bp->b_un.b_addr & (MB_MRSIZE-1)) != 0) {
			CPRINTF("bp=0x%x, addr=0x%x\n", bp, bp->b_un.b_addr);
			cmn_err(CE_PANIC, "mbad_setup: bad FS IO");
		}

		/*
		 * Kludge for MULTIBUS tape driver on systems with
		 * > 64 Meg of memory.  B_PHYS is used to know this
		 * is a tape request.
		 */

		if ((bp->b_flags & B_PHYS) == 0
		&&  (bp->b_bcount <= 0 || bp->b_bcount > MAXBSIZE)) {
			CPRINTF("bp=0x%x, bcount=%d\n", bp, bp->b_bcount);
			cmn_err(CE_PANIC, "mbad_setup: bad FS count");
		}
#endif	/* DEBUG  */
		break;


	case B_PTBIO:					/* Page-Table IO */
		/*
		 * Page-Table IO: like B_PTEIO, but can start/end with
		 * non-cluster aligned memory (but is always HW page
		 * aligned).  Count is multiple of NBPG.
		 *
		 * Separate case for greater efficiency in B_PTEIO.
		 *
		 * On SGS, this is identical to B_PTEIO, so fall into...
		 */

	case B_PTEIO:					/* swap/page IO */
		/*
		 * Pte-based IO -- already know pte of 1st page, which
		 * is cluster aligned, and b_count is a multiple of CLBYTES.
		 */

		flush_tlb();
		pte = bp->b_un.b_pte;
		count = (unsigned)(bp->b_bcount) >> MB_MRSIZEL2;
		val = mb->mb_dmabase + dmabase * MB_MRSIZE;
		paddr = PTETOPHYS(*pte);
		break;

	default:
		bad_b_iotype();
		/*NOTREACHED*/
	}

	/*
	 * Check count and set up the map registers.
	 * Careful to pick up new phys-address when cross page boundary.
	 */
	mbad_setup_range_check(nmap, dmabase, count);

	while (count--) {
#ifdef	DEBUG
		if (mbad_debug)
			CPRINTF("MBAd map[%d] = 0x%x\n",dmabase,PHYSTOMB(paddr));
#endif	/* DEBUG */
		ASSERT(paddr<MAX_MBAD_ADDR_MEM, "mbad_setup: address >= 64Meg");
		/*
		 *+ The kernel attempted to map addresses greater than 64 MB.
		 *+ The MULTIBUS Adapter cannot access physical memory
		 *+ at addresses greater than 64 MB.
		 */
		MBADMAP(mb, dmabase) = PHYSTOMB(paddr);
		++dmabase;
		paddr += MB_MRSIZE;
		if ((paddr & (NBPG-1)) == 0 && count) {
			++pte;
			ASSERT_DEBUG(PTEPF(*pte)!=0, "mbad_setup: null pfnum");
			paddr = PTETOPHYS(*pte);
		}
	}

	return (val);
}

#else	/* NBPG < MB_MRSIZE  */ /* eg, FGS */

/*
 * mbad_setup()
 *	Takes a "bp" (struct buf *), a mb_desc structure,
 *	a DMA register base number, and sets up the DMA.
 *
 * Returns DMA address to tell the multi-bus device.
 * Does *no* locking.
 * Panics if bad pte found; "can't" happen.
 *
 * Implementation assumes NBPG <= MB_MRSIZE <= CLBYTES.
 *
 * Don't bother checking MBAD physical addressing (vs 64Meg), since
 * all current small page-size systems are limited to < 64Meg already.
 */

unsigned
mbad_setup(bp, mb, dmabase, nmap)
	register struct	buf	*bp;		/* buffer header */
	register struct	mb_desc	*mb;		/* MBAd descriptor */
	register int	dmabase;		/* 1st map register */
	int	nmap;				/* # map registers (sanity) */
{
	register struct pte *pte;
	register int count;
	unsigned offset;
	unsigned val;
	struct pte *vtopte();

	/*
	 * Source/target pte's are found differently based on type
	 * of IO operation.
	 */

	switch(bp->b_iotype) {
	case B_RAWIO:					/* RAW IO */
		/*
		 * In this case, must look into alignment of physical
		 * memory, since we can start on arbitrary boundary.
		 */

		flush_tlb();
		pte = vtopte(bp->b_proc, btop(bp->b_un.b_addr));
		offset = PTEMBOFF(*pte) + ((int)bp->b_un.b_addr & (NBPG-1));
		pte -= btop(offset);
		count = (offset + bp->b_bcount + (MB_MRSIZE-1)) / MB_MRSIZE;
		break;

	case B_FILIO:					/* file-sys IO */
		/*
		 * Filesys/buffer-cache IO.  These are always cluster aligned
		 * both physically and virtually.
		 */

		pte = &Sysmap[btop(bp->b_un.b_addr)];
		count = (bp->b_bcount + MB_MRSIZE-1) / MB_MRSIZE;
		offset = 0;
#ifdef	DEBUG
		if (((int)bp->b_un.b_addr & (MB_MRSIZE-1)) != 0) {
			CPRINTF("bp=0x%x, addr=0x%x\n", bp, bp->b_un.b_addr);
			cmn_err(CE_PANIC, "mbad_setup: bad FS IO");
		}
		if (bp->b_bcount <= 0 || bp->b_bcount > MAXBSIZE) {
			CPRINTF("bp=0x%x, bcount=%d\n", bp, bp->b_bcount);
			cmn_err(CE_PANIC, "mbad_setup: bad FS count");
		}
#endif	/* DEBUG  */
		break;

	case B_PTEIO:					/* swap/page IO */
		/*
		 * Pte-based IO -- already know pte of 1st page, which
		 * is cluster aligned, and b_count is a multiple of CLBYTES.
		 */

		flush_tlb();
		pte = bp->b_un.b_pte;
		count = (bp->b_bcount + MB_MRSIZE-1) / MB_MRSIZE;
		offset = 0;
		break;

	case B_PTBIO:					/* Page-Table IO */
		/*
		 * Page-Table IO: like B_PTEIO, but can start/end with
		 * non-cluster aligned memory (but is always HW page
		 * aligned).  Count is multiple of NBPG.
		 *
		 * Separate case for greater efficiency in B_PTEIO.
		 */

		flush_tlb();
		pte = bp->b_un.b_pte;
		offset = PTEMBOFF(*pte);
		pte -= btop(offset);
		count = (offset + bp->b_bcount + (MB_MRSIZE-1)) / MB_MRSIZE;
		break;

	default:
		bad_b_iotype();
		/*NOTREACHED*/
	}

	/*
	 * Check count, figure return value, and set up the map registers.
	 * This code is a clone of mbad_dma().
	 */

	mbad_setup_range_check(nmap, dmabase, count);

	val = mb->mb_dmabase + dmabase * MB_MRSIZE + offset;

	for(; count--; pte += (MB_MRSIZE/NBPG), ++dmabase) {
		ASSERT_DEBUG(PTEPF(*pte) != 0, "mbad_setup: null pfnum");
		ASSERT_DEBUG((PTETOPHYS(*pte) & (MB_MRSIZE-1)) == 0,
			     "mbad_setup: unaligned");
#ifdef	DEBUG
		if (mbad_debug)
			CPRINTF("MBAd map[%d] = 0x%x\n", dmabase, PHYSTOMB(PTETOPHYS(*pte)));
#endif	/* DEBUG  */
		MBADMAP(mb, dmabase) = PHYSTOMB(PTETOPHYS(*pte));
	}

	return (val);
}

#endif	/* NBPG >= MB_MRSIZE  */

static void
bad_b_iotype()
{
		cmn_err(CE_PANIC, "mbad_setup: bad b_iotype");
		/*
		 *+ The setup function to program the MULTIBUS Adapter map
		 *+ registers for a DMA transfer was called with an invalid
		 *+ buffer type describing the I/O request.
		 */
}

static void
mbad_setup_range_check(nmap, dmabase, count) 
	int nmap, dmabase, count;
{
	ASSERT(count <= nmap, "mbad_setup: too big");
	/*
	 *+ The number of MULTIBUS Adapter map registers required to
	 *+ perform this DMA request was not sufficient for this transfer
	 *+ size.
	 */
	ASSERT(dmabase + count <= MB_MAPS, "mbad_setup: bad map");
	/*
	 *+ When setting up the MULTIBUS Adapter map registers for
	 *+ a DMA transfer, the kernel found it was using nonexistent
	 *+ registers.
	 */
}
