/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ident "$Header: pty.c 1.18 90/04/20 $"

/*
 *  pty.c
 *	STREAMS pty (pseudo-tty) driver
 */

/* $Log:	pty.c,v $
 */

/* #ident	"@(#)kern-port:io/pts.c	10.9"*/
/*
 *	         Copyright (c) 1987 AT&T
 *		   All Rights Reserved
 *
 *     THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T
 *   The copyright notice above does not evidence any actual
 *   or intended publication of such source code.
 */



#include "sys/param.h"
#include "sys/sysmacros.h"
#include "sys/signal.h"
#include "sys/types.h"
#include "sys/devsw.h"
#include "sys/systm.h"
#include "sys/mutex.h"
#include "sys/timer.h"
#include "sys/intctl.h"
#include "sys/file.h"
#include "sys/dir.h"
#include "sys/user.h"
#include "sys/proc.h"
#include "sys/session.h"
#include "sys/stream.h"
#include "sys/stropts.h"
#include "sys/errno.h"
#define STREAMS_DEBUG_ID	STRID_PTS /* enable debug macro defintions */
#include "sys/debug.h"
#include "sys/cmn_err.h"
#include "sys/strid.h"
#include "sys/termio.h"
#include "sys/strpty.h"
#include "sys/ptyms.h"

#define PTDEBUG  DPRINTF1

/*
 *   Slave STREAMS pty driver
 */


static struct module_info pts_info = {
/*  module ID    module name   min pkt size    max pkt siz   hiwat   lowat  */
     STRID_PTS,     "pts",         0,            INFPSZ,      300,    100
};

static int ptsopen(), ptsclose(), ptsrsrv(), ptswsrv();

static struct qinit ptsrint = {
/*  put	    server     open       close   admin(3bnet)	minfo     mstat */
   NULL,   ptsrsrv,   ptsopen,   ptsclose,   NULL,    &pts_info,   NULL
};
static struct qinit ptswint = {
/*  put	    server     open       close   admin(3bnet)	minfo     mstat */
   putq,   ptswsrv,   ptsopen,   ptsclose,   NULL,    &pts_info,   NULL
};
struct streamtab ptsinfo = {&ptsrint, &ptswint, NULL, NULL};


/*
 * The following externs are defined in pty_space.c
 */
extern struct pt_ttys ptms_tty[];	
extern int pt_cnt;	/* number of elements in ptms_tty[] */

extern int ptm_major;	/* major number of master driver */
extern int pts_major;	/* major number of slave driver */
extern ulong ptsflags;	/* initial line discipline flags */


extern struct devops ptmdevops;
extern struct devops ptsdevops;


/*
 * Open the slave device. 
 *
 * Reject a clone open and do not allow the driver to be pushed.
 * The streamhead serialzes open/close so that we needn't worry
 * about racing with another slave open or close until we return.
 */
static int 
ptsopen(rqp, dev, flag, sflag)
queue_t *rqp;
dev_t dev;
int flag;
int sflag;
{
        register struct pt_ttys *ptsp;
	int firstopen;		/* non-zero => this is first open */
	spl_t spl;

	PTDEBUG("pts: ptsopen(qp:0x%x, dev:0x%x, flag:0x%x, sflag:0x%x)\n",
			rqp, dev, flag, sflag);

        dev = minor(dev);
	
	/* Check for non-driver open */
	if (sflag != DRVROPEN) {
		PTDEBUG("pts: ptsopen(): Non-driver open: sflag = 0x%x\n", sflag);
		u.u_error = EINVAL;
		return(OPENFAIL);
	}
        if ((dev < 0) || (dev >= pt_cnt)) {
		PTDEBUG("pts: ptsopen(): invalid minor number (0x%x)\n", dev);
		u.u_error = ENODEV;
                return(OPENFAIL);
	}
        ptsp = &ptms_tty[dev];
	spl = p_lock(&ptsp->pt_lock, SPLSTR);

	/* 
	 * if first open of slave device then initialize
	 */
	if ((ptsp->pt_state & PTSOPEN) == 0) {

		PTDEBUG("pts: first open of slave\n");
		ASSERT_DEBUG(ptsp->pts_wrq == NULL, "pts: ptsopen()");

		firstopen = 1;
		ptsp->pt_state |= PTSOPEN;  /* mark slave as open */
		rqp->q_ptr = (char *) ptsp;
		WR(rqp)->q_ptr = (char *) ptsp;
		/* wait to set ptsp->pts_wrq until enable master below */

		ptsp->pts_cflag = ptsflags; /* initial line discipline flags */
		ptsp->pts_wd.ws_row = 0;   /* initial window size values */
		ptsp->pts_wd.ws_col = 0;
		ptsp->pts_wd.ws_xpixel = 0;
		ptsp->pts_wd.ws_ypixel = 0;
	} else {
		PTDEBUG("pts: slave already open\n");
		ASSERT_DEBUG(ptsp->pts_wrq == WR(rqp), "pts: ptsopen()");
		ASSERT_DEBUG(rqp->q_ptr == (char *) ptsp, "pts: ptsopen()");
		ASSERT_DEBUG(WR(rqp)->q_ptr == (char *) ptsp, "pts: ptsopen()");

		firstopen = 0;
	}


	/* Wait for "carrier" (i.e., master device to open), if required */ 

	if (!(flag & FNDELAY)) {
		while (!(ptsp->pt_state & PTMOPEN) && 
		       !(ptsp->pts_cflag & CLOCAL)) {
			/*
			 * Waiting for "carrier" (i.e., master device to open).
			 * If we get signalled out of the wait, and we are the 
			 * first opener of this slave device, clean up
			 * and return an error code.  If we are not the first
			 * opener, just return the error code.
			 */
			if (p_sema_v_lock(&ptsp->pts_openwait, PZERO+1 | PCATCH,
					  &ptsp->pt_lock) == PCATCH) {
				PTDEBUG("pts: caught signal\n");
				if (firstopen) {
					PTDEBUG("pts: backout first open\n");
					rqp->q_ptr = NULL;
					WR(rqp)->q_ptr = NULL;
					spl = p_lock(&ptsp->pt_lock, SPLSTR);
					ptsp->pt_state &= ~PTSOPEN;
					v_lock(&ptsp->pt_lock, spl);
				}
				return(OPENFAIL);
			}
			spl = p_lock(&ptsp->pt_lock, SPLSTR);
		}
	} 

	REQ_CTTY();	/* inform the stream head that we could be a */
			/* controlling terminal 		     */

	if (firstopen)
		ptsp->pts_wrq = WR(rqp);   /* enable master -> slave sending */

	if (ptsp->ptm_wrq) {
		/* Enable master write server to send any queued messages */
		PTDEBUG("pts: ptsopen(): qenable(master)\n");
		qenable(ptsp->ptm_wrq);
	}

	v_lock(&ptsp->pt_lock, spl);

	PTDEBUG("pts: ptsopen(): done\n");
	return(OPENOK);
}



/*
 * Close slave device.
 *
 * Streamhead allows only a single open/close at a time so don't
 * need to worry about racing with another slave open or close.
 */
static int 
ptsclose(rqp)
queue_t *rqp;
{
        register struct pt_ttys *ptsp;
	spl_t spl;

	PTDEBUG("pts: ptsclose(rqp:0x%x)\n", rqp);

	ASSERT_DEBUG(rqp->q_ptr != NULL, "pts: ptsclose()");

        ptsp = (struct pt_ttys *)rqp->q_ptr;
	spl = p_lock(&ptsp->pt_lock, SPLSTR);

	ASSERT_DEBUG(ptsp->pt_state & PTSOPEN, "pts: ptsclose()");
	ASSERT_DEBUG(ptsp->pts_wrq == WR(rqp), "pts: ptsclose()");

        ptsp->pts_wrq = NULL;	/* disable master -> slave sending */

	if (ptsp->ptm_wrq) {

		/* Let the open master device know we've closed */

		PTDEBUG("pts: ptsclose(): M_ERROR master\n");
		if (!putnextctl1(RD(ptsp->ptm_wrq), M_ERROR, EIO)) {
			cmn_err(CE_WARN,
			     "pty: master close M_ERROR putnextctl1 failed\n");
			/*
			 *+ An attempt to send an error indication to the
			 *+ pty master device during slave device close
			 *+ failed due to lack of system resources.
			 */
		}
	}

	v_lock(&ptsp->pt_lock, spl);
	disable_procs(rqp);	/* ask STREAMS to disable our servers */
	spl = p_lock(&ptsp->pt_lock, SPLSTR);

	ptsp->pt_state &= ~PTSOPEN;	/* mark slave as closed */
	rqp->q_ptr = NULL;
	WR(rqp)->q_ptr = NULL;

	v_lock(&ptsp->pt_lock, spl);

	PTDEBUG("pts: ptsclose(): done\n");
	return;
}


/*
 * Slave read server.
 */
static int 
ptsrsrv(qp)
queue_t *qp;
{
	ptxrsrv(qp, ISSLAVE);
}
		


/*
 * Slave write server.
 */
static int 
ptswsrv(qp)
queue_t *qp;
{
	ptxwsrv(qp, ISSLAVE);
}


#if ((FLUSHR | FLUSHW) != FLUSHRW)
The assumption in the code below won't work.
#endif

/*
 * Generic pty write server.  Used for both master and slave device.
 *
 * If there are messages on this queue that can be sent to 
 * other device, send them via putnext(). Else, if queued messages 
 * cannot be sent, leave them on this queue. If priority 
 * messages on this queue, send them to other device if there is
 * a connection (but regardless of canputnext()) or discard them
 * if there is no connection.
 */

static int 
ptxwsrv(qp, ismaster)
queue_t *qp;
int ismaster;	/* non-zero => qp is for the master device, else slave */
{
        register struct pt_ttys *ptxp;
        register queue_t *other_rdq;
        register queue_t *other_wrq;
	mblk_t *mp;
	spl_t spl;
	
	PTDEBUG("ptx: ptxwsrv(%s, qp:0x%x) - write service proc\n",
			(ismaster) ? "master" : "slave", qp);
	ASSERT_DEBUG(qp->q_ptr, "ptx: ptxwsrv()");

        ptxp = (struct pt_ttys *) qp->q_ptr;

	/*
	 * while there are messages on this write queue...
	 */
	while ((mp = getq(qp)) != NULL) {

		switch (mp->b_datap->db_type) {

		/*
		 * If write queue request, flush this write queue.
		 * Convert read flush to write flush and visa versa,
		 * then forward to other stream.
		 */
		case M_FLUSH:
		{
			unsigned char tmp;

			PTDEBUG("ptx: ptxwsrv(%s): M_FLUSH%s%s\n",
				(ismaster) ? "master" : "slave",
				(*mp->b_rptr & FLUSHR) ? "R" : "",
				(*mp->b_rptr & FLUSHW) ? "W" : "");

			if (*mp->b_rptr & FLUSHW)
				flushq(qp, FLUSHDATA);

			tmp = *mp->b_rptr;
			*mp->b_rptr &= ~(FLUSHW | FLUSHR);
			if (tmp & FLUSHR)
				*mp->b_rptr |= FLUSHW;
			if (tmp & FLUSHW)
				*mp->b_rptr |= FLUSHR;

			/* FALL THROUGH ... */
		}
		/*
		 * Send data messages to other device
		 * (Also used for priority FLUSH msgs.)
		 */
		case M_DATA:

			PTDEBUG("ptx: ptxwsrv(%s): M_DATA\n",
				(ismaster) ? "master" : "slave");

			spl = p_lock(&ptxp->pt_lock, SPLSTR);
			other_wrq = ismaster ? ptxp->pts_wrq 
					     : ptxp->ptm_wrq;
			/*
			 * If there's not a connection, discard priority 
			 * messages and hold non-priority messages.
			 */
			if (other_wrq == NULL) {
				v_lock(&ptxp->pt_lock, spl);
				PTDEBUG("ptx: ptxwsrv(%s): NO CONNECTION\n",
					(ismaster) ? "master" : "slave");
				if (pcmsg(mp->b_datap->db_type)) {
					PTDEBUG("ptx: ptxwsrv(%s): freeing priority msg type 0x%x\n",
						(ismaster) ? "master" : "slave",
						mp->b_datap->db_type);
					freemsg(mp);
					break;
				} else {
					PTDEBUG("ptx: ptxwsrv(%s): putbq\n",
						ismaster ? "master" : "slave");
					putbq(qp, mp);
					return;
				}
			}
			other_rdq = RD(other_wrq);
			/*
			 * If don't have control message and cannot put msg
			 * on other's read queue, put it back on this queue.
			 */
			if (!pcmsg(mp->b_datap->db_type) && 
			    !canputnext(other_rdq)) {
				PTDEBUG("ptx: ptxwsrv(%s): FLOW CONTROL - putbq\n",
					ismaster ? "master" : "slave");
				v_lock(&ptxp->pt_lock, spl);
				putbq(qp, mp);
				return;
			}
			/*
			 * else send the message up other's stream
			 */
			PTDEBUG("ptx: ptxwsrv(%s): send msg to %s\n",
				(ismaster) ? "master" : "slave",
				(ismaster) ? "slave"  : "master");
			/*
			 * Need to hold pt_lock during putnext to
			 * synchronize with close of other_rdq.  It's 
			 * ok to hold the lock since, (a) this bypasses
			 * processing by other_rdq and goes directly
			 * to it's q_next, and (b) should this putnext
			 * cause a message to loop back to other_wrq,
			 * the message will be queued at the write
			 * server and processed asynchronously rather
			 * than during this putnext.
			 *
			 * Note that the message here could be either
			 * a regular or priority message.
			 */
			putnext(other_rdq, mp);
			v_lock(&ptxp->pt_lock, spl);
			break;
		/*
		 * ioctls
		 */
		case M_IOCTL:
		case M_PCIOCTL:

			PTDEBUG("ptx: ptxwsrv(%s): IOCTL\n",
				(ismaster) ? "master" : "slave");
			ptxioctl(qp, mp, ismaster);
			break;

		case M_IOCDATA:
		{
			struct iocblk *iocp;
			uint tmp;

			PTDEBUG("ptx: ptxwsrv(%s): IOCDATA\n",
				(ismaster) ? "master" : "slave");

			iocp = (struct iocblk *) mp->b_rptr;
			switch (iocp->ioc_cmd) {

			case TIOCGWINSZ:
			case TIOCSWINSZ:

				spl = p_lock(&ptxp->pt_lock, SPLSTR);
				winsizioctl(RD(qp), mp, &ptxp->pts_wd,
					RD(ptxp->pts_wrq), &ptxp->pt_lock, spl,
					&tmp);
				break;
			
			case TIOCPTYDEV:

				PTDEBUG("ptx: ptxwsrv(%s) TIOCPTYDEV ack M_IOCDATA\n",
					(ismaster) ? "master" : "slave");
				ASSERT_DEBUG(ismaster, "ptx: ptxwsrv()");

				iocp = (struct iocblk *)mp->b_rptr;
				iocp->ioc_error = 0;
				iocp->ioc_count = 0;
				iocp->ioc_rval = 0;
				mp->b_datap->db_type = M_IOCACK;
				qreply(qp, mp);
				break;

			default:
				PTDEBUG("ptx: ptxwsrv(%s): discarding unexpected IOCDATA, ioc_cmd=0x%x\n",
					(ismaster) ? "master" : "slave",
					iocp->ioc_cmd);
				freemsg(mp);
				break;
			}
			break;
		}
		/*
		 * unknown messages get junked
		 */
		default:
			PTDEBUG("ptx: ptxwsrv(%s): discarding unexpected message type 0x%x\n",
				(ismaster) ? "master" : "slave",
				mp->b_datap->db_type);
			freemsg(mp);
			break;
		}
	}
	PTDEBUG("ptx: ptxwsrv(%s): queue now empty; done\n",
		(ismaster) ? "master" : "slave");
}


/*
 * Generic pty read server.  Used for both master and slave device.
 *
 * Enable the write side of the other stream. This triggers the 
 * other stream to send any messages queued on its write side to
 * the read side of this stream.
 */
static int 
ptxrsrv(qp, ismaster)
queue_t *qp;
int ismaster;	/* non-zero => qp is for the master device, else slave */
{
        register struct pt_ttys *ptxp;
        register queue_t *other_wrq;
	spl_t spl;
	
	PTDEBUG("ptx: ptxrsrv(%s, qp:0x%x) - read service proc\n",
			(ismaster) ? "master" : "slave", qp);
	ASSERT_DEBUG(qp->q_ptr, "ptx: ptxrsrv()");

        ptxp = (struct pt_ttys *) qp->q_ptr;
	spl = p_lock(&ptxp->pt_lock, SPLSTR);

	other_wrq = ismaster ? ptxp->pts_wrq : ptxp->ptm_wrq;

	if (other_wrq == NULL) {
		v_lock(&ptxp->pt_lock, spl);
		PTDEBUG("ptx: ptxrsrv(%s): no %s; done\n",
			(ismaster) ? "master" : "slave",
			(ismaster) ? "slave"  : "master");
		return;
	}
	/*
	 * Need to hold pt_lock during qenable to
	 * synchronize with close of other_wrq.  It's 
	 * ok to hold the lock since qenable causes
	 * the server to be run asynchronously rather
	 * than during the qenable call.
	 */
	qenable(other_wrq);
	v_lock(&ptxp->pt_lock, spl);

	PTDEBUG("ptx: ptxrsrv(%s): done\n", (ismaster) ? "master" : "slave");
}



/*
 * Process M_IOCTL and M_PCIOCTL messages.  
 * 
 * Acknowlege message and send it back upstream if all goes well.
 */

static int
ptxioctl(wr_q, mp, ismaster)
	queue_t *wr_q;	/* write queue */
	mblk_t *mp;
	int ismaster;	/* non-zero => wr_q is for master device, else slave */
{
        register struct pt_ttys *ptxp;
	struct copyreq	*send_buf_p;
	struct iocblk *iocp;
	struct termio *tptr;
	struct termios *tsptr;
	mblk_t *bp;
	spl_t spl;
	uint tmp;

	PTDEBUG("ptx: ptxioctl(%s, wr_q:0x%x mp:0x%x) - write service proc\n",
			(ismaster) ? "master" : "slave", wr_q, mp);
	ASSERT_DEBUG(wr_q->q_ptr, "ptx: ptxioctl()");
	ASSERT_DEBUG(mp->b_rptr != NULL, "ptx: ptxioctl()");
	ASSERT_DEBUG(((mp->b_datap->db_type == M_IOCTL) || (mp->b_datap->db_type == M_PCIOCTL)), "ptx: ptxioctl()");

        ptxp = (struct pt_ttys *) wr_q->q_ptr;
	iocp = (struct iocblk *) mp->b_rptr;

	switch (iocp->ioc_cmd) {

	case TIOCGWINSZ:
		PTDEBUG("ptx: ptxioctl(%s) TIOCGWINSZ\n",
				(ismaster) ? "master" : "slave");
	case TIOCSWINSZ:
		PTDEBUG("ptx: ptxioctl(%s) TIOCSWINSZ\n",
				(ismaster) ? "master" : "slave");

		spl = p_lock(&ptxp->pt_lock, SPLSTR);
		winsizioctl(RD(wr_q), mp, &ptxp->pts_wd, RD(ptxp->pts_wrq),
				&ptxp->pt_lock, spl, &tmp);
		return;

	case TCSBRK:				/* Do 0.25 secs of BREAK */

		if (ismaster)
			goto badioctl;

		PTDEBUG("ptx: ptxioctl(%s) TCSBRK\n",
				(ismaster) ? "master" : "slave");

		/* We don't send a break character */
		mp->b_datap->db_type = M_IOCACK;
		iocp->ioc_count = 0;
		qreply(wr_q, mp);
		return;

	case TCSETA:				/* Set parameters */
	case TCSETAW:
	case TCSETAF:				/* Set params with read flush */

		if (ismaster)
			goto badioctl;

		PTDEBUG("ptx: ptxioctl(%s) TCSETA*\n",
				(ismaster) ? "master" : "slave");

		ASSERT_DEBUG(iocp->ioc_count == sizeof(struct termio), 
				"ptx: ptioctl()");
		tptr = (struct termio *) mp->b_cont->b_rptr;
		ptxtcset(wr_q, mp, ismaster, (ulong)tptr->c_cflag,
			(iocp->ioc_cmd == TCSETAF));
		return;

	case TCSETP:				/* Set parameters */
	case TCSETPW:
	case TCSETPF:				/* Set params with read flush */

		if (ismaster)
			goto badioctl;

		PTDEBUG("ptx: ptxioctl(%s) TCSETP*\n",
				(ismaster) ? "master" : "slave");

		ASSERT_DEBUG(iocp->ioc_count == sizeof(struct termios),
				"ptx: ptioctl()");
		tsptr = (struct termios *) mp->b_cont->b_rptr;
		ptxtcset(wr_q, mp, ismaster, tsptr->c_cflag, 
			(iocp->ioc_cmd == TCSETPF));
		return;

	case TCGETA:					/* Get parameters */

		if (ismaster)
			goto badioctl;

		PTDEBUG("ptx: ptxioctl(%s) TCGETA\n",
				(ismaster) ? "master" : "slave");

		if (mp->b_cont == NULL) {
			if ((bp = allocb(sizeof(struct termio), BPRI_MED)) == NULL) {
				cmn_err(CE_WARN,
				     "pts: TCGETA data block allocb failed\n");
				/*
				 *+ An attempt to allocate a streams data buffer
				 *+ for TCGETA parameters failed.  
				 */
				mp->b_datap->db_type = M_IOCNAK;
				qreply(wr_q, mp);
				return;
			}
			/* clear unset fields */ 
			bzero((char *)bp->b_rptr, sizeof(struct termio));
			mp->b_cont = bp;	/* Link data buffer */
		} else {
			if (mp->b_cont->b_datap->db_lim - 
			    mp->b_cont->b_datap->db_base < sizeof(struct termio)) {
				cmn_err(CE_WARN,
				 "pts: TCGETA received data block bad size\n");
				/*
				 *+ A TCGETA ioctl with a preallocated buffer
				 *+ does not have enough allocated space to
				 *+ store the requested information.  
				 */
				mp->b_datap->db_type = M_IOCNAK;
				qreply(wr_q, mp);
				return;
			}
		}
		tptr = (struct termio *) mp->b_cont->b_wptr;

		spl = p_lock(&ptxp->pt_lock, SPLSTR);
		tptr->c_cflag = ptxp->pts_cflag;	/* Load cflags */
		v_lock(&ptxp->pt_lock, spl);

		mp->b_cont->b_wptr += sizeof(struct termio);
		iocp->ioc_count = sizeof(struct termio);
		mp->b_datap->db_type = M_IOCACK;
		qreply(wr_q, mp);
		return;

	case TCGETP:					/* Get parameters */

		if (ismaster)
			goto badioctl;

		PTDEBUG("ptx: ptxioctl(%s) TCGETP\n",
				(ismaster) ? "master" : "slave");

		if (mp->b_cont == NULL) {
			if ((bp = allocb(sizeof(struct termios), BPRI_MED)) == NULL) {
				cmn_err(CE_WARN,
				     "pts: TCGETP data block allocb failed\n");
				/*
				 *+ An attempt to allocate a streams data buffer
				 *+ for TCGETP parameters failed.  
				 */
				mp->b_datap->db_type = M_IOCNAK;
				qreply(wr_q, mp);
				return;
			}
			/* clear unset fields */ 
			bzero((char *)bp->b_rptr, sizeof(struct termios));
			mp->b_cont = bp;	/* Link data buffer */
		} else {
			if (mp->b_cont->b_datap->db_lim - 
			    mp->b_cont->b_datap->db_base < sizeof(struct termios)) {
				cmn_err(CE_WARN,
				 "pts: TCGETP received data block bad size\n");
				/*
				 *+ A TCGETP ioctl with a preallocated buffer
				 *+ does not have enough allocated space to
				 *+ store the requested information.  
				 */
				mp->b_datap->db_type = M_IOCNAK;
				qreply(wr_q, mp);
				return;
			}
		}
		tsptr = (struct termios *) mp->b_cont->b_wptr;

		spl = p_lock(&ptxp->pt_lock, SPLSTR);
		tsptr->c_cflag = ptxp->pts_cflag;	/* Load cflags */
		v_lock(&ptxp->pt_lock, spl);

		mp->b_cont->b_wptr += sizeof(struct termios);
		iocp->ioc_count = sizeof(struct termios);
		mp->b_datap->db_type = M_IOCACK;
		qreply(wr_q, mp);
		return;

	case TIOCPTYDEV:			/* get slave/master dev_t */
	{
		struct ptydev *xptydev;

		if (!ismaster)
			goto badioctl;

		if ((bp = allocb(sizeof(struct ptydev), BPRI_MED)) == NULL)  {
			PTDEBUG("ptx: ptxioctl(%s) TIOCPTYDEV nack allocb failed\n",
					(ismaster) ? "master" : "slave");
			mp->b_datap->db_type = M_IOCNAK;
			iocp->ioc_count = 0;
			iocp->ioc_error = EAGAIN;
			iocp->ioc_rval = -1;
			qreply(wr_q, mp);
			return;
		}

		if (iocp->ioc_count == TRANSPARENT) {
			PTDEBUG("ptx: ptxioctl(%s) TIOCPTYDEV TRANSPARENT\n",
					(ismaster) ? "master" : "slave");
			send_buf_p = (struct copyreq *)mp->b_rptr;
			send_buf_p->cq_addr = (caddr_t)(*(long *)(mp->b_cont->b_rptr));
			freemsg(mp->b_cont);
			mp->b_cont = bp;
			bp->b_wptr += sizeof(struct ptydev);
			send_buf_p->cq_private = NULL;
			send_buf_p->cq_flag = 0;
			send_buf_p->cq_size = sizeof(struct ptydev);
			mp->b_datap->db_type = M_COPYOUT;
		} else {
			PTDEBUG("ptx: ptxioctl(%s) TIOCPTYDEV data\n",
					(ismaster) ? "master" : "slave");
			if (mp->b_cont)
				freemsg(mp->b_cont);
			mp->b_cont = bp;
			bp->b_wptr += sizeof(struct ptydev);
			mp->b_datap->db_type = M_IOCACK;
			iocp->ioc_count = sizeof(struct ptydev);
			iocp->ioc_error = 0;
			iocp->ioc_rval = 0;
		}

		xptydev = (struct ptydev *)mp->b_cont->b_rptr;
		xptydev->ptym_major = ptm_major;
		xptydev->ptym_minor = ptxp - ptms_tty;
		xptydev->ptys_major = pts_major;
		xptydev->ptys_minor = ptxp - ptms_tty;

		qreply(wr_q, mp);
		return;
	}
	default:					/* nak the ioctl */
badioctl:
		PTDEBUG("ptx: ptxioctl(%s): IOCNAK unsupported ioc_cmd=0x%x\n",
			(ismaster) ? "master" : "slave", iocp->ioc_cmd);
		mp->b_datap->db_type = M_IOCNAK;
		qreply(wr_q, mp);
		return;
	}
}


/*
 * Process TCSET[AP]* ioctls.
 *
 * The only value we save is cflag.  The only cflag we care about is CLOCAL.
 */

static int
ptxtcset(wr_q, mp, ismaster, cflag, doflush)
	queue_t *wr_q;	/* write queue */
	mblk_t *mp;
	int ismaster;	/* non-zero => wr_q is for master device, else slave */
	ulong cflag;	/* new cflag value */
	int doflush;	/* non-zero => flush the read side */
{
        register struct pt_ttys *ptxp;
        register queue_t *other_wrq;
	struct iocblk 	*iocp;
	spl_t spl;

	ASSERT_DEBUG(wr_q->q_ptr, "ptx: ptxtcset()");
	ASSERT_DEBUG(mp->b_rptr,  "ptx: ptxtcset()");
	ASSERT_DEBUG(!ismaster,  "ptx: ptxtcset()"); /* only valid on slave */

        ptxp = (struct pt_ttys *) wr_q->q_ptr;
	iocp = (struct iocblk *) mp->b_rptr;

	spl = p_lock(&ptxp->pt_lock, SPLSTR);
	ptxp->pts_cflag = cflag; /* save new parameters */

	/* If no flushing requested, bypass the flush operation. */

	if (!doflush) {
		v_lock(&ptxp->pt_lock, spl);
		goto ackit;
	}

	/*
	 * If there's not a connection, just FLUSHR our side.
	 *
	 * If there is a connection, FLUSHW the other side which
	 * will be turned around at the other stream head and
	 * eventually return to our side as a FLUSHR.
	 */

	other_wrq = ismaster ? ptxp->pts_wrq 
			     : ptxp->ptm_wrq;

	if (other_wrq == NULL) {

		PTDEBUG("ptx: ptxioctl(%s): NO CONNECTION - FLUSHR %s only\n",
			(ismaster) ? "master" : "slave",
			(ismaster) ? "slave"  : "master");

		v_lock(&ptxp->pt_lock, spl);

		if (!putnextctl1(RD(wr_q), M_FLUSH, FLUSHR)) {
			cmn_err(CE_WARN, "pts: TCSETAF FLUSHR flush msg failed\n");
			/*
			 *+ The driver failed an attempt to send a flush message 
			 *+ upstream after processing a TCSETAF message
			 *+ when there was no pty connection established.
			 */
		}

	} else {

		PTDEBUG("ptx: ptxioctl(%s): - send FLUSHW to %s\n",
			(ismaster) ? "master" : "slave",
			(ismaster) ? "slave"  : "master");

		if (!putnextctl1(RD(other_wrq), M_FLUSH, FLUSHW)) {
			cmn_err(CE_WARN, "pts: TCSETAF FLUSHW flush msg failed\n");
			/*
			 *+ The driver failed an attempt to send a flush message 
			 *+ upstream after processing a TCSETAF message
			 *+ when there was a pty connection established.
			 */
		}
		v_lock(&ptxp->pt_lock, spl);
	}

ackit:
	iocp->ioc_error = 0;
	mp->b_datap->db_type = M_IOCACK;
	iocp->ioc_count = 0;
	qreply(wr_q, mp);
}


/*
 *   Master STREAMS pty driver
 */



static struct module_info ptm_info = {
/*  module ID    module name   min pkt size    max pkt siz   hiwat   lowat  */
     STRID_PTM,     "ptm",         0,            INFPSZ,      300,    100
};
static int ptmopen(), ptmclose(), ptmrsrv(), ptmwsrv();
static struct qinit ptmrint = {
/*  put	    server     open       close   admin(3bnet)	minfo     mstat */
   NULL,   ptmrsrv,   ptmopen,   ptmclose,   NULL,    &ptm_info,   NULL
};
static struct qinit ptmwint = {
/*  put	    server     open       close   admin(3bnet)	minfo     mstat */
   putq,   ptmwsrv,   ptmopen,   ptmclose,   NULL,    &ptm_info,   NULL
};
struct streamtab ptminfo = {&ptmrint, &ptmwint, NULL, NULL};


extern dev_t last_pty_dev;



/*
 * Open a minor of the master pty device. Find an unused entry in the
 * ptms_tty array if CLONEOPEN, else use the specified minor. 
 */
static int 
ptmopen(rqp, dev, flag, sflag)
queue_t *rqp;
dev_t dev;
int flag;
int sflag;
{
        register struct pt_ttys *ptmp;
	register int i;
	spl_t spl;

	PTDEBUG("ptm: ptmopen(rqp:0x%x, dev:0x%x, flag:0x%x, sflag:0x%x)\n",
			rqp, dev, flag, sflag);

	if (sflag == CLONEOPEN) {

		PTDEBUG("ptm: ptmopen(): CLONEOPEN\n");
		/*
		 * The streamhead serializes opens on the same major
		 * device.  Cloneopen for this major will not be invoked
		 * again until we return.  Therefore, we have implicit
		 * mutex on the "last_pty_dev" global variable.
		 */
		dev = last_pty_dev;
		for (i = 0; i < pt_cnt; i++) {
			dev++;
			if (dev == pt_cnt)
				dev = 0;
			ptmp = &ptms_tty[dev];
			spl = p_lock(&ptmp->pt_lock, SPLSTR);
			if (!(ptmp->pt_state & PTMOPEN))
					break;
			v_lock(&ptmp->pt_lock, spl);
		}
		if (i >= pt_cnt) {
			PTDEBUG("ptm: ptmopen(): no unallocated devices\n");
			u.u_error = ENODEV;
			return(OPENFAIL);
		}
		last_pty_dev = dev;

		PTDEBUG("ptm: ptmopen(): allocated minor 0x%x\n", dev);

	} else if (sflag == DRVROPEN) {

		dev = minor(dev);
		
		PTDEBUG("ptm: ptmopen(): DRVROPEN for minor 0x%x\n", dev);

		if ((dev < 0) || (dev >= pt_cnt)) {
			PTDEBUG("ptm: ptmopen(): invalid minor 0x%x\n", dev);
			u.u_error = ENODEV;
			return(OPENFAIL);
		}
		ptmp = &ptms_tty[dev];
		spl = p_lock(&ptmp->pt_lock, SPLSTR);
		if (ptmp->pt_state & PTMOPEN) {
			PTDEBUG("ptm: ptmopen(): master already open\n");
			v_lock(&ptmp->pt_lock, spl);
			u.u_error = EIO;
			return(OPENFAIL);
		}

	} else {

		PTDEBUG("ptm: ptmopen(): UNKNOWN sflag = 0x%x\n", sflag);
		u.u_error = EINVAL;
		return(OPENFAIL);
	}

	/* Now have a selected master device */

#ifdef DEBUG
	/* Check the status of the slave */

	if (ptmp->pt_state & PTSOPEN) {
		PTDEBUG("ptm: ptmopen(): slave already open\n");
		if (ptmp->pts_wrq == NULL)
			PTDEBUG("ptm: ptmopen(): slave already connected\n");
	} else {
		PTDEBUG("ptm: ptmopen(): slave not open\n");
		if (ptmp->pts_wrq != NULL)
			PTDEBUG("ptm: ptmopen(): BAD: pts_wrq != NULL\n");
	}
#endif

	/*
	 * set up the entries in the pt_ttys structure for this
	 * device.
	 */
	ptmp->pt_state |= PTMOPEN;
	ptmp->ptm_wrq = WR(rqp);
	WR(rqp)->q_ptr = (char *) ptmp;
	rqp->q_ptr = (char *) ptmp;

	/* wakeup any slave opens waiting for "carrier" */
	vall_sema(&ptmp->pts_openwait);

	if (ptmp->pts_wrq) {
		/* Enable slave write server to send any queued messages */
		PTDEBUG("ptm: ptmopen(): qenable(slave)\n");
		qenable(ptmp->pts_wrq);
	}

	v_lock(&ptmp->pt_lock, spl);

	PTDEBUG("ptm: ptmopen(): done\n");
	return(dev);
}


/*
 * Close a master device.
 *
 * Send a hang-up message up the slave's read queue to designate the
 * master/slave pair is tearing down.
 */
static int 
ptmclose(rqp)
queue_t *rqp;
{
        register struct pt_ttys *ptmp;
	spl_t spl;

	PTDEBUG("ptm: ptmclose(rqp:0x%x)\n", rqp);
	ASSERT_DEBUG(rqp->q_ptr, "ptm: ptmclose()");
		
        ptmp = (struct pt_ttys *)rqp->q_ptr;
	spl = p_lock(&ptmp->pt_lock, SPLSTR);

        ptmp->ptm_wrq = NULL;	/* disable slave -> master sending */

	/* Hangup slave if connected and CLOCAL is not set */

	if (ptmp->pts_wrq) {
		PTDEBUG("ptm: ptmclose(): slave still connected\n");
		ASSERT_DEBUG(ptmp->pt_state & PTSOPEN, "ptm: ptmclose()");

		if (!(ptmp->pts_cflag & CLOCAL)) {
			PTDEBUG("ptm: ptmclose(): M_HANGUP slave\n");
			if (!putnextctl(RD(ptmp->pts_wrq), M_HANGUP)) {
				cmn_err(CE_WARN,
				     "pty: master close M_HANGUP putnextctl failed\n");
				/*
				 *+ An attempt to send a hangup indication to a
				 *+ pty slave device during master device close
				 *+ failed due to lack of system resources.
				 */
			}
		} else {
			PTDEBUG("ptm: ptmclose(): CLOCAL prevents slave hangup\n");
		}
	} else {
		PTDEBUG("ptm: ptmclose(): slave is already disconnected\n");
	}

	v_lock(&ptmp->pt_lock, spl);
	disable_procs(rqp);	/* ask STREAMS to disable our servers */
	spl = p_lock(&ptmp->pt_lock, SPLSTR);

	ptmp->pt_state &= ~PTMOPEN;
	rqp->q_ptr = NULL;
	WR(rqp)->q_ptr = NULL;

	v_lock(&ptmp->pt_lock, spl);

	PTDEBUG("ptm: ptmclose(): done\n");
	return;
}


/*
 * Master read server.
 */
static int 
ptmrsrv(qp)
queue_t *qp;
{
	ptxrsrv(qp, ISMASTER);
}
		


/*
 * Master write server.
 */
static int 
ptmwsrv(qp)
queue_t *qp;
{
	ptxwsrv(qp, ISMASTER);
}


/*
 * Called from configure at boot time.
 */
/*ARGSUSED*/
void
pty_init(ndevs)
	int ndevs;
{
        register struct pt_ttys *ptsp;
        register int i;
        struct dev *dp;

	for (i = 0; i < pt_cnt; i++) {
		ptsp = &ptms_tty[i];
		init_lock(&ptsp->pt_lock);			/* unlocked */
		init_sema(&ptsp->pts_openwait, 0);		/* locked */
		/* add dev entry for master*/

		dp = (struct dev *)calloc(sizeof(struct dev));
		dp->d_dev = makedev(ptm_major, i);
		dp->d_ops = &ptmdevops;
		if (devsw_add(dp) != 0) {
			cmn_err(CE_WARN,
			"pty: failure to add dev master minor %d to devsw\n",
			i);
			/*
			*+ The kernel encountered a problem when it 
			*+ attempted to add a pty
			*+ dev to the devsw data structure.  This is
			*+ due to a driver that is adding a dev to
			*+ devsw; the major/minor number for the dev
			*+ already belongs to another dev.
			*/
		}

		/* add dev entry for slave*/

		dp = (struct dev *)calloc(sizeof(struct dev));
		dp->d_dev = makedev(pts_major, i);
		dp->d_ops = &ptsdevops;
		if (devsw_add(dp) != 0) {
			cmn_err(CE_WARN,
			"pty: failure to add dev slave minor %d to devsw\n",
			i);
			/*
			*+ The kernel encountered a problem when it
			*+ attempted to add a pty
			*+ dev to the devsw data structure.  This is
			*+ due to a driver that is adding a dev to
			*+ devsw; the major/minor number for the dev
			*+ already belongs to another dev.
			*/
		}
	}
	/*
	 * Add one clone for the master driver
	 */
	if (clonedev_add(makedev(ptm_major, 0))) {
		cmn_err(CE_WARN,
			"pty: failure to add master clone dev to devsw\n");
		/*
		 *+ The kernel encountered a problem when it
		 *+ attempted to add a pty
		 *+ dev to the clone devsw data structure.  This is
		 *+ due to a driver that is adding a dev to
		 *+ devsw; the major/minor number for the dev
		 *+ already belongs to another dev.
		 */
	}
}
