/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ident	"$Header: ssm.c 1.38 90/07/30 $"

/*
 * ssm.c
 *	Systems Services Module (SSM) driver code.
 */

/* $Log:	ssm.c,v $
 */

#include "sys/types.h"
#include "sys/param.h"
#include "sys/debug.h"
#include "sys/mutex.h"
#include "sys/boot.h"
#include "sys/vm.h"			  /* for VA_SLIC */
#include "sys/slic.h"
#include "sys/slicreg.h"
#include "sys/bic.h"
#include "sys/cfg.h"
#include "sys/clkarb.h"
#include "sys/SGSproc.h"
#include "sys/hwparam.h"
#include "sys/intctl.h"
#include "sys/systm.h"
#include "sys/conf.h"
#include "sys/clock.h"
#include "sys/mc_vmparam.h"
#include "sys/cmn_err.h"
#include "sys/mc_param.h"
#include "sys/pte.h"
#include "sys/autoconf.h"
#include "sys/session.h"
#include "sys/stream.h"
#include "sys/ssm_cb.h"
#include "sys/ssm_misc.h"
#include "sys/scsi.h"
#include "sys/scsi_adapter.h"
#include "sys/ssm_scsi.h"
#include "sys/ssm_vme.h"
#include "sys/ssm.h"
#include "sys/ssm_device.h"
/* 
 * Array of SSM descriptors alloc'd 
 * and filled by conf_ssm() and a
 * description of its elements.
 */
struct ssm_desc	*SSM_desc;		/* Array of SSM descriptors. */
struct ssm_desc	*SSM_cons;		/* SSM descriptor of console board. */
int NSSM;				/* # SSM's to map at boot time. */
int ssm_max_idx;			/* Max config index # for an SSM */
uint SSM_vec;				/* Bit-vector of existing SSM's */
/* static unchar	SSM_basevec;	 * Base interrupt vector of SSM's */
static struct pte *ssm_vme_pt;		/* Address of second level page
					 * tables for VME memory */
static label_t	ssm_vme_nmi_jmp;	  /* for NMI recovery during probe */
static int ssm_probed;			  /* for error reporting */

static void ssm_pic_set();		       /* Forward Reference */
static int ssm_vme_nmi();		       /* Forward Reference */

static int ssm_boot(), ssm_return_fw(), ssm_inittodr(), ssm_resettodr(),
	   ssm_wdtinit(), ssm_wdtreset();

struct firmwaresw ssm_interface = {
	ssm_boot,	ssm_return_fw,	ssm_inittodr,	ssm_resettodr,
	ssm_wdtinit,	ssm_wdtreset
};

extern caddr_t calloc();
extern int ssm_wdt_timeout;
extern struct device_config sb8000_config[];

#ifdef DEBUG
int ssm_debug = 0;			  /* levels 0=off */
#endif /* DEBUG */

ssm_probe( device )
	struct gen_device *device;
{
	register struct ssm_device *dev = (struct ssm_device *)device;
	struct ctlr_toc *toc;
	register struct	ctlr_desc *cd;
	int i, ssmnum = dev - &ssm_devices[0];
	struct bus *bus;
	static int vme_bustype = -1;
	extern char *bus_type[];

	/*
	 * first of all, scan the bus_type array for the matching 'vme'
	 * string. From that, use the index as a value to search for
	 * in the busses array, below. If not found, then vme not configured
	 */
	if( vme_bustype == -1 ) {
		for( i = 0; bus_type[i] != (char*) NULL; i++ ) {
			if( strcmp( "vme", bus_type[i] ) == 0 ) {
				vme_bustype = i;
				break;
			}
		}
	}

        toc = &CD_LOC->c_toc[SLB_SSMBOARD];
        cd = &CD_LOC->c_ctlrs[toc->ct_start];

	/*
	 * Only 1 SSM1 board will ever be found in a system, but scan
	 * down the array looking for a slic match  on that one
	 */
        for (i = 0; i < toc->ct_count; i++, cd++) {
		if (dev->config->dev_addr == cd->cd_slic) {
			dev->cd = cd;	/* This is the matching SSM */
			if (FAILED(cd)) 
				return 0;	/* Short circuit search */
			if (cd->cd_ssm_cons)
				CPRINTF( "# Console is on ssm%d\n", ssmnum);
			return 1;	/* Found */
		}
	}

        /* for this device, looking into the table of contents for the SSM2
        ** board, then address the controler descriptor for the SSM2. Scan
        ** all the SSM2s the firmware found for a slic address match.
        ** if found, remember the cd ptr, test for console control and
        ** report we found the board
        */
        toc = &CD_LOC->c_toc[SLB_SSM2BOARD];
        cd = &CD_LOC->c_ctlrs[toc->ct_start];
        for (i = 0; i < toc->ct_count; i++, cd++) {
                if (dev->config->dev_addr == cd->cd_slic) {
                        dev->cd = cd;   /* This is the matching SSM2 */
                        if (FAILED(cd))
                                return 0;       /* Short circuit search */
                        if (cd->cd_ssm_cons)
				CPRINTF( "# Console is on ssm%d\n", ssmnum);
		
			/* 
			 * scan bus table for a match with this SSM2, if
			 * found then check if VME is present. If not, then
			 * NULL out the owner field for the bus
			 */
			if( vme_bustype != -1 ) {
			   for( bus = busses; bus->bustype != NOBUS; bus++ ) {
				if( (bus->bustype == vme_bustype) &&
				    (bus->owner == device) ) {
					if( cd->cd_ssm_biff_type !=
							CFG_SSM_BTYPE_VBIF ) {
						/* 
						 * make sure no devices on list
						 */
						ASSERT(bus->devices == 0, "ssm_probe: Probe order wrong");
						/*
						 *+ there should not be any
						 *+ devices on the devices list
						 *+ at probe time
						 */

						/*
						 * null out the owner field
						 * of the bus entry-NO VME
						 */
						bus->owner = 0;
					}
				}
			   }
			}
					
                        return 1;       /* Found */
                }
        }

	return 0;			/* Not found */
}

/*
 * ssm_init()
 *	Initialize system services module.  
 *
 *	Assume that there may be more than 
 * 	one of these in the system at a time
 *	(only one is the console though). 
 */
void 
ssm_init( driver)
	struct driver_module *driver;
{
        struct ctlr_toc *toc1, *toc2;
        int ssm_cnt;
	register struct	ctlr_desc *cd;
	register struct	ssm_desc *ssm;
	register struct	ssm_device *dev = (struct ssm_device *)driver->devices;
	int i, j;
	extern	int	light_show;
	extern	unsigned sec0eaddr;
	extern void scboot(), wpboot(), ssboot();
#ifdef	MODELD
	extern struct ssm_misc *ssm_misc_init(/* iscons, slicaddr */);
#endif	/* MODELD */

	ssm_max_idx = -1;

        toc1 = CD_LOC->c_toc + SLB_SSMBOARD;
        toc2 = CD_LOC->c_toc + SLB_SSM2BOARD;

        ssm_cnt = toc1->ct_count + toc2->ct_count;

	if( ssm_cnt == 0 )
		return;            /* if no SSMs at all, short circuit */

	/*
	 * Allocate an array of descriptors, 
 	 * one per potential SSM. 
	 */
	ssm = SSM_desc = (struct ssm_desc *)
		calloc(driver->numdev * sizeof(struct ssm_desc));

	/*
	 * Initialize each of the descriptors.
	 * Save the location of the master board
	 * for access by misc. functions.
	 */
	for (i = 0; i < driver->numdev; i++, dev++, ssm++) {
		if (!(cd = dev->cd)) 
			continue;

		/*
		 * Save table of contents information 
		 * about the SSM.
		 */
                ssm->ssm_cd = cd;               /* keep a ptr to board's desc */
                if( ssm->ssm_is_cons ) {
			/*
			 * We're in charge - set up to use 
			 * front-panel LED's and get system 
			 * ID from the SSM.
			 */
			SSM_cons = ssm;
			fp_lights = 1;		/* use front panel LEDs */
			if (light_show > 1)
				fp_lights = -1;	/* front-panel and proc LEDs */
			FP_IO_ONLINE;		/* ASSUME all drives online */
			sec0eaddr = cd->cd_ssm_sysid & 0x00FFFFFF;
		}

		/* turn off this board's VME memory accessability via PIC
		*/
		ssm_pic_set(ssm->ssm_slicaddr, 0L );
		
		if ( !dev->driver )
			continue;	  /* not configured or not alive */

		/*
		 * Initialize the SSM's interfaces, unless 
		 * the board is deconfigured or failed diags.  
		 * 
		 * Note that a reported failure can only 
		 * occur if there are multiple SSM type 
		 * boards in the system.
		 */
		if (!FAILED(cd)) {
			/* 
			 *calc the pic flush register address here for
			 * use in the ssm_vme_dma_flush routine for
			 * speed reasons
			 */
			switch( cd->cd_type ) {
				case SLB_SSMBOARD:
#ifdef SSM2_PIC_BUG
					dev->pic_flush_method =
							SSM_PIC_FLUSH_REG;
#endif  /* SSM2_PIC_BUG */
					dev->pic_flush_reg =
						PIC_REGISTER_ADDR( PA_SSMVME(i), 
								PIC_BCR_NARROW);
					break;

				case SLB_SSM2BOARD:
#ifdef SSM2_PIC_BUG
					dev->pic_flush_method =
							SSM_PIC_FLUSH_MEMORY;
#endif  /* SSM2_PIC_BUG */
					dev->pic_flush_reg =
						PIC_REGISTER_ADDR( PA_SSMVME(i), 
								PIC_BCR_WIDE);
					break;

				default:
					/*
					 * don't use this board if bad type
					 */
					continue;
			}

			/* 
			 * Allocate console and printer command blocks and 
			 * notify the SSM of their location.
			 */
			ssm->ssm_cons_cbs =  init_ssm_cons(ssm->ssm_slicaddr);
			ssm->ssm_prnt_cbs = init_ssm_prnt(ssm->ssm_slicaddr);

#ifdef	MODELD
			ssm->ssm_mc = ssm_misc_init(ssm == SSM_cons,
					            ssm->ssm_slicaddr);
#else	/* !MODELD */
			/* 
			 * Allocate the message passing structure 
			 * for misc. SSM commands and notify the 
			 * SSM of its location.
			 */
			ssm->ssm_mc = (struct ssm_misc *) 
				ssm_alloc(sizeof(struct ssm_misc), 
					  SSM_ALIGN_XFER, SSM_BAD_BOUND);
			init_lock(&ssm->ssm_mc->sm_lock);
			ssm_send_misc_addr(ssm->ssm_mc, ssm->ssm_slicaddr);
#endif	/* MODELD */

			/* 
			 * setup first mapping value for this board, even
			 * if the VME is not present (we'll determine that
			 *  in probe_setup) we still need the first map at
			 * which to start allocating maps later
			 *  Set starting values for s2v and v2s mapping -
			 *  the next values will be calc'd from these
			 */
			ssm->ssm_s2v_map = SSM_S2V_1ST_MAP;

			ssm->ssm_v2s_map[ SSM_VME_A16_MAPS ] = 0;
			ssm->ssm_v2s_map[ SSM_VME_A24_MAPS ] = 0;
			ssm->ssm_v2s_map[ SSM_VME_A32_LO_MAPS ] = 0;
			ssm->ssm_v2s_map[ SSM_VME_A32_HI_MAPS ] = 0;


			if( cd->cd_ssm_biff_type == CFG_SSM_BTYPE_VBIF ) {
    				/* IF there is a VBIF on this SSM, then...
		         	 * Initialize mapping of VME interrupts and set 
			 	 * phy address at which SSM should respond to. 
				 * Memory address is calc by
			 	 *  IOBASE + 32Mb + (i * 32Mb)    
				 *	[ i is the board index 0,1,2, etc]
		   	 	 * if the VME is not present
			 	 * these routines will return an error, 
				 * but we don't really care here. 
				 * If it is there, then set the numbers for 
				 *the board
				 * For SSM1, use old method for compatibility
				 * for SSM2, ask fw to do the programming so
				 * both kernel and fw know where its at.
		         	 */
		        	ssm_clr_vme_imap( ssm->ssm_mc, 
						ssm->ssm_slicaddr);

				/*
				 * depending on the SSM board in question
				 * set its physical address
				 */
				switch( cd->cd_type ) {
					case SLB_SSM2BOARD:
						ssm_set_vme_mem_window( 
							ssm->ssm_mc,
							ssm->ssm_slicaddr, i );
						break;

					case SLB_SSMBOARD:
						ssm_pic_set(ssm->ssm_slicaddr, 
							(long) PA_SSMVME(i) );
						break;
				
					default:
						cmn_err( CE_PANIC,
						     "ssm_init:bad board type");
						/*
						 *+ During ssm_init, the board
						 *+ being probed was not a
						 *+ SSM1 or SSM2: contact service
						 */
						break;
				}
						

#ifdef SSM2_PIC_BUG
				/* 
				 * For wide PIC flush bug, we need to do memory
				 * type of flush. If PIC is update and the bug
				 * no longer exists, remove this
				 *
				 * need to setup a s2v and a v2s map for
				 * performing a PIC flush if the board
				 * requires memory based flushing
				 */
				if(dev->pic_flush_method==SSM_PIC_FLUSH_MEMORY){
					ssm_set_pic_flush_address( dev, ssm );
				}
#endif  /* SSM2_PIC_BUG */
			}

			NSSM++;
			SSM_vec |= 1 << i;
			ssm_max_idx = i;
		}
	}

	/* 
	 * Report any SSM's that were 
 	 * deconfigured or failed diags.
	 * Save the largest valid index into
	 * SSM_desc also.
	 */
        if (NSSM < ssm_cnt) {
                CPRINTF("Not using System Services Modules:  slic");
                for (i = 0, cd = CD_LOC->c_ctlrs + toc1->ct_start;
                     i < toc1->ct_count; i++, cd++) {
                        for (j = 0, dev = (struct ssm_device *)driver->devices;
                             j <= ssm_max_idx; j++, dev++)
                                if (dev->cd == cd)
                                        break;   /* Matching dev found */
                        if (j > ssm_max_idx || !SSM_EXISTS(i))
                                CPRINTF(" %d", cd->cd_slic);
                }
                for (i = 0, cd = CD_LOC->c_ctlrs + toc2->ct_start;
                     i < toc2->ct_count; i++, cd++) {
                        for (j = 0, dev = (struct ssm_device *)driver->devices;
                             j <= ssm_max_idx; j++, dev++)
                                if (dev->cd == cd)
                                        break;   /* Matching dev found */
                        if (j > ssm_max_idx || !SSM_EXISTS(i))
                                CPRINTF(" %d", cd->cd_slic);
                }
                CPRINTF(".\n");
        }

	/* boot console/SSM serial lines, parallel driver */
	scboot(driver);
	wpboot(driver);
	ssboot(driver);

	if (NSSM) {
		/*
		 * Allocate 2nd level page table space for
		 * virtual address mappings of VME memory
		 * for all alive SSMs.
		 */
		callocrnd(NBPG);
		ssm_vme_pt = (struct pte *) calloc( NSSM * 
			NPTE_SSMVME * sizeof(struct pte));
		ASSERT(((ulong)ssm_vme_pt & NBPG - 1) == 0,
		       "SSM/VME page table not on page boundary");
		/*
		 *+ When the kernel was allocating space for SSM/VME page tables,
		 *+ the space was not allocated on a page boundary.
		 */
	}
}

/*
 * ssm_vme_probe_setup()
 * 	Probe the vme device located on the SSM.
 *
 * 	Handles wild-card SSM specification, 
 * 	finding an SSM with the device if it 
 *	exists.  
 *
 *	Returns a pointer to the SSM descriptor
 *	the device lives on or NULL if not found.
 */
ssm_vme_probe_setup(dev)
	register struct gen_device *dev;
{
	struct driver_module *driver = dev->driver;
	register int ssm_index;
	struct ssm_desc *ssm;
	int i, dma_admod;
	ushort csrmap;
	extern int (*probe_nmi)();

	ssm_index = (struct ssm_device *)dev->bus->owner - ssm_devices;
	ssm = SSM_desc + ssm_index;

#ifdef DEBUG
	if (ssm_debug)
		CPRINTF("Probing vme driver %s  at ssm %d vme %lx(%x)...",
		       driver->name, ssm_index, dev->config->dev_addr,
		       dev->config->space);
#endif /* DEBUG */
	
	if (ssm_index>ssm_max_idx || !SSM_EXISTS(ssm_index))
                /* do we have an illegal value, or maybe the board doesn't
                ** exist
                */
		return 0;

	/*
	 * We take over NMI vector temporarily here, to handle
	 * absent devices (see machine/locore.s for NMI handling).
	 */
	
	probe_nmi = ssm_vme_nmi;
	ssm_probed = ssm_index;		  /* info for ssm_vme_nmi */
	for ( i = 0, csrmap = ssm->ssm_s2v_map; i < dev->nspaces; i++ ) {
		/*
		 * Map csrs for probing.  We assume a csr block never crosses
		 * a SSM_S2V_MAP_SIZE (defined in ssm_vme.h) boundary,
		 * currently, 16384.  Check to be sure we've got a map before
		 * we go probing.  Because of the number of maps available,
		 * (currently 1984) we 2make no attempt to share maps.
		 * NOTE:  we still do the set up even if there is no probe
		 * routine.  WE still need the map set up, etc.
		 */
	
		if ( csrmap >= SSM_S2V_NMAPS) {
			cmn_err( CE_WARN,
			  "%s: Out of S to VME maps on ssm%d, probe failed.",
				"ssm_vme_probe_setup", ssm_index);
			/*
			 *+ While trying to probe a VME device, the SSM ran out
			 *+ of Sequent-to-VME maps.  With the current
			 *+ assumptions built into ssm_vme_probe_setup 
			 *+ and given the size of the VME card cage, 
			 *+ this should not happen. 
			 *+ (The sum of the number of
			 *+ register "space" regions on all VME devices would
			 *+ have to total 1984.)  This error
			 *+ represents a data corruption problem.
			 */
			continue;
		}
	
		if ( dev->config[i].dev_len == 0 )
			continue;	  /* not a real CSR */

		dev->devreg[i] = (char *)ssm_s2v_map( ssm_index,
						     (int)csrmap++,
						     dev->config[i].dev_addr,
						     dev->config[i].space ); 
#ifdef DEBUG
		if (ssm_debug)
			CPRINTF( "sequent address = %lx, map = %d\n",
			       dev->devreg[i], (int)csrmap - 1);
#endif /* DEBUG */
	}

	/*
	 * setjmp will return true if the following probe NMI's,
	 * meaning the device isn't there.
	 */
	if (setjmp(&ssm_vme_nmi_jmp)) {
		probe_nmi = NULL;
		return 0;
	}

	if ( driver->probe && (*driver->probe)(dev) ) {
		/* 
		 * Success! Finish filling in the device
		 * description, and return. 
		 */
		probe_nmi = NULL;

		/* Make temporary CSR mapping permanent */
		ssm->ssm_s2v_map = csrmap; 
		
		/*
		 * Now, allocate DMA maps, not to be confused with
		 * ssm_alloc_maps, which allocates kernel page
		 * table entries.
		 */
		if ( ! ssm_allocate_maps( dev) )
			return 0;
		else {
			CPRINTF("# %s configuration unit %d on %s is csr 0x%x",
			       driver->name,
			       (dev->config - dev->on_bus->config_array)
			       /dev->nspaces,
			       bus_type[dev->bus->bustype],
			       dev->config->dev_addr);

			/*
			 * Find dma admod (which is the last non-zero space
			 * field in the set of config records that comprise
			 * this config unit
			 */
			for ( i = 0, dma_admod = 0; i < dev->nspaces; i++ )
				if ( dev->config[i].space )
					dma_admod= dev->config[i].space;

			CPRINTF(" admod 0x%x DMA admod 0x%x\n        ",
			       dev->config->space, dma_admod);
			return 1;
		}
	}

	
	/*
	 * If at this point in the code then
	 * all valid possiblilities for SSM
	 * have been tried without locating the device.
	 */
	probe_nmi = NULL;
	return 0;
}

/*
 * ssm_vme_nmi()
 *	NMI handler while probing VME devices.
 *
 * We clear the access error from the processor and longjmp (returning
 * to ssm_vme_probe_setup, above).
 *
 * This is entered directly from HW vector table; however, since
 * entry is from drivers 'probe' procedure, ssm_vme_probe_setup() already
 * considers scratch registers as volitile (thus, no need to save/restore
 * them).
 */

static
ssm_vme_nmi()
{
#if	defined(ns32000) || defined(KXX)
#ifdef	MODELD
	int	timeout;
#endif	/* MODELD */
	unchar	regval =  rdslave(va_slic->sl_procid, SL_G_ACCERR);

	/*
	 * Writing anything to the processor's access-error register
	 * clears it.
	 */

	wrslave(va_slic->sl_procid, SL_G_ACCERR, 0xbb);
#ifdef	MODELD
	/*
	 * Need a "clearnmi" for 32K.
	 */
	timeout = ((~regval)&SLB_ATMSK) == SLB_AETIMOUT;
#endif	/* MODELD */

#else	/* SGS HW */
#ifdef	MODELD
	/*
	 * SGS and SGS2 HW.
	 */
	int	timeout = clearnmi();
#else	/* !MODELD */
	extern	struct	ctlr_desc *slic_to_config[];
	unchar	my_slic = va_slic->sl_procid;
	unchar	bic_slic
		= BIC_SLIC(my_slic, slic_to_config[my_slic]->cd_flags);
	unchar	accerr_reg = (my_slic & 0x01) ? BIC_ACCERR1 : BIC_ACCERR0;
	unchar	regval = rdSubslave(bic_slic, PROC_BIC, accerr_reg);

	/*
	 * Writing anything to the BIC's access-error register clears it.
	 */

	wrSubslave(bic_slic, PROC_BIC, accerr_reg, 0xbb);
#endif	/* MODELD */
#endif	/* KXX */

	/*
	 * Determine why we got an NMI.  Hopefully it was NOT
	 * due to a Sequent Bus Timeout.
	 */
#ifdef	MODELD
	if (timeout) {
#else	/* !MODELD */
	if (((~regval)&SLB_ATMSK) == SLB_AETIMOUT) {
#endif	/* MODELD */
		/*
		 * We COULD deconfigure the board here, and keep on running.
		 * But, we panic instead, just to play it safe, and protect
		 * naive system administrators.
		 */
		cmn_err(CE_PANIC,
			"VME TIMEOUT: ssm%d: VMEbus is NOT FUNCTIONING, SQT BUS TIMEOUT",
			ssm_probed);
		/*
		 *+ The VMEbus is not functioning correctly.  Corrective
		 *+ action:  contact service.
		 */
	}

	/*
	 * "return" to ssm_vme_probe_setup.
	 */

	longjmp(&ssm_vme_nmi_jmp);
}

/*
 * SSM_error()
 *	Error interrupt from an SSM board.
 *
 * 	SSM has detected a hard error.  
 * 	Determine the source SSM board, 
 *	report the error, and panic.
 */
/* static								*/
/* SSM_error(vector)							*/
/* 	unchar	vector;							*/
/* {									*/
/* 	register int board = (vector -= SSM_basevec) / ERRST_NERRS;	*/
/* 	register int error = vector % ERRST_NERRS;			*/
/* 	switch (error) {						*/
/* 	case ERRST_BUS_LOCK:						*/
/* 		printf("SSM%d detects system bus lock.\n", board);	*/
/*		panic("bus lock"); 						*/
/* 	case ERRST_DEVFLT:						*/
/* 		printf("SSM%d detects host processor or ", board);	*/
/*		printf("controller device fault.\n"); 			*/
/*		panic("device fault"); 						*/
/* 	case ERRST_SSMFLT:						*/
/* 		printf("SSM%d reset after fatal error detection.\n", board); */
/*		panic("fatal error detected"); 						*/
/* 	case ERRST_VME_BERR:						*/
/* 		printf("SSM%d detects VME bus error.\n", board);	*/
/*		panic("VME bus error"); 						*/
/* 	}								*/
/* }									*/

/*
 * ssm_vme_alloc_maps()
 *	Map VME memory being used into the
 *	kernel's virtual address space and
 * 	notify the SSM that all its VME
 *	devices have been initialized.
 *
 *	probe_ssm_devs() determined the number of 
 *	windows to map and allocated enough level
 *	two page table space for them.
 */
/*ARGSUSED*/
void
ssm_vme_alloc_maps( bus ) 
	struct bus *bus;
{
	extern struct pte *kl1pt_master;
	struct pte *kl1pt = kl1pt_master;
	register struct pte *pte;
	register ulong paddr, endpaddr;
	int i, k, l1idx;
	static already_done = 0;

	if ( already_done )
		return;
	else 	already_done = 1;

	/* We do it once for all ssm's */

	if (!ssm_vme_pt) 
		return;			/* No SSMs alive, nothing to do */

	for (pte = ssm_vme_pt, i = 0; i <= ssm_max_idx; i++) {
		if ( !SSM_EXISTS(i) )
			continue;

		if ( !SSM_VME_EXISTS(i) )
			/*
			 * this ssm exists, but there is no VME on it
			 */
			continue;

		/* 
		 * Map the SSM's VME memory into the kernel's virtual 
		 * address space.  Each level1 page table entry 
		 * maps NPTEPG of level2 page table entries.
		 */
		for (l1idx = L1IDX(VA_SSMVME(i)+KSOFF), paddr = PA_SSMVME(i),
		     endpaddr = paddr + SSMVME_ADDR_SPACE; 
		     paddr < endpaddr; l1idx++) {
			/* Fill in a level1 page table entry */
			*(long *)(kl1pt + l1idx) = (long)pte 
				| PG_V | PG_R | PG_M | PG_KW;

			/* Fill in the corresponding level 2 entries */ 
			for (k = NPTEPG; k--; paddr += NBPG) 
				*(long *)(pte++) = PHYSTOPTE(paddr)
					| PG_V | PG_R | PG_M | PG_KW;
		}

		/*
		 * Notify the SSM that its VME device
		 * initialization is complete.  And that
		 * it may enable VME bus monitoring.
		 */
		ssm_vme_imap_ready(SSM_desc[i].ssm_mc,SSM_desc[i].ssm_slicaddr);
	}
}


/*
 * ssm_pic_set( slic, window )
 * sets the PIC of the ssm whose slic is given to respond to the 32 megabyte
 * window which starts at (window & 0xff000000).  If the most significant bit
 * of window is 0, the PIC is turned off.
 */
static void
ssm_pic_set( slic, window )
	unchar slic;
	long window;
{
	unchar pic_addr = (window & SSM_VME_PIC_MASK)>>PIC_IO_ADDR_SHIFT;

	wrSubslave( slic, SSM_IO_PIC_ADDR, PIC_IO_ADDR_REG, pic_addr );
	
}


/*
 * ssm_boot()
 *	SSM dependent reboot the machine.
 * 	Sets up a return of control to the SSM firmware.  
 * 	If called by panic it returns specifying that the 
 *	alternate boot name is to be booted (normally the 
 *	Memory dumper).  Prior to returning to the 
 *	firmware it attempts to sync the disks.
 * 	Only ONE engine is alive at this point.
 */

static ushort reboot_flags = RB_HALT;

static
ssm_boot(paniced, howto)
	int paniced;
	register int howto;
{
	spl_t s_ipl;
	extern int wd_tid;
	extern bool_t dblpanic;
	extern reboot_sync();
	extern short upyet;

	if ((!upyet) || ((paniced == RB_PANIC) && dblpanic))
		return_fw();

	/*
	 * Now tell FW how to reboot
	 */
	if (paniced == RB_PANIC) {
		/*
		 * Setup reboot to perform a system dump
		 * and then turn on the error light.
		 */
		reboot_flags = RB_AUXBOOT;
		s_ipl = splhi();
		ssm_set_fpst(ssm_get_fpst(SM_NOLOCK) | FPST_ERR, SM_NOLOCK);
		splx(s_ipl);
	} else {
		/*
		 * Setup reboot to use 'howto' flags passed in.
		 * Then disable the watchdog timer to prevent
  		 * the error light from changing state.
		 */ 
		ASSERT(paniced == RB_BOOT, "ssm_boot: invalid boot specifier");
		/*
		 *+ The ssm_boot routine was called to halt the system.
		 *+ Neither the RB_PANIC or the RB_BOOT bit was set in the
		 *+ parameter that tells ssm_boot what action to take.
		 */

		reboot_flags = howto;

		if (wd_tid) {
			untimeout(wd_tid);		/* Stop wdt reset */
			wd_tid = 0;
		}
		s_ipl = splhi();
		ssm_init_wdt(60 * 1000);		/* Disable wdt on SSM */
		splx(s_ipl);

	}

	reboot_sync(howto);		/* Sync the disks */
	return_fw();
}

/*
 * ssm_return_fw()
 *	Return control to the SSM firmware.
 *	Prior to doing so, invoke custom
 *	panic handlers (if panicing) and
 *	turn off the light show.
 */
static
ssm_return_fw()
{
	extern shutdown_light_show(); 

	shutdown_light_show();		/* Turn off proc lights, etc. */

	ssm_reboot(reboot_flags, 0, (char *)NULL);
	for (;;); 			/* SSM will take control.  */
	/*NOTREACHED*/
}

static
ssm_wdtinit()
{
	spl_t s_ipl = splhi();

	ssm_init_wdt((ulong)(ssm_wdt_timeout * 1000));	/* ssm_wdt_timeout milliseconds */
	splx(s_ipl);
	wd_tid = timeout(ssm_wdtreset, (caddr_t)0, hz/2);
}

static
ssm_wdtreset()
{
	ssm_poke_wdt();
	wd_tid = timeout(ssm_wdtreset, (caddr_t)0, hz/2);
}


/*
 * Routines to manipulate the SSM based time-of-day register.
 * TOD clock interrupt handling done by todclock in kern_clock.c
 *
 * Inittodr initializes the time-of-day hardware which provides
 * date functions. This starts the time-of-day clock.
 *
 * Base provides the time from the filesystem, and the time-of-day
 * clock provides the rest.
 */
static
ssm_inittodr(base)
	time_t base;
{
	extern bool_t check_time();
	uint todr;

	/* 
	 * Verify that TOD clock passed diagnostics.
	 */
	ASSERT(SSM_cons, "inittodr: can't find console SSM");
	/*
	 *+ ssm_inittodr was called to initialize the time of day hardware,
	 *+ but an SSM has not been selected as the controlling console.
	 */

	if (SSM_cons->ssm_diag_flags & CFG_SSM_TOD) {
		/*
		 * Clear todr if TOD failed powerup 
		 * diagnostics.  Will use file system
		 * time.
		 */
		cmn_err(CE_WARN, 
			"WARNING: TOD on SSM failed powerup diagnostics");
		/*
		 *+ The time of day clock on the SSM controlling the 
		 *+ system failed its power-up diagnostics.  
		 *+ The system date should be checked/reset after
		 *+ the system is booted and the failure is repaired.
		 */

		todr = 0;
	} else {
		/*
		 * Fetch the current time-of-day from 
		 * the SSM tod clock.
		 */
		todr = ssm_get_tod(SM_NOLOCK);
	}
	if(check_time((time_t) todr, base))
		ssm_resettodr();

	/* 
	 * Start the tod clock.
	 */
	ssm_tod_freq(SM_NOLOCK, TODFREQ, SL_GROUP | TMPOS_GROUP,
		TODCLKVEC, SL_MINTR | TODCLKBIN);
}

/*
 * ssm_resettodr restores the time-of-day hardware after a time change.
 *
 * Reset the TOD based on the time value; used when the TOD
 * has a preposterous value and also when the time is reset
 * by the stime system call.
 */
static
ssm_resettodr()
{
	ssm_set_tod((ulong)time.tv_sec);
}
