/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ident	"$Header: ssm_vme.c 1.23 90/07/30 $"

/*
 * ssm_vme.c
 *	SSM vme support routines.
 */

/* $Log:	ssm_vme.c,v $
 */

#include "sys/types.h"
#include "sys/param.h"
#include "sys/hwparam.h"
#include "sys/cfg.h"
#include "sys/ssm_misc.h"
#include "sys/ssm_vme.h"
#include "sys/cmn_err.h"
#include "sys/slic.h"
#include "sys/systm.h"
#include "sys/vmmac.h"
#include "sys/pte.h"
#include "sys/autoconf.h"
#include "sys/debug.h"

#include "sys/ssm.h"
#include "sys/ssm_device.h"

extern short upyet;
#ifdef notyet
/*
 * ssm_vme_map_setup takes a sequent bus address and a length in bytes for a
 * dma transfer as well as some bus mapping information, and returns as a long
 * a vme bus address suitable for the dma transfer, or -1.  Note that because
 * the SSM has a 1Mbyte window from VME to Sequent bus, a -1 address is only
 * valid in A32 when the window is in the highest possible location.  We
 * disallow this for software ease.  The window is determined by the addresses
 * to which the SSM's I/O PIC is programmed to respond, and the comparator
 * value previously programmed.  The comparator value is read out of the SSM,
 * and the variable ssm has the base address of the SSM (or any address to
 * which the SSM's I/O PIC will respond.  From that, we can calculate the start
 * of the 1Mbyte window.
 *
 * Space is the 6 bit code that the VMEbus uses to designate an address space.
 * First_map is an index into the ssm's mapping registers of the first map that
 * can be used for this dma setup.  N_maps is the number of maps that can be
 * used.  Both first_map and n_maps are updated to reflect what was used, so
 * that successive calls to ssm_vme_map_setup may use the same variables until
 * the maps run out.    Also, length is updated--in case there are not enough
 * maps to map the entire transfer, at least you can still transfer SOMETHING!
 * Note that the index is per address space--there are 32 maps for A16, 512 for
 * A24, and 1024 for A32. However, the A32 space is divided into two 512 index
 * chunks.  These chunks are NOT contiguous unless the 1 Mbyte window has been
 * placed at VME address 7ff00000. Unless it becomes a problem, it will be
 * treated like two 512 index chunks.
 *
 */
long
ssm_vme_map_setup( addr, p_length, space, p_first_map, p_n_maps, ssm)
	caddr_t addr, ssm;
	long *p_length;
	int space, *p_first_map, *p_n_maps;
{
	int map_set = 0, map_limit = 0, maps_needed = 0, index, i;
	long  addr_mask, window_base, map_base;
	unsigned long msb = 0;
    
	ssm = (caddr_t )((long)ssm & SSM_VME_PIC_MASK);

	/* window base is read out of the comparator */
	window_base = *(unsigned long *)((long)ssm) & SSM_VME_COMP_MASK;
    
	if ( *p_first_map < 0  || *p_n_maps <= 0  || *p_length <= 0) return -1;
    
	switch( space)
	{
	case 0x9:
	case 0xa:
	case 0xb:
	case 0xd:
	case 0xe:
	case 0xf:
		/* extended address (A32) */
		if ( *p_first_map < SSM_VME_NMAPS ) 
		{
			map_set = 2;
			map_limit = SSM_VME_NMAPS;
		} else
		{
			map_set = 3;
			map_limit = 2 * SSM_VME_NMAPS;
			msb = 1<<31;
		}
		addr_mask = 0xffffffff;
		break;
	case 0x29:
	case 0x2d:
		/* short addresses (A16) */
		map_set = 0;
		map_limit = SSM_VME_A16_MAPPINGS;
		addr_mask = 0x0000ffff;
		break;
	case 0x39:
	case 0x3a:
	case 0x3b:
	case 0x3d:
	case 0x3e:
	case 0x3f:
		/* standard addresses (A24) */
		map_set = 1;
		map_limit = SSM_VME_NMAPS;
		addr_mask = 0x00ffffff;
		break;
	default:
		/* reserved */
		return -1;
		/*NOTREACHED*/
		/* break; */
	}

	if ( *p_first_map >= map_limit ) return -1;
	if ( *p_first_map + *p_n_maps > map_limit )
		*p_n_maps = map_limit - *p_first_map;
	
	maps_needed = ((long)addr + *p_length -((long)addr
						& SSM_VME_MAP_ADDR_MASK)
		       + SSM_VME_BYTES_PER_MAP - 1)/SSM_VME_BYTES_PER_MAP;
	if ( maps_needed > *p_n_maps ) 
	{
		*p_length -= (maps_needed - *p_n_maps) * SSM_VME_BYTES_PER_MAP;
		maps_needed = *p_n_maps;
	}
    
	*p_n_maps -= maps_needed;
	map_base = *p_first_map << 11;	  /* Get this before its incremented */
	for ( index = 0; index < maps_needed; index++)
	{
		/*
		 * A16 needs special treatment.  For each mapping, there are 16
		 * registers that need to have the same value in them.  The
		 * first will be done below, the rest are done here. 
		 */
		if ( map_set == 0 )
			for( i = 1; i < 16; i++)
				*(long *)((long)ssm | 1<<16 | map_set<<14
					  | i<<7  | *p_first_map<<2)
					= ((long)addr & SSM_VME_MAP_ADDR_MASK)
						+ (index<<11)
					   | 0xf<<7 | SSM_MAP_HIT;

		*(long *)((long)ssm | 1<<16 | map_set<<14
			  | (*p_first_map)++<<2)
			= ((long)addr & SSM_VME_MAP_ADDR_MASK) + (index<<11)
				| 0xf<<7 | SSM_MAP_HIT;
	}
	return ( msb | window_base | map_base
		| (long)addr & SSM_VME_MAP_SIZE-1) & addr_mask;
}
#endif /* notyet */


#ifdef SSM2_PIC_BUG
/*
 * ssm_set_pic_flush_address( dev, ssm )
 * Given a ptr to the ssm_device and a ptr to the ssm structure
 * setup for the memory method of flushing the PIC on the SSM2
 *  use the s2v and v2s current map values, allocate a map from
 *  each and initialize the maps for loop-back accessing.
 * 
 *  In this manner, reading/writting the sequent address will cause
 *  the VME to respond and then read/write sequent memory. In so 
 *  doing the PIC will flush its prior values and respond to the current
 *  requests - thus the PIC will be flushed
 */
ssm_set_pic_flush_address( dev, ssm )
struct ssm_device *dev;
struct ssm_desc *ssm;
{
	int ssm_index;
	long sqnt_mem;
	int v2s_map, i, s2v_map;
	ulong phy_ssm;

	ssm_index = ssm - SSM_desc;
	phy_ssm = PA_SSMVME( ssm_index );

	s2v_map = ssm->ssm_s2v_map++;
	
	dev->pic_flush_memory = 
		(long*) ssm_s2v_map(	ssm_index, 
					s2v_map,
					0x0,	/* VME address */
					0x29);	/* A16 space */
		
	callocrnd(2048);		/* needs to be on 2k boundary */
	sqnt_mem = (long) calloc( 64 );	/* just 64 bytes : phy == virt still */

	/*
 	 * now get next v2s A16 maps, need 16 to cover the floating lines
	 */
	v2s_map = ssm->ssm_v2s_map[ SSM_VME_A16_MAPS ];
	ssm->ssm_v2s_map[ SSM_VME_A16_MAPS ] += 16;	

	for( i=0; i < 16; i++ ) {
		*(ulong *)(phy_ssm | 1<<16 | i<<7 | v2s_map<<2 )
			= (sqnt_mem | 0xF<<7 | SSM_MAP_HIT);
	}
}
#endif  /* SSM2_PIC_BUG */




/*
 * ssm_allocate_maps()
 * Given a pointer to a gen_device, allocate the dma maps it needs from its
 * ssm.  Returns 1 on success, 0 on failure. Since failures are serious (but
 * not fatal) conditions, this routine  prints out an appropriate error
 * message. 
 */
long
ssm_allocate_maps( dev )
	struct gen_device *dev;
{
	int mapset;
	int nmaps;
	struct ssm_vme_map_origin *mo;
	int ssm_index = (struct ssm_device *)dev->bus->owner - ssm_devices;
	struct ssm_desc *ssm = SSM_desc + ssm_index;
	int space, i;
	
	nmaps = dev->nsegs
		* ((dev->segsize + SSM_VME_MAP_SIZE-1)/SSM_VME_MAP_SIZE
		   + (dev->segalign%SSM_VME_MAP_SIZE ? 1 : 0));
	/* The space (A16, A24 or A32) is decided based on the last non-zero
	 * space in the config.  Therefore, you can always configure that
	 * separately by adding extra config records to the config.  There
	 should always be at least one good space.
	 */
	for ( i = 0, space = 0; i < dev->nspaces; i++ )
		if ( dev->config[i].space )
			space = dev->config[i].space;

	dev->map_origin = calloc( sizeof (struct ssm_vme_map_origin)
				 +  (nmaps - 1) * sizeof (ulong) );
	if ( ! dev->map_origin ) {
		cmn_err( CE_PANIC,
			"%s: no memory for map struct for %s%d",
			"ssm_allocate_maps", dev->driver->name,
			DEVICEN(dev) );
		/*
		 *+ The system ran out of memory.  
		 *+ ssm_allocate_maps was unable to allocate
		 *+ the data structure necessary to assign SSM VME DMA
		 *+ maps to the named device.
		 */
		return 0;
	}

	mo = (struct ssm_vme_map_origin *)dev->map_origin;

	switch(VME_SPACE(space)) {
		/* Look at the addr modifier to determine which maps to use */
	case VME_A32_SPACE:
		/* A32 space is divided into two sets of
		 * of registers - A32 HI and A32 LO - based
		 * on the most significant bit.  Try allocating 
		 * from both spaces, MSB=1 (HI) first */
		mapset = SSM_VME_A32_HI_MAPS;
		if ( nmaps + ssm->ssm_v2s_map[mapset]  > SSM_VME_NMAPS )
			mapset = SSM_VME_A32_LO_MAPS;
		break;
	case VME_USER_SPACE:
		/* User defined.  We don't define any, so can't get any maps */
		if ( nmaps > 0 ) {
			CPRINTF ( "ssm%d: Bad VME address modifier = %x.\n",
				ssm_index, space);
			return 0;
		}
		break;
	case VME_A16_SPACE:
		/*
		 * A16.  Need 16 maps per mapping to account
		 * for floating address lines. Note that everywhere 
		 * else we are counting in terms of mappings, so we 
		 * only use nmaps in the limit testing.  We use 
		 * dev->sdv_maps_req for the number of mappings 
		 * everywhere else.
		 */
		nmaps *= 16;
		mapset = SSM_VME_A16_MAPS;
		break;
	case VME_A24_SPACE:
		mapset = SSM_VME_A24_MAPS;
		break;
	}

	if ( nmaps + ssm->ssm_v2s_map[mapset]  > SSM_VME_NMAPS ) {
		cmn_err(CE_WARN, "ssm%d: out of A%d DMA maps.\n", ssm_index,
		       16 + 8*mapset );
		/*
		 *+ An attempt has been made to
		 *+ allocate more maps of the specified type than currently exist.
		 *+ Corrective action:  check the map allocations for each
		 *+ of the VME devices.
		 */
		return 0;
	} else {
		dev->n_maps = nmaps;
		dev->map_extent = SSM_VME_MAP_SIZE;
		mo->basemap =  ssm->ssm_v2s_map[mapset];
		mo->nmaps = nmaps;
		mo->mapset = mapset;
		ssm->ssm_v2s_map[mapset] += nmaps;
		return 1;
	}
}

/*
 * ssm_s2v_map
 * Given the I/O base of an SSM, the number of the map to use, the vme address
 * and the vme address modifier, sets up the map and returns the sequent
 * address to use to access the vme address.  Assumes ssm_d, map and address
 * are valid, masks the admod to six bits.
 */
long
ssm_s2v_map( ssm_index, map, address, admod)
	int ssm_index, map;
	long address;	/* vme address */
	int admod;	/* address modifier */
{
	long sqnt_addr;
	long ssm;

	ssm = PA_SSMVME( ssm_index );
	admod &= 0x3f;

	/*
	 * The expression for the map uses the I/O address of the ssm, 2<<16
	 * which addresses the sequent to vme bus map rams and the number of
	 * the map, shifted over for each map takes a long word.
	 *
	 * The value programmed is the upper portion of the VME address, the
	 * VME address modifier (6 bit) shifted over, 0xf<<4 which need to be
	 * ones in order to parity protect the maps and the hit bit (1), which
	 * says this map is valid.
	 */

	*(long *)(ssm | 2<<16 | map<<2)
		= address & SSM_S2V_MAP_MASK    | admod<<8 | 0xf<<4 | 1;
	
	/*
	 * Calculate the sequent address to access the vme address just mapped.
	 * Note that the address bits from "address" that aren't in the map are
	 * in this address.  The sequent address is the SSM's base I/O address
	 * with the map number shifted over and the leftover bits from the VME
	 * address.
	 */
	sqnt_addr = ssm | map<<14 |    address & ~SSM_S2V_MAP_MASK;

	return sqnt_addr;
}


/*
 * ssm_vme_dma_setup()
 * 	For the given vme dev, using the maps assigned 
 * 	to that dev, create a DMA mapping with the given 
 * 	id to be used for doing a DMA transfer of the
 * 	virtually contiguous buffer described by the 
 * 	length and virtual address arguments.  
 *
 * 	The caller must ensure that only one invocation of 
 * 	this routine is active at any one time for the
 * 	associated device, since it accesses the device's
 *	dma map pool and returns the VME start address in 
 *	dev->bus_map_addr.  If a DMA request already exists 
 * 	with the given id, the new "mapping" is added to it.  
 *	If the entire request cannot be mapped, as much of 
 *	it as can be mapped is mapped.  The actual length 
 *	mapped (which is less than or equal the requested 
 *	length) is returned.  
 * 
 * 	The ID must be non-zero, and length positive or a
 * 	length of -1 is returned. 
 */
ssm_vme_dma_setup(dev, length, address, id)
	struct gen_device *dev;
	caddr_t address;
	long id, length;
{
	long map_start, map_offset, nmaps, addr_mask, msb = 0, i, j;
	ulong *first, *last, *longest;
	long longest_size = 0;
	ulong ssm = PA_SSMVME((struct ssm_device *)dev->bus->owner - 
			ssm_devices)&SSM_VME_PIC_MASK;
	ulong window_base = *(ulong *)ssm & SSM_VME_COMP_MASK;
	struct ssm_vme_map_origin *mo
		= (struct ssm_vme_map_origin *)dev->map_origin;
	ulong *id0 = &mo->id[0];
	ulong *idn = &mo->id[mo->nmaps];
	char *map_addr;
	paddr_t paddress;

	if ( !id || length <= 0 )
		return -1;

#define KVTOP(addr) \
	(PTETOPHYS(Sysmap[btop(addr)]) + ((int)(addr) & (NBPG-1)))

	paddress = (upyet) ? (paddr_t)KVTOP(address) : (paddr_t)address;
	map_start = paddress & SSM_VME_MAP_ADDR_MASK;
	map_offset = paddress & ~SSM_VME_MAP_ADDR_MASK;
	nmaps = (length + map_offset + SSM_VME_BYTES_PER_MAP-1)
		/ SSM_VME_BYTES_PER_MAP;

	/* allocate first fit, remembering longest just in case */
	for ( first = id0; first < idn - (nmaps - 1); first++ )
		if ( !*first ) {
			for ( last = first + 1; last < first + nmaps; last++ )
				if( *last ) {
					if ( last - first > longest_size ) {
						longest = first;
						longest_size = last - first;
					}
					break;
				}
			if ( last == first + nmaps ) 
				break;
		}
	
	if ( last != first + nmaps ) {
		if ( !nmaps )
			return -1;
		nmaps = longest_size;
		first = longest;
		last = first + nmaps;
		length = nmaps*SSM_VME_BYTES_PER_MAP - map_offset;
	}
	
	switch( mo->mapset ) {
	case SSM_VME_A16_MAPS:		
		addr_mask = 0x0000ffff;
		break;
	case SSM_VME_A24_MAPS:	
		addr_mask = 0x00ffffff;
		break;
	case SSM_VME_A32_HI_MAPS:
		msb = 1<<31;
	case SSM_VME_A32_LO_MAPS:
		addr_mask = 0xffffffff;
		break;
        }

	i = first - id0;
	map_addr = (char *)
		((msb | window_base | (mo->basemap+i)<<11 | map_offset)
			& addr_mask);
	for ( ; i < last - id0; i++ ) {
		if ( mo->mapset == SSM_VME_A16_MAPS )
			/*
			 * A16 needs special treatment.  For
			 * each mapping, there are 16 registers
			 * that need to have the same value in
			 * them except for floating addres lines.  
			 * The first will be done below,
			 * the rest are done here.  
			 */
			for (j = 1; j < 16; j++) {
				*(ulong *)(ssm | 1<<16 | j<<7
					   | (mo->basemap+i)<<2)
					= (map_start| 0xf<<7 | SSM_MAP_HIT);
			}
		*(ulong *)(ssm | 1<<16 | mo->mapset<<14 | (mo->basemap+i)<<2)
			= (map_start | 0xf<<7 | SSM_MAP_HIT);
		*first++ = id;
		address += SSM_VME_MAP_SIZE; 
		paddress = (upyet) ? (paddr_t)KVTOP(address) : (paddr_t)address;
		map_start = paddress & SSM_VME_MAP_ADDR_MASK;
	}
	dev->bus_map_addr = map_addr;
	return (length);
}


/*
 * ssm_vme_dma_remap()
 * 	For the given vme dev, using the maps assigned 
 * 	to that dev, create a DMA mapping with the given 
 * 	id to be used for doing a DMA transfer of the
 * 	virtually contiguous buffer described by the 
 * 	length and virtual address arguments.  
 * 	
 *	Works like ssm_vme_dma_setup() but does not require
 *	mutex of device, since it will attempt to reuse the
 *	first cluster of maps from the mapping pool with the
 *	given id.  Also, the return value is different.  
 * 
 * 	The ID must be non-zero, the length positive, and the
 *	dma maps to reuse must be capable of mapping the buffer,
 *	in which case the return value is the VME address for
 *	starting the transfer.  Otherwise, the return value is
 *	NULL.
 */
char *
ssm_vme_dma_remap(dev, length, address, id)
	struct gen_device *dev;
	caddr_t address;
	long id, length;
{
	long map_start, map_offset, nmaps, addr_mask, msb = 0, i, j;
	ulong *first, *last;
	ulong ssm = PA_SSMVME((struct ssm_device *)dev->bus->owner - 
			ssm_devices)&SSM_VME_PIC_MASK;
	ulong window_base = *(ulong *)ssm & SSM_VME_COMP_MASK;
	struct ssm_vme_map_origin *mo
		= (struct ssm_vme_map_origin *)dev->map_origin;
	ulong *id0 = &mo->id[0];
	ulong *idn = &mo->id[mo->nmaps];
	char *map_addr;
	paddr_t paddress;

	if ( !id || length <= 0 )
		return ((char *)NULL); /* Invalid arguments */

#define KVTOP(addr) \
	(PTETOPHYS(Sysmap[btop(addr)]) + ((int)(addr) & (NBPG-1)))

	paddress = (upyet) ? (paddr_t)KVTOP(address) : (paddr_t)address;
	map_start = paddress & SSM_VME_MAP_ADDR_MASK;
	map_offset = paddress & ~SSM_VME_MAP_ADDR_MASK;
	nmaps = (length + map_offset + SSM_VME_BYTES_PER_MAP-1)
		/ SSM_VME_BYTES_PER_MAP;

	/*
	 * For eficiency, attempt to reuse only first block
	 * of sequential maps associated with this i.d.  
	 * If this fails, the caller can still invoke the
	 * dma_done() and dma_setup() interfaces.
	 */
	for ( first = id0; first < idn - (nmaps - 1) && *first != id; first++ )
			continue;	/* Looking for first map */

	for ( last = first; last < first + nmaps; last++ )
		if( *last != id ) 
			return ((char *)NULL);	/* Not enough maps */

	/* 
	 * Maps located, fill them in.
	 */
	switch( mo->mapset ) {
	case SSM_VME_A16_MAPS:		
		addr_mask = 0x0000ffff;
		break;
	case SSM_VME_A24_MAPS:	
		addr_mask = 0x00ffffff;
		break;
	case SSM_VME_A32_HI_MAPS:
		msb = 1<<31;
	case SSM_VME_A32_LO_MAPS:
		addr_mask = 0xffffffff;
		break;
        }

	i = first - id0;
	map_addr = (char *)
		((msb | window_base | (mo->basemap+i)<<11 | map_offset)
			& addr_mask);
	for ( ; i < last - id0; i++ ) {
		if ( mo->mapset == SSM_VME_A16_MAPS )
			/*
			 * A16 needs special treatment.  For
			 * each mapping, there are 16 registers
			 * that need to have the same value in
			 * them except for floating addres lines.  
			 * The first will be done below,
			 * the rest are done here.  
			 */
			for (j = 1; j < 16; j++) {
				*(ulong *)(ssm | 1<<16 | j<<7
					   | (mo->basemap+i)<<2)
					= (map_start| 0xf<<7 | SSM_MAP_HIT);
			}
		*(ulong *)(ssm | 1<<16 | mo->mapset<<14 | (mo->basemap+i)<<2)
			= (map_start | 0xf<<7 | SSM_MAP_HIT);
		first++;
		address += SSM_VME_MAP_SIZE; 
		paddress = (upyet) ? (paddr_t)KVTOP(address) : (paddr_t)address;
		map_start = paddress & SSM_VME_MAP_ADDR_MASK;
	}
	return (map_addr);
}

void ssm_vme_dma_done(dev, id)
	struct gen_device *dev;
	ulong id;
{
	struct ssm_vme_map_origin *mo
		= (struct ssm_vme_map_origin *)dev->map_origin;
	ulong *map;
	
	for ( map = &mo->id[0]; map < &mo->id[mo->nmaps]; map++ )
		if ( *map == id )
			*map = 0;
}

/* ssm_vme_dma_flush()
 *
 *  called by a device driver to flush/finishup the last dma (data
 *  transfer - this is a time critical function. It occurs alot
 *  thus, the PIC flush register address is calculated at init time
 *  and stored in the ssm_device structure for fast reference
 *
 */
void ssm_vme_dma_flush(dev)
	struct gen_device *dev;
{
	register struct ssm_device *ssmdev;
	register long *flush_reg;
	static time_t last_flush_done;	/* last time PIC flushed */
	time_t last_flush_cmd;		/* time 'we' flushed PIC */

	ssmdev = (struct ssm_device *) dev->bus->owner;

#ifdef SSM2_PIC_BUG
	switch( ssmdev->pic_flush_method ) {
		case SSM_PIC_FLUSH_MEMORY:
			/*
			 * read from the mapped memory, this will cause the
			 * pic to reload its address, then read 32-bytes further
			 * and this again reload the PIC address and flush out
		    	 * any previous data, and thus flush the pic
			 */
			*(ssmdev->pic_flush_memory+8) = 0;
			*ssmdev->pic_flush_memory = 0;
			break;

		case SSM_PIC_FLUSH_REG:

#endif /* SSM2_PIC_BUG */
			flush_reg = ssmdev->pic_flush_reg;

			ASSERT_DEBUG(flush_reg!=0, 
			    "ssm_vme_dma_flush: pic_flush_reg addr is zero\n");
			/*
			 *+ flushing pic register and the address of that
			 *+ register is zero when it should not be
			 */

		
			/* tell the SSMx PIC to flush its buffer(s) onto the 
			 * system bus
	 		 */
			*flush_reg = PIC_BCR_FLUSH;

			/*
			 * use the system time of secs to time us out,if the PIC
			 * fails to FLUSH, or we don't see the PIC flush bit
			 * reset due to 'constant' flushing by other procs
			 * remember time 'we' flush it & when we see it FLUSHED
			 * remember globally when it was flushed. In this way
			 * if 2 secs pass, and it hasn't been flushed since, we
			 * can panic out
			 */
			last_flush_cmd = time.tv_sec;
			while( time.tv_sec < (last_flush_cmd+2) ) {
				if ( !(*flush_reg & PIC_BCR_FLUSH) ) {
					/*
					 * flush is done, remember current time
					 * and leave
					 */
					last_flush_done = time.tv_sec;
					return;		  /* flushed! */
				}
			}

			/*
			 * if last_flush_done is less than the cmd we issued
			 * then hw problem 
			 * else
			 *  other cpus have been here and we can leave safely
			 */
			if( last_flush_done < last_flush_cmd ) {
				cmn_err( CE_PANIC, "SSM VME PIC unflushable");
				/*
				 *+ SSM PIC didn't flush after a long time and
				 *+ it should have
				 */
			}

#ifdef  SSM2_PIC_BUG
			break;

		default:
			cmn_err( CE_PANIC, "ssm_vme_dma_flush: bad pic method");
			/*
			 *+ Flushing the SSM PIC, it was found that METHOD of
			 *+ flushing was not a MEMORY nor a REGISTER type of
			 *+ of flush
			 */
			break;
	}
#endif /* SSM2_PIC_BUG */
}

ssm_vme_intr_setup(bus, bus_level, bus_vector, slic_bin, slic_vector,
		   slic_dest)
	struct bus *bus;
	int bus_level, bus_vector;
	unchar slic_bin, slic_vector, slic_dest;
{
	struct ssm_desc *ssm = 
		&SSM_desc[ (struct ssm_device *)bus->owner - ssm_devices ];
	struct ssm_vme_int *intr;
	extern void ssm_set_vme_imap();

	/* Validate bus level */
	if ( bus_level < 1 || bus_level > 7 )
		return BUS_INTLVL_BAD;

	/*
	 * Validate the bus vector.  Note that the SSM only checks the last 8
	 * bits.  Note also that the SSM *ignores* the interrupt level when
	 * translating the vector onto the SLIC bus--that is, you can't use the
	 * same vector on two different levels.
	 */ 
	for ( intr = ssm->ssm_vints; intr; intr = intr->next)
		if ( (bus_vector & 0xff) == (intr->vector & 0xff) )
			return BUS_INTVEC_BAD;

	/* Add this one to the list */
	intr = (struct ssm_vme_int *) calloc( sizeof (struct ssm_vme_int) );
	if ( !intr) {
		cmn_err( CE_WARN, "ssm_vme_intr_setup: no memory for map");
		/*
		 *+ There was no kernel memory left when a VMEbus device tried
		 *+ to set up its interrupts.  The request fails.
		 */
		return BUS_INTMAP_BAD;
	}
	intr->level = bus_level;
	intr->vector = bus_vector;
	intr->next = ssm->ssm_vints;
	ssm->ssm_vints = intr;

	ssm_set_vme_imap(ssm->ssm_mc, ssm->ssm_slicaddr,
			 (unchar)bus_level,
			 (unchar)bus_vector, slic_dest, slic_vector,
			 SL_MINTR | slic_bin);

	return BUS_INT_OK;
}
