/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

/*	Copyright (c) 1984 AT&T	*/
/*	  All Rights Reserved  	*/

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/

#ident "$Header: tirdwr.c 1.10 90/06/28 $"

/* $Log:	tirdwr.c,v $
 */

/*
 * Transport Interface Library read/write module - issue 1
 */

#include "sys/types.h"
#include "sys/param.h"
#include "sys/sysmacros.h"
#include "sys/mutex.h"
#include "sys/intctl.h"
#include "sys/session.h"
#include "sys/stream.h"
#include "sys/stropts.h"
#include "sys/tihdr.h"
#include "sys/signal.h"
#include "sys/user.h"
#include "sys/debug.h"
#include "sys/errno.h"
#include "sys/strid.h"
#include "sys/tirdwr.h"

extern struct trw_trw trw_trw[];
extern int trw_cnt;
extern lock_t tirdwr_lock;
extern nulldev();


/* stream data structure definitions */
int tirdwropen(), tirdwrclose(), tirdwrrput(), tirdwrwput();
static struct module_info tirdwr_info = {STRID_TIRDWR, "tirdwr", 0, INFPSZ, 4096, 1024};
static struct qinit tirdwrrinit = {tirdwrrput, NULL, tirdwropen, tirdwrclose, nulldev, &tirdwr_info, NULL};
static struct qinit tirdwrwinit = {tirdwrwput, NULL, tirdwropen, tirdwrclose, nulldev, &tirdwr_info, NULL};
struct streamtab tirdwrinfo = {&tirdwrrinit, &tirdwrwinit, NULL, NULL};



/*
 * Called from main at boot time.
 */
void
tirdwr_init()
{
	/*
	 * This lock mutexes the tim_tim table.
	 */
	init_lock(&tirdwr_lock);
}

/*
 * tirdwropen - open routine gets called when the
 *	       module gets pushed onto the stream.
 */

/*ARGSUSED*/
tirdwropen(q, dev, flag, sflag)
	dev_t dev;
	register queue_t *q;
{
	register struct trw_trw *trwptr;
	register spl_t spl;

	ASSERT_DEBUG(q != NULL, "tirdwropen: q is null");

	spl = p_lock(&tirdwr_lock, SPLSTR);

	/* If not first open then just return */
	if (q->q_ptr) {
		v_lock(&tirdwr_lock, spl);
		return(1);
	}

	/* First open */
	/* find free data structure */
	for (trwptr=trw_trw; trwptr < &trw_trw[trw_cnt]; trwptr++) {
		if (!(trwptr->trw_flags & USED)) {
			break;
		}
	}
	if (trwptr >= &trw_trw[trw_cnt]) {
		v_lock(&tirdwr_lock, spl);
		u.u_error = ENOSPC;
		return(OPENFAIL);
	}

	/* Since stream head is involved, must lock stream */

	(void) p_lock(&q->q_str->sd_mutex, SPLSTR);

	if (!check_strhead(q)) {
		v_lock(&q->q_str->sd_mutex, spl);
		v_lock(&tirdwr_lock, spl);
		u.u_error = EPROTO;
		return(OPENFAIL);
	}

	if ((trwptr->trw_mp = (mblk_t *)allocb(sizeof(struct T_discon_req), BPRI_LO)) == NULL) {
		v_lock(&q->q_str->sd_mutex, spl);
		v_lock(&tirdwr_lock, spl);
		u.u_error = EAGAIN;
		return(OPENFAIL);
	}

	strip_strhead(q);

	/* initialize data structure */
	trwptr->trw_flags = USED;
	trwptr->trw_rdq = q;
	q->q_ptr = (caddr_t)trwptr;
	WR(q)->q_ptr = (caddr_t)trwptr;
	WR(q)->q_maxpsz = WR(q)->q_next->q_maxpsz;
	q->q_maxpsz = q->q_next->q_maxpsz;

	v_lock(&q->q_str->sd_mutex, spl);
	v_lock(&tirdwr_lock, spl);
	init_lock(&trwptr->trw_lock);

	/* And now we are all set to enable the procedures */
	enable_procs(q);
	return(1);
}


/*
 * tirdwrclose - This routine gets called when the module
 *              gets popped off of the stream.
 */

tirdwrclose(q)
	register queue_t *q;
{
	register struct trw_trw *trwptr;
	register mblk_t *mp;
	register spl_t spl;
	register union T_primitives *pptr;


	ASSERT_DEBUG(q != NULL, "tirdwrclose: q is null");


	/* assign saved data structure from queue */
	trwptr = (struct trw_trw *)q->q_ptr;
	ASSERT_DEBUG(trwptr != NULL, "tirdwrclose: trwptr is null");
	 
	spl = p_lock(&trwptr->trw_lock, SPLSTR);
	if ((trwptr->trw_flags & ORDREL) && !(trwptr->trw_flags & FATAL)) {
		mp = trwptr->trw_mp;
		trwptr->trw_mp = NULL;
		v_lock(&trwptr->trw_lock, spl);

		pptr = (union T_primitives *)mp->b_rptr;
		mp->b_wptr = mp->b_rptr + sizeof(struct T_ordrel_req);
		pptr->type = T_ORDREL_REQ;
		mp->b_datap->db_type = M_PROTO;
		putnext(WR(q), mp);
	} else {
		v_lock(&trwptr->trw_lock, spl);
	}

	/* Turn off all service/put procs. */
	disable_procs(q);

	freemsg(trwptr->trw_mp);
	trwptr->trw_mp = NULL;
	trwptr->trw_rdq = NULL;

	spl = p_lock(&tirdwr_lock, SPLSTR);
	trwptr->trw_flags = 0;
	q->q_ptr = NULL;
	v_lock(&tirdwr_lock, spl);
}


/*
 * tirdwrrput - Module read queue put procedure.
 *             This is called from the module or
 *	       driver downstream.
 */

tirdwrrput(q, mp)
	register queue_t *q;
        register mblk_t *mp;
{
	register union T_primitives *pptr;
	register struct trw_trw *trwptr;
	register mblk_t *tmp;
	register spl_t spl;

	ASSERT_DEBUG(q != NULL, "tirdwrrput: q is null");

	trwptr = (struct trw_trw *)q->q_ptr;

	ASSERT_DEBUG(trwptr != NULL, "tirdwrrput: trwptr is null");

	spl = p_lock(&trwptr->trw_lock, SPLSTR);
	if ((trwptr->trw_flags & FATAL) && !(trwptr->trw_flags & WAITACK)) {
		v_lock(&trwptr->trw_lock, spl);
		freemsg(mp);
		return;
	}
	switch(mp->b_datap->db_type) {

	default:
		v_lock(&trwptr->trw_lock, spl);
		putnext(q, mp);
		return;

	case M_DATA:
		v_lock(&trwptr->trw_lock, spl);
		if (msgdsize(mp) == 0) {
			freemsg(mp);
			return;
		}
		putnext(q, mp);
		return;

	case M_PCPROTO:
	case M_PROTO:
		/* is there enough data to check type */
		ASSERT_DEBUG((mp->b_wptr - mp->b_rptr) >= sizeof(long),
			     "tirdwrrput: invalid message in stream head");

		pptr = (union T_primitives *)mp->b_rptr;

		switch (pptr->type) {

		case T_EXDATA_IND:
			trwptr->trw_flags |= FATAL;
			v_lock(&trwptr->trw_lock, spl);
			send_fatal(q, mp);
			return;
		case T_DATA_IND:
			v_lock(&trwptr->trw_lock, spl);
			if (msgdsize(mp) == 0) {
				freemsg(mp);
				return;
			}

			tmp = (mblk_t *)unlinkb(mp);
			freemsg(mp);
			putnext(q, tmp);
			return;

		case T_ORDREL_IND:
			trwptr->trw_flags |= ORDREL;
			v_lock(&trwptr->trw_lock, spl);
			mp->b_datap->db_type = M_DATA;
			mp->b_wptr = mp->b_rptr;
			putnext(q, mp);
			return;

		case T_DISCON_IND:
			trwptr->trw_flags |= DISCON;
			trwptr->trw_flags &= ~ORDREL;
			v_lock(&trwptr->trw_lock, spl);
			if (msgdsize(mp) != 0) {
				tmp = (mblk_t *)unlinkb(mp);
				putnext(q, tmp);
			}
			mp->b_datap->db_type = M_HANGUP;
			mp->b_wptr = mp->b_rptr;
			putnext(q, mp);
			return;

			/* flow thru */
		default:
			trwptr->trw_flags |= FATAL;
			v_lock(&trwptr->trw_lock, spl);
			send_fatal(q, mp);
			return;
		}
	}
	/*NOTREACHED*/
}


/*
 * tirdwrwput - Module write queue put procedure.
 *             This is called from the module or
 *	       stream head upstream.
 */

tirdwrwput(q, mp)
	register queue_t *q;
	register mblk_t *mp;
{
	register struct trw_trw *trwptr;
	register spl_t spl;

	ASSERT_DEBUG(q != NULL, "tirdwrwput: q is null");

	trwptr = (struct trw_trw *)q->q_ptr;

	ASSERT_DEBUG(trwptr != NULL, "tirdwrwput: trwptr is null");

	spl = p_lock(&trwptr->trw_lock, SPLSTR);
	if (trwptr->trw_flags & FATAL) {
		v_lock(&trwptr->trw_lock, spl);
		freemsg(mp);
		return;
	}

	switch(mp->b_datap->db_type) {

	default:
		v_lock(&trwptr->trw_lock, spl);
		putnext(q, mp);
		return;

	case M_DATA:
		v_lock(&trwptr->trw_lock, spl);
		if (msgdsize(mp) == 0) {
			freemsg(mp);
			return;
		}
		putnext(q, mp);
		return;

	case M_PROTO:
	case M_PCPROTO:
		trwptr->trw_flags |= FATAL;
		v_lock(&trwptr->trw_lock, spl);
		send_fatal(q, mp);
		return;
	}
	/*NOTREACHED*/
}


/* 
 * Sends an error message upto the stream head.
 * 
 * Assumes no locks are held (none required).
 */
static
send_fatal(q, mp)
	register queue_t *q;
	register mblk_t *mp;
{
	mp->b_datap->db_type = M_ERROR;
	*mp->b_datap->db_base = EPROTO;
	mp->b_rptr = mp->b_datap->db_base;
	mp->b_wptr = mp->b_datap->db_base + sizeof(char);
	freemsg(unlinkb(mp));
	if (q->q_flag & QREADR) {
		putnext(q, mp);
	} else {
		qreply(q, mp);
	}
}


/*
 * Checks the stream head of messages that make no sense
 * in the read-write context.  Returns 1 if none exist or
 * if they are cleanable (see strip_strhead), 0 otherwise.
 *
 * Assumes that the stream is lcoked.
 */
static
check_strhead(q)
	queue_t *q;
{
	register mblk_t *mp;
	register union T_primitives *pptr;

	for (mp = q->q_next->q_first; mp != NULL; mp = mp->b_next) {

		switch(mp->b_datap->db_type) {
		case M_PROTO:
			pptr = (union T_primitives *)mp->b_rptr;
			if ((mp->b_wptr - mp->b_rptr) < sizeof(long)) {
				return(0);
			}
			switch (pptr->type) {

			case T_EXDATA_IND:
				return(0);
			case T_DATA_IND:
				if (mp->b_cont && (mp->b_cont->b_datap->db_type != M_DATA)) {
					return(0);
				}
				break;
			default:
				return(0);
			}
			break;

		case M_PCPROTO:
			return(0);

		case M_DATA:
		case M_SIG:
			break;
		default:
			return(0);
		}
	}
	return(1);
}



/*
 * Strips the stream head of messages that make no sense
 * in the read-write context.
 *
 * Assumes that the stream is locked.
 */
static
strip_strhead(q)
	queue_t *q;
{
	register mblk_t *mp;
	register mblk_t *emp;
	register mblk_t *tmp;
	register union T_primitives *pptr;

	q = q->q_next;
	for (mp = q->q_first; mp != NULL;) {

		switch(mp->b_datap->db_type) {
		case M_PROTO:
			pptr = (union T_primitives *)mp->b_rptr;
			switch (pptr->type) {

			case T_DATA_IND:
				if (msgdsize(mp) == 0) {
strip0:
					tmp = mp->b_next;
					l_rmvq(q, mp);
					freemsg(mp);
					mp = tmp;
					break;
				}
				emp = mp->b_next;
				l_rmvq(q, mp);
				tmp = (mblk_t *)unlinkb(mp);
				freeb(mp);
				l_insq(q, emp, tmp);
				mp = emp;
				break;
			default:
				mp = mp->b_next;
				break;
			}
			break;

		case M_DATA:
			if (msgdsize(mp) == 0) {
				goto strip0;
			}
			mp = mp->b_next;
			break;

		default:
			mp = mp->b_next;
			break;
		}
	}
}
