/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ident		"$Header: wd.c 1.42 90/05/04 $"

/*
 * wd.c 
 *	SCSI disk device driver for SSM
 */

/* $Log:	wd.c,v $
 */

#include "sys/types.h"
#include "sys/devsw.h"
#include "sys/vtoc.h"
#include "sys/sysmacros.h"
#include "sys/signal.h"
#include "sys/param.h"
#include "sys/debug.h"
#include "sys/mutex.h"
#include "sys/user.h"
#include "sys/conf.h"
#include "sys/buf.h"
#include "sys/uio.h"
#include "sys/vnode.h"
#include "sys/systm.h"
#include "sys/vm.h"
#include "sys/stats_disk.h"
#include "sys/ioctl.h"
#include "sys/file.h"
#include "sys/timer.h"
#include "sys/proc.h"
#include "sys/scsi.h"
#include "sys/scsidisk.h"
#include "sys/scsiioctl.h"
#include "sys/cmn_err.h"
#include "sys/errno.h"

#include "sys/engine.h"
#include "sys/slic.h" 
#include "sys/slicreg.h" 
#include "sys/ssm_comblock.h"
#include "sys/clkarb.h"

#include "sys/autoconf.h"
#include "sys/pte.h"
#include "sys/intctl.h"
#include "sys/plocal.h"
#include "sys/ssm.h"
#include "sys/ssm_device.h"
#include "sys/ssm_scsi.h"
#include "sys/scsi_adapter.h"
#include "sys/wd.h"
#include "sys/wd_device.h"

wd_t 	**wd_info;
int	wd_base_vector;
int	wd_ndevs;

static	void wd_error();

extern unchar wd_bin;
extern	wdmaxmin;
extern	struct wd_bconf wdbconf[];
extern	int wdretrys;
extern	caddr_t get_cb();
extern	struct cb_desc *alloc_cb_desc();
extern	ulong	scb_buf_iovects();
extern	char *ssm_alloc();
extern  dev_deref();

int wd_probe(), wdstrat();
void wd_init(), wd_intr();

struct dev *wd_dev;			  /* to be allocated in wd_init */
extern int wd_b_major;
extern struct devops wddevops;

caddr_t	wd_compcode[] = {
	"Command Block Busy",
	"OK",
	"Bad Command Block",
	"No Target",
	"SCSI Bus Error",
	"Unrecognized Completion Code"
};

int	wdncompcodes = sizeof(wd_compcode) / sizeof(wd_compcode[0]);

caddr_t	wd_addsense[] = {
	"No Additional Sense",
	"No Index/Sector Signal",
	"No Seek Complete",
	"Write Fault",
	"Drive Not Ready",
	"Drive Not Selected",
	"No Track Zero Found",
	"Multiple Drives Selected",
	"Communication Failure",
	"Track Following Error",
	"Error 0x0a",
	"Error 0x0b",
	"Error 0x0c",
	"Error 0x0d",
	"Error 0x0e",
	"Error 0x0f",
	"ID Error",
	"Unrecovered Read",
	"No Addr Mark in ID",
	"No Addr Mark in Data",
	"No Record Found",
	"Seek Positioning Error",
	"Data Synch Mark Error",
	"Recovered Read, Retries",
	"Recovered Read, ECC",
	"Defect List Error",
	"Parameter Overrun",
	"Synchronous Transfer Error",
	"Defect List Not Found",
	"Compare Error",
	"Recoevered ID, ECC",
	"Error 0x1f",
	"Invalid Command Code",
	"Illegal Block Address",
	"Illegal Function",
	"Error 0x23",
	"Illegal Field in CDB",
	"Inavalid LUN",
	"Invalid Parameter",
	"Write Protected",
	"Medium Changed",
	"Power On or Reset",
	"Mode Select Changed",
	"Error 0x2b",
	"Error 0x2c",
	"Error 0x2d",
	"Error 0x2e",
	"Error 0x2f",
	"Inacompatible Cartridge",
	"Format Corrupted",
	"No Spare Available",
	"Error 0x33",
	"Error 0x34",
	"Error 0x35",
	"Error 0x36",
	"Error 0x37",
	"Error 0x38",
	"Error 0x39",
	"Error 0x3a",
	"Error 0x3b",
	"Error 0x3c",
	"Error 0x3d",
	"Error 0x3e",
	"Error 0x3f",
	"Ram Failure",
	"Data Path Diag Failure",
	"Power On Diag Failure",
	"Message Reject Error",
	"Internal Controller Error",
	"Select/Reselect Failed",
	"Unsuccessful Soft Reset",
	"Parity Error",
	"Initiator Detected Error",
	"Inappropriate/Illegal Message",
	""
};

int	wdnaddsense = sizeof(wd_addsense) / sizeof(wd_addsense[0]);

#ifdef DEBUG
int wd_debug = 0;				  /*  0=off */
#define WDPRINTF	if (wd_debug) CPRINTF 
#endif /* DEBUG */


/*
 * wd_probe - probe device
 *
 * This procedure polls a device with the test unit ready command
 * to determine if the device is present.
 * If a device is found an SCSI_INQUIRY command is issued to make sure
 * we found a SCSI disk and not a tape.
 */
wd_probe(dev)
	struct wd_device *dev;
{
	struct	scsi_cb	*cbs;
	struct	scb_init sinit;
	struct	scinq_dev *wd_inq;
	short	devno = SSM_SCSI_DEVNO(dev->config->space, 
					dev->config->dev_addr);
	unchar	compcode;
	static	daddr_t wd_sense_data = NULL;
	static	daddr_t wd_probe_data = NULL;
	
	sinit.si_mode = SCB_PROBE;
	sinit.si_flags = 0;
	sinit.si_ssm_slic = dev->bus->owner->config->dev_addr;
	sinit.si_scsi = 0;		/* bus no. - currently always 0 */ 
	sinit.si_target = (unchar)dev->config->space;
	sinit.si_lunit = (unchar)dev->config->dev_addr;
	sinit.si_control = 0;
	init_ssm_scsi_dev(&sinit);

	if (sinit.si_id >= 0) {
		cbs = sinit.si_cbs;	/* CB to use for SCSI probe commands */
		ASSERT(cbs, "wd_probe: CB not allocated");
		/*
		 *+ wd command blocks were not allocated.
		 *+ The system did not initialize properly.
		 */

		cbs->sw.cb_unit_index = SCVEC(sinit.si_id, 0);  
                cbs->sw.cb_slic = dev->bus->owner->config->dev_addr;
	} else {
#ifdef	DEBUG
		if ( wd_debug > 2 )
			CPRINTF("wd_probe: returning not found 1\n");
#endif	/* DEBUG */
		return (SPROBE_NOTFOUND);/* An initialization error occurred */
	}
	
#ifdef DEBUG
		if (wd_debug > 1)
			CPRINTF("wd_probe: wd%d cb @ 0x%x, SSM devno = 0x%x\n",
			       DEVICEN(dev), sinit.si_cbs, sinit.si_id);
#endif /* DEBUG */

	/* 
	 * Allocate a request sense buffer and a data buffer if they have 
	 * not been allocated.
 	 */
	if (!wd_sense_data) 
		wd_sense_data = (daddr_t) ssm_alloc(SIZE_MAXDATA,
						    SSM_ALIGN_XFER,
						    (uint) SSM_BAD_BOUND);
	if (!wd_probe_data)
		wd_probe_data = (daddr_t) ssm_alloc(SIZE_MAXDATA,
						    SSM_ALIGN_XFER,
						    (uint) SSM_BAD_BOUND);

	/*
	 * wdp_command is responsible for obtaining 
	 * request sense information if CHECK CONDITION occurs
     	 * after the test unit ready command.
	 */
	compcode = wdp_command(cbs, SCSI_TEST, 
	    (daddr_t)wd_sense_data, (daddr_t)wd_probe_data, devno);
	switch (compcode) {
		case SCB_NO_TARGET:
			if (!SCSI_CHECK_CONDITION(cbs->sh.cb_status)) {
#ifdef	DEBUG
				if (wd_debug > 2)
					CPRINTF("wd_probe: returning not found 2\n");
#endif	/* DEBUG */
				return (SPROBE_NOTARGET);
			}
			else {
#ifdef	DEBUG
				if (wd_debug > 2)
					CPRINTF("wd_probe: returning not found 3\n");
#endif	/* DEBUG */
			return (SPROBE_NOTFOUND);
		}
		case SCB_BAD_CB:
		case SCB_SCSI_ERR:
#ifdef	DEBUG
			if (wd_debug > 2)
				CPRINTF("wd_probe: returning not found 4\n");
#endif	/* DEBUG */
			return (SPROBE_NOTARGET);
		case SCB_OK:
			break;
		default:
#ifdef	DEBUG
			if (wd_debug > 2)
				CPRINTF("wd_probe: returning not found 5\n");
#endif	/* DEBUG */
			return (SPROBE_NOTFOUND);
		}
			


	/*
	 * Discover type of device found, make sure it is not a 
	 * SCSI tape.
 	 */
	if (wdp_command(cbs, SCSI_INQUIRY, 
	    wd_sense_data, wd_probe_data, devno) != SCB_OK) {
#ifdef	DEBUG
		if ( wd_debug > 2)
			CPRINTF("wd_probe: returning not found 6\n");
#endif	/* DEBUG */
		return (SPROBE_NOTFOUND);
	}

	if (SCSI_CHECK_CONDITION(cbs->sh.cb_status)) {	
#ifdef	DEBUG
		if (wd_debug > 2)
			CPRINTF("wd_probe: returning not found 7\n");
#endif	/* DEBUG */
		return (SPROBE_NOTFOUND);
	}
	/*
	 * Make sure it is a direct access device.
	 * and that is is an CCS drive.
	 */
	wd_inq = (struct scinq_dev *) wd_probe_data;
	if ((wd_inq->sc_hdr.ih_devtype == INQ_DIRECT) && 
 	    (wd_inq->sc_hdr.ih_reserved == INQ_RES_CCS)) {
#ifdef	DEBUG
		if (wd_debug > 2)
			CPRINTF("wd_probe: returning found 8\n");
#endif	/* DEBUG */

		return (SPROBE_ONELUN);
	}
	else {
#ifdef	DEBUG
		if (wd_debug > 2)
			CPRINTF("wd_probe: returning not found 9\n");
#endif	/* DEBUG */
		return(SPROBE_NOTFOUND);
	}

}


/*
 * wd_init - initialize all channels associated with this driver.
 *
 * This procedure initializes and allocates all device driver data
 * structures. It uses the configuration information passed in from
 * autoconfig() and from the device drivers binary configuration tables.
 */
void
wd_init(driver)
	struct driver_module *driver;
{
	register wd_t *ip;		/* info structure pointer */ 
	int	ndevs = driver->numdev;
	register struct scsi_cb *cb;
	register struct wd_bconf *bconf;
	struct scb_init sinit;
	int	dev, x;
	int	capacity;
	ulong	*maps;
	char	vendor[8], product[16];
	struct wd_device *wd;
	int vec;

	wd_info = (wd_t **) calloc(ndevs * sizeof(wd_t *));
	wd_ndevs = ndevs;
	wd_base_vector = ivecpeek(wd_bin); 

#ifdef DEBUG
	if (wd_debug > 2)
		CPRINTF( "wd_init: wd bin = 0x%x, base vector = 0x%x\n",
		       wd_bin, wd_base_vector);
#endif /* DEBUG */	

	wd_dev = (struct dev *)calloc( ndevs * sizeof (struct dev) );
	if ( !wd_dev ) {
		cmn_err(CE_WARN,
		"wd_init: No memory for dev structures\n");
		/*
		 *+ The wd driver was unable to allocate
		 *+ memory for its internal data structures.
		 *+ No wd devices will be accessible.  Something
		 *+ is allocating memory incorrectly or your
		 *+ system needs more memory to support the
		 *+ current configuration.
		 */
		for ( dev = 0, wd = (struct wd_device *) driver->devices;
		      dev < ndevs; dev++, wd++ )
			wd->driver = (struct driver_module *) 0;
		return;
	}

	for (dev = 0, wd = (struct wd_device *)driver->devices; 
	     dev < ndevs; dev++, wd++) {
		/*
		 * Allocate vector for every device whether alive
		 * or not (only have to init it if alive, though.)
		 */
		vec = ivecall(wd_bin);	
		(void) ivecall(wd_bin);	  /* dual ints */
		if (wd->driver == (struct driver_module *)NULL)
			continue;
		ivecinit(wd_bin, vec, wd_intr);
		ivecinit(wd_bin, vec+1, wd_intr);

		/*
		 * Verify that this device has been included in
		 * the device's binary configuration table. 
		 */
		if (dev > wdmaxmin) {
			cmn_err(CE_WARN,
			"wd%d: non-binary configured device found in config table ... deconfiguring", 
				dev);
			/*
			 *+ The number of wd devices 
			 *+ configured into the kernel is greater than the number
			 *+ of binary configuration entries.
			 *+ Excess devices will be deconfigured.
			 *+ This can be corrected in the system
			 *+ configuration file or in wd_space.c.
			 */
			wd->driver = (struct driver_module *)NULL;
			continue;
		}

		/*
		 * Verify that there are enough iovects allocted for each cb
		 * to allow for maximum data transfer. If this condition
 		 * is not met the driver may hang.
		 */
		bconf = &wdbconf[dev];
		if (wd->n_maps < (bconf->bc_iovects * NCBPERSCSI)) {
			cmn_err(CE_WARN,
			"wd%d: not enough iovects for device ... deconfiguring",
				dev);
			/*
			 *+ There are not enough iovects for 
			 *+ this device to allow the maximum
			 *+ transfer size.  Because this is a requirement
			 *+ for the wd driver, the device has
			 *+ been deconfigured.  Corrective action:  correct your
			 *+ system configuration for this device.
			 */
			wd->driver = (struct driver_module *)NULL;
			continue;
		}

		/*
		 * allocate info struct
 		 */	
		wd_info[dev] = ip = (wd_t *)calloc(sizeof(wd_t));

		ip->wd_devno = SSM_SCSI_DEVNO(wd->config->space, 
					      wd->config->dev_addr);
		ip->wd_ctlr = (char *)wd->bus->owner 
				      - (char *)&ssm_devices[0];

		/* 
		 * Initialize cb's here,
 		 * including request sense buffers and iovects.
		 */
		sinit.si_mode = SCB_BOOT;
		sinit.si_flags = 0;
		sinit.si_ssm_slic = wd->bus->owner->config->dev_addr;
		sinit.si_scsi = 0;		/* bus no. */ 
		sinit.si_target = (unchar)wd->config->space;
		sinit.si_lunit = (unchar)wd->config->dev_addr;
		sinit.si_host_bin = wd_bin;
		sinit.si_host_basevec = vec; 
		sinit.si_control = 0;


		init_ssm_scsi_dev(&sinit);
		if (sinit.si_id < 0) {
			cmn_err(CE_WARN,
			"wd%d: initialization error - exceeded SCSI ",
				dev);
			/*
			 *+ There are more devices on the SCSI
			 *+ bus than the SSM can handle.  This
			 *+ device will not be usable as a result.
			 */
			wd->driver = (struct driver_module *)NULL;
			continue;
		} else 

		ip->wd_cbdp = (struct cb_desc *) alloc_cb_desc(
			(char *) sinit.si_cbs,
			(short) NCBPERSCSI,
			(short) sizeof(struct scsi_cb),
		 	(short) FLD_OFFSET(struct scsi_cb, sw.cb_state));
		
#ifdef DEBUG
		if (wd_debug > 1)
			CPRINTF("wd_init: wd%d cb @ 0x%x, SSM devno = 0x%x\n",
			       DEVICEN(wd), sinit.si_cbs, sinit.si_id);
#endif /* DEBUG */

		/*
		 * for each cb
	 	 *	allocate request sense buffers
		 * 	allocate data buffers
		 * 	fill in all sw fields of cbs
		 */
		for(x = 0, cb = (struct scsi_cb *) ip->wd_cbdp->cb_cbs,
			maps = ((struct ssm_scsi_dma *)wd->map_origin)->map;
		    x < ip->wd_cbdp->cb_count; 
		    x++, cb++, maps += bconf->bc_iovects) {

			cb->sh.cb_sense = (ulong) ssm_alloc(SIZE_MAXDATA, 
					SSM_ALIGN_XFER, SSM_BAD_BOUND);
			cb->sw.cb_data = (ulong) ssm_alloc(SIZE_MAXDATA, 
					SSM_ALIGN_XFER, SSM_BAD_BOUND);
			cb->sh.cb_slen = SIZE_MAXDATA;
			
			cb->sw.cb_iovstart = (ulong *) maps;
			cb->sw.cb_iovnum = bconf->bc_iovects;
			cb->sw.cb_unit_index = SCVEC(sinit.si_id, x);  
			cb->sw.cb_scsi_device = ip->wd_devno;
			cb->sw.cb_slic = wd->bus->owner->config->dev_addr;
		}

		/*
		 * Use first cb in cb pair for inquiry and capacity commannds
		 */
		cb = (struct scsi_cb *) ip->wd_cbdp->cb_cbs;
		
#ifdef	DEBUG
		if (wd_debug > 1)
			CPRINTF("wd_init: start inquiry command\n");
#endif	/* DEBUG */
		/*
		 * Get inquiry information from drive.
		 */
		wd_command(cb, SCSI_INQUIRY, 0);
		if (wd_cbfin(cb) != SCB_OK) {
			cmn_err(CE_WARN,
			"wd%d: problem doing INQUIRY ... deconfiguring",
				dev);
			/*
			 *+ The wd device failed to respond to
			 *+ the SCSI INQUIRY command and is
			 *+ being deconfigured.  This can indicate
			 *+ a problem with the SSM or with the drive,
			 *+ as the drive responded validly when
			 *+ previously probed.
			 */
			if (fp_lights)
				FP_IO_ERROR;
			continue;
		}

		bcopy(((struct scinq_dev *) cb->sw.cb_data)->sc_vendor, vendor, 8);
		bcopy(((struct scinq_dev *)cb->sw.cb_data)->sc_product,
		      product,16);

#ifdef	DEBUG
		if (wd_debug > 1)
			CPRINTF("wd_init: start capacity command\n");
#endif	/* DEBUG */
		/*
		 * Get the configured capacity of the drive.
		 */
		if ((capacity = wd_readc(cb)) <= 0) {
			cmn_err(CE_WARN,
			"wd%d: unable to READ CAPACITY ... deconfiguring\n",
				dev);
			/*
			 *+ The wd device failed to respond to a
			 *+ SCSI READ CAPACITY command and is
			 *+ being deconfigured.  This can indicate
			 *+ a problem with the SSM or with the drive,
			 *+ as the drive responded validly when
			 *+ previously probed and inquired of.
			 */
			wd->driver = (struct driver_module *)NULL;
			if (fp_lights)
				FP_IO_ERROR;
			continue;
		}

		/*
		 * capacity actually shows last addressable block,
		 * so increment to get actual limit.
		 */
		ip->wd_capacity = capacity + 1;
        
		ip->wd_alive = (wd->driver) ? 1 : 0;

		ip->wd_retrys = wdretrys;
		ip->wd_iovcount = bconf->bc_iovects;
		ip->wd_inuse = 0;
		ip->wd_flags =  ip->wd_alive ? WD_IS_ALIVE : 0;
		
		init_lock(&ip->wd_lock);
		bufinit(&ip->wd_rbufh);
	
		if (ip->wd_alive) {
			/* If we get here, device is O.K.--add it to devsw */
			wd_dev[dev].d_dev = makedev( wd_b_major,
					    unittominor(dev)|V_RAW_MINOR);
			wd_dev[dev].d_refcnt = 1;
			wd_dev[dev].d_flags = D_ASYNC;
			wd_dev[dev].d_ops = &wddevops;
#ifdef DEBUG
			if (wd_debug > 2)
				CPRINTF( "wd_init: Adding wd%d to devsw\n",
				       dev);
#endif /* DEBUG */
			if ( devsw_add( &wd_dev[dev] ) < 0 ) {
				CPRINTF( "wd%d: can't add dev 0x%x to devsw\n",
				       dev, &wd_dev[dev]);
			} else {
				dev_deref(&wd_dev[dev]);
			}
		}

		ip->wd_dstats = ds_alloc_stats("wd", (ulong)ip->wd_capacity,
			dev, DS_BLK_HIST);
	}
	
#ifdef	DEBUG
	if (wd_debug > 1) 
		CPRINTF("wd_init: exiting boot\n");
#endif	/* DEBUG */
}

		

/*
 * wdopen - open a channel
 *
 * Check for validity of device/drive
 *
 * No locking is necessary because wdopen looks only at static
 * information that is recorded at boot time.
 */
wdopen(dp, flags)
	struct dev *dp;
	int flags;
{
	dev_t dev = dp->d_dev;
	register int drive = SCSID_DRIVE(dev);
	register wd_t *ip;
	int	length;
	int err = 0;
	spl_t	s;

	if (drive >= wd_ndevs)
		return (ENXIO);

	ip = wd_info[drive];
	if (ip == NULL)	
		return (ENXIO);
	length = ip->wd_capacity;

	if ((length <= 0)			/* partition has length */
	    || (!(ip->wd_flags & WD_IS_ALIVE)))	/* disk is alive */
		return (ENXIO);

	s = p_lock(&ip->wd_lock, WDSPL);
	if (ip->wd_flags & WD_EXCLUSIVE) {
#ifdef DEBUG
		WDPRINTF("wdopen: exclusive set - EBUSY\n");
#endif /* DEBUG */
		err = EBUSY;	
	} else if (flags & FEXCL) {
#ifdef DEBUG
		WDPRINTF("wdopen: FEXCL\n");
#endif /* DEBUG */
		if (!suser()) {
#ifdef DEBUG
			WDPRINTF("wdopen: !suser\n");
#endif /* DEBUG */
			err = EPERM;
		} else if (ip->wd_nopen) {
#ifdef DEBUG
			WDPRINTF("wdopen: EBUSY\n");
#endif /* DEBUG */
			err = EBUSY;
		} else {
			ip->wd_flags |= WD_EXCLUSIVE;
		}
	}
	if (!err)
		ip->wd_nopen++;
#ifdef DEBUG
	WDPRINTF("wdopen: dev 0x%x -- %d opens\n", dev, ip->wd_nopen);
#endif /* DEBUG */
	v_lock(&ip->wd_lock, s);
	return(err);
}

/*
 * wdclose - close a channel 
 *
 * Decrement reference count, make sure temporary settings
 * are set back to their default values.
 */
/*ARGSUSED*/
wdclose(dp, flag)
	struct dev *dp;
	int flag;
{
	dev_t dev = dp->d_dev;
	wd_t *ip;
	spl_t s;

	ip = wd_info[SCSID_DRIVE(dev)];
	s = p_lock(&ip->wd_lock, WDSPL);
	ip->wd_nopen--;
	if (ip->wd_flags & WD_EXCLUSIVE) {
		ip->wd_flags &= ~WD_EXCLUSIVE;
	}
	v_lock(&ip->wd_lock, s);
#ifdef DEBUG
	WDPRINTF("wdclose: dev 0x%x - nopens = %d\n", dev, ip->wd_nopen);
#endif /* DEBUG */
}

/*
 * wdminphys - correct for too large a request
 */
/*ARGSUSED*/
wdminphys(dp, bp)
	struct dev *dp;
	register struct buf *bp;
{
	/*  commented out because doesn't do anything
	scb_num_iovects(bp);
	*/

	if (bp->b_bcount > WD_MAX_XFER * DEV_BSIZE)
		bp->b_bcount = WD_MAX_XFER * DEV_BSIZE;
}

/*
 * wdread - standard raw read procedure
 */
wdread(dp, uio, flags)
	struct dev *dp;
	struct uio *uio;
	int flags;
{
	dev_t dev = dp->d_dev;
	int err, diff;
	off_t lim;
		
	ASSERT(SCSID_DRIVE(dev) < wd_ndevs, "wdread dev");
	/*
	 *+ A bad device number was passed in to wdread.
	 *+ The system is in an inconsistent state.
	 */

		lim = wd_info[SCSID_DRIVE(dev)]->wd_capacity;
	lim <<= DEV_BSHIFT;
	err = physck(lim, uio, B_READ, &diff);
	if (err != 0) {
		if (err == -1)	/* not an error, but request of 0 bytes */
			err = 0;
		return(err);
	}
	if (flags & IO_ASYNC) {
		err = aphysio(dp, B_READ, uio, diff, flags);
	} else {
		err = physio(dp, (struct buf *)0, B_READ, uio);
		uio->uio_resid += diff;
	}
	return(err);
}

/*
 * wdwrite - standard raw write procedure
 */
wdwrite(dp, uio, flags)
	struct dev *dp;
	struct uio *uio;
	int flags;
{
	dev_t dev = dp->d_dev;
	int err, diff;
	off_t lim;
		
	ASSERT(SCSID_DRIVE(dev) < wd_ndevs, "wdwrite dev");
	/*
	 *+ A bad device number was passed in to wdwrite.
	 *+ The system is in an inconsistent internal state.
	 */

		lim = wd_info[SCSID_DRIVE(dev)]->wd_capacity;
	lim <<= DEV_BSHIFT;
	err = physck(lim, uio, B_WRITE, &diff);
	if (err != 0) {
		if (err == -1)	/* not an error, but request of 0 bytes */
			err = 0;
		return(err);
	}
	if (flags & IO_ASYNC) {
		err = aphysio(dp, B_WRITE, uio, diff, flags);
	} else {
		err = physio(dp, (struct buf *)0, B_WRITE, uio);
		uio->uio_resid += diff;
	}
	return(err);
}

/*
 * wdstrat - SSM SCSI disk strategy routine
 * 	will check block sizes and limits.
 *
 * Locking of the wd_info structure is performed before
 * calls to disksort (needs exclusive access to bufh) and
 * wdstart (needs exclusive access to command block(s) it gets).
 */
/*ARGSUSED*/
wdstrat(dp, bp, offset)
	struct dev *dp;
	register struct buf *bp;
	daddr_t	offset;
{
	register wd_t *ip;
	spl_t	s;
	int	length;

	bp->b_status = B_STATUS_OK;	  /* base for ORing */
	ip = wd_info[SCSID_DRIVE(bp->b_dev)];
	ASSERT(ip != NULL, "wdstrat: null info ptr");
	/*
	 *+ The system is attempting to do I/O on an invalid
	 *+ wd device.
	 */

	/* 
	 * length of the partition is computed as the size of the disk minus
	 * partition start if the length is specified as SCSID_END.
	 */

		length = ip->wd_capacity;

	/*
 	 * Fail request on: bad byte count, transfer not entirely 
	 * within a partition, or if it is not aligned properly.
	 * Or if drive went bad.
	 */
	if ((bp->b_bcount <= 0)
	||  ((bp->b_bcount & (DEV_BSIZE -1)) != 0)		/* size */
	||  ((bp->b_iotype == B_RAWIO) &&
	      (((int)bp->b_un.b_addr & (SSM_ALIGN_XFER - 1)) != 0))
	||  (bp->b_blkno + offset < 0)      
	||  ((bp->b_blkno + offset + howmany(bp->b_bcount, DEV_BSIZE)) > length)
	||  (((wd_info[SCSID_DRIVE(bp->b_dev)])->wd_flags & WD_BAD) 
	     == WD_BAD)) {
		bp->b_resid = bp->b_bcount;	
		bp->b_flags |= B_ERROR;
		bp->b_error = EINVAL;
		if (((wd_info[SCSID_DRIVE(bp->b_dev)])->wd_flags & WD_BAD) 
		    == WD_BAD)
			bp->b_status |= B_S_XFER_ERR|B_S_UNIT_ERR|B_S_PERM_ERR;
#ifdef	DEBUG
		if (wd_debug > 3) {
			CPRINTF("wdstrat: Printing bp\n");
			wd_dump(sizeof(struct buf), bp);
		}
#endif	/* DEBUG */
		iodone(bp);
		return;
	}

	bp->b_disksortkey = bp->b_blkno + offset;	/* req'd by disksort */

	s = p_lock(&ip->wd_lock, WDSPL);
	disksort(&ip->wd_bufh, bp);
	DS_QUEUED_TIME(bp);
	wdstart(ip);
	DS_Q(ip->wd_dstats);
	v_lock(&ip->wd_lock, s);
}

/*
 * wdstart - start a request to a channel.
 * 	if no command blocks are available just return. 
 *
 * Mutual exclusion must take place before calling wdstart
 */
wdstart(ip)
	register wd_t *ip;
{
	register struct buf *bufh = &ip->wd_bufh;
	register struct scsi_cb *cb;
	register struct buf *bp;
	spl_t s;

	if ((bp = bufh->av_forw) == NULL)	/* Check for requests */
		return;
#ifdef	DEBUG
	if (wd_debug > 3) {
		CPRINTF("wdstart: bp = \n");
		wd_dump(sizeof(struct buf), bp);
	}
#endif	/* DEBUG */
	cb = (struct scsi_cb *) get_cb(ip->wd_cbdp); 
	if (cb == NULL) 
		return;

	if (ip->wd_inuse == 0) {
		if (fp_lights) {
			s = splhi();
			FP_IO_ACTIVE;
			splx(s);
		}
		DS_IOSTART_TIME(ip->wd_dstats);
	}

 	ip->wd_inuse ++;
	bufh->av_forw = bp->av_forw;
	if (bp->b_flags & B_IOCTL) {
#ifdef DEBUG
		WDPRINTF("wdstart: call doioctl\n");
#endif /* DEBUG */
		wd_doioctl(cb, bp);
	} else {
		ds_access_stats(ip->wd_dstats, (ulong)bp->b_disksortkey);
		wd_rw(cb, bp);
	}
}

/*
 * wd_intr 
 * 	Command block completion interrupt handler.
 */
void
wd_intr(vector)
	int vector;
{
	register wd_t	*ip; 
	register struct	buf	*bp;
	register struct scsi_cb *cb;
	register uint level = vector - wd_base_vector;
	struct scrsense *rsense;
	spl_t s, s1;
	unchar	command;

#ifdef	DEBUG
	if (wd_debug > 3) {
		CPRINTF("wd_intr: Just got interrupted\n");
		CPRINTF("wd_intr: vector = 0x%x, level = 0x%x\n",
			vector, level);
	}
#endif	/* DEBUG */

	ASSERT((level >> WD_LEVEL_SHIFT) < wd_ndevs,
	       "wd_intr: interrupt vector exceeds range of configured devices");
	/*
	 *+ An interrupt was received by the wd driver.
	 *+ The interrupt vector translates to a unit 
	 *+ number that exceeds the range of units configured by
	 *+ this device driver.  This indicates an internal 
	 *+ system error such as a corrupt interrupt vector 
	 *+ table.
	 */
	ip = wd_info[level >> WD_LEVEL_SHIFT];
	ASSERT(ip != 0, 
	"wd_intr: interrupt vector matches a unfound configured device");
	/*
	 *+ An interrupt was received by the wd driver. 
	 *+ The interrupt vector translates to a
	 *+ unit that was specified in the system 
	 *+ configuration, but was not found at
	 *+ system startup time.  This indicates an internal
	 *+ system error such as a corrupt interrupt vector 
	 *+ table.
	 */
	s = p_lock(&ip->wd_lock, WDSPL);

	cb = (struct scsi_cb *) ip->wd_cbdp->cb_cbs + 
		(level & WD_LEVEL_CB);
#ifdef	DEBUG
	if (wd_debug > 3) {
		CPRINTF("wd_intr: cb = 0x%x\n", cb);
		wd_printcb(cb);
	}
#endif	/* DEBUG */

	ASSERT(cb->sw.cb_state & CB_BUSY, "wd_intr: interrupt to command block not in use");
	/*
	 *+ A interrupt was received by the wd driver
	 *+ for a device that appears to be idle. This
	 *+ can indicate a problem with the SSM or an 
	 *+ internal error on the driver's part. 
	 */

	bp = cb->sw.cb_bp;

	switch(cb->sh.cb_compcode) {
	case SCB_BAD_CB:
		wd_error(cb, bp, "Hard Error: ");
		goto error;

	case SCB_NO_TARGET:
	case SCB_SCSI_ERR:
		wd_error(cb, bp, "Hard Error: ");
		wd_shutdrive(ip);
		v_lock(&ip->wd_lock, s);
		return ;
	
	case SCB_OK:
		if (SCSI_CHECK_CONDITION(cb->sh.cb_status)) {
			rsense = (struct scrsense *) cb->sh.cb_sense;
			if (SCSI_RS_ERR_CLASS(rsense) != RS_CLASS_EXTEND
			||  SCSI_RS_ERR_CODE(rsense) != RS_CODE_EXTEND) {
				bp->b_status |= B_S_XFER_ERR;
				wd_error(cb, bp,
					 "Hard Error bad sense data: ");
				goto error;
			} 
        	 	switch (rsense->rs_key & RS_SENSEKEYMASK) {
			case RS_RECERR:
				wd_error(cb, bp, "Soft Error: ");
				bp->b_resid = 0;
				bp->b_status |= B_S_MEDIA_ERR;
				goto finish;

			default:
				if (bp->b_flags & B_IOCTL
				    || cb->sw.cb_errcnt >= 4) {
					wd_error(cb, bp, "Hard Error: ");
					bp->b_status |= B_S_XFER_ERR;
					goto error;
				}

				wd_error(cb, bp, "Soft Error: ");
				cb->sw.cb_errcnt += 1;
				/*
				 * Restart command
				 */
				command = (cb->sw.cb_bp->b_flags & B_READ) 
				   ? SCSI_READ : SCSI_WRITE;
				wd_command(cb, command, SCB_IENABLE);
				v_lock(&ip->wd_lock, s);
				return ;
			}

		} else			/* Success */
			bp->b_resid = 0;
			goto finish;
	}

error:	
	bp->b_flags |= B_ERROR;
	bp->b_error = EIO;
	bp->b_resid = cb->sh.cb_count;

finish:
	free_cb((caddr_t) cb, ip->wd_cbdp);
#ifdef	DEBUG
	if (wd_debug > 3) {
		CPRINTF("wd_intr: Printing bp\n");
		wd_dump(sizeof(struct buf), bp);
	}
#endif	/* DEBUG */
	DS_IODONE_TIME(ip->wd_dstats);
	if ((ip->wd_bufh.av_forw) != NULL)	/* Check for requests */
		wdstart(ip);
	if (--ip->wd_inuse == 0 && fp_lights) {
		s1 = splhi();
		FP_IO_INACTIVE;
		splx(s1);
	}
	ds_completion_stats(ip->wd_dstats, bp);
	iodone(bp);
	v_lock(&ip->wd_lock, s);
	return;
}


/*
 * wd_rw - start a read or write disk access command.
 *
 * Called only the first time a request is made on the buffer,
 * retries go through wd_command.
 *
 * Assumes exclusive ownership of cb
 */
static
wd_rw(cb, bp)
	struct scsi_cb	*cb;
	struct buf	*bp;
{
	unchar	command;

	cb->sw.cb_bp = bp;
	cb->sw.cb_errcnt = 0;

	command = (bp->b_flags & B_READ) ? SCSI_READ : SCSI_WRITE;

	wd_command(cb, command, SCB_IENABLE);
}


/* 
 * wd_readc - get the capacity of a SCSI disk
 *
 * boot must already have initialized cb's and 
 * request sense buffers for them.
 */
static
wd_readc(cb)
	struct scsi_cb *cb;
{
	struct screadcap *capdat;
	ulong	nblocks = 0;		/* number of blocks on disk */

	cb->sw.cb_errcnt = 0;
	
	wd_command(cb, SCSI_READC, 0);
	if (wd_cbfin(cb) != SCB_OK)
		return (-1);

	capdat = (struct screadcap *)cb->sw.cb_data;

	/*
	 * Reverse the byte order
	 */

	nblocks = (capdat->rc_nblocks0 << 24) |
		  (capdat->rc_nblocks1 << 16) |
		  (capdat->rc_nblocks2 << 8)  |
		  (capdat->rc_nblocks3) ;

	return (nblocks);
}

/*
 * wd_command - complete and start an SSM SCSI command
 *
 * Commands other than SCSI_READ that need to get data returned from
 * the SCSI device will have said data returned in the request sense
 * buffer. The sc_compcode field must be checked to make sure that
 * the data returned is from the command and not request sense data.
 *
 * The software portion of a SSM SCSI cb must be filled in prior
 * to calling here.
 * 
 * assumes cb is owned exclusively.
 */
static
wd_command(cb, command, flags)
	register struct	scsi_cb	*cb;
	unchar	command;
	unchar	flags;

{
	register struct buf	*bp;
	spl_t	sx;
#ifdef	DEBUG
	if (wd_debug > 3) {
		CPRINTF("wd_command: In wd_command\n");
		CPRINTF("wd_command: command = 0x%x\n", command);
		CPRINTF("wd_command: cb = 0x%x\n", cb);
	}
#endif	/* DEBUG */
	bp = cb->sw.cb_bp;
	bzero((caddr_t) SWBZERO(cb), SWBZERO_SIZE);

	switch (command) {

	case SCSI_TEST:
		cb->sh.cb_count = SCSI_L_TEST;
		cb->sh.cb_cmd = SCB_READ | flags;
		cb->sh.cb_scmd[0] = command;
		cb->sh.cb_scmd[1] |= cb->sw.cb_scsi_device << SCSI_LUNSHFT; 
		cb->sh.cb_clen = SCSI_CMD6SZ;
		break;

	case SCSI_INQUIRY:
		cb->sh.cb_count = INQ_LEN_DEV;
		cb->sh.cb_addr = cb->sw.cb_data;
		cb->sh.cb_cmd = SCB_READ | flags;
		cb->sh.cb_scmd[0] = command;
		cb->sh.cb_scmd[1] |= cb->sw.cb_scsi_device << SCSI_LUNSHFT;
		cb->sh.cb_scmd[4] = INQ_LEN_DEV;
		cb->sh.cb_clen = SCSI_CMD6SZ;
		break;

	case SCSI_READC:
		cb->sh.cb_count = SCSI_L_READC;
		cb->sh.cb_addr = cb->sw.cb_data;
		cb->sh.cb_cmd = SCB_READ | flags;
		/*
 		 * PMI 0
		 * Returns capacity of entire disk
		 */
		cb->sh.cb_scmd[0] = command;
		cb->sh.cb_scmd[1] |= cb->sw.cb_scsi_device << SCSI_LUNSHFT; 
		cb->sh.cb_clen  = SCSI_CMD10SZ;
		break;

	case SCSI_REZERO:
		cb->sh.cb_cmd = SCB_READ | flags;
		cb->sh.cb_scmd[0] = command;
		cb->sh.cb_scmd[1] |= cb->sw.cb_scsi_device << SCSI_LUNSHFT;
		cb->sh.cb_clen = SCSI_CMD6SZ;
		break;

	case SCSI_READ:
	case SCSI_WRITE:
		if ((cb->sh.cb_addr = scb_buf_iovects(bp, cb->sw.cb_iovstart))
		    == NULL) 
			cmn_err(CE_PANIC,
				"wd%d:  wd_command: scb_buf iovects error",
				SCSID_DRIVE(bp->b_dev));	
			/*
			 *+ The wd driver was unable to perform
			 *+ DMA mapping for a requested data
			 *+ transfer.  Either the transfer length
			 *+ or the alignment exceeds the resources for
			 *+ this device.
			 */
		cb->sh.cb_iovec = cb->sw.cb_iovstart;
		cb->sh.cb_count = roundup(bp->b_bcount, (ulong) DEV_BSIZE);
		cb->sh.cb_cmd = (command == SCSI_READ) ? SCB_READ : SCB_WRITE;
		cb->sh.cb_cmd |= flags;
		cb->sh.cb_scmd[0] = command;
		cb->sh.cb_scmd[1] |= cb->sw.cb_scsi_device << SCSI_LUNSHFT;
		cb->sh.cb_scmd[1] |= (unchar) ((int) bp->b_disksortkey >> 16);
		cb->sh.cb_scmd[2] = (unchar) ((int) bp->b_disksortkey >> 8);
		cb->sh.cb_scmd[3] = (unchar) bp->b_disksortkey;
		cb->sh.cb_scmd[4] = (unchar) howmany(bp->b_bcount, DEV_BSIZE);
		cb->sh.cb_clen = SCSI_CMD6SZ;
		break;

	default:
		cmn_err(CE_WARN, "wd%d:  wd_command: unsupported SCSI command %d",
			SCSID_DRIVE(bp->b_dev), command);
		/*
		 *+ The wd driver attempted to generate a SCSI command
		 *+ that it does not support.  This is a driver internal error.
		 */
		return;
	}

#ifdef	DEBUG
	if (wd_debug > 3) {
		CPRINTF("wd_command: About to interrupt ssm, cb =\n");
		wd_printcb(cb);
	}
#endif	/* DEBUG */
	/*
	 * start the SSM SCSI command by interrupting the SSM
 	 */
	sx = splhi();
	mIntr(cb->sw.cb_slic, SCSI_BIN, cb->sw.cb_unit_index);
	splx(sx);
}

/*
 * wdsize
 *	Used for swap-space partition calculation.
 *
 * Called by config
 *
 * Doesn't assume that the device is open or alive at all
 * Hence there are more parameter checks than are really needed.
 * Partitions can be designated as going out to the end of the
 * drive by specifying the length as SCSID_END.  When this is done,
 * the actual drive capacity is used.
 */
wdsize(dev)
	dev_t	dev;
{
	int	drive;
	register wd_t	*ip;

	drive = SCSID_DRIVE(dev);
	ip = wd_info[drive];

	if(ip == NULL				/* set in boot? */
	|| drive >= wd_ndevs			/* Binary conf'd? */
	|| !ip->wd_alive)			/* Passed probeing? */
		return (-1);
		return(ip->wd_capacity);
}

/*
 * wdioctl
 */
/* ARGSUSED */
wdioctl(dp, cmd, addr)
	struct dev *dp;
	int	cmd;
	caddr_t	addr;
{
	wd_t *ip;
	struct scsiioctl sioctl;
	struct scsidev scsidev;
	dev_t dev = dp->d_dev;

	ip = wd_info[SCSID_DRIVE(dev)];

	switch (cmd) {

	case RIOFIRSTSECT: {
		int firstsect = 0;

		if (copyout((caddr_t)&firstsect, addr, sizeof(int)))
			return(EFAULT);
		return(0);
		/*NOTREACHED*/
		break;
	}

	case RIODRIVER:		/* success indicates we're a driver! */
		return(0);

	case SIOCDEVDATA:
#ifdef DEBUG
		WDPRINTF("wdioctl: SIOCDEVDATA\n");
#endif /* DEBUG */
		scsidev.scsi_devno = ip->wd_devno;
		scsidev.scsi_ctlr = ip->wd_ctlr;
		if (copyout((caddr_t)&scsidev, (caddr_t)addr,
			    sizeof(struct scsidev)))
			return(EFAULT);
		return(0);

	case SIOCSCSICMD:

#ifdef DEBUG
		WDPRINTF("wdioctl: SCSICMD\n");
#endif /* DEBUG */
		if (!suser())
			return(EACCES);
		/* Get ioctl args */
		if (copyin((caddr_t)addr, (caddr_t)&sioctl, 
			   sizeof(struct scsiioctl)))
			return(EFAULT);
		/* SCSI commands */
		return(wd_ioctlcmd(dp, ip, &sioctl));
				
	default:
#ifdef DEBUG
		WDPRINTF("wdioctl: default -- EINVAL\n");
#endif /* DEBUG */
		return(EINVAL);			/* not supported yet */
	}
	/*NOTREACHED*/
	return(EINVAL);
}

static
wd_doioctl(cb, bp)
	struct scsi_cb	*cb;
	struct buf	*bp;
{
	spl_t s;

	cb->sw.cb_bp = bp;
	cb->sw.cb_errcnt = 0;

	bzero((caddr_t)SWBZERO(cb), SWBZERO_SIZE);
	if (bp->b_bcount) {
#ifdef DEBUG
		WDPRINTF("wd_doioctl: get scb_buf_iovects\n");
#endif /* DEBUG */
		if ((cb->sh.cb_addr = scb_buf_iovects(bp, cb->sw.cb_iovstart)) 
		     == NULL) {
			cmn_err(CE_PANIC, 
				"wd%d:  wd_doioctl: scb_buf_iovects error",
				SCSID_DRIVE(bp->b_dev));	
			/*
			 *+ The wd driver was unable to perform
			 *+ DMA mapping for a requested data
			 *+ transfer.  Either the transfer length
			 *+ or the alignment exceeds the resources for
			 *+ this device.
			 */
		} else
			cb->sh.cb_iovec = cb->sw.cb_iovstart;
	}
	cb->sh.cb_count = bp->b_bcount;
	cb->sh.cb_cmd = ((bp->b_flags & B_READ) ? SCB_READ : SCB_WRITE) 
			| SCB_IENABLE;
	cb->sh.cb_clen = SCSI_IOCTLLEN(bp);		
#ifdef DEBUG
	WDPRINTF("wd_doioctl: cmdlen = %d, cmd = 0x%x\n",
		  cb->sh.cb_clen, cb->sh.cb_cmd);
#endif /* DEBUG */
	bcopy((caddr_t)(&((struct scsiioctl *)bp->b_sioctl)->sio_cmd),
	      (caddr_t)cb->sh.cb_scmd, cb->sh.cb_clen);	

	s = splhi();
	mIntr(cb->sw.cb_slic, SCSI_BIN, cb->sw.cb_unit_index);
	splx(s);
}

static
wd_ioctlcmd(dp, ip, sio)
	struct dev *dp;
	wd_t *ip;
	struct scsiioctl *sio;
{
	struct buf *bp;			/* ioctl buffer */
	int dir = 0;			/* zero == READ */
	int error;
	spl_t s;

	bp = &ip->wd_rbufh;

	/* make sure command is for the current device! */
	if (SCSI_UNIT(sio->sio_cmd6.cmd_lun >> SCSI_LUNSHFT)
	     != SCSI_UNIT(ip->wd_devno)) {
#ifdef DEBUG
		WDPRINTF("wd_ioctlcmd: bad lun 0x%x for 0x%x, cmd_lun 0x%x\n",
			 SCSI_UNIT(sio->sio_cmd6.cmd_lun >> SCSI_LUNSHFT),
			 SCSI_UNIT(ip->wd_devno), sio->sio_cmd6.cmd_lun);
#endif /* DEBUG */
		return(EINVAL);
	}
	if ((int)sio->sio_datalength < 0) {		/* invalid parm */
#ifdef DEBUG
		WDPRINTF("wd_ioctlcmd: negative data length\n");
#endif /* DEBUG */
		return(EINVAL);
	}

	switch(sio->sio_cmd6.cmd_opcode) {

	case SCSI_FORMAT:
#ifdef DEBUG
		WDPRINTF("wd_ioctlcmd: SCSI_FORMAT\n");
#endif /* DEBUG */
		sio->sio_cmdlen = SCSI_CMD6SZ;
		break;

	case SCSI_INQUIRY:
#ifdef DEBUG
		WDPRINTF("wd_ioctlcmd: SCSI_INQUIRY\n");
#endif /* DEBUG */
		sio->sio_cmdlen = SCSI_CMD6SZ;
		break;

	case SCSI_READC:
#ifdef DEBUG
		WDPRINTF("wd_ioctlcmd: SCSI_READC\n");
#endif /* DEBUG */
		sio->sio_cmdlen = SCSI_CMD10SZ;
		break;

	case SCSI_REASS:
#ifdef DEBUG
		WDPRINTF("wd_ioctlcmd: SCSI_REASS\n");
#endif /* DEBUG */
		sio->sio_cmdlen = SCSI_CMD6SZ;
		dir = 1;
		break;

	case SCSI_READ_DEFECTS:
#ifdef DEBUG
		WDPRINTF("wd_ioctlcmd: SCSI_READ_DEFECTS\n");
#endif /* DEBUG */
		sio->sio_cmdlen = SCSI_CMD10SZ;
		break;

	default:
#ifdef DEBUG
		WDPRINTF("wd_ioctlcmd: default - not supported yet\n");
#endif /* DEBUG */
		return(EINVAL);

	} /* end switch */

#ifdef DEBUG
	WDPRINTF("addr = 0x%x, datalength = %d\n", sio->sio_addr,
		  sio->sio_datalength);
#endif /* DEBUG */
	/*
	 * Make sure user has access to specified data buffer
	 */
	if (sio->sio_datalength)
		if (!useracc((char *)sio->sio_addr,
			     (uint)sio->sio_datalength,
			     ((dir) ? B_WRITE : B_READ))) {
#ifdef DEBUG
			WDPRINTF("wd_ioctl: no user access\n");
#endif /* DEBUG */
			return(EFAULT);
		}

	/*
	 * Allocate buffer and set fields
	 */
	bufalloc(bp);
	bp->b_flags = ((dir) ? B_WRITE : B_READ) | B_IOCTL;
	bp->b_bcount = sio->sio_datalength;
	bp->b_un.b_addr = (caddr_t)sio->sio_addr;
	bp->b_sioctl = (long)sio;
	bp->b_devp = dp;
	bp->b_blkno = 0;
	bp->b_error = 0;
	bp->b_proc = u.u_procp;
	bp->b_iotype = B_RAWIO;
	BIODONE(bp) = 0;
	/*
	 * Lock process and specified pages in memory
	 */
#ifdef DEBUG
	WDPRINTF("wd_ioctl: locking proc and pages\n");
#endif /* DEBUG */
	++u.u_procp->p_noswap;
	if (bp->b_bcount)
		vslock(bp->b_un.b_addr, (int)bp->b_bcount,
			(bool_t)(bp->b_flags & B_READ));

	/*
	 * Insert request at front of queue
	 */
	s = p_lock(&ip->wd_lock, WDSPL);
#ifdef DEBUG
	WDPRINTF("wd_ioctl: insert in queue\n");
#endif /* DEBUG */
	bp->av_forw = ip->wd_bufh.av_forw;
	ip->wd_bufh.av_forw = bp;
	DS_QUEUED_TIME(bp);
	wdstart(ip);
	DS_Q(ip->wd_dstats);
	v_lock(&ip->wd_lock, s);

#ifdef DEBUG
	WDPRINTF("wd_ioctl: wait\n");
#endif /* DEBUG */
	biowait(bp);
	--u.u_procp->p_noswap;
	error = geterror(bp);
	buffree(bp);
#ifdef DEBUG
	WDPRINTF("wd_ioctl: return %d\n", error);
#endif /* DEBUG */
	return(error);
}
		
/* 
 * wd_cbfin 
 * polls for SSM command completion and returns completion status.
 */
static
wd_cbfin(cb)
	struct scsi_cb	*cb;
{
	int error = 0;
	struct scrsense *rsense;

#ifdef	DEBUG
	if (wd_debug > 3) {
		CPRINTF("wd_cbfin: In wd_cbfin\n");
	}
#endif	/* DEBUG */
	while (error++ < 4) {
		/*
	 	 * Firmware will timeout if commands take to long
		 */
#ifdef	DEBUG
		if (wd_debug > 3) {
			CPRINTF("wd_cbfin: Waiting for compcode\n");
		}
#endif	/* DEBUG */
		while (cb->sh.cb_compcode == SCB_BUSY)
			;
#ifdef	DEBUG
		if (wd_debug > 3) {
			CPRINTF("wd_cbfin: Got compcode\n");
		}
#endif	/* DEBUG */
		switch(cb->sh.cb_compcode) {
		case SCB_BAD_CB:
		case SCB_NO_TARGET:
		case SCB_SCSI_ERR:
			 goto restart ;
		
		case SCB_OK:
			if (SCSI_CHECK_CONDITION(cb->sh.cb_status)) {
				rsense = (struct scrsense *) cb->sh.cb_sense;
				if (SCSI_RS_ERR_CLASS(rsense) != RS_CLASS_EXTEND
				|| SCSI_RS_ERR_CODE(rsense) != RS_CODE_EXTEND) 
					goto restart ;
				else if ((rsense->rs_key & RS_SENSEKEYMASK)
					 == RS_RECERR)
					return SCB_OK;
				else
					goto restart ;
			} else			/* No check condition */
				return SCB_OK;

		default:
			goto restart ;
	
		}
restart:		wd_command(cb, cb->sh.cb_scmd[0], 0);
	}

	switch(cb->sh.cb_compcode) {
	case SCB_BAD_CB:
	case SCB_NO_TARGET:
	case SCB_SCSI_ERR:
		 return -1 ;
	
	case SCB_OK:
		 if (SCSI_CHECK_CONDITION(cb->sh.cb_status)) {
			rsense = (struct scrsense *) cb->sh.cb_sense;
			if (SCSI_RS_ERR_CLASS(rsense) != RS_CLASS_EXTEND
			||  SCSI_RS_ERR_CODE(rsense) != RS_CODE_EXTEND) {
				return -1 ;
			} 
       	 	 	if ((rsense->rs_key & RS_SENSEKEYMASK) == RS_RECERR) 
				return SCB_OK;

		return -1 ;
		}

		if (SCSI_CHECK_CONDITION(cb->sh.cb_status)) {
			rsense = (struct scrsense *) cb->sh.cb_sense;
			if (SCSI_RS_ERR_CLASS(rsense) != RS_CLASS_EXTEND
			||  SCSI_RS_ERR_CODE(rsense) != RS_CODE_EXTEND) 
				return -1;
			else if ((rsense->rs_key & RS_SENSEKEYMASK)==RS_RECERR)
				return SCB_OK;
			else
				return -1;
		} else			/* No check condition */
			return SCB_OK;

	default:
		return -1 ;
	}
	
}

/* 
 * wdp_cbfin 
 * polls for SSM command completion and retries upto four times.
 * This is used during probe of devices and does not attempt to
 * interpret errors.
 */
static
wdp_cbfin(cb)
	struct scsi_cb	*cb;
{

	/*
 	 * Firmware will timeout if commands take to long
	 */
	while (cb->sh.cb_compcode == SCB_BUSY)
		;

	if (cb->sh.cb_compcode == SCB_OK && 
	    !SCSI_CHECK_CONDITION(cb->sh.cb_status))
		return (cb->sh.cb_compcode);
	else
		return (cb->sh.cb_compcode);
}

/*
 * wdp_command - start a probe command
 *
 * neccesary because cb's are not initialized yet
 */
static
wdp_command(cb, cmd, sense, buffer, devno)
	struct scsi_cb	*cb;
	unchar	cmd;
	daddr_t	sense;
	daddr_t	buffer;
	short	devno;
{
	cb->sh.cb_sense = sense;
	cb->sh.cb_slen = SIZE_MAXDATA;
	cb->sw.cb_data = buffer;
	cb->sw.cb_scsi_device = devno;
	wd_command(cb, cmd, 0);
	return (wdp_cbfin(cb));
}

/*
 * wd_printcb
 *
 * prints out the fields of a command block
 */
static
wd_printcb(cb)
	struct scsi_cb *cb;
{
	int x;

	cmn_err(CE_CONT, "command: ");
#ifdef	DEBUG
	if (wd_debug > 3) {
		CPRINTF("0x%x 0x%x 0x%x \n", cb->sh.cb_cmd,
			cb->sh.cb_reserved0[0],
			cb->sh.cb_clen); 
	}
#endif	/* DEBUG */
	for (x = 0; x < cb->sh.cb_clen; x++) 
		cmn_err(CE_CONT, "0x%x ", cb->sh.cb_scmd[x]);
	cmn_err(CE_CONT, "\n");
#ifdef DEBUG
	if (wd_debug > 3) {
		CPRINTF("0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x\n",
			*(ulong*) &cb->sh.cb_reserved1[0],	 
			cb->sh.cb_addr,		
			(ulong) cb->sh.cb_iovec,
			cb->sh.cb_count,
			cb->sh.cb_status, 
			*(ushort*) &cb->sh.cb_reserved2[0],
			cb->sh.cb_compcode);
		CPRINTF("0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x\n", 
			(ulong) cb->sw.cb_bp, 
			(ulong) cb->sw.cb_iovstart,
			cb->sw.cb_errcnt,
			cb->sw.cb_unit_index,
			cb->sw.cb_scsi_device,
			cb->sw.cb_state,
			cb->sw.cb_data,
			cb->sw.cb_iovnum,
			cb->sw.cb_slic); 
	}
#endif /* DEBUG */
}

/*
 * wd_printsense - prints sense information returned from a
 * 	SCSI request sense command
 */
static
wd_printsense(rsense)
	struct scrsense *rsense;
{
	int more;
	unchar *moredat;
	long lba, vend;

	cmn_err(CE_CONT, "Sense: ");
	if (SCSI_RS_ERR_CLASS(rsense) != RS_CLASS_EXTEND) {
		/* 
		 * Report information for Non-extended format.
		 */
		moredat = &rsense->rs_seg;
		vend = *moredat >> 5 & 7;
		if (rsense->rs_error & RS_VALID) {
			lba = ((long)*moredat++ & 0x1f) << 16;
			lba |= (long)*moredat++ << 8;
			lba |= (long)*moredat;
			cmn_err(CE_CONT, 
				"Error 0x%x, Vendor Unique 0x%x, LBA 0x%x ",
				rsense->rs_error, vend, lba);
		} else {
			cmn_err(CE_CONT, "Error 0x%x, Vendor Unique 0x%x ",
				rsense->rs_error, vend);
		}
		return;
	} 

	/* 
	 * Report information for extended format.
	 */
	if (SCSI_RS_ERR_CODE(rsense) != RS_CODE_EXTEND) 
		cmn_err(CE_CONT, "Vendor unique extended sense format\n\t");
	
	cmn_err(CE_CONT, "Error 0x%x, Seg Num 0x%x, Key 0x%x, ",
		rsense->rs_error, rsense->rs_seg, rsense->rs_key);
	if (rsense->rs_error & RS_VALID) {
		lba = (long)rsense->rs_info[0] << 24;
		lba |= (long)rsense->rs_info[1] << 16;
		lba |= (long)rsense->rs_info[2] << 8;
		lba |= (long)rsense->rs_info[3];
		cmn_err(CE_CONT, "Info 0x%x, ", lba);
	}

	more = (rsense->rs_addlen > SIZE_MAXDATA - sizeof (struct scrsense)) ?
		SIZE_MAXDATA - sizeof(struct scrsense) : (int)rsense->rs_addlen;
	if (more) {
		cmn_err(CE_CONT, "Addlen 0x%x:\n\t", rsense->rs_addlen);
		for (moredat = &rsense->rs_addlen + 1; more--; )
			cmn_err(CE_CONT, "0x%x ", *moredat++);
		cmn_err(CE_CONT, "\n");
	} else
		cmn_err(CE_CONT, "Addlen  0x%x\n", rsense->rs_addlen);
}

/*
 * wd_shutdrive
 * Fail all I/O requests queued to the drive.
 * Currently active I/O requests will finish and fail asyncronously.
 * The drive is marked as WD_BAD
 */
static
wd_shutdrive(ip)
	register wd_t *ip;
{
	register struct buf *bp, *nextbp;
	struct dev *dp;
	int dev;
	spl_t s;

	if (ip->wd_flags & WD_BAD)
		return;

	bp = ip->wd_bufh.av_forw;
	while (bp != NULL) {
		bp->b_flags |= B_ERROR;
		bp->b_error = EIO;
		bp->b_resid = bp->b_bcount;
		nextbp = bp->av_forw;
		bp->b_status |= B_S_UNIT_ERR | B_S_XFER_ERR | B_S_PERM_ERR;
#ifdef	DEBUG
		if (wd_debug > 3) {
			CPRINTF("wd_shutdrive: Printing bp\n");
			wd_dump(sizeof(struct buf), bp);
		}
#endif	/* DEBUG */
		DS_IODONE_TIME(ip->wd_dstats);
		ds_completion_stats(ip->wd_dstats, bp);
		iodone(bp);
		bp = nextbp;
	}

	ip->wd_bufh.av_forw = NULL;
	ip->wd_flags |= WD_BAD;
	for (dev=0; dev < wd_ndevs; dev++)
		/* Find our dev based on our info pointer */
		if (wd_info[dev] == ip) {
			dp = wd_dev + dev;
			break;
		}
	if ( dp && dp->d_parent )
		DEV_SHUTDOWN(dp->d_parent, dp);
	if (fp_lights) {
		s = splhi();
		if (ip->wd_inuse)
			FP_IO_INACTIVE;
		FP_IO_ERROR;
		splx(s);
	}
	disk_offline();
}

/*
 * wd_error - disk access error reporting
 * 	Block number reported is taken from 
 *	request sense data when available.
 * 	Otherwise, taken from bp->b_disksortkey, 
 *	since bp->b_blkno may not have received 
 *	offsetting from upper-level drivers.
 */
static
void
wd_error(cb, bp, str)
	struct scsi_cb *cb;
	struct buf *bp;
	char *str;
{
	struct scrsense *rsense;
	char	*command, *key, *code, *completion;
	long lba;

	rsense = (struct scrsense *) cb->sh.cb_sense;	

	command = (bp->b_flags & B_READ) ? "READ" : "WRITE";

	if (cb->sh.cb_compcode >= wdncompcodes)
		completion = wd_compcode[wdncompcodes - 1];
	else
		completion = wd_compcode[cb->sh.cb_compcode];

	if (rsense->rs_key & RS_SENSEKEYMASK) {
		key = scsi_errors[rsense->rs_key & RS_SENSEKEYMASK];

		if (((caddr_t) cb->sh.cb_sense)[12] >= wdnaddsense)
			code = wd_addsense[wdnaddsense -1];
		else
			code = wd_addsense[((caddr_t) cb->sh.cb_sense)[12]];

		/* 
		 * If the sense data is valid then retrieve the
		 * logical block address in question from the
		 * returned sense data.  Otherwise use the
		 * starting block address from the buf-structure.
		 */
		if (rsense->rs_error & RS_VALID) {
			lba = (long)rsense->rs_info[0] << 24;
			lba |= (long)rsense->rs_info[1] << 16;
			lba |= (long)rsense->rs_info[2] << 8;
			lba |= (long)rsense->rs_info[3];
		} else
			lba = bp->b_disksortkey;

		cmn_err(CE_WARN, "wd%d: %s %s, \n\t%s, %s, lba=%d, %s\n",
			SCSID_DRIVE(bp->b_dev), str, command, key, code, lba,
			completion);
		/*
		 *+ The wd driver is reporting information about a
		 *+ hard or soft disk error that occurred.  The 
		 *+ information includes an error type, the SCSI
		 *+ command that caused it, the reported sense data
		 *+ key and code, the logical block address being
		 *+ referenced, and its completion status.
		 */
	} else {
		cmn_err(CE_CONT, "wd%d: %s %s, lba=%d, %s\n",
			SCSID_DRIVE(bp->b_dev), str, command, bp->b_disksortkey,
			completion);
	}
		
	
	cmn_err(CE_CONT, "wd%d: ", SCSID_DRIVE(bp->b_dev));
	wd_printcb(cb);

	if (rsense->rs_key & RS_SENSEKEYMASK) {
		cmn_err(CE_CONT, "wd%d: ", SCSID_DRIVE(bp->b_dev));
		wd_printsense(rsense);
	}
}

#ifdef	DEBUG
wd_dump(len, ptr)
	int len;
	char * ptr;
{
	int	x;
	
	for (x = 0; x < len; x++) {
		CPRINTF("%x ", (unchar) ptr[x]);
		if ((x % 25 == 0) && (x != 0))
			CPRINTF("\n");
	}
	CPRINTF("\n");
}
#endif	/* DEBUG */

/*
 * print information.
 *	not complete yet.  Does not print much info
 */
wdprint (dp,str)
	struct dev *dp;
	char *str;
{
	cmn_err(CE_CONT, "%s on scsi disk %d\n", str, SCSID_DRIVE(dp->d_dev));
}

/*
 * wdfirstsect
 *
 *	return the first usable sector of the device.  Always 0 on an sd.
 */
/* ARGSUSED */
int
wdfirstsect(dp)
	struct	dev *dp;
{
	return(0);
}
