/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ident	"$Header: zd.c 1.67 90/10/22 $"

/*
 * zd.c
 *
 * ZDC SMD-E Disk Driver
 *
 * NOTE:
 * This driver, although apparently separate, is heavily dependent on the zdc
 * driver. This is so much tha case that there are functions in zdinit.c that
 * it could be argued belong here and vica versa.
 */

/* $Log:	zd.c,v $
 */

#include "sys/types.h"
#include "sys/devsw.h"
#include "sys/sysmacros.h"
#include "sys/signal.h"
#include "sys/param.h"
#include "sys/debug.h"
#include "sys/mutex.h"
#include "sys/file.h"
#include "sys/user.h"
#include "sys/errno.h"
#include "sys/buf.h"
#include "sys/systm.h"
#include "sys/uio.h"
#include "sys/ioctl.h"
#include "sys/timer.h"
#include "sys/proc.h"
#include "sys/stats_disk.h"
#include "sys/vm.h"
#include "sys/cmn_err.h"
#include "sys/vnode.h"

#include "sys/slicreg.h"
#include "sys/clkarb.h"

#include "sys/intctl.h"
#include "sys/pte.h"
#include "sys/mftpr.h"
#include "sys/plocal.h"

#include "sys/zdc.h"
#include "sys/zdbad.h"
#include "sys/zdc_ioconf.h"
#include "sys/zdioctl.h"
#include "sys/zd.h"
#include "sys/zd_device.h"

#define	ZDUNIT(dev)	(minortounit(minor(dev)))
#define	ZDPART(dev)	(minortopart(minor(dev)))
#define PARTCHR(dev)	(ZDPART(dev))
#define	PARTSTRLEN	5
#define	ZDBLKNO(daddr, dd)	\
	(daddr.da_cyl * dd->zdd_tracks * dd->zdd_sectors +  \
	 daddr.da_head * dd->zdd_sectors +  \
	 daddr.da_sect)
#define	NUDGE_ZDC(ctlp, cbp, s)	{ \
	(s) = splhi(); \
	mIntr((ctlp)->zdc_slicaddr, CBBIN, (unchar)((cbp) - (ctlp)->zdc_cbp)); \
	splx((s)); \
	}
#define	PZOPEN		(PZERO - 1)		/* not signallable */
#define	b_diskaddr	b_disksortkey
#define	b_psect		b_error
/*
 * PTECLOFF returns offset in cluster of memory pointed at by pte.
 */
#define	PTECLOFF(pte)	((unsigned)((*(int*)(&(pte))) & ((~(NBPG-1))&CLOFSET)))

extern	struct	zdc_ctlr *zdctrlr;	/* zdctrlr array */
extern	struct	zd_unit	 *zdunit;	/* zdunit array */
extern	int	zdc_iovpercb;		/* no of iovecs per cb */
extern	int	zdc_AB_throttle;	/* Channel A&B DMA throttle */
extern	short	zdcretry;		/* retry count on errors */
extern	unchar	zdctrl;			/* additional icb_ctrl bits */
extern	unchar	base_cb_intr;		/* base interrupt for zdc driver */
extern	unchar	base_err_intr;		/* base controller interrupt */
extern	lock_t	zdcprlock;		/* coordinate error printfs */
extern zd_max_unit;			  /* set at probe time. */
extern struct zd_device zd_devices[];

#ifdef	DEBUG
int	zddebug = 0;
static struct zd_unit *zd_intr_shutdown = (struct zd_unit *)0;
static ulong zd_status_returns = 0;
static ulong zd_express = 0;
int zd_debug = 0;		/* online format == 1, patrol seeks == 2 */ 
#define ZDPRINTF	if (zd_debug == 1) CPRINTF
#define ZDPRINTF2	if (zd_debug == 2) CPRINTF
#endif	/* DEBUG  */

caddr_t	zd_compcodes[] = {
	"Command in progress",
	"Successful completion",
	"Write protect fault",
	/*
	 *+ The driver attempted to write to a write-protected disk.
	 */
	"Drive Fault",
	/*
	 *+ A drive fault was reported from the indicated disk drive.
	 *+ Frequent occurrences indicate a suspect disk drive.
	 *+ The driver will issue a ZDC_RESET command and retry the
	 *+ request 'zdretry' times.  If the failure persists,
	 *+ the drive is marked as bad and will no longer be
	 *+ accessible through the driver.
	 */
	"Seek error",
	/*
	 *+ A seek error was reported from the disk drive.
	 *+ Frequent occurrences indicate a suspect disk drive.
	 *+ The driver will issue a ZDC_RESET command and retry the
	 *+ request 'zdretry' times.  If the failure persists,
	 *+ the drive is marked as bad and will no longer be
	 *+ accessible through the driver.
	 */
	"Seek timeout",
	/*
	 *+ A seek was started and did not complete.  This can
	 *+ indicate a problem with either the disk drive or the
	 *+ DCC hardware.
	 *+ The driver will issue a ZDC_RESET command and retry the
	 *+ request 'zdretry' times.  If the failure persists,
	 *+ the drive is marked as bad and will no longer be
	 *+ accessible through the driver.
	 */
	"Channel timeout",
	/*
	 *+ A channel operation was started for which an interrupt
	 *+ was not received within 100 milliseconds.  This indicates
	 *+ that this DCC channel hardware is suspect.
	 *+ The driver will retry the request 'zdretry' times.  
	 *+ If the failure persists, the I/O request is failed
	 *+ and the driver shuts down the channel.  This means
	 *+ that all drives on the channel are marked as bad and
	 *+ will no longer be accessible through the driver.
	 */
	"DMA timeout",
	/*
	 *+ A DCC DMA transfer was started for which an interrupt was
	 *+ not received within 100 milliseconds.  This indicates
	 *+ suspect DMA hardware on the DCC.  The driver logs the
	 *+ failure and fails the I/O request.  The driver then
	 *+ proceeds to shut down the channel.  This means that
	 *+ all drives on the channel are marked as bad and will
	 *+ no longer be accessible through the driver.
	 */
	"Header ECC error",
	/*
	 *+ A CRC error in the header of the sector was detected.
	 *+ Corrective action:  if this sector on this disk 
	 *+ continues to receive such
	 *+ errors, add the sector to the bad block
	 *+ list (use the addbad option of the 'format' program
	 *+ to do this).
	 *+ The driver will issue a ZDC_RESET command and retry the
	 *+ request 'zdretry' times.  If the failure persists,
	 *+ the I/O request is failed. 
	 */
	"Soft ECC error",
	/*
	 *+ An ECC error in the data portion of the sector was
	 *+ detected and successfully read (via retries) without making a
	 *+ correction.  Corrective action:  if this sector on this disk continues to
	 *+ receive such errors, add the sector to the
	 *+ bad block list (use the addbad option of the 'format'
	 *+ program to do this).  The driver will complete this
	 *+ I/O request successfully.
	 */
	"Correctable ECC error",
	/*
	 *+ An ECC error in the data portion of the sector was
	 *+ detected and successfully corrected.  Corrective action:  if this sector
	 *+ on this disk continues to receive such errors, add the
	 *+ sector to the bad block list (use the
	 *+ addbad option of the 'format' program to do this).  The driver
	 *+ will complete this I/O request successfully.
	 */
	"Uncorrectable ECC error",
	/*
	 *+ An ECC error in the data portion of the sector was
	 *+ detected and could not be corrected.  The sector
	 *+ is corrupted and should be
	 *+ added to the bad block list (use the addbad option of
	 *+ the 'format' program to do this).  The driver will fail this I/O
	 *+ request.
	 */
	"Sector not found",
	/*
	 *+ The DCC was unable to locate a
	 *+ sector with header bytes matching the address
	 *+ indicated in the error message.  This can be caused by
	 *+ either a bad header or a seek to the wrong cylinder.
	 *+ Corrective action:
	 *+ if this sector on this disk continues to receive such
	 *+ errors, add the sector to the bad block list
	 *+ (use the addbad option of the 'format' program to do this).  The
	 *+ type given to addbad should be an error in the
	 *+ header.  The driver will issue a ZDC_RESET command
	 *+ and retry the request 'zdretry' times.  If the failure
	 *+ persists, the I/O request is failed.
	 */
	"Bad data sector",
	/*
	 *+ DCC auto-revectoring was inhibited.
	 *+ Because this shouldn't be done by the driver, the DCC
	 *+ microcode is suspect.  The driver will proceed to
	 *+ shut down the controller, which means that all drives
	 *+ on the controller will become inaccessible.
	 */
	"Sector overrun",
	/*
	 *+ The DCC unexpectedly encountered
	 *+ a sector pulse.  Corrective action:  if this sector 
	 *+ on this disk continues
	 *+ to receive such errors, add the sector to
	 *+ the bad block list (use the addbad option of the 'format'
	 *+ program to do this).  The type given to addbad should be an
	 *+ error in the header.  The driver will issue a ZDC_RESET
	 *+ command and retry the request 'zdretry' times.
	 *+ If the failure persists, the I/O request is failed.
	 */
	"No data synch",
	/*
	 *+ The indicated sector on the disk is suspect.  
	 *+ Corrective action:  if this
	 *+ sector on this disk continues to receive such errors,
	 *+ add the sector to the bad block list (use
	 *+ the addbad option of the 'format' program to do this).  The
	 *+ error type given to addbad should be an error in
	 *+ the data.  The driver will issue a ZDC_RESET command
	 *+ and retry the request 'zdretry' times.  If the failure
	 *+ persists, the I/O request is failed.
	 */
	"Fifo data lost",
	/*
	 *+ Repeated occurrences of this error indicate a suspect
	 *+ channel on the DCC.  The driver will retry the request 'zdretry' 
	 *+ times.  If the failure persists, the I/O
	 *+ request is failed and driver shuts down the channel.
	 *+ This means that all drives on the channel are marked 
	 *+ as bad and will be inaccessible through the driver.
	 */
	"Illegal cb_cmd",
	/*
	 *+ The DCC received an illegal
	 *+ command from the driver.  If the command block
	 *+ contains a valid command, the driver assumes that the DCC
	 *+ is unreliable and shuts down the controller,
	 *+ making all drives on the controller inaccessible.
	 *+ If the command block contains an invalid command,
	 *+ the driver panics the system.
	 */
	"Illegal cb_mod",
	/*
	 *+ The DCC received an illegal
	 *+ command modifier from the driver.  If the command
	 *+ block contains a valid command modifier, the driver
	 *+ assumes that the DCC is unreliable and shuts down the
	 *+ controller, making all drives on the controller
	 *+ inaccessible.  If the command block contains an
	 *+ invalid command modifier, the driver panics the
	 *+ system.
	 */
	"Illegal disk address",
	/*
	 *+ The DCC received an illegal
	 *+ disk address from the driver.  If the command
	 *+ block contains a valid disk address, the driver
	 *+ assumes that the DCC is unreliable and shuts down the
	 *+ controller, making all drives on the controller
	 *+ inaccessible.  If the command block contains an
	 *+ invalid disk address, the driver panics the system.
	 */
	"cb_addr not 16-byte aligned",
	/*
	 *+ The DCC received an address
	 *+ that was not 16-byte aligned from the driver.
	 *+ If the address in the command block is 16-byte
	 *+ aligned, the driver assumes that the DCC is unreliable
	 *+ and shuts down the controller, making all
	 *+ drives on the controller inaccessible.  If the
	 *+ command block contains an address that is not
	 *+ 16-byte aligned, the driver panics the system.
	 */
	"Illegal cb_count",
	/*
	 *+ The DCC received an illegal
	 *+ count request from the driver.  If the command
	 *+ block field 'cb_count' is nonzero and is a multiple
	 *+ of 16, the driver assumes that the DCC is unreliable and 
	 *+ shuts down the controller, making all drives on 
	 *+ the controller inaccessible.  If the command block 
	 *+ contains an invalid count, the driver panics the
	 *+ system.
	 */
	"cb_iovec not 32-byte aligned",
	/*
	 *+ The DCC received an I/O vector
	 *+ address that was not 32-byte aligned.  If 
	 *+ the command block field 'cb_iovec' is 32-byte
	 *+ aligned, the driver assumes that the DCC is unreliable and 
	 *+ shuts down the controller, making all drives 
	 *+ on the controller inaccessible.  If the 'cb_iovec'
	 *+ field is not 32-byte aligned, the driver panics the 
	 *+ system.
	 */
	"Non-zero cb_iovec and page size invalid",
	/*
	 *+ The DCC microcode was not
	 *+ initialized with the system page size.  Because the
	 *+ driver always initializes the DCC with the system
	 *+ page size, the DCC is suspect.  The driver will
	 *+ shut down the controller, making all drives on that
	 *+ controller inaccessible.
	 */
	"Illegal icb_pagesize",
	/*
	 *+ The DCC was initialized with an illegal
	 *+ page size.  The DCC is suspect.  The driver will
	 *+ shut down the controller, making all drives on that
	 *+ controller inaccessible.
	 */
	"icb_dumpaddr not 16-byte aligned",
	/*
	 *+ This message can occur only during initialization.  The
	 *+ DCC is suspect.  The driver will shut down the
	 *+ controller, making all drives on that controller
	 *+ inaccessible.
	 */
	"Bad drive",
	/*
	 *+ The disk drive is not online
	 *+ and formatted to match the channel format.  
	 *+ Typically, the drive went offline
	 *+ unexpectedly.  The drive is marked as bad and becomes
	 *+ inaccessible.  The current I/O request and all
	 *+ queued requests are failed.
	 */
	"In-use CB reused",
	/*
	 *+ The DCC received a request on an active command block.
	 *+ This indicates suspect software or DCC microcode.
	 *+ The driver immediately panics the system.
	 */
	"Access error during DMA",
	/*
	 *+ This error can be caused by either a DCC software/microcode
	 *+ bug or a hardware failure.  If the failure is due
	 *+ to an uncorrectable memory error, the driver panics
	 *+ the system.  Otherwise, the driver fails the current
	 *+ I/O request, marks all drives on the DCC channel
	 *+ as bad, and fails all I/O requests to those drives.
	 *+ The driver then clears the DCC's access error register
	 *+ to notify the DCC microcode that the access error
	 *+ has been cleared.
	 */
	"Channel not configured",
	"Channel was reset",
	/*
	 *+ This indicates a suspect disk drive.  The current
	 *+ I/O request is failed, the drive is marked as bad, and
	 *+ all I/O requests to those drives are failed.
	 */
	"Unexpected status from DDC",
	/*
	 *+ An unrecognized completion status was received
	 *+ from a drive.  This indicates a suspect disk drive.
	 *+ The driver marks the drive as bad and fails all
	 *+ I/O requests to that drive, including the current
	 *+ request.
	 */
	"Unknown Completion code"
	/*
	 *+ The driver received an unrecognized completion
	 *+ code from the DCC.  The driver will shutdown the
	 *+ controller, making all drives on that controller
	 *+ inaccessible.
	 */
};
int	zdncompcodes = sizeof(zd_compcodes) / sizeof(zd_compcodes[0]);

/*
 * Variables associated with patrol seeks.
 */
static lock_t	patrollock;		/* lock for patrol_on, patrolcount */
					/* can be nested in zu_lock locks */ 
static bool_t	patrol_on = 0;		/* 1 : zdpatrol timeout is scheduled */
static int	patrolcount = 0;	/* count of devices to be patrolled */
char	*zdpatrolbuf = (char *)NULL;	/* read buffer for patrols */
static int 	zdpatrol();		/* routine that does the patrols */

/*
 * zdopen
 *
 * Return:
 *	0 - success.
 *	EACCES - open for write but drive write-protected.
 *	ENXIO - all other failures.
 */
zdopen(dp, mode)
	struct	dev *dp;
	int	mode;
{
	register struct zd_unit *up;
	register struct zdcdd	*dd;		/* channel configuration */
	dev_t	dev = dp->d_dev;
	int	retval;				/* return value */
	unchar	oldcfg;				/* previous cfg state */
	int size;
	struct	zdcdd	*zdget_chancfg();
	struct zdc_ctlr *ctlrp;
	int chan;
	spl_t   s_ipl;

	up = &zdunit[ZDUNIT(dev)];

#ifdef DEBUG 
	ZDPRINTF("zdopen: dev = 0x%x\n", dev);
	if (zddebug)
		CPRINTF("O");
#endif	/* DEBUG  */

	/*
	 * Fail open if the unit is bad or if the unit was not bound
	 * to a drive during configuration.  Check of ZDUNIT is just
	 * sanity, since it no devs should be present which can get
	 * to zdopen without a valid dev_t.
	 */
	if (ZDUNIT(dev) > zd_max_unit ||
	    up->zu_state == ZU_NOTFOUND || up->zu_state == ZU_BAD)
		return (ENXIO);

	/*
	 * Now get serious
	 */
	(void)p_sema(&up->zu_ocsema, PZOPEN);

	(void)p_lock(&up->zu_lock, SPLZD);
	/*
	 * check to see if still good drive.
	 */
	if (up->zu_state == ZU_BAD) {
		v_lock(&up->zu_lock, SPL0);
		v_sema(&up->zu_ocsema);
		return (ENXIO);
	}

        /*
         * Verify that a the channel is not owned.
         * If so, error out.
         */
        chan = up->zu_drive & 1;
        ctlrp = &zdctrlr[up->zu_ctrlr];
        s_ipl = p_lock(&ctlrp->zdc_ctlrlock, SPLHI);
        if (ctlrp->zdc_chan_owner[chan]) {
                v_lock(&ctlrp->zdc_ctlrlock, s_ipl);
                v_lock(&up->zu_lock, SPL0);
                v_sema(&up->zu_ocsema);
                return (EACCES);
        }

	if (up->zu_nopen == ZU_MAX_OPENS) {
		/* Don't allow the open counter to overflow. */
                v_lock(&ctlrp->zdc_ctlrlock, s_ipl);
                v_lock(&up->zu_lock, SPL0);
                v_sema(&up->zu_ocsema);
                return (EBUSY);
	}

	up->zu_nopen++;
        v_lock(&ctlrp->zdc_ctlrlock, s_ipl);
	retval = 0;
	if (up->zu_nopen > 1) {
		/*
		 * Already opened at least once.
		 */
		if ((mode & FEXCL) || (up->zu_flags & ZUF_EXCLUSIVE)) {
			/*
			 * Trying to open exclusively when device already
			 * open, or trying to open when device open
			 * exclusively!
			 */
			 retval = EBUSY;
		} else if (up->zu_state == ZU_NO_RW) {
			/*
			 * Only one formatter at a time please...
			 */
			retval = EACCES;
		} else if ((mode & FWRITE) && (up->zu_cfg & ZD_READONLY)) {
			/*
			 * Fail open for write on a write-protected drive.
			 */
			retval = EACCES;
		}
		v_lock(&up->zu_lock, SPL0);
		if (retval != 0)
			--up->zu_nopen;
		v_sema(&up->zu_ocsema);
#ifdef DEBUG
		ZDPRINTF("zdopen: state = %d\n", up->zu_state);
#endif /* DEBUG */
		return (retval);
	}

	/*
	 * First open!
	 * Drop lock, we don't need as no other I/O can be in progress
	 * on this drive.
	 */
	up->zu_flags |= ZUF_OPENING;
	v_lock(&up->zu_lock, SPL0);

	up->zu_state = ZU_GOOD;		/* assume good */
		
	/*
	 * Free memory for bad block list from previous open. 
	 * If drive has removable media, there may be a different
	 * bad block list for this open...
	 */
	if (up->zu_zdbad != NULL) {
		size = (up->zu_zdbad->bz_nsnf * sizeof(struct bz_bad))
			+ sizeof(struct zdbad) - sizeof(struct bz_bad);
		size = roundup(size, DEV_BSIZE);
		wmemfree((caddr_t)up->zu_zdbad, size);
		up->zu_zdbad = NULL;
	}

	oldcfg = up->zu_cfg;
	if ((retval = zdprobe_drive(dp, up)) != 0) {
		/*
		 * Error whilst attempting probe!
		 */
		goto out;
	}
	/*
	 * Error if drive still not there or not online.
	 */
	if (up->zu_cfg == ZD_NOTFOUND ||		/* Not there */
	    (up->zu_cfg & ZD_ONLINE) != ZD_ONLINE) {	/* Offline */
		if ((oldcfg & ZD_ONLINE) == ZD_ONLINE) {
			/*
			 * It had been online.
			 */
			disk_offline();
		}
		retval = ENXIO;
		goto out;
	}

	if (oldcfg == ZD_NOTFOUND || ((oldcfg & ZD_ONLINE) != ZD_ONLINE)) {
		/*
		 * It had been offline.
		 */
		disk_online();
	}

	if ((mode & FWRITE) && (up->zu_cfg & ZD_READONLY)) {
		/*
		 * Fail open for write on a write-protected drive.
		 */
		retval = EACCES;
		goto out;
	}

	dd = zdget_chancfg(dp, up);
	if (dd == NULL) {
		retval = EIO;
		goto out;
	}
	/*
	 * Does channel configuration match the drive configuration.
	 * If bound drive_type and mismatch, then drive is marked to
	 * disallow normal read/write operations.
	 */
	if (dd->zdd_sectors == 0) {
		up->zu_state = ZU_NO_RW;
		cmn_err(CE_WARN, 
			"zd%d: drive type mismatch - check configuration.",
			up - zdunit);
		/*
		 *+ There was a mismatch in the drive types between the
		 *+ channel configuration and the drive configuration.
		 *+ Normal read/write operations will be disallowed on
		 *+ this drive.
		 */
		goto out;
	}

	/*
	 * Make sure drive size and first sect are set,
	 * since the channel config may have been set
	 * since system boot.
	 */
	up->zu_drive_size = dd->zdd_sectors * dd->zdd_tracks * dd->zdd_cyls;
	up->zu_firstsect = dd->zdd_sectors * dd->zdd_tracks;

	/*
	 * If the drive is not formatted or is formatted differently than
	 * the other drives on the channel, then cannot read the bad block
	 * list. In this case, only format operations via ioctl will be
	 * allowed. Read and write operations will return error.
	 */
	if ((up->zu_cfg & (ZD_FORMATTED|ZD_MATCH)) != (ZD_FORMATTED|ZD_MATCH)) {
		up->zu_state = ZU_NO_RW;
		cmn_err(CE_WARN, "zd%d: %s.", up - zdunit,
			((up->zu_cfg & ZD_FORMATTED) != ZD_FORMATTED)
				? "drive unformatted"
				: "drive/channel mismatch");
		/*
		 *+ Either the drive was not formatted or it was not formatted in the
		 *+ same manner as other drives on the channel. This prevents the bad
		 *+ block list from being read in.
		 */
		goto out;
	}

	up->zu_drive_type = dd->zdd_drive_type;

	/*
	 * Online and formatted drive - get bad block list.
	 * If cannot correctly read bad block list - only allow ioctl
	 * operations.
	 */
	if (zdgetbad(dp, up) < 0) {
		cmn_err(CE_WARN, "zd%d: Cannot read bad block list.", 
			up - zdunit);
		/*
		 *+ The driver couldn't read the bad block list.  Only ioctl
		 *+ operations will be allowed on this drive.
		 */
		up->zu_state = ZU_NO_RW;
	}

out:
	if (mode & FEXCL) {		/* open for exclusive use */
		if (!suser())
			retval = EACCES;
		(void)p_lock(&up->zu_lock, SPLZD);
		up->zu_flags |= ZUF_EXCLUSIVE;
		v_lock(&up->zu_lock, SPL0);
	}
#ifdef DEBUG
	ZDPRINTF("state = %d\n", up->zu_state);
#endif /* DEBUG */
	(void)p_lock(&up->zu_lock, SPLZD);
	if (retval != 0)
		--up->zu_nopen;
	else {
		/*
		 * If open successful and ZUF_SEEKERR set, then
		 * patrol seeks need to be re-enabled.
		 */
		if (up->zu_flags & ZUF_SEEKERR) {
			up->zu_flags &= ~ZUF_SEEKERR;
			up->zu_flags |= ZUF_PATROL;
			(void)p_lock(&patrollock, SPLZD);
			patrolcount++;
			v_lock(&patrollock, SPL0);
		}
	}
	up->zu_flags &= ~ZUF_OPENING;
	v_lock(&up->zu_lock, SPL0);
	v_sema(&up->zu_ocsema);
	return (retval);
}

/*
 * zdclose
 *	Close the device.
 *	If last close, free memory holding bad block list.
 */
/*ARGSUSED*/
zdclose(dp, flag)
	struct dev *dp;
	int flag;
{
	register struct zd_unit *up;
	dev_t dev = dp->d_dev;
        struct zdc_ctlr *ctlrp;
        int chan;

	up = &zdunit[ZDUNIT(dev)];
#ifdef	DEBUG
	if (zddebug)
		CPRINTF("C");
#endif	/* DEBUG  */
	(void)p_sema(&up->zu_ocsema, PZOPEN);
	(void)p_lock(&up->zu_lock, SPLZD);
	if (--up->zu_nopen == 0) {
		up->zu_flags &= ~(ZUF_EXCLUSIVE|ZUF_OPENING);
                /*
                 * Don't let it go away with the controller
                 * channel reserved.  Release it if it failed
                 * to restore it.
                 */
                chan = up->zu_drive & 1;
                ctlrp = &zdctrlr[up->zu_ctrlr];
                if (ctlrp->zdc_chan_owner[chan] == up)
                        ctlrp->zdc_chan_owner[chan] = (struct zd_unit *)NULL;
	}
	v_lock(&up->zu_lock, SPL0);
	v_sema(&up->zu_ocsema);
}

/*
 * zdprobe_drive
 *	Probe for the status of a particular drive then update
 *	appropriate fields in controller and unit structures.
 */
static int
zdprobe_drive(dp, up)
	struct dev *dp;
	register struct zd_unit *up;
{
	register struct buf *bp;	/* ioctl buffer */
	register struct cb  *cbp;	/* cb argument */
	int	error;

	bp  = &up->zu_ioctl;
	cbp = &up->zu_ioctlcb;
	bufalloc(bp);			/* will always get since 1st open */
	bp->b_flags = B_IOCTL;
	bp->b_bcount = 0;		/* data in CB */
	bp->b_un.b_addr = NULL;
	bp->b_devp = dp;
	bp->b_blkno = 0;
	bp->b_error = 0;
	bp->b_proc = u.u_procp;
	bp->b_iotype = B_FILIO;
	BIODONE(bp) = 0;
	/* Fill out CB */
	cbp->cb_cmd = ZDC_PROBEDRIVE;
	/*
	 * Insert at head of list and zdstart!
	 */
	(void)p_lock(&up->zu_lock, SPLZD);
	bp->av_forw = NULL;
	up->zu_bhead.av_forw = bp;
	DS_QUEUED_TIME(bp);
	zdstart(up);
	DS_Q(up->zu_dstats);
	v_lock(&up->zu_lock, SPL0);
	biowait(bp);
	error = geterror(bp);
	if (error) {
		buffree(bp);
		return (error);
	}
	/*
	 * extract drive cfg data.
	 */
	up->zu_cfg = ((struct probe_cb *)cbp)->pcb_drivecfg[up->zu_drive];
	zdctrlr[up->zu_ctrlr].zdc_drivecfg[up->zu_drive] = up->zu_cfg;
	buffree(bp);
	return (0);
}

/*
 * zdget_chancfg
 *	Get the channel configuration for the channel on which this
 *	drive resides. Fills in channel configuration in controller structure.
 */
static struct zdcdd *
zdget_chancfg(dp, up)
	struct dev *dp;
	register struct zd_unit *up;
{
	register struct buf	*bp;	/* ioctl buffer */
	register struct cb	*cbp;	/* cb argument */
	struct zdcdd	*dd;		/* channel configuration data */
	int	chan;			/* Channel A (0) or channel B (1) */

	bp  = &up->zu_ioctl;
	cbp = &up->zu_ioctlcb;
	bufalloc(bp);			/* will always get since 1st open */
	bp->b_un.b_addr = wmemall(sizeof(struct zdcdd), 1);
	bzero(bp->b_un.b_addr, sizeof(struct zdcdd));
	bp->b_flags = B_READ | B_IOCTL;
	bp->b_bcount = sizeof(struct zdcdd);
	bp->b_devp = dp;
	bp->b_blkno = 0;
	bp->b_error = 0;
	bp->b_proc = u.u_procp;
	bp->b_iotype = B_FILIO;
	BIODONE(bp) = 0;
	/* Fill out CB */
	cbp->cb_cmd = ZDC_GET_CHANCFG;
	cbp->cb_addr = (ulong)bp->b_un.b_addr;
	cbp->cb_count = sizeof(struct zdcdd);
	/*
	 * Insert at head of list and zdstart!
	 */
	(void)p_lock(&up->zu_lock, SPLZD);
	bp->av_forw = NULL;
	up->zu_bhead.av_forw = bp;
	DS_QUEUED_TIME(bp);
	zdstart(up);
	DS_Q(up->zu_dstats);
	v_lock(&up->zu_lock, SPL0);
	biowait(bp);
	if (bp->b_flags & B_ERROR) {
		wmemfree(bp->b_un.b_addr, sizeof(struct zdcdd));
		buffree(bp);
		return (NULL);
	}
	/*
	 * Extract Channel configuration data.
	 */
	chan = up->zu_drive & 1;
	dd = (chan) ? &zdctrlr[up->zu_ctrlr].zdc_chanB
		    : &zdctrlr[up->zu_ctrlr].zdc_chanA;
	bcopy(bp->b_un.b_addr, (caddr_t)dd, sizeof(struct zdcdd));
	wmemfree(bp->b_un.b_addr, sizeof(struct zdcdd));
	buffree(bp);
	/*
	 * On the first get of the channel configuration set the
	 * dma throttle for the channel. Make sure that the configuration
	 * is valid.
	 */
	if (dd->zdd_sectors != 0 &&
	    (zdctrlr[up->zu_ctrlr].zdc_dma_throttle & (1 << chan)) == 0)
		set_dma_throttle(&zdctrlr[up->zu_ctrlr], dd, chan);
	return (dd);
}

/*
 * set_dma_throttle
 *	If 1st open on channel, set the dma throttle
 */
/*ARGSUSED*/
static
set_dma_throttle(ctlrp, dd, chan)
	register struct zdc_ctlr *ctlrp;	/* controller */
	struct zdcdd	*dd;			/* channel configuration data */
	int	chan;				/* channel A (0) or B (1) */
{
	register int	throttle;	/* throttle count */
	spl_t	s_ipl;

	s_ipl = p_lock(&ctlrp->zdc_ctlrlock, SPLHI);
	if ((ctlrp->zdc_dma_throttle & (1 << chan)) == 0) {
		throttle = zdc_AB_throttle;
		if (throttle > SLB_TVAL)
			throttle = SLB_TVAL;
		wrslave(ctlrp->zdc_slicaddr, (unchar)(SL_G_CHAN0 + chan),
					(unchar)(SLB_TH_ENB | throttle));
		ctlrp->zdc_dma_throttle |= (1 << chan);
	}
	v_lock(&ctlrp->zdc_ctlrlock, s_ipl);
}

/*
 * zdpatrol 
 *	Perform a seek on all drives.  This routine is called
 *	via timeout to perform a seek at regular intervals.
 *	This implements the patrol seek recommended by Fujitsu
 *	for prolonging the life of the drive by not allowing
 *	the head to stay in one place for long periods.
 */
static
zdpatrol(up)
	register struct zd_unit *up;	/* start checking at this drive */
{
	register struct buf *bp;	/* ioctl buffer */
	register struct cb  *cbp;	/* cb argument */
	struct zd_unit *oldup;
	struct zdcdd *dd;
	spl_t	s_ipl;
	int interval;

#ifdef DEBUG
	ZDPRINTF2("<Z>");
#endif /* DEBUG */
	s_ipl = p_lock(&patrollock, SPLZD);
	if (patrolcount == 0) {
		/*
		 * If no devices have patrol seek enabled, die!
		 */
		patrol_on = 0;
		v_lock(&patrollock, s_ipl);
		return;
	}
	interval = (PATROLTIME/2)/patrolcount;
	v_lock(&patrollock, s_ipl);	
	oldup = up;
	do {
		if (up->zu_flags & ZUF_PATROL)
			break;
		up++;
		if (up > &zdunit[zd_max_unit])
			up = zdunit;
	} while (up != oldup);

	s_ipl = p_lock(&up->zu_lock, SPLZD);
	/*
	 * Skip device if in process of opening (data structures
	 * may be changing..), if active since last check, or if
	 * patrol turned off, or if opened exclusively.  (currently
	 * only the format program opens exclusively -- seeking during
	 * reformat may be a bad idea...)
	 */
	if ((up->zu_flags & (ZUF_ACTIVE|ZUF_OPENING|ZUF_EXCLUSIVE))
	    || !(up->zu_flags & ZUF_PATROL)) {
		up->zu_flags &= ~ZUF_ACTIVE;
		v_lock(&up->zu_lock, s_ipl);
#ifdef DEBUG
		ZDPRINTF2("<c%d>", up-zdunit);
#endif /* DEBUG */
		goto next;	
	}
	dd = (up->zu_drive & 1) ? &zdctrlr[up->zu_ctrlr].zdc_chanB
				: &zdctrlr[up->zu_ctrlr].zdc_chanA;
	/*
	 * Insert at head of list and zdstart!
	 */
	up->zu_flags |= ZUF_ACTIVE;
#ifdef DEBUG
	ZDPRINTF2("<p%d x%x>", up-zdunit, up->zu_lastcyl);
#endif /* DEBUG */
	bp  = &up->zu_seek;
	cbp = &up->zu_seekcb;
	cbp->cb_cyl = (up->zu_lastcyl + 250) % dd->zdd_cyls;
	if (cbp->cb_cyl == 0)		
		/* reading 1st sector of disk will fail, so don't */
		cbp->cb_cyl++;
	up->zu_lastcyl = cbp->cb_cyl;
	bp->av_forw = up->zu_bhead.av_forw; 
	up->zu_bhead.av_forw = bp;
	DS_QUEUED_TIME(bp);
	zdstart(up);
	DS_Q(up->zu_dstats);
	v_lock(&up->zu_lock, s_ipl);

next:
	/*
	 * On next invocation, zdpatrol should start searching
	 * for a drive to patrol starting with the drive following
	 * the one we just checked.
	 */
	up++;
	if (up > &zdunit[zd_max_unit])
		up = zdunit;
#ifdef DEBUG
	ZDPRINTF2("<T %d>", interval);
#endif /* DEBUG */
	if (!interval)	/* more than PATROLTIME/2 drives! */
		interval = 1;
	(void)timeout(zdpatrol, up, interval*HZ);
	return;
}


/*
 * getchksum
 *	Calculate bad block list checksum
 */
static
getchksum(lptr, nelem, seed)
	register long *lptr;
	register int nelem;
	long seed;
{
	register long sum;

	sum = seed;
	while (nelem-- > 0) {
		sum ^= *lptr;
		++lptr; 
	}
	return (sum);
}

/*
 * zdgetbad
 *	- get the bad block list for this unit.
 * Return:
 *	 0 - success
 *	-1 - failure
 */
static
zdgetbad(dp, up)
	struct dev *dp;
	struct zd_unit *up;
{
	register struct cb	*cbp;
	register struct	bz_bad	*fbzp, *tbzp;
	register struct buf	*bp;		/* ioctl buffer */
	register int	size;
	struct	zdbad	*zdp;			/* bad block list */
	struct	zdcdd	*dd;			/* disk description */
	int	zdpsize;
	int	block;
	caddr_t	addr;
	static struct	zdbad	null_bbl;	/* fake bad block list */

	/*
	 * Initialize a fake bad block list in case we receive a
	 * SNF failure whilst reading the bad block list!
	 */
	null_bbl.bz_nelem = 0;
	null_bbl.bz_nsnf = 0;
	up->zu_zdbad = &null_bbl;

	dd = (up->zu_drive & 1) ? &zdctrlr[up->zu_ctrlr].zdc_chanB
				: &zdctrlr[up->zu_ctrlr].zdc_chanA;
	size = 1;				/* get 1 for starters */
	zdp = (struct zdbad *)wmemall(DEV_BSIZE, 1);
	/*
	 * Read bad block list.
	 */
	bp = &up->zu_ioctl;
	bufalloc(bp);
	bp->b_devp = dp;
	bp->b_blkno = 0;
	bp->b_proc = u.u_procp;
	bp->b_iotype = B_FILIO;
	cbp = &up->zu_ioctlcb;
	cbp->cb_mod = 0;
	cbp->cb_cmd = ZDC_READ;
	cbp->cb_cyl = ZDD_DDCYL;
	addr = (caddr_t)zdp;
	for (block = 0; block < size; block++) {
		cbp->cb_head = 0;
		cbp->cb_sect = ZDD_NDDSECTORS + block;
		while (cbp->cb_head < MIN(dd->zdd_tracks, BZ_NBADCOPY)) {
			bp->b_bcount = DEV_BSIZE;
			bp->b_flags = B_READ | B_IOCTL;
			bp->b_un.b_addr = addr;
			bp->b_error = 0;
			BIODONE(bp) = 0;
			cbp->cb_count = DEV_BSIZE;
			cbp->cb_addr = (ulong)addr;
			/*
			 * Queue and start I/O
			 */
			(void)p_lock(&up->zu_lock, SPLZD);
			bp->av_forw = NULL;
			up->zu_bhead.av_forw = bp;
			DS_QUEUED_TIME(bp);
			zdstart(up);
			DS_Q(up->zu_dstats);
			v_lock(&up->zu_lock, SPL0);
			biowait(bp);
			if ((bp->b_flags & B_ERROR) != B_ERROR)
				break;
			/*
			 * If could not read - try next track.
			 */
			cbp->cb_head++;
			cbp->cb_sect = block;
		}

		/*
		 * If cannot read the first block of bad block list.
		 * give up and return.
		 */
		if (bp->b_flags & B_ERROR) {
			cmn_err(CE_WARN, 
				"zd%d: Cannot read block %d of bad block list.",
				up - zdunit, block);
			/*
			 *+ The driver couldn't read the first block of the bad
			 *+ block list.
			 */
			if (block == 0)
				wmemfree((caddr_t)zdp, DEV_BSIZE);
			else
				wmemfree((caddr_t)zdp, size << DEV_BSHIFT);
			buffree(bp);
			up->zu_zdbad = (struct zdbad *)NULL;
			return (-1);
		}
		if (block == 0) {
			size = (zdp->bz_nelem * sizeof(struct bz_bad))
				+ sizeof(struct zdbad) - sizeof(struct bz_bad);
			size = howmany(size, DEV_BSIZE);
			if (size > ((dd->zdd_sectors - ZDD_NDDSECTORS) >> 1)) {
				cmn_err(CE_WARN, 
					"zd%d: Bad block list corrupted!",
					up - zdunit);
				/*
				 *+ The bad block list was corrupted.
				 */
				wmemfree((caddr_t)zdp, DEV_BSIZE);
				up->zu_zdbad = (struct zdbad *)NULL;
				buffree(bp);
				return (-1);
			}
			zdpsize = size << DEV_BSHIFT;
			/*
			 * copy block 0 to new location in free memory, so
			 * that bad block list will be contiguous.
			 */
			addr = wmemall(zdpsize, 1);
			bcopy((caddr_t)zdp, addr, (unsigned)DEV_BSIZE);
			wmemfree((caddr_t)zdp, DEV_BSIZE);
			zdp = (struct zdbad *)addr;
		}
		addr += DEV_BSIZE;
	} /* end of for */ 

	/*
	 * done with I/O - free buf header.
	 */
	buffree(bp);

	/*
	 * Confirm data integrity via checksum.
	 */
	size = (zdp->bz_nelem * sizeof(struct bz_bad)) / sizeof(long);
	if (zdp->bz_csn != getchksum((long *)zdp->bz_bad, size,
				(long)(zdp->bz_nelem ^ zdp->bz_nsnf))) {
		cmn_err(CE_WARN, "zd%d: Checksum failed!", up - zdunit);
		/*
		 *+ The checksum on the bad block list failed.
		 */
		wmemfree((caddr_t)zdp, zdpsize);
		up->zu_zdbad = (struct zdbad *)NULL;
		return (-1);
	}

	/*
	 * Copy only BZ_SNF entries into unit's bad block list.
	 * 
	 * Calculate size of needed bad block list.
	 * That is, only BZ_SNF entries are needed.
	 */
	size = (zdp->bz_nsnf * sizeof(struct bz_bad))
		+ sizeof(struct zdbad) - sizeof(struct bz_bad);
	size = roundup(size, DEV_BSIZE);
	up->zu_zdbad = (struct zdbad *)wmemall(size, 1);
#ifdef	DEBUG
	bzero((caddr_t)up->zu_zdbad, (uint)size);
#endif	/* DEBUG  */
	*up->zu_zdbad = *zdp;
	tbzp = up->zu_zdbad->bz_bad;
	for (fbzp = zdp->bz_bad; fbzp < &zdp->bz_bad[zdp->bz_nelem]; fbzp++) {
		if (fbzp->bz_rtype == BZ_SNF)
			*tbzp++ = *fbzp;
	}
	wmemfree((caddr_t)zdp, zdpsize);
	return (0);
}

/*
 * zdstrat
 * 	zd disk read/write routine.
 * Perform various checks and queue request and call start routine to
 * initiate I/O to the device.
 */
/*ARGSUSED*/
zdstrat(dp, bp, offset)
	struct dev *dp;
	register struct buf *bp;
	daddr_t	offset;
{
	register struct zd_unit *up;
	register struct zdcdd	*dd;		/* channel configuration */
	register int sector;
	register int nspc;
	struct	diskaddr diskaddress;
	spl_t s_ipl;

#ifdef	DEBUG
	if (zddebug > 1)
		CPRINTF("zdstrat(%c): bp=%x, dev=%x, cnt=%d, blk=%x, vaddr=%x\n",
			(bp->b_flags & B_READ) ? 'R' : 'W',
			bp, bp->b_dev, bp->b_bcount, bp->b_blkno + offset,
			bp->b_un.b_addr);
	else	if (zddebug)
			CPRINTF("%c", (bp->b_flags & B_READ) ? 'R' : 'W');
#endif	/* DEBUG  */

	bp->b_status = B_STATUS_OK;	  /* default and base for ORing */

#ifdef DEBUG
	if (zd_status_returns) {
		bp->b_status = zd_status_returns;
		zd_status_returns = (zd_status_returns + 1) & B_S_ALL;
	}
	if (zd_express) {
		bp->b_flags |= B_EXPRESS;
		zd_express--;
	}
#endif /* DEBUG */

	up = &zdunit[ZDUNIT(bp->b_dev)];
	/*
	 * Error if NO_RW operations are permitted.
	 */
	if (up->zu_state == ZU_NO_RW) {
		bp->b_flags |= B_ERROR;
		bp->b_error = EIO;
		bp->b_resid = bp->b_bcount;
		bp->b_status |= B_S_UNIT_ERR;
		biodone(bp);
		return;
	}

	/*
	 * IO request validity check.
	 *
	 * Fail request if bogus byte count, or if address not aligned to
	 * ADDRALIGN boundary.  Note that we check to be sure block number
	 * is non-negative, which may be needed if the unpartitioned
	 * block device is used.
	 */
	if (bp->b_bcount <= 0
	||  ((bp->b_bcount & (DEV_BSIZE -1)) != 0)		/* size */
	||  (bp->b_blkno + offset + howmany(bp->b_bcount, DEV_BSIZE)
					> up->zu_drive_size)	/* offset */
	||  ((bp->b_iotype == B_RAWIO) &&
	     (((int)bp->b_un.b_addr & (ADDRALIGN - 1)) != 0))	/* alignment */
	||  (bp->b_blkno + offset < 0)) {
		bp->b_flags |= B_ERROR;
		bp->b_error = EINVAL;
		bp->b_resid = bp->b_bcount;
		biodone(bp);
		return;
	}

	dd = (up->zu_drive & 1)	? &zdctrlr[up->zu_ctrlr].zdc_chanB
				: &zdctrlr[up->zu_ctrlr].zdc_chanA;
	nspc = dd->zdd_sectors * dd->zdd_tracks;
	sector = bp->b_blkno + offset;
	diskaddress.da_cyl = sector / nspc;
	sector %= nspc;
	diskaddress.da_head = sector / dd->zdd_sectors;
	diskaddress.da_sect = sector % dd->zdd_sectors;
	bp->b_diskaddr = *(long *)&diskaddress;
	bp->b_psect = zdgetpsect(&diskaddress, dd);

	s_ipl = p_lock(&up->zu_lock, SPLZD);
	if (up->zu_state == ZU_BAD) {
		v_lock(&up->zu_lock, s_ipl);
		/*
		 * Controller/channel/Drive has gone bad!
		 */
		bp->b_flags |= B_ERROR;
		bp->b_error = EIO;
		bp->b_resid = bp->b_bcount;
		bp->b_status = B_S_XFER_ERR | B_S_UNIT_ERR | B_S_PERM_ERR;
		biodone(bp);
		return;
	}
	disksort(&up->zu_bhead, bp);
	DS_QUEUED_TIME(bp);
	up->zu_flags |= ZUF_ACTIVE;
	zdstart(up);
	DS_Q(up->zu_dstats);
	v_lock(&up->zu_lock, s_ipl);
}

/*
 * zdgetpsect
 *	Determine physical sector in track where the I/O  transfer is to
 *	occur. Used by ZDC firmware for RPS optimization.
 */
static
zdgetpsect(dp, dd)
	register struct	diskaddr *dp;
	register struct zdcdd	 *dd;
{
	if (dd->zdd_tskew == 1)
		return (( ((dd->zdd_tracks - 1 + dd->zdd_cskew) * dp->da_cyl)
				+ dp->da_head + dp->da_sect)
					 % (dd->zdd_sectors + dd->zdd_spare));
	/*
	 * track skew != 1
	 */
	return (( (((dd->zdd_tskew*(dd->zdd_tracks-1)) + dd->zdd_cskew)
								 * dp->da_cyl)
		  + (dd->zdd_tskew*dp->da_head) + dp->da_sect)
					% (dd->zdd_sectors + dd->zdd_spare));
}

/*
 * zdfill_iovec
 *	- fill out cb_iovec for the I/O request.
 *
 * B_RAWIO, B_PTEIO, B_PTBIO cases must flush TLB to avoid stale mappings
 * thru Usrptmap[], since this is callable from interrupt procedures (SGS only).
 *
 * Panics if bad pte found; "can't" happen.
 */

static ulong
zdfill_iovec(bp, iovstart)
#ifndef	i386
	register				/* want optimial on 032's */
#endif
	struct buf *bp;
	ulong *iovstart;
{
	register struct pte *pte;
	register int	count;
	register ulong	*iovp;
	ulong	retval;
	unsigned offset;
	extern struct pte *vtopte();

	/*
	 * Source/target pte's are found differently based on type
	 * of IO operation.
	 */
	switch(bp->b_iotype) {

	case B_RAWIO:					/* RAW IO */
		/*
		 * In this case, must look into alignment of physical
		 * memory, since we can start on any ADDRALIGN boundary.
		 */
		flush_tlb();
		pte = vtopte(bp->b_proc, clbase(btop(bp->b_un.b_addr)));
		count = (((int)bp->b_un.b_addr & CLOFSET) + bp->b_bcount
							 + CLOFSET) / CLBYTES;
		retval = (ulong)bp->b_un.b_addr;
		break;

	case B_FILIO:					/* file-sys IO */
		/*
		 * Filesys/buffer-cache IO.  These are always cluster aligned
		 * both physically and virtually.
		 * Note: also used when to kernel memory acquired via wmemall().
		 * For example, channel configuration buffer in zdget_chancfg().
		 */
		pte = &Sysmap[btop(bp->b_un.b_addr)];
		count = (bp->b_bcount + CLOFSET) / CLBYTES;
		retval = (ulong)bp->b_un.b_addr;
		break;

	case B_PTEIO:					/* swap/page IO */
		/*
		 * Pte-based IO -- already know pte of 1st page, which
		 * is cluster aligned, and b_count is a multiple of CLBYTES.
		 */
		flush_tlb();
		pte = bp->b_un.b_pte;
		count = (bp->b_bcount + CLOFSET) / CLBYTES;
		retval = PTETOPHYS(*pte);
		break;

	case B_PTBIO:					/* Page-Table IO */
		/*
		 * Page-Table IO: like B_PTEIO, but can start/end with
		 * non-cluster aligned memory (but is always HW page
		 * aligned). Count is multiple of NBPG.
		 *
		 * Separate case for greater efficiency in B_PTEIO.
		 */
		flush_tlb();
		pte = bp->b_un.b_pte;
		retval = PTETOPHYS(*pte);
		offset = PTECLOFF(*pte);
		pte -= btop(offset);
		count = (offset + bp->b_bcount + CLOFSET) / CLBYTES;
		break;

	default:
		cmn_err(CE_PANIC, "zdfill_iovec: bad b_iotype");
		/*
		 *+ Before issuing a command to the controller,
		 *+ the DCC driver was filling out the command block 
		 *+ and found that it
		 *+ had been passed an invalid buffer type as an argument.
		 */
		/*NOTREACHED*/
	}

	/*
	 * Now translate PTEs and fill-in iovectors.
	 */
	iovp = iovstart;
	while (count--) {
		*iovp++ = PTETOPHYS(*pte);
		pte += CLSIZE;
	}
	return (retval);
}

/*
 * zdstart
 *	- intitiate I/O request to controller.
 * If controller is busy just return. Otherwise stuff appropriate CB
 * with request and notify ZDC.
 *
 * Called with unit structure locked at SPLZD.
 */
zdstart(up)
	register struct zd_unit *up;
{
	register struct cb	*cbp;
	register struct buf	*bp;
	register struct zdc_ctlr *ctlrp;
	spl_t	s_ipl;

#ifdef	DEBUG
	if (zddebug)
		CPRINTF("S");
#endif	/* DEBUG  */
	bp = up->zu_bhead.av_forw;
	ctlrp = &zdctrlr[up->zu_ctrlr];
	cbp = up->zu_cbptr;
	if (cbp->cb_bp == NULL && cbp[1].cb_bp == NULL) {
		/*
		 * Drive is idle.
		 * If fp_lights - turn activity light on.
		 * Get starting time.
		 */
		if (fp_lights) {
			s_ipl = splhi();
			FP_IO_ACTIVE;
			splx(s_ipl);
		}
		DS_IOSTART_TIME(up->zu_dstats);
	}
	if (cbp->cb_bp == NULL || cbp[1].cb_bp == NULL) {
		/*
		 * Fill CB.
		 */
		if (cbp->cb_bp)
			++cbp;
		if (bp->b_flags & B_IOCTL) {
			/*
			 * copy 1st half of cb (what fw will see)
			 */
			if (bp->b_flags & B_ASYNC)
				/*
				 * We are using B_IOCTL|B_ASYNC to
				 * distinguish patrol seeks from other
				 * ioctl's
				 */
				bcopy((caddr_t)&up->zu_seekcb, 
					(caddr_t)cbp, FWCBSIZE);
			else
				bcopy((caddr_t)&up->zu_ioctlcb, 
					(caddr_t)cbp, FWCBSIZE);
			if (cbp->cb_cmd != ZDC_READ_LRAM
			&&  cbp->cb_cmd != ZDC_WRITE_LRAM)
				cbp->cb_addr = zdfill_iovec(bp, cbp->cb_iovstart);
			cbp->cb_mod |= up->zu_mod;
		} else {
			*(long *)&cbp->cb_diskaddr = bp->b_diskaddr;
			up->zu_lastcyl = cbp->cb_cyl;
			cbp->cb_psect = (unchar)bp->b_psect;
			cbp->cb_count = bp->b_bcount;
			cbp->cb_mod = up->zu_mod;
			cbp->cb_cmd = (bp->b_flags & B_READ) ? ZDC_READ : ZDC_WRITE;
			cbp->cb_addr = zdfill_iovec(bp, cbp->cb_iovstart);
			/*
			 * Do cylinder/block access related stats
			 */
			ds_access_stats(up->zu_dstats,
					(ulong)cbp->cb_diskaddr.da_cyl);
		}
		bp->b_resid = bp->b_bcount;
		cbp->cb_transfrd = 0;
		cbp->cb_bp = bp;
		cbp->cb_errcnt = 0;
		cbp->cb_state = ZD_NORMAL;
		cbp->cb_iovec = cbp->cb_iovstart;
		/*
		 * Notify ZDC of job request.
		 */
#ifdef	DEBUG
		if (zddebug > 2)
			zddumpcb(cbp);
#endif	/* DEBUG  */
		NUDGE_ZDC(ctlrp, cbp, s_ipl);
		up->zu_bhead.av_forw = bp->av_forw;
	}
}

/*
 * zd_intr()
 *	Normal request completion interrupt handler.
 */
void
zd_intr(vector)
	unchar	vector;
{
	register struct cb	 *cbp;
	register struct	zd_unit  *up;
	register struct	buf	 *donebp;
	register struct	zdc_ctlr *ctlrp;
	struct	zdcdd *dd;
	int	zdcvec;
	daddr_t	origblkno;
	struct	diskaddr diskaddress;
	int	i;
	unchar	val;
	spl_t	s_ipl;
	int 	waiter = 1;

#ifdef	DEBUG
	if (zddebug)
		CPRINTF("<I>");
#endif	/* DEBUG  */
	zdcvec = vector - base_cb_intr;
	ctlrp = &zdctrlr[zdcvec >> NCBZDCSHFT];
	cbp = ctlrp->zdc_cbp + (zdcvec & (NCBPERZDC-1));
	up = &zdunit[cbp->cb_unit];

#ifdef DEBUG
	if ( zd_intr_shutdown == up ) {
		zd_intr_shutdown = (struct zd_unit *)0;
		zdshutdrive(up);
	}
#endif /* DEBUG */

	if (cbp->cb_bp == NULL) {
		if (ctlrp->zdc_state == ZDC_DEAD) {
			cmn_err(CE_NOTE, 
				"zdc%d drive %d: Spurious interrupt from dead controller.",
				ctlrp - zdctrlr,
				(cbp - ctlrp->zdc_cbp) / NCBPERDRIVE);
			/*
			 *+ The system received an interrupt from a DCC
			 *+ controller that was marked as being dead.
			 */
			return;
		}
		if (cbp->cb_unit < 0) {
			cmn_err(CE_NOTE, 
				"zdc%d drive %d: Interrupt from unknown unit.",
				ctlrp - zdctrlr,
				(cbp - ctlrp->zdc_cbp) / NCBPERDRIVE);
			/*
			 *+ The system received an interrupt from an unknown unit
			 *+ on a DCC controller.
			 */
			return;
		}
		cmn_err(CE_NOTE, "zd%d: Spurious interrupt.", cbp->cb_unit);
		/*
		 *+ The system received a spurious interrupt from a DCC controller.
		 */
		return;
	}
	
	/*
	 * Record completion code to save possible error for user to
	 * get through ioctl ZIOGERR.
	 */
	up->zu_l_compcode = cbp->cb_compcode;

	/*
	 * Separate normal case from error cases for performance.
	 */
	if (cbp->cb_compcode == ZDC_DONE) {
		if (cbp->cb_state == ZD_NORMAL) {
			donebp = cbp->cb_bp;
			donebp->b_resid = 0;
			goto donext;
		}
		if (cbp->cb_state & ZD_RESET) {
			/*
			 * Reset completed.
			 */
			cbp->cb_state &= ~ZD_RESET;
			if (cbp->cb_errcnt++ < zdcretry) {
				zdretry(cbp, up, ctlrp);
				return;
			}
			/*
			 * Fail I/O request.
			 */
			donebp = cbp->cb_bp;
			donebp->b_flags |= B_ERROR;
			donebp->b_error = EIO;
			donebp->b_status |= B_S_XFER_ERR;
			goto donext;
		}
		if (cbp->cb_state & ZD_REVECTOR) {
			/*
			 * Completed revector. Continue with rest of transfer.
			 */
			zdcontinue(cbp, up, ctlrp);
			return;
		}
	}

	/*
	 * E R R O R  O C C U R R E D !
	 */

#ifdef DEBUG
	ZDPRINTF2("<IE 0x%x>", cbp->cb_compcode);
#endif /* DEBUG */
	/*
	 * If this is from a patrol seek, don't retry
	 * or report error -- drive may have gone offline
	 * when closed which is okay.  So just set flag
	 * and disable seeks.
	 */
	if ((cbp->cb_bp->b_flags & (B_IOCTL|B_ASYNC)) == (B_IOCTL|B_ASYNC)) {
#ifdef DEBUG
		ZDPRINTF2("<IS>");
#endif /* DEBUG */
		donebp = cbp->cb_bp;
		switch (cbp->cb_compcode) {
		case ZDC_SNF:
		case ZDC_SOFTECC:
		case ZDC_CORRECC:
		case ZDC_ECC:
		case ZDC_HDR_ECC:
		case ZDC_SO:
		case ZDC_NDS:
			/*
			 * Just ignore these errors
			 */
			goto donext;
		default:
			/*
			 * Turn off seeks, since drive is
			 * having problems (may have gone offline...)
			 */
			s_ipl = p_lock(&up->zu_lock, SPLZD);
			up->zu_flags |= ZUF_SEEKERR;
			up->zu_flags &= ~ZUF_PATROL;
			v_lock(&up->zu_lock, s_ipl);
			s_ipl = p_lock(&patrollock, SPLZD);
			--patrolcount;
			v_lock(&patrollock, s_ipl);
			goto donext;
		}
	}
	/*
	 * Since b_blkno may not be the physical block number, compute it
	 * from the diskaddr in the buf hdr.
	 */
	dd = (up->zu_drive & 1) ? &ctlrp->zdc_chanB : &ctlrp->zdc_chanA;
	*(long *)&diskaddress = cbp->cb_bp->b_diskaddr;
	origblkno = ZDBLKNO(diskaddress, dd);

	/*
	 * Short circuit ZDC_READ_HDRS failure so
	 * revector is not attempted.
	 */
        if (cbp->cb_cmd == ZDC_READ_HDRS) {
#ifdef DEBUG
                ZDPRINTF("zd%d: ZDC_READ_HDRS failure.\n", cbp->cb_unit);
                zddumpcb(cbp);
#endif /* DEBUG */
                donebp = cbp->cb_bp;
                donebp->b_flags |= B_ERROR;
                donebp->b_error = EIO;
		donebp->b_status |= B_S_XFER_ERR;
                goto donext;
        }

	/*
	 * If not doing retry reset or revectoring, update b_resid.
	 */
	if (cbp->cb_state == ZD_NORMAL)
		cbp->cb_bp->b_resid = cbp->cb_count;

	if (cbp->cb_state & ZD_RESET) {
		/*
		 * Error occurred during attempted reset for previous
		 * error. Give up and shutdown the drive.
		 */
		s_ipl = p_lock(&zdcprlock, SPLZD);
		cmn_err(CE_WARN, "zd%d: Reset Failed.", cbp->cb_unit);
		/*
		 *+ An error occurred when the DCC driver attempted to reset
		 *+ the controller to recover from a previous error.
		 */
		if (cbp->cb_compcode >= zdncompcodes) {
			cmn_err(CE_WARN, "zd%d: Bad cb_compcode 0x%x on reset.",
				cbp->cb_unit, cbp->cb_compcode);
			/*
			 *+ While attempting to reset the controller
			 *+ to recover from a previous error,
			 *+ the DCC driver received an invalid
			 *+ completion code from the DCC firmware.
			 */
			cbp->cb_compcode = zdncompcodes - 1;
		}
		hard_error(cbp);
		zddumpstatus(cbp);
		v_lock(&zdcprlock, s_ipl);
		zdshutdrive(up);
	} else switch (cbp->cb_compcode) {

	/*
	 * Command completed successfully but an ecc error occurred.
	 */
	case ZDC_SOFTECC:
	case ZDC_CORRECC:
		/*
		 * Corrected or Soft ECC error.
		 */
		/*
		 * We still need to find some way to report the filesystem
		 * block number, to make it easy to do an ncheck/icheck
		 * and find the affected file.  One possible way to do this is
		 * to report the bad block/inode/device in the buf cache
		 * code.
		 */
		s_ipl = p_lock(&zdcprlock, SPLZD);
		cmn_err(CE_WARN, "zd%ds%d: %s at (%d, %d, %d).",
			cbp->cb_unit, PARTCHR(cbp->cb_bp->b_dev),
			zd_compcodes[cbp->cb_compcode],
			cbp->cb_cyl, cbp->cb_head, cbp->cb_sect);
		/*
		 *+ The DCC driver received the specified error
		 *+ completion status from the DCC firmware for the
		 *+ indicated drive partition.  The '%s' translates
		 *+ to the error code received.  Information about
		 *+ that error appears in this chapter.
		 *+ The final '(%d, %d, %d)' indicates the cylinder,
		 *+ head, and sector in the failing command block.
		 */
		zddumpstatus(cbp);
		v_lock(&zdcprlock, s_ipl);
		if (cbp->cb_state & ZD_REVECTOR) {
			/*
			 * Completed revector. Continue with rest of transfer.
			 */
			zdcontinue(cbp, up, ctlrp);
			return;
		}
		donebp = cbp->cb_bp;
		donebp->b_status |= B_S_MEDIA_ERR;
		goto donext;		/* Not fatal */

	/*
	 * Fail the job immediately - no retries.
	 */
	case ZDC_DRVPROT:
	case ZDC_ECC:
                if (cbp->cb_cmd == ZDC_REC_DATA) {
                        donebp = cbp->cb_bp;
                        goto donext;            /* Not an error */
                }
		cbp->cb_bp->b_status |= B_S_XFER_ERR;
		if (cbp->cb_compcode == ZDC_DRVPROT)
			cbp->cb_bp->b_status |= B_S_UNIT_ERR;
		s_ipl = p_lock(&zdcprlock, SPLZD);
		hard_error(cbp);
		zddumpstatus(cbp);
		v_lock(&zdcprlock, s_ipl);
		break;

	case ZDC_CH_RESET:		/* Shutdown drive */
	case ZDC_BADDRV:
	case ZDC_DDC_STAT:
		s_ipl = p_lock(&zdcprlock, SPLZD);
		hard_error(cbp);
		zddumpstatus(cbp);
		v_lock(&zdcprlock, s_ipl);
		zdshutdrive(up);
		break;

	case ZDC_DMA_TO:		/* Shutdown Channel */
		s_ipl = p_lock(&zdcprlock, SPLZD);
		hard_error(cbp);
		v_lock(&zdcprlock, s_ipl);
		zdshutchan(ctlrp, up->zu_drive & 1);
		break;

	case ZDC_NOCFG:
		if (cbp->cb_cmd == ZDC_GET_CHANCFG) {
			donebp = cbp->cb_bp;
			donebp->b_status |= B_S_UNIT_ERR;
			goto donext;		/* Not an error */
		}
		/*
		 * FW assumed insane...
		 */
	case ZDC_REVECT:		/* FW hosed - shutdown controller */
	case ZDC_ILLVECIO:
	case ZDC_ILLPGSZ:
	case ZDC_ILLDUMPADR:
		s_ipl = p_lock(&zdcprlock, SPLZD);
		hard_error(cbp);
		zddumpstatus(cbp);	/* for ZDC_REVECT */
		v_lock(&zdcprlock, s_ipl);
		zdshutctlr(ctlrp);
		break;

	default:
		/*
		 * Unknown completion code - controller bad?
		 */
		s_ipl = p_lock(&zdcprlock, SPLZD);
		cmn_err(CE_WARN, "zd%d: Bad cb_compcode 0x%x.",
			cbp->cb_unit, cbp->cb_compcode);
		/*
		 *+ The DCC interrupted the system with an illegal completion code.
		 */
		cbp->cb_compcode = zdncompcodes - 1;	/* nice message */
		hard_error(cbp);
		v_lock(&zdcprlock, s_ipl);
		zdshutctlr(ctlrp);
		break;

	case ZDC_ILLCMD:
		s_ipl = p_lock(&zdcprlock, SPLZD);
		hard_error(cbp);
		if (cbp->cb_cmd == 0 || cbp->cb_cmd > ZDC_MAXCMD) {
			cmn_err(CE_CONT, "zd%d: cb_cmd 0x%x corrupted.\n",
				cbp->cb_unit, cbp->cb_cmd);
			v_lock(&zdcprlock, s_ipl);
			cmn_err(CE_PANIC, "zdintr: cb corrupted - command");
			/*
			 *+ The DCC interrupted the system with an illegal
			 *+ command error.  The command in the command block 
			 *+ passed to the controller was 
			 *+ corrupted.
			 */
		}
		v_lock(&zdcprlock, s_ipl);
		/*
		 * Cb looks ok - assume controller unreliable
		 */
		zdshutctlr(ctlrp);
		break;

	case ZDC_ILLMOD:
		s_ipl = p_lock(&zdcprlock, SPLZD);
		hard_error(cbp);
		switch(cbp->cb_cmd) {

		case ZDC_READ:
		case ZDC_READ_SS:
		case ZDC_READ_HDRS:
		case ZDC_LONG_READ:
		case ZDC_REC_DATA:
			if ((cbp->cb_mod & ~ZDC_NOECC) 
			     != up->zu_ioctlcb.cb_mod) {
				cmn_err(CE_CONT, "zd%d: cb_mod 0x%x corrupted.\n",
					cbp->cb_unit, cbp->cb_mod);
				v_lock(&zdcprlock, s_ipl);
				cmn_err(CE_PANIC, "zdintr: cb corrupted - modifier");
				/*
				  *+ The DCC interrupted the system with an illegal
				  *+ modifier error.  The modifier in the command 
				 *+ block passed to the controller was 
				 *+ corrupted.
			 	 */
			}
		}
		v_lock(&zdcprlock, s_ipl);
		/*
		 * Cb looks ok - assume controller unreliable
		 */
		zdshutctlr(ctlrp);
		break;

	case ZDC_ILLALIGN:
		s_ipl = p_lock(&zdcprlock, SPLZD);
		hard_error(cbp);
		if ((cbp->cb_addr & (ADDRALIGN - 1)) != 0) {
			cmn_err(CE_CONT, "zd%d: cb_addr 0x%x corrupted.\n",
				cbp->cb_unit, cbp->cb_addr);
			v_lock(&zdcprlock, s_ipl);
			cmn_err(CE_PANIC, "zdintr: cb corrupted - alignment");
			/*
			 *+ The DCC interrupted the system with an illegal
			 *+ alignment error.  The command block passed to the
			 *+ controller had incorrectly aligned addresses.
			 */
		}
		v_lock(&zdcprlock, s_ipl);
		/*
		 * Cb looks ok - assume controller unreliable
		 */
		zdshutctlr(ctlrp);
		break;

	case ZDC_ILLCNT:
		s_ipl = p_lock(&zdcprlock, SPLZD);
		hard_error(cbp);
		if (cbp->cb_count == 0 || (cbp->cb_count & (CNTMULT-1)) != 0) {
			cmn_err(CE_CONT, "zd%d: cb_count 0x%x corrupted.\n",
				cbp->cb_unit, cbp->cb_count);
			v_lock(&zdcprlock, s_ipl);
			cmn_err(CE_PANIC, "zdintr: cb corrupted - count");
			/*
			 *+ The DCC interrupted the system with an illegal
			 *+ count error.  The command block passed to the
			 *+ controller had an invalid transfer count.
			 */
		}
		v_lock(&zdcprlock, s_ipl);
		/*
		 * Cb looks ok - assume controller unreliable
		 */
		zdshutctlr(ctlrp);
		break;

	case ZDC_ILLIOV:
		s_ipl = p_lock(&zdcprlock, SPLZD);
		hard_error(cbp);
		if (((uint)cbp->cb_iovec & (IOVALIGN - 1)) != 0) {
			cmn_err(CE_CONT, "zd%d: cb_iovec 0x%x corrupted.\n",
				cbp->cb_unit, cbp->cb_iovec);
			v_lock(&zdcprlock, s_ipl);
			cmn_err(CE_PANIC, "zdintr: cb corrupted - iovec");
			/*
			 *+ The DCC interrupted the system with an illegal
			 *+ iovec error.  The command block passed to the
			 *+ controller had a corrupted iovec structure.
			 */
		}
		v_lock(&zdcprlock, s_ipl);
		/*
		 * Cb looks ok - assume controller unreliable
		 */
		zdshutctlr(ctlrp);
		break;

	case ZDC_ILLCHS:
		s_ipl = p_lock(&zdcprlock, SPLZD);
		hard_error(cbp);
		if (cbp->cb_cyl >= dd->zdd_cyls ||
		    cbp->cb_head >= dd->zdd_tracks ||
		    cbp->cb_sect >= dd->zdd_sectors + dd->zdd_spare) {
			cmn_err(CE_CONT, "zd%d: cb_diskaddr (%d, %d, %d) corrupted.\n",
				cbp->cb_unit,
				cbp->cb_cyl, cbp->cb_head, cbp->cb_sect);
			v_lock(&zdcprlock, s_ipl);
			cmn_err(CE_PANIC, "zdintr: cb corrupted - disk address");
			/*
			 *+ The DCC interrupted the system with an illegal
			 *+ disk address error.  The command block passed to
			 *+ the controller had an invalid disk address.
			 */
		}
		v_lock(&zdcprlock, s_ipl);
		/*
		 * Cb looks ok - assume controller unreliable
		 */
		zdshutctlr(ctlrp);
		break;

	case ZDC_CBREUSE:
		cmn_err(CE_CONT, "zd%d: CBREUSE error cb = 0x%x\n", cbp->cb_unit, cbp);
		cmn_err(CE_PANIC, "zdc: ZDC_CBREUSE error");
		/*
		 *+ The DCC interrupted the system with an error.  A command
		 *+ block that was already in use was reused before the previous
		 *+ command completed.
		 */

	case ZDC_ACCERR:
		val = rdslave(ctlrp->zdc_slicaddr,
			      (unchar)((up->zu_drive & 1) ? SL_G_ACCERR1
							  : SL_G_ACCERR0));
		cmn_err(CE_CONT, "zd%d: Access error on transfer starting at physical address 0x%x.\n",
				cbp->cb_unit, cbp->cb_addr);
		s_ipl = splhi();
		access_error(val);
		splx(s_ipl);
		val = ~val;
		if (((val & SLB_ATMSK) == SLB_AEFATAL) &&
		    ((val & SLB_AEIO) != SLB_AEIO)) {
			/*
			 * Uncorrectable memory error!
			 * Clear access error to restart controller and
			 * panic the system.
			 */
			wrslave(ctlrp->zdc_slicaddr,
				(unchar)((up->zu_drive & 1) ? SL_G_ACCERR1
							    : SL_G_ACCERR0),
				(unchar)0xbb);
			cmn_err(CE_PANIC, "zdc access error");
			/*
			 *+ The DCC had a fatal access error and was unable
			 *+ to recover.
			 */
		}

		/*
		 * Shutdown the channel and restart the controller.
		 */
		zdshutchan(ctlrp, up->zu_drive & 1);
		wrslave(ctlrp->zdc_slicaddr,
			(unchar)((up->zu_drive & 1) ? SL_G_ACCERR1
						    : SL_G_ACCERR0),
			(unchar)0xbb);
		break;

	case ZDC_CH_TO:			/* Retry without reset */
	case ZDC_FDL:
		s_ipl = p_lock(&zdcprlock, SPLZD);
		if (cbp->cb_errcnt++ < zdcretry) {
#ifdef	DEBUG
			zddumpcb(cbp);
#endif	/* DEBUG  */
			cmn_err(CE_WARN, "zd%ds%d: Error (%s); cmd 0x%x at (%d, %d, %d).",
				cbp->cb_unit, PARTCHR(cbp->cb_bp->b_dev),
				zd_compcodes[cbp->cb_compcode],
				cbp->cb_cmd, cbp->cb_cyl, cbp->cb_head,
				cbp->cb_sect);
			/*
			 *+ The DCC driver received the specified error
			 *+ completion status from the DCC firmware for the
			 *+ indicated drive partition.  The '%s' translates
			 *+ to the error code received.  Information about
			 *+ that error appears in this chapter.
			 *+ The final '(%d, %d, %d)' indicates the cylinder,
			 *+ head, and sector in the failing command block.
			 */
			zddumpstatus(cbp);
			v_lock(&zdcprlock, s_ipl);
			zdretry(cbp, up, ctlrp);
			return;
		}
		/*
		 * Exceeded retry count.
		 * Shutdown the channel.
		 */
		hard_error(cbp);
		zddumpstatus(cbp);
		v_lock(&zdcprlock, s_ipl);
		zdshutchan(ctlrp, up->zu_drive & 1);
		break;

	case ZDC_DRVFLT:
		/*
		 * cb_count and cb_diskaddr are unreliable on drive fault
		 * so retry entire request.
		 */
		cbp->cb_bp->b_resid = cbp->cb_bp->b_bcount - cbp->cb_transfrd;
		i = ((cbp->cb_bp->b_bcount - cbp->cb_bp->b_resid)
				>> DEV_BSHIFT) + origblkno;
		cbp->cb_diskaddr.da_cyl =
			i / (dd->zdd_sectors * dd->zdd_tracks)
								;
		i %= (dd->zdd_sectors * dd->zdd_tracks);
		cbp->cb_diskaddr.da_head = i / dd->zdd_sectors;
		cbp->cb_diskaddr.da_sect = i % dd->zdd_sectors;
		/* Fall into */
	case ZDC_SEEKERR:	/* reset and retry */
	case ZDC_SEEK_TO:
		s_ipl = p_lock(&zdcprlock, SPLZD);
		if (cbp->cb_errcnt < zdcretry) {
#ifdef	DEBUG
			zddumpcb(cbp);
#endif	/* DEBUG  */
			cmn_err(CE_WARN, "zd%ds%d: Error (%s); cmd 0x%x at (%d, %d, %d).",
				cbp->cb_unit, PARTCHR(cbp->cb_bp->b_dev),
				zd_compcodes[cbp->cb_compcode],
				cbp->cb_cmd, cbp->cb_cyl, cbp->cb_head,
				cbp->cb_sect);
			/*
			 *+ The DCC driver received the specified error
			 *+ completion status from the DCC firmware for the
			 *+ indicated drive partition.  The '%s' translates
			 *+ to the error code received.  Information about
			 *+ that error appears in this chapter.
			 *+ The final '(%d, %d, %d)' indicates the cylinder,
			 *+ head, and sector in the failing command block.
			 */
			zddumpstatus(cbp);
			v_lock(&zdcprlock, s_ipl);
			/*
			 * reset drive
			 */
			cbp->cb_state |= ZD_RESET;
			cbp->cb_cmd = ZDC_RESET;
			NUDGE_ZDC(ctlrp, cbp, s_ipl);
			return;
		}

		/*
		 * Exceeded retry count.
		 * Shutdown the drive.
		 */
		hard_error(cbp);
		zddumpstatus(cbp);
		v_lock(&zdcprlock, s_ipl);
		zdshutdrive(up);
		break;

	case ZDC_SNF:			/* Revector request */
		if (zdrevector(cbp, up, ctlrp))
			return;
		if (cbp->cb_cmd == ZDC_REC_DATA) {
			donebp = cbp->cb_bp;
			goto donext;		/* Not an error */
		}
		goto bad;
	case ZDC_HDR_ECC:		/* Reset and retry */
		if ((up->zu_mod & ZDC_NOECC) == 0) {
			if (zdrevector(cbp, up, ctlrp))
				return;
		}
		if (cbp->cb_cmd == ZDC_REC_DATA) {
			donebp = cbp->cb_bp;
			goto donext;		/* Not an error */
		}
		/* else fall into... */
	case ZDC_SO:
	case ZDC_NDS:
bad:
		s_ipl = p_lock(&zdcprlock, SPLZD);
		if (cbp->cb_errcnt == zdcretry || (up->zu_mod & ZDC_NOECC)) {
			hard_error(cbp);
			v_lock(&zdcprlock, s_ipl);
			break;
		} else {
#ifdef	DEBUG
			zddumpcb(cbp);
#endif	/* DEBUG  */
			cmn_err(CE_WARN, "zd%ds%d: Error (%s); cmd 0x%x at (%d, %d, %d).",
				cbp->cb_unit, PARTCHR(cbp->cb_bp->b_dev),
				zd_compcodes[cbp->cb_compcode],
				cbp->cb_cmd, cbp->cb_cyl, cbp->cb_head,
				cbp->cb_sect);
			/*
			 *+ The DCC driver received the specified error
			 *+ completion status from the DCC firmware for the
			 *+ indicated drive partition.  The '%s' translates
			 *+ to the error code received.  Information about
			 *+ that error appears in this chapter.
			 *+ The final '(%d, %d, %d)' indicates the cylinder,
			 *+ head, and sector in the failing command block.
			 */
		}
		zddumpstatus(cbp);
		v_lock(&zdcprlock, s_ipl);
		cbp->cb_state |= ZD_RESET;
		cbp->cb_cmd = ZDC_RESET;
		NUDGE_ZDC(ctlrp, cbp, s_ipl);
		return;

	} /* end of switch */

	/*
	 * Fail this I/O request
	 */
	donebp = cbp->cb_bp;
	donebp->b_flags |= B_ERROR;
	donebp->b_error = EIO;
	if (up->zu_state == ZU_BAD) {
		donebp->b_status |= B_S_UNIT_ERR | B_S_PERM_ERR;
	}

donext:
#ifdef	DEBUG
	if (zddebug > 2)
		zddumpcb(cbp);
#endif	/* DEBUG  */
	if (donebp->b_flags & B_IOCTL) {
		/*
		 * The combination of B_IOCTL|B_ASYNC is used
		 * to distinguish patrol seeks from other ioctl's,
		 * since they must be handled differently
		 */
		if (donebp->b_flags & B_ASYNC) {
#ifdef DEBUG
			ZDPRINTF2("<I>");
#endif /* DEBUG */
			/* asynchronous ioctl -- no one is waiting for it */
			waiter = 0;
		} else
			bcopy((caddr_t)cbp, (caddr_t)&up->zu_ioctlcb, FWCBSIZE);
	}
	/*
	 * If more requests - start them.
	 */
	s_ipl = p_lock(&up->zu_lock, SPLZD);
	DS_IODONE_TIME(up->zu_dstats);
	cbp->cb_bp = NULL;
	if (up->zu_bhead.av_forw != NULL) {
		zdstart(up);
	} else {
		if (up->zu_cbptr->cb_bp == NULL
		&&  up->zu_cbptr[1].cb_bp == NULL) {
			/*
			 * Going idle.
			 */
			if (fp_lights) {
				s_ipl = splhi();
				FP_IO_INACTIVE;
				splx(s_ipl);
			}
		}
	}
	ds_completion_stats(up->zu_dstats, donebp);
	v_lock(&up->zu_lock, s_ipl);

	if (waiter)
		biodone(donebp);
}

/*
 * hard_error
 *	report hard error
 */
static
hard_error(cbp)
	register struct cb *cbp;
{
#ifdef	DEBUG
	zddumpcb(cbp);
#endif	/* DEBUG  */
	cmn_err(CE_WARN,"zd%ds%d: Hard Error (%s); cmd 0x%x at (%d, %d, %d).",
		cbp->cb_unit, PARTCHR(cbp->cb_bp->b_dev),
		zd_compcodes[cbp->cb_compcode], cbp->cb_cmd,
		cbp->cb_cyl, cbp->cb_head, cbp->cb_sect);
	/*
	 *+ The DCC driver received an error completion status
	 *+ from the DCC firmware for the indicated drive partition
	 *+ and is treating it as a hard error.  This can be due to the 
	 *+ command reaching its retry limit or to the particular status.
	 *+ The '%s' translates
	 *+ to the error code received.  Information about
	 *+ that error appears in this chapter.
	 *+ The final '(%d, %d, %d)' indicates the cylinder,
	 *+ head, and sector in the failing command block.
	 */
}

/*
 * zdgetrpl()
 *	Search bad block list for replacement sector for bad sector (SNF).
 *
 * Return:
 *	pointer to replacement sector
 *	otherwise NULL pointer if no replacement sector found.
 *	
 */
static struct diskaddr *
zdgetrpl(badsect, zdp)
	register struct diskaddr *badsect;
	register struct zdbad	 *zdp;
{
	register struct bz_bad	 *bb;

	if (zdp == NULL)		/* if formatting, bbl may be NULL */
		return(NULL);
	for (bb = zdp->bz_bad; bb < &zdp->bz_bad[zdp->bz_nsnf]; bb++) {
		if (bb->bz_cyl < badsect->da_cyl)
			continue;
		if (bb->bz_cyl > badsect->da_cyl)
			break;
		/* cylinder matched */
		if (bb->bz_head < badsect->da_head)
			continue;
		if (bb->bz_head > badsect->da_head)
			break;
		/* head matched */
		if (bb->bz_sect == badsect->da_sect)
			return (&bb->bz_rpladdr);
	}
	return (NULL);
}

/*
 * zdrevector
 *	Attempt to revector the sector-not-found.
 * If the sector-not-found is in the bad block list perform the I/O on
 * its replacement sector. Then continue with the following sector if
 * necessary.
 * 
 * Return values:
 *	 0 - No replacement sector in bad block list.
 *	 1 - Replacement sector found in bad block list.
 */
static int
zdrevector(cbp, up, ctlrp)
	register struct cb *cbp;
	struct	zd_unit	*up;
	struct zdc_ctlr *ctlrp;
{
	register struct zdcdd	 *dd;		/* disk description */
	register struct diskaddr *replcmnt;	/* replacement address */
	register int	count;			/* # bytes of iovecs */
	int	transfrd;			/* bytes already transferred */
	int	cbcount;
	ulong	*from;
	spl_t	s_ipl;

#ifdef	DEBUG
	if (zddebug || zd_debug)
		CPRINTF("zd%d: revectoring (%d, %d, %d)", cbp->cb_unit,
			cbp->cb_cyl, cbp->cb_head, cbp->cb_sect);
#endif	/* DEBUG  */
	/*
	 * Flush tlb to protect accesses of zu_zdbad from interrupt
	 * This is because of the problem with memory allocated
	 * with wmemall, which may cause page table discrepancies
	 * across processors.
	 */
	FLUSH_TLB();

	/*
	 * Determine if bad block is in bad block list.
	 */
	replcmnt = zdgetrpl(&cbp->cb_diskaddr, up->zu_zdbad);
	if (replcmnt == NULL) {
		/*
		 * Not in bad block list. Caller should retry
		 * bad block.
		 */
#ifdef	DEBUG
		if (zddebug || zd_debug)
			CPRINTF(": not in table.\n");
#endif	/* DEBUG  */
		return (0);
	}

	/*
	 * Save state of current request, do new request for replacement.
	 */
	transfrd = cbp->cb_bp->b_bcount - cbp->cb_bp->b_resid;
	/* adjust for previously transferred sectors */
	transfrd -= cbp->cb_transfrd;
	from = cbp->cb_iovstart
			+ (((cbp->cb_addr & CLOFSET) + transfrd) >> CLSHIFT);
	cbp->cb_transfrd += transfrd;
	cbp->cb_addr += transfrd;

	/*
	 * figure number of iovecs to copy down.
	 */
	count = 0;
	cbcount = cbp->cb_count;
	if (cbp->cb_addr & CLOFSET) {
		++count;
		cbcount -= CLBYTES - (cbp->cb_addr & CLOFSET);
	}
	count += (cbcount + CLOFSET) >> CLSHIFT;
	count *= sizeof(ulong *);
	cbp->cb_iovec = cbp->cb_iovstart;
	bcopy((caddr_t)from, (caddr_t)cbp->cb_iovstart, (unsigned)count);

	dd = (up->zu_drive & 1) ? &ctlrp->zdc_chanB : &ctlrp->zdc_chanA;
	if (cbp->cb_count != DEV_BSIZE) {
		cbp->cb_state |= ZD_REVECTOR;
		if (++cbp->cb_sect == dd->zdd_sectors) {
			cbp->cb_sect = 0;
			if (++cbp->cb_head == dd->zdd_tracks) {
				cbp->cb_head = 0;
				cbp->cb_cyl++;
			}
		}
		*(int *)&cbp->cb_contaddr = *(int *)&cbp->cb_diskaddr;
		cbp->cb_contiovsz = count;
	}
	cbp->cb_count = DEV_BSIZE;
	cbp->cb_diskaddr = *replcmnt;
	cbp->cb_psect = zdgetpsect(&cbp->cb_diskaddr, dd);
#ifdef	DEBUG
	if (zddebug || zd_debug)
		CPRINTF(" to (%d, %d, %d).\n",
			cbp->cb_cyl, cbp->cb_head, cbp->cb_sect);
#endif	/* DEBUG  */

	/*
	 * Nudge controller.
	 */
	NUDGE_ZDC(ctlrp, cbp, s_ipl);
	return (1);
}

/*
 * zdcontinue
 *	Continue rest of I/O request after the revectored sector.
 */
static
zdcontinue(cbp, up, ctlrp)
	register struct cb	 *cbp;
	struct zd_unit *up;
	register struct	zdc_ctlr *ctlrp;
{
	ulong	*from;
	spl_t	s_ipl;

	/*
	 * Revectoring completed successfully.
	 * Restart rest of I/O request.
	 */
	cbp->cb_state &= ~ZD_REVECTOR;
	cbp->cb_bp->b_resid -= DEV_BSIZE;
	cbp->cb_transfrd += DEV_BSIZE;
	cbp->cb_errcnt = 0;
	cbp->cb_count = cbp->cb_bp->b_resid;
	cbp->cb_iovec = cbp->cb_iovstart;
	from = cbp->cb_iovstart
			+ (((cbp->cb_addr & CLOFSET) + DEV_BSIZE) >> CLSHIFT);
	cbp->cb_addr += DEV_BSIZE;
	if (from != cbp->cb_iovstart) {
		/*
		 * Must start with next iovector.
		 */
		bcopy((caddr_t)from, (caddr_t)cbp->cb_iovstart,
			(unsigned)(cbp->cb_contiovsz - sizeof(ulong *)));
	}
	*(int *)&cbp->cb_diskaddr = *(int *)&cbp->cb_contaddr;
	cbp->cb_psect = zdgetpsect(&cbp->cb_diskaddr,
					(up->zu_drive & 1) ? &ctlrp->zdc_chanB
							   : &ctlrp->zdc_chanA);
	/*
	 * Nudge controller.
	 */
	NUDGE_ZDC(ctlrp, cbp, s_ipl);
}

/*
 * zdretry
 *	Retry the request.
 * If the request is a normal read/write request then retry the command
 * from where the error occurred. If B_IOCTL request, retry from the
 * initial CB.
 */
static
zdretry(cbp, up, ctlrp)
	register struct cb *cbp;
	struct	zd_unit  *up;
	struct	zdc_ctlr *ctlrp;
{
	register int count;		/* # iovecs remaining */
	register int transfrd;		/* # bytes transferred */
	register int cbcount;
	ulong	*from;			/* 1st retry iovec */
	spl_t	s_ipl;

	/*
	 * Retry job
	 */
	if (cbp->cb_bp->b_flags & B_IOCTL) {
		/*
		 * copy 1st half of cb (what fw will see)
		 */
		bcopy((caddr_t)&up->zu_ioctlcb, (caddr_t)cbp, FWCBSIZE);
	} else {
		cbp->cb_count = cbp->cb_bp->b_resid;
		cbp->cb_cmd = (cbp->cb_bp->b_flags & B_READ) ? ZDC_READ : ZDC_WRITE;
		cbp->cb_sect &= ~ZD_AUTOBIT;	 /* drop Auto-revetor bit */
		cbp->cb_psect = zdgetpsect(&cbp->cb_diskaddr,
					(up->zu_drive & 1) ? &ctlrp->zdc_chanB
							   : &ctlrp->zdc_chanA);
		transfrd = cbp->cb_bp->b_bcount - cbp->cb_count;
		transfrd -= cbp->cb_transfrd;
		from = cbp->cb_iovstart
			+ (((cbp->cb_addr & CLOFSET) + transfrd) >> CLSHIFT);
		cbp->cb_transfrd += transfrd;
		cbp->cb_addr += transfrd;

		/*
		 * determine number of iovecs remaining.
		 */
		count = 0;
		cbcount = cbp->cb_count;
		if (cbp->cb_addr & CLOFSET) {
			++count;
			cbcount -= CLBYTES - (cbp->cb_addr & CLOFSET);
		}
		count += (cbcount + CLOFSET) >> CLSHIFT;

		/*
		 * Copy down iovecs to 32 byte aligned boundary.
		 * That is, cb_iovstart.
		 */
		bcopy((caddr_t)from, (caddr_t)cbp->cb_iovstart,
			(unsigned)(count * sizeof(ulong *)));
	}
	cbp->cb_iovec = cbp->cb_iovstart;

	/*
	 * Nudge controller.
	 */
	NUDGE_ZDC(ctlrp, cbp, s_ipl);
}

/*
 * zdc_cb_cleanup
 *	Cleanup (cancel) jobs active on disabled controller.
 * Called as timeout routine to allow completed CB's to drain.
 * That is, those for which a completion interrupt was sent just
 * before the error interrupt occurred.
 */
zdc_cb_cleanup(ctlrp)
	register struct zdc_ctlr *ctlrp;
{
	register struct cb	 *cbp;
	register struct buf	 *bp;
	spl_t	s_ipl;

	/*
	 * Controller dead. Fail all active I/O requests.
	 */
	cbp = ctlrp->zdc_cbp;
	for (; cbp < &ctlrp->zdc_cbp[NCBPERZDC]; cbp++) {
		if (cbp->cb_bp != NULL) {
			bp = cbp->cb_bp;
			bp->b_flags |= B_ERROR;
			bp->b_error = EIO;
			bp->b_resid = bp->b_bcount;
			bp->b_status |= B_S_XFER_ERR|B_S_UNIT_ERR|B_S_PERM_ERR;
			/* 
			 * The zd_unit structure in not locked here.
			 * Normally it would be.  Since we are just
			 * cleaning up before declaring the drive dead
			 * and the stats are questionable anyway, it is
			 * not being locked.
			 */
			DS_IODONE_TIME(zdunit[cbp->cb_unit].zu_dstats);
			ds_completion_stats(zdunit[cbp->cb_unit].zu_dstats,
					    bp);
			biodone(bp);
			bp = NULL;
			/*
			 * Decrement I/O activity
			 */
			if (fp_lights) {
				s_ipl = splhi();
				FP_IO_INACTIVE;
				splx(s_ipl);
			}
		}
	}
}

/*
 * zdc_error
 *	Controller error interrupt.
 */
void
zdc_error(vector)
	unchar	vector;
{
	register int	val;
	register struct zdc_ctlr *ctlrp;
	spl_t	s_ipl;

	ctlrp = &zdctrlr[vector - base_err_intr];
	val = rdslave(ctlrp->zdc_slicaddr, SL_Z_STATUS);
	if (((val & SLB_ZPARERR) != SLB_ZPARERR) &&
	    ((val & ZDC_READY) == ZDC_READY)) {
		cmn_err(CE_WARN, "zdc%d - stray error interrupt!", 
			ctlrp - zdctrlr);
		/*
		 *+ The DCC driver received a stray controller
		 *+ error interrupt.  The interrupt is ignored.
		 */
		return;
	}
	cmn_err(CE_WARN, "zdc%d: controller interrupt - SL_Z_STATUS == 0x%x.",
		ctlrp - zdctrlr, val);
	/*
	 *+ The DCC driver received a controller error interrupt.
	 *+ The driver will shut down the controller.
	 */
	if ((val & ZDC_ERRMASK) == ZDC_OBCB) {
		/*
		 * Assume stray interrupt happened and tell fw to continue.
		 */
		s_ipl = splhi();	/* in case zdc_err_bin not splhi() */
		mIntr(ctlrp->zdc_slicaddr, CLRERRBIN, 0xbb);
		splx(s_ipl);
		return;
	}

	splx(SPLZD);		/* Do the rest at SPLZD */

	/*
	 * Fail all queued I/O requests.
	 */
	zdshutctlr(ctlrp);

	/*
	 * Allow completed CBs to drain then
	 * fail all other "active" I/O requests.
	 */
	(void) timeout(zdc_cb_cleanup, (caddr_t)ctlrp, 5 * hz);
}

/*
 * zdshutctlr
 *	Fail all I/O requests queued to the drives on this controller.
 *	Currently active I/O requests will finish asyncronously.
 *	The controller is marked as ZDC_DEAD and all units on the
 *	controller are marked as ZU_BAD.
 */
static
zdshutctlr(ctlrp)
	register struct zdc_ctlr *ctlrp;
{
	register struct cb *cbp;

	cmn_err(CE_NOTE, "zdc%d: Controller disabled.", ctlrp - zdctrlr);
	/*
	 *+ The DCC driver is disabling the specified controller.
	 */

	cbp = ctlrp->zdc_cbp;
	for (; cbp < &ctlrp->zdc_cbp[NCBPERZDC]; cbp += NCBPERDRIVE) {
		if (cbp->cb_unit >= 0)
			zdshutdrive(&zdunit[cbp->cb_unit]);
	}
	ctlrp->zdc_state = ZDC_DEAD;
}

/*
 * zdshutchan
 *	Fail all I/O requests queued to the drives on this channel.
 *	Currently active I/O requests will finish asyncronously.
 *	All units on the channel are marked as ZU_BAD.
 */
static
zdshutchan(ctlrp, channel)
	register struct zdc_ctlr *ctlrp;
	int	channel;		/* 0 == Channel A, 1 == channel B */
{
	register struct cb *cbp;

	cmn_err(CE_NOTE, "zdc%d: Channel %c disabled.", ctlrp - zdctrlr,
			(channel == 0) ? 'A' : 'B');
	/*
	 *+ The DCC driver is disabling the specified channel.
	 */

	cbp = &ctlrp->zdc_cbp[channel * NCBPERDRIVE];
	for (; cbp < &ctlrp->zdc_cbp[NCBPERZDC]; cbp += 2 * NCBPERDRIVE) {
		if (cbp->cb_unit >= 0)
			zdshutdrive(&zdunit[cbp->cb_unit]);
	}
}

/*
 * zdshutdrive
 *	Fail all I/O requests queued to the drive.
 *	Currently active I/O requests will finish asyncronously.
 *	The drive is marked as ZU_BAD.
 */
static
zdshutdrive(up)
	register struct	zd_unit *up;
{
	register struct	buf	*bp, *nextbp;
	spl_t	s_ipl;

	/*
	 * Lock the device and fail all jobs currently queued.
	 */
	s_ipl = p_lock(&up->zu_lock, SPLZD);
	if (up->zu_state == ZU_BAD) {
		v_lock(&up->zu_lock, s_ipl);
		return;
	}
	cmn_err(CE_NOTE, "zd%d: Drive disabled.", up - zdunit);
	/*
	 *+ The DCC driver is disabling the specified drive.
	 */
	bp = up->zu_bhead.av_forw;
	while (bp != NULL) {
		bp->b_flags |= B_ERROR;
		bp->b_error = EIO;
		bp->b_resid = bp->b_bcount;
		nextbp = bp->av_forw;
		bp->b_status |= B_S_UNIT_ERR | B_S_XFER_ERR | B_S_PERM_ERR;
		DS_IODONE_TIME(up->zu_dstats);
		ds_completion_stats(up->zu_dstats, bp);
		biodone(bp);
		bp = nextbp;
	}
	up->zu_bhead.av_forw = NULL;
	up->zu_state = ZU_BAD;
	if (up->zu_flags & ZUF_PATROL) {	/* turn off patrol seeks */
		up->zu_flags &= ~ZUF_PATROL;
		v_lock(&up->zu_lock, s_ipl);
		s_ipl = p_lock(&patrollock, SPLZD);
		patrolcount--;
		v_lock(&patrollock, s_ipl);
	} else
		v_lock(&up->zu_lock, s_ipl);

	/*
	 * inform higher-level drivers that the drive has gone
	 * away.  First arg is parent, second is child.  Think of it as sending
	 * the parent the message "SHUTDOWN child"--that is "Shutdown the drive
	 * that failed.
	 */
	if ( up->zu_physdev->d_parent )
		DEV_SHUTDOWN(up->zu_physdev->d_parent, up->zu_physdev);
	/*
	 * Turn on error light on front panel.
	 */
	if (fp_lights) {
		s_ipl = splhi();
		FP_IO_ERROR;
		splx(s_ipl);
	}
	/*
	 * It had been online and usable.
	 */
	disk_offline();
}

/*
 * zdminphys - correct for too large a request.
 *
 * Note correction for non-cluster-aligned transfers.
 */
/*ARGSUSED*/
zdminphys(dp, bp)
	struct dev *dp;
	struct buf *bp;
{
	if (bp->b_bcount > ((zdc_iovpercb - 1) * CLBYTES))
		bp->b_bcount = (zdc_iovpercb - 1) * CLBYTES;
}

/*ARGSUSED*/
zdread(dp, uio, flags)
	struct dev *dp;
	struct uio *uio;
{
	int err, diff;
	off_t lim;
	dev_t dev = dp->d_dev;

	lim = zdunit[ZDUNIT(dev)].zu_drive_size;
	lim <<= DEV_BSHIFT;
	err = physck(lim, uio, B_READ, &diff);
	if (err != 0) {
		if (err == -1)	/* not an error, but request of 0 bytes */
			err = 0;
		return (err);
	}
	if (flags & IO_ASYNC) {
		err = aphysio(dp, B_READ, uio, diff, flags);
	} else {
		err = physio(dp, (struct buf *)0, B_READ, uio);
		uio->uio_resid += diff;
	}
	return (err);
}

/*ARGSUSED*/
zdwrite(dp, uio, flags)
	struct dev *dp;
	struct uio *uio;
{
	int err, diff;
	off_t lim;
	dev_t dev = dp->d_dev;

	lim = zdunit[ZDUNIT(dev)].zu_drive_size;
	lim <<= DEV_BSHIFT;
	err = physck(lim, uio, B_WRITE, &diff);
	if (err != 0) {
		if (err == -1)	/* not an error, but request of 0 bytes */
			err = 0;
		return (err);
	}
	if (flags & IO_ASYNC) {
		err = aphysio(dp, B_WRITE, uio, diff, flags);
	} else {
		err = physio(dp, (struct buf *)0, B_WRITE, uio);
		uio->uio_resid += diff;
	}
	return (err);
}

#ifdef DEBUG
/*
 * zd_sync()
 *	This is a debug test routine only.
 */
void
zd_sync()
{
	cmn_err( CE_NOTE, "zd_sync");
}
#endif /* DEBUG */

/*
 * zdioctl
 *	This routine provides support for various IOCTL's,
 *	most of which are used to support online formatting.
 *	Those which change the configuration or state of the
 *	device are only accessible to the superuser.
 */
/*ARGSUSED*/
zdioctl(dp, cmd, uaddr, flag)
	struct	dev *dp;
	int	cmd;
	caddr_t uaddr;
	int	flag;
{
	struct cb cb;
	register struct cb *cbp = &cb;	/* cb argument */
	register struct buf *bp;	/* ioctl buffer */
	register struct zd_unit *up;	/* unit structure */
	register struct zd_unit *tup;	/* temporary unit structure address */
	dev_t	dev = dp->d_dev;
	int	error;
	int	firstsect;
	struct zddev zddev;
	unchar	ustate;
	struct zdcdd *dd;
	struct zdc_ctlr *ctlrp;
	int setokay = 0;
	int chan, i;
	spl_t	s_ipl;

	up = &zdunit[ZDUNIT(dev)];
	bp = &up->zu_ioctl;
	dd = (up->zu_drive & 1) ? &zdctrlr[up->zu_ctrlr].zdc_chanB
				: &zdctrlr[up->zu_ctrlr].zdc_chanA;

	switch(cmd) {

	case RIOFIRSTSECT:	/* get first user-usable sector address */
		firstsect = DEV_FIRSTSECT(dp);
		if (firstsect < 0) {
			return(EIO);
		}
		if (copyout((caddr_t)&firstsect, (caddr_t)uaddr, sizeof(int)))
			return(EFAULT);
		return(0);

	case RIODRIVER:		/* success indicates we're a driver! */
		return(0);

	case RIOPATROL:		/* start patrol seeks for this drive */
		if (!suser())
			return(EACCES);
 		if (up->zu_state == ZU_NO_RW) {
 			cmn_err(CE_WARN, "zd%d: patrol seeks not allowed.",
 			up - zdunit);
 			/*
			  *+ An attempt was made to start patrol seeks on a disk
			  *+ where read and write operations are disallowed.
			  *+ This usually indicates that the drive is not
			  *+ formatted.  Patrol seeks should be restarted after
			  *+ the drive has been formatted.
 			 */
 			return(0);
 		}
		s_ipl = p_lock(&up->zu_lock, SPLZD);
		if (!(up->zu_flags & ZUF_PATROL)) {
			up->zu_flags &= ~ZUF_SEEKERR;
			up->zu_flags |= ZUF_PATROL;
			v_lock(&up->zu_lock, s_ipl);
			s_ipl = p_lock(&patrollock, SPLZD);
			patrolcount++;
			if (!patrol_on) {
				/* start timeout */
				patrol_on = 1;
				(void)timeout(zdpatrol, up, PATROLTIME/2*HZ);
			}
			v_lock(&patrollock, s_ipl);
		} else
			v_lock(&up->zu_lock, s_ipl);
		return(0);

	case RIONPATROL:	/* disable patrol seeks for this drive */
		if (!suser())
			return(EACCES);
		s_ipl = p_lock(&up->zu_lock, SPLZD);
		if (up->zu_flags & ZUF_PATROL) {
			up->zu_flags &= ~ZUF_PATROL; 
			v_lock(&up->zu_lock, s_ipl);
			s_ipl = p_lock(&patrollock, SPLZD);
			--patrolcount;
			v_lock(&patrollock, s_ipl);
		} else
			v_lock(&up->zu_lock, s_ipl);
		return(0);

#ifdef DEBUG
#ifndef ZIOCSHUTDOWN
#define ZIOCSHUTDOWN	(ZIOC | 255)
#endif /* !ZIOCSHUTDOWN */
#ifndef ZIOCISHUT
#define ZIOCISHUT	(ZIOC | 254)
#endif /* !ZIOCISHUT */
#ifndef ZIOC_STAT_TEST
#define ZIOC_STAT_TEST	(ZIOC | 253)
#endif /* !ZIOC_STAT_TEST */
#ifndef ZIOC_SYNCON
#define ZIOC_SYNCON	(ZIOC | 252)
#endif /* !ZIOC_SYNCON */
#ifndef ZIOC_SYNCOFF
#define ZIOC_SYNCOFF	(ZIOC | 251)
#endif /* !ZIOC_SYNCOFF */
#ifndef ZIOC_BEXPRESS
#define ZIOC_BEXPRESS	(ZIOC | 250)
#endif /* ZIOC_BEXPRESS */
	case ZIOCSHUTDOWN:
		printf( "zdioctl: ZIOCSHUTDOWN, dev_t = 0x%x, dev @ 0x%x\n",
		       dev, dp);
		zdshutdrive(up);
		return(0);
	case ZIOCISHUT:
		printf( "zdioctl: ZIOCISHUT, dev_t = 0x%x, dev @ 0x%x\n",
		       dev, dp);
		zd_intr_shutdown = up;
		return(0);
	case ZIOC_STAT_TEST:
		zd_status_returns++;
		return(0);
	case ZIOC_SYNCON:
		d_sync_add(zd_sync);
		return(0);
	case ZIOC_SYNCOFF:
		return(d_sync_remove(zd_sync)? ENXIO : 0);
	case ZIOC_BEXPRESS:
		zd_express++;
		return(0);
#endif /* DEBUG */
	case ZIOSEVERE:			/* turn on severe burn-in mode */
#ifdef DEBUG
		ZDPRINTF("zdioctl: ZIOSEVERE\n");
#endif /* DEBUG */
		if (!suser() || !(up->zu_flags & ZUF_EXCLUSIVE))
			return(EACCES);
		/*
		 * Inhibit ZDC ECC correction
		 */
		up->zu_mod |= ZDC_NOECC;
		return(0);

	case ZIONSEVERE:		/* turn off severe burn-in mode */
#ifdef DEBUG
		ZDPRINTF("zdioctl: ZIONSEVERE\n");
#endif /* DEBUG */
		if (!suser())
			return(EACCES);
		up->zu_mod &= ~ZDC_NOECC;
		return(0);
	
	case ZIODEVDATA:		/* return device-specific data */
#ifdef DEBUG
		ZDPRINTF("zdioctl: ZIODEVDATA\n");
#endif /* DEBUG */
		zddev.zd_drive = up->zu_drive;
		zddev.zd_ctlr = up->zu_ctrlr;
		zddev.zd_cfg = up->zu_cfg;
		zddev.zd_state = up->zu_state;
		if (copyout((caddr_t)&zddev, (caddr_t)uaddr, 
			     sizeof(struct zddev)))
			return(EFAULT);
		return(0);

	case ZIOSETSTATE:		/* set device state */
#ifdef DEBUG
		ZDPRINTF("zdioctl: ZIOSETSTATE\n");
#endif /* DEBUG */
		if (!suser() || !(up->zu_flags & ZUF_EXCLUSIVE))
			return(EACCES);
		if (copyin((caddr_t)uaddr, (caddr_t)&ustate, sizeof(unchar)))
			return(EFAULT);
		if ((int)ustate < 0 || (int)ustate > ZUMAXSTATE) {
#ifdef DEBUG
			ZDPRINTF("zdioctl: bad state %d\n", ustate);
#endif /* DEBUG */
			return(EIO);
		}
		up->zu_state = ustate;
		return(0);
		
	case ZIOGERR:			/* return last completion status */
#ifdef DEBUG
		ZDPRINTF("zdioctl: ZIOGERR\n");
#endif /* DEBUG */
		if (copyout((caddr_t)&up->zu_l_compcode, (caddr_t)uaddr,
			    sizeof(int)))
			return(EFAULT);
		return(0);

	case ZIOSETBBL:			/* get bad block list from userland */
		{
		struct zdbad *zdp, *tzdp;
		struct bz_bad *fbzp, *tbzp;
		int size, zdpsize;
		caddr_t addr;

#ifdef DEBUG
		ZDPRINTF("zdioctl: ZIOSETBBL\n");
#endif /* DEBUG */
		if (!suser() || !(up->zu_flags & ZUF_EXCLUSIVE))
			return(EACCES);
		/*
		 * First get just first block to determine how much
		 * space we need.
		 */
		zdp = (struct zdbad *)wmemall(DEV_BSIZE, 1);
		if (copyin((caddr_t)uaddr, (caddr_t)zdp, DEV_BSIZE))
			return(EFAULT);
		size = (zdp->bz_nelem * sizeof(struct bz_bad))
			+ sizeof(struct zdbad) - sizeof(struct bz_bad);
		size = howmany(size, DEV_BSIZE);
		if (size > ((dd->zdd_sectors - ZDD_NDDSECTORS) >> 1)) {
			/*
			 * Bad block list too big -- must be invalid! 
			 */
#ifdef DEBUG
			ZDPRINTF("zdioctl: bbl too big\n");
#endif /* DEBUG */
			wmemfree((caddr_t)zdp, DEV_BSIZE);
			return(EIO);
		}
		/*
		 * Allocate full size, copy 1st block to it, then
		 * copy remainder from userland.
		 */
		zdpsize = size * DEV_BSIZE;
		addr = wmemall(zdpsize, 1);
		bcopy((caddr_t)zdp, addr, (unsigned)DEV_BSIZE);
		wmemfree((caddr_t)zdp, DEV_BSIZE);
		zdp = (struct zdbad *)addr;
		if (copyin((caddr_t)uaddr+DEV_BSIZE, (caddr_t)addr+DEV_BSIZE,
			   (uint)(zdpsize-DEV_BSIZE)))
			return(EFAULT);
		/*
		 * Confirm data integrity via checksum.
		 */
		size = (zdp->bz_nelem * sizeof(struct bz_bad)) / sizeof(long);
		if (zdp->bz_csn != getchksum((long *)zdp->bz_bad, size,
				(long)(zdp->bz_nelem ^ zdp->bz_nsnf))) {
#ifdef DEBUG
			ZDPRINTF("zdioctl: bad checksum\n");
#endif /* DEBUG */
			wmemfree((caddr_t)zdp, zdpsize);
			return(EIO);
		}
		/*
		 * Copy only BZ_SNF entries into unit's bad block list.
		 */
		size = (zdp->bz_nsnf * sizeof(struct bz_bad))
			+ sizeof(struct zdbad) - sizeof(struct bz_bad);
		size = roundup(size, DEV_BSIZE);
		tzdp = (struct zdbad *)wmemall(size, 1);
#ifdef DEBUG
		bzero((caddr_t)tzdp, (uint)size);
#endif /* DEBUG */
		*tzdp = *zdp;
		tbzp = tzdp->bz_bad;
		for (fbzp = zdp->bz_bad; fbzp < &zdp->bz_bad[zdp->bz_nelem]; fbzp++) {
			if (fbzp->bz_rtype == BZ_SNF)
				*tbzp++ = *fbzp;
		}
		wmemfree((caddr_t)zdp, zdpsize);
		if (up->zu_zdbad) {
			/*
			 * Free memory used for former bad block list.
			 */
			zdp = up->zu_zdbad;
			size = (zdp->bz_nsnf * sizeof(struct bz_bad))
				+ sizeof(struct zdbad) - sizeof(struct bz_bad);
			size += roundup(size, DEV_BSIZE);
			wmemfree((caddr_t)zdp, size);
		}

		up->zu_zdbad = tzdp;
		return(0);
		}
	
	case ZIOCBCMD:			/* send zd command block to device */
		if (copyin((caddr_t)uaddr, (caddr_t)cbp, sizeof(struct cb)))
			return(EFAULT);
		switch(cbp->cb_cmd) {

		case ZDC_SET_CHANCFG:	/* set disk channel configuration */
#ifdef DEBUG
			ZDPRINTF("zdioctl: ZDC_SET_CHANCFG\n");
#endif /* DEBUG */
			if (!suser() || !(up->zu_flags & ZUF_EXCLUSIVE))
				return(EACCES);
			/*
			 * Grab lock and determine if SET_CHANCFG is
			 * legal.  It is only allowed if the channel
			 * configuration is not set, or if this drive
			 * is the only formatted drive on the
			 * channel.
			 */
			ctlrp = &zdctrlr[up->zu_ctrlr];
			s_ipl = p_lock(&ctlrp->zdc_ctlrlock, SPLHI);

			/*
			 * setokay is both a flag and a state mechanism.
			 * As a flag, 0 means SET_CHANCFG not permitted,
			 * non-zero means permitted.
			 * As state:
			 *	0 - SET_CHANCFG not done
			 *	1 - SET_CHANCFG done, chancfg previously set
			 *     -1 - SET_CHANCFG done, chancfg not previously set
			 */
			setokay = 1;
			chan = up->zu_drive & 1;
			if (ctlrp->zdc_chan_owner[chan] == up) {
				/*
				 * Channel has been reserved by
				 * this drive for exclusive use.
				 * Let it go.  It must also take
				 * responsibility for restoring.
				 */
				dd->zdd_sectors = -1;
				setokay = -1;
			} else if (dd->zdd_sectors == 0) {
				/*
				 * Channel configuration isn't set.
				 * Use zdd_sectors as a flag to lock
				 * out any other device attempting to
				 * SET_CHANCFG while this one is in
				 * progress.
				 */
				dd->zdd_sectors = -1;
				setokay = -1;
			} else if (dd->zdd_sectors == (unsigned) -1) {
				setokay = 0;
			} else if ((up->zu_cfg & (ZD_FORMATTED|ZD_MATCH))
				 != (ZD_FORMATTED|ZD_MATCH)) {
				setokay = 0;
			} else {
				for (i = (up->zu_drive+2)%(ZDC_MAXDRIVES/2);
				     i != up->zu_drive;
				     i = (i + 2)%(ZDC_MAXDRIVES/2)) {
					if ((ctlrp->zdc_drivecfg[i]
					    & (ZD_FORMATTED|ZD_MATCH))
					    == (ZD_FORMATTED|ZD_MATCH)) {
						setokay = 0;
					}
				}
			}
			v_lock(&ctlrp->zdc_ctlrlock, s_ipl);
			if (!setokay)
				return(EACCES);
			/* fall through */
		case ZDC_GET_CHANCFG:	/* return disk channel configuration */
#ifdef DEBUG
			if (cbp->cb_cmd == ZDC_GET_CHANCFG)
				ZDPRINTF("zdioctl: ZDC_GET_CHANCFG\n");
#endif /* DEBUG */
			if (cbp->cb_count != sizeof(struct zdcdd) ||
			    (cbp->cb_addr & (ADDRALIGN-1)) != 0) {
				if (setokay < 0)
					dd->zdd_sectors = 0;
				return (EINVAL);
			}
			if (!useracc((char *)cbp->cb_addr, 
						(uint)cbp->cb_count, B_WRITE))
				return (EFAULT);
			bufalloc(bp);
			bp->b_flags = (cbp->cb_cmd == ZDC_GET_CHANCFG) ? B_READ
					: B_WRITE;
			bp->b_bcount = cbp->cb_count;
			bp->b_un.b_addr = (caddr_t)cbp->cb_addr;
			break;

		case ZDC_WRITE_SS:	/* write special sector */
#ifdef DEBUG
			ZDPRINTF("zdioctl: ZDC_WRITE_SS\n");
#endif /* DEBUG */
			if (!suser() || !(up->zu_flags & ZUF_EXCLUSIVE))
				return(EACCES);
			/* fall through */
		case ZDC_READ_SS:	/* read special sector */
#ifdef DEBUG
			if (cbp->cb_cmd == ZDC_READ_SS)
				ZDPRINTF("zdioctl: ZDC_READ_SS\n");
#endif /* DEBUG */
			if (dd->zdd_sectors == 0) {
#ifdef DEBUG
				ZDPRINTF("no READ_SS - zdd_sectors is 0\n");
#endif /* DEBUG */
				return(EACCES);
			}
			if (cbp->cb_count != ZDD_SS_SIZE ||
			    (cbp->cb_addr & (ADDRALIGN-1)) != 0 ||
			    cbp->cb_diskaddr.da_cyl != 0 ||
			    cbp->cb_diskaddr.da_head != 0 ||
			    cbp->cb_diskaddr.da_sect >= ZDD_NDDSECTORS)
				return (EINVAL);
			if (!useracc((char *)cbp->cb_addr,
						(uint)cbp->cb_count, B_WRITE))
				return (EFAULT);
			bufalloc(bp);
			bp->b_flags = (cbp->cb_cmd == ZDC_READ_SS) ? B_READ
					: B_WRITE;
			bp->b_bcount = cbp->cb_count;
			bp->b_un.b_addr = (caddr_t)cbp->cb_addr;
			break;

		case ZDC_FMT_SS:	/* format special sector */
#ifdef DEBUG
			ZDPRINTF("zdioctl: ZDC_FMT_SS\n");
#endif /* DEBUG */
			if (!suser() || !(up->zu_flags & ZUF_EXCLUSIVE)
				     || dd->zdd_sectors == 0) {
#ifdef DEBUG
				if (dd->zdd_sectors == 0)
					ZDPRINTF("FMT_SS err: zdd_sectors is 0\n");
#endif /* DEBUG */
				return(EACCES);
			}	
			if (cbp->cb_count != CNTMULT 
			    || (cbp->cb_addr & (ADDRALIGN-1)) != 0
			    || cbp->cb_diskaddr.da_cyl != 0
			    || cbp->cb_diskaddr.da_head != 0
			    || cbp->cb_diskaddr.da_sect >= ZDD_NDDSECTORS
			    || cbp->cb_iovec !=0)
				return (EINVAL);
			if (!useracc((char *)cbp->cb_addr,
						(uint)cbp->cb_count, B_WRITE))
				return (EFAULT);
			bufalloc(bp);
			bp->b_flags = B_WRITE;
			bp->b_bcount = cbp->cb_count;
			bp->b_un.b_addr = (caddr_t)cbp->cb_addr;
			break;

		case ZDC_WHDR_WDATA:	/* write sector header data */
		case ZDC_REC_DATA:	/* recover data from sector */
		case ZDC_READ:		/* (normal) read */
		case ZDC_WRITE:		/* (normal) write */
			if (!suser() || !(up->zu_flags & ZUF_EXCLUSIVE)
				     || dd->zdd_sectors == 0) {
#ifdef DEBUG
				if (dd->zdd_sectors == 0)
					ZDPRINTF("WHDR (cmd = 0x%x) - zdd_sectors is 0\n",
						cbp->cb_cmd);
#endif /* DEBUG */
				return(EACCES);
			}
			if (cbp->cb_cmd == ZDC_WHDR_WDATA) {
				if (cbp->cb_count != (CNTMULT+dd->zdd_ddc_regs.dr_sector_bc))
					return(EINVAL);
			} else {
				if (cbp->cb_count & (DEV_BSIZE-1))
					return(EINVAL);
			}
			if ((cbp->cb_addr & (ADDRALIGN-1))
			    || cbp->cb_cyl >= dd->zdd_cyls
			    || cbp->cb_head >= dd->zdd_tracks
			    || cbp->cb_sect >= (dd->zdd_sectors + dd->zdd_spare)
			    || cbp->cb_iovec != 0) {
#ifdef DEBUG
				ZDPRINTF("bad READ/WRITE/REC_DATA\n");
#endif /* DEBUG */
				return(EINVAL);
			}
			if (!useracc((caddr_t)cbp->cb_addr,
				     (uint)cbp->cb_count, B_WRITE))
				return(EFAULT);
			bufalloc(bp);
			if (cbp->cb_cmd == ZDC_WRITE 
			    || cbp->cb_cmd == ZDC_WHDR_WDATA)
				bp->b_flags = B_WRITE;
			else
				bp->b_flags = B_READ;
			bp->b_bcount = cbp->cb_count;
			bp->b_un.b_addr = (caddr_t)cbp->cb_addr;
			break;
		case ZDC_READ_HDRS:	/* read all sector headers */
			if (!suser())
				return(EACCES);
			else if (dd->zdd_sectors == 0) {
#ifdef DEBUG
				ZDPRINTF("READ_HDR - zdd_sectors is 0\n");
#endif /* DEBUG */
				return(EINVAL);
			}
			if (cbp->cb_count % CNTMULT
			    || (cbp->cb_addr & (ADDRALIGN-1)) != 0
			    || cbp->cb_iovec !=0)
				return (EINVAL);
			if (!useracc((char *)cbp->cb_addr,
						(uint)cbp->cb_count, B_WRITE))
				return (EFAULT);
			bufalloc(bp);
			bp->b_flags = B_READ;
			bp->b_bcount = cbp->cb_count;
			bp->b_un.b_addr = (caddr_t)cbp->cb_addr;
			break;
		case ZDC_FMTTRK:	/* format specified track */
			if (!suser() || !(up->zu_flags & ZUF_EXCLUSIVE)
				     || dd->zdd_sectors == 0) {
#ifdef DEBUG
				if (dd->zdd_sectors == 0)
					ZDPRINTF("FMTTRK - zdd_sectors is 0\n");
#endif /* DEBUG */
				return(EACCES);
			}
			if (cbp->cb_count % CNTMULT
			    || (cbp->cb_addr & (ADDRALIGN-1)) != 0
			    || cbp->cb_iovec !=0)
				return (EINVAL);
			if (!useracc((char *)cbp->cb_addr,
						(uint)cbp->cb_count, B_WRITE))
				return (EFAULT);
			bufalloc(bp);
			bp->b_flags = B_WRITE;
			bp->b_bcount = cbp->cb_count;
			bp->b_un.b_addr = (caddr_t)cbp->cb_addr;
			break;
			
		case ZDC_WRITE_LRAM:	/* write zdc local ram */
			if (!suser())
				return(EACCES);
			/* Fall into */
		case ZDC_READ_LRAM:	/* read zdc local ram */
			/*
			 * Check if LRAM address reasonable.
			 */
			if (cbp->cb_addr >= (ZDC_LRAMSZ / sizeof(int)))
				return (EINVAL);
			bufalloc(bp);
			bp->b_flags = (cbp->cb_cmd == ZDC_READ_LRAM) ? B_READ
								     : B_WRITE;
			bp->b_bcount = 0;		/* data in CB */
			bp->b_un.b_addr = NULL;
			break;

		default:
			return (EINVAL);
		}

		/*
		 * Do ioctl.
		 */
		bcopy((caddr_t)cbp, (caddr_t)&up->zu_ioctlcb, FWCBSIZE);
		bp->b_flags |= B_IOCTL;
		bp->b_devp = dp;
		bp->b_blkno = 0;
		bp->b_error = 0;
		bp->b_proc = u.u_procp;
		bp->b_iotype = B_RAWIO;
		BIODONE(bp) = 0;
		++u.u_procp->p_noswap;
		if (bp->b_bcount > 0) {
			/*
			 * lock down pages
			 */
			vslock(bp->b_un.b_addr, (int)bp->b_bcount,
				(bool_t)(bp->b_flags & B_READ));
		}
		s_ipl = p_lock(&up->zu_lock, SPLZD);
		if (up->zu_state == ZU_BAD) {
			v_lock(&up->zu_lock, s_ipl);
			/*
			 * Controller/channel/Drive has gone bad!
			 */
			buffree(bp);
			--u.u_procp->p_noswap;
			if (setokay < 0)
				dd->zdd_sectors = 0;
			return (EIO);
		}
		/*
		 * Insert at head of list and zdstart!
		 */
		bp->av_forw = up->zu_bhead.av_forw;
		up->zu_bhead.av_forw = bp;
 		DS_QUEUED_TIME(bp);
		zdstart(up);
		DS_Q(up->zu_dstats);
		v_lock(&up->zu_lock, s_ipl);
		biowait(bp);
		--u.u_procp->p_noswap;			/* re-swappable */
		error = geterror(bp);
		bcopy((caddr_t)&up->zu_ioctlcb, (caddr_t)cbp, FWCBSIZE);
		buffree(bp);
		if (copyout((caddr_t)cbp, (caddr_t)uaddr, sizeof( struct cb))) {
			if (setokay < 0)
				dd->zdd_sectors = 0;
			return(EFAULT);
		}
		if (setokay < 0)
			dd->zdd_sectors = 0;
		if (cbp->cb_cmd == ZDC_SET_CHANCFG) {
			/*
			 * Copy in the Channel Configuration to
			 * make sure it's set before anything else
			 * is done, and set drive size, etc.
			 */
			if (copyin((caddr_t)cbp->cb_addr, (caddr_t)dd, 
			    	sizeof(struct zdcdd)))
				return(EFAULT);
			up->zu_drive_size = dd->zdd_sectors * dd->zdd_tracks
					    * dd->zdd_cyls;
			up->zu_firstsect = dd->zdd_sectors * dd->zdd_tracks;
		}	
		return (error);

	case ZIORESERVE:
		/*
		 * Reserve this channel for exclusive use.
		 * Must be super user, have exclusive access
		 * to the current device, and it must be the
		 * only open device on the channel. It will
		 * inihibit further opens on the channel.
		 */
		if (!suser() || !(up->zu_flags & ZUF_EXCLUSIVE))
			return(EACCES);
		chan = up->zu_drive & 1;
		ctlrp = &zdctrlr[up->zu_ctrlr];
		s_ipl = p_lock(&ctlrp->zdc_ctlrlock, SPLHI);
		for (i=0, tup = &zdunit[0]; i <= zd_max_unit; i++, tup++) {
			if (tup != up && tup->zu_ctrlr == up->zu_ctrlr
			&&  (tup->zu_drive & 1) == chan
			&&  tup->zu_nopen) {
				v_lock(&ctlrp->zdc_ctlrlock, s_ipl);
				return(EACCES);
			}
		}
		if (ctlrp->zdc_chan_owner[chan]) {
			v_lock(&ctlrp->zdc_ctlrlock, s_ipl);
			return(EACCES);
		}
		ctlrp->zdc_chan_owner[chan] = up;
		v_lock(&ctlrp->zdc_ctlrlock, s_ipl);
		return (0);
	case ZIORELEASE:
		/*
		 * Converse of ZIORESERVE.
		 */
		if (!suser() || !(up->zu_flags & ZUF_EXCLUSIVE))
			return(EACCES);
		chan = up->zu_drive & 1;
		ctlrp = &zdctrlr[up->zu_ctrlr];
		if (ctlrp->zdc_chan_owner[chan] != up) {
			return(EACCES);
		}
		ctlrp->zdc_chan_owner[chan] = (struct zd_unit *)NULL;
		return (0);

	default:
		return (EINVAL);
	}
	/*NOTREACHED*/
	return(0);
}

/*
 * zdsize()
 *	Used for swap-space partition calculation.
 */
zdsize(dp)
	register struct dev *dp;
{
	register dev_t dev = dp->d_dev;
	register struct zd_unit *up;

	up = &zdunit[ZDUNIT(dev)];

	if (ZDUNIT(dev) > zd_max_unit
	||  up->zu_state != ZU_GOOD
	||  ((up->zu_cfg & (ZD_FORMATTED|ZD_MATCH)) != (ZD_FORMATTED|ZD_MATCH))
	||  up->zu_drive_size == 0)
			return (-1);
	return (up->zu_drive_size);
}

/*
 * Print out status bytes when appropriate.
 */
zddumpstatus(cbp)
	register struct cb *cbp;
{
	register int i;

	if (cbp->cb_cmd == ZDC_INIT || cbp->cb_cmd == ZDC_PROBE ||
	    cbp->cb_cmd == ZDC_PROBEDRIVE)
		return;

	switch(cbp->cb_compcode) {

	case ZDC_DRVPROT:
	case ZDC_DRVFLT:
	case ZDC_SEEKERR:
	case ZDC_HDR_ECC:
	case ZDC_SOFTECC:
	case ZDC_CORRECC:
	case ZDC_ECC:
	case ZDC_SNF:
	case ZDC_REVECT:
	case ZDC_SO:
	case ZDC_NDS:
	case ZDC_FDL:
	case ZDC_DDC_STAT:
		break;

	default:
		if (cbp->cb_cmd == ZDC_READ_LRAM || cbp->cb_cmd == ZDC_WRITE_LRAM)
			break;
		return;
	}
	if (cbp->cb_bp != NULL)
		cmn_err(CE_CONT, "zd%ds%d: cb_status:", cbp->cb_unit,
			PARTCHR(cbp->cb_bp->b_dev));
	else
		cmn_err(CE_CONT, "zd%d: cbstatus:", cbp->cb_unit);
	for (i=0; i < NSTATBYTES; i++)
		cmn_err(CE_CONT, " 0x%x", cbp->cb_status[i]);
	cmn_err(CE_CONT, "\n");
}

#ifdef	DEBUG
/*
 * Print out the contents of a cb.
 */
zddumpcb(cbp)
	register struct cb *cbp;
{
	register struct init_cb *icbp;
	register int i;
	register int count;
	int	cbcount;

	CPRINTF("zd%d: cb at 0x%x, cb_cmd 0x%x, cb_compcode 0x%x\n",
		cbp->cb_unit, cbp, cbp->cb_cmd, cbp->cb_compcode);
	if (cbp->cb_cmd == ZDC_INIT) {
		icbp = (struct init_cb *)cbp;
		CPRINTF("icb_ctrl 0x%x, icb_pagesize 0x%x, icb_dumpaddr 0x%x\n",
			icbp->icb_ctrl, icbp->icb_pagesize, icbp->icb_dumpaddr);
		CPRINTF("icb_dest 0x%x, icb_bin 0x%x, icb_vecbase 0x%x\n",
			icbp->icb_dest, icbp->icb_bin, icbp->icb_vecbase);
		CPRINTF("icb_errdest 0x%x, icb_errbin 0x%x, icb_errvector 0x%x\n\n",
			icbp->icb_errdest, icbp->icb_errbin,
			icbp->icb_errvector);
		return;
	}
	if (cbp->cb_cmd == ZDC_PROBE || cbp->cb_cmd == ZDC_PROBEDRIVE) {
		CPRINTF("pcb_drivecfg:");
		for (i = 0; i < ZDC_MAXDRIVES; i++)
			CPRINTF(" 0x%x",
				 ((struct probe_cb *)cbp)->pcb_drivecfg[i]);
		CPRINTF("\n\n");
		return;
	}
	CPRINTF("cb_mod 0x%x, cb_diskaddr (%d, %d, %d), cb_psect 0x%x\n",
		cbp->cb_mod, cbp->cb_cyl, cbp->cb_head, cbp->cb_sect,
		cbp->cb_psect);
	CPRINTF("cb_addr 0x%x, cb_count 0x%x, cb_iovec 0x%x, cb_reqstat 0x%x\n",
		cbp->cb_addr, cbp->cb_count, cbp->cb_iovec, cbp->cb_reqstat);
	zddumpstatus(cbp);
	CPRINTF("cb_bp 0x%x, cb_errcnt 0x%x, cb_iovstart 0x%x, cb_state 0x%x\n",
		cbp->cb_bp, cbp->cb_errcnt, cbp->cb_iovstart, cbp->cb_state);
	CPRINTF("cb_contaddr 0x%x, cb_contiovsz 0x%x, cb_transfrd 0x%x\n",
		cbp->cb_contaddr, cbp->cb_contiovsz, cbp->cb_transfrd);
	if (cbp->cb_iovec != NULL) {
		count = 0;
		cbcount = cbp->cb_count;
		if (cbp->cb_addr & CLOFSET) {
			++count;
			cbcount -= CLBYTES - (cbp->cb_addr & CLOFSET);
		}
		count += (cbcount + CLOFSET) >> CLSHIFT;
		CPRINTF("iovecs:");
		for (i = 0; i < count; i++) {
			CPRINTF(" 0x%x", cbp->cb_iovstart[i]);
			if ((i % 8) == 7)
				CPRINTF("\n\t");
		}
		CPRINTF("\n");
	}
	CPRINTF("\n");
}
#endif	/* DEBUG  */


zdprint (dp,str)
	struct dev *dp;
	char *str;
{
	dev_t	dev = dp->d_dev;

	cmn_err(CE_CONT,
		"%s on zdc disk %d, partition %d\n",
		str,ZDUNIT(dev), ZDPART(dev)); 
}

/*
 * zdfirstsect
 *	Return the first sector which is usable (ie, not reserved) on
 *	this particular device.  This hook is needed because some disks
 *	(like zd disks) require reserved space at the front of the
 *	disk, yet the rest of the kernel needs to know how to get to
 *	the beginning of the disk.  This requires that clients of
 *	physical disk drivers interrogate the serving driver for the
 *	"First Sector of usable space" on the device.
 */

int
zdfirstsect(dp)
	struct	dev	*dp;
{
	register dev_t dev = dp->d_dev;
	register struct zd_unit *up;

	up = &zdunit[ZDUNIT(dev)];

	if (ZDUNIT(dev) > zd_max_unit
	||  (up->zu_state != ZU_GOOD)
	||  ((up->zu_cfg & (ZD_FORMATTED|ZD_MATCH)) != (ZD_FORMATTED|ZD_MATCH))
	||  up->zu_firstsect == 0)
			return (-1);
	return (up->zu_firstsect);
}

