/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ident	"$Header: startup.c 1.63 90/10/01 $"

/*
 * startup.c
 *	Various startup/initializations.  i386 version.
 *
 * Factored out of machdep.c to keep smaller modules and more related
 * functions together.
 */

/* $Log:	startup.c,v $
 */

#include "sys/signal.h"
#include "sys/types.h"
#include "sys/devsw.h"
#include "sys/param.h"
#include "sys/debug.h"
#include "sys/mutex.h"
#include "sys/systm.h"
#include "sys/user.h"
#include "sys/map.h"
#include "sys/vm.h"
#include "sys/timer.h"
#include "sys/session.h"
#include "sys/proc.h"
#include "sys/buf.h"
#include "sys/conf.h"
#include "sys/vfs.h"
#include "sys/vnode.h"
#include "sys/ufsinode.h"
#include "sys/ufsmount.h"
#ifdef QUOTA
#include "sys/ufsquota.h"
#endif /* QUOTA */
#include "sys/file.h"
#include "sys/uio.h"
#include "sys/dio.h"
#include "sys/callo.h"
#include "sys/cmap.h"
#include "sys/var.h"
#include "sys/cmn_err.h"
#include "sys/dnlc.h"
#include "sys/arch.h"

#include "sys/engine.h"
#include "sys/slic.h"
#include "sys/slicreg.h"
#include "sys/clkarb.h"
#include "sys/cfg.h"

#include "sys/pte.h"
#include "sys/trap.h"
#include "sys/hwparam.h"
#include "sys/plocal.h"
#include "sys/mmu.h"
#include "sys/psl.h"
#include "sys/intctl.h"
#include "sys/mftpr.h"
#include "sys/gate.h"

#include "sys/autoconf.h"

#ifdef	MODELD
#include "sys/scan.h"
#endif	/* MODELD */

/*
 * Define these here, as we fill them out.
 */

int		Usrptsize = 0;		/* # pte's for user page-tables */

struct	pte	*Sysmap = NULL;		/* start of KL2PT */
struct	pte	*Usrptmap = NULL;	/* KL2PT subset for User PT's */
struct	pte	*Bufmap = NULL;		/* KL2PT subset for cache buffers */
struct	pte	*Endmap;		/* End of contiguous level-2 PT */

struct	pte	*usrpt = NULL;		/* virt-addr mapped by Usrptmap[0] */

struct	pte	*kl1pt_master = NULL;	/* pointer to "master" kl1pt */

int		maxmem = 0;		/* initial free-space */
caddr_t		maxkmem = 0;		/* max kmem virtual address */
bool_t		calloc_ok = 1;		/* flag for legal calloc's */
unchar		*egafont_p[5] = { 0 };
ulong		basemem, expnmem;

static	caddr_t curmem;			/* memory pointer for calloc() */

#define	csalloc(n,type)	(type *) calloc((int)(n)*sizeof(type))
#define	valloc(name, type, num) \
		(name) = csalloc(num, type)
#define	valloclim(name, type, num, lim) \
		(name) = csalloc(num, type); (lim) = ((name)+(num))

/*
 * Heuristic limits.
 */

ulong	max_ker_vaddr =	VA_PLOCAL;		/* max kernel virt-address */

/*
 * Tunable flags -- for performance experimentation.
 */

/*
 * Variable(s) to hold kernel virtual address of the SLIC; this changes when we
 * move from powerup environment (PHYS=VIRT) to kernel page tables.  We declare
 * va_slic_lmask to optimize hot code that would otherwise have to compute a
 * constant offset at runtime.
 */

struct	cpuslic	*va_slic = (struct cpuslic *)LOAD_CPUSLICADDR;
unchar	*va_slic_lmask = &((struct cpuslic *)LOAD_CPUSLICADDR)->sl_lmask;
int	slic_delay = 0;
unchar	*plocal_slic_delay = (unchar *)&slic_delay;

/*
 * Architecture dependent variables
 */

unchar	arch_type;
struct	arch_functions arch_functions;

/*
 * Various externals we need.
 */

extern	int	bufpages;
extern	int	bufpct;

extern	int	resphysmem;

extern	caddr_t	topmem;			/* top of memory */
extern	uint	totalmem;		/* total memory (topmem-holes) */

extern	struct	ctlr_desc *slic_to_config[];

#ifdef	MODELD
int	l1pt_cacheable = 0;		/* bitmask (patchable) used to turn off caching of L1 PT's */
int	l2pt_cacheable = 0;		/* same for L2 */
int	starttype;			/* starts out as zero */
#endif	/* MODELD */

/*
 * sysinit()
 *	Do basic system initializations.
 *
 * Includes:
 *	Allocate per-processor local data.
 *	Do heuristics for DS sizing.
 *	Allocate KL2PT (Sysmap et.al.).
 *	Allocate engine structures (per processor).
 *	Fill out (most of) KL2PT, KL1PT.
 *	Init interrupt and trap vectors.
 *
 * Called by 1st processor to start, *very* early after system is alive.
 * Running with segmentation enabled, 1-1 map with physical memory.
 * Paging not yet enabled.
 */

sysinit(startmem, puarea)
	caddr_t	startmem;		/* pointer to first byte of free mem */
	caddr_t	puarea;			/* phys address of U-block base */
{
	register caddr_t addr;
	register struct pte *pte;
	register int i;
	unsigned procid;		/* logical processor */
	struct	ppriv_pages *ppriv;
	extern	lock_t	time_lck;
	extern	lock_t	panic_lock;
	struct bus *bus;
#ifdef	MODELD
	extern int etext;
	unsigned cr0;
#endif	/* MODELD */
 

	/*
	 * Set up architechture dependent changes
	 */

	set_arch_type();

#ifdef MODELD
	if (arch_type != SYSTYP_S1)
		starttype = rdslave(va_slic->sl_procid, SL_G_BOARDTYPE);

	if (starttype == SLB_SGS2PROCBOARD) {
		/*
		 * Enable the onchip cache.  Note that we're explicitly
		 * clearing the "Alignment Mask" bit to prevent alignment
		 * checking in privledge level 3.
		 */
		INVAL_ONCHIP_CACHE();
		cr0 = READ_MSW();
		WRITE_MSW(cr0 & ~(CR0_CD|CR0_NW|CR0_AM) | CR0_WP);
	}
#endif	/* MODELD */

	/*
 	 * Zap physical U-area; memory not zeroed as we come up,
	 * and need a clean U-area.  Other (per-engine) are zeroed
	 * when allocated, below.
	 */

	bzero((char *)UBTOUSER(puarea), sizeof(struct user));

	/*
	 * Initialize temporary HW interrupt table.
	 * Need an IDT and GDT to take NMI's during probing.
	 * Can't build on "self" local data, since this isn't
	 * yet allocated, and won't be until after configure(),
	 * which determines number of processors.
	 */

	if (arch_type != SYSTYP_S1) {
#ifdef	MODELD
		init_temp_IDT(starttype);
#else	/* !MODELD */
		init_temp_IDT();
#endif	/* MODELD */
		init_temp_GDT();
	}

	/*
	 * Initialize monitor IO (see consio.c)
	 */

	initmonio();
	curmem = startmem;			/* init's calloc() */

#ifdef	MODELD
	if (starttype == SLB_SGS2PROCBOARD) {
		/*
		 * Init the scan interface and enable NMI.
		 * We need to do this only when booting on an SGS2
		 * processors.  Otherwise, the scan interface will
		 * be initialized as part of autoconfig before we
		 * online the first SGS2 processor.
		 */
		scan_init((int)(CD_LOC->c_cons->cd_slic));
		enable_snmi(va_slic->sl_procid);
	}
#endif	/* MODELD */

	/*
	 * Initialize the devsw data structure.
	 */
	
	devsw_init();

	/*
	 * Configure the HW and initialize interrupt table.
	 * Configure() is called with the loader/boot page-tables in use.
	 *
	 * Configure allocates and fills out:
	 *	engine[] array		; one per processor
	 *	Nengine			; # processors
	 *	topmem			; top of physical memory
	 * plus sets up device drivers/etc.
	 */

	init_lock(&panic_lock);		/* initialize panic lock */
	(void) splhi();			/* insure interrupts OFF */
	CONFIGURE(&arch_functions);

	/*
	 * Initial TLB flush mechanism
	 */

	InitFlushTLB();

	/*
	 * Allow custom system-calls to allocate data-structures.
	 */

	cust_sys_boot();

	/*
	 * Allocate fixed-size system data-structures.
	 */

	alloc_tables();

	/*
	 * Allocate processor-local IO mapping pages.
	 */
	
	alloclocalIO();

	/*
	 * Allocate per-processor local data.
	 */

	callocrnd(KL1PT_BYTES);				/* for L1PT's */
	ppriv = (struct ppriv_pages *) calloc((int)(Nengine * SZPPRIV));

	/*
	 * Allocate heuristicly-sized system data-structures, including
	 * system level-2 page-table.
	 */

	heuristics(totalmem - (uint)curmem);


	/*
	 * Fill out engine structures; these were allocated by
	 * configure() who also filled out the e_slicaddr fields,
	 * and turned on E_FPU387 and/or E_FPA in e_flags if appropriate.
	 *
	 * Figure out my procid from engine structures.
	 *
	 * If there were any mono-P drivers for existing HW, then
	 * they were bound to 'me' (ie, booting processor); thus
	 * set flag to avoid ever taking `me' offline.
	 */

	procid = Nengine;
	for (i = 0; i < Nengine; i++) {
		engine[i].e_local = &ppriv[i];	/* remember private parts */
		engine[i].e_flags |= E_OFFLINE;	/* not up yet */
		if (arch_type == SYSTYP_S1)
			procid = 0;
		else if (va_slic->sl_procid == engine[i].e_slicaddr)
			procid = i;
	}
	ASSERT(procid < Nengine, "sysinit: unknown processor");
	/*
	 *+ During initialization of the system, the system configuration
	 *+ data structures implied that the booting processor was not a
	 *+ member of this system.
	 */

	init_lock(&engtbl_lck);

	/*
	 * "My" engine structure identifies "master" KL1PT.
	 * Init this for KL2PT already known (it got zapped above).
	 */

	kl1pt_master = (struct pte *) &engine[procid].e_local->pp_kl1pt[0][0];

	ASSERT(((int)kl1pt_master & (KL1PT_BYTES-1)) == 0, "kl1pt mis-aligned");
	/*
	 *+ The master level-1 page table for the system was not aligned on
	 *+ the appropriate boundary.
	 */
	ASSERT(((int)Sysmap & (NBPG-1)) == 0, "Sysmap mis-aligned");
	/*
	 *+ The kernel data structure Sysmap was not aligned on a page boundary.
	 */

	for (i = 0, pte = Sysmap; pte < Endmap; pte += NPTEPG, i++)
		*(int*)(&kl1pt_master[i+L1IDX(KSOFF)]) = 
#ifdef	MODELD
					PHYSTOPTE(pte)|PG_V|PG_KW|PG_R|PG_M |
					l2pt_cacheable;
#else	/* !MODELD */
					PHYSTOPTE(pte)|PG_V|PG_KW|PG_R|PG_M;
#endif	/* MODELD */

	/*
	 * Set up Sysmap to map existing physical memory.
#ifdef	MODELD
	 * Map the text segment as being read only.  Need to make sure the
	 * CR0 register is initialized with the WP bit in order for this
	 * to be effective.  This will prevent the kernel from overwriting
	 * the text segment.  A read of a later edition of the i486 manual
	 * cast doubts on whether this will be allowed.  XXX.jds
#else	!MODELD
	 * 80386 doesn't allow write-protect kernel text thru page-tables only.
	 * Oh, well.  Can make low 8K or so invalid, which helps some.
#endif	MODELD
	 *
	 * Note that physical address space holes are mapped with zeroes in
	 * the page-table.  Could reclaim this memory as free space and zero
	 * the level-1 pte, but (eg) /dev/kmem and meminit() get more
	 * difficult.  Since this isn't all that big, ignore for now (relevant
	 * cmap array space is much larger).
	 *
	 * On the PC/3AT also map the area between basemem and expnmem
	 * as this can contain memory mapped boards and the display memory.
	 *
	 * This assumes kernel is linked above CD_LOC.
	 */
#ifdef	MODELD
	for (i = btop(CD_LOC), addr = (caddr_t)CD_LOC; addr < (caddr_t)&etext;
							addr += NBPG, i++) {
		if (page_exists(i)) {
			*(int *)(&Sysmap[i]) = PHYSTOPTE(addr)
					     | PG_V|PG_KR|PG_R|PG_M;
		}
	}
	for (; addr < topmem+resphysmem; addr += NBPG, i++) {
		if (page_exists(i)
		|| (arch_type == SYSTYP_S1
		 && addr < (caddr_t)MC_EXPBASE_AT386)) {
			*(int *)(&Sysmap[i]) = PHYSTOPTE(addr)
					     | PG_V|PG_KW|PG_R|PG_M;
		}
	}
#else	/* !MODELD */
	for (i = btop(CD_LOC), addr = (caddr_t)CD_LOC; addr < topmem+resphysmem;
							addr += NBPG, i++) {
		if (page_exists(i)
		|| (arch_type == SYSTYP_S1
		 && addr < (caddr_t)MC_EXPBASE_AT386)) {
			*(int *)(&Sysmap[i]) = PHYSTOPTE(addr)
					     | PG_V|PG_KW|PG_R|PG_M;
		}
	}
#endif	/* MODELD */

	/*
	 * Initialize IO controller mapping.
	 */

	for ( bus = busses; bus->bustype; bus++ )
		if ( bus->alloc_maps )
			(*bus->alloc_maps)(bus);

	/*
	 * Initialize time_lck. This must be done before the 1st
	 * tod clock interrupt as this triggers a softclock interrupt
	 * which uses this lock.
	 */

	init_lock(&time_lck);

	/*
	 * Init "eng_wait" semaphore, used to coordinate bringing
	 * processors on-line and taking processors off-line.
	 * Initialize here since selfinit does a V. Need to reinitialize
	 * after selfinit().
	 */

	init_sema(&eng_wait, 0);

	/*
	 * System is mapped enough to do self-init's.
	 * Need to do this now, since allockds() needs to run mapped.
	 */

	selfinit(procid);

	/*
	 * Enable Ecc error reporting in memory controllers.
	 *
	 * Note: errors will merely be latched.  No interrupts
	 * are generated other than NMI for uncorrectable errors.
	 */

	MEMENABLE(&arch_functions);

	/*
	 * Reinitialize since selfinit does a V sema.
	 */

	init_sema(&eng_wait, 0);

	/*
	 * Good {morning,afternoon,evening,night}.
	 */

	CPRINTF(version);

#define MEG	(1024*1024)
	CPRINTF("real mem  = %d.%d%d megabytes.\n", totalmem/MEG,
		((totalmem%MEG)*10)/MEG,
		((totalmem%(MEG/10))*100)/MEG);

	/*
	 * Initialize callouts
	 */

	callfree = callout;
	for (i = 1; i < v.v_call; i++)
		callout[i-1].c_next = &callout[i];

	/*
	 * Initialize memory allocator and user page table maps.
	 */

	maxmem = meminit(curmem);
	i = ctob(maxmem);
	CPRINTF("avail mem = %d.%d%d megabytes.\n", i/MEG,
		((i%MEG)*10)/MEG,
		((i%(MEG/10))*100)/MEG);

	i = bufpages * CLBYTES;
	CPRINTF("using %d buffers containing %d.%d%d megabytes of memory.\n",
		v.v_buf,
		i/MEG,
		((i%MEG)*10)/MEG,
		((i%(MEG/10))*100)/MEG);


	uptinit((long)Usrptsize, UPTMAPMULT*v.v_proc);

	/*
	 * Hand-craft page-table for process[0] -- swapper.
	 * Need to set this up before call main() from start.s, since
	 * currently running on boot-processor's private resources.
	 *
	 * Use of vgetpt() inherits current l. mapping.
	 *
	 * NOTE: currently running on proc[0]'s Uarea stack, but at
	 * physical address.  "u." references are to processor private Uarea.
	 * All proc[] fields were zeroed when struct proc's were allocated.
	 */

	proc[0].p_szpt = SZPT(&proc[0]);		/* includes Uarea map */
	proc[0].p_ptb1 = (unsigned) l.priv_pt;		/* for vmaccess() */
	u.u_procp = &proc[0];				/* ditto */
	(void) vgetpt(&proc[0], vmemall);		/* alloc page-table */

	*(int *) (UAREAPTES(&proc[0])+0) = PHYSTOPTE(puarea+0*NBPG) |
					PG_V|PG_KW|PG_R|PG_M;
#if	UPAGES == 2
	*(int *) (UAREAPTES(&proc[0])+1) = PHYSTOPTE(puarea+1*NBPG) |
					PG_V|PG_KW|PG_R|PG_M;
#endif
#if	UPAGES > 2
	ERROR -- assumes UPAGES <= 2
#endif

	WRITE_PTROOT(proc[0].p_ptb1);

#ifdef	DEBUG
	/*
	 * Set up process 0's stack with known pattern -- allows sense of
	 * stack overflow more easily.  Needs vgetu() to copy entire Uarea
	 * if this is to propogate.
	 */
	for (i = VA_UBLOCK; i <= VA_UAREA - 512; i += sizeof(int))
		*(int *)i = ((i >> sizeof(int)) & 1) ? 0xdeadbeef : 0xfeedface;
#endif	/* DEBUG  */
}

/*
 * set_arch_type()
 *
 * Set global "arch_type" from CFG data.
 */

set_arch_type()
{
	int i;
	extern sb8000_configure(), sb8000_memenable();
	extern at386_configure(), at386_memenable();

	arch_type = CD_LOC->c_sys.sd_type;

	switch (arch_type) {

	default:
		cmn_err(CE_PANIC,
			"set_arch_type: unsupported type %d\n", arch_type);
		/*
		 *+ The config data tables contain a value for
		 *+ the system type that is unsupported.
		 */
		break;

	case SYSTYP_S27:
	case SYSTYP_S81:
	case SYSTYP_S16:
		/*
		 * SYMMETRY hardware platforms
		 */
		arch_functions.arch_configure = sb8000_configure;
		arch_functions.arch_memenable = sb8000_memenable;
		break;

	case SYSTYP_S1:
		/*
		 * PC/3AT hardware platforms
		 */
		arch_functions.arch_configure = at386_configure;
		arch_functions.arch_memenable = at386_memenable;
		basemem = get_basemem();
		expnmem = get_expnmem();
		for (i = 0; i < 5; i++)
			egafont_p[i] =
				(unchar *)phystokv(ftop(BIOS_EGAFONTPTR[i]));
		break;
	}
}

/*
 * alloc_tables()
 *	Allocate (most) fixed-size kernel data-structures.
 *
 * This used to be called startup() in 4.2bsd, and was called by main().
 */

static
alloc_tables()
{
	int nctmp;

	valloclim(inode, struct inode, v.v_inode, v.ve_inode);
	valloclim(file, struct file, v.v_file, v.ve_file);
	valloclim(proc, struct proc, v.v_proc, procNPROC);
	valloc(sess, struct session, v.v_proc);
	valloclim(mounttab, struct mount, v.v_mount, v.ve_mount);
	valloc(callout, struct callout, v.v_call);
	valloc(mfile, struct mfile, nmfile);
	valloc(ucred_free, struct ucred, nucred);
	valloc(swapmap, struct map, nswapmap = v.v_proc * 2);
	valloc(uptmap, struct map, UPTMAPMULT*v.v_proc);
	valloc(ncache, struct ncache, ncsize);
	nctmp = 1;
	while(nctmp < nchsize)
		nctmp <<= 1;
	nchsize = nctmp;
	valloc(nc_hash, struct nc_hash, nchsize);
#ifdef QUOTA
	valloclim(dquot, struct dquot, ndquot, dquotNDQUOT);
	valloc(dqhead, struct dqhead, ndqhash);
#endif
	valloc(abuf, struct AIO_buf, nabuf);
	shminit();
	seminit();
	msginit();
	ofile_boot();
	strinit();
	spec_init();
}

/*
 * heuristics()
 *	Allocate heuristicly sized system data structures.
 */

static
heuristics(freesize)
	uint	freesize;
{
	register char	*paddr;
	register int	mapidx;
	register int	j;
	register unsigned i;
	register int	base;
	register int	residual;
	extern	int	avg_size_process;		/* in bytes */
	extern	int	buf_hash_mult, buf_hash_div;

	/*
	 * Determine how many buffers to allocate.
	 * Use bufpct % of memory, with min of 16.
	 * We allocate 1/2 as many swap buffer headers as file i/o buffers.
	 *
	 * Since max physical memory is about 256Meg, 10% of this is
	 * 25Meg.  With 4K page-size, and one buf-page per virtual
	 * buffer, this is 50Meg address space.  Since user stuff
	 * starts at 1Gig, there is plenty of room -- don't bother
	 * limiting.
	 *
	 * On K20, must be sure to not map over processor-local IO stuff,
	 * which lives at virtual 28Meg-64k.  Since max memory is 28Meg,
	 * the same reasoning ==> no real problem.
	 */

	if (bufpages == 0) {
		/*
		 *  We are careful to avoid overflow for large memory
		 *  systems....
		 */
		bufpages = (freesize / (100*CLBYTES)) * bufpct
			 + (freesize % (100*CLBYTES)) * bufpct / (100*CLBYTES);
	}
	if (v.v_buf == 0) {
		v.v_buf = MIN(bufpages,
			   (max_ker_vaddr-(ulong)curmem-512*1024)/MAXBSIZE);
		if (v.v_buf < 16) {
			v.v_buf = 16;
		}
	}
	if (v.v_pbuf == 0) {
		v.v_pbuf = v.v_buf / 2;
		if (v.v_pbuf > 256) {
			v.v_pbuf = 256;		/* sanity */
		}
	}
	if (v.v_hbuf == 0)
		v.v_hbuf = imax(v.v_buf * buf_hash_mult / buf_hash_div, MIN_BUFHSZ);

	fifo_init(bufpages);
	valloclim(bufhash, struct bufhd, v.v_hbuf, bufhashBUFHSZ);
	valloc(buf, struct buf, v.v_buf);
	valloc(swbuf, struct buf, v.v_pbuf);

	/*
	 * Figure size of Usrptmap[] (user page-tables part of KL2PT).
	 * Base this on v.v_proc, and layout of a process (U-area, page-table,
	 * sizing page-table based on "average" size process 
	 * (see space/param_space.c).
	 * Each pte in Usrptmap[] maps *alot* of space, and there is *alot*
	 * of kernel virtual space; thus no big deal if overallocate a bit.
	 *
	 * Tuning of "v.v_proc" might be more appropriate; suspect the
	 * heuristic in param_space.c is too conservative.
	 *
	 * The idea is to not prematurely swap due to lack of Usrptmap
	 * space, but not overallocate Usrptmap either.
	 */

	if (Usrptsize == 0) {
		Usrptsize = v.v_proc *
				( UPAGES		/* U-area */
				+ UL1PT_PAGES		/* UL1PT */
				+ SZL2PT(btoc(avg_size_process),1)
				+ CLSIZE		/* Slop */
				);
	}
	Usrptsize = clrnd(Usrptsize);

	/*
	 * Allocate KL2PT.  Form:
	 *
	 *	Sysmap
	 *	Usrptmap
	 *	Bufmap
	 *
	 * Sysmap[] is sized to map all of physical memory.
	 *
	 * Create entire L2 page-table in one allocation since code assumes
	 * contiguous KL2PT.  Could make filling out of kl1pt smarter but
	 * holes skipped here should be rare.
	 *
	 */

	callocrnd(NBPG);

	i = clrnd((int) btop(topmem + resphysmem))	/* Sysmap */
	  + Usrptsize					/* Usrptmap */
	  + v.v_buf * btop(MAXBSIZE);			/* Bufmap */

	if (i > (j = topmem - curmem)) {
		cmn_err(CE_PANIC, "heuristics(): size of user level 2 page table exceeds available memory \n(Sysmap:0x%x + Usrptmap:0x%x + Bufmap:0x%x = 0x%x) > 0x%x",
				clrnd((int) btop(topmem + resphysmem)),
				Usrptsize, v.v_buf * btop(MAXBSIZE), i, j);
		/*
		 *+ Insufficent memory is available to allocate the user level
		 *+ 2 page tables. This indicates a resource configuration 
		 *+ problem. 
		 *+ Corrective action:  decrease the number of user process slots
		 *+ or add more memory.
		 */
	}
	Sysmap = csalloc(i, struct pte);
	Usrptmap = Sysmap + clrnd((int) btop(topmem + resphysmem));
	Bufmap = Usrptmap + Usrptsize;
	Endmap = &Sysmap[i];

	usrpt = (struct pte *) ptob(Usrptmap - Sysmap);
	buffers = ptob(Bufmap - Sysmap);

	maxkmem = ptob(Endmap - Sysmap);
	ASSERT((ulong)maxkmem <= max_ker_vaddr, "heuristics: maxkmem");
	/*
	 *+ When sizing memory and allocating kernel data structures,
	 *+ the kernel's notion of the highest kernel address overflowed
	 *+ into virtual space used to map processor-local information.
	 */

	/*
	 * Fill out buf-headers by distributing bufpages among the headers.
	 * binit() fills in b_addr.
	 */

	callocrnd(CLBYTES);

	base = bufpages / v.v_buf;
	residual = bufpages % v.v_buf;
	mapidx = 0;

	for (i = 0; i < residual; i++) {
		valloc(paddr, char, ptob((base + 1) * CLSIZE));
		for (j = 0; j < (base + 1) * CLSIZE; j++) {
			*(int *)(&Bufmap[mapidx+j]) = PHYSTOPTE(paddr)
						    | PG_V|PG_R|PG_M | PG_KW;
			paddr += NBPG;
		}
		mapidx += MAXBSIZE / NBPG;
	}
	for (i = residual; i < v.v_buf; i++) {
		valloc(paddr, char, ptob(base * CLSIZE));
		for (j = 0; j < base * CLSIZE; j++) {
			*(int *)(&Bufmap[mapidx+j]) = PHYSTOPTE(paddr)
						    | PG_V|PG_R|PG_M | PG_KW;
			paddr += NBPG;
		}
		mapidx += MAXBSIZE / NBPG;
	}

	/*
	 * Allocate space for core map.
	 * Allocate a cmap structure for each page of memory < topmem (including
	 * holes).  meminit() handles reclaiming useless cmap[] pages.
	 * Cluster align cmap[] to make reclaim of holes more likely.
	 */

	callocrnd(CLBYTES);
	valloclim(cmap, struct cmap, (int) topmem / CLBYTES, ecmap);

	ASSERT(curmem < topmem - 8*UPAGES, "heuristics: no memory");
	/*
	 *+ After allocating kernel data structures, the kernel found that
	 *+ no memory was available for user processes.
	 */

	/*
	 * There should be NO allocations after this!
	 */

	callocrnd(CLBYTES);
	calloc_ok = 0;				/* insure no more calloc()'s */
}

/*
 * selfinit()
 *	Do self-init's.  Done by each processor as it comes alive.
 *
 * If system just booting, sysinit() has already been called to
 * allocate kernel data-structures and init common parts kernel
 * level-2 page-tables.
 *
 * Includes:
 *	Init self KL1PT.
 *	Init per-processor descriptor tables (GDT, IDT).
 *	Init private parts of processor page-tables.
 *	Map U-area.
 *	Turn ON page-table mapping.
 *	(K20 only) Copy stuff to private RAM and start using this.
 *	Init interrupts.
 *	Init any other local processor resources.
 *
 * Caller arranged that we're already running on a U-area stack.
 */

selfinit(procid)
	unsigned procid;			/* logical processor # */
{
	register struct	ppriv_pages *ppriv;	/* my private parts */
	register struct	pte	*kl1pt;		/* my KL1PT */
	register int	i;
	struct	priv_map *pm;			/* my private mapping pages */
	struct	engine	*eng;			/* my engine structure */
	extern	bool_t	light_show;
	int	ver386;
#ifdef	MODELD
	unsigned cr0;
#endif	/* MODELD */

	eng = &engine[procid];
	ppriv = eng->e_local;

	/*
	 * Copy private KL1PT from "master" copy -- that of 1st
	 * processor booted.  This is redundant for 1st booted processor.
	 */

	kl1pt = &ppriv->pp_kl1pt[0][0];
	if (kl1pt != kl1pt_master)
		bcopy((caddr_t)kl1pt_master, (caddr_t)kl1pt, KL1PT_BYTES);

	ASSERT_DEBUG(((int)kl1pt & (KL1PT_BYTES-1)) == 0, "kl1pt mis-aligned");

	/*
	 * Make private PT self-referencing.
	 */

  	*(int*) &kl1pt[L1IDX(VA_PT+KSOFF)] =
#ifdef	MODELD
				PHYSTOPTE(kl1pt) | PG_V|PG_R|PG_M|PG_KW |
				l2pt_cacheable;
#else	/* !MODELD */
				PHYSTOPTE(kl1pt) | PG_V|PG_R|PG_M|PG_KW;
#endif	/* MODELD */

	/*
	 * Make private-mapping pages map private stuff and Uarea, and
	 * map these in the level-1 page-table.
	 */

	pm = (struct priv_map *) &ppriv->pp_pmap[0][0];
	
  	*(int*) &kl1pt[L1IDX(VA_PLOCAL+KSOFF)] =
#ifdef	MODELD
				PHYSTOPTE(pm) | PG_V|PG_R|PG_M|PG_KW |
				l2pt_cacheable;
#else	/* !MODELD */
				PHYSTOPTE(pm) | PG_V|PG_R|PG_M|PG_KW;
#endif	/* MODELD */
	for (i = 0; i < PL_PAGES; i++)
		*(int *) &(pm->pm_plocal[i]) =
			PHYSTOPTE(&ppriv->pp_local[i][0]) |PG_V|PG_R|PG_M|PG_KW;

  	*(int*) &kl1pt[L1IDX(VA_UBLOCK+KSOFF)] =
#ifdef	MODELD
			PHYSTOPTE(pm) | PG_V|PG_R|PG_M|PG_KW | l2pt_cacheable;
#else	/* !MODELD */
			PHYSTOPTE(pm) | PG_V|PG_R|PG_M|PG_KW;
#endif	/* MODELD */
	for (i = 0; i < UPAGES; i++)
		*(int *) &(pm->pm_uarea[i]) =
			PHYSTOPTE(&ppriv->pp_uarea[i][0]) |PG_V|PG_R|PG_M|PG_KW;

	/*
	 * Map processor local-IO stuff.
	 */

	if (arch_type != SYSTYP_S1)
		maplocalIO(kl1pt);

#if	KSOFF != 0

	/*
	 * Double map the first 256 MB (arbitrary) of kernel virtual space
	 * so we can enable paging before segmentation.
	 * NOTE: be careful not to zap any legit mapping (such as VA_USER).
	 */

	bcopy((caddr_t) &kl1pt[L1IDX(KSOFF)],
	      (caddr_t) &kl1pt[L1IDX(0)],
	      (((256*1024*1024)/NBPG)/NPTEPG) * sizeof (struct pte));
#endif

	/*
	 * Initialize descriptor tables, and turn ON mapping!
	 */

#ifdef	MODELD
	WRITE_PTROOT((uint)kl1pt | l1pt_cacheable);	/* set up page-table root */
	cr0 = READ_MSW();
	if (eng->e_flags & E_FPU387)		/* if 387... */
		WRITE_MSW(cr0 | CR0_PG|CR0_PE|CR0_EM|CR0_MP);/* ON paging! protection on! emulate floating point (trap7) */
	else
		WRITE_MSW(cr0 | CR0_PG|CR0_PE|CR0_EM);
#else	/* !MODELD */
	WRITE_PTROOT((uint)kl1pt);		/* set up page-table root */
	if (eng->e_flags & E_FPU387)		/* if 387... */
		WRITE_MSW(CR0_PG|CR0_PE|CR0_EM|CR0_MP);	/* ON paging! protection on! emulate floating point (trap7) */
	else
		WRITE_MSW(CR0_PG|CR0_PE|CR0_EM);	/* ON paging! protection on! emulate floating point (trap7) */
#endif	/* MODELD */
	bzero((caddr_t) &ppriv->pp_local[0][0], sizeof(struct plocal));
	init_desc_tables((struct plocal *) &ppriv->pp_local[0][0]);

	/*
	 * Mapping moves address of the SLIC so rewrite
	 * variables to reflect this change.
	 */

	va_slic = (struct cpuslic *) VA_SLIC;
	va_slic_lmask = &((struct cpuslic *) VA_SLIC)->sl_lmask;
	plocal_slic_delay = (unchar *)&l.slic_delay;

#if	KSOFF != 0

	/*
	 * Zap double mapping since both mapping and segmentation are now on.
	 * Be careful to not zap any legit mapping!
	 */

	bzero((caddr_t) &kl1pt[L1IDX(0)],
		(((256*1024*1024)/NBPG)/NPTEPG) * sizeof(struct pte));
	FLUSH_TLB();
#endif

	/*
	 * Fill out relevant fields in "l" structure.
	 * Some of these (eg, eng, priv_pt) are derivable, but placed
	 * here to allow fast access during system execution.
	 */

	l.me = procid;				/* "me" */
	l.eng = eng;				/* "me", too */
	l.priv_pt = kl1pt;			/* for use_private() */
	l.privstk = (caddr_t)UBTOUSER(ppriv->pp_uarea);
						/* top of private stack */
	l.plocal_pte = (int) ppriv->pp_pmap[0] | PG_V|PG_R|PG_M|PG_KW;
	l.noproc = 1;				/* no process running yet! */
	l.fpuon = (CR0_PG|CR0_PE);		/* how to turn FPU on */
	if (eng->e_flags & E_FPU387) {		/* if 387... */
		l.fpuon |= CR0_ET|CR0_MP;		/*	... set for 387 */

		/* i486--turn on numeric error mode for compatibility */
		if (eng->e_flags & E_SGS2)
			l.fpuon |= CR0_NE;
		l.fpuoff = l.fpuon | CR0_EM;		/* off ==> emulate math */
		init_fpu();				/* 387 needs fninit */
	} else {
		l.fpuon |= CR0_EM;		/*	... set for 387 */
		l.fpuoff = l.fpuon;		/* off ==> emulate math */
	}

	/*
	 * If processor is early rev, set l.fpu_pgxbug to imply checking
	 * for this chip-bug should be done. Read config proms to determine
	 * if D step 386 and set if not.
	 */
	if (arch_type == SYSTYP_S1)
		l.fpu_pgxbug = 1;	/* assume early rev chip */
	else {
		ver386 = SL_P2_386_VER(eng->e_slicaddr);
		l.fpu_pgxbug = (ver386 != SL_P2_386_D);
	}

	/*
	 * Do other processor-local inits.
	 */

	localinit();

	/*
	 * Fill out the engine structure.
	 * Need to do after localinit(), to avoid switcher posting
	 * dispatch SW interrupt to us until we've enabled (properly)
	 * interrupts.
	 */

	eng->e_state = E_GLOBAL;			/* not bound */
	eng->e_pri = eng->e_npri = PIDLE;		/* idle */
	eng->e_head = eng->e_tail = (struct proc *)eng;	/* no affinity */
	eng->e_count = 0;				/* no affinity */
	eng->e_flags &= ~E_OFFLINE;			/* on-line, now! */

#ifdef	MODELD
	if (eng->e_flags&E_SGS2)
		/*
		 * Enable NMI's.
		 */
		enable_snmi(eng->e_slicaddr);
#endif	/* MODELD */

	/*
	 * We're *on*!  Need to goose "eng_wait" sema to complete the init.
	 * This allows another on-line or off-line to start.
	 */

	v_sema(&eng_wait);

	if (arch_type == SYSTYP_S1)
		return;

	/*
	 * Say hello.
	 *
	 * Processor up -- turn processor LED on.
	 */

	if (light_show) {
		DISABLE();
		if (fp_lights)
			FP_LIGHTON(l.me);
		*(int *) VA_LED = 1;
		ENABLE();
	}
}

/*
 * alloclocalIO()
 *	Allocate processor-local IO mapping pages.
 *
 * These are shared by all processors.
 *
 * When/if need to map more such things, should make this table-driven.
 */

static	struct	pte	*slic_map;		/* page to map the SLIC */

static	struct	pte	*led_map;		/* page to map the LED */

ulong volatile *k_usclk = (ulong *) VA_ETC;	/* usclk ptr for kernel use */


static
alloclocalIO()
{

	switch (arch_type) {

	case SYSTYP_S1:
		/*
		 * Allocate mapping for FPA.
		 * Not mapped in maplocalIO(); rather let page fault turn it on
		 * per process.
		 */

		alloc_fpa();
		break;

	case SYSTYP_S27:
	case SYSTYP_S81:
	case SYSTYP_S16:
		/*
		 * Allocate mapping for FPA.
		 * Not mapped in maplocalIO(); rather let page fault turn it on
		 * per process.
		 */

		alloc_fpa();

		/*
		 * Allocate SLIC map.
		 */

		callocrnd(NBPG);
		slic_map = csalloc(NPTEPG, struct pte);
		*(int *) &slic_map[L2IDX(PHYS_SLIC)] =
#ifdef	MODELD
				PHYSTOPTE(PHYS_SLIC) | PG_V|PG_R|PG_M|PG_KW|
						       PG_PCD|PG_PWT;
#else	/* !MODELD */
				PHYSTOPTE(PHYS_SLIC) | PG_V|PG_R|PG_M|PG_KW;
#endif	/* MODELD */
		/*
		 * Allocate map to talk to processor LED.
		 */

		callocrnd(NBPG);
		led_map = csalloc(NPTEPG, struct pte);
		*(int *) &led_map[L2IDX(PHYS_LED)] =
#ifdef	MODELD
				PHYSTOPTE(PHYS_LED) | PG_V|PG_R|PG_M|PG_KW|
						       PG_PCD|PG_PWT;
#else	/* !MODELD */
				PHYSTOPTE(PHYS_LED) | PG_V|PG_R|PG_M|PG_KW;
#endif	/* MODELD */
		/*
		 * Use space in the led map to map the ETC.
		 */

		*(int *) &led_map[L2IDX(PHYS_ETC)] =
#ifdef	MODELD
				PHYSTOPTE(PHYS_ETC) | PG_V|PG_R|PG_M|PG_KW|
						       PG_PCD|PG_PWT;
#else	/* !MODELD */
				PHYSTOPTE(PHYS_ETC) | PG_V|PG_R|PG_M|PG_KW;
#endif	/* MODELD */

		break;
	}
}

/*
 * maplocalIO()
 *	Make sure processor-local resources are mapped by given page-table.
 */

static
maplocalIO(kl1pt)
	register struct pte *kl1pt;
{
	/*
	 * Map the SLIC.
	 *
	 * Note that this consumes addresses 24Meg-28meg on the K20;
	 * if the system has more than 24Meg of memory, this 4Meg is not
	 * 1-1 mapped.  space/param_space.h limits K20 usable memory to
	 * 24Meg to deal with this.  If K20 were to become production,
	 * could just take over the top 64K if there was 28Meg.
	 */

	if (*(int *) &kl1pt[L1IDX(VA_SLIC+KSOFF)]
#ifdef	MODELD
			!= (PHYSTOPTE(slic_map) | PG_V|PG_R|PG_M|PG_KW |
			    l2pt_cacheable)) {
#else	/* !MODELD */
			!= (PHYSTOPTE(slic_map) | PG_V|PG_R|PG_M|PG_KW)) {
#endif	/* MODELD */
		ASSERT(*(int*)&kl1pt[L1IDX(VA_SLIC+KSOFF)] == 0,"maplocalIO: SLIC");
		/*
		 *+ On initialization, the SLIC was mapped to an invalid address.
		 */
		*(int *) &kl1pt[L1IDX(VA_SLIC+KSOFF)] = 
#ifdef	MODELD
				PHYSTOPTE(slic_map) | PG_V|PG_R|PG_M|PG_KW |
				l2pt_cacheable;
#else	/* !MODELD */
				PHYSTOPTE(slic_map) | PG_V|PG_R|PG_M|PG_KW;
#endif	/* MODELD */
	}

	/*
	 * Map the LED (1st time only).
	 */

	if (*(int *) &kl1pt[L1IDX(VA_LED+KSOFF)]
#ifdef	MODELD
			!= (PHYSTOPTE(led_map) | PG_V|PG_R|PG_M|PG_KW |
			    l2pt_cacheable)) {
#else	/* !MODELD */
			!= (PHYSTOPTE(led_map) | PG_V|PG_R|PG_M|PG_KW)) {
#endif	/* MODELD */
		ASSERT(*(int*)&kl1pt[L1IDX(VA_LED+KSOFF)] == 0, "maplocalIO: LED");
		/*
		 *+ On initialization, the LED was mapped to an invalid address.
		 */
		*(int *) &kl1pt[L1IDX(VA_LED+KSOFF)] = 
#ifdef	MODELD
				PHYSTOPTE(led_map) | PG_V|PG_R|PG_M|PG_KW |
				l2pt_cacheable;
#else	/* !MODELD */
				PHYSTOPTE(led_map) | PG_V|PG_R|PG_M|PG_KW;
#endif	/* MODELD */
	}

}

/*
 * localinit()
 *	Init local processor resources.
 *
 * This involves:
 *	turning on the cache,
 *	setting up SLIC interrupt control,
 *	set up and start SLIC local clock.
 *
 * Assumes configure() set up interrupt vector for local clock handler,
 * since we turn on the local clock here.
 */

static
localinit()
{
	extern	int	lcpuspeed;
#ifdef	MODELD
	extern	int	i486_lcpuspeed;
#endif	/* MODELD */
	register struct cpuslic *sl = va_slic;
	/*
	 * start_engine() brought processor up with PROC_CTL register
	 * in apprioriate state (including NMI's enabled).
	 */

	/*
	 * If processor has an FPA, initialize it.
	 */

	if (l.eng->e_flags & E_FPA) {
		l.fpa = 1;
		init_fpa();
		online_fpa();
	}
	/*
	 * compute l.cpu_speed for delay loops. The vaule should correspond
	 * to the "mips" rateing for the processor at its running rate.
	 * e_clock_rate: is a actual board running rate
	 * cpu_speed: is the number of mips if sys_clock_rate = 100
	 *
	 */

#ifdef	MODELD
	if (l.eng->e_flags & E_SGS2)
		/*
		 * Use i486 ipc rather than i386.
		 */
		l.cpu_speed = (i486_lcpuspeed * l.eng->e_cpu_speed) / 100;
	else
#endif	/* MODELD */

	 l.cpu_speed = (lcpuspeed * l.eng->e_cpu_speed) / 100;

	 /*
	  * Fill in l.slic_delay field.  For models B, C, and D the overhead
	  * of the "countdown" loop for the slic lmask delay more than covers
	  * the needed time.  For faster models, the count may need to be
	  * non-zero.
	  */
	l.slic_delay = 0;

	/*
	 * Set up interrupt control and start local clock.
	 */

	(void) splhi();				/* all intrs masked */
	ENABLE();				/* but ON at processor */

	switch (arch_type) {

	case SYSTYP_S1:
		at386_startrtclock();		/* start local timer */
		break;

	case SYSTYP_S27:
	case SYSTYP_S81:
	case SYSTYP_S16:
#ifdef	DEBUG
		if ((sl->sl_ictl & (SL_HARDINT|SL_SOFTINT)) != 0) {
			printf("localinit: pending interrupts 0x%x\n", 
					sl->sl_ictl & (SL_HARDINT|SL_SOFTINT));
			cmn_err(CE_PANIC, "localinit");
		}
#endif
		ASSERT((sl->sl_ictl & (SL_HARDINT|SL_SOFTINT)) == 0,
			"localinit: pending ints");
		/*
	 *+ On initialization, the processor found interrupts pending at the SLIC.
		 */
		sl->sl_ictl = 0x00;		  /* not using `m' bit */
		sl->sl_procgrp = TMPOS_GROUP;	  /* set group ID */
		setgm(sl->sl_procid, SL_GM_ALLON);/* set group-mask all ON */
#ifdef	CHECKSLIC
		ASSERT(sl->sl_gmask == SL_GM_ALLON, "localinit: GM not set");
#endif
		startrtclock();			  /* start local timer */
		break;
	}
	(void) spl0();			  /* intr's unmasked -- go! */
}

/*
 * init_temp_IDT()
 *	Set up temporary IDT to allow taking NMI's during probing.
 */

static	struct	gate_desc temp_IDT[T_NMI+1];	/* temporary IDT */
static	struct	segment_desc temp_GDT[GDT_KDS+1];/* temporary GDT */
static	struct	desctab	desctab;		/* to init IDT, GDT */
static	ushort	selector;			/* temporary selector */

static
#ifdef	MODELD
init_temp_IDT(type)
	int type;
#else	/* !MODELD */
init_temp_IDT()
#endif	/* MODELD */
{
	register struct gate_desc *gd = &temp_IDT[T_NMI];
	extern	void t_nmi();

	/*
	 * Get stand-alone CS selector, so can build interrupt gate.
	 */

	asm("movw %cs, %ax");
	asm("movw %ax, selector");

	/*
	 * Construct interrupt gate for NMI.
	 */

	BUILD_GATE_DESC(gd, selector, t_nmi, INTR_GATE_TYPE, DPL_KERNEL, 0);

	/*
	 * Set up IDT by building a "table-descriptor" and loading
	 * IDT register with it.
	 */

	BUILD_TABLE_DESC(&desctab, temp_IDT, T_NMI+1);
	asm("lidt desctab");

	/*
	 * Enable NMI's on the processor.
	 */

#ifdef	MODELD
	if (type != SLB_SGS2PROCBOARD)
#endif	/* MODELD */
	wrslave(va_slic->sl_procid, PROC_CTL,
		PROC_CTL_NO_SSTEP | PROC_CTL_NO_HOLD | PROC_CTL_NO_RESET);
}

static
init_temp_GDT()
{
	/*
	 * Construct interrupt gate for NMI.
	 */

	BUILD_MEM_DESC(&temp_GDT[GDT_KCS], 0, SD_MAX_SEG,
					CODE_SEG_TYPE, DPL_KERNEL);
	BUILD_MEM_DESC(&temp_GDT[GDT_KDS], 0, SD_MAX_SEG,
					DATA_SEG_TYPE, DPL_KERNEL);

	BUILD_TABLE_DESC(&desctab, temp_GDT, GDT_KDS+1);
	asm("lgdt desctab");
}

/*
 * init_desc_tables()
 *	Set up processor-local copy of GDT and IDT, and start using them.
 */

void	t_diverr(), t_dbg(), t_nmi(), t_int3(), t_into(), t_check(),
	t_und(), t_dna(), t_syserr(), t_res(), t_badtss(), t_notpres(),
	t_stkflt(), t_gpflt(), t_pgflt(), t_coperr();
	
void	t_svc0(), t_svc1(), t_svc2(), t_svc3(), t_svc4(), t_svc5(), t_svc6();

void	bin0int(), bin1int(), bin2int(), bin3int(),
	bin4int(), bin5int(), bin6int(), bin7int();

void	t_nmi_at386(), t_res_at386(), t_fpa_at386(),
	t_pgflt_at386();

void	irq0int(),  irq1int(),  irq2int(),  irq3int(),
	irq4int(),  irq5int(),  irq6int(),  irq7int(),
	irq8int(),  irq9int(),  irq10int(), irq11int(),
	irq12int(), irq13int(), irq14int(), irq15int();

#ifdef	FPA
void	t_fpa();
#endif	/* FPA  */

struct	idt_init {
	int	idt_desc;
	int	idt_type;
	void	(*idt_addr)();
	int	idt_dpl;
};

struct	idt_init idt_init[] = {
	/*	Desc #		Type		Address		Privilege */
	{	T_DIVERR,	TRAP_GATE_TYPE,	t_diverr,	DPL_KERNEL },
	{	T_DBG,		TRAP_GATE_TYPE,	t_dbg,		DPL_KERNEL },
	{	T_INT3,		TRAP_GATE_TYPE,	t_int3,		DPL_USER },
	{	T_INTO,		TRAP_GATE_TYPE,	t_into,		DPL_USER },
	{	T_CHECK,	TRAP_GATE_TYPE,	t_check,	DPL_USER },
	{	T_UND,		TRAP_GATE_TYPE,	t_und,		DPL_KERNEL },
	{	T_DNA,		TRAP_GATE_TYPE,	t_dna,		DPL_KERNEL },
	{	T_SYSERR,	INTR_GATE_TYPE,	t_syserr,	DPL_KERNEL },
	{	T_BADTSS,	INTR_GATE_TYPE,	t_badtss,	DPL_KERNEL },
	{	T_NOTPRES,	TRAP_GATE_TYPE,	t_notpres,	DPL_KERNEL },
	{	T_STKFLT,	TRAP_GATE_TYPE,	t_stkflt,	DPL_KERNEL },
	{	T_GPFLT,	TRAP_GATE_TYPE,	t_gpflt,	DPL_KERNEL },
	{	T_SVC0,		TRAP_GATE_TYPE,	t_svc0,		DPL_USER },
	{	T_SVC1,		TRAP_GATE_TYPE,	t_svc1,		DPL_USER },
	{	T_SVC2,		TRAP_GATE_TYPE,	t_svc2,		DPL_USER },
	{	T_SVC3,		TRAP_GATE_TYPE,	t_svc3,		DPL_USER },
	{	T_SVC4,		TRAP_GATE_TYPE,	t_svc4,		DPL_USER },
	{	T_SVC5,		TRAP_GATE_TYPE,	t_svc5,		DPL_USER },
	{	T_SVC6,		TRAP_GATE_TYPE,	t_svc6,		DPL_USER },
	{ 0 }
};

struct	idt_init idt_sb8000[] = {
	{	T_NMI,		INTR_GATE_TYPE,	t_nmi,		DPL_KERNEL },
	{	T_COPERR,	TRAP_GATE_TYPE,	t_coperr,	DPL_KERNEL },
	{	T_PGFLT,	TRAP_GATE_TYPE,	t_pgflt,	DPL_KERNEL },
	/*
	 * SLIC Bin0-7 Interrupts.  All must be interrupt-gate to avoid
	 * re-enter interrupt handler before ready to; SLIC asserts interrupt
	 * request line until we read the 8-bit vector number.
	 */
	/*	Desc #		Type		Address		Privilege */
	{	T_BIN0,		INTR_GATE_TYPE,	bin0int,	DPL_KERNEL },
	{	T_BIN1,		INTR_GATE_TYPE,	bin1int,	DPL_KERNEL },
	{	T_BIN2,		INTR_GATE_TYPE,	bin2int,	DPL_KERNEL },
	{	T_BIN3,		INTR_GATE_TYPE,	bin3int,	DPL_KERNEL },
	{	T_BIN4,		INTR_GATE_TYPE,	bin4int,	DPL_KERNEL },
	{	T_BIN5,		INTR_GATE_TYPE,	bin5int,	DPL_KERNEL },
	{	T_BIN6,		INTR_GATE_TYPE,	bin6int,	DPL_KERNEL },
	{	T_BIN7,		INTR_GATE_TYPE,	bin7int,	DPL_KERNEL },
	/*
	 * Weitek FPA exception vector.
	 */
	{	T_FPA,		INTR_GATE_TYPE,	t_fpa,		DPL_KERNEL },
	{ 0 }
};

struct	idt_init idt_at386[] = {
	{	T_NMI,		INTR_GATE_TYPE,	t_nmi_at386,	DPL_KERNEL },
	{	T_PGFLT,	TRAP_GATE_TYPE,	t_pgflt_at386,	DPL_KERNEL },
	/*
	 * PC/3AT hardware interrupts.
	 */
	/*	Desc #		Type		Address		Privilege */
	{	T_IRQ0,		INTR_GATE_TYPE, irq0int,	DPL_KERNEL },
	{	T_IRQ1,		INTR_GATE_TYPE, irq1int,	DPL_KERNEL },
	{	T_IRQ2,		INTR_GATE_TYPE, irq2int,	DPL_KERNEL },
	{	T_IRQ3,		INTR_GATE_TYPE, irq3int,	DPL_KERNEL },
	{	T_IRQ4,		INTR_GATE_TYPE, irq4int,	DPL_KERNEL },
	{	T_IRQ5,		INTR_GATE_TYPE, irq5int,	DPL_KERNEL },
	{	T_IRQ6,		INTR_GATE_TYPE, irq6int,	DPL_KERNEL },
	{	T_IRQ7,		INTR_GATE_TYPE, irq7int,	DPL_KERNEL },
	{	T_IRQ8,		INTR_GATE_TYPE, irq8int,	DPL_KERNEL },
	{	T_IRQ9,		INTR_GATE_TYPE, irq9int,	DPL_KERNEL },
	{	T_IRQ10,	INTR_GATE_TYPE, irq10int,	DPL_KERNEL },
	{	T_IRQ11,	INTR_GATE_TYPE, irq11int,	DPL_KERNEL },
	{	T_IRQ12,	INTR_GATE_TYPE, irq12int,	DPL_KERNEL },
	{	T_IRQ13,	INTR_GATE_TYPE, irq13int,	DPL_KERNEL },
	{	T_IRQ14,	INTR_GATE_TYPE, irq14int,	DPL_KERNEL },
	{	T_IRQ15,	INTR_GATE_TYPE, irq15int,	DPL_KERNEL },
	/*
	 * Weitek FPA exception vector.
	 */
	{	T_FPA,		INTR_GATE_TYPE,	t_fpa_at386,	DPL_KERNEL },
	{ 0 }
};


static
init_desc_tables(pl)
	register struct plocal *pl;
{
	register struct gate_desc *gd;
	register struct idt_init *id;
	void setup_seg_regs();
	void sys_call(), sig_clean();

	/*
	 * Build GDT.  Use per-processor TSS to avoid problems with
	 * BUSY TSS when on-line processors.
	 */

	BUILD_MEM_DESC(&pl->gdt[GDT_KCS], KSOFF, SD_MAX_SEG,
					CODE_SEG_TYPE, DPL_KERNEL);
	BUILD_MEM_DESC(&pl->gdt[GDT_KDS], KSOFF, SD_MAX_SEG,
					DATA_SEG_TYPE, DPL_KERNEL);
	BUILD_MEM_DESC(&pl->gdt[GDT_UCS], USOFF, SD_MAX_SEG,
					CODE_SEG_TYPE, DPL_USER);
	BUILD_MEM_DESC(&pl->gdt[GDT_UDS], USOFF, SD_MAX_SEG,
					DATA_SEG_TYPE, DPL_USER);
	BUILD_SYS_DESC(&pl->gdt[GDT_TSS], (caddr_t)&pl->tss + KSOFF,
			sizeof(struct tss_seg), TSS_SEG_TYPE, DPL_KERNEL);
	BUILD_SYS_DESC(&pl->gdt[GDT_LDT], (caddr_t)pl->ldt + KSOFF,
				sizeof(pl->ldt), LDT_SEG_TYPE, DPL_KERNEL);


	/*
	 * Build IDT for different machines.
	 */

	switch (arch_type) {

	case SYSTYP_S1:
		/*
		 * Initialize IDT to all "t_res_at386",
		 * then fill out from table.
		 */
		for (gd = pl->idt; gd < &pl->idt[IDT_SIZE]; gd++)
			BUILD_GATE_DESC(gd, KERNEL_CS, t_res_at386,
				INTR_GATE_TYPE, DPL_KERNEL, 0);
		for (id = idt_at386; id->idt_addr != NULL; id++) {
			gd = &pl->idt[id->idt_desc];
			BUILD_GATE_DESC(gd, KERNEL_CS, id->idt_addr,
						id->idt_type, id->idt_dpl, 0);
		}
		break;

	case SYSTYP_S27:
	case SYSTYP_S81:
	case SYSTYP_S16:
		/*
		 * Initialize IDT to all "t_res",
		 * then fill out from table.
		 */
		for (gd = pl->idt; gd < &pl->idt[IDT_SIZE]; gd++)
			BUILD_GATE_DESC(gd, KERNEL_CS, t_res, INTR_GATE_TYPE,
				DPL_KERNEL, 0);
		for (id = idt_sb8000; id->idt_addr != NULL; id++) {
			gd = &pl->idt[id->idt_desc];
			BUILD_GATE_DESC(gd, KERNEL_CS, id->idt_addr,
						id->idt_type, id->idt_dpl, 0);
		}
		break;
	}

	/*
	 * Fill out common IDT entries.
	 */

	for (id = idt_init; id->idt_addr != NULL; id++) {
		gd = &pl->idt[id->idt_desc];
		BUILD_GATE_DESC(gd, KERNEL_CS, id->idt_addr,
					id->idt_type, id->idt_dpl, 0);
	}

	/*
	 * Build LDT.
	 */

	BUILD_GATE_DESC(&pl->ldt[LDT_SCALL], KERNEL_CS, sys_call,
				CALL_GATE_TYPE, DPL_USER, 1);
	BUILD_GATE_DESC(&pl->ldt[LDT_SIGCALL], KERNEL_CS, sig_clean,
				CALL_GATE_TYPE, DPL_USER, 1);

	/*
	 * Init GDTR and IDTR to start using new stuff.
	 */

	BUILD_TABLE_DESC(&desctab, (caddr_t)pl->idt + KSOFF, IDT_SIZE);
	asm("lidt desctab");

	BUILD_TABLE_DESC(&desctab, (caddr_t)pl->gdt + KSOFF, GDT_SIZE);
	asm("lgdt desctab");

	/*
	 * Fill out TSS so setup_seg_regs() can load the register.
	 * TSS was zeroed by selfinit() when whole struct plocal got zapped.
	 * Make I/O instructions generate a fault.
	 */

	pl->tss.ts_ss0 = KERNEL_DS;
	pl->tss.ts_esp0 = VA_UAREA;
	pl->tss.ts_bmapoff = 0xffff;

	/*
	 * Initialize the segment registers, TSS register, LDT register.
	 */

	setup_seg_regs();
}

/*
 * calloc()
 *	Allocate zeroed memory at boot time.
 *
 * Done via bumping "curmem" value.
 *
 * Skips holes in physical memory after memory is configured (topmem != 0).
 * Assumes allocations to that point are in memory contiguous from physical 0.
 *
 * callocrnd() is used to round up so that next allocation occurs
 * on a given boundary.
 */

caddr_t
calloc(size)
	int	size;
{
	caddr_t	val;

	ASSERT(calloc_ok, "calloc: too late");
	/*
	 *+ The calloc function to allocate kernel memory was called too
	 *+ late in system initialization.  After this point, other kernel
	 *+ routines for memory allocation should have been used.
	 */

	size = (size + (sizeof(int) - 1)) & ~(sizeof(int)-1);

	/*
	 * If ok to check, insure memory exists and skip hole if necessary.
	 * Skipping hole puts curmem on hole boundary, thus arbitrary alignment.
	 * Doing this in terms of CLBYTES is independent of memory bit-map
	 * grain size.
	 */

	if (topmem != 0) {
		while (!cmem_exists(curmem, size)) {
			curmem = (caddr_t) (((int)curmem + CLBYTES) & ~CLOFSET);
			ASSERT(curmem < topmem, "calloc: not enough memory");
			/*
			 *+ The kernel tried to allocate memory and sufficient
			 *+ resources were not available.
			 */
		}
	}

	/*
	 * Allocate and clear the memory.
	 */

	val = (caddr_t) curmem;
	curmem += size;

	bzero(val, (unsigned)size);
	return(val);
}

callocrnd(bound)
	int	bound;
{
	curmem = (caddr_t)roundup((int)curmem, bound);
}

/*
 * cmem_exists()
 *	Check for existence of memory from a given address for a given size.
 */

static
cmem_exists(paddr, size)
	caddr_t	paddr;
	register int	size;
{
	register int	pg;

	size += (int) paddr & (NBPG-1);
	for (pg = btop(paddr); size > 0; size -= NBPG, pg++)
		if (!page_exists(pg))
			return(0);
	return(1);
}

