/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ident	"$Header: kern_exit.c 1.62 90/10/01 $"

/*
 * kern_exit.c
 *	Exit() and wait() system calls.
 */

/* $Log:	kern_exit.c,v $
 */

#include "sys/types.h"
#include "sys/devsw.h"
#include "sys/param.h"
#include "sys/debug.h"
#include "sys/signal.h"
#include "sys/mutex.h"
#include "sys/systm.h"
#include "sys/map.h"
#include "sys/user.h"
#include "sys/wait.h"
#include "sys/errno.h"
#include "sys/timer.h"
#include "sys/session.h"
#include "sys/proc.h"
#include "sys/buf.h"
#include "sys/vm.h"
#include "sys/file.h"
#include "sys/vnode.h"
#include "sys/var.h"
#include "sys/stream.h"
#include "sys/uio.h"
#include "sys/dio.h"

#include "sys/engine.h"

#include "sys/pte.h"
#include "sys/reg.h"
#include "sys/psl.h"
#include "sys/plocal.h"
#include "sys/gate.h"
#include "sys/intctl.h"

#include "sys/cmn_err.h"

#ifdef DEBUG
extern	int	jc_debug;
#endif

struct proc *zombie_todo_list;	/* List of proc entries to free */
lock_t zombie_todo;		/* Interlock for list */
unsigned zombies_done;		/* # entries ever queued on zombie_todo_list */

/*
 * empty_zombie_todo()
 *	Called when zombie_todo_list contains entries which couldn't be
 *	freed on the usual freeproc() path.  Take zombie_todo lock, free
 *	each entry, unlock and return.
 */
empty_zombie_todo()
{
	register struct proc *q, *pp;
	spl_t s1, s2;

	s1 = p_lock(&proc_list, SPL_PROC_LIST);
	s2 = p_lock(&zombie_todo, SPLHI);

	/*
	 * For each entry, pull it off the list and free it
	 */
	while (zombie_todo_list) {
		pp = zombie_todo_list;
		zombie_todo_list = pp->p_sptr;

		/*
		 * Update v.ve_proc if we're freeing the highest slot
		 */
		if (pp+1 == v.ve_proc) {
			for (q = pp-1; q > &proc[1]; q--)
				if (q->p_stat != NULL)
					break;
			v.ve_proc = q+1;
		}

		/*
		 * Actually mark it free
		 */
		pp->p_stat = NULL;
	}
	v_lock(&zombie_todo, s2);
	v_lock(&proc_list, s1);
}

/*
 * rexit()
 *	Exit system call: pass back caller's arg.
 */
rexit()
{
	register struct a {
		int	rval;
	} *uap;

	uap = (struct a *)u.u_ap;
	exit((uap->rval & 0377) << 8);
}

/*
 * exit()
 *	Calling process dies.
 *
 * Assumes always called at spl0.
 * Assumes SPL0 <= SPLSESS <= SPL_PROC_LIST <= SPLHI
 * Allow custom system calls a chance to deal with exit; call when
 * process image and all file descriptors still exist.
 * Release resources.
 * Save u. area statistics for parent to look at.
 * Enter zombie state.
 * Wake up parent and init processes, and dispose of children.
 */

exit(rv)
	int rv;
{
	register struct proc *p;	/* must be a register for use_stack */
	register struct proc *q;
	register struct proc *next;
	register int i;
	struct	session	*s;
	sigset_t	mask;
	int	x;
	int	flag = 0;
	int	tmp_orph;

	p = u.u_procp;
	s = p->p_session;

	/*
	 * If a vfork child, give VM resources back to the parent.
	 * Done here to avoid deadlock with SMPTRC (debugger is parent).
	 */

	if (p->p_flag & SVFORK) {
		/*
		 * Ok to race with parent switching VM resources here,
		 * since we're not swappable and p_sema doesn't need
		 * those resources.  No need to clear SVFORK and SWTOK,
		 * since new fork into this proc-slot writes new flags
		 * (and zaps p_noswap).
		 */
		++p->p_noswap;
		cust_sys_exit();
		v_sema(&p->p_pptr->p_vfork);
		(void)p_sema(&p->p_vfork, PZERO-1);
	}

	/*
	 * If process traced by a multi-process debugger, tell it we're
	 * about to die, then remove self from debugger's list.  These
	 * can race with debugger exiting and turning off SMPTRC.
	 */

	if (p->p_flag & SMPTRC) {
		(void) p_lock(&p->p_state, SPLHI);	/* was SPL0 */
		if (p->p_flag & SMPTRC)
			mpt_stop(PTS_EXIT);
		v_lock(&p->p_state, SPL0);
		mpt_remove_process(p);
	}

	/*
	 * Set all signals to be ignored.
	 */
	(void) p_lock(&p->p_state, SPLHI); 	/* was SPL0 */
	SIGM_CLR(mask);
	for (i=1; i<NSIG; i++) {
		SIGM_ADDSET(mask, sigmask(i));
		u.u_sa[i].sa_handler = SIG_IGN;
	}
	SIGM_ADDSET(p->p_sigignore, mask);
	SIGM_DELSET(p->p_sigcatch, mask);
	SIGM_DELSET(p->p_sig, mask);
	SIGM_DELSET(p->p_sigmask, mask);
	v_lock(&p->p_state, SPL0);
	
	/*
	 * POSIX IEEE std 1003.1-1988 3.2.2
	 * (6) If the process is a controlling process, the SIGHUP signal
	 * shall be sent to each process in the foreground process group of
	 * the controlling terminal belonging to the calling process.
	 *
	 * To be the controlling process (2.3) we must be a session leader,
	 * and there must exist a controlling terminal.
	 */
	if (p->p_pid == p->p_sid) {
		(void)p_lock(&s->s_lock, SPLSESS);	/* was SPL0 */
		if (s->s_vp != NULL) {
			ASSERT_DEBUG(s->s_ctty, "exit() non-null s_ctty");
			signal(s->s_ctty->c_pgrp, SIGHUP);
			/*
	 		 * POSIX IEEE std 1003.1-1988 3.2.2
			 * (7) If the process is a controlling process, the
			 * controlling terminal associated with the session
			 * is disassociated from the session, allowing it to
			 * be acquired by a new controlling process.
			 */
			relinquish_ctty(s);	/* s_lock released */
		} else
			v_lock(&s->s_lock, SPL0);
	}

	disable_fpu();				/* not using FPU anymore */
#ifdef	FPA
	if (p->p_fpa) 
		disable_fpa();			/* not using FPA anymore */
#endif	/* FPA  */

	/* 
	 * unset the interval timer.
	 */
	untimeout(p->p_rtid);

	ofile_deref(u.u_ofile_tab);
	if (u.u_fpref != NULL) {
		/*
		 * Got here without exiting from syscall that bump'd reference.
		 * Only current case is in procxmt() for PT_KILL.
		 */
		closef(u.u_fpref);
	}

	/*
	 * If it exist de-reference the asyncronous I/O object.
	 */

	if (p->p_aio != NULL) {
		aio_deref(p->p_aio, p);
		p->p_aio = NULL;
	}

	/*
	 * Release virtual memory, if not a vfork.  Handled vfork case above.
	 */

	if ((p->p_flag & SVFORK) == 0) {
		p->p_noswap++;				/* avoid swaps */
		cust_sys_exit();
		vrelvm();
	}

	VN_RELE(u.u_cdir);
	if (u.u_rdir) {
		VN_RELE(u.u_rdir);
	}
	semexit();
	acct(rv);
	crfree(u.u_cred);

	l.multprog--;
	/*
	 * Code in panic/pause_self assumes that noproc is set before
	 * using the private stack/pages. This is true currently
	 * but if it changes, need to mod the way fpa state is
	 * saved on panic.
	 */
	l.noproc = 1;		/* Set so hardclock doesn't mess u.u_ru */

	ASSERT(p->p_pid != 1, "exit: init died");
	/*
	 *+ The init process has exited.
	 */
	p->p_xstat = rv;

	/*
	 * Decrement processor affinity count if applicable.
	 * Do this here, since we've done all the device closing/etc
	 * that this process will do.
	 */

	if (p->p_affinity != ANYENG) {
		(void) p_lock(&engtbl_lck, SPLHI);	/* was SPL0 */
		l.eng->e_count--;
		v_lock(&engtbl_lck, SPL0);
	}

	/*
	 * remove this proccess form its sessions linkage.
	 * if no other members are in this session then free its
	 * session structure.
	 */
	(void)p_lock(&s->s_lock, SPLSESS);		/* was SPL0 */
#ifdef DEBUG
	if (jc_debug)
		printf("exit() proc_leave(%d)\n", p->p_pid);
#endif /* DEBUG */
	tmp_orph = p->p_flag & SORPH;
	q = p->p_pgrpb;
	if (proc_leave(p)) {
		freesess(s);
	} else {
		/*
		 * we may have orphaned our process group
		 * see POSIX IEEE 3.2.2.2 (8) below.
		 */
		if ((q->p_flag & SORPH) && !(tmp_orph))
			any_stopped(q);
	}

	/*
	 * Remove process from hash list.
	 */

	(void)p_lock(&proc_list, SPL_PROC_LIST);	/* was SPLSESS */

	i = PIDHASH(p->p_pid);
	x = p - proc;
	if (pidhash[i] == x)
		pidhash[i] = p->p_idhash;
	else {
		for(i = pidhash[i]; i != 0; i = proc[i].p_idhash)
			if (proc[i].p_idhash == x) {
				proc[i].p_idhash = p->p_idhash;
				goto found;
			}
		cmn_err(CE_PANIC, "exit: pidhash");
		/*
		 *+ An exiting process's pid was not found
		 *+ on the process pid hash list.
		 */
	}
found:
	/* 
	 * If the process has any children, tell init about them.
	 *
	 * The session structure is still locked.
	 * proc_list and p_state must be locked (in that order) when
	 * changing the p_pptr of an active process (non-ZOMBIE).
	 * Therefore, use p_pptr (read it) locking one or the other is
	 * sufficient. p_state is also needed locked for signal purposes;
	 * lpsignal() and access to p_sig in particular.
	 *
	 * if one of our children is not in our process group and he
	 * is not a member of the orphaned process then maybe he is now.
	 * if he was in our process group then we have already taken care
	 * of us leaveing.
	 */

	for (q = p->p_cptr; q; q = next) {
		next = q->p_sptr;
		q->p_sptr = proc[1].p_cptr;
		proc[1].p_cptr = q;

		(void) p_lock(&q->p_state, SPLHI);	/* was SPL_PROC_LIST */
		/*
		 * POSIX 3.2.2.2
		 * 
		 * Children of a terminating process shall be assigned a 
		 * new  parent process ID, corresponding to an 
		 * implementation-defined system process.
		 */
		q->p_pptr = &proc[1];
		q->p_ppid = 1;

		tmp_orph = q->p_flag & SORPH;
		if (q->p_stat == SZOMB)
			flag = 1;
		else if ((q->p_flag & STRC) == STRC &&
					(q->p_flag & SMPTRC) == 0) {
			/*
			 * Traced processes are killed since their existence
			 * means someone is screwing up.
			 */
			q->p_flag &= ~STRC;
			lpsignal(q, SIGKILL);
		} 

		v_lock(&q->p_state, SPL_PROC_LIST);	/* now SPL_PROC_LIST */

		/*
		 * drop lock since we may wish to orphan the child.
		 * (session is still locked)
		 */
		if ((q->p_sid == p->p_sid) && (q->p_pgrp != p->p_pgrp) &&
							!(q->p_flag&SORPH)){
#ifdef DEBUG
			if (jc_debug)
				printf("exit() pid %d,pgrp %d may have orphaned pid %d in pgrp %d\n", p->p_pid, p->p_pgrp, q->p_pid, q->p_pgrp);
#endif
			if (q->p_pgrpf == q) {
				(void) p_lock(&q->p_state, SPLHI);
				q->p_flag |= SORPH;
				v_lock(&q->p_state, SPL_PROC_LIST);	/* now SPL_PROC_LIST */
			} else {
				orphan(q, p);
			}
			/*
			 *
			 * POSIX 3.2.2.2
			 *
			 * If the implementation supports job control, and if
			 * the exit of the process causes a process group to
			 * become orphaned, and if any member of the 
			 * newly-orphaned process group is stopped, then a 
			 * SIGHUP signal followed by a SIGCONT signal shall be
			 * sent to each process in the newly-orphaned process 
			 * group.
			 */
			if ((q->p_flag & SORPH) && !(tmp_orph))
				any_stopped(q);
		} 
	}

	if (flag)
		psignal(&proc[1], SIGCLD);

	if (p->p_cptr) {
		v_event(&proc[1].p_zombie);
		p->p_cptr = NULL;			/* no more kids */
	}

	/*
	 * If process is a multi-process debugger, release debug "children".
	 * Any SMPSTOP actual children were left alone, above.
	 * Note: proc_list still held.
	 */

	if (p->p_flag & SMPDBGR) {
		for(q = &proc[p->p_mptc]; q != &proc[0]; q = &proc[q->p_mpts]) {
			(void) p_lock(&q->p_state, SPLHI); /* was SPLPROC_LIST */
			q->p_flag &= ~(SMPTRC | STRC);
			/*
			 * SMPSTOP ==> hung in mpt_stop() -- let it continue.
			 * unless it was started by the above setrun() and
			 * not yet cleared SMPSTOP.
			 */
			if ((q->p_flag & SMPSTOP) && (q->p_stat == SSTOP))
				setrun(q);
			v_lock(&q->p_state, SPL_PROC_LIST);
		}
	}

	v_lock(&proc_list, SPLSESS);	/* was SPL_PROC_LIST now SPLSESS */
	v_lock(&s->s_lock, SPL0);

	/*
	 * Free any auxiliary data structures
	 */
	shadow_exit(p);

	/*
	 * Need to release U-area (which contains current stack), and
	 * page-tables (which may contain per-process kernel level-1 PT,
	 * depending on the HW).
	 *
	 * Since TMP, can't assume memory stays around thru swtch, thus must
	 * switch to private stack.  Must not block once this is done.
	 * Also must not use old values of stack variables or arguments
	 * since stack is gone.
	 *
	 * This code assumes memfree()/etc use lock/gate to lock the
	 * memory free-lists.  It won't work if mem_alloc is a semaphore,
	 * since must not block once on private stack.
	 */

	ASSERT_DEBUG(p->p_szpt != 0, "exit: szpt");

	use_private();				/* get on private stack */
	vrelpt(p);				/* Release page-tables */
	/*
	 * get stats
	 *	Use the proc structure's pointer to the uarea.
	 *	This is done becuase the use_private() call ALSO
	 *	makes the symbol "u" reference the processor's
	 *	PRIVATE uarea.
	 * NOTE all locals to exit have now been lost!
	 */
	p->p_un.pu_pst = p->p_uarea->u_pst;
	pstadd(&p->p_un.pu_pst, &p->p_uarea->u_cpst);

	vrelu(p, 0);				/* Release U-area */

	/*
	 * Just about dead.  Lock self state so we can become a zombie
	 * and tell parent that we just died.  Parent wait() synch's
	 * with this by locking process state.
	 */

	(void) p_lock(&p->p_state, SPLHI);
	q = p->p_pptr;
	/*
	 * don't zero the pgrp 'cos waipid still needs it.
	 * p->p_pgrp = 0;
	 */
	p->p_stat = SZOMB;
	p->p_flag &= ~STRC;
	p->p_flag |= SORPH;
	p->p_cursig = 0;	/* make sure zombie is not set running */
	SIGM_CLR(p->p_sig);
	v_event(&q->p_zombie);
	v_lock(&p->p_state, SPL0);	
	/*
	 * POSIX 3.2.2.2
	 *
	 * If the implementation supports SIGCHLD signal, a SIGCHLD shall
	 * be sent to the parent process.
	 */
	psignal(q, SIGCLD);

	swtch((struct proc *)NULL);	/* cast is for lint! */
	/* Dead Babe */
}


/*
 * wait()
 *	Wait system call.
 * pid returned in rval1
 * status returned in rval2
 */
wait()
{
	/*
	 * POSIX IEEE std 1003.1-1988 3.2
	 * The waitpid function shall behave identically to wait()
	 * function, if the pid argument has a value of -1 and the options
	 * argument has a value of zero.
	 */
	u.u_error = wait1(-1,0);
	return;
}

/*
 * waitpid()
 *	Waitpid system call.
 * pid returned in rval1
 * status returned in rval2
 */
waitpid()
{
	register struct a {
		int	pid;
		int	opt;
	} *uap;

	uap = (struct a *)u.u_ap;

#ifdef DEBUG
	if (jc_debug)
		printf("waitpid pid 0x%x opt 0x%x\n",uap->pid,uap->opt);
#endif
	if ((uap->opt & ~(WNOHANG|WUNTRACED)) != 0)
		u.u_error = EINVAL;
	else
		u.u_error = wait1((pid_t)uap->pid, uap->opt);
	return;
}

/*
 * wait1()
 *	Wait() and waitpid() system call implementation.
 *
 * Search for a terminated (zombie) child,
 * finally lay it to rest, and collect its status.
 * Look also for stopped (traced) children,
 * and pass back status from them.
 */

wait1(pid,options)
	pid_t	pid;
	int options;
{
	register struct proc *p;
	register struct proc *q;
	register struct proc *child;
	int		     found;
	spl_t		     s;

	p = u.u_procp;
	u.u_r.r_val1 = 0;
	/*
	 * re-init event since we're about to
	 * scan the children.  A concurrent exit of a child can
	 * be missed, but it will v_event() our p_zombie, so
	 * we'll not block below.
	 */

	clear_event(&p->p_zombie);

	for(;;) {
		found = 0;

		q = NULL;
		for(child = p->p_cptr; child; q = child, child = child->p_sptr) {
			/*
			 * POSIX IEEE std 1003.1-1988 3.2
			 * if pid is greater than zero, it specifies the
			 * process ID of a single child process for
			 * which [the] status is requested.
			 */
			if ((pid > 0) && (child->p_pid != pid))
				continue;
			/*
			 * POSIX IEEE std 1003.1-1988 3.2
			 * if pid is equal to zero, [the] status is
			 * requested for any child process whose process
			 * ID is equal to that of the calling process.
			 */
			if ((pid == 0) && (child->p_pgrp != p->p_pgrp))
				continue;
			/*
			 * POSIX IEEE std 1003.1-1988 3.2
			 * if pid is less than -1, [the] status is
			 * requested for any child process whose process
			 * group ID is equal to the absolute value of pid
			 */
			if ((pid < -1) && (child->p_pgrp != -pid))
				continue;

			if (child->p_stat == SZOMB) {
				freeproc(child, q, 1);
				return (0);
			}
			found = 1;

			/*
			 * If stopped child, then return it if tracing or
			 * interested in stopped children.
			 * Ok to check state w/o entering child state lock
			 * since signal/v_sema to parent is done after child
			 * changes state to SSTOP (HW specific proof of this).
			 * We must check p_stat again since the child may
			 * have received a SIGCONT...
			 */

			if ((child->p_stat == SSTOP)
			&&  (child->p_flag&SWTED) == 0
			&&  (child->p_flag&STRC || options&WUNTRACED)) {
				s = p_lock(&child->p_state, SPLHI);
				if (child->p_stat == SSTOP) {
					child->p_flag |= SWTED;
					u.u_r.r_val1 = child->p_pid;
					u.u_r.r_val2 = (child->p_cursig<<8) | WSTOPPED;
					v_lock(&child->p_state, s);
					return (0);
				}
				/* lost the race. Child no longer asleep */
				v_lock(&child->p_state, s);
			}
		}

		/*
		 * POSIX IEEE std 1003.1-1988 3.2.1.4
		 * [ECHILD] The process or process groups specified by pid
		 * does not exist or is not a child od the calling process.
		 */

		if ((p->p_cptr == NULL) || ((pid != -1) && !found)) 
			return (ECHILD);

		/*
		 * We have children.  Return immediately if options say to.
		 */

		if (options&WNOHANG) 
			return (0);

		/*
		 * Wait for a child to change state.
		 */

		(void)p_event(&p->p_zombie, PWAIT);
	}
}

/*
 * Remove zombie children from the process table.
 * Called by wait1(), if flag == 1, and issig()
 * if flag == 0.
 *	child - process being freed
 *	q - process before child in p-c-s list
 *	    or NULL if first entry
 */
freeproc(child, q, flag)
	register struct proc *child;
	register struct proc *q;
{
	register struct proc *p;
	spl_t	s;
	int gone;
	int am_init;

	p = u.u_procp;
	am_init = (p == &proc[1]);

	/*
	 * Synch with exiting child: insure it's gone.
	 * Then copy status (only if called by wait)
	 * and statistics.
	 * If signal attempt is made after this, it 
	 * will no longer find the pid.
	 */

	s = p_lock(&child->p_state, SPLHI);
	if (flag) {
		u.u_r.r_val1 = child->p_pid;
		u.u_r.r_val2 = child->p_xstat;
	}
	child->p_pid = 0;
	child->p_sid = 0;
	child->p_session = &sess[0];
	child->p_pgrp = 0;
	child->p_flag = 0;
	child->p_ppid = 0;
	child->p_xstat = 0;
	v_lock(&child->p_state, s);
	pstadd(&u.u_cpst, &child->p_un.pu_pst);

	/*
	 * Put child to rest.
	 */

	child->p_sig = 0;
	child->p_wchan = NULL;
	child->p_cursig = 0;
	child->p_sigcatch = 0;
	SIGM_CLR(child->p_sigmask);
	child->p_sigignore = 0;
	child->p_pptr = NULL;

	/*
	 * If we're init, we need to lock proc_list before we
	 * manipulate our p-c-s chain.  Unlike all other process slots
	 * on the system, init's may be manipulated by other processes
	 * "on the fly".
	 */
	if (am_init) {
		s = p_lock(&proc_list, SPL_PROC_LIST);
	}

	/*
	 * Remove our child from his list if we can.  For a "normal" parent
	 * we have the previous pointer "q" unless it's the first on the
	 * chain, in which case we use p_cptr.  For init, it is possible for
	 * us to race with someone inserting a new entry on to our chain
	 * even as we try to remove one.  In this case, it is possible for
	 * "q" to be NULL (because it was the first on the chain when we
	 * looked), but for p_cptr to NOT point to it (because somebody was
	 * inserted after "q" was calculated).  In this fun case, we must
	 * then re-scan the chain now that it's locked, and figure out who
	 * is really before us now.
	 */
	gone = 1;
	if (q != NULL)
		q->p_sptr = child->p_sptr;
	else if (child == p->p_cptr)
		p->p_cptr = child->p_sptr;
	else
		gone = 0;

	ASSERT(am_init || gone, "freeproc: corrupt child chain");
	/*
	 *+ While a process was trying to free a child process, it
	 *+ discovered that its child chain was corrupt.
	 */

	/*
	 * See if we can take the proc_list right now.  Take special actions
	 * if we can't.
	 */
	if (!am_init &&
		((s = cp_lock(&proc_list, SPL_PROC_LIST)) == CPLOCKFAIL)) {
		/*
		 * We couldn't take the proc list.  We are a "regular" sort
		 * of process, and could be on the brink of deadlock with paths
		 * like kill()--it takes the proc_list, then goes for a p_state.
		 * We have our p_state locked, and are going for the proc_list.
		 * Since our process-oriented work is done, instead queue to
		 * the zombie_todo_list, and let it be placed on the actual
		 * proc freelist when softclock comes around.
		 */
		s = p_lock(&zombie_todo, SPLHI);
		child->p_sptr = zombie_todo_list;
		zombie_todo_list = child;
		v_lock(&zombie_todo, s);
		zombies_done++;
		return;
	}

	/*
	 * We are here on the normal path when we could get the lock, or
	 * on init's path in any case.  For init, do a brute-search of
	 * the child list to find our subject.
	 */
	if (!gone) {
		for (q = p->p_cptr; q; q = q->p_sptr)
			if (q->p_sptr == child)
				break;
		ASSERT(q != NULL, "freeproc: list1");
		/*
		 *+ init has discovered an error in the parent-child-sibling
		 *+ linked list of an exiting process.
		 */
		q->p_sptr = child->p_sptr;
	}

	/* child cleared self p_cptr in exit() */

	/*
	 * Adjust v.ve_proc down if we were the
	 * high-water mark, and free child proc[]
	 * table entry.
	 */
	if (child+1 == v.ve_proc) {
		for(q = child-1; q > &proc[1]; q--) {
			if (q->p_stat != NULL)
				break;
		}
		v.ve_proc = q+1;
	}
	child->p_stat = NULL;	/* free entry */
	v_lock(&proc_list, s);
}
