/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ident	"$Header: kern_mman.c 1.26 90/11/16 $"

/*
 * kern_mman.c
 *	Various memory management procedures.
 *
 * Original Berkeley "ifdef notdef"'s around pseudo-implementations
 * of several of the stubs here has been removed in the interest of
 * readability.
 */

/* $Log:	kern_mman.c,v $
 */

#include "sys/types.h"
#include "sys/devsw.h"
#include "sys/param.h"
#include "sys/debug.h"
#include "sys/signal.h"
#include "sys/systm.h"
#include "sys/mutex.h"
#include "sys/map.h"
#include "sys/user.h"
#include "sys/fcntl.h"
#include "sys/errno.h"
#include "sys/timer.h"
#include "sys/proc.h"
#include "sys/vnode.h"
#include "sys/file.h"
#include "sys/cmap.h"
#include "sys/conf.h"
#include "sys/vm.h"
#include "sys/seg.h"
#include "sys/buf.h
#include "sys/uio.h
#include "sys/dio.h

#include "sys/pte.h"
#include "sys/reg.h"
#include "sys/plocal.h"
#include "sys/intctl.h"
#include "sys/mftpr.h"
#include "sys/cmn_err.h"

getpagesize()
{
	u.u_r.r_val1 = NBPG * CLSIZE;
}

/*
 * brk and sbrk system calls
 *	Adjust size of program data segment.
 *
 * Fail if expanding into a shmem or mmap segment (need for SysV semantics)
 * This requires that we keep track of the data space separately (u_dsegsz).
 * If no mmap or shmem regions, do the standard thing.
 */

sbreak()
{
	struct a {
		uint	nva;
	};
	register int n, d;
	register struct mmap *um;
	struct pte *pte;
	struct proc *p = u.u_procp;

	/*
	 * set n to new data size
	 * set d to new-old
	 */

	n = btoc(((struct a *)u.u_ap)->nva);
	if (n <= u.u_tsize) {
		/*
		 * Can't make any part of text space disappear.
		 */
		u.u_error = EINVAL;
		return;
	}

	/*
	 * ensure doesn't grow into mapped segments
	 */
	for (um = u.u_mmap; um < u.u_mmapmax; um++) {
		if (um->mm_pgcnt &&
		   (um->mm_type == MMT_SHM || um->mm_type == MMT_MMAP) &&
		   n >= um->mm_1stpg &&
		   u.u_dsegsz <= um->mm_1stpg) {
			u.u_error = ENOMEM;
			return;
		}
	}

	/* Make sure there is no pending asynchronous I/O for this process */
	if (p->p_aio) {
		aio_pend_wait(p->p_aio, p, 0);
	}

	if (u.u_dsegsz == u.u_dsize) {
		/*
		 * This is the normal case - no mmap or shared mem segments.
		 * The only other time when this is true is when maps are
		 * established completely within the process's current
		 * address space, ie map over existing memory.
		 */
		d = clrnd(n - u.u_dsize);
		if (u.u_error = chksize((uint)u.u_dsize+d, (uint)u.u_ssize))
			return;
		if (d > 0 && !vsalloc(&u.u_dmap, u.u_dsize, (size_t)d)) {
			u.u_error = EAGAIN;
			return;
		}

		if (expand(d, 0, 0, PG_ZFOD)) {
			vexpandRS(d);		/* grow Rset, if appropriate */
			/*
			 * Shrink is probably rare, but maybe more used due to
			 * mapped files -- free any swap space beyond what's
			 * needed.
			 */
			if ( d < 0)
				vsfree(&u.u_dmap, u.u_dsize, dtoc(maxdmap)-u.u_dsize);
		} else {
			u.u_error = ENOMEM;
		}
	} else {
		/*
		 * Must have maps. If expanding, alloc ZFOD pages. We know
		 * that the page tables don't need to be grown. This is
		 * basically mmap(ZEROFILL).
		 */
		d = clrnd(n - u.u_dsegsz);
		if (n > u.u_dsegsz) {
			if (!vsalloc(&u.u_dmap, u.u_dsegsz, (size_t)d)) {
				u.u_error = EAGAIN;
				return;
			}
			pte = dvtopte(p, u.u_dsegsz);
			ptefill(pte, PG_ZFOD, (int)d);
			l.cnt.v_nzfod += d;
		} else {
			/* shrinking */
			pte = dvtopte(u.u_procp, n);
			u.u_procp->p_rssize -= vmemfree(pte, -d, 1);
			vsfree(&u.u_dmap, (size_t)n, (size_t)-d);
			FLUSH_USER_TLB(p->p_ptb1);
		}
	}
	u.u_dsegsz = n;
}

/*
 * grow()
 *	grow the stack to include the SP
 *
 * True return if successful.  The (rare) failure is due to no swap space
 * when trying to swap out to expand page-table.
 */

grow(sp)
	unsigned sp;
{
	register int si;

	si = clrnd(btoc((USRSTACK-sp)) - u.u_ssize + SINCR);

	if (sp >= USRSTACK-ctob(u.u_ssize)
	||  chksize((uint)u.u_dsize, (uint)u.u_ssize+si) != 0
	||  !vsalloc(&u.u_smap, u.u_ssize, (size_t)si)
	||  !expand(0, si, 0, PG_ZFOD))
		return (0);
	
	vexpandRS(si);			/* grow Rset, if appropriate */
	return (1);
}

/*
 * mmap()
 *	Re-map part of process address space.
 */

mmap()
{
	struct a {
		caddr_t	addr;		/* starting virt-addr */
		int	len;		/* length (bytes) to map */
		int	prot;		/* RO, RW encoding */
		int	share;		/* private/shared modifications */
		int	fd;		/* open file to be mapped */
		ulong	pos;		/* where in file to begin map */
	} *uap = (struct a *)u.u_ap;
	register struct mmap *um;
	register struct pte *pte;
	register size_t	npgcnt;
	register ulong	paddr;
	register int	j;
	struct	proc	*p = u.u_procp;
	struct	file	*fp;
	struct	vnode	*vp;
	struct	mapops	*mapops;
	size_t		pgcnt;
	int		pgs;
	ulong		pospg;
	ulong		firstpg;
	ulong		lastpg;
	ulong		firstdp;
	ulong		nfirstpg;
	ulong		handle;
	int		map_type;
	int		proto_pte;
	int		share = uap->share;	/* convenience copy of arg */
	int		prot = uap->prot;	/* convenience copy of arg */
	int		max_prot;
	int		ods;
	int		nds;
	bool_t		um_extend = 0;
	extern	int	ummap_max_hole;		/* max u_mmap[] concat hole */

	/*
	 * Must have good alignment of arguments, reasonable address space,
	 * not a vfork child...
	 */

	npgcnt = pgcnt = btop(uap->len);
	nfirstpg = firstpg = btop(uap->addr);
	lastpg = firstpg + npgcnt - 1;
	pospg = btop(uap->pos);

	if (uap->len <= 0
	||  (prot & PROT_RDWR) == 0
	||  (((int)uap->addr | uap->len | uap->pos) & CLOFSET) != 0
	||  lastpg < firstpg
	||  isassv(p, firstpg)
	||  (p->p_flag & SVFORK)) {
		u.u_error = EINVAL;
		return;
	}

	/* Make sure there is no pending asynchronous I/O for this process */
	if (p->p_aio) {
		aio_pend_wait(p->p_aio, p, 0);
	}

	/*
	 * Based on share arg, check further and do special case stuff.
	 */

	switch(share) {

	case MAP_PRIVATE:
	case MAP_SHARED:
		/*
		 * Must have valid file-descriptor for vnode,
		 * enough access in the file-descriptor...
		 */

		u.u_error = getvnodefp(uap->fd, &fp);
		if (u.u_error)
			return;
		vp = fp->f_vnode;

		/*
		 * Don't allow mapping a file with the manadatory
		 * modes set.
		 */
		if (vp->v_flag & VMNDLCK) {
			u.u_error = EINVAL;
			return;
		}

		if (share == MAP_PRIVATE) {
			max_prot = PROT_READ;
			if (prot & PROT_WRITE)
				prot &= ~PROT_WRITE;
			else
				share = MAP_SHARED;
		} else
			max_prot = (fp->f_flag&FWRITE) ? PROT_RDWR : PROT_READ;
		if ((prot & max_prot) != (prot & PROT_RDWR)
		||  (fp->f_flag & (FWRITE|FREAD)) == FWRITE) {
			u.u_error = EACCES;
			return;
		}

		/*
		 * Insist on execute access for PROT_EXEC, and !PROT_WRITE.
		 * Execute access avoids issue of mapping RO file and
		 * dis-allowing writes by holding VTEXT mapping.
		 * Mapping function worries about open RW connections, etc.
		 */

		if (prot & PROT_EXEC) {
			if (prot & PROT_WRITE) {
				u.u_error = EINVAL;
				return;
			}
			/*
			 * VOP_ACCESS wants a locked vnode.
			 */
			VN_LOCKNODE(vp);
			u.u_error = VOP_ACCESS(vp, VEXEC, u.u_cred);
			VN_UNLOCKNODE(vp);
			if (u.u_error)
				return;
			max_prot |= PROT_EXEC;
		}

		/*
		 * Ok so far.  If file-descriptor is already mapped, try
		 * to extend previous u_mmap[]. The extension of an
		 * existing u_mmap entry occurs only if the fd wasn't
		 * previously closed and reopened, since the FD_MAPPED
		 * bit gets cleared on every open. This is ok for current
		 * use. Note that this is an optimization to ensure we
		 * don't run out of u_mmap[]s and isn't really visible
		 * users (other than getting premature EMFILES). If not
		 * already mapped (or there was an intervening close/open
		 * between the two mmaps), locate an unused u_mmap[] in
		 * this process.
		 *
		 * Could extend for MAP_PRIVATE if check mm_cor, if this case
		 * becomes important.
		 *
		 * This assumes mapping the same thing gives equivalent map.
		 * There is no good way to check this.
		 */

		um = u.u_mmapmax;
		if (share == MAP_SHARED && vp->v_flag & VMAPPED) {
			/*
			 * Check mm_pgcnt, mm_vp and mm_prot (mm_prot
			 * insures don't mix (eg) "exec" and RO map).
			 * Close() undoes maps, FD_MAPPED is cleared during
			 * open & only left on across exec (but all maps are
			 * cleared during exec).
			 *
			 * Position in file offset from current u_mmap[] must
			 * agree with offset in process virtual space.  This
			 * allows arbitrary overlap (including completely
			 * replacing previous map).
			 *
			 * Also, insure the map isn't too far away: don't
			 * want to waste underlying map function resources.
			 */
			for (um = u.u_mmap; um < u.u_mmapmax; um++) {
				if (um->mm_pgcnt != 0
				&&  um->mm_vp == vp
				&&  um->mm_prot == max_prot
				&&  (pospg-um->mm_off) == (firstpg-um->mm_1stpg)
				&&  firstpg+npgcnt+ummap_max_hole >= um->mm_1stpg
				&&  firstpg <= um->mm_1stpg+um->mm_size+ummap_max_hole) {
					/*
					 * Got a match!  Set up to map whole
					 * region; later unmap previous.
					 */
					if (pospg > um->mm_off) {
						pospg = um->mm_off;
						firstpg = um->mm_1stpg;
					}
					if (lastpg < um->mm_1stpg + um->mm_size)
						lastpg = um->mm_1stpg + um->mm_size - 1;
					pgcnt = lastpg - firstpg + 1;
					break;
				}
				/*
				 * Disallow mapping over shared memory,
				 * test is from MFINTERSECT(). Note that
				 * the test above will always fail since
				 * shared mem maps have mm_fdidx set to -1.
				 */
				if (um->mm_type == MMT_SHM &&
				    um->mm_1stpg < firstpg + npgcnt &&
				    um->mm_1stpg + um->mm_size > firstpg) {
					u.u_error = ENOMEM;
					return;
				}
			}
		}
		if (um == u.u_mmapmax) for (um=u.u_mmap; um < u.u_mmapmax; um++)
			if (um->mm_pgcnt == 0)
				break;
		if (um == u.u_mmapmax) {
			if (u.u_mmapmax >= &u.u_mmap[NUMMAP]) {
				u.u_error = EMFILE;
				return;
			}
			um->mm_pgcnt = 0;		/* in case error */
			++u.u_mmapmax;
		}
		break;

	case MAP_ZEROFILL:
		/*
		 * Don't check the fd argument -- unused.
		 */
		if ((prot & PROT_BITS) != PROT_RDWR) {
			u.u_error = EINVAL;
			return;
		}
		break;

	default:
		u.u_error = EINVAL;
		return;
	}

	/*
	 * If don't fit within existing data-space, create new space,
	 * filling holes with invalid pages.  This doesn't allow
	 * overlap into stack (for now).
	 */

	ods = p->p_dsize;
	nds = vtodp(p, lastpg+1);
	firstdp = vtodp(p, nfirstpg);

	if (u.u_error = chksize((uint)nds, (uint)u.u_ssize))
		return;
	if (share!=MAP_SHARED && !vsalloc(&u.u_dmap,(size_t)firstdp,npgcnt)) {
		u.u_error = EAGAIN;
		return;
	}
	if (nds > ods) {
		if (!expand(nds-ods, 0, 0, PG_INVAL)) {
			u.u_error = ENOMEM;
			return;
		}
		vexpandRS(nds-ods);
	}

	/*
	 * Based on type of vnode, figure handle and mapping proc
	 * and see about the mapping.
	 *
	 * Note that ask for max possible access, since user may
	 * later mprotect() a RO page into a RW page.
	 */

	if (share != MAP_ZEROFILL) {
		mapops = vnode_mmap[(int)vp->v_type];
		VN_LOCKNODE(vp);		/* insure consistent state */
		map_type = (*mapops->map_new)(vp, pospg, pgcnt, max_prot,
				&handle, (int (*)())0, 0);
		VN_UNLOCKNODE(vp);

		if (map_type > 0) {				/* error? */
			if (nds > ods)				/* grew above */
				(void) expand(ods-nds, 0, 0, 0);/* shrink */
			u.u_error = map_type;
			return;
		} else if (map_type != MM_PAGED && share == MAP_PRIVATE)
			share = MAP_SHARED;		/* for now */
	}

	/*
	 * Loose old address-space. No need to clear pte's since we're
	 * calling ptefill below.
	 *
	 * Dis-allow swaps starting now, since u_mmap[] may be temporarily
	 * inconsistent (and mapping function procedures may allow swapping).
	 */

	++p->p_noswap;

	pte = dptopte(p, firstdp);
	p->p_rssize -= vmemfree(pte, (int)npgcnt, 0);

	/*
	 * MAP_ZEROFILL is special-case -- just fill in PT.
	 */

	if (share == MAP_ZEROFILL) {
		ptefill(pte, PG_ZFOD, (int)npgcnt);
		l.cnt.v_nzfod += npgcnt;
		--p->p_noswap;
		return;
	}

	/*
	 * If the u_mmap[] is being extended, MM_UNMAP old version.
	 * If totally overlapped, this went away already in vmemfree().
	 * Note: `handle' may change (eg, mmreg_alloc()).
	 */

	if (um->mm_pgcnt != 0) {
		ASSERT(	um->mm_paged ? map_type == MM_PAGED :(um->mm_noio ? map_type == MM_PHYS :map_type == MM_NPMEM), "mmap: skew");
		/*
		 *+ While unmapping a page, the kernel discovered that
		 *+ the type of the mapped page is inconsistent with
		 *+ other attributes of the page description.
		 */
		(*mapops->map_unmap)(um->mm_handle, um->mm_off,
				um->mm_size, um->mm_prot);
		um_extend = 1;
	}

	/*
	 * Set up mmap structure.
	 *
	 * Note that mm_prot is set for maximum possible file-descriptor access,
	 * not requested access, to anticipate mprotect() function.
	 * Dis-allowed write-only file above (inconsistent with MMU capability).
	 *
	 * MM_PAGED and MM_NPMEM allow IO services, MM_PHYS doesn't.
	 */

	um->mm_off = pospg;
	um->mm_1stpg = firstpg;
	um->mm_size = pgcnt;
	um->mm_pgcnt += npgcnt;				/* += for extending */
	um->mm_handle = handle;
	um->mm_ops = mapops;
	um->mm_vp = vp;
	um->mm_paged = (map_type == MM_PAGED);
	um->mm_noio = (map_type == MM_PHYS);
	um->mm_prot = max_prot;
	um->mm_cor = (share == MAP_PRIVATE);
	um->mm_type = MMT_MMAP;

	/*
	 * If shared map, don't need swap space under these addresses any more.
	 * Special case: if mapping thru end of data-space, make sure
	 * to delete last chunk of swap.
	 */

	if (share == MAP_SHARED) {
		if (firstdp + npgcnt == p->p_dsize)
			vsfree(&u.u_dmap, (size_t)firstdp, (size_t)(dtoc(maxdmap)-firstdp));
		else
			vsfree(&u.u_dmap, (size_t)firstdp, npgcnt);
	}

	/*
	 * Re-paint address space with new mapping.
	 */

	proto_pte = MAPXTOPTE(um-u.u_mmap)
		  | (((prot & PROT_WRITE) || share == MAP_PRIVATE) ? RW : RO)
		  | PG_R;

	if (map_type == MM_PAGED) {
		/*
		 * Set up paged map -- fill with (invalid) proto_pte,
		 * to fault in pages on demand.
		 */
#ifdef	MMU_MBUG
		proto_pte |= PG_M;			/* redundant */
#endif	/* MMU_MBUG  */
		ptefill(pte, proto_pte, (int)npgcnt);
	} else {
		/*
		 * Phys-mapped address-space or non-paged memory:
		 * call driver for each kernel-size page.
		 */
		proto_pte |= PG_M|PG_V;
		firstpg = btop(uap->pos);

		for (pgs = npgcnt; pgs > 0; pgs -= CLSIZE, firstpg += CLSIZE) {
			paddr = (*mapops->map_refpg)(handle, firstpg);
			for (j = 0; j < CLSIZE; j++, pte++, paddr += NBPG)
				*(int*)pte = PHYSTOPTE(paddr) | proto_pte;
		}

		if (um->mm_noio && !um_extend)
			++u.u_pmapcnt;
	}
	--p->p_noswap;					/* ok to swap now */

	/*
	 * Flush MMU TLB, since process mapping changed a bunch.
	 */

	FLUSH_USER_TLB(p->p_ptb1);
}

/*
 * munmap()
 *	Release some address-space and replace with invalid pages.
 */

munmap()
{
	register struct a {
		caddr_t	addr;		/* starting virt-addr */
		int	len;		/* length (bytes) to unmap */
	} *uap = (struct a *)u.u_ap;
	register struct	proc *p = u.u_procp;
	register struct pte *pte;
	register size_t	pgcnt;
	register unsigned firstpg;
	unsigned	lastpg;

	/*
	 * Must have good alignment of arguments and valid address space.
	 *
	 * For now, insist addresses be entirely in data space.
	 * This restriction should be relaxed when mmap() is more general.
	 */

	pgcnt = btop(uap->len);
	firstpg = btop(uap->addr);
	lastpg = firstpg + pgcnt - 1;

	if (uap->len <= 0 || (((int)uap->addr | uap->len) & CLOFSET)
	||  !isadsv(p, firstpg) || !isadsv(p, lastpg)) {
		u.u_error = EINVAL;
		return;
	}

	/* Make sure there is no pending asynchronous I/O for this process */
	if (p->p_aio) {
		aio_pend_wait(p->p_aio, p, 0);
	}

	/*
	 * No argument errors.
	 * Loose old address-space.
	 *
	 * Release of u_mmap[] entries is automatic in vmemfree(),
	 * when mm_pgcnt reaches zero. Freed pte entries get marked
	 * invalid.
	 */

	pte = dvtopte(p, firstpg);
	p->p_rssize -= vmemfree(pte, (int)pgcnt, 1);

	/*
	 * Flush MMU TLB, since process mapping changed a bunch.
	 */

	FLUSH_USER_TLB(p->p_ptb1);
}

/*
 * munmapmm()
 *	unmap this map from user's address space
 *
 * Called by core() to unmap any phys-mapped stuff. Assumes um->mm_pgcnt
 * is non-0.
 *
 * Note that unmapping pages doesn't turn off FD_MAPPED bit when freeing
 * u_mmap[] entry, relying on efficiency of this procedure.
 * Could fix if necessary (see vmmapfree()).
 */

munmapmm(um)
	struct mmap *um;
{
	register struct	proc *p = u.u_procp;
	register struct	pte *pte;
	register int	i;
	register int	cnt;
	int		mapx;

	/*
	 * Run thru the page-table and zap all references.
	 *
	 * Note: any private reclaimable pte's are !PTEMAPPED() and don't look
	 * mapped even if being concurrently reallocated.
	 */

	/* Make sure there is no pending asynchronous I/O for this process */
	if (p->p_aio) {
		aio_pend_wait(p->p_aio, p, 0);
	}

	pte = dvtopte(p, um->mm_1stpg);
	mapx = um - u.u_mmap;
	for (cnt = um->mm_size; cnt > 0; cnt -= CLSIZE) {
		if (!pte->pg_fod
		&&  PTEMAPPED(*pte)
		&&  PTETOMAPX(*pte) == mapx) {
			p->p_rssize -= vmmapfree(p, pte, 1);
			for (i = 0; i < CLSIZE; i++, pte++)
				*(int*)pte = PG_INVAL;
		} else
			pte += CLSIZE;
	}
	ASSERT(um->mm_pgcnt == 0, "munmapmm: skew");
	/*
	 *+ A mapped region had a nonzero page count after all
	 *+ of the pages had been freed.
	 */

	/*
	 * No need to zap ofile flags: will get zapped on next open.
	 * Flush MMU TLB, since process mapping may have changed a bunch.
	 */

	FLUSH_USER_TLB(p->p_ptb1);
}
