/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ident	"$Header: kern_xxx.c 1.29 90/04/20 $"

/*
 * kern_xxx.c
 *	Misc admin system calls
 *
 *	This is essentially utssys.c and sys3b.c/sysi86.c.
 */

/* $Log:	kern_xxx.c,v $
 */

#include "sys/types.h"
#include "sys/devsw.h"
#include "sys/param.h"
#include "sys/debug.h"
#include "sys/signal.h"
#include "sys/systm.h"
#include "sys/mutex.h"
#include "sys/user.h"
#include "sys/errno.h"
#include "sys/timer.h"
#include "sys/vfs.h"
#include "sys/proc.h"
#include "sys/statfs.h"
#include "sys/uadmin.h"
#include "sys/ustat.h"
#include "sys/utsname.h"
#include "sys/conf.h"
#include "sys/var.h"
#include "sys/boot.h"
#include "sys/sysi86.h"
#include "sys/swap.h"

#include "sys/engine.h"

#include "sys/intctl.h"

extern time_t c_correct;

sema_t uadmin_sema;			/* Need to serialize uadmin() */

utssys()
{
	register struct a {
		char	*cbuf;
		int	mv;
		int	type;
	} *uap;
	struct	ustat	ustat_buf;

	uap = (struct a *)u.u_ap;
	switch (uap->type) {
	case 0:		/* uname */
		if (copyout((char *)&utsname, uap->cbuf,
		    sizeof(struct utsname)))
			u.u_error = EFAULT;
		return;

	/* case 1 was umask */

	case 2:		/* ustat */
		u.u_error = vfs_ustat((dev_t)uap->mv, &ustat_buf);
		if (u.u_error == 0)
			u.u_error = copyout((caddr_t)&ustat_buf,
					uap->cbuf, sizeof(ustat_buf));
		return;

	default:
		u.u_error = EFAULT;
	}
}

/*
 * administrivia system call
 */

uadmin()
{
	register struct a {
		int	cmd;
		int	fcn;
		int	mdep;
	} *uap = (struct a *)u.u_ap;
	register struct proc *p;
	spl_t	s;
	int mode = RB_AUTOBOOT;
	extern int (*rootfsunmount)();	/* pointer to root unmounting routine */

	if (!cp_sema(&uadmin_sema))
		return;

	if (!suser()) {
		v_sema(&uadmin_sema);
		return;
	}

	switch (uap->cmd) {
	case A_SHUTDOWN:
	case A_REBOOT:
		if (nonline != 1) {
			u.u_error = EINVAL;
			break;
		}

		if (uap->cmd == A_SHUTDOWN) {
			if (uap->fcn != AD_BOOT_NOKILL) {
				s = p_lock(&proc_list, SPL_PROC_LIST);
				for (p = &proc[3]; p < v.ve_proc; p++) {
					if (p->p_stat == NULL)
						continue;
					if ((p != u.u_procp) &&
					   !(p->p_flag & SSYS))
						psignal(p, SIGKILL);
				}
				v_lock(&proc_list, s);
	
				/*
				 * let the processes die
				 */
				(void)p_sema(&lbolt, PZERO);
			}

			xumount(rootvfs);
			/*
			 * no need to sync since boot() does it.
			 * rootfsunmount doesn't really flush the
			 * various caches, etc. It marks the superblock
			 * as having been cleanly unmounted and writes
			 * it out so fsck won't complain when the system
			 * is rebooted.
			 */
			(*rootfsunmount)();
		}
		/* translate AD_* commands to RB_* */
		if (uap->fcn == AD_HALT)
			mode |= RB_HALT;
		FW_BOOT(RB_BOOT, mode|uap->mdep);
		/* no return */

	case A_REMOUNT:
		/*
		 * SysV code is incredibly buggy! There isn't an easy
		 * solution here so for now we just punt.
		 */
		u.u_error = EBUSY;
		break;

	case A_CLOCK:
		/*
		 * Correction (in seconds) from local time to gmt is
		 * passed.  This is used by wtodc() when called later
		 * from stime() so that rtc is in local time.  Nothing
		 * else is done right now in order to avoid needing
		 * code in kernel to convert rtc format to internal
		 * format.  That can be done at user level during startup.
		 */
		c_correct = uap->fcn;
		break;

	default:
		u.u_error = EINVAL;
	}
	v_sema(&uadmin_sema);
}

/*
 * Machine specific calls
 */
sysi86()
{
	struct a {
		int	cmd;
		int	arg1;
		int	arg2;
		int	arg3;
	} *uap = (struct a *) u.u_ap;

	switch (uap->cmd) {
	case SI86SWAP:
	{
		swpi_t	swpbuf;

		swpbuf.si_cmd = SI_ADD;
		swpbuf.si_buf = (char *)uap->arg1;
		swpbuf.si_swplo = 0;
		swpbuf.si_nblks = 0;
		swapfunc(&swpbuf);
		break;
	}
	case SI86SWPI:
	{
		swpi_t	swpbuf;

		if (copyin((caddr_t)uap->arg1, (caddr_t)&swpbuf, sizeof(swpbuf))){
			u.u_error = EFAULT;
			return;
		}
		swapfunc(&swpbuf);
		break;
	}
	case STIME:
	{
		timeval_t atv;

		atv.tv_sec = (time_t)uap->arg1;
		atv.tv_usec = (time_t)0;
		setthetime(&atv, 0);
		break;
	}
	case SETNAME:
	{
		int len;
		char sysnamex[sizeof(utsname.sysname)];

		if (!suser())
			break;
		
		u.u_error = copyinstr((caddr_t)uap->arg1, sysnamex,
				  sizeof(sysnamex), &len);
		if (u.u_error) {
			if (u.u_error == ENAMETOOLONG)
				u.u_error = EINVAL;
			break;
		}
		/* name is already null-terminated */
		bcopy(sysnamex, (char *)utsname.sysname, (unsigned)len);
		bcopy(sysnamex, (char *)utsname.nodename, (unsigned)len);
		break;
	}
	case SI86MEM:
	{
		extern uint totalmem;

		u.u_r.r_val1 = totalmem;
		break;
	}
	case SI86FPHW:
		if (suword((char *)uap->arg1, fp_type()))
			u.u_error = EFAULT;
		break;
	case SI86KWPT:
		u.u_error = kwatchpt(uap->arg1, uap->arg2 ,uap->arg3);
		break;
	default:
		u.u_error = EINVAL;
		break;
	}
}

swapfunc(si)
	swpi_t *si;
{
	switch (si->si_cmd) {
	case SI_LIST:
		if (copyout((char *)swdevt, (char *)si->si_buf,
		    sizeof(struct swdevt)*MSFILES))
			u.u_error = EFAULT;
		break;
	
	case SI_ADD:
		if (suser())
			swapon(si);
		break;

	case SI_DEL:
		u.u_error = EINVAL;
		break;
	}
}
