/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

/*	Copyright (c) 1984 AT&T	*/
/*	  All Rights Reserved  	*/

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/

#ident	"$Header: str_mm.c 1.19 89/08/23 $"

/* $Log:	str_mm.c,v $
 */


/*
 * STREAMS MESSAGE MGT ROUTINES
 *                                                                        
 * This file contains routines which manipulate streams messages.
 * The bufcall/unbufcall routines are in "stream.c".
 */


#include "sys/types.h"
#include "sys/param.h"
#include "sys/mutex.h"
#include "sys/intctl.h"
#include "sys/sysmacros.h"
#include "sys/stropts.h"
#include "sys/session.h"
#include "sys/stream.h"
#include "sys/strstat.h"
#include "sys/conf.h"
#include "sys/var.h"
#include "sys/debug.h"
#include "sys/cmn_err.h"


extern	struct	strstat strst;
extern	mblk_t	*mbfreelist;
extern	dblk_t	*dbfreelist[];
extern	struct	dbalcst dballoc[];
extern	char	strbcflag;
extern	lock_t	bc_mutex;
extern	lock_t	mb_mutex;
extern	lock_t	mm_mutex[];

#ifdef DEBUG
int failallocb;			/* test hook variable */
int failallocb_skip;		/* test hook variable */
#endif /* DEBUG */



/*
 * Allocate a message and data block.  Tries to get a block large
 * enough to hold 'size' bytes.  A message block and data block are
 * allocated together, initialized, and a pointer to the message block
 * returned.  Data blocks are always allocated in association with a
 * message block, but there may be several message blocks per data
 * block (see dupb).  If no message blocks or data blocks (of the
 * required size) are available, NULL is returned.
 *
 * Message Mgt assumed unlocked.
 */

mblk_t *
allocb(size, pri)
#ifdef DEBUG
	int size;
#else
	register size;
#endif /* DEBUG */
	uint pri;
{
	register dblk_t *databp;
	register mblk_t *bp;
	register spl_t s_ipl;
	register class;
#ifdef DEBUG
	int *p;

	/* Hooks which allow alloc failure testing */
	if (failallocb) {
		if (!failallocb_skip) {
			return(NULL);
		}
		failallocb_skip = 0;
	}
#endif /* DEBUG */


	if ((class = getclass(size)) >= NCLASS) {
		cmn_err(CE_NOTE, "allocb: size requested too large\n");
		/*
		 *+ A streams buffer could not be allocated because the
		 *+ size specified is greater than the largest configured
		 *+ buffer class.
		 */
		return(NULL);
	}


	/*
	 * Get buffer - if can't in class then try one class higher.
	 */

	s_ipl = p_lock(&mm_mutex[class], SPLMM);
	if ((dballoc[class].dba_cnt < bcmax(class, pri)) && 
	    (databp = dbfreelist[class])) {
		goto gotdp;
	}
	v_lock(&mm_mutex[class], s_ipl);

	strst.dblk[class].fail++;
	strst.dblock.fail++;
	if (class++ == NCLASS-1) {
		return(NULL);
	}

	s_ipl = p_lock(&mm_mutex[class], SPLMM);
	if ((dballoc[class].dba_cnt < bcmax(class, pri)) && 
	    (databp = dbfreelist[class])) {
		goto gotdp;
	}
	v_lock(&mm_mutex[class], s_ipl);

	strst.dblk[class].fail++;
	strst.dblock.fail++;
	return(NULL);

gotdp:
	/*
	 * The data buffer is in hand, try for a message block.
	 * The mm_mutex lock is already held.
	 */

	(void) p_lock(&mb_mutex, SPLMM);
	if (!(bp = mbfreelist)) {
		v_lock(&mb_mutex, SPLMM);
		v_lock(&mm_mutex[class], s_ipl);
		strst.mblock.fail++;
		return(NULL);
	}
	mbfreelist = bp->b_next;
	BUMPUP(strst.mblock);
	BUMPUP(strst.dblock);
	v_lock(&mb_mutex, SPLMM);

	dbfreelist[class] = databp->db_freep;
	dballoc[class].dba_cnt++;
	BUMPUP(strst.dblk[class]);
	v_lock(&mm_mutex[class], s_ipl);

	/*
	 * Initialize message block and data block descriptors.
	 */

	bp->b_next = NULL;
	bp->b_prev = NULL;
	bp->b_cont = NULL;
	bp->b_datap = databp;
	bp->b_rptr = databp->db_base;
	bp->b_wptr = databp->db_base;
	databp->db_freep = NULL;
	databp->db_type = M_DATA;
	databp->db_func = NULL;
	databp->db_arg = 0;
	databp->db_ref = 1;

#ifdef DEBUG
	/* Colour the buffer with the caller's return address */
	p = (int *)&size;
	bp->b_datap->db_func = (int (*)())p[-1];
	bp->b_datap->db_arg = 1;
#endif /* DEBUG */

	return(bp);
}


/*
 * Similar to allocb but with a private buffer.  A message block and
 * data block are allocated together, initialized (to point to the
 * private buffer specified via the base and size parameters), and a
 * pointer to the message block returned.  Data blocks are always
 * allocated in association with a message block, but there may be
 * several message blocks per data block (see dupb).  If no message
 * blocks or data blocks (of the required size) are available, NULL is
 * returned.
 *
 * At the time when freeb() is invoked on the dblock corresponding
 * to the private buffer, func(arg) will be invoked to allow the owner
 * of the private buffer a chance to do housekeeping (such as freeing
 * the private buffer) before the structure is dismantled.  Note that
 * this dblock may be part of a chain of many dblocks in a message,
 * this function is invoked during the appropriate freeb() and not
 * during freemsg() of the enveloping message.  func(arg) will be
 * invoked without any locks and at SPLSTR.
 *
 * Message Mgt assumed unlocked.
 */

mblk_t *
allocb_priv(base, size, pri, func, arg)
	unsigned char *base;	/* address of private buffer */
	int size;		/* size of private buffer */
	uint pri;		/* priority of allocation */
	int (*func)();		/* function to call at freeb() time */
	long arg;		/* argument to func() */
{
	register dblk_t *databp;
	register mblk_t *bp;
	register spl_t s_ipl;

#ifdef DEBUG
	/* Hooks which allow alloc failure testing */
	if (failallocb) {
		if (!failallocb_skip) {
			return(NULL);
		}
		failallocb_skip = 0;
	}
#endif /* DEBUG */


	/*
	 * Get a free data block descriptor.
	 */

	s_ipl = p_lock(&mm_mutex[NCLASS], SPLMM);
	if ((dballoc[NCLASS].dba_cnt < bcmax(NCLASS, pri)) && 
	    (databp = dbfreelist[NCLASS])) {
		goto gotdp;
	}
	v_lock(&mm_mutex[NCLASS], s_ipl);

	strst.dblk[NCLASS].fail++;
	strst.dblock.fail++;
	return(NULL);

gotdp:
	/*
	 * The data buffer is in hand, try for a message block.
	 * The mm_mutex lock is already held.
	 */

	(void) p_lock(&mb_mutex, SPLMM);
	if (!(bp = mbfreelist)) {
		v_lock(&mb_mutex, SPLMM);
		v_lock(&mm_mutex[NCLASS], s_ipl);
		strst.mblock.fail++;
		return(NULL);
	}
	mbfreelist = bp->b_next;
	BUMPUP(strst.mblock);
	BUMPUP(strst.dblock);
	v_lock(&mb_mutex, SPLMM);

	dbfreelist[NCLASS] = databp->db_freep;
	dballoc[NCLASS].dba_cnt++;
	BUMPUP(strst.dblk[NCLASS]);
	v_lock(&mm_mutex[NCLASS], s_ipl);


	/*
	 * Initialize message block and data block descriptors.
	 */

	bp->b_next = NULL;
	bp->b_prev = NULL;
	bp->b_cont = NULL;
	bp->b_datap = databp;
	bp->b_rptr = base;
	bp->b_wptr = base;
	databp->db_freep = NULL;
	databp->db_class = NCLASS;
	databp->db_base = base;
	databp->db_lim = (unsigned char *)((int)base + size);
	databp->db_type = M_DATA;
	databp->db_func = func;
	databp->db_arg = arg;
	databp->db_ref = 1;
	return(bp);
}


/*
 * Test if block of given size can be allocated with a request of
 * the given priority.  This routine can probably be safely changed
 * to not lock anything since the returned value is not a reliable
 * indication anyway (the subsequent allocb on a MP may still fail
 * even if testb returns positive, for example).  For now, locks are
 * used while testing the freelists.
 *
 * To support private buffers, a size value of SIZE_PRIV is also 
 * allowed: its tests for availability of free mblock/dblocks in
 * the NCLASS slot.
 *
 * Message Mgt assumed unlocked.
 */

int
testb(size, pri)
	register size;
	uint pri;
{
	register class;
	register spl_t s_ipl;

	if (size == SIZE_PRIV) {
		class = NCLASS;
	} else {
		if ((class = getclass(size)) >= NCLASS) {
			return(0);
		}
	}

	s_ipl = p_lock(&mm_mutex[class], SPLMM);
	if ((dballoc[class].dba_cnt < bcmax(class, pri)) && 
	    dbfreelist[class]) {
		v_lock(&mm_mutex[class], s_ipl);
		return(1);
	}
	v_lock(&mm_mutex[class], s_ipl);

	if (class++ >= NCLASS-1) {
		return(0);
	}

	s_ipl = p_lock(&mm_mutex[class], SPLMM);
	if ((dballoc[class].dba_cnt < bcmax(class, pri) ) && 
	    dbfreelist[class]) {
		return(1);
	}
	v_lock(&mm_mutex[class], s_ipl);

	return(0);
}


/*
 * Get class of buffer.  Returns NCLASS if size is greater than the
 * the largest block class.
 *
 * No locking assumptions necessary.
 */

int
getclass(size)
	register size;
{
	register class;

	ASSERT_DEBUG(size >= 0, "streams: getclass: negative size");

	for (class = 0; (class < NCLASS) && (size > rbsize[class]) ; class++);
	return(class);
}




/*
 * Free a message block and decrement the reference count on its 
 * data block. If reference count == 0 also return the data block.
 * If the data block has a private buffer then invokes
 * (*db_func)(db_arg) at SPLSTR (without locks) before freeing it.
 *
 * Message Mgt assumed unlocked.
 */

freeb(bp)
	register mblk_t *bp;
{
	register class;
	register struct dbalcst *dbp;
	register spl_t s_ipl;
	register sched = 0;
	register dblock_freed = 0;

	ASSERT_DEBUG(bp && bp->b_datap, "streams: l_freeb: null bp or b_datap");
	ASSERT_DEBUG(bp->b_datap->db_ref != 0, "streams: l_freeb: invalid bp");

	class = bp->b_datap->db_class;
	dbp = &dballoc[class];

	s_ipl = p_lock(&mm_mutex[class], SPLMM);

	if (--bp->b_datap->db_ref == 0) {
		if (bp->b_datap->db_class == NCLASS) {
			/*
			 * Private buffer type.
			 * Only current use is in NFS.  Most cases
			 * of func() are "v_sema", which doesn't require a
			 * drop of mm_mutex.  Other cases can cause
			 * long latency and/or badly nest SPL's,
			 * so drop mm_mutex and re-acquire when done.
			 */
			if (bp->b_datap->db_func == (int (*)()) v_sema) {
				v_sema((sema_t *)(bp->b_datap->db_arg));
			} else {
				v_lock(&mm_mutex[NCLASS], SPLSTR);
				(*bp->b_datap->db_func)(bp->b_datap->db_arg);
				(void) p_lock(&mm_mutex[NCLASS], SPLMM);
			}
		}
		bp->b_datap->db_freep = dbfreelist[class];
		dbfreelist[class] = bp->b_datap;
		dbp->dba_cnt--;
		strst.dblk[class].use--;
		dblock_freed = 1;

		(void) p_lock(&bc_mutex, SPLMM);
		if (!strbcflag && ( dbp->dba_hip ||
		    (dbp->dba_medp && (dbp->dba_cnt < dbp->dba_med)) ||
		    (dbp->dba_lop && (dbp->dba_cnt < dbp->dba_lo)))) {
			sched = 1;
			strbcflag = 1;
		}			
		v_lock(&bc_mutex, SPLMM);
	}
	v_lock(&mm_mutex[class], s_ipl);

	bp->b_datap = NULL;

	s_ipl = p_lock(&mb_mutex, SPLMM);
	bp->b_next = mbfreelist;
	mbfreelist = bp;
	strst.mblock.use--;
	if (dblock_freed) {
		strst.dblock.use--;
	}
	v_lock(&mb_mutex, s_ipl);

	if (sched) {
		setqsched();
	}
}


/*
 * Free all message blocks in a message using freeb().  
 * The message may be NULL.
 *
 * Message Mgt assumed unlocked.
 */

freemsg(bp)
	register mblk_t *bp;
{
	register mblk_t *tp;

	while (bp) {
		tp = bp->b_cont;
		freeb(bp);
		bp = tp;
	}
}



/*
 * Duplicate a message block
 *
 * Allocate a message block and assign proper values to it (read and
 * write pointers) and link it to existing data block.  Increment
 * reference count of data block.
 *
 * Message Mgt assumed unlocked.
 */

mblk_t *
dupb(bp)
	register mblk_t *bp;
{
	register spl_t s_ipl;
	register mblk_t *nbp;
	register int class;

	ASSERT_DEBUG(bp, "streams: l_dupb: null bp");

	class = bp->b_datap->db_class;
	/*
	 * Grab mm_mutex[class] lock to avoid race with freeb() when updating
	 * bp->b_datap->db_ref.  Note that we grab this before mb_mutex to
	 * avoid a deadlock situation with allocb(), which aquires these locks
	 * in the same order.
	 */
	s_ipl = p_lock(&mm_mutex[class], SPLMM);
	(void) p_lock(&mb_mutex, SPLMM);

	if (!(nbp = mbfreelist)) {
		v_lock(&mb_mutex, SPLMM);
		v_lock(&mm_mutex[class], s_ipl);
		strst.mblock.fail++;
		return(NULL);
	}
	mbfreelist = nbp->b_next;
	(nbp->b_datap = bp->b_datap)->db_ref++;
	BUMPUP(strst.mblock);

	v_lock(&mb_mutex, SPLMM);
	v_lock(&mm_mutex[class], s_ipl);

	nbp->b_next = NULL;
	nbp->b_prev = NULL;
	nbp->b_cont = NULL;
	nbp->b_rptr = bp->b_rptr;
	nbp->b_wptr = bp->b_wptr;
	return(nbp);
}


/*
 * Duplicate a message block by block (uses dupb), returning a pointer
 * to the duplicate message.  Returns a non-NULL value only if the
 * entire message was dup'd.
 *
 * Message Mgt assumed unlocked.
 */

mblk_t *
dupmsg(bp)
	register mblk_t *bp;
{
	register mblk_t *head, *nbp;

	if (!bp || !(nbp = head = dupb(bp))) {
		return(NULL);
	}

	while (bp->b_cont) {
		if (!(nbp->b_cont = dupb(bp->b_cont))) {
			freemsg(head);
			return(NULL);
		}
		nbp = nbp->b_cont;
		bp = bp->b_cont;
	}
	return(head);
}



/*
 * Copies data from message block to newly allocated message block and
 * data block.  The copy is rounded out to full word boundaries so
 * that the (usually) more efficient word copy can be done.  Returns
 * new message block pointer, or NULL if error.
 *
 * Message Mgt assumed unlocked.
 */

mblk_t *
copyb(bp)
	register mblk_t *bp;
{
	register mblk_t *nbp;
	register dblk_t *dp, *ndp;

	ASSERT_DEBUG(bp, "streams: copyb: null bp");
	ASSERT_DEBUG(bp->b_wptr >= bp->b_rptr, "streams: copyb: invalid bp");

	dp = bp->b_datap;
	if (!(nbp = allocb(dp->db_lim - dp->db_base, BPRI_MED))) {
		return(NULL);
	}
	ndp = nbp->b_datap;	
	ndp->db_type = dp->db_type;
	nbp->b_rptr = ndp->db_base + (bp->b_rptr - dp->db_base);
	nbp->b_wptr = ndp->db_base + (bp->b_wptr - dp->db_base);
	bcopy((caddr_t)bp->b_rptr, (caddr_t)nbp->b_rptr, (uint)(nbp->b_wptr - nbp->b_rptr));
	return(nbp);
}


/*
 * Copies data from message to newly allocated message using new data
 * blocks.  Returns pointer to new message.  NULL if error.
 *
 * Message Mgt assumed unlocked.
 */

mblk_t *
copymsg(bp)
	register mblk_t *bp;
{
	register mblk_t *head, *nbp;

	if (!bp || !(nbp = head = copyb(bp))) {
		return(NULL);
	}

	while (bp->b_cont) {
		if (!(nbp->b_cont = copyb(bp->b_cont))) {
			freemsg(head);
			return(NULL);
		}
		nbp = nbp->b_cont;
		bp = bp->b_cont;
	}
	return(head);
}


/* 
 * Link a message block to tail of message.
 *
 * No locking assumptions necessary.
 */

linkb(mp, bp)
	register mblk_t *mp;
	register mblk_t *bp;
{
	ASSERT_DEBUG(mp && bp, "streams: linkb: null mp or bp");

	for (; mp->b_cont; mp = mp->b_cont);
	mp->b_cont = bp;
}


/*
 * Unlink a message block from head of message return pointer to new
 * message.  NULL if message becomes empty.
 *
 * No locking assumptions necessary.
 */

mblk_t *
unlinkb(bp)
	register mblk_t *bp;
{
	register mblk_t *bp1;

	ASSERT_DEBUG(bp, "streams: unlinkb: null bp");

	bp1 = bp->b_cont;
	bp->b_cont = NULL;
	return(bp1);
}


/* 
 * Remove a message block "bp" from message "mp".
 *
 * Return pointer to new message or NULL if no message remains.
 * Return -1 if bp is not found in message.
 *
 * No locking assumptions necessary.
 */

mblk_t *
rmvb(mp, bp)
	register mblk_t *mp;
	register mblk_t *bp;
{
	register mblk_t *tmp;
	register mblk_t *lastp = NULL;


	for (tmp = mp; tmp; tmp = tmp->b_cont) {
		if (tmp == bp) {
			if (lastp) {
				lastp->b_cont = tmp->b_cont;
			} else {
				mp = tmp->b_cont;
			}
			tmp->b_cont = NULL;
			return(mp);
		}
		lastp = tmp;
	}
	return((mblk_t *)-1);
}


/*
 * Return size of message of block type (bp->b_datap->db_type)
 *
 * No locking assumptions necessary.
 */
int
xmsgsize(bp)
	register mblk_t *bp;
{
	register unsigned char type;
	register count = 0;
	
	type = bp->b_datap->db_type;

	for (; bp; bp = bp->b_cont) {
		if (type != bp->b_datap->db_type) {
			break;
		}
		ASSERT_DEBUG(bp->b_wptr >= bp->b_rptr, "streams: xmsgsize: invalid bp");
		count += bp->b_wptr - bp->b_rptr;
	}
	return(count);
}


/*
 * Get number of data bytes in message.
 *
 * No locking assumptions necessary.
 */
int
msgdsize(bp)
	register mblk_t *bp;
{
	register int count = 0;

	for (; bp; bp = bp->b_cont)
		if (bp->b_datap->db_type == M_DATA) {
			ASSERT_DEBUG(bp->b_wptr >= bp->b_rptr, "streams: msgdsize: invalid bp");
			count += bp->b_wptr - bp->b_rptr;
		}
	return(count);
}


/*
 * Macro to check pointer alignment.
 * (true if alignment is sufficient for worst case)
 */

#define str_aligned(X)	(((uint)(X) & (sizeof(int) - 1)) == 0)

/*
 * Concatenate and align first len bytes of common message type.  Len
 * value of -1 means concat everything.  Returns 1 on success, 0 on
 * failure.  After the pullup, mp points to the pulled up data.  This is
 * convenient but messy to implement.
 *
 * Message Mgt assumed unlocked.
 */

int
pullupmsg(mp, len)
	mblk_t *mp;
	register len;
{
	register mblk_t *bp;
	register mblk_t *new_bp;
	register mblk_t *tmp;
	register n;
	register spl_t s_ipl;

	ASSERT_DEBUG(mp, "streams: pullupmsg: null mp");

	/*
	 * Quick checks for success or failure:
	 */
	if (len == -1) {
		if (mp->b_cont == NULL && str_aligned(mp->b_rptr)) {
			return(1);
		}
		len = xmsgsize(mp);
	} else {
		ASSERT_DEBUG(mp->b_wptr >= mp->b_rptr, "streams: pullupmsg: invalid mp");
		if (mp->b_wptr - mp->b_rptr >= len && str_aligned(mp->b_rptr)) {
			return(1);
		}
		if (xmsgsize(mp) < len) {
			return(0);
		}
	}

	/* 
	 * (len requested < xmsgsize) => so there are enough bytes in
	 * mp for the operation to succeed.
	 *
	 * Allocate a new mblk header.  It is used later to
	 * interchange mp and new_bp.
	 */

	s_ipl = p_lock(&mb_mutex, SPLMM);
	if ((tmp = mbfreelist) == NULL) {
		v_lock(&mb_mutex, s_ipl);
		return(0);
	}
	mbfreelist = tmp->b_next;
	v_lock(&mb_mutex, s_ipl);

	/*
	 * Allocate the new mblk.  We might be able to use the
	 * existing mblk, but we don't want to modify it in case its
	 * shared.  The new dblk takes on the type of the old dblk.
	 */
	if ((new_bp = allocb(len, BPRI_MED)) == NULL) {
		s_ipl = p_lock(&mb_mutex, SPLMM);
		tmp->b_next = mbfreelist;
		mbfreelist = tmp;
		v_lock(&mb_mutex, s_ipl);
		return(0);
	}
	new_bp->b_datap->db_type = mp->b_datap->db_type;

	/*
	 * Scan mblks and copy over data into the new mblk.  Two ways
	 * to fall out: exact count match: while (len && bp) bp points to
	 * the next mblk containing data or is null.  Inexact match:
	 * if (bp->b_rptr != ...)  In this case, bp points to an mblk
	 * that still has data in it.
	 */ 
	bp = mp;
	while (len && bp) {
		mblk_t *b_cont;

		ASSERT_DEBUG(bp->b_wptr >= bp->b_rptr, "streams: pullupmsg: invalid bp");
		n = imin((bp->b_wptr - bp->b_rptr), len);
		bcopy((caddr_t) bp->b_rptr, (caddr_t) new_bp->b_wptr, (uint) n);
		new_bp->b_wptr += n;
		bp->b_rptr += n;
		len -= n;
		if (bp->b_rptr != bp->b_wptr) {
			/* len bytes copied; bp's buffer still has data left */
			break;
		}
		b_cont = bp->b_cont;
		if (bp != mp) {	/* don't free the head mblk */
			freeb(bp);
		}
		bp = b_cont;
	}

	/*
	 * At this point: new_bp points to a dblk that contains the
	 * pulled up data.  The head mblk, mp, is preserved and may or
	 * may not have data in it.  The intermediate mblks are freed,
	 * and bp points to the last mblk that was pulled-up or is
	 * null.
	 *
	 * Now the tricky bit.  After this, mp points to the new dblk
	 * and tmp points to the old dblk.  New_bp points nowhere.
	 */

	*tmp = *mp;
	*mp = *new_bp;
	new_bp->b_datap = NULL;

	/*
	 * If the head mblk (now tmp) still has data in it, link it to
	 * mp otherwise link the remaining mblks to mp and free the
	 * old head mblk.
	 */

	if (tmp->b_rptr != tmp->b_wptr)
		mp->b_cont = tmp;
	else {
		mp->b_cont = bp;
		freeb(tmp);
	}

	/*
	 * Free new_bp
	 */
	s_ipl = p_lock(&mb_mutex, SPLMM);
	new_bp->b_next = mbfreelist;
	mbfreelist = new_bp;
	v_lock(&mb_mutex, s_ipl);
 
	return(1);
}

/*
 * Trim bytes from message:
 *  len > 0, trim from head
 *  len < 0, trim from tail
 * Returns 1 on success, 0 on failure.
 *
 * No locking assumptions necessary.
 */

int
adjmsg(mp, len)
	mblk_t *mp;
	register len;
{
	register mblk_t *bp;
	register n;
	int fromhead;

	ASSERT_DEBUG(mp, "streams: adjmsg: null mp");

	fromhead = 1;
	if (len < 0) {
		fromhead = 0;
		len = -len;
	}
	if (xmsgsize(mp) < len)
		return(0);

	if (fromhead) {
		bp = mp;
		while (len && bp) {
			ASSERT_DEBUG(bp->b_wptr >= bp->b_rptr, "streams: adjmsg: invalid bp");
			n = imin((bp->b_wptr - bp->b_rptr), len);
			bp->b_rptr += n;
			len -= n;
			bp = bp->b_cont;
		}
	} else {
		register mblk_t *save_bp;
		register unsigned char type;

		type = mp->b_datap->db_type;
		bp = mp;
		while (len) {
			bp = mp;
			save_bp = NULL;
			while (bp && bp->b_datap->db_type == type) {
				ASSERT_DEBUG(bp->b_wptr >= bp->b_rptr, "streams: adjmsg: invalid bp");
				if (bp->b_wptr - bp->b_rptr > 0) {
					save_bp = bp;
				}
				bp = bp->b_cont;
			}
			if (!save_bp) {
				break;
			}
			n = imin((save_bp->b_wptr - save_bp->b_rptr), len);
			save_bp->b_wptr -= n;
			len -= n;
		}
	}
	return(1);
}

#ifdef DEBUG

/*
 * Test Hook.
 * Fail allocb from now on.  If skip is true, then this takes
 * effect after the next allocb call, otherwise immediately.
 */

fail_allocb(skip)
	register skip;
{
	failallocb = 1;
	failallocb_skip = skip;
}


/*
 * Test Hook.
 * Undo the effect of fail_allocb.
 */

unfail_allocb()
{
	failallocb = 0;
}


/*
 * Test Hook.
 * Returns number of allocated buffers for the given size class.
 *
 * Assumes memory mgt is unlocked.
 */

int
allocbcnt(size)
	uint size;
{
	register class, count;
	register spl_t s_ipl;

	if (size == (uint)SIZE_PRIV) {
		class = NCLASS;
	} else {
		if ((class = getclass((int) size)) >= NCLASS) {
			return(0);
		}
	}

	s_ipl = p_lock(&mm_mutex[class], SPLMM);
	count = dballoc[class].dba_cnt;
	v_lock(&mm_mutex[class], s_ipl);
	return(count);
}

/*
 * Test Hook.
 * Returns number of free mblocks.
 *
 * Assumes memory mgt is unlocked.
 */

int
freemblkbcnt()
{
	register count;
	register spl_t s_ipl;
	register mblk_t *free;

	count = 0;
	s_ipl = p_lock(&mb_mutex, SPLMM);
	free = mbfreelist;
	while (free) {
		count++;
		free = free->b_next;
	}	
	v_lock(&mb_mutex, s_ipl);
	return(count);
}

/*
 * Dump block (for debug)
 */

dumpb(bp)
	register mblk_t *bp;
{
	int i, count;
	unchar *ptr;

	printf("Base to read pointer:\n");
	count = bp->b_rptr - bp->b_datap->db_base;
	for (ptr = bp->b_datap->db_base, i = 0; i < count; i++) {
		if ((i % 20) == 0)
			printf("\n");
		printf(" %x", *ptr++);
	}
	printf("\n");

	printf("read pointer to write pointer:\n");
	count = bp->b_wptr - bp->b_rptr;
	for (ptr = bp->b_rptr, i = 0; i < count; i++) {
		if ((i % 20) == 0)
			printf("\n");
		printf(" %x", *ptr++);
	}
	printf("\n");

	printf("write pointer to limit:\n");
	count = bp->b_datap->db_lim - bp->b_wptr;
	for (ptr = bp->b_wptr, i = 0; i < count; i++) {
		if ((i % 20) == 0)
			printf("\n");
		printf(" %x", *ptr++);
	}
	printf("\n");
}

/*
 * Dump message
 * Message Mgt assumed unlocked.
 */

dumpmsg(bp)
	register mblk_t *bp;
{
	printf("dumpmsg: begin\n");
	if (!bp) {
		return(NULL);
	}

	while (bp) {
		printf("next message: \n");
		dumpb(bp);
		bp = bp->b_cont;
	}
	printf("dumpmsg: end\n");
	return(1);
}
#endif /* DEBUG */
