/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ident	"$Header: subr_devsw.c 1.28 90/10/22 $"


/*
 * subr_devsw.c
 *	devsw data structure handling routines
 *
 */

/* $Log:	subr_devsw.c,v $
 */

#include "sys/types.h"
#include "sys/devsw.h"
#include "sys/param.h"
#include "sys/cmn_err.h"
#include "sys/intctl.h"
#include "sys/mc_mutex.h"
#include "sys/debug.h"
#include "sys/vtoc.h"
#include "sys/sysmacros.h"
#include "sys/signal.h"
#include "sys/user.h"
#include "sys/file.h"
#include "sys/vnode.h"
#include "sys/swap.h"
#include "sys/genalloc.h"
#include "sys/errno.h"
#include "sys/systm.h"

extern	short	upyet;
extern	int	nodev();
extern	int	cmpt_type;
extern	int	cmpt_swp_type;
extern	dev_t	cmpt_c_major;
extern	dev_t	vtoc_c_major;
extern	dev_t	rootdev;
extern	struct dev *rootdevp;
extern	struct swdevt swdevt[];
extern	int	dev_mem_init;
extern	int	dev_mem_click;

static	caddr_t	devall();
static struct dev *devsw_lookup();

/*
 * The devsw is a data structure which contains objects which describe
 * all of the directly-accessable devices on the system.  "Directly-
 * accessable" in this context means that the device has been configured
 * into the kernel and was "found", but have not been subsumed into
 * the structure of a virtual device driver.  This data structure
 * takes the place of the old cdevsw and bdevsw tables, which were
 * implicitly configurable at kernel compile time, but static during
 * kernel runtime.
 */

#define DEVHSZ	2048
#if     ((DEVHSZ&(DEVHSZ-1)) == 0)
#define DEVHASH(dev)        ((major(dev)+minor(dev))&(DEVHSZ-1))
#else
#define DEVHASH(dev)        (((unsigned)(major(dev)+minor(dev)))%DEVHSZ)
#endif

#ifndef lint
int	devsw_hsz = DEVHSZ;
#endif	/* lint */

union devsw devsw[DEVHSZ];

/*
 * memory management for devs and devops structs.
 */

struct	genallocator	devlist;
struct	genallocator	devopslist;

/*
 * Mutex.  Note that since these routines are called from the bottom
 * half of the kernel, the data structures must be protected by a lock.
 * Note that we skip the mutex operations if we are not up yet.  This
 * is because most driver probe routines run at high SPL, and the
 * p_lock would reduce it to a lower SPL.
 */

lock_t	devsw_lck;

#define	LOCK_DEVSW(s)	{ if (upyet) (s) = p_lock(&devsw_lck, SPLDEV); }
#define	UNLOCK_DEVSW(s)	{ if (upyet) v_lock(&devsw_lck, (s)); }

/*
 * devsw_init()
 *	Initialize the devsw hash table entries
 *
 * if this looks a lot like the way the inode table is initialized, there's
 * a reason.
 */

devsw_init()
{
	register union devsw *ds = devsw;
	register int i;

	init_lock(&devsw_lck);

	for (i = DEVHSZ; --i >= 0; ds++) {
		ds->ds_head[0] = ds;
		ds->ds_head[1] = ds;
	}

	/*
	 * initialize dev allocation mechanism
	 */

	if (geninit(&devlist, sizeof(struct dev), dev_mem_init, calloc, 
		dev_mem_click, devall) != 0) {

		cmn_err(CE_PANIC, "devsw_init: can't init devlist");
		/*
		 *+ The kernel encountered a problem when allocating memory for
		 *+ device objects.  More memory needs to be added
		 *+ to the system or a system-initialization
		 *+ routine is allocating too much memory.
		 */
	}

	/*
	 * initialize devops allocation mechanism
	 */

	if (geninit(&devopslist, sizeof(struct devops), dev_mem_init, calloc, 
		dev_mem_click, devall) != 0) {

		cmn_err(CE_PANIC, "devsw_init: can't init devopslist");
		/*
		 *+ The kernel encountered a problem when allocating memory for
		 *+ driver access routine structures.
		 *+ More memory needs to be added
		 *+ to the system or a system-initialization
		 *+ routine is allocating too much memory.
		 */
	}
}

/*
 * devall - allocate more memory for devs or devops structs
 *
 * Passed to the genalloc routines to be used to increase the allocation
 * for these data structures.
 */

static caddr_t
devall(nbytes)
	int	nbytes;
{
	return(wmemall(nbytes, 1));
}

/*
 * devsw_add - place the dev into the devsw data structure.
 *
 * The devsw hash table is searched under lock for the appropriate
 * slot to add the dev, and it is added.  Failure can occur if the
 * hash resolution is already filled (ie, trying to add a dev which has
 * already been added.)
 */

devsw_add(devp)
	struct dev *devp;
{
	register struct	dev *dp;
	register union	devsw *ds;
	spl_t	s;

	if (!devp) {
		return(-1);
	}

	ds = &devsw[DEVHASH(devp->d_dev)];

	LOCK_DEVSW(s);

	for (dp = ds->ds_chain[0]; dp != (struct dev *)ds; dp = dp->d_chain[0]) {
		if (dp == devp) {
			cmn_err(CE_WARN, "devsw_add: dev already in devsw");
			/*
			 *+ A device driver or other kernel routine
			 *+ is trying to add a dev to devsw even though
			 *+ the dev is already in devsw.
			 */
			UNLOCK_DEVSW(s);
			return(-1);
		}
		if(dp->d_dev == devp->d_dev) {
			cmn_err(CE_WARN, "devsw_add: duplicate dev_t");
			/*
			 *+ A device driver is attempting to add a dev to devsw;
			 *+ however, the major/minor number of the dev matches another entry
			 *+ in devsw.  This is serious; this situation can cause
			 *+ device accesses to go through the wrong driver.
			 */
			UNLOCK_DEVSW(s);
			return(-1);
		 }
	}
	devp->d_refcnt = 1;
	insque(devp, ds);
	UNLOCK_DEVSW(s);
	return(0);
}

/*
 * devsw_remove()
 *	Remove a dev from devsw.
 *
 * Lock the device table and call devsw_lremove.
 */

devsw_remove(devp)
	struct dev *devp;
{
	int	ret;
	spl_t	s;

	LOCK_DEVSW(s);
	ret = devsw_lremove(devp);
	UNLOCK_DEVSW(s);
	return(ret);
}

/*
 * devsw_lremove()
 *	Locked devsw_remove.
 *
 * The caller locks the devsw table at SPLDEV.  We only allow removal
 * of a dev if the reference count is == 1 (ie, caller is only active
 * reference), and the dev is really in the table.
 */

static
devsw_lremove(devp)
	struct dev *devp;
{
	struct dev *d1;

	if (devp == NULL || devp->d_refcnt != 1)
		return(-1);

	d1 = devsw_lookup(devp->d_dev);
	if (d1 != devp) {
		if (d1 == (struct dev *)NULL) {
			cmn_err(CE_WARN,
			   "devsw_lremove: tried to remove a dev not in devsw");
			/*
			 *+ A request was made to remove an entry from
			 *+ the devsw data structure even though devsw
			 *+ did not contain the dev.  This is due to
			 *+ a problem with a kernel routine that is
			 *+ erroneously assuming that the subject
			 *+ dev is in devsw.
			 */
		} else {
			cmn_err(CE_PANIC,
			   "devsw_lremove: dup dev");
			/*
			 *+ A request was made to remove an entry from
			 *+ the devsw data structure, but a search of
			 *+ devsw produced a different dev.  This indicates
			 *+ a serious inconsistency in the devsw data
			 *+ structure.
			 */
		}
	}
	remque(devp);
	return(0);
}

/*
 * dev_ref()
 *	Return a reference to a dev
 *
 * The devsw data structure is searched for a dev which matches the
 * major/minor number passed in as an argument.  Before the dev is
 * returned, its reference count is incremented.  Both lookup and
 * ref count incrementing is done under lock.
 */

struct dev *
dev_ref(devnum)
	dev_t	devnum;
{
	struct	dev *dp;
	spl_t	s;

	LOCK_DEVSW(s);

	dp = devsw_lookup(devnum);
	if (dp != (struct dev *)NULL) {
		ASSERT((int)(dp->d_refcnt) >= 0, "dev_ref: invalid refcnt");
		/*
		 *+ A reference was made to a dev whose reference
		 *+ count is less than zero.
		 */
		dp->d_refcnt++;
	}
	UNLOCK_DEVSW(s);
	return(dp);
}

/*
 * dev_deref()
 *	release a reference to a dev
 *
 * It is the responsibility of the caller to dispose of the pointer in
 * a responsible manner, since the reference count will reflect that
 * the reference is no longer held.  Care is taken to be sure that
 * the ref count is sane.
 */

dev_deref(devp)
	struct	dev *devp;
{
	spl_t	s;

	if (devp == NULL)
		return;

	LOCK_DEVSW(s);
	ASSERT((int)(devp->d_refcnt) > 0, "dev_deref: invalid refcnt");
	/*
	 *+ An attempt was made to release a reference to a dev, but
	 *+ the reference count was found to be less than or equal to
	 *+ zero, a condition that is inconsistent.
	 */
	devp->d_refcnt--;
	UNLOCK_DEVSW(s);
}

/*
 * devsw_lookup()
 *	Locked devsw search routine
 *
 * translates a major/minor number into a dev pointer.  Assumes that
 * caller has locked the devsw table.
 */

static struct dev *
devsw_lookup(devno)
	dev_t	devno;
{
	register struct	dev *dp;
	register union	devsw *ds;

	ds = &devsw[DEVHASH(devno)];

	for (dp = ds->ds_chain[0]; dp != (struct dev *)ds; dp = dp->d_chain[0]) {
		if (dp->d_dev == devno) {
			return(dp);
		}
	}
	return((struct dev *)NULL);
}


/*
 * translate an open file descriptor into a dev pointer.  Care
 * must be taken to call this in a user context, so the u
 * structure is sensible.  This hook is provided for AFS drivers
 * which are passed an fd in the V_BUILD ioctl.
 */

struct dev *
fdtodev(fdes)
	int fdes;
{
	register struct file *fp;
	register struct vnode *vp;

	/*
	 * expansion of GETF, done because GETF doesn't return 0
	 */
	GETFP(fp, fdes);
	if (fp == NULL) {
		return(0);
	}
	vp = fp->f_vnode;
	return(vp->v_rdevp);
}


/*
 * rootinit()
 *	Put partitioning in place on rootdev and swapdev
 *
 * Attempt to put a partitioning driver in place on the root device and
 * if needed, on the swap device.  Must be done before any IO is done
 * to either device.
 */

rootinit()
{
	struct	dev *dp, *dpp;
	dev_t	swp_dev = swdevt[0].sw_dev;

	if (cmpt_swp_type == -1)
		cmpt_swp_type = cmpt_type;

	/*
	 * Now establish the devs for rootdev and swapdev.  Since this
	 * is the first time they are looked at, a partitioning
	 * driver needs to be pushed on first.
	 */
	dp = dev_ref(rootdev | V_RAW_MINOR);
	if (dp == (struct dev *)NULL) {
		cmn_err(CE_WARN,
			"Root disk device is not available\n");
		/*
		 *+ The disk device that holds the root partition could not
		 *+ be found by the system.  This can mean that the
		 *+ disk was not present in the kernel's configuration
		 *+ file, or that it is configured, but not present on the system.
		 */
		cmn_err(CE_CONT, "Returning to Firmware\n");
		return_fw();
	}
	if (cmpt_type < 0) {
		dpp = dev_ref(makedev(vtoc_c_major, 0));
		if (dpp == (struct dev *)NULL) {
			cmn_err(CE_WARN, "vtoc driver has not been configured\n");
			/*
			 *+ The kernel needs to be reconfigured with
			 *+ the VTOC pseudodriver in the configuration.
			 */
			cmn_err(CE_CONT, "Returning to Firmware\n");
			return_fw();
		}
		(void)vtoc_vbuild(dpp, dp);
	} else {
		dpp = dev_ref(makedev(cmpt_c_major, 0));
		if (dpp == (struct dev *)NULL) {
			cmn_err(CE_WARN, "cmpt driver has not been configured\n");
			/*
			 *+ The kernel needs to be reconfigured with
			 *+ the cmpt pseudodriver in the configuration.
			 */
			cmn_err(CE_CONT, "Returning to Firmware\n");
			return_fw();
		}
		(void) cmpt_vbuild(dpp, dp, cmpt_type);
	}
	dev_deref(dpp);

	/*
	 * Note: a root may still be possible on a non-partitioned device.
	 * Thus, attempt a dev_ref on the rootdev whether or not the
	 * partitioning succeeded.
	 */
	dev_deref(dp);
	dp = dev_ref(rootdev);
	if (!dp) {
		cmn_err(CE_WARN,
			"No root device, or no VTOC present on root disk.\n");
		/*
		 *+ The root disk device could not be accessed by the
		 *+ system.  This is usually due to a missing VTOC on
		 *+ the root disk device.  Corrective action:  add a 
		 *+ VTOC to the root disk.
		 */
		cmn_err(CE_CONT,"Returning to Firmware\n");
		return_fw();
	}
	rootdevp = dp;

	/*
	 * if the rootdev and swp_dev are on different devices,
	 * we should push partitioning driver for the swap device
	 * as well.  Note, we don't check the result of the
	 * V_BUILD, since we may be trying to swap on the
	 * raw device.
	 */
	if (major(rootdev) != major(swp_dev) ||
	    minortounit(minor(rootdev)) != minortounit(minor(swp_dev))) {
		dp = dev_ref(swp_dev | V_RAW_MINOR);
		if (cmpt_type < 0) {
			dpp = dev_ref(makedev(vtoc_c_major, 0));
			(void)vtoc_vbuild(dpp, dp);
		} else {
			dpp = dev_ref(makedev(cmpt_c_major, 0));
			(void)cmpt_vbuild(dpp, dp, cmpt_swp_type);
		}
		dev_deref(dpp);
		dev_deref(dp);
	}
		
	dp = dev_ref(swp_dev);
	if (dp == (struct dev *)NULL) {
		cmn_err(CE_WARN,"Swap device has not been configured\n");
		/*
		 *+ The initial swap device (specified by
		 *+ the -s flag when booting the system) was not configured. 
		 *+ The system requires a swap device in order to boot.
		 */
		cmn_err(CE_CONT,"Returning to Firmware\n");
		return_fw();
	}
	swdevt[0].sw_devp = dp;

}
