/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ident	"$Header: subr_prf.c 1.24 90/01/17 $"

/*
 * subr_prf.c
 *	Various output subroutines.
 */

/* $Log:	subr_prf.c,v $
 */

#include "sys/signal.h"
#include "sys/types.h"
#include "sys/devsw.h"
#include "sys/param.h"
#include "sys/debug.h"
#include "sys/mutex.h"
#include "sys/systm.h"
#include "sys/buf.h"
#include "sys/conf.h"
#include "sys/boot.h"
#include "sys/vm.h"
#include "sys/user.h"
#include "sys/errno.h"
#include "sys/timer.h"
#include "sys/session.h"
#include "sys/proc.h"
#include "sys/ioctl.h"
#include "sys/cmn_err.h"
#include "sys/sysmacros.h"
#include "sys/uio.h"

#include "sys/engine.h"

#include "sys/pte.h"
#include "sys/intctl.h"
#include "sys/plocal.h"
#include "sys/vnode.h"

#define	TRUE	1
#define	FALSE	0

extern lock_t prf_lock;		/* Printf lock */
extern short upyet;

#ifdef DEBUG
extern	int	jc_debug;
#endif

/*
 * Scaled down version of C Library printf.
 * Used to print diagnostic information directly on console tty.
 * Since it is not interrupt driven, all system activities are
 * suspended.  Printf should not be used for chit-chat.
 *
 * One additional format: %b is supported to decode error registers.
 * Usage is:
 *	printf("reg=%b\n", regval, "<base><arg>*");
 * Where <base> is the output base expressed as a control character,
 * e.g. \10 gives octal; \20 gives hex.  Each arg is a sequence of
 * characters, the first of which gives the bit number to be inspected
 * (origin 1), and the next characters (up to a control character, i.e.
 * a character <= 32), give the name of the register.  Thus
 *	printf("reg=%b\n", 3, "\10\2BITTWO\1BITONE\n");
 * would produce output:
 *	reg=2<BITTWO,BITONE>
 */

#ifndef KLINT
/*VARARGS1*/
printf(fmt, x1)
	char *fmt;
	unsigned x1;
{
	spl_t	s;

	if (!upyet)
		prf(fmt, &x1, 0);
	else {
		s = p_lock(&prf_lock, SPLHI);
		prf(fmt, &x1, 0);
		v_lock(&prf_lock, s);
	}
}
#endif /*KLINT*/

/*
 * uprintf()
 *	Print on current users terminal.
 *
 * Guarantees not to sleep (so can be called by interrupt routines)
 * and does no watermark checking - (so no verbose messages).
 */

#ifndef KLINT
/*VARARGS1*/
uprintf(fmt, x1)
	char *fmt;
	unsigned x1;
{
	prf(fmt, &x1, 2);
}
#endif /*KLINT*/

prf(fmt, adx, touser)
	register char *fmt;
	register uint *adx;
{
	register int b, c, i;
	char *s;
	int any;

loop:
	while ((c = *fmt++) != '%') {
		if (c == '\0')
			return;
		putchar(c, touser);
	}
again:
	c = *fmt++;
	/* THIS CODE IS VAX DEPENDENT IN HANDLING %l? AND %c */
	switch (c) {

	case 'l':
		goto again;
	case 'x': case 'X':
		b = 16;
		goto number;
	case 'd': case 'D':
	case 'u':		/* what a joke */
		b = 10;
		goto number;
	case 'o': case 'O':
		b = 8;
number:
		printn((ulong)*adx, b, touser);
		break;
	case 'c':
		b = *adx;
		for (i = 24; i >= 0; i -= 8)
			if (c = (b >> i) & 0x7f)
				putchar(c, touser);
		break;
	case 'b':
		b = *adx++;
		s = (char *)*adx;
		printn((ulong)b, *s++, touser);
		any = 0;
		if (b) {
			putchar('<', touser);
			while (i = *s++) {
				if (b & (1 << (i-1))) {
					if (any)
						putchar(',', touser);
					any = 1;
					for (; (c = *s) > 32; s++)
						putchar(c, touser);
				} else
					for (; *s > 32; s++)
						;
			}
			if (any)
				putchar('>', touser);
		}
		break;

	case 's':
		s = (char *)*adx;
		while (c = *s++)
			putchar(c, touser);
		break;

	case '%':
		putchar('%', touser);
		break;
	}
	adx++;
	goto loop;
}

/*
 * Printn prints a number n in base b.
 * We don't use recursion to avoid deep kernel stacks.
 */

printn(n, b, touser)
	ulong n;
{
	char prbuf[11];
	register char *cp;

	if (b == 10 && (int)n < 0) {
		putchar('-', touser);
		n = (unsigned)(-(int)n);
	}
	cp = prbuf;
	do {
		*cp++ = "0123456789abcdef"[n%b];
		n /= b;
	} while (n);
	do
		putchar(*--cp, touser);
	while (cp > prbuf);
}

/*
 * Warn that a system table is full.
 */

tablefull(tab)
	char *tab;
{
	 cmn_err(CE_WARN, "%s: table is full\n", tab);
	/*
	 *+ A table has filled up (for example, a file table or
	 *+ inode table).  Corrective action:  reconfigure
	 *+ the kernel to increase the table size.
	 */
}

/*
 * Print a character on console or user's terminal.
 *	sy_c_major is defined in conf.c  It is major
 *	device number for the "/dev/tty" driver.
 *
 * WARNING:
 *	When printing to the user's terminal, this function must not be
 *	called above SPLSESS (see sycheck() and syrelease()).
 */

/*ARGSUSED*/
putchar(c, touser)
	int c;
{
	extern struct vnode *sycheck();
	proc_t *p;
	struct session *s;
	struct vnode *vp;
	int retval;
	struct uio auio;
	struct iovec aiov;
#ifdef DEBUG
	int	vopret;
#endif

	if (touser) {
		auio.uio_iov = &aiov;
		auio.uio_iovcnt = 1;
		aiov.iov_base = (caddr_t)&c;
		auio.uio_resid = aiov.iov_len = 1;
		auio.uio_segflg = UIOSEG_KERNEL;
		p = u.u_procp;
		s = p->p_session;
		if ((vp = sycheck(s, TRUE, &retval)) != NULL) {
	                /*
			 * WARNING:
			 *  We must use the returned pointer of 'vp' only.
			 *  The vnode (and any associated stream) originally
			 *  referenced by 's_vp' cannot be "put" until we have
			 *  released our 's_ctty_count' reference (created by
			 *  sycheck()), but 's_vp' could now be NULL (since our
			 *  session is not locked, and the last file descriptor
			 *  referring to 'vp' might be getting closed).
			 */
			/* The controlling tty had better be VCHR */
			ASSERT_DEBUG((vp->v_type==VCHR),
				     "putchar not to a character special");
#ifdef DEBUG
			if (jc_debug)
				printf("putchar (0x%x) ccty 0x%x touser\n",c,
					vp->v_rdev);
			vopret = VOP_RDWR(vp, &auio, UIO_WRITE, 0, u.u_cred);
			if (jc_debug)
				printf("putchar VOP failed 0x%x\n",vopret);
#else
			(void) VOP_RDWR(vp, &auio, UIO_WRITE, 0, u.u_cred);
#endif
			syrelease(s, FALSE);
			return;
		}
		/*
		 * else print to console.
		 */
	}
	if (c == 0)
		return;
	cnputc(c);
}

#ifdef DEBUG

/*
 * Assail prints its arguments and possible a backtrace.
 * it then panics the cpu.
 */

assfail( assstr, file, line, s )
	char	*assstr;
	char	*file;
	int	line;
	char	*s;
{
	/*
	 * back trace !!!!!
	 */
	cmn_err(CE_PANIC,"assertion failed: %s, file: %s, line: %d\n%s\n",
		assstr,file,line,s);
}

#endif /* DEBUG */
