/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ident	"$Header: vfs_io.c 1.25 89/08/07 $"

/*
 * vfs_io.c
 *	Virtual File-System system calls.
 */

/* $Log:	vfs_io.c,v $
 */

#include "sys/signal.h"
#include "sys/types.h"
#include "sys/devsw.h"
#include "sys/param.h"
#include "sys/debug.h"
#include "sys/user.h"
#include "sys/errno.h"
#include "sys/mutex.h"
#include "sys/timer.h"
#include "sys/proc.h"
#include "sys/uio.h"
#include "sys/vfs.h"
#include "sys/vnode.h"
#include "sys/file.h"
#include "sys/stat.h"
#include "sys/ioctl.h"
#include "sys/vmmeter.h"

#include "sys/vmparam.h"
#include "sys/pte.h"
#include "sys/intctl.h"
#include "sys/plocal.h"
#include "sys/cmn_err.h"

static	bool_t	vno_dolock[] = {
	/* VNON, VREG, VDIR, VBLK, VCHR, VLNK, VSOCK, VFIFO, VBAD */
	     0,    1,    1,    0,    0,    0,    0,     0,    0
};

vno_rw(fp, rw, uiop)
	register struct file *fp;
	enum	uio_rw	rw;
	struct	uio	*uiop;
{
	register struct vnode *vp;
	int	error;

	vp = fp->f_vnode;
	/*
	 * If write, make sure filesystem is writable.
	 */
	if ((rw == UIO_WRITE) && isrofile(vp))
		return (EROFS);
	/*
	 * If necessary, do IO with locked vnode.
	 *	duplicate the reading and update of f_offset to
	 *	so the accesses can be locked in one case.
	 *
	 *	The actual offset used for FAPPEND case is set
	 *	by VOP_RDWR from ip->i_size.  Value returned in
	 *	uio_offset is always correct.
	 */
	if (vno_dolock[(int)vp->v_type]) {
		VN_LOCKNODE(vp);
		uiop->uio_offset = fp->f_offset;
		error = VOP_RDWR(vp, uiop, rw,
				(((fp->f_flag & FAPPEND) ? IO_APPEND : 0) |
				((fp->f_flag & FSYNC) ? IO_SYNC : 0) |
				((fp->f_flag & FNDELAY) ? IO_NDELAY : 0)) ,
				fp->f_cred);
		fp->f_offset = uiop->uio_offset;
		VN_UNLOCKNODE(vp);
	} else {
		uiop->uio_offset = fp->f_offset;
		error = VOP_RDWR(vp, uiop, rw,
				(((fp->f_flag & FAPPEND) ? IO_APPEND : 0) |
				((fp->f_flag & FSYNC) ? IO_SYNC : 0) |
				((fp->f_flag & FNDELAY) ? IO_NDELAY : 0)) ,
				fp->f_cred);
		fp->f_offset = uiop->uio_offset;
	}
	return (error);
}

vno_ioctl(fp, com, data)
	struct	file	*fp;
	int	com;
	caddr_t	data;
{
	struct	vnode	*vp;
	int	error = 0;

	vp = fp->f_vnode;
	if (vp->v_type != VCHR) {
		error = ENOTTY;
	} else {
		u.u_r.r_val1 = 0;
		error = VOP_IOCTL(vp, com, data, fp->f_flag, fp->f_cred);
	}
	return (error);
}

/*
 * vno_stat
 * 	stat a vnode.
 *
 * Called with locked vnode (vp).
 */

int
vno_stat(vp, sb)
	register struct vnode *vp;
	register struct stat *sb;
{
	register int error;
	struct vattr vattr;

	error = VOP_GETATTR(vp, &vattr, u.u_cred);
	if (!error) {
		sb->st_mode = vattr.va_mode;
		sb->st_uid = vattr.va_uid;
		sb->st_gid = vattr.va_gid;
		sb->st_dev = vattr.va_fsid;
		sb->st_ino = vattr.va_nodeid;
		sb->st_nlink = vattr.va_nlink;
		sb->st_size = vattr.va_size;
		sb->st_atimeval = vattr.va_atime;
		sb->st_mtimeval = vattr.va_mtime;
		sb->st_ctimeval = vattr.va_ctime;
		sb->st_rdev = (dev_t)vattr.va_rdev;
	}
	return (error);
}

vno_close(fp)
	register struct file *fp;
{
	register struct vnode *vp;
	int	flag;

	/*
	 * No lock needed on fp here, because this is the last
	 * reference to it.  So, no one else should be touching
	 * it at this time.
	 */

	vp = fp->f_vnode;
	flag = fp->f_flag;
	ffree(fp);

	u.u_error = vn_close(vp, flag);
	VN_RELE(vp);
	return (u.u_error);
}

