/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ident	"$Header: clock.c 1.12 89/08/07 $"

/*
 * clock.c
 *	General clock routines.
 *
 * Included are the per processor real-time clock initialization and
 * a routine to set the systems internal software time stamp.
 */

/* $Log:	clock.c,v $
 */

#include "sys/types.h"
#include "sys/signal.h"
#include "sys/param.h"
#include "sys/debug.h"
#include "sys/systm.h"

#include "sys/cfg.h"
#include "sys/clock.h"
#include "sys/slic.h"
#include "sys/conf.h"

#include "sys/mc_vmparam.h"
#include "sys/hwparam.h"
#include "sys/intctl.h"
#include "sys/cmn_err.h"

/*
 * startrtclock()
 *	Start the real-time clock.
 *
 * Startrtclock restarts the real-time clock, which provides
 * hardclock interrupts to kern_clock.c.  On Sequent HW, this
 * is one-time only per processor (eg, no restart, clock reprimes
 * itself).
 *
 * Called by localinit() during selfinit().
 * This turns on the processor-local SLIC timer.
 *
 * For testing/performance measurement convenience, enable_local_clock
 * allows the per-processor clock to be left OFF.  Need to patch the
 * kernel binary or system memory to effect this.
 */

static	bool_t	enable_local_clock = 1;		/* default ON */

startrtclock()
{
	register struct cpuslic *sl = va_slic;

	if (!enable_local_clock)
		return;

	sl->sl_trv = ((sys_clock_rate * 1000000) / (SL_TIMERDIV * hz)) - 1;
	/* clear prescaler, load reload value */
	sl->sl_tcont = 0;
	sl->sl_tctl = SL_TIMERINT | LCLKBIN;	/* timer on in given bin */
}


/*
 * check_time checks and set the systems internal time.
 *
 * It sets the internal time to the time form the time of day hardware
 * if it is a reasonable time.  If the time from the time of day hardware
 * is unreasonable, it uses the time form the root filesystem.  If this
 * time is unreasonable it uses a default time. Check_time returns 0 if
 * the time of day hardware time is used, otherwise it returns 1.
 */

bool_t
check_time(todr, base)
	time_t todr;	/* Time from time of day hardware */
	time_t base;	/* Time from root filesystem */
{
	long deltat;

	if (todr < 13*SECYR) {
		cmn_err(CE_WARN, "TOD value bad -- Setting TOD to file system time");
		/*
		 *+ An invalid time of day value was returned from the
		 *+ console controller.  The system time of day is
		 *+ being loaded from the filesystem.
		 *+ Corrective action:  use the date command to set the
		 *+ time of day to a correct value.
		 */
		if (base < 13*SECYR) {
			cmn_err(CE_WARN, "preposterous time in file system -- Setting TOD to default time");
			/*
			 *+ The time value in the filesystem is wrong.
			 *+ The time of day is being set from a precalculated
			 *+ value.
			 *+ Corrective action:  use the date command to set the
			 *+ time of day to a correct value.
			 */
			time.tv_sec = 13*SECYR + 19*SECDAY + (2*SECDAY)/3;
		} else
			/*
			 * Believe the time in the file system for lack of
			 * anything better, resetting the TOD.
			 */
			time.tv_sec = base;

		cmn_err(CE_CONT, "\nCHECK AND RESET THE DATE!\n");
		return(1);
	} else {
		/*
		 * See if we gained/lost two or more days;
		 * if so, assume something is amiss.
		 */
		time.tv_sec = todr;
		deltat = time.tv_sec - base;
		if (deltat < 0)
			deltat = -deltat;
		if (deltat >= 2*SECDAY) {
			cmn_err(CE_WARN, "clock %s %d days",
			    time.tv_sec < base ? "lost" : "gained", deltat / SECDAY);
			/*
			 *+ The time of day clock has gained or lost more than
			 *+ two days.  Corrective action:  use the date command
			 *+ to reset the time to the correct value.
			 */
			cmn_err(CE_CONT, "\nCHECK AND RESET THE DATE!\n");
			return(1);
		}
	}
	return(0);
}
