/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ident	"$Header: swapgeneric.c 1.34 90/10/22 $"

/*
 * swapgeneric.c
 *	GENERIC swap/root configuration.
 */

/* $Log:	swapgeneric.c,v $
 */

#include "sys/types.h"
#include "sys/devsw.h"
#include "sys/vtoc.h"
#include "sys/cmn_err.h"
#include "sys/sysmacros.h"
#include "sys/signal.h"
#include "sys/param.h"
#include "sys/debug.h"
#include "sys/conf.h"
#include "sys/swap.h"
#include "sys/buf.h"
#include "sys/vm.h"
#include "sys/systm.h"
#include "sys/boot.h"

#include "sys/cfg.h"

#define	MINARGLEN	5

extern	dev_t rootdev;
extern	struct dev *rootdevp;
extern	struct	swdevt swdevt[];
#ifdef DEBUG
int	boot_debug;
#endif /* DEBUG */
extern  struct genericconf {
	char	*gc_name;
	dev_t	gc_root;
} genericconf[];


int need_r_opt=0;	/* flag indicating whether a -r option is necessary. */
			/* no need to mutex need_r_opt since we're running on*/
			/* just one proc during this phase of the boot. */

struct genericconf * setconf_dev();
extern int cmpt_type;
extern int cmpt_swp_type;


/*
 * a physical device spec is  {phys_spec} :== {dev} ( {unit}, {part} )
 * eg zd(0,6)
 *    meaning: the zdc drive, unit 0 partition 6
 *
 * a logical device spec is  {logical_spec} :== {dev}{unit}|{dev}{unit}{part}
 * eg zd0
 *    meaning: the first zdc drive
 *
 * a device spec is {dev_spec} :== {phys_spec} | {logical_spec}
 * Setconf() looks at the boot string and sets:
 *   rootdev = bootdev
 *   swdevt[0] = rootdev+1
 * The argumnents  [ -s {dev_spec} ] [ -r {dev_spec} ] are also proceesed.
 */
/*
 * also set the type number for the root disk.  -t {type_num}
 * and the type of the swap disk, -S {type_num}
 * this is a temporary while the VTOC driver is under development.
 */

setconf()
{
	register struct genericconf *gc;
	register char *p;
#ifdef IOPT
	extern char	*_initstr;
#endif
	dev_t	swp_dev;		/* initial swap device */
	int	M_opt = 0;		/* "-M" option, for booting miniroot */

	/*
	 * look for an opening paren.
	 * {dev}({unit},{part}){kern} [-r {rootdev} -s {swp_dev} ... ]
	 */
	gc = (struct genericconf *)NULL;
	for (p=CD_LOC->c_boot_name; *p != '\0'; p++) {
		if (*p == '(') {
			if (gc=setconf_dev(p-2)){
				rootdev = gc->gc_root;
				swp_dev = makedev(major(gc->gc_root),minor(gc->gc_root)+1);
			}
			break;
		}
	}
	/*
	 * now look for arguments.
	 */
	while (p < &CD_LOC->c_boot_name[BNAMESIZE]) {
		p++;
		if (*p == '-') {
			p++;
			switch (*p++) {
			case 's':
				/*
				 * swapdevice 
				 */
				while ((*p == ' ') || (*p == '\0'))
					p++;
				if (gc=setconf_dev(p)){
					swp_dev = gc->gc_root;
				}
				break;
			case 'r':
				/*
				 * root device
				 */
				while ((*p == ' ') || (*p == '\0'))
					p++;
				if (gc=setconf_dev(p)){
					rootdev = gc->gc_root;
					need_r_opt = 0;
				}
				break;

			case 'D':
#ifdef DEBUG
				boot_debug = 1;
#endif /* DEBUG */
				break;
#ifdef IOPT
			case 'i':
				while ((*p == ' ') || (*p == '\0'))
					p++;
				bcopy(p, _initstr, 20);
				break;
#endif
			case 't':
				cmpt_type = 0;
				while ((*p == ' ') || (*p == '\0'))
					p++;
				while (*p >= '0' && *p <= '9') {
					cmpt_type = (cmpt_type*10)+(*p++ - '0');
				} 
				break;
			case 'S':
				cmpt_swp_type = 0;
				while ((*p == ' ') || (*p == '\0'))
					p++;
				while (*p >= '0' && *p <= '9') {
					cmpt_swp_type =
						(cmpt_swp_type*10)+(*p++ - '0');
				} 
				break;
			case 'M':
				/*
				 * Allow root and swap on the same partition
				 * while running the miniroot.  Otherwise,
				 * panic if root and swap share a partition.
				 */
				M_opt = 1;
				break;
			default:
				cmn_err(CE_WARN, "unknown argument - %s\n",p);
				/*
				 *+ An unrecognized argument was
				 *+ passed into the kernel on the boot
				 *+ string.
				 */
				break;
			}
		}
	}
	if (gc == (struct genericconf *)NULL) {
		cmn_err(CE_NOTE, "Generic devices:");
		/*
		 *+ One of the devices specified in the boot string
		 *+ is not available as a bootable device.  
		 *+ A list of available devices follows this message.
		 */
		for (gc = genericconf; gc->gc_name; gc++)
			cmn_err(CE_CONT," %s", gc->gc_name);
		cmn_err(CE_CONT, ".\n");

		cmn_err(CE_CONT, "Returning to Firmware\n");
		return_fw();
	}

	if (M_opt == 0) {
		if (swp_dev == rootdev) {
			cmn_err(CE_NOTE, "ROOT and SWAP are on the SAME partition!");
			/*
		 *+ The system has been booted with root and swap on
		 *+ the same partition.  This is allowed only when
		 *+ the "-M" option has been specified (as when booting
		 *+ the miniroot during an installation).
			 */
			return_fw();
		}
	} else {
		if (swp_dev != rootdev) {
			cmn_err(CE_NOTE, "miniroot ROOT and SWAP must share partition!");
			/*
		 *+ The system is a miniroot (because the "-M" option was
		 *+ specified), but root and swap are on different
		 *+ partitions.
			 */
			return_fw();
		}
	}
	if (need_r_opt) {
	  	cmn_err(CE_NOTE, "The \"-r\" option is needed if booting from a disk unit other than 0");
		/*
		 *+ The standalone disk ID in the boot command specified a disk
		 *+ other than disk unit 0, but did not include the -r 
		 *+ option on the line.  If the -r option is not specified
		 *+ with the associated kernel disk device description 
		 *+ (for example, zd1s0), then the kernel cannot determine the minor 
		 *+ number associated with the root disk partition.  Panicking
		 *+ is better than writing and swapping on the wrong partition.
		 */
		return_fw();
	}
	swdevt[0].sw_dev = swp_dev;
	return;
}


#define	P_PER_D	256	/* each disk can contain 256 partitions */

/*
 * parse an option for a device specifier.
 */
struct genericconf *
setconf_dev(name)
	char	*name;
{
	register struct genericconf *gc;
	int	unit;
	int	part;
	char	*p;
	int	error;

	for (gc = genericconf; gc->gc_name; gc++) {
		/*
		 * Match the first 2 characters for a disk device.
		 */
		if (gc->gc_name[0] == name[0] && 
		    gc->gc_name[1] == name[1]) {
			p = &name[2];
			if (*p == '(') {
				/*
				 * xx(nnn,mmm) spec xx = disk type
				 *               nnn = unit
				 *               mmm = partition
				 */
				/*
				 * Determine unit number.
				 */
				error = 1;	/* assume an error */
				unit = 0;
				while (*(++p) != ',') {
					if (*p >= '0' && *p <= '9') {
						unit = (unit*10) + (*p - '0');
						error = 0;
					} else {
						cmn_err(CE_WARN, "bad/missing unit number\n");
						/*
					 *+ A device specified in the boot
					 *+ string had a bad or
					 *+ missing unit number.
						 */
						error = 1;
						break;
					}
				}
				if (error)
					break;
				/*
				 * Determine partition number.
				 */
				error = 1;	/* assume an error */
				part = 0;
				while (*(++p) != ')') {
					if (*p >= '0' && *p <= '9') {
						part = (part*10) + (*p - '0');
						error = 0;
					} else {
						cmn_err(CE_WARN, "bad/missing partition number\n");
						/*
					 *+ A device specified in the
					 *+ boot string had a bad or
					 *+ missing partition number.
						 */
						error = 1;
						break;
					}
				}
				if (error)
					break;
				/*
				 * Now find the minor.  Note that we cannot
				 * correctly determine the minor number
				 * if the unit number found is non-zero.
				 * This is due to the wierd standalone unit
				 * numbers used for SCSI disks (sd and wd) 
				 * where the first disk unit # is 0, the 
				 * second is 8, the third is 16, and so on.
				 * Since there are 256 minor numbers per unit
				 * we must multiply the unit number by 256
				 * and add this to the partition to get the
				 * minor number.  Rather than hard coding
				 * standalone device ID information into
				 * the kernel, we're going to assume here
				 * that the unit number is zero.  The
				 * correct way to boot when this is not
				 * the case is to use the "-r" and "-s"
				 * boot string options.  If the unit number
				 * found is non-zero, then we will insist
				 * that the "-r" option is present (naming
				 * the kernel's notion of the unit number).
				 */
				if (unit) {
					need_r_opt=1; /* panic if we don't...*/
						      /*    find a -r option */
				}
				/*
				 * Assume a zero unit number.
				 */
				gc->gc_root = makedev(major(gc->gc_root),part);
				return gc;
			}
			/*
			 * parse up to 3 decimal digits for unit
			 */
			if (*p >= '0' && *p <= '9') {
				unit = *p++ - '0';
				if (*p >= '0' && *p <= '9') {
					unit = (unit*10) + (*p++ - '0');
					if (*p >= '0' && *p <= '9') {
						unit = (unit*10) + (*p++ - '0');
					}
				}
			} else {
				cmn_err(CE_WARN, "bad/missing unit number %s\n",
							name);
				/*
				 *+ A device specified in the boot string
				 *+ had a bad or missing unit number.
				 */
				return (struct genericconf *)NULL;
			}
			if (*p == 's') {
				/*
				 * xxn[nn]sm[mm] spec xx = disk device
				 *                     n = unit
				 *                     m = partition
				 */
				p++;
				/*
				 * parse up to 3 decimal digits for partition
				 */
				if (*p >= '0' && *p <= '9') {
					part = *p++ - '0';
					if (*p >= '0' && *p <= '9') {
						part = (part*10) + (*p++ - '0');
						if (*p >= '0' && *p <= '9') {
							part = (part*10) + (*p++ - '0');
						}
					}
				}
				if (*p != '\0') {
					cmn_err(CE_WARN, "bad partition number %s\n",
						name);
					/*
					 *+ A device specified in the
					 *+ boot string had a bad or
					 *+ missing unit number.
					 */
					return (struct genericconf *)NULL;
				}
				gc->gc_root = makedev(major(gc->gc_root),
							(unit*P_PER_D)+part);
				return (gc);
			} else {
				/*
				 * xxna spec xx = disk device
				 *             n = unit
				 *             a = partition
				 */
				if (*p >= 'a' && *p <= 'h')
					part = *p-'a';
				else {
					cmn_err(CE_WARN, "bad/missing partition number %s\n",name);
					/*
					 *+ A device specified in the
					 *+ boot string had a bad or
					 *+ missing partition number.
					 */
					return (struct genericconf *)NULL;
				}
				gc->gc_root = makedev(major(gc->gc_root),
							(unit*P_PER_D)+part);
				return (gc);
			}
		}
	}
	return (struct genericconf *)NULL;
}
