/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#include "sys/types.h"
#include "sys/devsw.h"
#include "sys/param.h"
#include "sys/conf.h"
#include "sys/systm.h"
#include "sys/mutex.h"
#include "sys/signal.h"
#include "sys/user.h"
#include "sys/timer.h"
#include "sys/proc.h"
#include "sys/debug.h"
#include "sys/errno.h"
#include "sys/vfs.h"
#include "sys/vnode.h"
#include "sys/seg.h"
#include "sys/ipc.h"
#include "sys/sysmacros.h"
#include "sys/shm.h"
#include "sys/cmn_err.h"
#include "sys/buf.h"
#include "sys/uio.h"
#include "sys/dio.h"
#include "sys/cmap.h"
#include "sys/vm.h"
#include "sys/pte.h"
#include "sys/engine.h"
#include "sys/plocal.h"
#include "sys/intctl.h"
#include "sys/mftpr.h"


#ident	"$Header: shm.c 1.21 90/05/03 $"

/* $Log:	shm.c,v $
 */

extern struct shmid_ds	*shmem;		/* shared memory headers */
extern struct shminfo	shminfo;	/* shared memory info structure */
extern timeval_t	time;	/* system idea of date */
extern struct mapops	mmap_reg;
extern struct vnodeops	shm_vnodeops;

sema_t	shm_sema;
#define PSHM	PPIPE

struct ipc_perm *ipcget();
struct shmid_ds *shmconv();
int shmfree();


/*
 * Shmat (attach shared segment) system call.
 */
static
shmat(uap)
	struct a {
		int	shmid;
		uint	addr;
		int	flag;
	}	*uap;
{
	register struct shmid_ds *sp;
	register struct mmap *um;
	struct mmap *ump = (struct mmap *)0;
	register struct proc *p = u.u_procp;
	struct pte *pte;
	int newds, oldds;
	int shmaddr;
	int val, firstpg, pgcnt, n;
	int proto_pte;
	int handle;

	if (u.u_procp->p_flag & SVFORK) {
		u.u_error = EINVAL;
		return;
	}

	if ((sp = shmconv(uap->shmid, 0)) == NULL)
		return;
	if (ipcaccess(&sp->shm_perm, SHM_R)) {
		VN_UNLOCKNODE(MTOV(sp));
		return;
	}
	if ((uap->flag & SHM_RDONLY) == 0) {
		if (ipcaccess(&sp->shm_perm, SHM_W)) {
			VN_UNLOCKNODE(MTOV(sp));
			return;
		}
	}

	/*
	 * This is a simplified version of mmap()
	 */

	/*
	 * get next available slot and find possible attach address.
	 * if there are no mmap/shmem segments, attach at shm_gap bytes
	 * beyond end of data space (allow for possible sbrks to succeed).
	 */

	shmaddr = u.u_dsegsz;
	for (um = u.u_mmap; um < u.u_mmapmax; um++) {
		if (um->mm_pgcnt == 0) {
			if (ump == (struct mmap *)0)
				ump = um;
		} else if (um->mm_type == MMT_SHM || um->mm_type == MMT_MMAP) {
			if (shmaddr < um->mm_1stpg + um->mm_size)
				shmaddr = um->mm_1stpg + um->mm_size;
		}
	}
	if (ump == (struct mmap *)0) {
		if (u.u_mmapmax >= &u.u_mmap[NUMMAP]) {
			u.u_error = EMFILE;
			VN_UNLOCKNODE(MTOV(sp));
			return;
		}
		um->mm_pgcnt = 0;
		++u.u_mmapmax;
	} else
		um = ump;

	if (shmaddr == u.u_dsegsz)
		shmaddr += btoc(shm_gap);

	if (uap->addr == 0)
		firstpg = shmaddr;
	else if (uap->flag & SHM_RND)
		firstpg = btop(uap->addr & ~(SHMLBA - 1));
	else {
		if (uap->addr & CLOFSET) {
			u.u_error = EINVAL;
			VN_UNLOCKNODE(MTOV(sp));
			return;
		}
		firstpg = btop(uap->addr);
	}

	pgcnt = btoc(sp->shm_segsz);
	oldds = p->p_dsize;
	newds = vtodp(p, firstpg + pgcnt);

	if (u.u_error = chksize((uint)newds, (uint)u.u_ssize)) {
		VN_UNLOCKNODE(MTOV(sp));
		return;
	}

	/*
	 * check for overlap with other regions. Only need to check
	 * if attach address is within process's current dsize.
	 */
	if (vtodp(p, firstpg) < oldds) {
		pte = dvtopte(p, firstpg);
		n = imin(pgcnt, oldds - vtodp(p, firstpg));
		for (; n > 0; n -= CLSIZE, pte += CLSIZE) {
			if (*(int *)pte != PG_INVAL)
				break;
		}
		if (n) {
			u.u_error = EINVAL;
			VN_UNLOCKNODE(MTOV(sp));
			return;
		}
	}

	if (newds > p->p_dsize) {
		if (!expand(newds - oldds, 0, 0, PG_INVAL)) {
			u.u_error = ENOMEM;
			VN_UNLOCKNODE(MTOV(sp));
			return;
		}
		vexpandRS(newds - oldds);
	}

	val = (*mmap_reg.map_new)(MTOV(sp), 0, pgcnt, PROT_RDWR, &handle,
				shmfree, sp);
	if (val > 0) {
		(void) expand(oldds - newds, 0, 0, 0);
		u.u_error = val;
		VN_UNLOCKNODE(MTOV(sp));
		return;
	}
	ASSERT(sp->shm_handle == handle, "shmat: handle mismatch");
	/*
	 *+ The handle returned by the IFREG mapper on a shmat() system
	 *+ call didn't match the handle returned by the mapper when
	 *+ shmget() set up the map initially.
	 */

	++p->p_noswap;
	um->mm_off = 0;
	um->mm_1stpg = firstpg;
	um->mm_pgcnt = um->mm_size = pgcnt;
	um->mm_handle = handle;
	um->mm_ops = &mmap_reg;
	um->mm_vp = 0;
	um->mm_paged = 1;
	um->mm_noio = 0;
	um->mm_prot = (uap->flag & SHM_RDONLY) ? PROT_READ : PROT_RDWR;
	um->mm_cor = 0;
	um->mm_type = MMT_SHM;

	proto_pte = MAPXTOPTE(um-u.u_mmap)
			| ((um->mm_prot == PROT_RDWR) ? RW : RO)
			| PG_R;
	ptefill(dvtopte(p, firstpg), proto_pte, pgcnt);
	--p->p_noswap;

	u.u_r.r_val1 = (int) ptob(um->mm_1stpg);
	sp->shm_atime = time.tv_sec;
	sp->shm_lpid = p->p_pid;
	VN_UNLOCKNODE(MTOV(sp));

	FLUSH_USER_TLB(p->p_ptb1);
}

/*
 * Convert user supplied shmid into a ptr to the associated
 * shared memory header.
 *
 *	added flag argument, "lflag" to assist removing shmid's.
 *	It makes deadlock avoidance easier by leaving shm_sema
 *	locked on exit when we know we may have to delete the shmid.
 *	The only place this flag is non-zero is during the IPC_RMID
 *	case of shmctl().
 *
 *	If success, exit with the shmid vnode semaphore locked.
 */
static
struct shmid_ds *
shmconv(s, lflag)
register int	s;	/* shmid */
bool_t 	lflag;
{
	register struct shmid_ds	*sp;	/* ptr to associated header */

	if (s < 0)
	{
		u.u_error = EINVAL;
		return(NULL);
	}
	sp = &shmem[s % shminfo.shmmni];
	(void)p_sema(&shm_sema, PSHM);
	if (!(sp->shm_perm.mode & IPC_ALLOC)  
		|| s / shminfo.shmmni != sp->shm_perm.seq) {
		v_sema(&shm_sema);
		u.u_error = EINVAL;
		return(NULL);
	}
	VN_LOCKNODE(MTOV(sp));
	if (lflag == 0)
		v_sema(&shm_sema);
	return(sp);
}

/*
 * Shmctl system call.
 */
static
shmctl(uap)
	register struct a {
		int		shmid,
				cmd;
		struct shmid_ds	*arg;
	}	*uap;
{
	register struct shmid_ds *sp;	/* shared memory header ptr */
	struct shmid_ds ds;	/* hold area for IPC_SET */
	struct map_stat mstat;
	int pgcnt;

	if ((sp = shmconv(uap->shmid, uap->cmd == IPC_RMID?1:0)) == NULL)
		return;

	switch (uap->cmd) {

	/* Remove shared memory identifier. */
	case IPC_RMID:
		if (u.u_uid != sp->shm_perm.uid && u.u_uid != sp->shm_perm.cuid
			&& !suser()) {
			v_sema(&shm_sema);
			break;
		}
		/*
		 * If the segment size is 0 and the key is set to 
		 * IPC_PRIVATE then the identifier has already been
		 * removed (i.e., we have been through this code before,
		 * possibly as a result of the user performing two
		 * ipcrm's on the same identifier while a process still
		 * had the segment attached).
		 */
		if (sp->shm_segsz == 0 && sp->shm_perm.key == IPC_PRIVATE) {
			v_sema(&shm_sema);
			break;
		}
		pgcnt = btoc(sp->shm_segsz);
		/*
		 * We set the key to private so it "disappears" from the list
		 * of attachable shmid_ds structures but still can't be
		 * reallocated. The mode field in shm_perm gets cleared on
		 * the last unmap (in shmfree) which will make this
		 * shmid_ds structure available again.
		 */
		sp->shm_perm.key = IPC_PRIVATE;
		sp->shm_segsz = 0;
		if (((int)(++(sp->shm_perm.seq) * shminfo.shmmni + (sp - shmem))) < 0)
			sp->shm_perm.seq = 0;
		VN_UNLOCKNODE(MTOV(sp));
		v_sema(&shm_sema);
		/*
		 * Make sure segment is in core (got swapped out in shmget)
		 */
		while ((*mmap_reg.map_swpin)(sp->shm_handle, 0, pgcnt) != 0) {
			WAIT_MEM;
		}
		(*mmap_reg.map_unmap)(sp->shm_handle, 0, pgcnt,
				PROT_RDWR|PROT_LASTFD);
		return;

	/* Set ownership and permissions. */
	case IPC_SET:
		if (u.u_uid != sp->shm_perm.uid && u.u_uid != sp->shm_perm.cuid
			 && !suser()) 
			 break;

		if (copyin((char *)uap->arg, (char *)&ds.shmid_ic,
			   sizeof(ds.shmid_ic))) {
			u.u_error = EFAULT;
			return;
		}
		sp->shm_perm.uid = ds.shm_perm.uid;
		sp->shm_perm.gid = ds.shm_perm.gid;
		sp->shm_perm.mode = (ds.shm_perm.mode & 0777) |
			(sp->shm_perm.mode & ~0777);
		sp->shm_ctime = time.tv_sec;
		break;

	/* Get shared memory data structure. */
	case IPC_STAT:
		if (ipcaccess(&sp->shm_perm, SHM_R))
			break;

		/*	The following is needed because
		**	a user can look at it.  In
		**	particular, the regression tests
		**	require it.
		*/

		(void)(*mmap_reg.map_stat)(sp->shm_handle, &mstat);
		/*
		 * The number of shmats to this segment is actually one
		 * less than the refcount on the handle since we do an
		 * additional "new" operation in shmget.
		 */
		sp->shm_nattch = mstat.ms_count - 1;
		sp->shm_cnattch = sp->shm_nattch;

		if (copyout(&sp->shmid_ic, uap->arg, sizeof(sp->shmid_ic)))
			u.u_error = EFAULT;
		break;

	/* Lock segment in memory */
	/* LOCK and UNLOCK are not implemented */
	case SHM_LOCK:
		if (suser())
			u.u_error = EINVAL;
		break;

	/* Unlock segment */
	case SHM_UNLOCK:
		if (suser())
			u.u_error = EINVAL;
		break;

	default:
		u.u_error = EINVAL;
	}
	VN_UNLOCKNODE(MTOV(sp));
}

/*
 * Detach shared memory segment
 */
static
shmdt(uap)
	register struct a {
		caddr_t	addr;
	} *uap;
{
	register struct mmap *um;
	struct proc *p = u.u_procp;
	struct pte *pte;
	int firstpg, pgcnt;

	/*
	 * Find matching shmem address in process map list. Ensure
	 * that the addr passed in is on a page boundary.
	 */

	if ((int)uap->addr & CLOFSET) {
		u.u_error = EINVAL;
		return;
	}
	firstpg = btop(uap->addr);
	for (um = u.u_mmap; um < u.u_mmapmax; um++) {
		if (um->mm_pgcnt && (um->mm_type == MMT_SHM) &&
		    um->mm_1stpg == firstpg)
			break;
	}
	if (um >= u.u_mmapmax) {
		u.u_error = EINVAL;
		return;
	}
#ifdef notnow
	/*
	 * SysV used to search the entire shmid table for a match on the
	 * region being detached and set dtime/lpid only if the shmid
	 * pointed to the same region (it may have been deleted and
	 * reallocated to another region). This is not possible in our
	 * case since we still have a reference to the map. Deleted
	 * shmid structures get marked IPC_PRIVATE (see shmctl(RMID)
	 * above) so they essentially "disappear" until the last unmap.
	 * However, we don't have a handle to the shmid_ds structure
	 * at this point - we only have a pointer to the mfile structure.
	 * We ignore setting dtime/lpid for now. If it becomes important,
	 * should somehow get sp and update. A real kludge is to know
	 * that um->mm_handle is a pointer to a struct mfile which
	 * points to the shm vp which points to the shmid_ds struct.
	 * There are other, better methods.
	 */
#endif

	pgcnt = um->mm_size;
	pte = dvtopte(p, um->mm_1stpg);
	p->p_rssize -= vmemfree(pte, pgcnt, 1);

	FLUSH_USER_TLB(p->p_ptb1);
}

/*
 * Shmget (create new shmem) system call.
 */
static
shmget(uap)
	register struct a {
		key_t	key;
		uint	size,
			shmflg;
	}	*uap;
{
	register struct shmid_ds	*sp;	/* shared memory header ptr */
	int				s;	/* ipcget status */
	int				val;	/* return value from mapper */
	int				handle;	/* handle returned by mapper */

	if (u.u_procp->p_flag & SVFORK) {
		u.u_error = EINVAL;
		return;
	}

	(void)p_sema(&shm_sema, PSHM);
	if ((sp = (struct shmid_ds *)ipcget(uap->key, (int)uap->shmflg,
		  (struct ipc_perm *)shmem, shminfo.shmmni, sizeof(*sp),
		  &s)) == NULL) {
		v_sema(&shm_sema);
		return;
	}
	if (s) {

		/*
		 * This is a new shared memory segment.
		 * Allocate a region and init shmem table
		 */
		if (uap->size < shminfo.shmmin || uap->size > shminfo.shmmax) {
			u.u_error = EINVAL;
			sp->shm_perm.mode = 0;
			v_sema(&shm_sema);
			return;
		}
		/*
		 * Need to set size before map_new since shm_bmap uses it.
		 * This shouldn't be necessary when shm_filsys becomes a
		 * real swap_filsys.
		 */
		VN_LOCKNODE(MTOV(sp));
		sp->shm_segsz = uap->size;
		sp->shm_blocks = 0;
		MTOV(sp)->v_flag |= VNOLINKS;
		MTOV(sp)->v_devvp = swapdev_vp;
		val = (*mmap_reg.map_new)(MTOV(sp), 0, btoc(uap->size),
			PROT_RDWR, &handle, shmfree, sp);
		if (val > 0) {
			sp->shm_segsz = 0;
			sp->shm_perm.mode = 0;
			u.u_error = val;
			VN_UNLOCKNODE(MTOV(sp));
			v_sema(&shm_sema);
			return;
		}
		sp->shm_atime = sp->shm_dtime = 0;
		sp->shm_ctime = time.tv_sec;
		sp->shm_lpid = 0;
		sp->shm_cpid = u.u_procp->p_pid;
		sp->shm_handle = handle;
		VN_UNLOCKNODE(MTOV(sp));
		/*
		 * Force the segment to be swapped out. This ensures that
		 * when no processes are attached to it, the in-core count
		 * drops to 0. Otherwise the count will be at least 1 and
		 * it will never get swapped out.
		 */
		(*mmap_reg.map_swpout)(handle, 0, btoc(uap->size));
	} else {
		/*
		 * Found an existing segment.  Check size
		 */
		if (uap->size && uap->size > sp->shm_segsz) {
			u.u_error = EINVAL;
			v_sema(&shm_sema);
			return;
		}
	}

	u.u_r.r_val1 = sp->shm_perm.seq * shminfo.shmmni + (sp - shmem);
	v_sema(&shm_sema);
}

/*
 * Called by the mapper when removing the last reference to the
 * underlying object. The argument (sp) is passed to the map_new
 * function of the mapper which then uses it on unmapping. Since
 * it's a virtual address, this won't work if we start swapping
 * the shmid_ds array.
 *
 * Called with the map's vnode locked.
 *
 */
static
shmfree(sp)
	struct shmid_ds *sp;
{
	/*
	 * shm_perm.mode is really mutexed by shm_sema. We can't
	 * acquire it here since unmap can be called by exit(),
	 * shmdt() and shmctl(), and shmctl() calls it with shm_sema
	 * held. Since we're only assigning 0 to it, the worst
	 * that can happen is a simultaneous ipcget() will skip
	 * this entry as being unavailable. This shouldn't be a
	 * problem.
	 */
	sp->shm_perm.mode = 0;
}

static struct vfs shm_vfs;

shminit()
{
	register int i;
	register struct shmid_ds *sp;
	register struct vnode *vp;
	extern int dmmin;

	init_sema(&shm_sema, 1);
	shmem = (struct shmid_ds *)
		calloc(shminfo.shmmni * sizeof(struct shmid_ds));
	sp = shmem;
	for (i = 0; i < shminfo.shmmni  ; i++, sp++) {
		vp = MTOV(sp);
		init_lock(&vp->v_mutex);
		init_rwsema(&vp->v_nodemutex, RWSEMA_SWP);
		VN_INIT(vp, &shm_vfs, VBLK, swapdev);
		vp->v_count = 0;
		vp->v_data = (caddr_t)sp;
		vp->v_op = &shm_vnodeops;
		vp->v_mapx = VMAPX_NULL;
	}
	shm_vfs.vfs_bsize = dtob(dmmin);
}


/*
 * System entry point for shmat, shmctl, shmdt, and shmget system calls.
 */

int	(*shmcalls[])() = {shmat, shmctl, shmdt, shmget};

shmsys()
{
	register struct a {
		uint	id;
	}	*uap = (struct a *)u.u_ap;
	int		shmat(),
			shmctl(),
			shmdt(),
			shmget();

	if (uap->id > 3) {
		u.u_error = EINVAL;
		return;
	}

	/* Make sure there is no pending asynchronous I/O for this process */
	if (u.u_procp->p_aio) {
		aio_pend_wait(u.u_procp->p_aio, u.u_procp, 0);
	}

	(*shmcalls[uap->id])(&u.u_arg[1]);
}
