/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

/* SystemV-compatible FIFO implementation */

#ident "$Header: ffo_vnodeops.c 1.33 90/01/25 $"

/* $Log:	ffo_vnodeops.c,v $
 */
#include "sys/types.h"
#include "sys/devsw.h"
#include "sys/sysmacros.h"
#include "sys/signal.h"
#include "sys/param.h"
#include "sys/debug.h"
#include "sys/systm.h"
#include "sys/mutex.h"
#include "sys/user.h"
#include "sys/errno.h"
#include "sys/buf.h"
#include "sys/vfs.h"
#include "sys/vnode.h"
#include "sys/timer.h"
#include "sys/proc.h"
#include "sys/conf.h"
#include "sys/uio.h"
#include "sys/file.h"
#include "sys/fifo.h"
#include "sys/snode.h"
#include "sys/fifonode.h"
#include "sys/ufsinode.h"
#include "sys/statfs.h"
#include "sys/cmn_err.h"
#include "sys/vmmeter.h"
#include "sys/vm.h"		/* For plocal */
#include "sys/pte.h"		/* For plocal */
#include "sys/plocal.h"		/* For l struct (rablock, etc) */
#include "sys/intctl.h"		/* For SPLs */
#include "sys/poll.h"
#include "sys/stropts.h"

struct  fifonode *fifostuff;
struct fifonode *fifofree;
lock_t fifofree_list;

static int fifo_open();
static int fifo_close();
static int fifo_rdwr();
static int fifo_getattr();
static int fifo_setattr();
static int fifo_inactive();
static int fifo_invalop();
static int fifo_nostrmsgioop();
static int fifo_nostrpollop();
static int fifo_badop();
static int fifo_fsync();

extern int spec_access();
extern int spec_link();
extern int spec_fsync();
extern int spec_fdclose();
extern int spec_realvp();
extern int spec_fcntl();

struct vnodeops fifo_vnodeops = {
	fifo_open,	  /* vn_open     */
	fifo_close,	  /* vn_close    */
	fifo_rdwr,	  /* vn_rdwr     */
	fifo_badop,	  /* vn_ioctl    */
	fifo_getattr,	  /* vn_getattr  */
	fifo_setattr,	  /* vn_setattr  */
	spec_access,	  /* vn_access   */
	fifo_invalop,	  /* vn_lookup   */
	fifo_invalop,	  /* vn_creat    */
	fifo_invalop,	  /* vn_remove   */
	spec_link,	  /* vn_link     */
	fifo_invalop,	  /* vn_rename   */
	fifo_invalop,	  /* vn_mkdir    */
	fifo_invalop,	  /* vn_rmdir    */
	fifo_invalop,	  /* vn_readdir  */
	fifo_invalop,	  /* vn_symlink  */
	fifo_invalop,	  /* vn_readlink */
	fifo_fsync,	  /* vn_fsync    */
	fifo_inactive,	  /* vn_inactive */
	fifo_badop,	  /* vn_bmap     */
	fifo_badop,	  /* vn_strategy */
	fifo_badop,	  /* vn_bread    */
	fifo_badop,	  /* vn_brelse   */
	fifo_badop,	  /* vn_minphys  */
	fifo_badop,	  /* vn_getdents */
	spec_fdclose,	  /* vn_fdclose  */
	spec_fcntl,	  /* vn_fcntl    */
	spec_realvp,	  /* vn_realvp   */
	fifo_badop,	  /* vn_vhangup  */
	fifo_nostrmsgioop,/* vn_strmsgio */
	fifo_nostrpollop  /* vn_strpoll  */
};

static int pipe_statfs();

static
struct vfsops pipe_vfsops = {
	fifo_badop,	/* vfs_mount */
	fifo_badop,	/* vfs_unmount */
	fifo_badop,	/* vfs_root */
	pipe_statfs,	/* vfs_statfs */
	fifo_badop,	/* vfs_sync */
};

static struct vfs pipe_vfs;

/*
 * fifo_open()
 *	Open a fifo -- sleep until there are at least one reader & one writer.
 *
 * Caller passes *vpp unlocked.  It remains unlocked at exit.
 */

/*ARGSUSED*/
static int
fifo_open(vpp, flag, cred)
	struct vnode **vpp;
	int flag;
	struct ucred *cred;	/* unused */
{
	register struct fifonode *fp;
	spl_t	ipl;
	register int error=0;
	label_t saved_qsave;

	/*
	 * Setjmp in case open is interrupted.
	 * If it is, close and return error.
         * But first, let's save the current u_qsave on the stack
         * (to be restored before returning from fifo_open).
	 */
	saved_qsave = u.u_qsave;
	if (setjmp(&u.u_qsave)) {
		/*
	         * Restore the saved u_qsave area before returning.
	         */
		u.u_qsave = saved_qsave;

		(void) fifo_close(*vpp, flag & FMASK, cred);
		error = EINTR;
		return (error);
	}
	fp = VTOF(*vpp);

	ipl = FIFOLOCK(fp);
	if (flag & FREAD) {
		if (fp->fn_rcnt++ == 0)
			/* if any writers waiting, wake them up */
			vall_sema(&fp->fn_rwait);
	}

	if (flag & FWRITE) {
		if ((flag & FNDELAY) && (fp->fn_rcnt == 0)) {
			FIFOUNLOCK(fp, ipl);
			error = ENXIO;
			goto out;
		}
		if (fp->fn_wcnt++ == 0)
			/* if any readers waiting, wake them up */
			vall_sema(&fp->fn_wwait);
	}

	if (flag & FREAD) {
		while (fp->fn_wcnt == 0) {
			/* if no delay, or data in fifo, open is complete */
			if ((flag & FNDELAY) || fp->fn_size) {
				FIFOUNLOCK(fp, ipl);
				goto out;
			}
			FIFO_WWAIT(fp);
			ipl = FIFOLOCK(fp);
		}
	}

	if (flag & FWRITE) {
		while (fp->fn_rcnt == 0) {
			FIFO_RWAIT(fp);
			ipl = FIFOLOCK(fp);
		}
	}
	FIFOUNLOCK(fp, ipl);

out:
        /*
         * Restore the saved u_qsave area before returning.
         */
	u.u_qsave = saved_qsave;
	return (error);
}

/*
 * fifo_close()
 *	Close a fifo.
 *
 * Caller passes vp unlocked.  It remains unlocked at exit.
 *
 * On final close, all buffered data goes away
 */

/*ARGSUSED*/
static int
fifo_close(vp, flag, cred)
	struct vnode *vp;
	int flag;
	struct ucred *cred;	/* unused */
{
	register struct fifonode *fp;
	spl_t	ipl;

	fp = VTOF(vp);
	ipl = FIFOLOCK(fp);
	if ((flag & FREAD) && (--fp->fn_rcnt == 0))  {
		vall_sema(&fp->fn_wwait);
	}

	if ((flag & FWRITE) && (--fp->fn_wcnt == 0)) {
		vall_sema(&fp->fn_rwait);
	}

	if ((fp->fn_rcnt == 0) && (fp->fn_wcnt == 0)) {
		/* free buffer associated with this fifo */
		if (fp->fn_bp != NULL) {
			brelse(fp->fn_bp);
			fp->fn_bp = NULL;
			fp->fn_data = NULL;
		}

		/* update times only if there were bytes flushed from fifo */
		if (fp->fn_size != 0)
			FIFOMARK(fp, SUPD|SCHG);
		fp->fn_rptr = 0;
		fp->fn_wptr = 0;
		fp->fn_size = 0;
		fp->fn_flag = 0;
	}
	FIFOUNLOCK(fp, ipl);
	return (0);
}


/*
 * fifo_rdwr()
 *	Read or write a fifo vnode.
 *
 * Caller passes vp unlocked.
 */

/*ARGSUSED*/
static int
fifo_rdwr(vp, uiop, rw, ioflag, cred)
	struct vnode *vp;
	struct uio *uiop;
	enum uio_rw rw;
	int ioflag;
	struct ucred *cred;	/* unused */
{
	struct fifonode *fp;
	uint count;
	unsigned i;
	int rval = 0;
	int ocnt;
	int n;
	spl_t	ipl;

	ASSERT_DEBUG(uiop->uio_offset == 0, "fifo_rdwr: non-zero uio_offset");

	fp = VTOF(vp);

	if (rw == UIO_WRITE) {
		/*
		 * we know uio_resid is positive, due to
		 * check in write system call.
		 */
		ocnt = uiop->uio_resid;	/* save original request size */
		ipl = FIFOLOCK(fp);
		while (uiop->uio_resid && rval == 0) {
			count = uiop->uio_resid;
			if (fp->fn_flag & FN_WBUSY) {
				FIFO_WMUTEX(fp);
				(void)FIFOLOCK(fp);
				continue;
			}
			if (fp->fn_rcnt == 0) {
				/* no readers anymore! */
				psignal(u.u_procp, SIGPIPE);
				rval = EPIPE;
				break;
			}
			if ((count + fp->fn_size) > PIPE_BUF) {
				if (ioflag & IO_NDELAY) {	/* NO DELAY */
					if (count <= PIPE_BUF) {
						/*
						 * Write will be satisfied
						 * atomically, later.
						 */
						if (ocnt == uiop->uio_resid)
							rval = EAGAIN;
						break;
					} else if (fp->fn_size >= PIPE_BUF) {
					    /*
					     * Write will never be atomic.
					     * At this point, it cannot even be
					     * partial.   However, some portion
					     * of the write may already have
					     * succeeded.  If so, uio_resid
					     * reflects this.
					     */
						if (ocnt == uiop->uio_resid)
							rval = EAGAIN;
						break;
					}
				} else {
					/*
					 * Sleep until there is room for this
					 * request. On wakeup, go back to the
					 * top of the loop.  Before blocking,
					 * give the next waiting process
					 * a chance.
					 */
					if ( (count <= PIPE_BUF) ||
					    (fp->fn_size >= PIPE_BUF) ) {
						(void)cv_sema(&fp->fn_wmutex);
						FIFO_WWAIT(fp);
						(void)FIFOLOCK(fp);
						continue;
					}
				}
				/* at this point, can do a partial write */
				count = PIPE_BUF - fp->fn_size;
			}
			fp->fn_flag |= FN_WBUSY;
			FIFOUNLOCK(fp, ipl);
			if (fp->fn_bp == NULL) {
				fp->fn_bp = geteblk(PIPE_BUF);
				fp->fn_data = fp->fn_bp->b_un.b_addr;
			}
			n = 0;
			while (count) {
				i = PIPE_BUF - fp->fn_wptr;
				i = MIN(count, i);
				if (rval = uiomove(&fp->fn_data[fp->fn_wptr],
					(int) i, UIO_WRITE, uiop)){
					break;
				}
				fp->fn_wptr += i;
				n += i;
				count -= i;
				if (fp->fn_wptr >= PIPE_BUF) {
					fp->fn_wptr = 0;
				}
			}
			(void)FIFOLOCK(fp);
			fp->fn_size += n;
			FIFOMARK(fp, SUPD|SCHG);	/* update mod times */

			/* wake up any sleeping readers */
			vall_sema(&fp->fn_rwait);
			fp->fn_flag &= ~FN_WBUSY;
			continue;
		}
		(void)cv_sema(&fp->fn_wmutex);
		FIFOUNLOCK(fp, ipl);

	} else {
		/*
		 * Handle zero-length reads specially here
		 */
		if ((count = uiop->uio_resid) == 0) {
			goto rdwrdone;
		}
		ipl = FIFOLOCK(fp);
		while ((fp->fn_size == 0) || (fp->fn_flag & FN_RBUSY)) {
			/*
			 * if read side of fifo is busy,
			 *	wait for it to be not busy
			 *	Should this be signalable?
			 *	Should it interact differently with
			 *		FNDELAY?
			 */
			if (fp->fn_flag & FN_RBUSY) {
				FIFO_RMUTEX(fp);
				(void)FIFOLOCK(fp);
				continue;
			}
			/*
			 * no data in pipe and no writers...(EOF)
			 */
			if (fp->fn_wcnt == 0)
				goto rddone;
			/*
			 * No data in pipe, but writer is there;
			 * if No-Delay, return EAGAIN
			 */
			if (ioflag & IO_NDELAY) {
				rval = EAGAIN;
				goto rddone;
			}
			(void)cv_sema(&fp->fn_rmutex);
			FIFO_RWAIT(fp);
			(void)FIFOLOCK(fp);
			/* loop to make sure there is still a writer */
		}
		ASSERT_DEBUG(fp->fn_bp != NULL, "pipe UIO_READ: null bp");
		ASSERT_DEBUG(fp->fn_data != NULL, "pipe UIO_READ: null data");

		/* smaller of pipe size and read size */
		count = MIN(count, fp->fn_size);
		fp->fn_flag |= FN_RBUSY;
		FIFOUNLOCK(fp, ipl);
		n = 0;

		while (count) {
			i = PIPE_BUF - fp->fn_rptr;
			i = MIN(count, i);
			if (rval = uiomove(&fp->fn_data[fp->fn_rptr], (int)i,
						UIO_READ, uiop)) {
				break;
			}
			n += i;
			fp->fn_rptr += i;
			count -= i;
			if (fp->fn_rptr == PIPE_BUF) 
				fp->fn_rptr = 0;
		}
    
		(void)FIFOLOCK(fp);
		/*
		 * update fifo size based on bytes transferred.
		 * update access time.
		 * Mark the read side of the fifo as not busy, and
		 *	wake up a single reader if there is one.
		 * wake up writers.
		 */
		fp->fn_size -= n;
		FIFOMARK(fp, SACC);
		vall_sema(&fp->fn_wwait);
		fp->fn_flag &= ~FN_RBUSY;
rddone:
		(void)cv_sema(&fp->fn_rmutex);
		FIFOUNLOCK(fp, ipl);
	}

rdwrdone:
	uiop->uio_offset = 0;		/* guarantee that f_offset stays 0 */
	return (rval);
}

/*
 * fifo_getattr()
 *	Return attributes of fifo vnode.
 *
 * Caller passes vp locked.  It remains locked at exit.
 *
 * VOP_GETATTR requires realvp to be locked before call, and unlocked after.
 *	Note: there's no deadlock here
 */

static int
fifo_getattr(vp, vap, cred)
	struct vnode *vp;
	struct vattr *vap;
	struct ucred *cred;
{
	register int error;
	register struct snode *sp;

	sp = VTOS(vp);
	if (sp->s_realvp != NULL) {
		VN_LOCKNODE(sp->s_realvp);
		error = VOP_GETATTR(sp->s_realvp, vap, cred);
		VN_UNLOCKNODE(sp->s_realvp);
	} else {
		error = pipe_getattr(vp, vap);
	}
	if (!error) {
		/* set current times from snode, even if older than vnode */
		vap->va_atime = sp->s_atime;
		vap->va_mtime = sp->s_mtime;
		vap->va_ctime = sp->s_ctime;

		/* size should reflect the number of unread bytes in pipe */
		vap->va_size = (VTOF(vp))->fn_size;
		vap->va_blocksize = PIPE_BUF;
	}
	return (error);
}

/*
 * pipe_getattr()
 *	Code to fake out a getatter on a regular pipe
 *	(not a named one).  These pipes don't have any
 *	real file system backing them up.
 */

static int
pipe_getattr(vp, vap)
	struct vnode *vp;
	struct vattr *vap;
{

	/*
	 * Copy from fifonode 
	 */
	vap->va_type = IFTOVT(VTOF(vp)->fn_mode);
	vap->va_mode = VTOF(vp)->fn_mode;
	vap->va_uid = VTOF(vp)->fn_uid;
	vap->va_gid = VTOF(vp)->fn_gid;
	vap->va_fsid = VTOS(vp)->s_dev;
	vap->va_nodeid = VTOF(vp) - fifostuff;
	vap->va_nlink = 0;
	vap->va_flags = 0;
	vap->va_rdev = VTOS(vp)->s_dev;
	vap->va_blocks = PIPE_NBLKS;
	vap->va_blocksize = vp->v_vfsp->vfs_bsize;
	return (0);
}

/*
 * fifo_setattr()
 *	Set attributes of fifo vnode.
 *
 * Caller passes vp locked.  It remains locked at exit.
 *
 * Just like spec_setattr() except returns a different error if no s_realvp
 * (i.e., fifo is an unnamed pipe) for System V compatibility.
 */

static int 
fifo_setattr(vp, vap, cred)
	struct vnode *vp;
	struct vattr *vap;
	struct ucred *cred;
{
	if (VTOS(vp)->s_realvp == NULL)
		return(EINVAL);

	return(spec_setattr(vp, vap, cred));
}

/*
 * fifo_fsync()
 *	Synch-out a fifo vnode.
 *
 * Caller passes vp locked.  It remains locked at exit.
 *
 * Just like spec_fsync() except returns a different error if no s_realvp
 * (i.e., fifo is an unnamed pipe) for System V compatibility.
 */

static int
fifo_fsync(vp, cred)
	struct vnode *vp;
	struct ucred *cred;
{
	if (VTOS(vp)->s_realvp == NULL)
		return(EINVAL);

	return(spec_fsync(vp, cred));
}

/*
 * fifo_inactive()
 *	Vnode is no longer referenced.
 *
 * We are called with the vnode locked.
 * The snode itself MAY be locked, as indicated by the flag==NODEISLOCKED.
 *
 * There's a very careful design to avoid races.
 * See specvp() for more information.
 *
 * Note DYNIX doesn't pass a cred to VOP_INACTIVE.
 * Pass a zero cred to spec_fsync, since ufs_fsync doesn't use the cred.
 *
 * On entry, vnode may or may not be locked.
 * Spin lock in vnode is asserted.
 * sunsave() removes spec vnode from spec_list, zaps
 * fields in the snode to indicate it's deallocated, and
 * VN_RELE's the s_realvp.
 */

static int
fifo_inactive(vp, flag)
	struct vnode *vp;
	int flag;
{

	if ((flag == NODEISLOCKED) || VN_TRYLOCKNODE(vp)) {
		VN_UNLOCK(vp);
		/*
		 * The existence of a "shadow vnode" indicates
		 * this is a named pipe.  So, the shadow vnode
		 * must be "sync'd", and this fifonode
		 * must be removed from the snode hash list.
		 *
		 * In the event sunsave loses a race with a lookup
		 * we do NOT want to free this fifonode.  In this
		 * case, sunsave returns a "failure" indication and
		 * the fifonode is VN_UNLOCKNODE'd.
		 *
		 * Otherwise, free the fifonode. 
		 * If the snode is to be freed, sunsave returns the fifonode
		 * still locked and also VN_RELE's any shadow vnodes.
		 */
		if (VTOS(vp)->s_realvp != NULL) {
			(void)spec_fsync(vp, u.u_cred);
			if (sunsave(VTOS(vp))) {
				fifo_free(VTOF(vp));
			}
		} else {
			/*
			 * The fifonode is for a (unnamed) pipe.
			 * It is NOT on the snode hash list.
			 * So, there can be no races with
			 * lookups.
			 *
			 * It has NO shadow vnodes.  Just need
			 * to put it back into the free pool.
			 */
			fifo_free(VTOF(vp));
		}

	} else {
		VN_UNLOCK(vp);
	}
	return (0);
}

static int
fifo_invalop()
{
	return (EINVAL);
}

/*ARGSUSED*/
static int
fifo_nostrmsgioop(vp, mctl, mdata, rw, uflag, ioflag)
	struct vnode *vp;
	struct strbuf *mctl;
	struct strbuf *mdata;
	enum uio_rw rw;
	int uflag;
	int ioflag;
{
	return (ENOSTR);
}

/*ARGSUSED*/
static int
fifo_nostrpollop(vp, events, anyyet)
	struct vnode *vp;
	short events;
	int anyyet;
{
	return (POLLNVAL);
}

static int
fifo_badop()
{
	cmn_err(CE_PANIC, "fifo_badop\n");
	/*
	 *+ An improper operation was requested on a fifo vnode.
	 *+ This indicates a kernel software bug.
	 */
}

/*
 * fifosp()
 *	Construct a fifonode that can masquerade as an snode.
 */

int
fifosp(spp)
struct snode **spp;
{
	struct fifonode *fp;
	spl_t	ipl;

	ipl = p_lock(&fifofree_list, SPLFS);

	if ((fp = fifofree) != NULL) {
		fifofree = (struct fifonode *)fp->fn_snode.s_next;
		v_lock(&fifofree_list, ipl);
		*spp = FTOS(fp);
		return (0);
	} else {
		resource_ovf.ov_fifo++;
		v_lock(&fifofree_list, ipl);
		tablefull("fifonode");
		return(ENFILE);
	}
}

/*
 * pipesp()
 *	Allocate and initialize a fifonode for use in a pipe (NOT a NAMED pipe).
 */

int
pipesp(vpp)
struct vnode **vpp;
{
	struct snode *sp;
	int error;

	error = fifosp(&sp);
	/*
	 * if snode allocation failed,
	 */
	if (error != 0) 
		return(error);

	/*
	 * No need to insert into any list.  Just initialize appropriate
	 * fields.
	 */
	SN_INIT(sp, &pipe_vfs, VFIFO, PIPE_DEV);
	STOV(sp)->v_count = 2;
	sp->s_next = NULL;
	sp->s_atime = time;
	sp->s_mtime = sp->s_atime;
	sp->s_ctime = sp->s_atime;
	STOF(sp)->fn_uid = u.u_uid;
	STOF(sp)->fn_gid = u.u_gid;
	STOF(sp)->fn_rcnt = 1;
	STOF(sp)->fn_wcnt = 1;
	*vpp = STOV(sp);
	return (0);
}

/*
 * fifo_free()
 *	Put a fifonode back on the free list.
 *
 * The supplied fifonode's vnode semaphore is locked.
 */

static 
fifo_free(fn)
struct fifonode *fn;
{
	spl_t	ipl;

	VN_UNLOCKNODE(FTOV(fn));
	ipl = p_lock(&fifofree_list, SPLFS);
	fn->fn_snode.s_next = (struct snode *)fifofree;
	fifofree = fn;
	v_lock(&fifofree_list, ipl);
}

/*
 * fifo_init()
 *	Do boot time initialization for fifos.
 */

fifo_init(bufpgs)
int bufpgs;
{
	struct fifonode *fp;
	int nfifo;

	/*
	 * initialize the pipe's vfs structure
	 */
	pipe_vfs.vfs_op = &pipe_vfsops;
	pipe_vfs.vfs_bsize = PIPE_BLKSZ;

	/*
	 * sanity check number of fifo's with
	 * amount of memory set aside for
	 * buffer cache.
	 */

	nfifo = (bufpgs * CLBYTES)/PIPE_BUF/PIPE_SANITY;
	nfifo = MIN(nfifo, fifoinfo.nfifo);
	if (nfifo < fifoinfo.nfifo)  {
		cmn_err(CE_WARN, "fifos reduced from %d to %d\n",
			fifoinfo.nfifo, nfifo);
		/*
		 *+ During boot time initialization, the number
		 *+ of nodes available for pipes was reduced.
		 *+ This is to prevent pipes from consuming too
		 *+ large a percentage of the buffer cache.  This
		 *+ percentage can be modified by changing the
		 *+ configuration parameters for pipes (LFIFOSANITY, FIFO-DIV).
		 */
	}
	fifoinfo.nfifo = nfifo;
	CPRINTF("max number of fifos is %d\n", fifoinfo.nfifo);
	/*
	 * initialize free list of fifonodes
	 *	also init structures in the fifonodes.
	 */
	fifofree = NULL;
	fifostuff = (struct fifonode *)
			calloc(fifoinfo.nfifo * sizeof (struct fifonode));
	for (fp = fifostuff; fp < &fifostuff[fifoinfo.nfifo]; fp++) {
		FTOV(fp)->v_op = &fifo_vnodeops;
		FTOV(fp)->v_data = (caddr_t)fp;
		init_rwsema(&FTOV(fp)->v_nodemutex, RWSEMA_SWP);
		init_lock(&FTOV(fp)->v_mutex);
		init_lock(&fp->fn_lock);
		init_sema(&fp->fn_rwait, 0);
		init_sema(&fp->fn_rmutex, 0);
		init_sema(&fp->fn_wwait, 0);
		init_sema(&fp->fn_wmutex, 0);
		fp->fn_snode.s_next = (struct snode *)fifofree;
		fp->fn_mode = IFIFO;
		fifofree = fp;
	}
	init_lock(&fifofree_list);
}

/*
 * pipe_statfs()
 *	Get file system statistics.
 */

static int
pipe_statfs(vfsp, vp, sbp)
	struct vfs *vfsp;
	struct vnode *vp;
	struct statfs *sbp;
{
	ASSERT(vp == NULL, "statfs on regular pipe with non-null vp");
	/*
	 *+ While the kernel was doing a statfs system call 
	 *+ on a regular pipe, the vnode pointer
	 *+ argument was not null.  This indicates a probable kernel software
	 *+ bug.
	 */
	ASSERT(vfsp == &pipe_vfs, "statfs on regular pipe with wrong vfsp");
	/*
	 *+ A statfs system call was done on a regular pipe.  An
	 *+ unexpected pointer value was used inside the kernel.  This
	 *+ indicates a kernel software bug.
	 */
	bzero((caddr_t)sbp, sizeof(*sbp));
	sbp->f_bsize = PIPE_BLKSZ;
	sbp->f_frsize = PIPE_BLKSZ;
	return(0);
}
