/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ident	"$Header: ufs_flock.c 1.20 90/09/14 $"

/* $Log:	ufs_flock.c,v $


/*
 *	Notes :
 *
 *		This file contains all of the file/record locking
 *		specific routines.
 *
 *		All record lock lists (referenced by a pointer in
 *		the inode) are ordered by starting position relative
 *		to the beginning of the file.
 *
 *		In this file the name "l_end" is a macro and is used
 *		in place of "l_len" because the end, not length, of
 *		the record lock is stored internally.
 *
 */

#include "sys/types.h"
#include "sys/devsw.h"
#include "sys/param.h"
#include "sys/cmn_err.h"
#include "sys/debug.h"
#include "sys/signal.h"
#include "sys/systm.h"
#include "sys/mutex.h"
#include "sys/user.h"
#include "sys/timer.h"
#include "sys/proc.h"
#include "sys/conf.h"
#include "sys/file.h"
#include "sys/fcntl.h"
#include "sys/flock.h"
#include "sys/errno.h"
#include "sys/vnode.h"
#include "sys/ufsinode.h"
#include "sys/uio.h"

#include "sys/gate.h"
#include "sys/intctl.h"

#define WAKEUP(ptr)	vall_sema(&(ptr)->wakesem)

/* region types */
#define	S_BEFORE	010
#define	S_START		020
#define	S_MIDDLE	030
#define	S_END		040
#define	S_AFTER		050
#define	E_BEFORE	001
#define	E_START		002
#define	E_MIDDLE	003
#define	E_END		004
#define	E_AFTER		005

extern	struct	flckinfo flckinfo;	/* configuration and acct info	*/
struct	filock	*frlock;		/* pointer to record lock free list */
lock_t	frlock_lock;

struct	filock	*sleeplcks;		/* head of chain of sleeping locks */
lock_t	slplck_lock;


/* build file lock free list
 */
flckinit()
{
	register i;

	init_lock(&frlock_lock);
	init_lock(&slplck_lock);

	for (i=0; i<flckinfo.recs; i++) {
		init_sema(&flox[i].wakesem, 0);
		if (frlock == NULL) {
			flox[i].next = flox[i].prev = NULL;
			frlock = &flox[i];
		} else {
			flox[i].next = frlock;
			flox[i].prev = NULL;
			frlock = (frlock->prev = &flox[i]);
		}
	}
}

/* Insert lock (lckdat) after given lock (fl); If fl is NULL place the
 * new lock at the beginning of the list and update the head ptr to
 * list which is stored at the address given by lck_list. 
 */
static struct filock *
insflck(lck_list, lckdat, fl)
	struct	filock	**lck_list;
	struct	flock	*lckdat;
	struct	filock	*fl;
{
	register struct filock *new;

	(void)p_lock(&frlock_lock, SPLFS);

	if ((new = frlock) != NULL) {
		++flckinfo.reccnt;
		++flckinfo.rectot;
		frlock = new->next;
		if (frlock != NULL)
			frlock->prev = NULL;
		v_lock(&frlock_lock, SPLFS);
		new->set = *lckdat;
		new->set.l_pid = u.u_procp->p_pid;
		new->oftp = u.u_ofile_tab;
		if (fl == NULL) {
			new->next = *lck_list;
			if (new->next != NULL)
				new->next->prev = new;
			*lck_list = new;
		} else {
			new->next = fl->next;
			if (fl->next != NULL)
				fl->next->prev = new;
			fl->next = new;
		}
		new->prev = fl;
	} else {
		++flckinfo.recovf;
		v_lock(&frlock_lock, SPLFS);
	}

	return (new);
}

/* Delete lock (fl) from the record lock list. If fl is the first lock
 * in the list, remove it and update the head ptr to the list which is
 * stored at the address given by lck_list.
 */
static
delflck(lck_list, fl)
	struct filock  **lck_list;
	struct filock  *fl;
{
	if (fl->prev != NULL)
		fl->prev->next = fl->next;
	else
		*lck_list = fl->next;
	if (fl->next != NULL)
		fl->next->prev = fl->prev;
	WAKEUP(fl);

	(void)p_lock(&frlock_lock, SPLFS);
	--flckinfo.reccnt;
	if (frlock == NULL) {
		fl->next = fl->prev = NULL;
		frlock = fl;
	} else {
		fl->next = frlock;
		fl->prev = NULL;
		frlock = (frlock->prev = fl);
	}
	v_lock(&frlock_lock, SPLFS);
}

/* regflck sets the type of span of this (un)lock relative to the specified
 * already existing locked section.
 * There are five regions:
 *
 *  S_BEFORE        S_START         S_MIDDLE         S_END          S_AFTER
 *     010            020             030             040             050
 *  E_BEFORE        E_START         E_MIDDLE         E_END          E_AFTER
 *      01             02              03              04              05
 * 			|-------------------------------|
 *
 * relative to the already locked section.  The type is two octal digits,
 * the 8's digit is the start type and the 1's digit is the end type.
 */
static int
regflck(ld, flp)
	struct flock *ld;
	struct filock *flp;
{
	register int regntype;

	if (ld->l_start > flp->set.l_start) {
		if ((ld->l_start-1) == flp->set.l_end)
			return(S_END|E_AFTER);
		if (ld->l_start > flp->set.l_end)
			return(S_AFTER|E_AFTER);
		regntype = S_MIDDLE;
	} else if (ld->l_start == flp->set.l_start)
		regntype = S_START;
	else
		regntype = S_BEFORE;

	if (ld->l_end < flp->set.l_end) {
		if (ld->l_end == (flp->set.l_start-1))
			regntype |= E_START;
		else if (ld->l_end < flp->set.l_start)
			regntype |= E_BEFORE;
		else
			regntype |= E_MIDDLE;
	} else if (ld->l_end == flp->set.l_end)
		regntype |= E_END;
	else
		regntype |= E_AFTER;

	return (regntype);
}

/* Adjust file lock from region specified by 'ld', in the record
 * lock list indicated by the head ptr stored at the address given
 * by lck_list. Start updates at the lock given by 'insrtp'. It is 
 * assumed the list is ordered on starting position, relative to 
 * the beginning of the file, and no updating is required on any
 * locks in the list previous to the one pointed to by insrtp.
 * Insrtp is a result from the routine blocked().  Flckadj() scans
 * the list looking for locks owned by the process requesting the
 * new (un)lock :
 *
 * 	- If the new record (un)lock overlays an existing lock of
 * 	  a different type, the region overlaid is released.
 *
 * 	- If the new record (un)lock overlays or adjoins an exist-
 * 	  ing lock of the same type, the existing lock is deleted
 * 	  and its region is coalesced into the new (un)lock.
 *
 * When the list is sufficiently scanned and the new lock is not 
 * an unlock, the new lock is inserted into the appropriate
 * position in the list.
 */
static
flckadj(lck_list, insrtp, ld)
	struct filock	**lck_list;
	register struct filock *insrtp;
	struct flock	*ld;
{
	register struct	filock	*flp, *nflp;
	int regtyp;

	nflp = (insrtp == NULL) ? *lck_list : insrtp;

	while (flp = nflp) {
		nflp = flp->next;
		if (flp->oftp == u.u_ofile_tab) {

			/* release already locked region if necessary */

			switch (regtyp = regflck(ld, flp)) {
			case S_BEFORE|E_BEFORE:
				nflp = NULL;
				break;
			case S_BEFORE|E_START:
				if (ld->l_type == flp->set.l_type) {
					ld->l_end = flp->set.l_end;
					delflck(lck_list, flp);
				}
				nflp = NULL;
				break;
			case S_START|E_END:
				/* don't bother if this is in the middle of
				 * an already similarly set section.
				 */
				if (ld->l_type == flp->set.l_type)
					return(0);
			case S_START|E_AFTER:
				insrtp = flp->prev;
				delflck(lck_list, flp);
				break;
			case S_BEFORE|E_END:
				if (ld->l_type == flp->set.l_type)
					nflp = NULL;
			case S_BEFORE|E_AFTER:
				delflck(lck_list, flp);
				break;
			case S_START|E_MIDDLE:
				insrtp = flp->prev;
			case S_MIDDLE|E_MIDDLE:
				/* don't bother if this is in the middle of
				 * an already similarly set section.
				 */
				if (ld->l_type == flp->set.l_type)
					return(0);
			case S_BEFORE|E_MIDDLE:
				if (ld->l_type == flp->set.l_type)
					ld->l_end = flp->set.l_end;
				else {
					/* setup piece after end of (un)lock */
					register struct	filock *tdi, *tdp;
					struct flock td;

					td = flp->set;
					td.l_start = ld->l_end + 1;
					tdp = tdi = flp;
					do {
						if (tdp->set.l_start < td.l_start)
							tdi = tdp;
						else
							break;
					} while (tdp = tdp->next);
					if (insflck(lck_list, &td, tdi) == NULL)
						return(ENOLCK);
				}
				if (regtyp == (S_MIDDLE|E_MIDDLE)) {
					/* setup piece before (un)lock */
					flp->set.l_end = ld->l_start - 1;
					WAKEUP(flp);
					insrtp = flp;
				} else
					delflck(lck_list, flp);
				nflp = NULL;
				break;
			case S_MIDDLE|E_END:
				/* don't bother if this is in the middle of
				 * an already similarly set section.
				 */
				if (ld->l_type == flp->set.l_type)
					return(0);
				flp->set.l_end = ld->l_start - 1;
				WAKEUP(flp);
				insrtp = flp;
				break;
			case S_MIDDLE|E_AFTER:
				if (ld->l_type == flp->set.l_type) {
					ld->l_start = flp->set.l_start;
					insrtp = flp->prev;
					delflck(lck_list, flp);
				} else {
					flp->set.l_end = ld->l_start - 1;
					WAKEUP(flp);
					insrtp = flp;
				}
				break;
			case S_END|E_AFTER:
				if (ld->l_type == flp->set.l_type) {
					ld->l_start = flp->set.l_start;
					insrtp = flp->prev;
					delflck(lck_list, flp);
				}
				break;
			case S_AFTER|E_AFTER:
				insrtp = flp;
				break;
			}
		}
	}

	if (ld->l_type != F_UNLCK) {
		if (flp = insrtp) {
			do {
				if (flp->set.l_start < ld->l_start)
					insrtp = flp;
				else
					break;
			} while (flp = flp->next);
		}
		if (insflck(lck_list, ld, insrtp) == NULL)
			return(ENOLCK);
	}

	return (0);
}

/* blocked checks whether a new lock (lckdat) would be
 * blocked by a previously set lock owned by another process.
 * Insrt is set to point to the lock where lock list updating
 * should begin to place the new lock.
 */
static struct filock *
blocked(flp, lckdat, insrt)
	struct filock *flp;
	struct flock  *lckdat;
	struct filock **insrt;
{
	register struct filock *f;

	*insrt = NULL;
	for (f = flp; f != NULL; f = f->next) {
		if (f->set.l_start < lckdat->l_start)
			*insrt = f;
		else
			break;
		if (f->oftp == u.u_ofile_tab) {
			if ((lckdat->l_start-1) <= f->set.l_end)
				break;
		} else if ((lckdat->l_start <= f->set.l_end)
			    && (f->set.l_type == F_WRLCK
			        || (f->set.l_type == F_RDLCK
			            && lckdat->l_type == F_WRLCK)))
			return(f);
	}

	for (; f != NULL; f = f->next) {
		if (lckdat->l_end < f->set.l_start)
			break;
		if (lckdat->l_start <= f->set.l_end
		    && (f->oftp != u.u_ofile_tab)
		    && (f->set.l_type == F_WRLCK
		        || (f->set.l_type == F_RDLCK
		            && lckdat->l_type == F_WRLCK)))
			return(f);
	}

	return(NULL);
}

/*
 * get and set file/record locks
 *
 * cmd & SETFLCK indicates setting a lock.
 * cmd & SLPFLCK indicates waiting if there is a blocking lock.
 * cmd & VNOFLCK indicates the associated vnode is locked.
 *
 * size is the size of the file, e.g., i_size for inodes.
 *	Usually set to zero for non-regular files.
 *
 */
reclock(vp, lckdat, cmd, flag, offset, size)
	struct	vnode  *vp;
	long size;
	struct	flock *lckdat;
	int	cmd;
	int	flag;
	off_t	offset;
{
	register struct filock  **lock_list, *sf;
	struct	filock *found, *insrt = NULL;
	int retval = 0;
	int contflg = 0;
	short whence;

	/* check access permissions */
	if ((cmd & SETFLCK)
	    && ((lckdat->l_type == F_RDLCK && (flag&FREAD) == 0)
		|| (lckdat->l_type == F_WRLCK && (flag&FWRITE) == 0)))
		return (EBADF);
	
	/* Convert start to be relative to beginning of file */
	whence = lckdat->l_whence;
	if (retval = convoff(size, lckdat, 0, offset))
		return (retval);

	/* Convert l_len to be the end of the rec lock l_end */
	if (lckdat->l_len < 0)
		return (EINVAL);
	if (lckdat->l_len == 0)
		lckdat->l_end = MAXEND;
	else
		lckdat->l_end += (lckdat->l_start - 1);

	/* check for arithmetic overflow */
	if (lckdat->l_start > lckdat->l_end)
		return (EINVAL);

	(void)p_lock(&vp->v_flck_lock, SPLFS);
	lock_list = (struct filock **)&vp->v_filocks;

	do {
		contflg = 0;
		switch (lckdat->l_type) {
		case F_RDLCK:
		case F_WRLCK:
			if ((found=blocked(*lock_list, lckdat, &insrt)) == NULL) {
				if (cmd & SETFLCK)
					retval = flckadj(lock_list, insrt, lckdat);
				else
					lckdat->l_type = F_UNLCK;
			} else if (cmd & SLPFLCK) {
				(void)p_lock(&slplck_lock, SPLFS);
				/* do deadlock detection here */
				if (deadflck(found)) {
					v_lock(&slplck_lock, SPLFS);
					retval = EDEADLK;
				} else if ((sf=insflck(&sleeplcks, lckdat,
					(struct filock *)NULL)) == NULL) {
					v_lock(&slplck_lock, SPLFS);
					retval = ENOLCK;
				} else {
					sf->stat.blk.oftp = found->oftp;
					v_lock(&slplck_lock, SPLFS);
					if (cmd & VNOFLCK)
						VN_UNLOCKNODE(vp);
					if (p_sema_v_lock(&found->wakesem, PCATCH|(PZERO+1), &vp->v_flck_lock) == PCATCH)
						retval = EINTR;
					else
						contflg = 1;
					(void)p_lock(&slplck_lock, SPLFS);
					sf->stat.blk.oftp = 0;
					delflck(&sleeplcks, sf);
					v_lock(&slplck_lock, SPL0);
					if (cmd & VNOFLCK)
						VN_LOCKNODE(vp);
					(void)p_lock(&vp->v_flck_lock, SPLFS);
				}
			} else if (cmd & SETFLCK) {
				retval = EAGAIN;
			} else {
				*lckdat = found->set;
			}
			break;
		case F_UNLCK:
			/* removing a file record lock */
			if (cmd & SETFLCK)
				retval = flckadj(lock_list, *lock_list, lckdat);
			break;
		default:
			/* invalid lock type */
			retval = EINVAL;
			break;
		}
	} while (contflg);

	v_lock(&vp->v_flck_lock, SPL0);

	/* Restore l_len */
	if (lckdat->l_end == MAXEND)
		lckdat->l_len = 0;
	else
		lckdat->l_len -= (lckdat->l_start-1);
	/*
	 * POSIX 1003.1 section 6.5.2.2
	 * XPG3 fcntl() description.
	 * POSIX 1003.1 and XPG3 require that the value of l_whence, after
	 * a successfull F_GETLK request, be set to SEEK_SET.
	 *
	 * BCS does not agree.
	 */
	if ((cmd & GETFLCK) && u.u_ptype != BCS_PROC)
		whence = 0;
	(void)convoff(size, lckdat, whence, offset);

	return(retval);
}

/* convoff - converts the given data (start, whence) to the
 * given whence.
 */
static int
convoff(size, lckdat, whence, offset)
	long size;
	struct flock *lckdat;
	int whence;
	off_t offset;
{
	if (lckdat->l_whence == 1)
		lckdat->l_start += offset;
	else if (lckdat->l_whence == 2)
		lckdat->l_start += size;
	else if (lckdat->l_whence != 0)
		return (EINVAL);
	if (lckdat->l_start < 0)
		return (EINVAL);
	if (whence == 1)
		lckdat->l_start -= offset;
	else if (whence == 2)
		lckdat->l_start -= size;
	else if (whence != 0)
		return (EINVAL);
	lckdat->l_whence = whence;
	return (0);
}

/* deadflck does the deadlock detection for the given record */
static int
	deadflck(flp)
	struct filock *flp;
{
	register struct filock *blck, *sf;
	struct ofile_tab *blckoftp;

	blck = flp;	/* current blocking lock pointer */
	blckoftp = blck->oftp;
	do {
		if (blckoftp == u.u_ofile_tab)
			return(1);
		/* if the blocking process is sleeping on a locked region,
		 * change the blocked lock to this one.
		 */
		for (sf = sleeplcks; sf != NULL; sf = sf->next) {
			if (blckoftp == sf->oftp) {
				blckoftp = sf->stat.blk.oftp;
				break;
			}
		}
		blck = sf;
	} while (blck != NULL);
	return(0);
}

/* Clean up record locks left around by process.		*/
/* if flag == USE_PID, match pid with pid in filock structure.	*/
/* if flag == IGN_PID, clean regardless of pid.		*/
cleanlocks(vp, flag)
	struct vnode *vp;
	int	flag;
{
	register struct filock *flp, *nflp, **lock_list;

	(void)p_lock(&vp->v_flck_lock, SPLFS);

	lock_list = (struct filock **)&vp->v_filocks;
	for (flp = *lock_list; flp!=NULL; flp=nflp) {
		nflp = flp->next;
		if ((flp->oftp == u.u_ofile_tab) || (flag == IGN_PID))
			delflck(lock_list, flp);
	};

	v_lock(&vp->v_flck_lock, SPL0);
}

/*
 * Enforce record locking protocol on a file, vp.
 */
chklock(vp, uio, rw, ioflag, size)
	register struct vnode *vp;
	long size;
	struct uio *uio;
	enum uio_rw rw;
	int ioflag;
{
	register int i;
	struct flock bf;

	bf.l_type = (rw == UIO_WRITE) ? F_WRLCK : F_RDLCK;
	bf.l_whence = 0;
	bf.l_start = uio->uio_offset;
	bf.l_len = uio->uio_resid;
	i = (ioflag & IO_NDELAY) ? VNOFLCK : SLPFLCK|VNOFLCK;
	if ((i = reclock(vp, &bf, i, 0, 0, size)) || bf.l_type != F_UNLCK)
		return(i ? i : EAGAIN);
	else
		return(0);
}

/*
 * Perform file locking fcntl operations on a vnode
 *	This code does copyin/copyout.  This is strictly not a good
 *	idea for something called from a VOP...  But, let it go for now.
 *
 *	Also, the flags passed in here are from the file descriptor.
 *	This is ALSO not right.  But for now.....
 *
 * "ismandatory": (non-zero) specifies mandatory file locking.
 *		  This causes this routine to lock the vnode before
 *		  doing some operations in order to synchronize with
 *		  file access operations which may check the file lock state.
 *		  It also currently causes requests for VN_MAPPED vnodes to
 *		  be refused (until this case is handled better).
 */
lck_fcntl(vp, cmd, arg, flag, offset, size, ismandatory)
	struct vnode *vp;
	int cmd, arg, flag;
	off_t offset;
	long size;
	int ismandatory;
{
	struct flock bf;
	int i, error;
	int vnolock = 0;

	error = 0;

	switch (cmd) {
	case F_GETLK:
		/*
		 * get record lock,
		 *	this case will never block.
		 *	Nor will it modify locking state of the file.
		 *	So, there's no need to assert vnode lock.
		 */
		if (copyin((caddr_t)arg, &bf, sizeof bf))
			error = EFAULT;
		else if ((i = reclock(vp, &bf, GETFLCK, flag, offset, size)) != 0)
			error = i;
		else if (copyout(&bf, (caddr_t)arg, sizeof bf))
			error = EFAULT;
		break;

	case F_SETLK:
		/*
		 * set record lock
		 *	if mandatory locking, we must lock the vnode.
		 */
		if (copyin((caddr_t)arg, &bf, sizeof bf)) {
			error = EFAULT;
			break;
		}
		if (ismandatory) {
			ASSERT(vp->v_flag & VMNDLCK,
				"F_SETLK entry mandatory locking");
			/*
			 *+ An internal inconsistency has been discovered
			 *+ between the vnode flags designating mandatory
			 *+ locking and the mode bits of the inode.
			 */

			/*
			 * don't allow mandatory locking on a mapped
			 * file.  (Suppose someone did a chmod on
			 * a file which had already been mapped).
			 */
			if (VN_MAPPED(vp)) {
				error = EINVAL;
				break;
			}

			VN_LOCKNODE(vp);
			vnolock = VNOFLCK;
		}
		if ((i = reclock(vp, &bf, vnolock | SETFLCK, flag,
						offset, size)) != 0) {
			/* The following if statement is to maintain
			** backward compatibility. Note: the return
			** value of reclock was not changed to
			** EAGAIN because rdwr() also calls reclock(),
			** and we want rdwr() to meet the standards.
			*/
			if ( i == EAGAIN )
				error = EACCES;
			else
				error = i;
		}
		if (vnolock) {
			ASSERT(vp->v_flag & VMNDLCK,
					"F_SETLK exit mandatory locking");
			/*
			 *+ An internal inconsistency has been discovered
			 *+ between the vnode flags designating mandatory
			 *+ locking and the mode bits of the vnode.
			 */
			VN_UNLOCKNODE(vp);
		}
		break;

	case F_SETLKW:
		/* set record lock and wait */
		if (copyin((caddr_t)arg, &bf, sizeof bf)) {
			error = EFAULT;
			break;
		}
		if (ismandatory) {
			ASSERT(vp->v_flag & VMNDLCK,
				"F_SETLKW entry mandatory locking");
			/*
			 *+ An internal inconsistency has been discovered
			 *+ between the vnode flags designating mandatory
			 *+ locking and the mode bits of the vnode.
			 */

			/*
			 * don't allow mandatory locking on a mapped
			 * file.  (Suppose someone did a chmod on
			 * a file which had already been mapped).
			 */
			if (VN_MAPPED(vp)) {
				error = EINVAL;
				break;
			}

			VN_LOCKNODE(vp);
			vnolock = VNOFLCK;
		}
		error = reclock(vp, &bf, vnolock | SETFLCK | SLPFLCK,
				flag, offset, size);
		if (vnolock) {
			ASSERT(vp->v_flag & VMNDLCK,
				"F_SETLKW exit mandatory locking");
			/*
			 *+ An internal inconsistency has been discovered
			 *+ between the vnode flags designating mandatory
			 *+ locking and the mode bits of the vnode.
			 */
			VN_UNLOCKNODE(vp);
		}
		break;

	case F_CHKFL:
		break;

	case F_FREESP:
		/*
		 * don't know if this will ever be supported or not.
		 *	system V says "experimental, may change".
		 */
	default:
		error = EINVAL;
	}
	return(error);
}
